use super::*;
use itertools::Itertools;
use v_htmlescape::escape;

/// Returns whether the given node name is valid according to given seeds.
///
/// # Arguments
/// * `node_name`: &str - The name of the nodes.
/// * `base_names`: Option<&[&str]> - The expected base of the node name, how the expected KG nodes may start.
/// * `base_length`: Option<usize> - The expected length of the node name when it includes the base name.
/// * `separator`: Option<&str> - The expected separator that must be included when the node name starts with the base name.
/// * `id_acronym`: Option<&str> - The expected ID the node name must start with AFTER the separator.
/// * `id_length`: Option<usize> - The expected length of the node name after the separator.
/// * `numeric_part_length`: Option<usize> - The expected length of the numeric part at the end of the node name.
pub(crate) fn is_valid_node_name_from_seeds(
    node_name: &str,
    base_names: Option<&[&str]>,
    base_length: Option<usize>,
    separator: Option<&str>,
    id_acronym: Option<&str>,
    id_length: Option<usize>,
    numeric_part_length: Option<usize>,
) -> Result<()> {
    if (base_names.is_none()
        || base_names
            .as_ref()
            .map_or(false, |base_names| base_names.is_empty()))
        && id_acronym.is_none()
    {
        panic!(concat!(
            "There is some error: both base name and ID acronym are None.",
            "At least one of the two must be not None."
        ));
    }
    let mut validated = false;
    let mut node_name = node_name.to_string();

    // First of all, we check if the node name starts with the base name
    // and the base name was provided.
    if let Some(base_names) = base_names {
        for base_name in base_names.iter() {
            if node_name
                .to_uppercase()
                .starts_with(&base_name.to_string().to_uppercase())
            {
                // If so, it must have the provided base length.
                if base_length
                    .as_ref()
                    .map_or(false, |len| node_name.len() != *len)
                {
                    return Err(format!(
                        "The given node name {} does not respect the given base length {:?}.",
                        node_name, base_length
                    ));
                }
                // We remove the base name from the given node name
                node_name = node_name[base_name.len()..node_name.len()].to_string();
                // The provided separator must exist in the given node name
                if let Some(separator) = separator {
                    // If the separator was provided, it must appear exactly once
                    if node_name.matches(separator).count() >= 1 {
                        node_name = node_name.split(separator).skip(1).join("");
                    } else {
                        return Err(format!(
                            "We expect for the given separator {} to appear exactly once in the given object '{}'.",
                            separator,
                            node_name
                        ));
                    }
                }

                // Finally, we check that the final `numeric_part_length` values must
                // be numeric digits, if this value was provided.
                if let Some(numeric_part_length) = numeric_part_length {
                    if numeric_part_length > node_name.len() {
                        return Err(format!(
                            concat!(
                                "The expected numeric part length is {}, but the node ",
                                "name provided `{}` has length {}."
                            ),
                            numeric_part_length,
                            node_name,
                            node_name.len(),
                        ));
                    }
                    if node_name[(node_name.len() - numeric_part_length)..node_name.len()]
                        .chars()
                        .any(|character| !character.is_ascii_digit())
                    {
                        return Err(format!(
                            "The last {} characters of the node name {} are expected to be digits.",
                            numeric_part_length, node_name,
                        ));
                    }
                }

                validated = true;
                break;
            }
        }
    }

    // If an ID acronym was provided, we expect the node to respect it.
    if let Some(id_acronym) = id_acronym {
        if !node_name
            .to_uppercase()
            .starts_with(&id_acronym.to_uppercase())
        {
            return Err(format!(
                "The given node name {} does not start with the expected ID acronym {}.",
                node_name, id_acronym
            ));
        }
        // If it starts with the provided acronym, it must have the provided
        // acronym length.
        if id_length
            .as_ref()
            .map_or(false, |len| node_name.len() != *len)
        {
            return Err(format!(
                "The given node name {} starts with the given acronym {} but does not have the expected length {:?}.",
                node_name,
                id_acronym,
                id_length
            ));
        }
        // Finally, we check that the final `numeric_part_length` values must
        // be numeric digits, if this value was provided.
        if let Some(numeric_part_length) = numeric_part_length {
            if node_name[(node_name.len() - numeric_part_length)..node_name.len()]
                .chars()
                .any(|character| !character.is_ascii_digit())
            {
                return Err(format!(
                    "The last {} characters of the node name {} are expected to be digits.",
                    numeric_part_length, node_name,
                ));
            }
        }
        validated = true;
    }

    if validated {
        Ok(())
    } else {
        Err(format!(
            "The given node name {node_name} was not validated successfully.",
            node_name = node_name
        ))
    }
}

/// Returns formatted url.
///
/// # Arguments
/// * `url_pattern`: &str - The URL pattern.
/// * `node_name`: &str - The node name.
/// * `separator`: Option<&str> - The expected separator.
pub(crate) fn format_url_from_node_name(
    url_pattern: &str,
    node_name: &str,
    separator: Option<&str>,
) -> String {
    url_pattern.replace(
        "{node_name}",
        separator
            .map_or(node_name.to_string(), |sep| {
                if node_name.contains(sep) {
                    node_name.split(sep).skip(1).join("")
                } else {
                    node_name.to_string()
                }
            })
            .as_str(),
    )
}

/// Returns built url for given element
fn get_url_formatted(url: &str, content: &str, repository: Option<&str>) -> String {
    format!(
        "<a href='{url}' target='_blank' title='Go to {repository} to get more informations about {content}'>{content}</a>",
        url = url,
        content = escape(content).to_string(),
        repository=repository.unwrap_or("website")
    )
}

/// Returns url describing the given node name if a pattern is known.
///
/// # Implementative details
/// Currently we have support for building the URLs for:
/// * [Sequence Ontology](http://www.sequenceontology.org/)
/// * [FlyBase](http://flybase.org/)
/// * [Mouse Genome Informations](http://www.informatics.jax.org/)
/// * [Pubmed NCBI](https://www.ncbi.nlm.nih.gov/)
/// * [NCBI Gene](https://www.ncbi.nlm.nih.gov/gene/)
/// * [NCBI ClinVar](https://www.ncbi.nlm.nih.gov/clinvar/)
/// * [WormBase](https://wormbase.org)
/// * [Ensembl](https://www.ensembl.org/index.html)
/// * [ZFIN](http://zfin.org/)
/// * [DOI](https://www.doi.org/)
/// * [CHEBI](https://www.ebi.ac.uk/chebi/init.do)
/// * [BioGrid](https://thebiogrid.org/)
/// * [OMIM](https://www.omim.org/)
/// * [Rat Genome DataBase](https://rgd.mcw.edu/rgdweb/homepage/)
/// * [MMRRC](https://www.mmrrc.org/)
/// * [GO](http://amigo.geneontology.org/amigo/landing)
/// * [UniProtKB](https://www.uniprot.org/)
/// * [Coriell](https://coriell.org/)
/// * [DrugCentral](https://drugcentral.org/)
/// * [NCBI MESH](https://www.ncbi.nlm.nih.gov/mesh/)
/// * [NCBI Taxonomy](https://www.ncbi.nlm.nih.gov/taxonomy)
/// * [JAX](https://www.jax.org/strain)
/// * [WikiData](https://www.wikidata.org/wiki/Wikidata:Main_Page)
/// * [Therapeutic Target Database](http://db.idrblab.net/ttd/)
/// * [Reactome](https://reactome.org/)
///
/// TODO! Update documentation.
/// TODO! Add support for URLs.
///
/// # Arguments
/// * `node_name`: &str - Node name to query for.
///
/// # Raises
/// * If there is no known url source for the given node name.
pub fn get_node_source_url_from_node_name(node_name: &str) -> Result<String> {
    if is_valid_flybase_node_name(node_name) {
        return Ok(unsafe { format_flybase_url_from_node_name(node_name) });
    }
    if is_valid_sequence_ontology_node_name(node_name) {
        return Ok(unsafe { format_sequence_ontology_url_from_node_name(node_name) });
    }
    if is_valid_mmrrc_node_name(node_name) {
        return Ok(unsafe { format_mmrrc_url_from_node_name(node_name) });
    }
    if is_valid_wormbase_gene_node_name(node_name) {
        return Ok(unsafe { format_wormbase_gene_url_from_node_name(node_name) });
    }
    if is_valid_mouse_genome_informatics_node_name(node_name) {
        return Ok(unsafe { format_mouse_genome_informatics_url_from_node_name(node_name) });
    }
    if is_valid_ncbi_dbgap_node_name(node_name) {
        return Ok(unsafe { format_ncbi_dbgap_url_from_node_name(node_name) });
    }
    if is_valid_pubmed_ncbi_node_name(node_name) {
        return Ok(unsafe { format_pubmed_ncbi_url_from_node_name(node_name) });
    }
    if is_valid_ncbi_gene_node_name(node_name) {
        return Ok(unsafe { format_ncbi_gene_url_from_node_name(node_name) });
    }
    if is_valid_ncbi_clinvar_node_name(node_name) {
        return Ok(unsafe { format_ncbi_clinvar_url_from_node_name(node_name) });
    }
    if is_valid_ensembl_node_name(node_name) {
        return Ok(unsafe { format_ensembl_url_from_node_name(node_name) });
    }
    if is_valid_zfin_node_name(node_name) {
        return Ok(unsafe { format_zfin_url_from_node_name(node_name) });
    }
    if is_valid_doi_node_name(node_name) {
        return Ok(unsafe { format_doi_url_from_node_name(node_name) });
    }
    if is_valid_chebi_node_name(node_name) {
        return Ok(unsafe { format_chebi_url_from_node_name(node_name) });
    }
    if is_valid_biogrid_node_name(node_name) {
        return Ok(unsafe { format_biogrid_url_from_node_name(node_name) });
    }
    if is_valid_omim_node_name(node_name) {
        return Ok(unsafe { format_omim_url_from_node_name(node_name) });
    }
    if is_valid_rat_genome_database_node_name(node_name) {
        return Ok(unsafe { format_rat_genome_database_url_from_node_name(node_name) });
    }
    if is_valid_gene_ontology_node_name(node_name) {
        return Ok(unsafe { format_gene_ontology_url_from_node_name(node_name) });
    }
    if is_valid_uniprotkb_node_name(node_name) {
        return Ok(unsafe { format_uniprotkb_url_from_node_name(node_name) });
    }
    if is_valid_coriell_node_name(node_name) {
        return Ok(unsafe { format_coriell_url_from_node_name(node_name) });
    }
    if is_valid_drugcentral_node_name(node_name) {
        return Ok(unsafe { format_drugcentral_url_from_node_name(node_name) });
    }
    if is_valid_ncbi_mesh_node_name(node_name) {
        return Ok(unsafe { format_ncbi_mesh_url_from_node_name(node_name) });
    }
    if is_valid_ncbi_taxonomy_node_name(node_name) {
        return Ok(unsafe { format_ncbi_taxonomy_url_from_node_name(node_name) });
    }
    if is_valid_jax_node_name(node_name) {
        return Ok(unsafe { format_jax_url_from_node_name(node_name) });
    }
    if is_valid_therapeutic_target_database_node_name(node_name) {
        return Ok(unsafe { format_therapeutic_target_database_url_from_node_name(node_name) });
    }
    if is_valid_wikidata_node_name(node_name) {
        return Ok(unsafe { format_wikidata_url_from_node_name(node_name) });
    }
    if is_valid_reactome_node_name(node_name) {
        return Ok(unsafe { format_reactome_url_from_node_name(node_name) });
    }

    if is_valid_jackson_lab_reference_node_name(node_name) {
        return Ok(unsafe { format_jackson_lab_reference_url_from_node_name(node_name) });
    }

    if is_valid_uberon_core_node_name(node_name) {
        return Ok(unsafe { format_uberon_core_url_from_node_name(node_name) });
    }

    if is_valid_gene_ontology_relations_node_name(node_name) {
        return Ok(unsafe { format_gene_ontology_relations_url_from_node_name(node_name) });
    }

    if is_valid_quantity_unit_dimension_and_type_node_name(node_name) {
        return Ok(unsafe {
            format_quantity_unit_dimension_and_type_url_from_node_name(node_name)
        });
    }

    if is_valid_open_biomedical_association_node_name(node_name) {
        return Ok(unsafe { format_open_biomedical_association_url_from_node_name(node_name) });
    }

    if is_valid_kegg_homo_sapiens_node_name(node_name) {
        return Ok(unsafe { format_kegg_homo_sapiens_url_from_node_name(node_name) });
    }

    if is_valid_dublin_core_node_name(node_name) {
        return Ok(unsafe { format_dublin_core_url_from_node_name(node_name) });
    }

    if is_valid_xsd_node_name(node_name) {
        return Ok(unsafe { format_xsd_url_from_node_name(node_name) });
    }

    if is_valid_kegg_ko_node_name(node_name) {
        return Ok(unsafe { format_kegg_ko_url_from_node_name(node_name) });
    }

    if is_valid_swiss_protein_node_name(node_name) {
        return Ok(unsafe { format_swiss_protein_url_from_node_name(node_name) });
    }

    if is_valid_ncbi_books_node_name(node_name) {
        return Ok(unsafe { format_ncbi_books_url_from_node_name(node_name) });
    }

    if is_valid_chembl_compound_node_name(node_name) {
        return Ok(unsafe { format_chembl_compound_url_from_node_name(node_name) });
    }

    if is_valid_hugo_gene_name_consortium_node_name(node_name) {
        return Ok(unsafe { format_hugo_gene_name_consortium_url_from_node_name(node_name) });
    }

    if is_valid_provenance_authoring_and_versioning_node_name(node_name) {
        return Ok(unsafe {
            format_provenance_authoring_and_versioning_url_from_node_name(node_name)
        });
    }

    if is_valid_feature_annotation_location_description_ontology_node_name(node_name) {
        return Ok(unsafe {
            format_feature_annotation_location_description_ontology_url_from_node_name(node_name)
        });
    }

    if is_valid_rbrc_node_name(node_name) {
        return Ok(unsafe { format_rbrc_url_from_node_name(node_name) });
    }

    if is_valid_dictybase_gene_node_name(node_name) {
        return Ok(unsafe { format_dictybase_gene_url_from_node_name(node_name) });
    }

    if is_valid_monarch_initiative_node_name(node_name) {
        return Ok(unsafe { format_monarch_initiative_url_from_node_name(node_name) });
    }

    if is_valid_edam_data_node_name(node_name) {
        return Ok(unsafe { format_edam_data_url_from_node_name(node_name) });
    }

    if is_valid_ncimr_node_name(node_name) {
        return Ok(unsafe { format_ncimr_url_from_node_name(node_name) });
    }

    if is_valid_ecogene_node_name(node_name) {
        return Ok(unsafe { format_ecogene_url_from_node_name(node_name) });
    }

    if is_valid_aspergillus_genomic_data_reference_node_name(node_name) {
        return Ok(unsafe {
            format_aspergillus_genomic_data_reference_url_from_node_name(node_name)
        });
    }

    if is_valid_european_mouse_mutant_archive_node_name(node_name) {
        return Ok(unsafe { format_european_mouse_mutant_archive_url_from_node_name(node_name) });
    }

    if is_valid_translation_of_embl_nucleotide_sequence_database_node_name(node_name) {
        return Ok(unsafe {
            format_translation_of_embl_nucleotide_sequence_database_url_from_node_name(node_name)
        });
    }

    if is_valid_mouse_phenome_database_strain_node_name(node_name) {
        return Ok(unsafe { format_mouse_phenome_database_strain_url_from_node_name(node_name) });
    }

    if is_valid_basic_formal_ontology_node_name(node_name) {
        return Ok(unsafe { format_basic_formal_ontology_url_from_node_name(node_name) });
    }

    if is_valid_dictybase_genes_node_name(node_name) {
        return Ok(unsafe { format_dictybase_genes_url_from_node_name(node_name) });
    }

    if is_valid_cell_line_ontology_node_name(node_name) {
        return Ok(unsafe { format_cell_line_ontology_url_from_node_name(node_name) });
    }

    if is_valid_animal_genome_horse_qtl_node_name(node_name) {
        return Ok(unsafe { format_animal_genome_horse_qtl_url_from_node_name(node_name) });
    }

    if is_valid_wormbase_node_name(node_name) {
        return Ok(unsafe { format_wormbase_url_from_node_name(node_name) });
    }

    if is_valid_animal_qtl_traits_node_name(node_name) {
        return Ok(unsafe { format_animal_qtl_traits_url_from_node_name(node_name) });
    }

    if is_valid_friend_of_a_friend_node_name(node_name) {
        return Ok(unsafe { format_friend_of_a_friend_url_from_node_name(node_name) });
    }

    if is_valid_omim_phenotypic_series_node_name(node_name) {
        return Ok(unsafe { format_omim_phenotypic_series_url_from_node_name(node_name) });
    }

    if is_valid_owl_node_name(node_name) {
        return Ok(unsafe { format_owl_url_from_node_name(node_name) });
    }

    if is_valid_arabidopsis_tair_node_name(node_name) {
        return Ok(unsafe { format_arabidopsis_tair_url_from_node_name(node_name) });
    }

    if is_valid_unified_medical_language_system_node_name(node_name) {
        return Ok(unsafe { format_unified_medical_language_system_url_from_node_name(node_name) });
    }

    if is_valid_gene_ontology_reference_node_name(node_name) {
        return Ok(unsafe { format_gene_ontology_reference_url_from_node_name(node_name) });
    }

    if is_valid_ncbi_clinical_variants_submitters_node_name(node_name) {
        return Ok(unsafe {
            format_ncbi_clinical_variants_submitters_url_from_node_name(node_name)
        });
    }

    if is_valid_gdc_project_node_name(node_name) {
        return Ok(unsafe { format_gdc_project_url_from_node_name(node_name) });
    }

    if is_valid_complex_portal_node_name(node_name) {
        return Ok(unsafe { format_complex_portal_url_from_node_name(node_name) });
    }

    if is_valid_ucsc_golden_path_node_name(node_name) {
        return Ok(unsafe { format_ucsc_golden_path_url_from_node_name(node_name) });
    }

    if is_valid_yeast_genome_reference_node_name(node_name) {
        return Ok(unsafe { format_yeast_genome_reference_url_from_node_name(node_name) });
    }

    if is_valid_xenbase_node_name(node_name) {
        return Ok(unsafe { format_xenbase_url_from_node_name(node_name) });
    }

    if is_valid_chembl_target_node_name(node_name) {
        return Ok(unsafe { format_chembl_target_url_from_node_name(node_name) });
    }

    if is_valid_coriell_collection_node_name(node_name) {
        return Ok(unsafe { format_coriell_collection_url_from_node_name(node_name) });
    }

    if is_valid_gdc_annotation_node_name(node_name) {
        return Ok(unsafe { format_gdc_annotation_url_from_node_name(node_name) });
    }

    if is_valid_animal_genome_chicken_qtl_node_name(node_name) {
        return Ok(unsafe { format_animal_genome_chicken_qtl_url_from_node_name(node_name) });
    }

    if is_valid_semantic_medline_database_node_name(node_name) {
        return Ok(unsafe { format_semantic_medline_database_url_from_node_name(node_name) });
    }

    if is_valid_animal_genome_rainbow_trout_qtl_node_name(node_name) {
        return Ok(unsafe { format_animal_genome_rainbow_trout_qtl_url_from_node_name(node_name) });
    }

    if is_valid_animal_genome_pig_qtl_node_name(node_name) {
        return Ok(unsafe { format_animal_genome_pig_qtl_url_from_node_name(node_name) });
    }

    if is_valid_livestock_product_trait_node_name(node_name) {
        return Ok(unsafe { format_livestock_product_trait_url_from_node_name(node_name) });
    }

    if is_valid_variation_representation_specification_node_name(node_name) {
        return Ok(unsafe {
            format_variation_representation_specification_url_from_node_name(node_name)
        });
    }

    if is_valid_semantic_science_node_name(node_name) {
        return Ok(unsafe { format_semantic_science_url_from_node_name(node_name) });
    }

    if is_valid_experimental_factor_ontology_node_name(node_name) {
        return Ok(unsafe { format_experimental_factor_ontology_url_from_node_name(node_name) });
    }

    if is_valid_animal_genome_sheep_qtl_node_name(node_name) {
        return Ok(unsafe { format_animal_genome_sheep_qtl_url_from_node_name(node_name) });
    }

    if is_valid_open_biomedical_ontology_node_name(node_name) {
        return Ok(unsafe { format_open_biomedical_ontology_url_from_node_name(node_name) });
    }

    if is_valid_intact_node_name(node_name) {
        return Ok(unsafe { format_intact_url_from_node_name(node_name) });
    }

    if is_valid_game_ontology_project_node_name(node_name) {
        return Ok(unsafe { format_game_ontology_project_url_from_node_name(node_name) });
    }

    if is_valid_apergillus_genome_data_node_name(node_name) {
        return Ok(unsafe { format_apergillus_genome_data_url_from_node_name(node_name) });
    }

    if is_valid_animal_genome_pubblications_node_name(node_name) {
        return Ok(unsafe { format_animal_genome_pubblications_url_from_node_name(node_name) });
    }

    if is_valid_snp_individual_node_name(node_name) {
        return Ok(unsafe { format_snp_individual_url_from_node_name(node_name) });
    }

    if is_valid_ncbi_protein_node_name(node_name) {
        return Ok(unsafe { format_ncbi_protein_url_from_node_name(node_name) });
    }

    if is_valid_orpha_node_name(node_name) {
        return Ok(unsafe { format_orpha_url_from_node_name(node_name) });
    }

    if is_valid_wormbase_vocabulary_node_name(node_name) {
        return Ok(unsafe { format_wormbase_vocabulary_url_from_node_name(node_name) });
    }

    if is_valid_kegg_path_node_name(node_name) {
        return Ok(unsafe { format_kegg_path_url_from_node_name(node_name) });
    }

    if is_valid_vertebrate_gene_names_consortium_node_name(node_name) {
        return Ok(unsafe {
            format_vertebrate_gene_names_consortium_url_from_node_name(node_name)
        });
    }

    if is_valid_yeast_genome_locus_node_name(node_name) {
        return Ok(unsafe { format_yeast_genome_locus_url_from_node_name(node_name) });
    }

    if is_valid_schema_node_name(node_name) {
        return Ok(unsafe { format_schema_url_from_node_name(node_name) });
    }

    if is_valid_provenance_ontology_node_name(node_name) {
        return Ok(unsafe { format_provenance_ontology_url_from_node_name(node_name) });
    }

    if is_valid_kegg_disease_node_name(node_name) {
        return Ok(unsafe { format_kegg_disease_url_from_node_name(node_name) });
    }

    if is_valid_database_snp_node_name(node_name) {
        return Ok(unsafe { format_database_snp_url_from_node_name(node_name) });
    }

    if is_valid_cell_ontology_node_name(node_name) {
        return Ok(unsafe { format_cell_ontology_url_from_node_name(node_name) });
    }

    if is_valid_mugen_node_name(node_name) {
        return Ok(unsafe { format_mugen_url_from_node_name(node_name) });
    }

    if is_valid_world_geodetic_system_node_name(node_name) {
        return Ok(unsafe { format_world_geodetic_system_url_from_node_name(node_name) });
    }

    if is_valid_chromosome_ontology_node_name(node_name) {
        return Ok(unsafe { format_chromosome_ontology_url_from_node_name(node_name) });
    }

    if is_valid_pharmgkb_node_name(node_name) {
        return Ok(unsafe { format_pharmgkb_url_from_node_name(node_name) });
    }

    if is_valid_the_arabidopsis_information_resource_locus_node_name(node_name) {
        return Ok(unsafe {
            format_the_arabidopsis_information_resource_locus_url_from_node_name(node_name)
        });
    }

    if is_valid_void_node_name(node_name) {
        return Ok(unsafe { format_void_url_from_node_name(node_name) });
    }

    if is_valid_rdf_node_name(node_name) {
        return Ok(unsafe { format_rdf_url_from_node_name(node_name) });
    }

    if is_valid_rgd_reference_node_name(node_name) {
        return Ok(unsafe { format_rgd_reference_url_from_node_name(node_name) });
    }

    if is_valid_mouse_phenome_database_assay_node_name(node_name) {
        return Ok(unsafe { format_mouse_phenome_database_assay_url_from_node_name(node_name) });
    }

    if is_valid_drugbank_node_name(node_name) {
        return Ok(unsafe { format_drugbank_url_from_node_name(node_name) });
    }

    if is_valid_coriell_family_node_name(node_name) {
        return Ok(unsafe { format_coriell_family_url_from_node_name(node_name) });
    }

    if is_valid_pombase_node_name(node_name) {
        return Ok(unsafe { format_pombase_url_from_node_name(node_name) });
    }

    if is_valid_online_mendelian_inheritance_in_animals_node_name(node_name) {
        return Ok(unsafe {
            format_online_mendelian_inheritance_in_animals_url_from_node_name(node_name)
        });
    }

    if is_valid_interpro_node_name(node_name) {
        return Ok(unsafe { format_interpro_url_from_node_name(node_name) });
    }

    if is_valid_panther_database_node_name(node_name) {
        return Ok(unsafe { format_panther_database_url_from_node_name(node_name) });
    }

    if is_valid_human_developmental_stages_node_name(node_name) {
        return Ok(unsafe { format_human_developmental_stages_url_from_node_name(node_name) });
    }

    if is_valid_mouse_pathology_ontology_node_name(node_name) {
        return Ok(unsafe { format_mouse_pathology_ontology_url_from_node_name(node_name) });
    }

    if is_valid_unified_phenotype_ontology_upheno_node_name(node_name) {
        return Ok(unsafe {
            format_unified_phenotype_ontology_upheno_url_from_node_name(node_name)
        });
    }

    if is_valid_human_phenotype_ontology_node_name(node_name) {
        return Ok(unsafe { format_human_phenotype_ontology_url_from_node_name(node_name) });
    }

    if is_valid_protein_ontology_pro_node_name(node_name) {
        return Ok(unsafe { format_protein_ontology_pro_url_from_node_name(node_name) });
    }

    if is_valid_human_disease_ontology_node_name(node_name) {
        return Ok(unsafe { format_human_disease_ontology_url_from_node_name(node_name) });
    }

    if is_valid_zebrafish_anatomy_and_development_ontology_node_name(node_name) {
        return Ok(unsafe {
            format_zebrafish_anatomy_and_development_ontology_url_from_node_name(node_name)
        });
    }

    if is_valid_an_ontology_of_core_ecological_entities_node_name(node_name) {
        return Ok(unsafe {
            format_an_ontology_of_core_ecological_entities_url_from_node_name(node_name)
        });
    }

    if is_valid_pathway_ontology_node_name(node_name) {
        return Ok(unsafe { format_pathway_ontology_url_from_node_name(node_name) });
    }

    if is_valid_ascomycete_phenotype_ontology_node_name(node_name) {
        return Ok(unsafe { format_ascomycete_phenotype_ontology_url_from_node_name(node_name) });
    }

    if is_valid_environment_ontology_node_name(node_name) {
        return Ok(unsafe { format_environment_ontology_url_from_node_name(node_name) });
    }

    if is_valid_nci_thesaurus_obo_edition_node_name(node_name) {
        return Ok(unsafe { format_nci_thesaurus_obo_edition_url_from_node_name(node_name) });
    }

    if is_valid_mouse_adult_gross_anatomy_node_name(node_name) {
        return Ok(unsafe { format_mouse_adult_gross_anatomy_url_from_node_name(node_name) });
    }

    if is_valid_neuro_behavior_ontology_node_name(node_name) {
        return Ok(unsafe { format_neuro_behavior_ontology_url_from_node_name(node_name) });
    }

    if is_valid_chemical_methods_ontology_node_name(node_name) {
        return Ok(unsafe { format_chemical_methods_ontology_url_from_node_name(node_name) });
    }

    if is_valid_zebrafish_developmental_stages_ontology_node_name(node_name) {
        return Ok(unsafe {
            format_zebrafish_developmental_stages_ontology_url_from_node_name(node_name)
        });
    }

    if is_valid_fungal_gross_anatomy_node_name(node_name) {
        return Ok(unsafe { format_fungal_gross_anatomy_url_from_node_name(node_name) });
    }

    if is_valid_zebrafish_phenotype_ontology_node_name(node_name) {
        return Ok(unsafe { format_zebrafish_phenotype_ontology_url_from_node_name(node_name) });
    }

    if is_valid_ontology_of_biological_attributes_node_name(node_name) {
        return Ok(unsafe {
            format_ontology_of_biological_attributes_url_from_node_name(node_name)
        });
    }

    if is_valid_confidence_information_ontology_node_name(node_name) {
        return Ok(unsafe { format_confidence_information_ontology_url_from_node_name(node_name) });
    }

    if is_valid_ncbi_gene_expression_omnibus_node_name(node_name) {
        return Ok(unsafe { format_ncbi_gene_expression_omnibus_url_from_node_name(node_name) });
    }

    if is_valid_genotype_ontology_node_name(node_name) {
        return Ok(unsafe { format_genotype_ontology_url_from_node_name(node_name) });
    }

    if is_valid_ontology_of_adverse_events_node_name(node_name) {
        return Ok(unsafe { format_ontology_of_adverse_events_url_from_node_name(node_name) });
    }

    if is_valid_experimental_condition_ontology_node_name(node_name) {
        return Ok(unsafe { format_experimental_condition_ontology_url_from_node_name(node_name) });
    }

    if is_valid_plant_ontology_node_name(node_name) {
        return Ok(unsafe { format_plant_ontology_url_from_node_name(node_name) });
    }

    if is_valid_environmental_conditions_treatments_and_exposures_ontology_node_name(node_name) {
        return Ok(unsafe {
            format_environmental_conditions_treatments_and_exposures_ontology_url_from_node_name(
                node_name,
            )
        });
    }

    if is_valid_units_of_measurement_ontology_node_name(node_name) {
        return Ok(unsafe { format_units_of_measurement_ontology_url_from_node_name(node_name) });
    }

    if is_valid_ontology_for_biomedical_investigations_node_name(node_name) {
        return Ok(unsafe {
            format_ontology_for_biomedical_investigations_url_from_node_name(node_name)
        });
    }

    if is_valid_c_elegans_phenotype_node_name(node_name) {
        return Ok(unsafe { format_c_elegans_phenotype_url_from_node_name(node_name) });
    }

    if is_valid_biological_spatial_ontology_node_name(node_name) {
        return Ok(unsafe { format_biological_spatial_ontology_url_from_node_name(node_name) });
    }

    if is_valid_scientific_evidence_and_provenance_information_ontology_node_name(node_name) {
        return Ok(unsafe {
            format_scientific_evidence_and_provenance_information_ontology_url_from_node_name(
                node_name,
            )
        });
    }

    if is_valid_drosophila_gross_anatomy_node_name(node_name) {
        return Ok(unsafe { format_drosophila_gross_anatomy_url_from_node_name(node_name) });
    }

    if is_valid_the_statistical_methods_ontology_node_name(node_name) {
        return Ok(unsafe {
            format_the_statistical_methods_ontology_url_from_node_name(node_name)
        });
    }

    if is_valid_infectious_disease_ontology_node_name(node_name) {
        return Ok(unsafe { format_infectious_disease_ontology_url_from_node_name(node_name) });
    }

    if is_valid_drosophila_development_node_name(node_name) {
        return Ok(unsafe { format_drosophila_development_url_from_node_name(node_name) });
    }

    if is_valid_mental_disease_ontology_node_name(node_name) {
        return Ok(unsafe { format_mental_disease_ontology_url_from_node_name(node_name) });
    }

    if is_valid_human_developmental_anatomy_abstract_node_name(node_name) {
        return Ok(unsafe {
            format_human_developmental_anatomy_abstract_url_from_node_name(node_name)
        });
    }

    if is_valid_c_elegans_gross_anatomy_ontology_node_name(node_name) {
        return Ok(unsafe {
            format_c_elegans_gross_anatomy_ontology_url_from_node_name(node_name)
        });
    }

    if is_valid_phenotype_and_trait_ontology_node_name(node_name) {
        return Ok(unsafe { format_phenotype_and_trait_ontology_url_from_node_name(node_name) });
    }

    if is_valid_evidence_ontology_node_name(node_name) {
        return Ok(unsafe { format_evidence_ontology_url_from_node_name(node_name) });
    }

    if is_valid_information_artifact_ontology_node_name(node_name) {
        return Ok(unsafe { format_information_artifact_ontology_url_from_node_name(node_name) });
    }

    if is_valid_brenda_tissue_enzyme_source_node_name(node_name) {
        return Ok(unsafe { format_brenda_tissue_enzyme_source_url_from_node_name(node_name) });
    }

    if is_valid_mouse_developmental_anatomy_ontology_node_name(node_name) {
        return Ok(unsafe {
            format_mouse_developmental_anatomy_ontology_url_from_node_name(node_name)
        });
    }

    if is_valid_mammalian_phenotype_ontology_node_name(node_name) {
        return Ok(unsafe { format_mammalian_phenotype_ontology_url_from_node_name(node_name) });
    }

    if is_valid_foodon_node_name(node_name) {
        return Ok(unsafe { format_foodon_url_from_node_name(node_name) });
    }

    if is_valid_vertebrate_trait_ontology_node_name(node_name) {
        return Ok(unsafe { format_vertebrate_trait_ontology_url_from_node_name(node_name) });
    }

    if is_valid_exposure_ontology_node_name(node_name) {
        return Ok(unsafe { format_exposure_ontology_url_from_node_name(node_name) });
    }

    if is_valid_population_and_community_ontology_node_name(node_name) {
        return Ok(unsafe {
            format_population_and_community_ontology_url_from_node_name(node_name)
        });
    }

    if is_valid_clinical_measurement_ontology_node_name(node_name) {
        return Ok(unsafe { format_clinical_measurement_ontology_url_from_node_name(node_name) });
    }

    if is_valid_mondo_disease_ontology_node_name(node_name) {
        return Ok(unsafe { format_mondo_disease_ontology_url_from_node_name(node_name) });
    }

    if is_valid_ontology_for_general_medical_science_node_name(node_name) {
        return Ok(unsafe {
            format_ontology_for_general_medical_science_url_from_node_name(node_name)
        });
    }

    if is_valid_common_anatomy_reference_ontology_node_name(node_name) {
        return Ok(unsafe {
            format_common_anatomy_reference_ontology_url_from_node_name(node_name)
        });
    }

    if is_valid_gazetteer_node_name(node_name) {
        return Ok(unsafe { format_gazetteer_url_from_node_name(node_name) });
    }

    if is_valid_uberon_multispecies_anatomy_ontology_node_name(node_name) {
        return Ok(unsafe {
            format_uberon_multispecies_anatomy_ontology_url_from_node_name(node_name)
        });
    }

    if is_valid_mental_functioning_ontology_node_name(node_name) {
        return Ok(unsafe { format_mental_functioning_ontology_url_from_node_name(node_name) });
    }

    if is_valid_relation_ontology_node_name(node_name) {
        return Ok(unsafe { format_relation_ontology_url_from_node_name(node_name) });
    }

    if is_valid_ontology_for_parasite_lifecycle_node_name(node_name) {
        return Ok(unsafe { format_ontology_for_parasite_lifecycle_url_from_node_name(node_name) });
    }

    if is_valid_dictyostelium_discoideum_anatomy_node_name(node_name) {
        return Ok(unsafe {
            format_dictyostelium_discoideum_anatomy_url_from_node_name(node_name)
        });
    }

    if is_valid_emotion_ontology_node_name(node_name) {
        return Ok(unsafe { format_emotion_ontology_url_from_node_name(node_name) });
    }

    if is_valid_eaglei_resource_ontology_node_name(node_name) {
        return Ok(unsafe { format_eaglei_resource_ontology_url_from_node_name(node_name) });
    }

    if is_valid_dublin_core_terms_node_name(node_name) {
        return Ok(unsafe { format_dublin_core_terms_url_from_node_name(node_name) });
    }

    if is_valid_dublin_core_types_node_name(node_name) {
        return Ok(unsafe { format_dublin_core_types_url_from_node_name(node_name) });
    }

    if is_valid_flybase_controlled_vocabulary_node_name(node_name) {
        return Ok(unsafe { format_flybase_controlled_vocabulary_url_from_node_name(node_name) });
    }

    if is_valid_foundational_model_of_anatomy_ontology_subset_node_name(node_name) {
        return Ok(unsafe {
            format_foundational_model_of_anatomy_ontology_subset_url_from_node_name(node_name)
        });
    }

    if is_valid_gene_ontology_obo_in_owl_node_name(node_name) {
        return Ok(unsafe { format_gene_ontology_obo_in_owl_url_from_node_name(node_name) });
    }

    if is_valid_monarch_initiative_archive_node_name(node_name) {
        return Ok(unsafe { format_monarch_initiative_archive_url_from_node_name(node_name) });
    }

    if is_valid_monarch_initiative_data_node_name(node_name) {
        return Ok(unsafe { format_monarch_initiative_data_url_from_node_name(node_name) });
    }

    if is_valid_protein_modification_node_name(node_name) {
        return Ok(unsafe { format_protein_modification_url_from_node_name(node_name) });
    }

    if is_valid_rdfs_node_name(node_name) {
        return Ok(unsafe { format_rdfs_url_from_node_name(node_name) });
    }

    if is_valid_unified_medical_language_system_semantic_code_node_name(node_name) {
        return Ok(unsafe {
            format_unified_medical_language_system_semantic_code_url_from_node_name(node_name)
        });
    }

    if is_valid_unified_medical_language_system_semantic_type_node_name(node_name) {
        return Ok(unsafe {
            format_unified_medical_language_system_semantic_type_url_from_node_name(node_name)
        });
    }

    if is_valid_unified_medical_language_system_semantic_group_node_name(node_name) {
        return Ok(unsafe {
            format_unified_medical_language_system_semantic_group_url_from_node_name(node_name)
        });
    }

    if is_valid_cord_pubmed_central_node_name(node_name) {
        return Ok(unsafe { format_cord_pubmed_central_url_from_node_name(node_name) });
    }

    if is_valid_website_node_name(node_name) {
        return Ok(format_website_url_from_node_name(node_name));
    }

    if is_valid_angular_link(node_name) {
        return Ok(format_angular_link_url_from_object(node_name));
    }

    if may_be_string_node_name(node_name) {
        if is_valid_string_pseudomonas_aeruginosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_cepacia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_oleovorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_campestris_campestris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_agrobacterium_rhizogenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sinorhizobium_fredii_ngr234_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_baumannii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alcaligenes_faecalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pantoea_agglomerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_klebsiella_oxytoca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_klebsiella_pneumoniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_proteus_vulgaris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_yersinia_pseudotuberculosis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_vulnificus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pasteurella_multocida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pasteurella_pneumotropica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydia_trachomatis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_ureolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_vulgaris_hildenborough_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_vulgaris_miyazaki_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_hydatis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erythrobacter_longus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_elongatus_pcc7942_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechocystis_sp_pcc6714_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechocystis_sp_pcc6803_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leuconostoc_lactis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dermacoccus_nishinomiyaensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_aureus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_xylosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_pyogenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_iniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_cereus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_mycoides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_atcc29733_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sticklandii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_animalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_adolescentis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_coryneforme_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_cuniculi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_pseudolongum_globosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_magnum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_minimum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oerskovia_turbata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_pseudotuberculosis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodococcus_fascians_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinoplanes_utahensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_aureofaciens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_bikiniensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_fradiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_glaucescens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_halstedii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sclerotialus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_virginiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_cellulosae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microtetraspora_glauca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptosporangium_amethystogenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pimelobacter_simplex_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_terrae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudonocardia_autotrophica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_agalactiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_californicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermoanaerobacter_kivui_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_solemya_velum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sodalis_pierantonius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fervidobacterium_islandicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_citrus_sinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synergistes_jonesii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phaeodactylum_tricornutum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ectocarpus_siliculosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_emiliania_huxleyi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydomonas_reinhardtii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_volvox_carteri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_auxenochlorella_protothecoides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_physcomitrella_patens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_theobroma_cacao_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_carica_papaya_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cucumis_melo_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cucumis_sativus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_populus_trichocarpa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arabidopsis_thaliana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brassica_rapa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brassica_oleracea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_malus_domestica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prunus_persica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cicer_arietinum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_glycine_max_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_medicago_truncatula_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phaseolus_vulgaris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_manihot_esculenta_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ricinus_communis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_linum_usitatissimum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_solanum_lycopersicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nicotiana_sylvestris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nicotiana_tomentosiformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_solanum_tuberosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erythranthe_guttata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nelumbo_nucifera_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hordeum_vulgare_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oryza_rufipogon_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oryza_sativa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oryza_brachyantha_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oryza_nivara_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oryza_punctata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oryza_glaberrima_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_setaria_italica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sorghum_bicolor_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_triticum_aestivum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_triticum_urartu_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_zea_mays_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_musa_acuminata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phytophthora_infestans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phytophthora_parasitica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_schizosaccharomyces_pombe_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_schizosaccharomyces_japonicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_schizosaccharomyces_octosporus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pichia_kudriavzevii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_millerozyma_farinosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_komagataella_pastoris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_scheffersomyces_stipitis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_meyerozyma_guilliermondii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharomyces_cerevisiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_torulaspora_delbrueckii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_yarrowia_lipolytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_zygosaccharomyces_rouxii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_debaryomyces_hansenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_taphrina_deformans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bipolaris_maydis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bipolaris_zeicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptosphaeria_maculans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_histoplasma_capsulatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blastomyces_dermatitidis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aspergillus_clavatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aspergillus_flavus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aspergillus_niger_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aspergillus_oryzae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_claviceps_purpurea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusarium_fujikuroi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neurospora_crassa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_podospora_anserina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sordaria_macrospora_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sclerotinia_sclerotiorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cryptococcus_neoformans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tremella_mesenterica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ustilago_maydis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodosporidium_toruloides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_puccinia_graminis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_schizophyllum_commune_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_agaricus_bisporus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coprinopsis_cinerea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_colletotrichum_orbiculare_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_candida_albicans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_candida_glabrata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_candida_maltosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_candida_parapsilosis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_candida_tropicalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coccidioides_immitis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusarium_oxysporum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusarium_graminearum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_trichophyton_rubrum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leishmania_major_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leishmania_infantum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leishmania_panamensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_trypanosoma_brucei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_trypanosoma_cruzi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_trichomonas_vaginalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_giardia_intestinalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_entamoeba_histolytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_naegleria_gruberi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dictyostelium_purpureum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cryptosporidium_muris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_toxoplasma_gondii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_plasmodium_berghei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_plasmodium_chabaudi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_plasmodium_cynomolgi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_plasmodium_falciparum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_plasmodium_knowlesi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_plasmodium_vivax_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_plasmodium_vinckei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_babesia_bovis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_babesia_equi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_theileria_annulata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_theileria_parva_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paramecium_tetraurelia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tetrahymena_thermophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ichthyophthirius_multifiliis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_exophiala_dermatitidis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hydra_vulgaris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_schistosoma_mansoni_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_echinococcus_multilocularis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caenorhabditis_briggsae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caenorhabditis_elegans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bursaphelenchus_xylophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_trichinella_spiralis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helobdella_robusta_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aplysia_californica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_daphnia_pulex_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acyrthosiphon_pisum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tribolium_castaneum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bombyx_mori_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aedes_aegypti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anopheles_gambiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_culex_quinquefasciatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_loa_loa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ceratitis_capitata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_drosophila_ananassae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_drosophila_erecta_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_drosophila_grimshawi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_drosophila_melanogaster_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_drosophila_mojavensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_drosophila_persimilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_drosophila_pseudoobscura_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_drosophila_virilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_drosophila_yakuba_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_drosophila_willistoni_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_musca_domestica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nasonia_vitripennis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_apis_mellifera_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_strongylocentrotus_purpuratus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ciona_intestinalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_branchiostoma_floridae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_latimeria_chalumnae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lepisosteus_oculatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_danio_rerio_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_astyanax_mexicanus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_esox_lucius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gadus_morhua_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_poecilia_reticulata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xiphophorus_maculatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oryzias_latipes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oreochromis_niloticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haplochromis_burtoni_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xenopus_tropicalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chelonia_mydas_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chrysemys_picta_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alligator_mississippiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_columba_livia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gallus_gallus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ornithorhynchus_anatinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sarcophilus_harrisii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_macropus_eugenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dasypus_novemcinctus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erinaceus_europaeus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_echinops_telfairi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pteropus_alecto_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tarsius_syrichta_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_callithrix_jacchus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_macaca_fascicularis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_macaca_mulatta_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_papio_anubis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gorilla_gorilla_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pan_paniscus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pan_troglodytes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pongo_abelii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_homo_sapiens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_canis_lupus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ailuropoda_melanoleuca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mustela_putorius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_felis_catus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_panthera_tigris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_odobenus_rosmarus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptonychotes_weddellii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_orcinus_orca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tursiops_truncatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_balaenoptera_acutorostrata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_trichechus_manatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_loxodonta_africana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_orycteropus_afer_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sus_scrofa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bos_taurus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ovis_aries_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ochotona_princeps_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oryctolagus_cuniculus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cricetulus_griseus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesocricetus_auratus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peromyscus_maniculatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mus_musculus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rattus_norvegicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cavia_porcellus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_octodon_degus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_heterocephalus_glaber_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccoglossus_kowalevskii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_trichoplax_adhaerens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_atta_cephalotes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dactylococcopsis_salina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_danaus_plexippus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodnius_prolixus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amborella_trichopoda_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobotrys_oligospora_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_monodelphis_domestica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_parastagonospora_nodorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_paucimobilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobium_yanoikuyae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelodiscus_sinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brachypodium_distachyon_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_naumovozyma_castellii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_naumovozyma_dairenensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_verticillium_dahliae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saimiri_boliviensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mnemiopsis_leidyi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharopolyspora_rectivirgula_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nostoc_sp_pcc7524_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_macacae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_yersinia_kristensenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_photodesmus_katoptron_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_colwellia_psychrerythraea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cobetia_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anolis_carolinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_herbidospora_cretacea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tarenaya_hassleriana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_talaromyces_stipitatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_allomyces_macrogynus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_elephantulus_edwardii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kluyveromyces_lactis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ursus_maritimus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eptesicus_fuscus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neospora_caninum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_griseoluteus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_yersinia_ruckeri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_navarrensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natrialba_asiatica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_janthinobacterium_lividum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gossypium_raimondii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vitis_vinifera_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gaeumannomyces_graminis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_trichoderma_virens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sporothrix_schenckii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_otolemur_garnettii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_takifugu_rubripes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caenorhabditis_remanei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_colletotrichum_graminicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_pseudolongum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clavibacter_michiganensis_sepedonicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_sputorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_stutzeri_atcc14405_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_pcc7002_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_wh_7803_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_calothrix_sp_pcc7103_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tetranychus_urticae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neolamprologus_brichardi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blautia_producta_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eremothecium_gossypii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aspergillus_terreus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_uncinocarpus_reesii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_catenuloplanes_japonicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_galbus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_thermophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracoccus_versutus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mixia_osmundae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blumeria_graminis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_strongyloides_ratti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_heliconius_melpomene_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chinchilla_lanigera_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thalassiosira_pseudonana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_citrobacter_amalonaticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thielavia_terrestris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_macrophomina_phaseolina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dactylosporangium_aurantiacum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_choerinum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_thermoamylovorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vanderwaltozyma_polyspora_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mucor_circinelloides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pythium_irregulare_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neosartorya_fischeri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_penicillium_digitatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_abscessus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_wigglesworthia_glossinidia_sp_gbr_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_cangingivalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_canoris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lodderomyces_elongisporus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_algidicarnis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aegilops_tauschii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phytoplasma_mali_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_talaromyces_marneffei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodococcus_opacus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chaetomium_globosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alligator_sinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_panicum_virgatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_micromonas_pusilla_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tuber_melanosporum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_equi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neurospora_tetrasperma_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oryza_glumipatula_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oryza_meridionalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_junii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_cipa165_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aspergillus_kawachii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fomitopsis_pinicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_stereum_hirsutum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_botrytis_cinerea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lentzea_albidocapillata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyanothece_sp_pcc8801_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bathycoccus_prasinos_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pneumocystis_jirovecii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pythium_vexans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sorex_araneus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rubrobacter_radiotolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phoenix_dactylifera_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_candida_dubliniensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_salina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anopheles_darlingi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ictidomys_tridecemlineatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capronia_epimyces_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capronia_coronata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prauserella_rugosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_wedmorensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyanothece_sp_atcc51142_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aureococcus_anophagefferens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_megasporus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_durus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_avium_avium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dictyostelium_discoideum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bipolaris_sorokiniana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrenophora_triticirepentis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyanidioschyzon_merolae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eremothecium_cymbalariae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nematostella_vectensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_conjunctivae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_candida_tenuis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anabaena_sp_90_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_weissella_hellenica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobium_chlorophenolicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_entamoeba_dispar_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_olivaceus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_lydicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_triplex_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_poecilia_formosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arabis_alpina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_trogontum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_jaculus_jaculus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_trichoderma_reesei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ciona_savignyi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfitobacter_sp_ee36_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haliaeetus_leucocephalus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrenophora_teres_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pectobacterium_betavasculorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_guillardia_theta_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_anguillarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_toyocaensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cylindrospermum_stagnale_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oscillatoria_acuminata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_syntrophus_aciditrophicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fragaria_vesca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spirillospora_albida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_celluloflavus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rickettsiella_grylli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fibrobacter_succinogenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_myotis_lucifugus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pantholops_hodgsonii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arabidopsis_lyrata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phytoplasma_australiense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ficedula_albicollis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_pastoris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_mit9107_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_gp2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_sb_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_wh_7805_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_paraplantarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlorocebus_sabaeus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ophiostoma_piceae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coniosporium_apollinis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhinopithecus_roxellana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pluralibacter_gergoviae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nomascus_leucogenys_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_azoarcus_sp_bh72_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_adp1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthroderma_benhamiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microsporum_gypseum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthroderma_otae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_trichophyton_equinum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_trichoderma_atroviride_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nostoc_punctiforme_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycosphaerella_pini_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_cc9311_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pythium_ultimum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halothece_sp_pcc7418_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyanothece_sp_pcc7424_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oryza_barthii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinoalloteichus_cyanogriseus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_piriformospora_indica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pantoea_stewartii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_niger_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_violens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_roseoverticillatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_clausii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_atroolivaceus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_bicolor_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_catenulae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_griseorubens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_albus_albus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_alboflavus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_aureocirculatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_californicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_lavenduligriseus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_mutabilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_purpeofuscus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_resistomycificus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_varsoviensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phytophthora_sojae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lechevalieria_aerocolonigenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_durhamensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_flavochromogenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_iakyrus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_katrae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_pyridomyceticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_albulus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_theileria_orientalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermococcus_kodakarensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_wh_5701_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aquamicrobium_defluvii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gasterosteus_aculeatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microplitis_demolitor_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_vlb120_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caulobacter_henricii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_penicillium_oxalicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ostreococcus_tauri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrococcus_horikoshii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eucalyptus_grandis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haemophilus_influenzae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bos_mutus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphaeroforma_arctica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ophiocordyceps_sinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sporisorium_reilianum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_boechera_stricta_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eutrema_salsugineum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_seoulensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cordyceps_militaris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_mit9302_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_mit9312_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_mit9313_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiomonas_intermedia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aromatoleum_aromaticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycetocola_saprophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_putida_gb1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leersia_perrieri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_subtile_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthobacter_autotrophicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_gallinarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_merycicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_ruminantium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pectobacterium_carotovorum_odoriferum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_myceliophthora_thermophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mortierella_verticillata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microtus_ochrogaster_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coniophora_puteana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_colletotrichum_higginsianum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_monosiga_brevicollis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capsella_rubella_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_trichosporon_asahii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudanabaena_sp_pcc7367_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serratia_grimesii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serratia_plymuthica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfitobacter_mediterraneus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_tuberculosis_h37rv_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudocercospora_fijiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gammaproteobacteria_bacterium_hdn1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lysobacter_antibioticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_wh_8102_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudozyma_flocculosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thauera_sp_mz1t_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_citrus_clementina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphaerulina_musiva_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_pylori_26695_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_pylori_j99_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serpula_lacrymans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cladophialophora_carrionii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_endosymbiont_of_acanthamoeba_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_pasteurianum_bc1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoalteromonas_tunicata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selaginella_moellendorffii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseovarius_nubinhibens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bubalus_bubalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salmonella_enterica_typhimurium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_camelina_sativa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiomicrospira_sp_milost1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiomicrospira_sp_milost2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_pcc7335_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracaedibacter_acanthamoebae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_metarhizium_acridum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_mit9201_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_mit9211_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_mit9215_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pandoraea_pnomenusa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_setosphaeria_turcica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_sp_ana3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bordetella_petrii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_m1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfococcus_oleovorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eutypa_lata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_sp_asf360_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_asf356_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_asf502_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fischerella_thermalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hammondia_hammondi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_calothrix_sp_pcc7507_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_coelicolor_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_wolbachia_sp_ooc_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusarium_pseudograminearum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bipolaris_oryzae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodococcus_jostii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_glarea_lozoyensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prunus_mume_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xenococcus_sp_pcc7305_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptolyngbya_sp_pcc7375_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gloeocapsa_sp_pcc73106_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acromyrmex_echinatior_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nostoc_sp_pcc7120_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharothrix_syringae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gloeophyllum_trabeum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serratia_sp_atcc39006_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptacidiphilus_neutrinimicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptacidiphilus_carbonis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptacidiphilus_albus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrosococcus_watsonii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_frankia_sp_cci3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_maylandia_zebra_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_bereziniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylosinus_sp_lw3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylosinus_sp_pw1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_buchnera_aphidicola_aps_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_myotis_brandtii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spizellomyces_punctatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_batrachochytrium_dendrobatidis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardioides_sp_cf8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gregarina_niphandrodes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_cc9605_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_cc9616_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_gulae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_stanieria_cyanosphaera_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptolyngbya_sp_pcc7376_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saprolegnia_diclina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geminocystis_herdmanii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_tv2a2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tetrapisispora_phaffii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_ors278_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pythium_iwayamai_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydophila_pneumoniae_ar39_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydophila_pneumoniae_cwl029_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusarium_verticillioides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chrysiogenes_arsenatis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pleurocapsa_sp_pcc7319_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pleurocapsa_sp_pcc7327_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nodosilinea_nodulosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coleofasciculus_chthonoplastes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudanabaena_sp_pcc6802_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lipotes_vexillifer_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ustilago_hordei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pediculus_humanus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracoccidioides_brasiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_persephonella_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bordetella_trematum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_strigamia_maritima_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nipponia_nippon_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_galdieria_sulphuraria_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bombus_impatiens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pteropus_vampyrus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinoplanes_sp_se50110_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caenorhabditis_brenneri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_zootermopsis_nevadensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kocuria_polaris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylosinus_sp_lw4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfitobacterium_hafniense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nectria_haematococca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnobacterium_bovis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_zobellia_uliginosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_stegastes_partitus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_as9601_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_griseofuscus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_magnaporthiopsis_poae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_kunkeei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_wallemia_sebi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kozakia_baliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sporocytophaga_myxococcoides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrosomonas_sp_al212_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gallibacterium_genomosp_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_escherichia_coli_o157h7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alteromonadales_bacterium_tw7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_magnetococcus_marinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_imitans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aphanomyces_invadans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_cremoricolorata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphaerochaeta_globosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphaerochaeta_pleomorpha_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_novosphingobium_resinovorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_scardovii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cedecea_neteri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dechloromonas_aromatica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_sacchari_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thalassiosira_oceanica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_putida_kt2440_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xylella_fastidiosa_9a5c_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_borealis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_auricularia_delicata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermodesulfobacterium_hydrogeniphilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erythrobacter_sp_sd21_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_beta_vulgaris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aspergillus_nidulans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_wolbachia_sp_dme_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anabaena_sp_pcc7108_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phytophthora_ramorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_sp_jls_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_wh_8109_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_wh_8016_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_ccmp1375_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_mit9515_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_mit9301_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_mit9314_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_mit9322_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_marinus_natl1a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_listeria_monocytogenes_egde_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_pneumoniae_tigr4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_pneumoniae_r6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanobacillus_picturae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_elizabethkingia_miricola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_thb2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_beauveria_bassiana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_epidermidis_rp62a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_epidermidis_atcc12228_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_agrobacterium_fabrum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_python_bivittatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_francisella_tularensis_schus4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfobacterium_autotrophicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfotalea_psychrophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrobaculum_aerophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acetobacter_malorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oscillatoria_nigroviridis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyanobium_sp_pcc7001_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_robinsoniella_peoriensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudopodoces_humilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydophila_pneumoniae_tw183_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_cetorum_mit007128_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chrysochloris_asiatica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_baumannia_cicadellinicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrococcus_furiosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alkalilimnicola_ehrlichii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylocystis_sp_sc2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dermacoccus_sp_ellin185_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_graminis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_odorifer_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_ec33_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_nucleatum_atcc25586_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_axonopodis_citri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caulobacter_crescentus_cb15_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_atypicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_jejuni_nctc11168_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capsaspora_owczarzaki_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanosarcina_mazei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlorobium_tepidum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_sp_atcc31555_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_perfringens_atcc13124_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haematobacter_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_pcc7336_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_pcc6312_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermococcus_nautili_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardioides_sp_js614_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_efficiens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_sordidicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_ai1a2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_glutamicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermosynechococcus_elongatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_anthracis_ames_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shigella_flexneri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anoxybacillus_gonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dickeya_dadantii_3937_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_buchnera_aphidicola_sg_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coccidioides_posadasii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_escherichia_coli_cft073_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_punctularia_strigosozonata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_listeria_ivanovii_londoniensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_gerneri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_tandoii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_tjernbergiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_towneri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminiclostridium_thermocellum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leuconostoc_mesenteroides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharophagus_degradans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oenococcus_oeni_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_trichodesmium_erythraeum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tropheryma_whipplei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tannerella_forsythia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blochmannia_floridanus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_melampsora_laricipopulina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfurihydrogenibium_azorense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_koribacter_versatilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_herminiimonas_arsenicoxydans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haemophilus_somnus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_syringae_b728a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ehrlichia_chaffeensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_fluorescens_pf01_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_alaskensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neptuniibacter_caesariensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amycolatopsis_japonica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amycolatopsis_vancoresmycina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_carnobacterium_sp_174_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fomitiporia_mediterranea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chaetomium_thermophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_mutans_ua159_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_agalactiae_nem316_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_allokutzneria_albata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlorogloeopsis_fritschii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_oneidensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaplasma_phagocytophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_yersinia_pestis_co92_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_larimichthys_crocea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enhygromyxa_salina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_rhizosphaerae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_croceibacter_atlanticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_cenocepacia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_marinum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_fluorescens_sbw25_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_leguminosarum_3841_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_longum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_psychraerophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_vietnamensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pectobacterium_atrosepticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salmonella_bongori_nctc12419_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_uberis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydophila_abortus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aquilegia_coerulea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_micromonospora_sp_atcc39149_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salmonella_enterica_ct18_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_protegens_pf5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_plantarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_putidum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_moniliophthora_roreri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanobacillus_iheyensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mastigocladopsis_repens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_rs9916_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_rs9917_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mannheimia_succiniciproducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_frankia_sp_bmg512_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neorickettsia_sennetsu_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natrinema_altunense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kocuria_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_togninia_minima_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_syringae_tomato_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_parahaemolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_subtilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aquifex_aeolicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_archaeoglobus_fulgidus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanobrevibacter_sp_abm4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_diazoefficiens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brucella_melitensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_buchnera_aphidicola_bp_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrus_x_bretschneideri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_myotis_davidii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_piezotolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinobacter_adhaerens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_faecalis_v583_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_thetaiotaomicron_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_grosmannia_clavigera_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bigelowiella_natans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coxiella_burnetii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_avermitilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydophila_caviae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinobacillus_pleuropneumoniae_1_4074_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomonas_neptunium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrosomonas_europaea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_beijerinckia_mobilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_alkaliantarctica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_cb0101_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidovorax_sp_js42_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haemophilus_ducreyi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_solibacter_usitatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodococcus_erythropolis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaplasma_marginale_maries_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoalteromonas_sp_sm9913_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_hepaticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacillus_kaustophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptacidiphilus_jiangxiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alcanivorax_sp_dg881_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_formosense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_scalindua_brodae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_alkylphenolia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erythrobacter_sp_nap1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidobacterium_capsulatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_verrucomicrobium_spinosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anabaena_variabilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halobacillus_dabanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ostreococcus_lucimarinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_gingivalis_w83_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodopirellula_baltica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidithiobacillus_ferrooxidans_atcc23270_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_mallei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydia_muridarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dehalococcoides_mccartyi_195_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_radiodurans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacter_sulfurreducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanocaldococcus_jannaschii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylococcus_capsulatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_photorhabdus_luminescens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_arthritidis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_genitalium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermotoga_maritima_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_denticola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_pallidum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_cholerae_o1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chromobacterium_violaceum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_lutea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cynoglossus_semilaevis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caedibacter_acanthamoebae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracaedibacter_symbiosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_wallemia_ichthyophaga_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_carboxydothermus_hydrogenoformans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dichelobacter_nodosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_smegmatis_mc2155_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_myxococcus_xanthus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_albus_8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruegeria_pomeroyi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_mitis_nctc_12261_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tupaia_chinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermococcus_sp_am4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_farcinica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_planctomycete_ksu1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gammaproteobacteria_bacterium_htcc2143_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gammaproteobacteria_bacterium_htcc2148_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gammaproteobacteria_bacterium_htcc2080_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coccomyxa_subellipsoidea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gloeobacter_violaceus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chroococcidiopsis_thermalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanicola_batsensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_c_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ehrlichia_ruminantium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dehalococcoides_mccartyi_cbdb1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bordetella_bronchiseptica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bordetella_pertussis_tohama_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_johnsonii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rickettsia_typhi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kitasatospora_arboriphila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_cosmeticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodopseudomonas_palustris_cga009_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychrobacter_arcticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_anthracis_sterne_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrosomonas_sp_is79a3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_buchnera_aphidicola_ctu_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_avium_paratuberculosis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_exiguobacterium_sibiricum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_hyopneumoniae_j_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_synoviae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermus_thermophilus_hb27_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phytoplasma_onion_yellows_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_verrucosispora_maris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pneumocystis_murina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_picrophilus_torridus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ralstonia_eutropha_jmp134_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_thermophilus_lmg18311_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_protochlamydia_amoebophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydophila_felis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_zymomonas_mobilis_zm4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capsella_grandiflora_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bdellovibrio_bacteriovorus_hd100_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_syringae_phaseolicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_ruminicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_moorella_thermoacetica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_byssochlamys_spectabilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylobacillus_flagellatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesoplasma_florum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_cibi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rubrobacter_xylanophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cupriavidus_metallidurans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_xenovorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_antarcticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_gingivicanis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chelativorans_sp_bnc1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thalassobacter_stenotrophicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sinorhizobium_meliloti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_loti_maff303099_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kineococcus_radiotolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanococcus_maripaludis_s2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ralstonia_solanacearum_gmi1000_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_propionibacterium_acnes_kpa171202_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_mobile_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_wynnii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natronomonas_moolapensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_elongatus_pcc6301_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_vietnamiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ehrlichia_canis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodospirillum_rubrum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanosarcina_barkeri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cytophaga_hutchinsonii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacter_metallireducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermobifida_fusca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinobacter_sp_elb17_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anabaena_cylindrica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptolyngbya_boryana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aggregatibacter_actinomycetemcomitans_hk1651_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aeropyrum_pernix_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_halodurans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_fragilis_nctc_9343_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_pseudomallei_k96243_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_acetobutylicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptoclostridium_difficile_630_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gluconacetobacter_diazotrophicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloarcula_marismortui_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_acidophilus_ncfm_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactococcus_lactis_il1403_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_legionella_pneumophila_philadelphia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_listeria_innocua_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylobacterium_extorquens_am1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_leprae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_mycoides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_penetrans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_pulmonis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrococcus_abyssi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodobacter_capsulatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodobacter_sphaeroides_241_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rickettsia_prowazekii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rickettsia_sibirica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyaloperonospora_arabidopsidis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfolobus_solfataricus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfolobus_tokodaii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldanaerobacter_subterraneus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermoplasma_acidophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermoplasma_volcanium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ureaplasma_parvum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_wolinella_succinogenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_candida_orthopsilosis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serratia_marcescens_db11_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_oleivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pediococcus_pentosaceus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_diplosphaera_colitermitum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidobacteriaceae_bacterium_taa166_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_licheniformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_novosphingobium_aromaticivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudogulbenkiania_ferrooxidans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_haemolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leifsonia_xyli_ctcb07_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caenorhabditis_japonica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_quintana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_henselae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoalteromonas_sp_bsw20308_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_idiomarina_loihiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_flavovariabilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_xylophagus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_fulvoviolaceus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amycolatopsis_rifamycinica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_btai1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_renibacterium_salmoninarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermodesulfovibrio_yellowstonii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermodesulfobacterium_commune_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_bovis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlorobium_limicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlorobium_phaeobacteroides_dsm266_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlorobium_phaeovibrioides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_aurescens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaeromyxobacter_dehalogenans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chromohalobacter_salexigens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_fb24_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_jannaschia_sp_ccs1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_beijerinckii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_borrelia_garinii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prosthecochloris_aestuarii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gluconobacter_oxydans_621h_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_photorhabdus_asymbiotica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blochmannia_pennsylvanicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erythrobacter_vulgaris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruegeria_sp_tm1040_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiobacillus_denitrificans_atcc25259_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_symbiobacterium_thermophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyanobacterium_stanieri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyanobium_gracile_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_wolbachia_sp_bma_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyphellophora_europaea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rickettsia_canadensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rickettsia_akari_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alkaliphilus_metalliredigens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_hyopneumoniae_232_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_micromonas_sp_rcc299_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_polaromonas_sp_js666_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_legionella_pneumophila_paris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_photobacterium_profundum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_frankia_sp_ean1pec_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_frankia_sp_eui1c_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_frankia_sp_cn3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermus_thermophilus_hb8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pundamilia_nyererei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanocella_paludicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thauera_sp_27_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_endozoicomonas_elysicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_lari_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_upsaliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fischerella_muscicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_jeikeium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_vrystaatense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coprothermobacter_proteolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dictyoglomus_thermophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloferax_volcanii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermomicrobium_roseum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermotoga_neapolitana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salinibacter_ruber_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neofusicoccum_parvum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_agrobacterium_vitis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_agrobacterium_radiobacter_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dehalococcoides_mccartyi_vs_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_polynucleobacter_necessarius_asymbioticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinobacterium_phsc20c1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_fischeri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_janibacter_sp_htcc2649_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dokdonia_sp_med134_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_polaribacter_irgensii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychroflexus_torquis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_robiginitalea_biformata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_polaribacter_sp_med152_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_maribacter_sp_htcc2170_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microscilla_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lyngbya_sp_pcc8106_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nodularia_spumigena_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_bl107_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lentisphaera_araneosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erythrobacter_litoralis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blastopirellula_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fulvimarina_pelagi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_loktanella_vestfoldensis_ska53_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanicaulis_sp_htcc2633_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanicola_granulosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_parvularcula_bermudensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseobacter_sp_med193_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseovarius_sp_217_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelagibaca_bermudensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_sp_ska58_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodobacteraceae_bacterium_htcc2083_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_maritimibacter_alkaliphilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alteromonas_macleodii_deep_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrococcus_mobilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychromonas_sp_cnpt3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_congregibacter_litoralis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gammaproteobacteria_bacterium_htcc2207_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_photobacterium_angustum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_sakei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mariprofundus_ferrooxydans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_beta_proteobacterium_kb13_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_borrelia_hermsii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_borrelia_turicatae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rickettsia_felis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_weihenstephanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_cytotoxicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_pumilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodopseudomonas_palustris_bisa53_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodopseudomonas_palustris_bisb18_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodopseudomonas_palustris_bisb5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodopseudomonas_palustris_haa2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacter_daltonii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_campestris_vesicatoria_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_herpetosiphon_aurantiacus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aliivibrio_salmonicida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_rcc307_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_cc9902_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_escherichia_coli_k12_w3110_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_axonopodis_phaseoli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiomicrospira_crunogena_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorothrix_hollandica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingopyxis_alaskensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nostoc_sp_pcc7107_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_denitrificans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_frigidimarina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_rufum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sulfidigenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracoccus_denitrificans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_magnaporthe_oryzae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_wsm1743_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_wsm1253_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_putrefaciens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlorobium_luteolum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nonlabens_sediminis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_geothermalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaplasma_marginale_florida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_ja33ab_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_ja23ba_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_simiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevibacterium_linens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phytoplasma_wheat_blue_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phytoplasma_aster_yellows_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_thermophilus_lmd9_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_azotobacter_vinelandii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrobacter_hamburgensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrobacter_winogradskyi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanospirillum_hungatei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrosococcus_oceani_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrosospira_multiformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_loihica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_jdr2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chloroflexus_aurantiacus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_gasseri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelodictyon_phaeoclathratiforme_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phytophthora_kernoviae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_axonopodis_vasculorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_amazonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfurimonas_denitrificans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_amyloliquefaciens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_frankia_alni_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_helveticus_cnrz32_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_breve_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chloroflexus_aggregans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoalteromonas_haloplanktis_tac125_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudozyma_hubeiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_crudilactis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acaryochloris_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amycolatopsis_jejuensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrospira_defluvii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfolobus_acidocaldarius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blattabacterium_sp_bge_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_simkania_negevensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydophila_pecorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydia_psittaci_6bc_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlorobium_phaeobacteroides_bs1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alpha_proteobacterium_bal199_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_drakei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_okhensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_fermentum_ifo3956_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_finegoldia_magna_atcc29328_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rickettsia_tamurae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrosomonas_eutropha_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychrobacter_cryohalolentis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_syntrophomonas_wolfei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_syntrophobacter_fumaroxidans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_s23321_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelagibacter_ubique_htcc1062_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rickettsia_bellii_rml369c_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyronema_omphalodes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_sp_ktr9_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelobacter_carbinolicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelobacter_propionicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodoferax_ferrireducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_ambifaria_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinobacillus_succinogenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanosphaera_stadtmanae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermoanaerobacter_pseudethanolicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spathaspora_passalidarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlorobium_chlorochromatii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_oklahomensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_saprophyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoalteromonas_atlantica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrococcus_sp_na2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sodalis_glossinidius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_planctomyces_maris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_cholerae_o395_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_coagulans_36d1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kutzneria_sp_744_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_hominis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_etli_cfn42_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natronomonas_pharaonis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_sp_lpu83_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodobacter_sphaeroides_atcc17025_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychrobacter_sp_prwf1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_reuteri_10023_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halorhodospira_halophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfotomaculum_reducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidiphilium_cryptum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leuconostoc_citreum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_polymyxa_e681_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hahella_chejuensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_akkermansia_muciniphila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_yersinia_intermedia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_yersinia_frederiksenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_gilvum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_vanbaalenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alkaliphilus_oremlandii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseobacter_sp_azwk3b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanocella_arvoryzae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinobacter_hydrocarbonoclasticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidothermus_cellulolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldicellulosiruptor_saccharolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_putida_f1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyramidobacter_piscolens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_delbrueckii_2038_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_jejuni_81176_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_tsurumiense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_micromonospora_chokoriensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_orientia_tsutsugamushi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_dorei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychromonas_ingrahamii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseiflexus_sp_rs1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnoclostridium_phytofermentans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidovorax_sp_kks102_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halobiforma_lacisalsi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevibacillus_brevis_nbrc100599_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_olindensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_oryzae_pxo99a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_bacilliformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_concisus_13826_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_fetus_8240_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_hominis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bordetella_avium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_exiguobacterium_sp_at1b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_ulcerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_reichenbachii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_escherichia_coli_536_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloquadratum_walsbyi_dsm16790_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lawsonia_intracellularis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_endocarpon_pusillum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_polaromonas_naphthalenivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ramlibacter_tataouinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_frankia_sp_bcu110501_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_mitis_b6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sinorhizobium_medicae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caulobacter_sp_k31_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_fuscans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phycicoccus_jejuensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodobacterales_bacterium_htcc2255_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arcobacter_butzleri_rm4018_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanoculleus_marisnigri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermofilum_pendens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salinispora_tropica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelotomaculum_thermopropionicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erwinia_typographi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodobacter_sp_sw2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_buchnera_aphidicola_bcc_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_pneumoniae_d39_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halothermothrix_orenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rivularia_sp_pcc7116_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_korarchaeum_cryptofilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_janthinobacterium_sp_marseille_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseobacter_denitrificans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_francisella_tularensis_holarctica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_johnsoniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracoccus_halophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_teredinibacter_turnerae_t7901_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kocuria_rhizophila_dc2201_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_stigmatella_aurantiaca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gemmatimonas_aurantiaca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_stutzeri_a1501_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_albilineans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidithiobacillus_ferrooxidans_atcc53993_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aeromonas_hydrophila_atcc7966_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hydrogenobaculum_sp_y04aas1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachancea_thermotolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ralstonia_eutropha_h16_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fervidobacterium_nodosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aeromonas_salmonicida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_verrucomicrobiae_bacterium_dg1235_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_acinonychis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_tribocorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseiflexus_castenholzii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erythrobacter_sp_jl475_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_oryzae_oryzicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrobaculum_islandicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_entomophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_labrenzia_aggregata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermophagus_xiamenensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_listeria_welshimeri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_novyi_nt_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanobacterium_arcticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitratiruptor_sp_sb1552_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfurovum_sp_nbc371_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_brevis_atcc367_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_archaeoglobus_sulfaticallidus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cupriavidus_sp_amp6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sagittula_stellata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodobacteraceae_bacterium_htcc2150_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_algoriphagus_machipongonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_planktothrix_agardhii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseobacter_sp_sk20926_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sanguinis_sk36_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_putida_w619_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_borrelia_afzelii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_delbrueckii_atcc11842_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermotoga_petrophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinocatenispora_sera_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_stenotrophomonas_maltophilia_r5513_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermosipho_melanesiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ehrlichia_sp_hf_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salinispora_arenicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_phymatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_granulibacter_bethesdensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_suis_05zyh33_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kordia_algicida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseobacter_sp_gai101_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseobacter_sp_ccs2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseobacter_litoralis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pedobacter_sp_bal39_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacteria_bacterium_bal38_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevundimonas_sp_bal3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacteriales_bacterium_alc1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyanothece_sp_ccy0110_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseovarius_sp_tm1035_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gammaproteobacteria_bacterium_htcc5015_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_octadecabacter_arcticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phaeobacter_inhibens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermococcus_barophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanibulbus_indolifex_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_plesiocystis_pacifica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_octadecabacter_antarcticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_verminephrobacter_eiseniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rickettsia_bellii_osu85389_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitratireductor_pacificus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_wittichii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_woodyi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pestalotiopsis_fici_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_yersinia_enterocolitica_8081_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_nucleatum_polymorphum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alcanivorax_borkumensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_crevioricanis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_maricaulis_maris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_cellulolyticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_multivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_leguminosarum_wsm2304_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_beggiatoa_alba_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gallionella_capsiferriformans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptothrix_cholodnii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyanothece_sp_pcc7425_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_beijerinckia_indica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylocapsa_acidiphila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylocella_silvestris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brachybacterium_phenoliresistens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_carnosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sulfidiphilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_k90mix_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marmoricola_aequoreus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_284_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_31_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_a2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_a4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidovorax_citrulli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldivirga_maquilingensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_pseudofirmus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudobacteroides_cellulosolvens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_bifidum_ncimb41171_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_elkanii_usda76_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_phytofirmans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_delftia_acidovorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_pealeana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dinoroseobacter_shibae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leeuwenhoekiella_blandensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacter_lovleyi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_metallosphaera_sedula_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylothermus_marinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_mendocina_ymp_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serratia_proteamaculans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterobacter_sp_638_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_comamonas_testosteroni_kf1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinomonas_sp_mwyl1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amphimedon_queenslandica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_terriglobus_saanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_dentium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermosinus_carboxydivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_psychrophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ralstonia_pickettii_12j_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kamptonema_formosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanococcus_maripaludis_c5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_parvibaculum_lavamentivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_petrotoga_mobilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacter_bemidjiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaeromyxobacter_sp_fw1095_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_helveticus_dpc4571_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharopolyspora_erythraea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_m313_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanococcus_vannielii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natrinema_sp_j72_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xenorhabdus_nematophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xenorhabdus_bovienii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardioidaceae_bacterium_broad1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanocorpusculum_labreanum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrobaculum_calidifontis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gramella_forsetii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_obeum_atcc29174_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_torques_atcc27756_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dorea_formicigenerans_atcc27755_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dorea_longicatena_dsm13814_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_ventriosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_piger_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_parvimonas_micra_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_odontolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoflavonifractor_capillosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_scindens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_hallii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_gnavus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_subdoligranulum_variabile_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_callidus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coprococcus_eutactus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_ovatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_parabacteroides_merdae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_uniformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_faecalibacterium_prausnitzii_a2165_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_l250_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerostipes_caccae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hoeflea_phototrophica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_caccae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_bolteae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_borrelia_duttonii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracoccus_sp_trp_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vesicomyosocius_okutanii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruthia_magnifica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halorubrum_halophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodospirillum_centenum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinopolyspora_erythraea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyperthermus_butylicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinobacillus_pleuropneumoniae_5b_l20_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halorubrum_lacusprofundi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermotoga_lettingae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactococcus_lactis_cremoris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylobacterium_extorquens_pa1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanococcus_aeolicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_tuberculosis_h37ra_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacillus_thermodenitrificans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanobrevibacter_smithii_atcc35061_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microvirga_lupini_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylibium_petroleiphilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactococcus_garvieae_lg2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_rudis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_epilithonimonas_lactis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_luteum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_sediminis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_sp_mit016451_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiomonas_arsenitoxydans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylobacterium_sp_446_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylobacterium_radiotolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanococcus_maripaludis_c7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_rhamnosiphila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_leptum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_spiroforme_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_dolichum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ammonifex_degensii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dactylellina_haptotyla_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_baudoinia_compniacensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_kluyveri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_gingivalis_atcc33277_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kazachstania_africana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neorickettsia_risticii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_vulgatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_parabacteroides_distasonis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_graevenitzii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neisseria_mucosa_c102_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_hominis_c80_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_capitis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sp_c150_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_idiomarina_salinarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfurihydrogenibium_sp_yo3aop1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptacidiphilus_jeojiense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrosopumilus_maritimus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_oleivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_azorhizobium_caulinodans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfatibacillum_alkenivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_selenitireducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ochrobactrum_anthropi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aciduliprofundum_boonei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelagibacter_sp_htcc7211_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodobacterales_bacterium_y4i_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruegeria_sp_r11_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_chol1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylobacterium_populi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acholeplasma_laidlawii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_coahuilensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacter_sp_m18_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacter_sp_m21_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinobacter_algicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amycolicicoccus_subflavus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinitoga_piezophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_clavuligerus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_cp53_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clavibacter_michiganensis_ncppb382_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrobaculum_neutrophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanococcus_maripaludis_c6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_botulinum_nctc2916_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_soli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alistipes_putredinis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerofustis_stercorihominis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerotruncus_colihominis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_intestinibacter_bartlettii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erysipelatoclostridium_ramosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_collinsella_stercoris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_borrelia_valaisiana_vs116_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinosynnema_mirum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brachybacterium_faecium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cellulomonas_flavigena_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardiopsis_dassonvillei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sanguibacter_keddieii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_stackebrandtia_nassauensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xylanimonas_cellulosilytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sorangium_cellulosum_so_ce56_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microcystis_aeruginosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_stercoris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_profundi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phenylobacterium_zucineum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amoebophilus_asiaticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_opitutus_terrae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_polynucleobacter_necessarius_stir1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kitasatospora_setae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rickettsia_rickettsii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobium_japonicum_ut26s_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_chlorophenolicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ignicoccus_hospitalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_arboricola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_glaciecola_sp_htcc2999_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_griseus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanococcus_voltae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanoregula_boonei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_7243faa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_sp_31syn3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_nucleatum_animalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_sp_5139bfaa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synergistes_sp_31syn1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiales_bacterium_1747faa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_fragilis_3112_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_albus_j1074_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_pristinaespiralis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natranaerobius_thermophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_macrococcus_caseolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_halifaxensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cloacimonas_acidaminovorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrospira_platensis_c1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylobacterium_nodulans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sviceus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_micrococcus_luteus_nctc_2665_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_mg1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erwinia_tasmaniensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelagibacter_ubique_htcc8051_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_ver3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_griseoflavus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodobacteraceae_bacterium_klh11_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_gordonii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pedobacter_oryzae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_kroppenstedtii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_escherichia_coli_bl21_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermobispora_bispora_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cryptobacterium_curtum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dethiosulfovibrio_peptidovorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halogeometricum_borinquense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_conexibacter_woesei_dsm14684_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_citrobacter_sp_302_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coprobacillus_sp_291_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_nucleatum_vincentii_3136a2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_nucleatum_vincentii_4113_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sp_2136faa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderiales_bacterium_1147_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_gonidiaformans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_mortiferum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_ulcerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_varium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_coprocola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cladophialophora_yegresii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacillus_sp_wch70_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermomonospora_curvata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_beutenbergia_cavernae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dyadobacter_fermentans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_slackia_heliotrinireducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_jonesia_denitrificans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharomonospora_viridis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_intestinalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_providencia_stuartii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_lactaris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_proteus_penneri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitratireductor_basaltis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrosococcus_halophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_colletotrichum_gloeosporioides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blautia_hydrogenotrophica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_achromobacter_arsenitoxydans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_stutzeri_ts44_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_modestobacter_marinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulforudis_audaxviator_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_verrucomicrobia_bacterium_lp2a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marvinbryantia_formatexigens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kytococcus_sedentarius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nakamurella_multipartita_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptosporangium_roseum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_catenulispora_acidiphila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphaerobacter_thermophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kribbella_flavida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_veillonella_parvula_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eggerthella_lenta_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylacidiphilum_infernorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_escherichia_coli_atcc8739_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_canis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_galeopterus_variegatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_finegoldii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_eggerthii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_pectinophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_myxococcus_fulvus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_plebeius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermosipho_africanus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_francisella_philomiragia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelosinus_sp_ufo1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ktedonobacter_racemifer_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomicrobium_mukohataei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfohalobium_retbaense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfotomaculum_acetoxidans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pedobacter_heparinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chitinophaga_pinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_soli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_intracellulare_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacteria_bacterium_ms0242a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_puniceispirillum_marinum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neisseria_lactamica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_moorea_producens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfurococcus_kamchatkensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_hispalense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anoxybacillus_flavithermus_wk1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_etli_ciat652_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinomonas_posidonica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_alamii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_glaciecola_arctica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychrobacter_sp_tb15_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_tb23_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_fermentans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thauera_sp_63_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chthoniobacter_flavus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyanothece_sp_pcc7822_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cellvibrio_japonicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_borrelia_spielmanii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_heliobacterium_modesticaldum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermus_aquaticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_avicenniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tyzzerella_nexilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_hiranonis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mitsuokella_multacida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_providencia_rustigianii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_citrobacter_youngae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_citreicella_sp_se45_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haliangium_ochraceum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_escherichia_albertii_tw07627_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eggerthella_sp_yy7918_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spirosoma_linguale_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_urealyticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_jejuia_pallidilutea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_meiothermus_ruber_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oligotropha_carboxidovorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rheinheimera_sp_a13l_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caulobacter_segnis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acetivibrio_cellulolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pedobacter_sp_v48_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldisericum_exile_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanimonas_sp_gk1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_escherichia_coli_k12_mg1655_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_buchneri_nrrlb30929_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_crossotus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_azobacteroides_pseudotrichonymphae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_crocodyli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinoplanes_missouriensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_riesia_pediculicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_eligens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_proteoclasticus_b316_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dictyoglomus_turgidum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlorobaculum_parvum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chloroherpeton_thalassium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pantoea_sp_ab_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_citromicrobium_sp_jlt1363_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_angulatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_holdemanella_biformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodothermus_marinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptobacillus_moniliformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halorhabdus_utahensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ectothiorhodospira_sp_phs1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_niph973_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_providencia_alcalifaciens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_providencia_rettgeri_dsm1131_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_collinsella_intestinalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_borrelia_bissettii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanosphaerula_palustris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kosmotoga_olearia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_atopobium_parvulum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tsukamurella_paurometabola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capnocytophaga_ochracea_dsm7271_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alicyclobacillus_acidocaldarius_dsm446_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_timonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldicellulosiruptor_bescii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_planctomyces_limnophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_caeni_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_accumulibacter_phosphatis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_stenotrophomonas_maltophilia_k279a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_denitrovibrio_acetiphilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kangiella_koreensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptotrichia_buccalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloferax_mediterranei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanothermococcus_thermolithotrophicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermococcus_onnurineus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_desulfuricans_atcc27774_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_atcc27244_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_coleocanis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_urogenitalis_dsm15434_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerococcus_lactolyticus_atcc51172_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerococcus_tetradius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_atopobium_vaginae_dsm15829_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_gleum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_accolens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_lipophiloflavum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_pseudogenitalium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_striatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_finegoldia_magna_atcc53516_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_antri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_buchneri_atcc11577_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_ultunensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_listeria_grayi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_parascrofulaceum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobacterium_spiritivorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_caprae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_equinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfomicrobium_baculatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfurospirillum_deleyianum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermanaerovibrio_acidaminovorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermobaculum_terrenum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidimicrobium_ferrooxidans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerococcus_prevotii_dsm20548_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sebaldella_termitidis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_salexigens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brachyspira_murdochii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geodermatophilus_obscurus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_bronchialis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_meiothermus_silvanus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_yersinia_aldovae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_proteus_mirabilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrococcus_yayanosii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thecamonas_trahens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodoluna_lacicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pirellula_staleyi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacteriaceae_bacterium_351910_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cylindrospermopsis_raciborskii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_raphidiopsis_brookii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidovorax_ebreus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_opportunistum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_carboxidivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_botulinum_a2_kyoto_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_botulinum_e1_bont_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blautia_hansenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_copri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_methylpentosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_liberibacter_asiaticus_psy62_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_canadensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_cinaedi_ccug18818_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_pullorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloterrigena_turkmenica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinoplanes_subtropicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_variovorax_paradoxus_s110_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_casei_bl23_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_beta_proteobacterium_cb_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_arbusti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_akl11_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_alj24_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_megaterium_qm_b1551_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_primitia_zas2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_azotonutricium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_holdemania_filiformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_celatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neisseria_cinerea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neisseria_flavescens_nrl30031h210_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neisseria_mucosa_atcc25996_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neisseria_polysaccharea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neisseria_subflava_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_filifactor_alocis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gemella_haemolysans_atcc10379_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_sputigena_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_veillonella_dispar_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eikenella_corrodens_atcc23834_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_periodonticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_deserti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_coprophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_pseudocatenulatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neisseria_sicca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hydrogenobaculum_sp_ho_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_vulneris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natrialba_magadii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_aurimucosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_glucuronolyticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mobiluncus_curtisii_atcc43063_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ferrimonas_balearica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nostoc_azollae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hirschia_maritima_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ponticaulis_koreensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erysipelotrichaceae_bacterium_5254faa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcaceae_bacterium_d16_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cryptococcus_gattii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_animalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dehalogenimonas_lykanthroporepellens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_bivia_jcvihmp010_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_melaninogenica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_endodontalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capnocytophaga_sputigena_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capnocytophaga_gingivalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_atopobium_rimae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_amycolatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_matruchotii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enhydrobacter_aerosaccus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_rectus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_showae_rm3277_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_gracilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cobetia_crustatorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_hylemonae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlorella_variabilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermosediminibacter_oceani_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dethiobacter_alkaliphilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_zymomonas_mobilis_atcc10988_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_galbibacter_marinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halothiobacillus_neapolitanus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfonatronospira_thiodismutans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_novosphingobium_sp_rr_217_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_d5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_necrophorum_d12_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_winghamensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oxalobacter_formigenes_hoxbls_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oxalobacter_formigenes_oxcc13_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_reuteri_dsm20016_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_laribacter_hongkongensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_kansasii_atcc12478_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_piperi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lentibacillus_jeotgali_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_doosanense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aeromonas_sp_ae122_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomadura_flavalba_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerococcus_hydrogenalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_gallicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dickeya_zeae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pectobacterium_carotovorum_pc1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pectobacterium_wasabiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amphibacillus_jilinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kyrpidia_tusciae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_viscosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gemella_haemolysans_m341_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gemella_morbillorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gemella_sanguinis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_oris_c735_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_sp_c561_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sp_m143_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sp_m334_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfurimonas_autotrophica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bilophila_wadsworthia_316_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geoglobus_acetivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brachyspira_hyodysenteriae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_luminiphilus_syltensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ureaplasma_urealyticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_gallinarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_casseliflavus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_faecium_c68_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_ghanaensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gammaproteobacteria_bacterium_nor53_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_catenulatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_metarhizium_robertsii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_rhamnosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bordetella_pertussis_18323_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dickeya_sp_ncppb569_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidaminococcus_intestini_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serratia_symbiotica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardiopsis_potens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_wolbachia_sp_culex_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fodinicurvata_sediminis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fodinicurvata_fenggangensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sedimentitalea_nanhaiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hamiltonella_defensa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_allochromatium_vinosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vulcanisaeta_distributa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halanaerobium_praevalens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arcobacter_nitrofigilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ilyobacter_polytropus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_archaeoglobus_profundus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aminobacterium_colombiense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_cellulovorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanocaldococcus_infernus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanocaldococcus_fervens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_asticcacaulis_excentricus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_magneticus_rs1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spirochaeta_smaragdinae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_francisella_sp_tx077308_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acetohalobium_arabaticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_gaemokensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_manliponensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaplasma_centrale_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_zhanjiangensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_isosphaera_pallida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_ruh2624_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_johnsonii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_lwoffii_sh145_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_radioresistens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroidetes_f0058_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_f0167_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_coleohominis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_jensenii_1153chn_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_jensenii_272chn_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptoniphilus_sp_f0131_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_sp_f0108_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_tasmaniensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanocaldococcus_vulcanius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_zymomonas_mobilis_pomaceae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dickeya_dadantii_ech703_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermoanaerobacterium_thermosaccharolyticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermoanaerobacter_italicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sideroxydans_lithotrophicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermovirga_lienii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hirschia_baltica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rubidibacter_lacunae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylovorus_glucosetrophus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomicrobium_denitrificans_atcc51888_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylotenera_mobilis_jlw8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coraliomargarita_akajimensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aminomonas_paucivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mobiluncus_curtisii_atcc35241_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mobiluncus_mulieris_atcc35243_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_mitis_sk321_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_mitis_sk564_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_mitis_sk597_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseburia_hominis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_kordi49_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_kordi52_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oribacterium_sinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_bergensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_noxia_atcc43541_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_weissella_paramesenteroides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_amylolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_genitalium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevibacterium_mcbrellneri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aeromicrobium_marinum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_sp_d20_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanobacillus_manasiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_terribacillus_aidingensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_chlororaphis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_papyrosolvens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizophagus_irregularis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kazachstania_naganishii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevundimonas_naejangsanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfurivibrio_alkaliphilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alteromonas_australica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ferroglobus_placidus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dickeya_dadantii_ech586_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cellulomonas_fimi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidaminococcus_fermentans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylothermus_hellenicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinobacillus_minor_202_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_spb78_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_aa4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_viridochromogenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_pratensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_abiotrophia_defectiva_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerobaculum_hydrogeniformans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_catonella_morbi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_botulinum_d_1873_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dialister_invisus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nonlabens_dokdonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_saphenum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pantoea_sp_at9b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pantoea_sp_sc1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermococcus_gammatolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanoregula_formicica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cellvibrio_gilvus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodopirellula_sp_swk7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tolumonas_auensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylosinus_trichosporium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_variovorax_paradoxus_eps_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_a3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_fructosivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_sp_u5l_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alicycliphilus_denitrificans_bc_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alicycliphilus_denitrificans_k601_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptostreptococcus_stomatis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_warneri_l37603_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neisseria_flavescens_sk114_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptotrichia_goodfellowii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_vincentii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_uenonis_603_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mobiluncus_mulieris_281_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptostreptococcus_anaerobius_653l_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptoniphilus_lacrimalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brenneria_sp_enid312_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nautilia_profundicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blattabacterium_sp_bplan_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_meiothermus_rufus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermococcus_sibiricus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldicellulosiruptor_obsidiansis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oribacterium_sp_f0262_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hydrogenobacter_thermophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_saccharolyticum_wm1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_bilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodoferax_saidenbachensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arenibacter_algicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_splendidus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_sp_f0295_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aquimarina_agarilytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_d14_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseburia_inulinivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylocystis_sp_atcc49242_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_granulicatella_elegans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_glumae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_tannerae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shuttleworthia_satelles_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinobacter_nanhaiticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phascolarctobacterium_succinatutens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobium_sp_syk6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_syringae_maculicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_hominis_sk119_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_sp_mn802worker_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shimwellia_blattae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiorhodovibrio_sp_970_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lutibaculum_baratangense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_butyricum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldicellulosiruptor_hydrothermalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldicellulosiruptor_kristjanssonii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldicellulosiruptor_owensensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thalassobium_sp_r2a62_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_olsenella_uli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermosphaera_aggregans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevundimonas_subvibrioides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_cellulosolvens_6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aggregatibacter_aphrophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_komagataeibacter_medellinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acetobacter_pasteurianus_ifo328301_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloarcula_hispanica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanobrevibacter_ruminantium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erwinia_pyrifoliae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erwinia_billingiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_grahamii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacillus_thermoglucosidasius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptotrichia_hofstadii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermincola_potens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidithiobacillus_caldus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidithiobacillus_thiooxidans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_violacea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_citrobacter_rodentium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinobacillus_minor_nm305_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_granulicatella_adiacens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_flueggei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermocrinis_albus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidobacteria_bacterium_kbs146_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deferribacter_desulfuricans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_starkeya_novella_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dechlorosoma_suillum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_segniliparus_rotundus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_sp_ccge1001_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_sp_ccge1002_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_sp_ccge1003_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterobacter_asburiae_lf7a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_dlviii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_flavefaciens_fd1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capnocytophaga_granulosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_scardovia_inopinata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_simonsiella_muelleri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neisseria_sp_f0314_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_desulfuricans_nd132_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyclobacterium_qasimii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_winogradskyella_psychrotolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tatumella_morbirosei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_lentocellum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microchaete_sp_pcc7126_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_aespoeensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_syntrophothermus_lipocalidus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marivirga_tractuosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_silicibacter_sp_trichch4b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_silicibacter_lacuscaerulensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanocaldococcus_sp_fs40622_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfarculus_baarsii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_micromonospora_aurantiaca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arcanobacterium_haemolyticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_neofelifaecis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_stutzeri_rch2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermaerobacter_marianensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_sp_fw1012b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_kroppenstedtii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_e14_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_jonquetella_anthropi_e333e1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_syntrophobotulus_glycolicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfosporosinus_acidiphilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanothermococcus_okinawensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodomicrobium_vannielii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylobacterium_sp_mb200_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermovibrio_ammonificans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leadbetterella_byssophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_truepera_radiovictrix_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_cellulosilyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_sp_f0332_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aneurinibacillus_aneurinilyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_ammoniagenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_veroralis_f0319_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_slackia_exigua_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinoplanes_sp_n902109_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erysipelothrix_rhusiopathiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodopseudomonas_palustris_dx1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_violaceusniger_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_sp_f0330_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfurispirillum_indicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_albugo_laibachii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerococcus_vaginalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aerococcus_viridans_atcc11563_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_oralis_atcc35037_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_zunongwangia_profunda_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudogymnoascus_destructans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_frankia_symbiont_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halanaerobium_hydrogeniformans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_xylanisolvens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_faecalibacterium_prausnitzii_sl33_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_3157faact1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_9143bfaa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_liberibacter_solanacearum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_legionella_drancourtii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_h2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_1456faa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erysipelotrichaceae_bacterium_3153_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesotoga_prima_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_olsenella_sp_f0356_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_legionella_longbeachae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fimbriimonas_ginsengisoli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloferax_mucosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_resistens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ethanoligenens_harbinense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_regiella_insecticola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylocapsa_aurea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acetobacter_aceti_atcc23746_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_dysgalactiae_atcc27957_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erwinia_amylovora_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spirochaeta_thermophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_viridosporus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_sp_6146afaa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_3146faa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_smithii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_subdoligranulum_sp_4354a2faa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_2a57ct2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_planktomarina_temperata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidilobus_saccharovorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylotenera_versatilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_afipia_sp_1nls2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodanobacter_denitrificans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_1nla3e_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermodesulfatator_indicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_salanitronis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_edwardsiella_tarda_atcc15947_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_sp_jpy347_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mannheimia_haemolytica_bovine_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_profftella_armatura_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microvirga_aerilata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomicrobium_denitrificans_1nes1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanithermus_profundus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_metallosphaera_yellowstonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylomirabilis_oxyfera_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aciduliprofundum_sp_mar08339_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_babela_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_alginolyticus_40b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudonocardia_dioxanivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_mimicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_grimontia_hollisae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_metschnikovii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_coralliilyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_sp_rc586_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_orientalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_photobacterium_damselae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_francisella_tularensis_novicida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_timonensis_cris5cb1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_buccalis_atcc35310_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_amnii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bulleidia_extructa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_segniliparus_rugosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alloprevotella_rava_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_johnsonella_ignava_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_infelix_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_mustelae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanoplanus_petrolearius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alistipes_finegoldii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_coprosuis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_scabiei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rothia_mucilaginosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_granulicella_mallensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_jejuni_414_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_listeria_seeligeri_12b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alpha_proteobacterium_himb114_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_sp_2009_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_citromicrobium_bathyomarinum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodococcus_equi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_sp_pr090111t3t6a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_rochalimae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylomicrobium_album_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_s9_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_comamonas_testosteroni_cnb2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_multisaccharivorax_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermotoga_thermarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cellulophaga_algicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oribacterium_sp_nk2b42_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_vignae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_fluorescens_nz007_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_africanus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fonticula_alba_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterobacteriaceae_bacterium_strain_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_archaeoglobus_veneficus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oscillibacter_valericigenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_proteolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_helcogenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylobacterium_oryzae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paludibacter_propionicigenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrolobus_fumarii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natronococcus_occultus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfurella_acetivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanomicrobium_mobile_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aggregatibacter_actinomycetemcomitans_d7s1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_clarridgeiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfotomaculum_ruminis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfotomaculum_nigrificans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrospira_platensis_nies39_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinobacillus_suis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mahella_australiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylobacter_tundripaludum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_larvae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermoanaerobacter_wiegelii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_albus_7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marssonina_brunnea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_lugdunensis_hku0901_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrobaculum_oguniense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amphibacillus_xylanus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_tropici_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_virgibacillus_alimentarius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gardnerella_vaginalis_6119v5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_diphtheriae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_megasphaera_genomosp_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiales_genomosp_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_ratti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coriobacterium_glomerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_neoaurum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_niastella_koreensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_sp_n418_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterobacter_lignolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pediococcus_claussenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_novosphingobium_sp_pp1y_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_noxia_f0398_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_oulorum_f0390_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_turicibacter_sanguinis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_bifidum_prl2010_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pantoea_ananatis_lmg20103_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_solobacterium_moorei_f0204_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_megasphaera_micronuciformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capnocytophaga_sp_f0087_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_anginosus_f0211_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_sp_f0337_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfomonile_tiedjei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_gallisepticum_f_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfuricurvum_kujiense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobiaceae_bacterium_sg6c_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_maricopensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_odoribacter_splanchnicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_frankia_sp_qa3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_colletotrichum_fioriniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_suis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_chubuense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_rhodesiae_nbb3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_smegmatis_js623_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_tusciae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_intrasporangium_calvum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_gxt6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pantoea_vagans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_thiocyanodenitrificans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_thiocyanoxidans_arh4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_iners_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leucobacter_chromiiresistens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_sanfranciscensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mucilaginibacter_paludis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lysinibacillus_fusiformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_asticcacaulis_biprosthecum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alteromonas_sp_sn2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterobacter_cloacae_atcc13047_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_waddlia_chondrophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ensifer_sojae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flexistipes_sinusarabici_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermobacillus_composti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_curdlanolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalimicrobium_aerophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalimicrobium_cyclicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinomonas_mediterranea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomicrobium_sp_mc1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_faecalibacterium_prausnitzii_l26_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_clariflavum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_atrophaeus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoalteromonas_haloplanktis_ant505_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_pneumoniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_idiomarina_xiamenensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rahnella_sp_y9602_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_achromobacter_piechaudii_atcc43553_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_collinsella_sp_4847faa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_2146faa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alistipes_indistinctus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_barnesiella_intestinihominis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_oleiciplenus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_citroniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_clostridioforme_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_orbiscindens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_symbiosum_wal14163_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_symbiosum_wal14673_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_collinsella_tanakaei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dialister_succinatiphilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dorea_formicigenerans_4653afaa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dysgonomonas_gadei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dysgonomonas_mossii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_odoribacter_laneus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_slackia_piriformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sutterella_wadsworthensis_3145b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sutterella_wadsworthensis_2159bfaa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidithiobacillus_ferrivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermus_scotoductus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_isoptericola_variabilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_lactis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_fulva_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoxanthomonas_suwonensis_111_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobacterium_sp_21_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylocystis_sp_sb2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_putrefaciens_ks1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_bovoculi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_moraxella_bovoculi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_caldaria_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ahrensia_sp_r2a130_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseibium_sp_trichskd4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alpha_proteobacterium_himb59_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gammaproteobacteria_bacterium_himb55_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rahnella_aquatilis_cip7865_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_sp_mm1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gallaecimonas_xiamenensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_olleya_sp_vcsm12_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_gobiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aspergillus_fumigatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aequorivita_sublithincola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermodesulfobium_narugense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_alligatoris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_faecalibacterium_prausnitzii_kle1255_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_azoarcus_sp_kh32c_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudogulbenkiania_sp_nh8b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halobacteroides_halobius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_alsr1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_crispatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_ljungdahlii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitratifractor_salsuginis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_bingchenggensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amycolatopsis_mediterranei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halobacterium_sp_dl1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermus_oshimai_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gammaproteobacteria_bacterium_himb30_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_phagedenis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_australicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_propionibacterium_freudenreichii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_escherichia_sp_tw09308_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_photobacterium_aphoticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylophaga_nitratireducenticrescens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylophaga_frappieri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyanobacterium_aponinum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_bnl1100_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fluviicola_taffensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microcoleus_vaginatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_planctomyces_brasiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfitobacterium_dehalogenans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloferacales_archaeon_dl31_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_herbaspirillum_seropedicae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ketogulonicigenium_vulgare_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_dysgalactiae_equisimilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brachyspira_pilosicoli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphaerochaeta_coccoides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_massilia_consociata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hippea_maritima_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfurospirillum_barnesii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haliscomenobacter_hydrossis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfotomaculum_kuznetsovii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_runella_slithyformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leuconostoc_kimchii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_stellenboschense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_achromobacter_xylosoxidans_a8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leuconostoc_gasicomitatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pedobacter_saltans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rothia_dentocariosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_sp_f0386_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_parasutterella_excrementihominis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paraprevotella_clara_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paraprevotella_xylaniphila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_succinatimonas_hippei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_clarus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_fluxus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_urinalis_228597_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_macacae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_ictaluri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_propionibacterium_acnes_hl037pa2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfurococcus_mucosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oscillochloris_trichoides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_ciceri_biovar_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bdellovibrio_bacteriovorus_w_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marichromatium_purpuratum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiocystis_violascens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioflavicoccus_mobilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiorhodococcus_drewsii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiorhodospira_sibirica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_parachlamydia_acanthamoebae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_citreicella_sp_357_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_propionibacterium_propionicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_denticola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_frateuria_aurantia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfotomaculum_gibsoniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_elongata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_hirae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_calditerrivibrio_nitroreducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiocapsa_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfurococcus_fermentans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermoproteus_tenax_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfosporosinus_meridiei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfosporosinus_orientis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfosporosinus_youngiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_mitis_2_f0392_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fervidobacterium_pennivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sinorhizobium_sp_ccbau05631_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_opitutaceae_bacterium_tav5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermodesulfobacterium_geofontis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ralstonia_sp_pba_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halalkalicoccus_jeotgali_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nesterenkonia_sp_f_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_methanolicus_mga3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_caribbeanicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptostreptococcaceae_bacterium_cm5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_stomatobaculum_longum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oribacterium_parvum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halosimplex_carlsbadense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haladaptatus_paucihalophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halopiger_xanaduensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halostagnicola_larsenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halovivax_ruber_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natrinema_pellirubrum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natronobacterium_gregoryi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_parafarraginis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spiroplasma_mirum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_micavibrio_aeruginosavorus_arl13_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylomonas_methanica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_scardovia_wiggsiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caloramator_australicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_moraxella_catarrhalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermoanaerobacterium_xylanolyticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_variabile_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_haemofelis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alpha_proteobacterium_himb5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ralstonia_solanacearum_psi07_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capnocytophaga_canimorsus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_agrobacterium_sp_h133_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gemmatimonadetes_bacterium_kbs708_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_arilaitensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaeroglobus_geminatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_sp_f0437_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_f0431_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sp_f0407_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_sp_al1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pediococcus_acidilactici_dsm20284_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_marshii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptoniphilus_duerdenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_sirexaae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteriovorax_marinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haemophilus_parainfluenzae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_intermedius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_constellatus_pharyngis_c1050_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_anginosus_c1051_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_anginosus_c238_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_nuruki_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_hortorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderiales_bacterium_joshi001_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microvirga_lotononidis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_herbaspirillum_frisingense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_sp_67h29bp_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_parascardovia_denticolens_dsm10105_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_yurii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_mitis_atcc6249_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sp_73h25ap_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oscillatoriales_cyanobacterium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_ultunense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gillisia_limnaea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_weeksella_virosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_belliella_baltica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_schizosaccharomyces_cryophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_disiens_fb03509an_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_atopobium_vaginae_pb189t14_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aerococcus_urinae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_veillonella_atypica_acs049vsch6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halobacillus_halophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chloroflexus_sp_y3961_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cellulophaga_lytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ornithobacterium_rhinotracheale_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermaerobacter_subterraneus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanobacterium_paludis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfotomaculum_carboxydivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfurobacterium_thermolithotrophum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_succinifaciens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinithermus_hydrothermalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cytophaga_fermentans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiothrix_nivea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_scophthalmi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_calcoaceticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfitobacterium_dichloroeliminans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfitobacterium_metallireducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_parauberis_ncfd2020_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_porcinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_criceti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_buccae_atcc33574_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capnocytophaga_ochracea_f0287_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_oralis_atcc33269_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_sp_jdm601_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptoniphilus_rhinitidis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gammaproteobacteria_bacterium_imcc3088_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methyloferula_stellata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_sp_nk3a76_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiales_bacterium_nk3b98_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erysipelotrichaceae_bacterium_nk3d112_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_bryantii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_nk4a136_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_nk4a144_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_nk4a179_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanobacterium_lacus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfobacter_postgatei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_asaccharolytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerococcus_prevotii_acs065vcol13_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptoniphilus_sp_f0436_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_veillonella_sp_f0412_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_sp_f0430_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyclobacterium_marinum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flexibacter_litoralis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfobacca_acetoxidans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldithrix_abyssi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_barnesiella_viscericola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_leachii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rikenella_microfusus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_listeria_ivanovii_pam55_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharomonospora_cyanea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharomonospora_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharomonospora_xinjiangensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_rhizoxinica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinobaculum_massiliae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinobaculum_schaalii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_europaeus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_turicensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_afipia_broomeae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_afipia_felis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alloiococcus_otitis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bergeyella_zoohelcum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dolosigranulum_pigrum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_infirmum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_facklamia_hominis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_facklamia_ignava_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_facklamia_languida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helcococcus_kunzii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_massilia_timonae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_veillonella_ratti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_micans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_urinalis_fb127cna2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_turicella_otitidis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_jonquetella_anthropi_dsm22815_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fukomys_damarensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_singulisphaera_acidiphila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_muricauda_ruestringensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerophaga_thermohalophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_polymyxa_sc2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hylemonella_gracilis_atcc19624_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnoanaerobaculum_saburreum_dsm3986_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kingella_kingae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lautropia_mirabilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoramibacter_alactolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_parasanguinis_atcc903_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_oralis_atcc49296_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_cardiffensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_sp_f0310_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptotrichia_wadei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_sp_f0400_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capnocytophaga_sp_f0234_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_centipeda_periodontii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dialister_micraerophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_italicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_neuii_bvs029a5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_sulci_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_multiformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_peroris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sanguinis_sk49_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sanguinis_sk355_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sanguinis_sk1057_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_salivae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_australis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_cristatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_infantis_atcc700779_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spirochaeta_africana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_acidophilus_30sc_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerobaculum_mobile_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_plautia_stali_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_moranella_endobia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_limosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_holophaga_foetida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gardnerella_vaginalis_101_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_downei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sp_f0418_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oribacterium_sp_f0425_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_vestibularis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_pettenkoferi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nonlabens_ulvanivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_brennaborense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_pylori_southafrica7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_saccharophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leuconostoc_fallax_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eremococcus_coleocola_acs139vcol8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptoniphilus_harei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_oris_pb013t23_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_hgf2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alistipes_sp_hgb5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_dentalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinokineospora_sp_eg49_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_syntrophorhabdus_aromaticivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alphaproteobacterium_sar116_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_pseudoporcinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dialister_microaerophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ewingella_americana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leclercia_adecarboxylata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudovibrio_sp_fobeg1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_weissella_cibaria_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_cf149_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lysobacter_arseniciresistens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_coryniformis_kctc3167_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfosporosinus_sp_ot_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sediminibacterium_sp_or53_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_vesicatoria_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_adhaeribacter_aquaticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldilinea_aerophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capnocytophaga_cynodegmi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_pimensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_echinicola_vietnamensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_joostella_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_meiothermus_chliarophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_orenia_marismortui_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_owenweeksia_hongkongensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_promicromonospora_kroppenstedtii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_terriglobus_roseus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermanaerovibrio_velox_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerolinea_thermophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloplanus_natans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halonatronum_saccharophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alkaliflexus_imshenetskii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_oralis_uo5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechocystis_sp_pcc7509_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leuconostoc_gelidum_kctc3527_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_ruminantium_lactilytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_sp_r4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharomonospora_glauca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_botulinum_bkt015925_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_solitalea_canadensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfurimonas_gotlandica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_emticicia_oligotrophica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flectobacillus_major_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_myroides_odoratus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_patulibacter_minatonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_niabella_soli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_senegalensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_brassicacearum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alcanivorax_dieselolei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_phenanthrenivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfolobus_islandicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fructobacillus_fructosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_saccharoperbutylacetonicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halococcus_morrhuae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acetobacterium_woodii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_rhodesiae_js60_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serratia_sp_m24t3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pantoea_ananatis_aj13355_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermocrinis_ruber_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_planococcus_donghaensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfosarcina_sp_bus5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidianus_hospitalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aminobacter_sp_j41_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_loti_r88b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sinorhizobium_arboris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracoccus_pantotrophus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoxanthomonas_suwonensis_j43_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_j33_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_j37_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cellulosimicrobium_cellulans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chelativorans_sp_j32_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_j14_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracoccus_sp_j39_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_luteimonas_sp_j29_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardioides_sp_j54_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldanaerobius_polysaccharolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizopus_delemar_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_leguminosarum_248_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_farciminis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_parauberis_kctc11537_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_felis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mogibacterium_sp_cm50_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_genosp_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_sp_icm47_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_atopobium_sp_bs2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_sp_fobrc6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_sp_fobrc9_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shuttleworthia_sp_msx8b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sp_cm6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_veillonella_sp_as16_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnoanaerobaculum_sp_msx33_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_taylorella_equigenitalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_peraridilitoris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiales_bacterium_9400853_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiales_bacterium_9401234_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiales_bacterium_9403326_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerococcus_sp_9402080_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroidetes_bacterium_scgc_aaa027n21_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acetobacter_pasteurianus_3p3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_sp_x2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_coagulans_26_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_fructivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermobrachium_celere_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_sp_wsm2230_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hippea_jasoniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hippea_alviniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hippea_sp_km1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arcobacter_butzleri_ed1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arcobacter_sp_l_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_sp_f0384_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_oris_f0423_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_veillonella_sp_f0422_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_parvimonas_sp_f0440_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tetragenococcus_halophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_brasiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_sinaloensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_ulcerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ignavibacterium_album_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_imtechella_halotolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanobacillus_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salpingoeca_rosetta_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_symbiobacter_mobilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_dilworthii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_avibacterium_paragallinarum_72_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_venezuelae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cupriavidus_taiwanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_testaceum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacteriaceae_bacterium_hqm9_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_morus_notabilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_nctc_7422_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_lwoffii_nctc_5866_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_ursingii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptacidiphilus_rugosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ornithinibacillus_scapharcae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruegeria_conchae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_fetus_venerealis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lacinutrix_sp_5h374_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_glaciecola_sp_4h37ye5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dokdonia_sp_4h375_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rubrivivax_gelatinosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_novosphingobium_nitrogenifigens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saprospira_grandis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_pseudintermedius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_heterobasidion_irregulare_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vulcanisaeta_moutnovskia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruegeria_halocynthiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gillisia_sp_cal575_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_terrae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_agnetis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_odyssella_thessalonicensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldalkalibacillus_thermarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rubrivivax_benzoatilyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lebetimonas_sp_js170_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_grahamii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_polymorphum_gilvum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_riemerella_anatipestifer_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharopolyspora_spinosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_youngiibacter_fragilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_suis_r61_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_griseoaurantiacus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_methanolicus_pb1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desmospora_sp_8437_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptoniphilus_indolicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_nigrescens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_pallens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_sp_kle1280_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_infantis_x_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_nordii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aeromonas_veronii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leucothrix_mucor_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_sp_td01_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_colicanis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_innocuum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eggerthia_catenaformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_parabacteroides_johnsonii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_sp_f0473_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sp_f0442_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_gladioli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leisingera_daeponensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leisingera_caerulea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudophaeobacter_arcticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leisingera_aquimarina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermoproteus_uzoniensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methyloversatilis_universalis_fam5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_megasphaera_sp_upii135e_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_anginosus_sk52_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_mitis_2_sk95_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cryobacterium_roopkundense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_photobacterium_leiognathi_mandapamensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_mendocina_nk01_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychrobacter_sp_1501_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leisingera_sp_ang1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_stercorea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelagibacter_sp_imcc9063_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_bizzozeronii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_solibacillus_silvestris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_cattleya_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_achromobacter_insuavis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesoflavibacter_zeaxanth_s86_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alteromonas_macleodii_black_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_collimonas_fungivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_buchnera_aphidicola_ua_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gallibacterium_anatis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_buchnera_aphidicola_ak_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_putida_csv86_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_oralis_sk255_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_infantis_sk1076_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ogataea_parapolymorpha_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_trabulsiella_guamensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tatumella_ptyseos_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leminorella_grimontii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kluyvera_ascorbata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_buttiauxella_agrestis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_metallosphaera_cuprina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_gallisepticum_s6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oscillibacter_ruminantium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_elgii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_sp_s17_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pusillimonas_sp_t77_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_mitis_sk1080_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_myroides_injenensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_taylorella_asinigenitalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acetonema_longum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_buchnera_aphidicola_g002_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylophaga_aminisulfidivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nannospalax_galili_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_endozoicomonas_montiporae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sporosarcina_newyorkensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_alkanivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_listeria_monocytogenes_scott_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterobacter_aerogenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neorhizobium_galegae_orientalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neorhizobium_galegae_officinalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haemophilus_haemolyticus_m19501_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haemophilus_haemolyticus_m21621_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haemophilus_haemolyticus_m21639_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthromitus_sp_sfbmousejapan_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_p838_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kocuria_rhizophila_p74_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_sp_kc8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caloramator_sp_ald01_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ralstonia_solanacearum_po82_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microlunatus_phosphovorus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aeromicrobium_massiliense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alistipes_senegalensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerococcus_senegalensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_timonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevibacterium_senegalense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_senegalense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kurthia_sp_dielmo_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kurthia_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kurthia_sp_jc8e_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_senegalensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptoniphilus_senegalensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salinisphaera_shabanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halorhabdus_tiamatea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloplasma_contractile_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_kefiranofaciens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_leguminosarum_cb782_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_senegalimassilia_anaerobia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_massiliosenegalensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_hw567_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_branchiophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_anatis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_lugdunensis_n920143_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_legionella_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_constellatus_pharyngis_sk1060_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_mitis_sk569_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_infantis_sk970_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevundimonas_diminuta_4704_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capnocytophaga_sp_f0382_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_durum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptostreptococcus_anaerobius_vpi4330_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_sp_f0040_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dehalobacter_sp_fth1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haemophilus_sputorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_idiomarina_sp_a28l_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_japonicum_usda6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_columbinum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_azorhizobium_doebereinerae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_elkanii_wsm1741_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_elkanii_wsm2783_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_japonicum_usda124_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_wsm2793_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_wsm3983_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_mimosarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_fluorescens_q287_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_loti_cj3sym_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_loti_usda3471_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_sp_wsm3224_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_sp_wsm3626_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_arr65_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_gallicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_giardinii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_leguminosarum_gb30_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_sullae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_leucaenae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ensifer_sp_wsm1721_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthromitus_sp_sfbratyit_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_colombiense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_wickerhamomyces_ciferrii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_columnare_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanocella_conradii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_sy8519_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevibacillus_laterosporus_lmg15441_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_fluorescens_hk44_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_sp_ibun_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinobacterium_stanieri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacteriaceae_bacterium_s85_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbulbifer_agarilyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_putida_s16_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermococcus_sp_4557_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serinicoccus_profundi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lysinimicrobium_mangrovi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oenococcus_kitaharae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_citricoccus_sp_ch26a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_weissella_koreensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoxanthomonas_spadix_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterobacter_cloacae_ecwsu1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brachyspira_intermedia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_lwoffii_wj10621_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bizionia_argentinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_salivarius_57i_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gayadomonas_joobiniege_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinobacter_lipolyticus_bf04cf4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aminicenantes_bacterium_aaa252a02_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_zymoseptoria_tritici_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sporichthya_polymorpha_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracoccidioides_sp_lutzii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_iowae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alicyclobacillus_acidocaldarius_tc41_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anditalea_andensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_endosymbiont_of_tevnia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_allobaculum_stercoricanis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinopolyspora_mortivallis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_propionibacterium_acnes_sk182bjcvi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_mojavensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_verticillium_alfalfae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfobacillus_acidophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_tubiashii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neisseria_weaveri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_polymyxa_m1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acetobacteraceae_bacterium_at5844_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermoplasmatales_archaeon_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_pseudopneumoniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_purpureus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychrobacter_sp_tb67_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldisphaera_lagunensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_grimontia_indica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_sp_bmg51109_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_teredinibacter_turnerae_t7902_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ensifer_sp_br816_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_megasphaera_elsdenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brachybacterium_paraconglomeratum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amycolatopsis_methanolica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amycolatopsis_nigrescens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_succinispira_mobilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_infantarius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_ruminis_atcc27782_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_glomeribacter_gigasporarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halonotius_sp_j07hn4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_nsp222_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloarchaeon_3a1dgr_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tetrapisispora_blattae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_buchneri_cd034_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_grimmiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_basilea_psittacipulmonis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cronobacter_condimenti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_curvatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brachybacterium_squillarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blattabacterium_sp_madar_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_amarae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blattabacterium_sp_cpu_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pantoea_rwandensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dietzia_alimentaria_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_faecis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_globiformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_effusa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_thermoresistibile_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_sp_hgb0015_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenisporosarcina_sp_hgh0030_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_mongolense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_chartreusis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_citrobacter_sp_s77_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_elodea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aquiluna_sp_imcc13023_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lonsdalea_quercina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelagibacterium_halotolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_agrobacterium_tumefaciens_ccnwgs0286_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_amorphae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_glaciecola_nitratireducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_frigoris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_exiguobacterium_antarcticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_peoriae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_novosphingobium_pentaromativorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_commensalibacter_intestini_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gluconobacter_morbifer_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fangia_hongkongensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aggregatibacter_actinomycetemcomitans_rhaa1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_sputi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mobilicoccus_pelagius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amycolatopsis_benzoatilytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amycolatopsis_balhimycina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinopolyspora_halophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rudanella_lutea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermicanus_aegyptius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloglycomyces_albus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salisaeta_longa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geminicoccus_roseus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodovibrio_salinarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermacetogenium_phaeum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_phyllosphaerae_52_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_phyllosphaerae_fa2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_melonis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_campestris_musacearum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_indicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_australis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_bovis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_schoenbuchensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_vinsonii_berkhoffii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermoanaerobacterium_saccharolyticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_alsatica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_doshiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_rattimassiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_melophagi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_tamiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_taylorii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_vinsonii_arupensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_washoensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fluoribacter_dumoffii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_sp_db56_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanolobus_psychrophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sp_sk140_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sp_sk643_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_mitis_sk579_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_oralis_sk1074_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haemophilus_paraphrohaemolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_necrophorum_funduliforme_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pasteurella_bettyae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnoanaerobaculum_saburreum_f0468_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_sp_f0472_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cellulomonas_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_herbaspirillum_massiliense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptoniphilus_timonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_timonella_senegalensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylobacterium_sp_gxf4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_pao19_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelagibacter_ubique_himb083_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_novosphingobium_lindaniclasticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_sp_yi23_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_limnohabitans_sp_rim28_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_limnohabitans_sp_rim47_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_man2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracoccus_sp_n5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylopila_sp_m107_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylobacterium_sp_10_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylobacterium_sp_77_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylophilaceae_bacterium_11_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrobaculum_sp_1860_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_faecium_nrrlb2354_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_sp_icm39_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_mste9_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rickettsia_australis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenirhodobacter_enshiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_enshiense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_rhizosphera_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_penicillium_rubens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_stutzeri_sdmlac_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tistrella_mobilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_asteroides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermus_sp_ccbus3uf1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_atopobium_sp_bv3ac4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_propionimicrobium_sp_bv2f7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptoniphilus_sp_bv3c26_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coriobacteriaceae_bacterium_bv3ac1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_megasphaera_sp_bv3c161_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alicyclobacillus_pomorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_budvicia_aquatica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_sputi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_antarcticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ignatzschineria_larvae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_polyisoprenivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychrobacter_sp_pamc21119_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_echinoides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_sp_pamc26605_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_sp_pamc26617_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_psychrotolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylophilus_sp_5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natronorubrum_tibetense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_citrobacter_farmeri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharomonospora_azurea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobium_baderi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_parasanguinis_fw213_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_fluorescens_f113_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_rossiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_escherichia_hermannii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_escherichia_vulneris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_31y_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_naeslundii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_macedonicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_acidiscabies_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hoeflea_sp_108_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_sp_ejy3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methyloglobulus_morosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_alvei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoalteromonas_citrea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoalteromonas_haloplanktis_atcc14393_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoalteromonas_rubra_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoalteromonas_spongiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_bataviensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_canis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_simiduia_agarivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sinorhizobium_fredii_hh103_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_extremaustralis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevibacillus_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerococcus_sp_ph9_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptoniphilus_grossensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_sp_ph3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kallipyga_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enorma_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_sp_gfaj1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_stenotrophomonas_maltophilia_pml168_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_hyorhinis_sk76_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingopyxis_sp_lc363_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacterium_sp_ms4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylopila_sp_73b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_sp_141_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acaricomes_phytoseiuli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acetobacter_nitrogenifigens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_bouvetii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinobacillus_capsulatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinobaculum_urinale_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinokineospora_enzanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomadura_atramentaria_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_dentalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_georgiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_israelii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_neuii_dsm8576_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_suimastitidis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_vaccimaxillae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomycetospora_chiangmaiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinoplanes_globisporus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinopolymorpha_alba_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aequorivita_capsosiphonis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aestuariibacter_salexigens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aestuariimicrobium_kwangyangense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_afifella_pfennigii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_agrococcus_lahaulensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_agromyces_italicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_agromyces_subbeticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_algicola_sagamiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_algoriphagus_mannitolivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_algoriphagus_marincola_dsm16067_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_algoriphagus_vanfongensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aliagarivorans_taiwanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alicyclobacillus_contaminans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alicyclobacillus_herbarius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alicyclobacillus_pohliae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alkanindiges_illinoisensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_allofustis_seminis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alloscardovia_omnicolens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alysiella_crassa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amorphus_coralli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaeroarcus_burkinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerobiospirillum_succiniciproducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerovorax_odorimutans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_andreprevotia_chitinilytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aquaspirillum_serpens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aquimarina_muelleri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arenibacter_certesii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arenibacter_latericius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arenimonas_composti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arenimonas_oryziterrae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arsenicicoccus_bolidensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_castelli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sanguinis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_asticcacaulis_benevestitus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_atopococcus_tabaci_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aureimonas_ureilytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_azospirillum_halopraeferens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_azovibrio_restrictus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_aidingensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_chagannorensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_fordii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_gelatini_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_barnesiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_graminisolvens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_massiliensis_b84634_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_pyogenes_dsm20611_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_salyersiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_balneola_vulgaris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bavariicoccus_seileri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_belnapia_moabensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blautia_wexlerae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brachymonas_chironomi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevibacillus_laterosporus_dsm25_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevundimonas_aveniformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevundimonas_bacteroides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_nodosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyricimonas_synergistica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_cuniculorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gemmobacter_nectariphilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_catelliglobosispora_koreensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_caeni_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_daeguense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_gregarium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_palustre_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiisalibacter_paucivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_aminophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_litorale_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_saccharogumia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sporosphaeroides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_stercorarium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_tyrobutyricum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_viride_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cohnella_laeviribosi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_conchiformibius_kuhniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_conchiformibius_steedae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_callunae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_capitovis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_caspium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_ciconiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_freiburgense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_halotolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_lubricantis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_massiliense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_mastitidis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_pilosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_propinquum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_ulceribovis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gryllotalpicola_ginsengisoli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cytophaga_aurantiaca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dasania_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_apachensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_aquatilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_frigens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_marmoris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_misasensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_demetria_terragena_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfobacter_curvatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfobulbus_japonicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfococcus_multivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfocurvus_vexinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfonatronovibrio_hydrogenovorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfotomaculum_alcoholivorax_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfotomaculum_alkaliphilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfotomaculum_hydrothermale_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfotomaculum_thermocisternum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_acrylicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_alcoholivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_alkalitolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_aminophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_bastinii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_desulfuricans_dsm642_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_frigidus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_gigas_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_hydrothermalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_longus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_oxyclinae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovirgula_thermocuniculi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfurispora_thermophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_donghicola_xiamenensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dyadobacter_alkalitolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_echinicola_pacifica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_elioraea_tepidiphila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_cecorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_columbae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterorhabdus_mucosicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_epilithonimonas_tenax_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eremococcus_coleocola_dsm15696_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_erysipelothrix_tonsillarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eudoraea_adriatica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ferrimicrobium_acidiphilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ferrimonas_futtsuensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_daejeonense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_filum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_frigidarium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_rivuli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_sasangense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_soli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_subsaxonicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_suncheonense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flexithrix_dorotheae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gelidibacter_mesophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gemella_cuniculi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geopsychrobacter_electrodiphilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geothrix_fermentans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gilvimarinus_chinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_glaciecola_punicea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_glaciibacter_superstes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_glycomyces_arizonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_hirsuta_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_shandongensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gracilibacillus_lacisalsi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gracilimonas_tropica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gramella_echinicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_granulicoccus_phenolivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gulosibacter_molinativorax_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hahella_ganghwensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halalkalibacillus_halophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haliea_salexigens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halobacillus_kuroshimensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_anticariensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_halocynthiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_lutea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halorubrum_ezzemoulense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hamadaea_tsunoensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helcococcus_sueciensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hellea_balneolensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_henriciella_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_humibacter_albus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hymenobacter_norwichensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_jeotgalicoccus_marinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_jeotgalicoccus_psychrophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_jonesia_quinghaiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kaistia_granuli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kangiella_aquimarina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kiloniella_laminariae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kordiimonas_gwangyangensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kribbella_catacumbae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kushneria_aurantia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lacticigenium_naphtae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_ceti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_harbinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_malefermentans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_psittaci_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_legionella_lansingensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_legionella_moravica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_legionella_shakespearei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptotrichia_shahii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptotrichia_trevisanii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leucobacter_chironomi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lewinella_cohaerens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lewinella_persica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_loktanella_hongkongensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_longispora_albida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lysobacter_concretionis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinimicrobium_agarilyticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinobacter_daepoensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinobacterium_litorale_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinomonas_ushuaiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinospirillum_insulare_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinospirillum_minutulum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_martelella_mediterranea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_megamonas_hypermegale_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_megamonas_rupellensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_meganema_perideroedes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_meiothermus_cerbereus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_meiothermus_taiwanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_meiothermus_timidus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesoflavibacter_zeaxanth_dsm18436_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesonia_mobilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_metascardovia_criceti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylophilus_methylotrophus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_gubbeenense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_indicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_luticocti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_moraxella_boevrei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_moraxella_caprae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_hassiacum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_neptunomonas_japonica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nesterenkonia_alba_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nevskia_ramosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nevskia_soli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_niabella_aurantiaca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardioides_halotolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nonomuraea_coxensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_novosphingobium_acidiphilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanicaulis_alexandrii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanicola_nanhaiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oligella_ureolytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_olivibacter_sitiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ornithinimicrobium_pekingense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_alginolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_daejeonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_fonticola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_ginsengihumi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sanguinis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_terrigena_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pannonibacter_phragmitetus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_parabacteroides_gordonii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paraoerskovia_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_patulibacter_americanus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelosinus_fermentans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_perlucidibaca_piscinae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pleomorphomonas_koreensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pleomorphomonas_oryzae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyrobacter_cryptus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_bennonis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_levii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_somerae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_albensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_corporis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_falsenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_loescheii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_maculosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_oris_dsm18711_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_paludivivens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_shahii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_timonensis_4401737_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_veroralis_dsm19559_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_propionibacterium_acidifaciens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_propionibacterium_jensenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_propionibacterium_thoenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_proteiniphilum_acetatigenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_proteocatella_sphenisci_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoclavibacter_soli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_resinovorans_dsm21078_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudonocardia_acaciae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudonocardia_asaccharolytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychrobacter_lutiphocae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychrobacter_phenylpyruvicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychroflexus_tropicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychroserpens_burtonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rathayibacter_toxicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rheinheimera_baltica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rheinheimera_perlucida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodonellum_psychrophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_riemerella_columbina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_robiginitomaculum_antarcticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseomonas_aerilata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruania_albidiflava_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rubritalea_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rubritepida_flocculans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rudaea_cellulosilytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_gauvreauii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharibacillus_kuerlensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharibacter_floricola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharospirillum_impatiens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salinarimonas_rosea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salinicoccus_albus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salinimicrobium_terrae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salinimonas_chungwhensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salipiger_mucosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salsuginibacillus_kocurii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sandarakinorhabdus_limnophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_schlesneria_paludicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sediminimonas_qiaohouensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_segetibacter_koreensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_bovis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serinicoccus_marinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shimazuella_kribbensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_silanimonas_lenta_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_simplicispira_psychrophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_solimonas_variicoloris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_solimonas_flava_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_smaragdicoccus_niigatensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_solimonas_soli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_solobacterium_moorei_dsm22971_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_astaxanthinifaciens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_sanxanigenens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingopyxis_baekryungensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spirochaeta_bajacaliforniensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spirosoma_luteum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spirosoma_panaciterrae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spirosoma_spitsbergense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spongiibacter_tropicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sporolactobacillus_vineae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sporomusa_ovata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sporosarcina_ureae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_stenoxybacter_acetivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_caballi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_castoreus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_devriesei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_didelphis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_entericus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_ferus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_henryi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_marimammalium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_merionis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_minor_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_orisratti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_ovis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_faecalicoccus_pleomorphus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_plurextorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_porci_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_thoraltensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_flavidovirens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_scabrisporus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sulphureus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_vitaminophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfurihydrogenibium_subterraneum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfurospirillum_arcachonense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tepidiphilus_margaritifer_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_terasakiella_pusilla_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tetragenococcus_muriaticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thalassobacter_arenae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thalassospira_xiamenensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thauera_linaloolentis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermithiobacillus_tepidarius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermodesulfatator_atlanticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermodesulfobacterium_hveragerdense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermodesulfobacterium_thermophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermodesulfovibrio_thiophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermomonas_fusca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermus_antranikianii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermus_igniterrae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermus_islandicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiobacillus_denitrificans_dsm12475_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiobacillus_thioparus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiothrix_disciformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiothrix_flexilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiothrix_lacustris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tuberibacillus_calidus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_uliginosibacterium_gangwonense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_varibaculum_cambriense_dsm15806_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_veillonella_magna_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vitreoscilla_stercoraria_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_weissella_halotolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_wenxinia_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_wohlfahrtiimonas_chitiniclastica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_woodsholea_maritima_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xenophilus_azovorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_yaniella_halotolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_zavarzinella_formosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_zymophilus_raffinosivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiomicrospira_arctica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiomicrospira_pelophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiomicrospira_sp_kp2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_stutzeri_dsm10701_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nafulsella_turpanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_sp_jc4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_protegens_cha0_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_morganella_morganii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_maltophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_medium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_socranskii_paredis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_olsenella_profusa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_socranskii_vpidr56br1116_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_pyruviciproducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deferrisoma_camini_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amycolatopsis_orientalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bosea_sp_117_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_doa9_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_weissella_confusa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_cyriacigeorgica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_glaciecola_lipolytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capnocytophaga_sp_f0381_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_sp_f0429_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_catoniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_veillonella_atypica_kon_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_acidurici_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacterium_sp_jkg1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_filamentous_cyanobacterium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_glaciecola_mesophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spiroplasma_melliferum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_hyorhinis_gdl1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alishewanella_jeotgali_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_glaciecola_psychrophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thaumarchaeota_sp_scgc_ab629i23_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrospina_sp_ab629b18_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_fragariae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dehalobacter_sp_cf_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrosomonas_cryotolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_vireti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_sp_83_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylobacterium_sp_88a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthobacter_sp_126_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_35w_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_37ma_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halococcus_hamelinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_sp_ccge510_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylotenera_mobilis_13_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_vini_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_brasiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_hygroscopicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_l1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_somaliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cellvibrio_sp_br_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_koehlerae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylocystis_parvus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_fragi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_cyclitrophicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_pentosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salinispora_pacifica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardiopsis_sp_cnt312_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_cnh099_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharomonospora_sp_cnq490_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_formosa_sp_ak20_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_endozoicomonas_numazuensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_curieae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_dispar_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_saccharolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_durans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_avium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_sulfureus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_intermedius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_providencia_burhodogranariea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_providencia_rettgeri_dmel1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phycisphaera_mikurensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_wigglesworthia_glossinidia_sp_gmo_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydia_gallinacea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corallococcus_coralloides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_novosphingobium_sp_ap12_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobium_sp_ap49_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_sp_cf080_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_sp_cf122_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_sp_cf136_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_herbaspirillum_sp_cf444_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_gm21_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_herbaspirillum_sp_yr522_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phyllobacterium_sp_yr531_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_cip102129_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_cip562_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rickettsia_helvetica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelagibacter_ubique_himb058_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lysinibacillus_varians_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blastococcus_saxobsidens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_asteroides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_pseudoalcaligenes_kf707_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_9mfcol31_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leifsonia_sp_109_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodospirillum_photometricum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesoaciditoga_lauensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_phlei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arcticibacter_svalbardensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfurospirillum_multivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phaeospirillum_molischianum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_micromonospora_lupini_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tsukamurella_sp_1534_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rahnella_aquatilis_hx2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermococcus_zilligii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_161mfsha21_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_162mfsha11_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_sp_292mf_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_paraoxydans_77mftsu32_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_mandelii_36mfcvi11_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptoclostridium_difficile_cd160_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leuconostoc_pseudomesenteroides_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_agalactiae_lmg14747_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_lappah108_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_mspmpm5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_hmica12_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_achromobacter_piechaudii_hle_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_agrobacterium_albertimagni_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylacidiphilum_fumariolicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tumebacillus_flagellatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_lappah95_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_amel2xe9_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_atexabd23_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_bolea5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_pstaah124_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_fxanac1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_variovorax_paradoxus_110b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_sp_155_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_135mfcol51_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_aljt_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_ald1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_almg11_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_ale31_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiomonas_sp_fbcd_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_sp_10_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfurobacterium_sp_tc51_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_persephonella_sp_if05l8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_persephonella_sp_km09lau8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_malodoratus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_raffinosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_villorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_asini_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_pallens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_haemoperoxidus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_moraviensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_phoeniculicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_caccae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_gilvus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_alj7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_alj11_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_alj16_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_equorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bordetella_sp_fb8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_sp_bmg111209_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ureibacillus_thermosphaericus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_auratus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_bicirculans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_maricaulis_sp_jl2009_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sp_acc21_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_nodatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptospirillum_ferrooxidans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phytoplasma_peanut_witchesbroom_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_francisella_noatunensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_hys_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodanobacter_spathiphylli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodanobacter_fulvus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodanobacter_thiooxydans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfuricella_denitrificans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_12_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fervidicoccus_fontis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_cetorum_mit995656_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_richelia_intracellularis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylotenera_sp_73s_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfurovum_sp_ar_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pectobacterium_sp_scc3193_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fibrella_aestuarina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterobacter_sp_r4368_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_xinjiangensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfocapsa_sulfexigens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_vitulinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_draconibacterium_orientale_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthobacteraceae_bacterium_501b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gammaproteobacteria_bacterium_bdw918_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiomicrospira_kuenenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinilabilia_salmonicolor_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_bryophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_123mfchir2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_sp_cny236_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_cnt372_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_cny243_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_propionibacterium_avidum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_calothrix_sp_pcc6303_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_propionibacterium_acidipropionici_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_142mfcol31_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_351mftsu51_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_303mfcol52_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_sp_348mftsu51_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_sp_360mftsu51_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_terracoccus_sp_273mftsu31_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfurimonas_sp_ast10_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_cinaedi_pagu611_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chamaesiphon_minutus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyanobacterium_sp_pcc7702_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_crinalium_epipsammum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fischerella_sp_pcc9431_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fischerella_sp_pcc9605_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geitlerinema_sp_pcc7407_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gloeocapsa_sp_pcc7428_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microcoleus_sp_pcc7113_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oscillatoria_sp_pcc10802_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spirulina_subsalsa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_pcc7502_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptolyngbya_sp_pcc6406_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_colletotrichum_sublineola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_psychrosaccharolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fischerella_sp_pcc9339_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_liberibacter_asiaticus_gxpsy_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingopyxis_sp_mc1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_herbaspirillum_sp_gw103_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aerococcus_viridans_ll1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevibacterium_massiliense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alcanivorax_sp_19m6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alcanivorax_hongdengensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alcanivorax_jadensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_sp_c448_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thalassospira_profundimaris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_sp_bjgmmb45_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_xiamenensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_dw54_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arenitalea_lutea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_photodesmus_blepharus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_lentus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharothrix_espanaensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_sp_g5847_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_m47t1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cladophialophora_psammophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_pseudoalcaligenes_cect5344_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pyrococcus_sp_st04_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gloeobacter_kilaueensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermogladius_cellulolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bdellovibrio_exovorus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_austwickia_chelonae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kineosphaera_limosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sinorhizobium_fredii_usda257_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_planococcus_antarcticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioclava_dalianensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fibrisoma_limi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blattabacterium_sp_bgi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_genomosp_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodovulum_sp_ph10_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_auris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_alkalescens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_bovigenitalium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_arginini_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_ovipneumoniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_yeatsii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ureaplasma_diversum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_rumoiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodovulum_sulfidophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_indibacter_alkaliphilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychroflexus_gondwanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_sp_acam123_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterovibrio_norvegicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterovibrio_calviensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_kanaloae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_venetianus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_melioribacter_roseus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chondromyces_apiculatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_sp_lig30_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobium_xenophagum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudaminobacter_salicylatoxidans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_parascardovia_denticolens_ipla20019_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tetrasphaera_elongata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_endolissoclinum_faulkneri_l2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_yannicii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_warneri_sg1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_vaccae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_termitidis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alishewanella_agri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_virgibacillus_halodenitrificans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_endophyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_oceanisediminis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_snodgrassella_alvi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gilliamella_apicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_maddingley_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_osyse_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_macauensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_stutzeri_ccug_29243_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halococcus_sp_197a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mangrovimonas_yunxiaonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_m2012083_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salmonella_bongori_n26808_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_afipia_birgiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_granulicella_tundricola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cycloclasticus_zancles_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aeropyrum_camini_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_janthinobacterium_sp_hh01_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_mutans_gs5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_terpenotabidum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerovibrio_sp_rm50_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminobacter_sp_rm87_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevibacillus_brevis_x23_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_birtlesii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteriovorax_sp_seq25v_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteriovorax_sp_bal6x_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_faecalis_atcc29212_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_moritella_dasanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_sp_f52_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halopiger_salifodinae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_moritella_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_ingluviei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_timonense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_oralis_hga0225_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sutterella_wadsworthensis_hga0223_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_sp_hpa0247_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_sp_kpl1859_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_sp_kpl1860_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dermabacter_sp_hfh0086_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sp_hph0090_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_atopobium_sp_f0494_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_propionibacterium_sp_f0372_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyricicoccus_pullicaecorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alistipes_onderdonkii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_sp_kpl1989_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_sp_kpl2004_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_reyranella_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_rattaustraliani_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_yersinia_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_translucens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_breoganii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardiopsis_alba_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_cnr698_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_aobensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_brevicatena_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_carnea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_exalbida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_higoensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_jiangxiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_otitidiscaviarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_niigatensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_takedensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_pneumoniae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_testacea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_concava_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_veterana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_transvalensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_lz4w_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobium_sp_c100_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomonas_sp_l53140_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanibaculum_indicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_uw4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_luteola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinomonas_sp_d104_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_celeribacter_baekdonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_commensalibacter_sp_mx01_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kinetoplastibacterium_crithidii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kinetoplastibacterium_desouzaii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kinetoplastibacterium_oncopeltii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kinetoplastibacterium_galatii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kinetoplastibacterium_blastocrithidii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cellvibrio_mixtus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_mageritense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tepidanaerobacter_acetatoxydans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_aa0539_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_janibacter_hoylei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gemmata_sp_iil30_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halogranum_salarium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lysinibacillus_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_psychrophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoxanthomonas_sp_gw2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chelatococcus_sp_gw1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_putida_nbrc14164_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_mesoamericanum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alistipes_marseilloanorexicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_massilioanorexius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blastococcus_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_anorexicamassiliense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_holdemania_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_stoquefichus_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_stutzeri_nf13_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterobacteriaceae_bacterium_b14_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_davawensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_suis_865192_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_suis_22083_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_suis_ys72_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_collinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_alcaligenes_nbrc14159_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_parafulva_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_liberibacter_crescens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactococcus_raffinolactis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoalteromonas_ruthenica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_hwasookii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_m240_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_edwardsiella_hoshinae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_elizabethkingia_meningoseptica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_achromobacter_xylosoxidans_nbrc15126_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_beijerinckii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_brisouii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_guillouiae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_gyllenbergii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_anc4105_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_anc3862_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_niph2100_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_niph899_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_niph758_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_niph809_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_anc3789_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_bohemicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cupriavidus_sp_bis7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseomonas_sp_b5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_acidipaludis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_bannensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_ferrariae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_terrae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_citrobacter_sedlakii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_indologenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_empedobacter_brevis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_alcalophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_stutzeri_kos6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_comamonas_aquatica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_novosphingobium_tardaugens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobium_herbicidovorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_parapaucimobilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_proteolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_halioticoli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_alginolyticus_nbrc15630_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_azureus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_ezurae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermotoga_hypogea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_cassavae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_sp_s4c9_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arcanobacterium_sp_s3pf19_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptostreptococcus_sp_mv1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanobacterium_sp_maddingley_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alpha_proteobacterium_imcc14465_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_chromogenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_rubi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_aichiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lysinibacillus_odysseyi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_fluorescens_ncimb11764_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_florum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_altibacter_lentus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_comamonas_granuli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_mobaraensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_malaquae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_paraffinivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_sihwensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_soli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterobacteriaceae_bacterium_lsjc7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_maris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_vitaeruminis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mangrovibacter_sp_mfb070_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gluconobacter_oxydans_h24_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pantoea_sp_a4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dickeya_sp_dw_0440_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oscillibacter_sp_kle1728_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_kle1755_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_nitroreducens_tx1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinobaculum_sp_f0552_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_capnocytophaga_sp_f0517_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptotrichia_sp_f0557_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_sp_w7784_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_sp_f0091_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_fslh7689_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_fslr7277_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_viridibacillus_arenosi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloarcula_japonica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halobiforma_nitratireducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halococcus_thailandensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halorubrum_saccharovorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halosarcina_pallida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloterrigena_salina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natrinema_pallidum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natronococcus_amylolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natronolimnobius_innermongolicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natronorubrum_bangense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hymenobacter_swuensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptolyngbya_sp_kiost1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinobacterium_llx17_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alpha_proteobacterium_l41a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_phenoliruptrix_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobacterium_sp_accc05744_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gammaproteobacteria_bacterium_wg36_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_sp_wg21_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blattabacterium_sp_bpaa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactococcus_garvieae_i113_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alkalibacterium_sp_ak22_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leuconostoc_gelidum_jb7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leuconostoc_carnosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microthrix_parvicella_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevibacterium_casei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_massiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydia_avium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrosopumilus_sp_ar2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_moraxella_macacae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salimicrobium_sp_mj3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_tetanomorphum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_legionella_anisa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloterrigena_limicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_natronorubrum_sulfidifaciens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_dfci1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenisporosarcina_sp_tg14_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitratireductor_aquibiodomus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitratireductor_indicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermotoga_sp_mc24_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_shenzhenensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactococcus_garvieae_dcc43_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pusillimonas_noertemannii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceaniovalibus_guishaninsula_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cardinium_endosymbiont_ceper1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfuromonas_sp_tf_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_sp_gd7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_megasphaera_sp_np3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nesterenkonia_sp_np1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nosocomiicoccus_sp_np2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coriobacteriaceae_bacterium_gd5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfobacula_sp_ts_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiales_bacterium_ve20213_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiales_bacterium_ve20218_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiales_bacterium_ve20209_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiales_bacterium_ve20208_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiales_bacterium_ve20214_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiales_bacterium_ve20221_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_sciuri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinobacterium_sp_ak27_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_sp_jm1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_epilithonimonas_sp_fh1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dyella_ginsengisoli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_catellicoccus_marimammalium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_sp_e463_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alpha_proteobacterium_jlt2015_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacillus_caldoxylosilyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_carnobacterium_maltaromaticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bhargavaea_cecembensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ralstonia_sp_au1208_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salinicoccus_carnicancri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_massiliensis_dnlkv3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_sp_142_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_m181_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_coe1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterorhabdus_caecimuris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_firmicutes_bacterium_m102_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oscillibacter_sp_13_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dorea_sp_52_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_32_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_101_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_murinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_plexicaudatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_parabacteroides_sp_asf519_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_disiens_jcm6334_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_pyogenes_jcm10003_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_enoeca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerotruncus_sp_g3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halyomorpha_halys_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_pleuritidis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_oulorum_jcm14966_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acetobacter_okinawensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_histicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_aurantiaca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_stercorirosoris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_fusca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_scopos_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_haliotis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_marina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methanomethylophilus_alvus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardiopsis_baichengensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_wolbachia_sp_dsi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylobacillus_glycogenes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_akibai_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_pini_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fulvivirga_imtechensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardiopsis_ganjiahuensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caenispirillum_salinarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanobacillus_kimchii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leucobacter_salsicius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_jeotgali_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydia_psittaci_10139811_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloquadratum_walsbyi_j07hqw2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_nigripulchritudo_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dokdonia_sp_pro95_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mariniradius_saccharolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodococcus_ruber_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_putida_mtcc5279_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cystobacter_fuscus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_concisus_atcc51562_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_g2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_showae_csunswcd_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_fetus_testudinum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_sp_1485e_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_magnetospirillum_sp_so1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_oligotrophicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_resinovorans_nbrc106553_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardiopsis_prasina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardiopsis_kunsanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardiopsis_valliformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_sp_2mfcol31_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardiopsis_xinjiangensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halobacillus_sp_bab2008_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_lehensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_cynos_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinoplanes_friuliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arsenophonus_endosymbiont_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bordetella_holmesii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_advenella_mimigardefordensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_parvularcula_oceani_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_photobacterium_leiognathi_lrivu41_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blastomonas_sp_aap53_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sandarakinorhabdus_sp_aap62_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyrobacter_sp_aap82_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfuricurvum_sp_rifrc1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thiorhodococcus_sp_ak35_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serratia_marcescens_fgi94_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gillisia_sp_heli29_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_maribacter_sp_heli7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_polaribacter_sp_hel185_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_polaribacter_sp_heli88_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_muricauda_sp_mar201075_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salegentibacter_sp_heli6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sorangium_cellulosum_so01572_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_nitratireducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_ramulus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coprothermobacter_platensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spiribacter_salinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_faecalis_13sdw01_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_liberibacter_americanus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halarchaeum_acidiphilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_pasteurianum_dsm525_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phascolarctobacterium_sp_cag207_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phascolarctobacterium_sp_cag266_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bibersteinia_trehalosi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_asiaticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haliea_rubra_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_urmo17wk12i8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldimonas_manganoxidans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_colwellia_piezophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfospira_joergensenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_succinimonas_amylolytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelagibacter_ubique_htcc9022_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_listeria_weihenstephanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_carnobacterium_sp_wn1359_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_ale6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_ale19_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_akl19_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrosospira_briensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracoccus_zeaxanthinifaciens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_parabrevis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomicrobium_zavarzinii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sediminibacterium_sp_c3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidobacteriaceae_bacterium_kbs83_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidobacteriaceae_bacterium_kbs89_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidobacteriaceae_bacterium_kbs96_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anoxybacillus_flavithermus_tno09006_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pantoea_sp_imh_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_g5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sabinae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aeromonas_diversa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoalteromonas_luteoviolacea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_cellulosilyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodococcus_sp_aw25m09_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidovorax_sp_mrs7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_legionella_oakridgensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_alr1721_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_sp_kjj_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pedobacter_sp_r2019_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_megasphaera_sp_nm10_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodococcus_rhodnii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_planomicrobium_glaciei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_gd11_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hepatobacter_penaei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sonorensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spiroplasma_taiwanense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spiroplasma_diminutum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spiroplasma_chrysopicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spiroplasma_syrphidicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spiroplasma_culicicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spiroplasma_sabaudiense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spiroplasma_apis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidovorax_sp_jhl9_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_gangotriensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_myxococcus_stipitatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodococcus_triatomae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acholeplasma_granularum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_russii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychromonas_ossibalaenae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_zimmermannella_faecalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_amphritea_japonica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acholeplasma_axanthum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_avibacterium_paragallinarum_221_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cesiribacter_andamanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanimonas_smirnovii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbulbifer_variabilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_thiocyanoxidans_arh2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_novispirillum_itersonii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_jasicida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_kordi100_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillaceae_bacterium_g5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_ae2015_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_vcd2006_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_ae3009_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_xpd2006_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_fc2001_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_ae3006_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_sp_agr2160_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_wcd3002_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_vcb2006_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_lc3010_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_wcd2001_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_xbb1001_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_nc3005_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_mc2013_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudobutyrivibrio_ruminis_cf1b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_paraputrificum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_cadaveris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudobutyrivibrio_ruminis_ad2017_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_fibrisolvens_nd3005_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dorea_longicatena_agr2136_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_ruminantium_atcc12561_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomonas_sp_t16b2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomonas_sp_cy54118_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomonas_sp_25b141_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomonas_sp_bhbn044_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomonas_sp_22ii122f38_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomonas_adhaerens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomonas_johnsonii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomonas_jannaschiana_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomonas_oceanitis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hyphomonas_polymorpha_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_agrobacterium_tumefaciens_cherry_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_poae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_asticcacaulis_sp_ac460_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_asticcacaulis_sp_ac402_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_asticcacaulis_sp_ac466_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_tigurinus_az3a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_tigurinus_1366_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alpha_proteobacterium_mf_105b01_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospira_multipara_atcc19207_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_frankia_sp_iso899_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tolumonas_sp_brl61_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardioides_sp_iso805n_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_conexibacter_woesei_iso977n_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylohalobius_crimeensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_a9_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_urogenitalis_s6c4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_sp_s6spd3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerococcus_lactolyticus_s7113_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiales_bacterium_s714_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_sp_s718_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_casei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lysinibacillus_sphaericus_ot4b31_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pandoraea_sp_sd62_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylophaga_lonarensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_raoultella_ornithinolytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_acidaminophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphaerotilus_natans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_zhouia_amylolytica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_sp_l2c084a000_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_sp_lnjc398b00_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_freneyi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_bivia_dnf00188_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_sp_s7_ms_2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_cnt318_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_taa040_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseovarius_mucosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_wulumuqiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrosospira_sp_apg3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinobacter_santoriniensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyclobacteriaceae_bacterium_ak24_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brachyspira_hampsonii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_taa204_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_josui_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_oryzae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dietzia_sp_ucdthp_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_putrefaciens_mput9231_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caulobacter_crescentus_or37_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sordellii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_propionibacterium_granulosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halopiger_sp_iih2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halopiger_sp_iih3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halanaerobium_saccharolyticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_equicursoris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_intestinale_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_denitrificans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_boroniphilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseibacterium_elongatum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacillus_stearothermophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aquimarina_sp_sw150_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aquimarina_megaterium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_komagataeibacter_xylinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_dentisani_7746_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_metallidurans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_sp_stm4661_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anoxybacillus_flavithermus_ak1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_intestinimonas_butyriciproducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_myxococcus_sp_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_dentisani_7747_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_afipia_sp_ohsuic4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_ohsuiii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thalassolituus_oleivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gracilibacillus_boraciitolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychrobacter_sp_jcm18900_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_sp_urha0056_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_sp_urha0036_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marmoricola_sp_urhb0036_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_sp_urhd0025_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alteromonas_sp_alt199_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_urha0002_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_taa486_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfotomaculum_guttoideum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_kansasii_732_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_oranimense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterococcus_mundtii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lysobacter_dokdonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfitobacter_donghicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_knackmussii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_01_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_namurensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_borrelia_miyamotoi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_oligofermentans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chthonomonas_calidirosea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_fulvissimus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salinisphaera_hydrothermalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiaceae_bacterium_l21thd2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cellulomonas_sp_krmcy2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_asbs410_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_magneticus_ifrc170_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aminiphilus_circumscriptus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aminobacterium_mobile_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_nicotinovorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_japonicum_22_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_japonicum_in8p8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caldicoprobacter_oshimai_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dechloromonas_agitata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulforegula_conservatrix_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geovibrio_sp_l21acebes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_sp_krocy2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanicola_sp_hl35_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_algoriphagus_marincola_hl49_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sporosarcina_sp_eur3222_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kineosporia_aurantiaca_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_thermolilacinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_candidate_division_tm6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_r1ns10_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_firmus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoalteromonas_agarivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_sp_l21syrab_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spirochaeta_sp_l21rpuld2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gluconobacter_frateurii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gracilibacillus_halophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_br18_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_h14_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ilumatobacter_coccineus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermocrinis_sp_gbs_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_borrelia_coriaceae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_borrelia_anserina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_borrelia_parkeri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermonema_rossianum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chitinivibrio_alkaliphilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aureispira_sp_ccbqb1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sp_hsism1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blattabacterium_sp_nci_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_corrugata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mannheimia_haemolytica_m42548_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phaeospirillum_fulvum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseivivax_sp_22iis10s_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aquimarina_sp_22iis11z7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioclava_sp_13d2w2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycoplasma_girerdii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinobacter_lipolyticus_sm19_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cetobacterium_somerae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_sp_nbimcp2c3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_suis_ea183292_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_atopobium_sp_f0209_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptotrichia_sp_f0581_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomyces_sp_f0311_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiales_bacterium_f0540_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptotrichia_sp_w9775_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mitsuokella_sp_w9106_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oribacterium_sp_f0263_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptostreptococcaceae_bacterium_w5053_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_sp_f0426_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_brachy_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_lecithinolyticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gemella_bergeriae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfovibrio_piezophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodococcus_defluvii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_pelagia_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halobonum_tyrrellensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_fennelliae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_catenovulum_agarivorans_ds2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_weissella_oryzae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermoactinomyces_daqus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermus_caliditerrae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobium_lactosutens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_haemolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serratia_fonticola_auap2c_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoalteromonas_haloplanktis_tb64_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_salinarchaeum_sp_harchtbsk1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_stutzeri_mf28_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alpha_proteobacterium_ma2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kurthia_huakuii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spiribacter_sp_uahsp71_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobium_sp_yl23_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseomonas_gilardii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_waksmanii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_atopobacter_phocae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_selenitireducens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_flectens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_xylanophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microvirga_flocculans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_carnimonas_nigrificans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_larrymoorei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_polaribacter_sp_hel13349_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_loktanella_cinnabarina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_calothrix_sp_3363_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_elizabethkingia_anophelis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_unc438cl73tsus30_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bryobacter_aggregatus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_laceyella_sacchari_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_saliperosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_flavefaciens_007c_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_limnosediminis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_septicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_indicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_bombi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfitobacter_sp_mm124_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfitobacter_sp_nb77_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfitobacter_sp_nb68_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharibacter_sp_am169_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_palaeococcus_pacificus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_rapamycinicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tatumella_sp_nml063099_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_exiguobacterium_pavilionensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_pylori_sa213a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_saccharobutylicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacillus_sp_jf8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobacterium_paucimobilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobium_ummariense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_eb01_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paucisalibacillus_sp_eb02_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_formosa_agariphila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_ff3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_ff4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerosalibacter_sp_nd1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroidaceae_bacterium_ms4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudoalteromonas_sp_ocn003_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leifsonia_rubra_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cellulophaga_baltica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_diazotrophicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thauera_terpenica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_argentoratense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kitasatospora_cheerisanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_megaterium_msp201_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_janthinobacterium_agaricidamnosum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tenacibaculum_maritimum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_akyn10_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_coprobacter_fastidiosus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_niveus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_olleya_marilimosa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioclava_sp_dt234_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteriovorax_sp_bsw11iv_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ensifer_sp_tw10_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioclava_pacifica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptoniphilus_sp_chdc_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychrobacter_aquaticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cardinium_endosymbiont_cbtq1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseovarius_sp_mctg156_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arcobacter_sp_af1028_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arcobacter_sp_af1440_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hymenobacter_sp_apr13_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alicyclobacillus_acidoterrestris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_syringae_cc1417_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_syringae_cc1513_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_syringae_cc1557_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_canis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_macacae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobacterium_antarcticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermofilum_sp_1910b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodobacteraceae_bacterium_himb11_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ralstonia_pickettii_dtp0602_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_coli_76339_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracoccus_aminophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_sp_umwgj_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_sp_umwwy_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_undicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_iranicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_firmicutes_bacterium_asf500_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gemmatimonas_sp_ap64_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfonauticus_sp_a7a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_candidate_division_zixi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_sp_fukuswis1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mucispirillum_schaedleri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_urha0041_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geodermatophilaceae_bacterium_urhb0048_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mesorhizobium_sp_urhc0008_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cellulomonas_sp_urhe0023_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bradyrhizobium_sp_urhd0069_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_blastococcus_sp_urhd0036_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_sp_23gom1509m_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfitobacter_sp_20gpm1509m_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_intrasporangiaceae_bacterium_urhb0013_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ahrensia_sp_13gom1096m_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tenacibaculum_sp_47agom205m_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_sp_urhb0044_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alcanivorax_sp_43bgom46m_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_solirubrobacterales_bacterium_urhd0059_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dongia_sp_urhe0060_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geodermatophilaceae_bacterium_urha0031_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodospirillales_bacterium_urhd0088_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anoxybacillus_flavithermus_yunnanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flaviramulus_ichthyoenteri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_darwinianus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aliihoeflea_sp_2ww_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevibacterium_sp_vcm10_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_asaia_platycodi_sf21_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caulobacteraceae_bacterium_pmmr1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alicyclobacillus_macrosporangiidus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sporosarcina_sp_d27_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermogemmatispora_sp_pm5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacillus_vulcani_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermomicrobiales_sp_ki4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anoxybacillus_tepidamans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidobacterium_sp_pmmr2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lysinibacillus_manganicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arenimonas_malthae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arenimonas_metalli_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lysinibacillus_sinduriensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_albus_ad2013_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_flavefaciens_ae3010_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_adlercreutzia_equolifaciens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_francisella_sp_w121067_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pontibacillus_halophilus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pontibacillus_marinus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pontibacillus_litoralis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pontibacillus_chungwhensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pontibacillus_yanchengensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lysobacter_defluvii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lysobacter_daejeonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_knoellia_flava_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_knoellia_aerolata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_knoellia_sinensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_knoellia_subterranea_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptolyngbya_sp_heron_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_intrasporangium_oryzae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gordonia_amicalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_schmidhempelia_bombi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylophilus_sp_oh31_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_mosselii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leifsonia_xyli_cynodontis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_mendocina_egdaq5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_bifermentans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_adety_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudozyma_brasiliensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_sp_hun102_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_sp_ab3007_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leeuwenhoekiella_sp_heli48_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leeuwenhoekiella_sp_mar2009132_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sediminibacter_sp_heli10_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcaceae_bacterium_ae2021_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_ac2031_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_maribacter_forsetii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_ruminantium_ac2024_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_ruminantium_ab3002_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_nectaris_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bordetella_hinzii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_sp_7101_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_sp_a11_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinomadura_madurae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sporolactobacillus_laevolacticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_moraviensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_taeanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_maepy2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haloferula_sp_bvorr071_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_verrucomicrobium_sp_bvorr034_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinobacter_sp_es1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leucobacter_sp_ph1c_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serratia_sp_h1n_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alcanivorax_sp_p2s70_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_sp_pbn3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alpha_proteobacterium_rs24_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_exiguobacterium_undae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_exiguobacterium_marinum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_exiguobacterium_oxidotolerans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_borrelia_valaisiana_tom4006_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_exiguobacterium_sp_mh3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_holospora_obtusa_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_enterobacter_cloacae_jd6301_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_fabifermentans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelagibacter_ubique_htcc7214_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pelagibacter_ubique_htcc7217_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_tuscaniense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oligella_urethralis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_veillonella_montpellierensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_buccalis_dnf00985_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_endolissoclinum_faulkneri_l5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfitobacter_sp_h3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_simplex_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_verrucomicrobium_sp_bvorr106_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_anginosus_dora7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_varibaculum_cambriense_dora20_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_glycinophilum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydia_suis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_staphylococcus_simulans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_beibuense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_synechococcus_sp_nkbg15041c_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_betaproteobacteria_bacterium_mola814_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sinorhizobium_americanum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vagococcus_lutrae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevibacillus_panacihumi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_nucleatum_w1481_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_mb2021_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_fcs014_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_mc2021_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_sp_ma2016_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oribacterium_sp_fc2011_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudobutyrivibrio_sp_lb2011_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_ac2028_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_ac3007_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_ma2020_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_mc2017_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acholeplasma_equifetale_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acholeplasma_hippikon_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acholeplasma_modicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidiphilium_angustum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acidocella_facilis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alkaliphilus_transvaalensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_anaerovibrio_lipolyticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_bogoriensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bartonella_elizabethae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bilophila_wadsworthia_atcc49260_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_crocinitomix_catalasitica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_cellulosolvens_ld2006_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_eubacterium_desmolans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_facklamia_sourekii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_fusobacterium_perfoetens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_pametensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_legionella_fairfieldensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_legionella_sainthelensi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_genavense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prolixibacter_bellariivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobacterium_sp_h1ai_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_mangenotii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacteroides_sp_ga6a2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_sp_c6a8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oribacterium_sp_p6a1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_sp_p6b1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudobutyrivibrio_sp_md2005_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_sp_fc2018_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_sp_nd2010_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serratia_sp_dd3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shinella_sp_dd12_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_c6a11_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_fe2018_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_md2004_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_nc2004_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_nd2006_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_p6a3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_p6b14_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_v9d3004_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_ysb2008_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_propionibacteriaceae_bacterium_p6a17_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcaceae_bacterium_ab4001_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_proteoclasticus_p6b7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_lundense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kandleria_vitulina_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospira_multipara_lb2003_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mitsuokella_jalaludinii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prevotella_brevis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_proteiniclasticum_ruminis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_flavefaciens_ma2007_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sharpea_azabuensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_succinivibrio_dextrinosolvens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobiales_bacterium_yim77505_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gammaproteobacteria_bacterium_mola455_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_sp_jcb_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_jcc_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nocardia_nova_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_tkp_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinobacter_sp_mctg268_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_sp_tg39a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanicola_sp_mctg156_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_botulinum_202f_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_baratii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porticoccus_hydrocarbonoclasticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_polycyclovorans_algicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_algiphilus_aromaticivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_paraoxydans_dh1b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leifsonia_aquatica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_myroides_odoratimimus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_borrelia_persica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_defluviimonas_sp_20v17_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_mandelii_pd30_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptococcus_sp_vt_162_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phaeobacter_gallaeciensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_sjs_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_apodemi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_casei_dsm20011_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_composti_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_farraginis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_fuchuensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_gallinarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_hamsteri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_hayakitensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_hominis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_kitasatonis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_nodensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_otakiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_pasteurii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_sucicola_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_suebicus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_vaginalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_versmoldensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_zeae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_advenella_kashmirensis_w13003_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hepatoplasma_crinochetorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodococcus_rhodochrous_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_stenotrophomonas_maltophilia_5bai2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_afipia_sp_p5210_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geobacillus_sp_g1w1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_magnetospirillum_gryphiswaldense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_etli_mimosae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_komagataeibacter_rhaeticus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mannheimia_varigena_usmarc1261_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rikenellaceae_bacterium_m3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mannheimia_varigena_usmarc1296_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dyadobacter_tibetensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_pseudomallei_abcpw111_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_moukalabense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodococcus_pyridinivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_criblamydia_sequanensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ochrobactrum_rhizosphaerae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_pullorum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_mongoliense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_actinocoloniiforme_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_bohemicum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_biavatii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_callitrichos_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bifidobacterium_reuteri_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_castellaniella_defragrans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_nitroreducens_hbp1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_bay1663_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_escherichia_albertii_kf1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_scopuliridis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_aromaticivorans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_cichorii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serratia_fonticola_rb25_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arcobacter_cibarius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_luteimonas_huabeiensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfitobacter_guttiformis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_yersinia_enterocolitica_lc20_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_novyi_b_nctc9691_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_botulinum_cd_bkt12695_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aquimarina_macrocephali_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sporolactobacillus_terrae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_brevibacillus_borstelensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_flexus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydia_sp_diamant_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlamydia_sp_rubis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xylella_fastidiosa_pls229_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sciscionella_sp_se31_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracoccus_yeei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aeromonas_hydrophila_yl17_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_tomitella_biformata_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_sp_cit045_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halobellus_rufus_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthrobacter_sp_unc362mftsu51_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mycobacterium_sp_unc280mftsu51_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caulobacter_sp_unc358mftsu51_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_knhs205_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_sp_unccl10_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_unc451mf_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_phyllobacterium_sp_unc302mfcol52_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodococcus_sp_unc23mfcrub11_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodococcus_sp_unc363mftsu51_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_sp_unc305mfcol52_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermus_sp_yim77409_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridiales_bacterium_dri13_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_carnobacterium_alterfunditum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_carnobacterium_divergens_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_carnobacterium_funditum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_carnobacterium_gallinarum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_carnobacterium_mobile_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_carnobacterium_pleistocenium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_rodentium_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kitasatospora_azatica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kitasatospora_mediocidica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseivivax_halodurans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_roseivivax_isoporae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptacidiphilus_oryzae_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_yeochonensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thermus_tengchongensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_kutzneria_albida_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_succinicans_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_ts2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_falsenii_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_sp_mrs1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_sp_crk32_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbacterium_sp_crk20_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_rl_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hafnia_alvei_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_aquatile_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_schleiferia_thermophila_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_microbulbifer_sp_hz11_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_pseudoalcaligenes_ad6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_chungangense_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_accumulibacter_sp_ba93_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pedobacter_borealis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cellulomonas_sp_hzm_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_photobacterium_phosphoreum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haladaptatus_cibarius_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halapricum_salinum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acetobacter_aceti_1023_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hylemonella_gracilis_niagara_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_sp_mp1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_comamonadaceae_bacterium_h1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lachnospiraceae_bacterium_ac2029_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_nitrososphaera_evergladensis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geomicrobium_sp_jcm19037_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_geomicrobium_sp_jcm19038_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_jcm19046_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_weeksella_sp_ff8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_haemophilus_sp_ff7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_jce_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_jeotgalicoccus_sp_13mg44air_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actibacterium_atlanticum_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanicola_sp_22iis11g_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_virgibacillus_sp_vm5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanobacillus_sp_s5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinosporangium_sp_nrrlb3428_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_glycomyces_sp_nrrlb16210_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_saccharothrix_sp_nrrlb16314_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrlf2580_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrlf2890_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrlf5008_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrlf5053_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrlf5065_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrlf5123_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrlf5126_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrlf5135_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrlf525_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrlf5630_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrlf6677_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrls118_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrls149_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrls1777_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrls237_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrls337_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrls340_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrls37_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrls474_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrls646_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrls87_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrls920_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrlwc3626_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrlwc3742_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrlwc3773_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_micromonospora_parva_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thioalkalivibrio_sp_hk1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lacinutrix_sp_pamc27137_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_scytonema_hofmanni_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhodovulum_sp_ni22_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_knhs209_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_sp_rit328_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_rit357_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_sp_scb241528o2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_sp_scb245a520k10_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_falsirhodobacter_sp_alg1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_sp_k24_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lysinibacillus_sp_bf4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_phoenicis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_streptomyces_sp_nrrlb24720_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_peptostreptococcaceae_bacterium_va2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_sp_hl48_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_marinobacter_sp_hl58_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oceanicaulis_sp_hl87_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyrobacter_sp_hl46_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_curtobacterium_sp_s6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_spirochaeta_sp_jc230_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pantoea_sp_psnih2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pantoea_sp_psnih1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_psychroserpens_sp_pamc27130_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_exiguobacterium_sp_ab2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_selenomonas_sp_ae3005_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ferriphaselus_sp_r1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_zetaproteobacteria_bacterium_tag1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_hmp27_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_desulfosporosinus_sp_hmp52_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_leptolyngbya_sp_jsc1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_corynebacterium_sp_atcc6931_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_ant303_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_sp_rl_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_sp_em1308_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_sp_em1321_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_gammaproteobacteria_bacterium_mfb021_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_cyanobium_sp_caciam14_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_listeriaceae_bacterium_fsla50209_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_sp_mit0601_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_sp_mit0602_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_sp_mit0701_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_mt2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_cl6_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_cl2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacillus_sp_kw12_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_20bn_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_lf2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacterium_uasb270_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_tca20_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_sp_yr295_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_sp_yr519_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chryseobacterium_sp_cf284_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_sp_cf097_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_sp_cf394_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_sp_ok494_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_luteibacter_sp_9143_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_luteibacter_sp_9135_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_massilia_sp_9096_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderia_sp_9120_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_mst1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_sp_mit0604_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_prochlorococcus_sp_mit0801_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alistipes_sp_627_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_devosia_sp_lc5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_methylotenera_sp_l2l1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingopyxis_sp_lc81_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bosea_sp_lc85_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_massilia_sp_lc238_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_actinobacterium_acamd5_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_polaromonas_sp_cg912_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacterium_sp_ol1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bacterium_sp_lf3_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_halomonas_sp_ko116_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_flavobacterium_sp_fl_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_butyrivibrio_sp_ae3004_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_arthromitus_sp_sfbmousenl_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_acinetobacter_sp_hr7_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_sp_r1200b2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_bosea_sp_unc402clcol_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_ruminococcus_sp_hun007_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_sp_cot239oh1446_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonas_sp_cot290oh860_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_sp_yqh10_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_idiomarina_sp_mccc1a10513_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_vibrio_sp_er1a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_porphyromonadaceae_cot184oh4590_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_oscillibacter_sp_er4_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chlorobium_sp_gbchlb_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_chloroflexus_sp_msg_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobium_sp_ba1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_pseudomonas_sp_ml96_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_serratia_sp_ag1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracoccus_sp_4681_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_planococcus_sp_pamc21323_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_atelocyanobacterium_thalassa_sio64986_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rickettsiales_bacterium_ac37b_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thalassospira_australica_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_maribius_sp_mola401_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_achromobacter_sp_rta_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_rhizobium_sp_ys1r_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_devosia_sp_172e8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_thalassotalea_sp_nd16a_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_fslp40081_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_fslr50345_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_fslr70273_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_fslr70331_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_fslh70357_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paenibacillus_sp_fslh70737_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingopyxis_sp_mwb1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfurospirillum_sp_scadc_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfuricurvum_sp_mlsb_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_alteromonas_sp_lor_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_aquabacterium_sp_nj1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobacterium_sp_ml3w_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_treponema_sp_omz838_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_deinococcus_sp_yim77859_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_clostridium_sp_knhs214_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_myxosarcina_sp_gi1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mollicutes_bacterium_hr1_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_mollicutes_bacterium_hr2_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_sp_wkb10_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_lactobacillus_sp_wkb8_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_sp_3524zxx_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_hoeflea_sp_bal378_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_francisella_sp_fsc1006_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_helicobacter_sp_mit115569_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_campylobacter_sp_mit975078_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_taxi_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingomonas_sp_37zxx_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sphingobacteriaceae_bacterium_dw12_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_listeriaceae_bacterium_fsla50281_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_xanthomonas_sp_nyagatare_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_caenorhabditis_tropicalis_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_burkholderiaceae_sp_mortierella_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_shewanella_sp_ecsmb14101_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_sulfurospirillum_sp_mes_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_paracoccus_sp_pamc22219_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }

        if is_valid_string_dickeya_sp_2b12_node_name(node_name) {
            return Ok(unsafe { format_string_url_from_node_name(node_name) });
        }
    }

    Err(format!(
        concat!(
            "There is no known url with a pattern for the provided node name {:?}.\n",
            "If you believe there should be one, please do open a pull request to ",
            "add it to the library!"
        ),
        node_name
    ))
}

/// Returns name of the graph repository from the given node name.
///
/// # Implementative details
/// Currently we have support for building the URLs for:
/// * [Sequence Ontology](http://www.sequenceontology.org/)
/// * [FlyBase](http://flybase.org/)
/// * [Mouse Genome Informations](http://www.informatics.jax.org/)
/// * [Pubmed NCBI](https://www.ncbi.nlm.nih.gov/)
/// * [NCBI Gene](https://www.ncbi.nlm.nih.gov/gene/)
/// * [NCBI ClinVar](https://www.ncbi.nlm.nih.gov/clinvar/)
/// * [WormBase](https://wormbase.org)
/// * [Ensembl](https://www.ensembl.org/index.html)
/// * [ZFIN](http://zfin.org/)
/// * [DOI](https://www.doi.org/)
/// * [CHEBI](https://www.ebi.ac.uk/chebi/init.do)
/// * [BioGrid](https://thebiogrid.org/)
/// * [OMIM](https://www.omim.org/)
/// * [Rat Genome DataBase](https://rgd.mcw.edu/rgdweb/homepage/)
/// * [MMRRC](https://www.mmrrc.org/)
/// * [GO](http://amigo.geneontology.org/amigo/landing)
/// * [UniProtKB](https://www.uniprot.org/)
/// * [Coriell](https://coriell.org/)
/// * [DrugCentral](https://drugcentral.org/)
/// * [NCBI MESH](https://www.ncbi.nlm.nih.gov/mesh/)
/// * [NCBI Taxonomy](https://www.ncbi.nlm.nih.gov/taxonomy)
/// * [JAX](https://www.jax.org/strain)
/// * [WikiData](https://www.wikidata.org/wiki/Wikidata:Main_Page)
/// * [Therapeutic Target Database](http://db.idrblab.net/ttd/)
/// * [Reactome](https://reactome.org/)
///
/// # Arguments
/// * `node_name`: &str - Node name to query for.
///
/// # Raises
/// * If there is no known url source for the given node name.
///
/// TODO: update the list of supported repositories.
pub fn get_node_repository_from_node_name(node_name: &str) -> Result<&str> {
    if is_valid_flybase_node_name(node_name) {
        return Ok("FlyBase");
    }
    if is_valid_sequence_ontology_node_name(node_name) {
        return Ok("Sequence Ontology");
    }
    if is_valid_mmrrc_node_name(node_name) {
        return Ok("MMRRC");
    }
    if is_valid_wormbase_gene_node_name(node_name) {
        return Ok("WormBase Gene");
    }

    if is_valid_mouse_genome_informatics_node_name(node_name) {
        return Ok("Mouse Genome Informatics");
    }
    if is_valid_ncbi_dbgap_node_name(node_name) {
        return Ok("NCBI dbGaP");
    }
    if is_valid_pubmed_ncbi_node_name(node_name) {
        return Ok("Pubmed NCBI");
    }
    if is_valid_ncbi_gene_node_name(node_name) {
        return Ok("NCBI Gene");
    }
    if is_valid_ncbi_clinvar_node_name(node_name) {
        return Ok("NCBI ClinVar");
    }
    if is_valid_ensembl_node_name(node_name) {
        return Ok("ENSEMBL");
    }
    if is_valid_zfin_node_name(node_name) {
        return Ok("ZFIN");
    }
    if is_valid_doi_node_name(node_name) {
        return Ok("DOI");
    }
    if is_valid_chebi_node_name(node_name) {
        return Ok("CHEBI");
    }
    if is_valid_biogrid_node_name(node_name) {
        return Ok("BIOGRID");
    }
    if is_valid_omim_node_name(node_name) {
        return Ok("OMIN");
    }
    if is_valid_rat_genome_database_node_name(node_name) {
        return Ok("Rat Genome Database");
    }
    if is_valid_gene_ontology_node_name(node_name) {
        return Ok("Gene Ontology");
    }
    if is_valid_uniprotkb_node_name(node_name) {
        return Ok("UniProtKB");
    }
    if is_valid_coriell_node_name(node_name) {
        return Ok("Coriell");
    }
    if is_valid_drugcentral_node_name(node_name) {
        return Ok("DrugCentral");
    }
    if is_valid_ncbi_mesh_node_name(node_name) {
        return Ok("NCBI Mesh");
    }
    if is_valid_ncbi_taxonomy_node_name(node_name) {
        return Ok("NCBI Taxonomy");
    }
    if is_valid_jax_node_name(node_name) {
        return Ok("JAX");
    }
    if is_valid_wikidata_node_name(node_name) {
        return Ok("WikiData");
    }
    if is_valid_therapeutic_target_database_node_name(node_name) {
        return Ok("Therapeutic Target Database");
    }
    if is_valid_reactome_node_name(node_name) {
        return Ok("Reactome");
    }

    if is_valid_jackson_lab_reference_node_name(node_name) {
        return Ok("Jackson Lab Reference");
    }

    if is_valid_uberon_core_node_name(node_name) {
        return Ok("Uberon Core");
    }

    if is_valid_gene_ontology_relations_node_name(node_name) {
        return Ok("Gene Ontology Relations");
    }

    if is_valid_quantity_unit_dimension_and_type_node_name(node_name) {
        return Ok("Quantity, Unit, Dimension and Type");
    }

    if is_valid_open_biomedical_association_node_name(node_name) {
        return Ok("Open Biomedical Association");
    }

    if is_valid_kegg_homo_sapiens_node_name(node_name) {
        return Ok("KEGG Homo sapiens");
    }

    if is_valid_dublin_core_node_name(node_name) {
        return Ok("Dublin Core");
    }

    if is_valid_xsd_node_name(node_name) {
        return Ok("XSD");
    }

    if is_valid_kegg_ko_node_name(node_name) {
        return Ok("KEGG-KO");
    }

    if is_valid_swiss_protein_node_name(node_name) {
        return Ok("Swiss Protein");
    }

    if is_valid_ncbi_books_node_name(node_name) {
        return Ok("NCBI Books");
    }

    if is_valid_chembl_compound_node_name(node_name) {
        return Ok("Chembl Compound");
    }

    if is_valid_hugo_gene_name_consortium_node_name(node_name) {
        return Ok("Hugo Gene Name Consortium");
    }

    if is_valid_provenance_authoring_and_versioning_node_name(node_name) {
        return Ok("Provenance, Authoring and Versioning");
    }

    if is_valid_feature_annotation_location_description_ontology_node_name(node_name) {
        return Ok("Feature Annotation Location Description Ontology");
    }

    if is_valid_rbrc_node_name(node_name) {
        return Ok("RBRC");
    }

    if is_valid_dictybase_gene_node_name(node_name) {
        return Ok("dictyBase Gene");
    }

    if is_valid_edam_data_node_name(node_name) {
        return Ok("EDAM Data");
    }

    if is_valid_ncimr_node_name(node_name) {
        return Ok("NCIMR");
    }

    if is_valid_ecogene_node_name(node_name) {
        return Ok("EcoGene");
    }

    if is_valid_aspergillus_genomic_data_reference_node_name(node_name) {
        return Ok("Aspergillus Genomic Data Reference");
    }

    if is_valid_european_mouse_mutant_archive_node_name(node_name) {
        return Ok("European Mouse Mutant Archive");
    }

    if is_valid_translation_of_embl_nucleotide_sequence_database_node_name(node_name) {
        return Ok("Translation of EMBL nucleotide sequence database");
    }

    if is_valid_mouse_phenome_database_strain_node_name(node_name) {
        return Ok("Mouse Phenome Database Strain");
    }

    if is_valid_basic_formal_ontology_node_name(node_name) {
        return Ok("Basic Formal Ontology");
    }

    if is_valid_dictybase_genes_node_name(node_name) {
        return Ok("Dictybase Genes");
    }

    if is_valid_cell_line_ontology_node_name(node_name) {
        return Ok("Cell Line Ontology");
    }

    if is_valid_animal_genome_horse_qtl_node_name(node_name) {
        return Ok("Animal Genome Horse QTL");
    }

    if is_valid_wormbase_node_name(node_name) {
        return Ok("WormBase");
    }

    if is_valid_animal_qtl_traits_node_name(node_name) {
        return Ok("Animal QTL Traits");
    }

    if is_valid_friend_of_a_friend_node_name(node_name) {
        return Ok("Friend of a Friend");
    }

    if is_valid_omim_phenotypic_series_node_name(node_name) {
        return Ok("OMIM Phenotypic Series");
    }

    if is_valid_arabidopsis_tair_node_name(node_name) {
        return Ok("Arabidopsis Tair");
    }

    if is_valid_unified_medical_language_system_node_name(node_name) {
        return Ok("Unified Medical Language System");
    }

    if is_valid_gene_ontology_reference_node_name(node_name) {
        return Ok("Gene Ontology Reference");
    }

    if is_valid_ncbi_clinical_variants_submitters_node_name(node_name) {
        return Ok("NCBI Clinical Variants Submitters");
    }

    if is_valid_gdc_project_node_name(node_name) {
        return Ok("GDC Project");
    }

    if is_valid_complex_portal_node_name(node_name) {
        return Ok("Complex Portal");
    }

    if is_valid_ucsc_golden_path_node_name(node_name) {
        return Ok("UCSC Golden Path");
    }

    if is_valid_yeast_genome_reference_node_name(node_name) {
        return Ok("Yeast Genome Reference");
    }

    if is_valid_xenbase_node_name(node_name) {
        return Ok("XENBASE");
    }

    if is_valid_chembl_target_node_name(node_name) {
        return Ok("CHEMBL Target");
    }

    if is_valid_coriell_collection_node_name(node_name) {
        return Ok("Coriell Collection");
    }

    if is_valid_gdc_annotation_node_name(node_name) {
        return Ok("GDC Annotation");
    }

    if is_valid_animal_genome_chicken_qtl_node_name(node_name) {
        return Ok("Animal Genome Chicken QTL");
    }

    if is_valid_semantic_medline_database_node_name(node_name) {
        return Ok("Semantic MEDLINE Database");
    }

    if is_valid_animal_genome_rainbow_trout_qtl_node_name(node_name) {
        return Ok("Animal Genome Rainbow-Trout QTL");
    }

    if is_valid_animal_genome_pig_qtl_node_name(node_name) {
        return Ok("Animal Genome Pig QTL");
    }

    if is_valid_livestock_product_trait_node_name(node_name) {
        return Ok("Livestock Product Trait");
    }

    if is_valid_variation_representation_specification_node_name(node_name) {
        return Ok("Variation Representation Specification");
    }

    if is_valid_semantic_science_node_name(node_name) {
        return Ok("Semantic Science");
    }

    if is_valid_experimental_factor_ontology_node_name(node_name) {
        return Ok("Experimental Factor Ontology");
    }

    if is_valid_animal_genome_sheep_qtl_node_name(node_name) {
        return Ok("Animal Genome Sheep QTL");
    }

    if is_valid_open_biomedical_ontology_node_name(node_name) {
        return Ok("Open Biomedical Ontology");
    }

    if is_valid_intact_node_name(node_name) {
        return Ok("INTACT");
    }

    if is_valid_game_ontology_project_node_name(node_name) {
        return Ok("Game Ontology Project");
    }

    if is_valid_apergillus_genome_data_node_name(node_name) {
        return Ok("Apergillus Genome Data");
    }

    if is_valid_animal_genome_pubblications_node_name(node_name) {
        return Ok("Animal Genome Pubblications");
    }

    if is_valid_monarch_initiative_node_name(node_name) {
        return Ok("Monarch Initiative");
    }

    if is_valid_snp_individual_node_name(node_name) {
        return Ok("SNP Individual");
    }

    if is_valid_ncbi_protein_node_name(node_name) {
        return Ok("NCBI Protein");
    }

    if is_valid_orpha_node_name(node_name) {
        return Ok("ORPHA");
    }

    if is_valid_wormbase_vocabulary_node_name(node_name) {
        return Ok("Wormbase Vocabulary");
    }

    if is_valid_kegg_path_node_name(node_name) {
        return Ok("KEGG Path");
    }

    if is_valid_vertebrate_gene_names_consortium_node_name(node_name) {
        return Ok("Vertebrate Gene Names Consortium");
    }

    if is_valid_yeast_genome_locus_node_name(node_name) {
        return Ok("Yeast Genome Locus");
    }

    if is_valid_schema_node_name(node_name) {
        return Ok("SCHEMA");
    }

    if is_valid_provenance_ontology_node_name(node_name) {
        return Ok("Provenance Ontology");
    }

    if is_valid_kegg_disease_node_name(node_name) {
        return Ok("KEGG Disease");
    }

    if is_valid_database_snp_node_name(node_name) {
        return Ok("Database SNP");
    }

    if is_valid_cell_ontology_node_name(node_name) {
        return Ok("Cell Ontology");
    }

    if is_valid_mugen_node_name(node_name) {
        return Ok("Mugen");
    }

    if is_valid_world_geodetic_system_node_name(node_name) {
        return Ok("World Geodetic System");
    }

    if is_valid_chromosome_ontology_node_name(node_name) {
        return Ok("Chromosome Ontology");
    }

    if is_valid_pharmgkb_node_name(node_name) {
        return Ok("PHARMGKB");
    }

    if is_valid_the_arabidopsis_information_resource_locus_node_name(node_name) {
        return Ok("The Arabidopsis Information Resource Locus");
    }

    if is_valid_owl_node_name(node_name) {
        return Ok("OWL");
    }

    if is_valid_void_node_name(node_name) {
        return Ok("VOID");
    }

    if is_valid_rdf_node_name(node_name) {
        return Ok("RDF");
    }

    if is_valid_rgd_reference_node_name(node_name) {
        return Ok("RGD Reference");
    }

    if is_valid_mouse_phenome_database_assay_node_name(node_name) {
        return Ok("Mouse Phenome Database Assay");
    }

    if is_valid_drugbank_node_name(node_name) {
        return Ok("DrugBank");
    }

    if is_valid_coriell_family_node_name(node_name) {
        return Ok("Coriell Family");
    }

    if is_valid_pombase_node_name(node_name) {
        return Ok("POMBASE");
    }

    if is_valid_online_mendelian_inheritance_in_animals_node_name(node_name) {
        return Ok("Online Mendelian Inheritance in Animals");
    }

    if is_valid_interpro_node_name(node_name) {
        return Ok("InterPro");
    }

    if is_valid_panther_database_node_name(node_name) {
        return Ok("Panther Database");
    }

    if is_valid_human_developmental_stages_node_name(node_name) {
        return Ok("Human Developmental Stages");
    }

    if is_valid_mouse_pathology_ontology_node_name(node_name) {
        return Ok("Mouse pathology ontology");
    }

    if is_valid_unified_phenotype_ontology_upheno_node_name(node_name) {
        return Ok("Unified phenotype ontology (uPheno)");
    }

    if is_valid_human_phenotype_ontology_node_name(node_name) {
        return Ok("Human Phenotype Ontology");
    }

    if is_valid_protein_ontology_pro_node_name(node_name) {
        return Ok("PRotein Ontology (PRO)");
    }

    if is_valid_human_disease_ontology_node_name(node_name) {
        return Ok("Human Disease Ontology");
    }

    if is_valid_zebrafish_anatomy_and_development_ontology_node_name(node_name) {
        return Ok("Zebrafish anatomy and development ontology");
    }

    if is_valid_an_ontology_of_core_ecological_entities_node_name(node_name) {
        return Ok("An ontology of core ecological entities");
    }

    if is_valid_pathway_ontology_node_name(node_name) {
        return Ok("Pathway ontology");
    }

    if is_valid_ascomycete_phenotype_ontology_node_name(node_name) {
        return Ok("Ascomycete phenotype ontology");
    }

    if is_valid_environment_ontology_node_name(node_name) {
        return Ok("Environment Ontology");
    }

    if is_valid_nci_thesaurus_obo_edition_node_name(node_name) {
        return Ok("NCI Thesaurus OBO Edition");
    }

    if is_valid_mouse_adult_gross_anatomy_node_name(node_name) {
        return Ok("Mouse adult gross anatomy");
    }

    if is_valid_neuro_behavior_ontology_node_name(node_name) {
        return Ok("Neuro Behavior Ontology");
    }

    if is_valid_chemical_methods_ontology_node_name(node_name) {
        return Ok("Chemical Methods Ontology");
    }

    if is_valid_zebrafish_developmental_stages_ontology_node_name(node_name) {
        return Ok("Zebrafish developmental stages ontology");
    }

    if is_valid_fungal_gross_anatomy_node_name(node_name) {
        return Ok("Fungal gross anatomy");
    }

    if is_valid_zebrafish_phenotype_ontology_node_name(node_name) {
        return Ok("Zebrafish Phenotype Ontology");
    }

    if is_valid_ontology_of_biological_attributes_node_name(node_name) {
        return Ok("Ontology of Biological Attributes");
    }

    if is_valid_confidence_information_ontology_node_name(node_name) {
        return Ok("Confidence Information Ontology");
    }

    if is_valid_ncbi_gene_expression_omnibus_node_name(node_name) {
        return Ok("NCBI Gene Expression Omnibus");
    }

    if is_valid_genotype_ontology_node_name(node_name) {
        return Ok("Genotype Ontology");
    }

    if is_valid_ontology_of_adverse_events_node_name(node_name) {
        return Ok("Ontology of Adverse Events");
    }

    if is_valid_experimental_condition_ontology_node_name(node_name) {
        return Ok("Experimental condition ontology");
    }

    if is_valid_plant_ontology_node_name(node_name) {
        return Ok("Plant Ontology");
    }

    if is_valid_environmental_conditions_treatments_and_exposures_ontology_node_name(node_name) {
        return Ok("Environmental conditions, treatments and exposures ontology");
    }

    if is_valid_units_of_measurement_ontology_node_name(node_name) {
        return Ok("Units of measurement ontology");
    }

    if is_valid_ontology_for_biomedical_investigations_node_name(node_name) {
        return Ok("Ontology for Biomedical Investigations");
    }

    if is_valid_c_elegans_phenotype_node_name(node_name) {
        return Ok("C. elegans phenotype");
    }

    if is_valid_biological_spatial_ontology_node_name(node_name) {
        return Ok("Biological Spatial Ontology");
    }

    if is_valid_scientific_evidence_and_provenance_information_ontology_node_name(node_name) {
        return Ok("Scientific Evidence and Provenance Information Ontology");
    }

    if is_valid_drosophila_gross_anatomy_node_name(node_name) {
        return Ok("Drosophila gross anatomy");
    }

    if is_valid_the_statistical_methods_ontology_node_name(node_name) {
        return Ok("The Statistical Methods Ontology");
    }

    if is_valid_infectious_disease_ontology_node_name(node_name) {
        return Ok("Infectious Disease Ontology");
    }

    if is_valid_drosophila_development_node_name(node_name) {
        return Ok("Drosophila development");
    }

    if is_valid_mental_disease_ontology_node_name(node_name) {
        return Ok("Mental Disease Ontology");
    }

    if is_valid_human_developmental_anatomy_abstract_node_name(node_name) {
        return Ok("Human developmental anatomy, abstract");
    }

    if is_valid_c_elegans_gross_anatomy_ontology_node_name(node_name) {
        return Ok("C. elegans Gross Anatomy Ontology");
    }

    if is_valid_phenotype_and_trait_ontology_node_name(node_name) {
        return Ok("Phenotype And Trait Ontology");
    }

    if is_valid_evidence_ontology_node_name(node_name) {
        return Ok("Evidence ontology");
    }

    if is_valid_information_artifact_ontology_node_name(node_name) {
        return Ok("Information Artifact Ontology");
    }

    if is_valid_brenda_tissue_enzyme_source_node_name(node_name) {
        return Ok("BRENDA tissue / enzyme source");
    }

    if is_valid_mouse_developmental_anatomy_ontology_node_name(node_name) {
        return Ok("Mouse Developmental Anatomy Ontology");
    }

    if is_valid_mammalian_phenotype_ontology_node_name(node_name) {
        return Ok("Mammalian Phenotype Ontology");
    }

    if is_valid_foodon_node_name(node_name) {
        return Ok("FOODON");
    }

    if is_valid_vertebrate_trait_ontology_node_name(node_name) {
        return Ok("Vertebrate trait ontology");
    }

    if is_valid_exposure_ontology_node_name(node_name) {
        return Ok("Exposure ontology");
    }

    if is_valid_population_and_community_ontology_node_name(node_name) {
        return Ok("Population and Community Ontology");
    }

    if is_valid_clinical_measurement_ontology_node_name(node_name) {
        return Ok("Clinical measurement ontology");
    }

    if is_valid_mondo_disease_ontology_node_name(node_name) {
        return Ok("Mondo Disease Ontology");
    }

    if is_valid_ontology_for_general_medical_science_node_name(node_name) {
        return Ok("Ontology for General Medical Science");
    }

    if is_valid_common_anatomy_reference_ontology_node_name(node_name) {
        return Ok("Common Anatomy Reference Ontology");
    }

    if is_valid_gazetteer_node_name(node_name) {
        return Ok("Gazetteer");
    }

    if is_valid_uberon_multispecies_anatomy_ontology_node_name(node_name) {
        return Ok("Uberon multi-species anatomy ontology");
    }

    if is_valid_mental_functioning_ontology_node_name(node_name) {
        return Ok("Mental Functioning Ontology");
    }

    if is_valid_relation_ontology_node_name(node_name) {
        return Ok("Relation Ontology");
    }

    if is_valid_ontology_for_parasite_lifecycle_node_name(node_name) {
        return Ok("Ontology for Parasite LifeCycle");
    }

    if is_valid_dictyostelium_discoideum_anatomy_node_name(node_name) {
        return Ok("Dictyostelium discoideum anatomy");
    }

    if is_valid_emotion_ontology_node_name(node_name) {
        return Ok("Emotion Ontology");
    }

    if is_valid_eaglei_resource_ontology_node_name(node_name) {
        return Ok("eagle-i resource ontology");
    }

    if is_valid_dublin_core_terms_node_name(node_name) {
        return Ok("Dublin Core Terms");
    }

    if is_valid_dublin_core_types_node_name(node_name) {
        return Ok("Dublin Core Types");
    }

    if is_valid_flybase_controlled_vocabulary_node_name(node_name) {
        return Ok("FlyBase Controlled Vocabulary");
    }

    if is_valid_foundational_model_of_anatomy_ontology_subset_node_name(node_name) {
        return Ok("Foundational Model of Anatomy Ontology (subset)");
    }

    if is_valid_gene_ontology_obo_in_owl_node_name(node_name) {
        return Ok("Gene Ontology OBO in OWL");
    }

    if is_valid_monarch_initiative_archive_node_name(node_name) {
        return Ok("Monarch Initiative Archive");
    }

    if is_valid_monarch_initiative_data_node_name(node_name) {
        return Ok("Monarch Initiative Data");
    }

    if is_valid_protein_modification_node_name(node_name) {
        return Ok("Protein modification");
    }

    if is_valid_rdfs_node_name(node_name) {
        return Ok("RDFS");
    }

    if is_valid_unified_medical_language_system_semantic_code_node_name(node_name) {
        return Ok("Unified Medical Language System Semantic Code");
    }

    if is_valid_unified_medical_language_system_semantic_type_node_name(node_name) {
        return Ok("Unified Medical Language System Semantic Type");
    }

    if is_valid_unified_medical_language_system_semantic_group_node_name(node_name) {
        return Ok("Unified Medical Language System Semantic Group");
    }

    if is_valid_cord_pubmed_central_node_name(node_name) {
        return Ok("CORD Pubmed Central");
    }

    if may_be_string_node_name(node_name) {
        if is_valid_string_pseudomonas_aeruginosa_node_name(node_name) {
            return Ok("STRING Pseudomonas aeruginosa");
        }

        if is_valid_string_burkholderia_cepacia_node_name(node_name) {
            return Ok("STRING Burkholderia cepacia");
        }

        if is_valid_string_pseudomonas_oleovorans_node_name(node_name) {
            return Ok("STRING Pseudomonas oleovorans");
        }

        if is_valid_string_xanthomonas_campestris_campestris_node_name(node_name) {
            return Ok("STRING Xanthomonas campestris campestris");
        }

        if is_valid_string_agrobacterium_rhizogenes_node_name(node_name) {
            return Ok("STRING Agrobacterium rhizogenes");
        }

        if is_valid_string_sinorhizobium_fredii_ngr234_node_name(node_name) {
            return Ok("STRING Sinorhizobium fredii NGR234");
        }

        if is_valid_string_acinetobacter_baumannii_node_name(node_name) {
            return Ok("STRING Acinetobacter baumannii");
        }

        if is_valid_string_alcaligenes_faecalis_node_name(node_name) {
            return Ok("STRING Alcaligenes faecalis");
        }

        if is_valid_string_pantoea_agglomerans_node_name(node_name) {
            return Ok("STRING Pantoea agglomerans");
        }

        if is_valid_string_klebsiella_oxytoca_node_name(node_name) {
            return Ok("STRING Klebsiella oxytoca");
        }

        if is_valid_string_klebsiella_pneumoniae_node_name(node_name) {
            return Ok("STRING Klebsiella pneumoniae");
        }

        if is_valid_string_proteus_vulgaris_node_name(node_name) {
            return Ok("STRING Proteus vulgaris");
        }

        if is_valid_string_yersinia_pseudotuberculosis_node_name(node_name) {
            return Ok("STRING Yersinia pseudotuberculosis");
        }

        if is_valid_string_vibrio_vulnificus_node_name(node_name) {
            return Ok("STRING Vibrio vulnificus");
        }

        if is_valid_string_pasteurella_multocida_node_name(node_name) {
            return Ok("STRING Pasteurella multocida");
        }

        if is_valid_string_pasteurella_pneumotropica_node_name(node_name) {
            return Ok("STRING Pasteurella pneumotropica");
        }

        if is_valid_string_chlamydia_trachomatis_node_name(node_name) {
            return Ok("STRING Chlamydia trachomatis");
        }

        if is_valid_string_campylobacter_ureolyticus_node_name(node_name) {
            return Ok("STRING Campylobacter ureolyticus");
        }

        if is_valid_string_desulfovibrio_vulgaris_hildenborough_node_name(node_name) {
            return Ok("STRING Desulfovibrio vulgaris Hildenborough");
        }

        if is_valid_string_desulfovibrio_vulgaris_miyazaki_node_name(node_name) {
            return Ok("STRING Desulfovibrio vulgaris Miyazaki");
        }

        if is_valid_string_flavobacterium_hydatis_node_name(node_name) {
            return Ok("STRING Flavobacterium hydatis");
        }

        if is_valid_string_erythrobacter_longus_node_name(node_name) {
            return Ok("STRING Erythrobacter longus");
        }

        if is_valid_string_synechococcus_elongatus_pcc7942_node_name(node_name) {
            return Ok("STRING Synechococcus elongatus PCC7942");
        }

        if is_valid_string_synechocystis_sp_pcc6714_node_name(node_name) {
            return Ok("STRING Synechocystis sp. PCC6714");
        }

        if is_valid_string_synechocystis_sp_pcc6803_node_name(node_name) {
            return Ok("STRING Synechocystis sp. PCC6803");
        }

        if is_valid_string_leuconostoc_lactis_node_name(node_name) {
            return Ok("STRING Leuconostoc lactis");
        }

        if is_valid_string_dermacoccus_nishinomiyaensis_node_name(node_name) {
            return Ok("STRING Dermacoccus nishinomiyaensis");
        }

        if is_valid_string_staphylococcus_aureus_node_name(node_name) {
            return Ok("STRING Staphylococcus aureus");
        }

        if is_valid_string_staphylococcus_xylosus_node_name(node_name) {
            return Ok("STRING Staphylococcus xylosus");
        }

        if is_valid_string_streptococcus_pyogenes_node_name(node_name) {
            return Ok("STRING Streptococcus pyogenes");
        }

        if is_valid_string_streptococcus_iniae_node_name(node_name) {
            return Ok("STRING Streptococcus iniae");
        }

        if is_valid_string_bacillus_cereus_node_name(node_name) {
            return Ok("STRING Bacillus cereus");
        }

        if is_valid_string_bacillus_mycoides_node_name(node_name) {
            return Ok("STRING Bacillus mycoides");
        }

        if is_valid_string_clostridium_sp_atcc29733_node_name(node_name) {
            return Ok("STRING Clostridium sp. ATCC29733");
        }

        if is_valid_string_clostridium_sticklandii_node_name(node_name) {
            return Ok("STRING Clostridium sticklandii");
        }

        if is_valid_string_lactobacillus_animalis_node_name(node_name) {
            return Ok("STRING Lactobacillus animalis");
        }

        if is_valid_string_bifidobacterium_adolescentis_node_name(node_name) {
            return Ok("STRING Bifidobacterium adolescentis");
        }

        if is_valid_string_bifidobacterium_coryneforme_node_name(node_name) {
            return Ok("STRING Bifidobacterium coryneforme");
        }

        if is_valid_string_bifidobacterium_cuniculi_node_name(node_name) {
            return Ok("STRING Bifidobacterium cuniculi");
        }

        if is_valid_string_bifidobacterium_pseudolongum_globosum_node_name(node_name) {
            return Ok("STRING Bifidobacterium pseudolongum globosum");
        }

        if is_valid_string_bifidobacterium_magnum_node_name(node_name) {
            return Ok("STRING Bifidobacterium magnum");
        }

        if is_valid_string_bifidobacterium_minimum_node_name(node_name) {
            return Ok("STRING Bifidobacterium minimum");
        }

        if is_valid_string_oerskovia_turbata_node_name(node_name) {
            return Ok("STRING Oerskovia turbata");
        }

        if is_valid_string_corynebacterium_pseudotuberculosis_node_name(node_name) {
            return Ok("STRING Corynebacterium pseudotuberculosis");
        }

        if is_valid_string_rhodococcus_fascians_node_name(node_name) {
            return Ok("STRING Rhodococcus fascians");
        }

        if is_valid_string_actinoplanes_utahensis_node_name(node_name) {
            return Ok("STRING Actinoplanes utahensis");
        }

        if is_valid_string_streptomyces_aureofaciens_node_name(node_name) {
            return Ok("STRING Streptomyces aureofaciens");
        }

        if is_valid_string_streptomyces_bikiniensis_node_name(node_name) {
            return Ok("STRING Streptomyces bikiniensis");
        }

        if is_valid_string_streptomyces_fradiae_node_name(node_name) {
            return Ok("STRING Streptomyces fradiae");
        }

        if is_valid_string_streptomyces_glaucescens_node_name(node_name) {
            return Ok("STRING Streptomyces glaucescens");
        }

        if is_valid_string_streptomyces_halstedii_node_name(node_name) {
            return Ok("STRING Streptomyces halstedii");
        }

        if is_valid_string_streptomyces_sclerotialus_node_name(node_name) {
            return Ok("STRING Streptomyces sclerotialus");
        }

        if is_valid_string_streptomyces_virginiae_node_name(node_name) {
            return Ok("STRING Streptomyces virginiae");
        }

        if is_valid_string_streptomyces_cellulosae_node_name(node_name) {
            return Ok("STRING Streptomyces cellulosae");
        }

        if is_valid_string_microtetraspora_glauca_node_name(node_name) {
            return Ok("STRING Microtetraspora glauca");
        }

        if is_valid_string_streptosporangium_amethystogenes_node_name(node_name) {
            return Ok("STRING Streptosporangium amethystogenes");
        }

        if is_valid_string_pimelobacter_simplex_node_name(node_name) {
            return Ok("STRING Pimelobacter simplex");
        }

        if is_valid_string_gordonia_terrae_node_name(node_name) {
            return Ok("STRING Gordonia terrae");
        }

        if is_valid_string_pseudonocardia_autotrophica_node_name(node_name) {
            return Ok("STRING Pseudonocardia autotrophica");
        }

        if is_valid_string_mycoplasma_agalactiae_node_name(node_name) {
            return Ok("STRING Mycoplasma agalactiae");
        }

        if is_valid_string_mycoplasma_californicum_node_name(node_name) {
            return Ok("STRING Mycoplasma californicum");
        }

        if is_valid_string_thermoanaerobacter_kivui_node_name(node_name) {
            return Ok("STRING Thermoanaerobacter kivui");
        }

        if is_valid_string_solemya_velum_node_name(node_name) {
            return Ok("STRING Solemya velum");
        }

        if is_valid_string_sodalis_pierantonius_node_name(node_name) {
            return Ok("STRING Sodalis pierantonius");
        }

        if is_valid_string_fervidobacterium_islandicum_node_name(node_name) {
            return Ok("STRING Fervidobacterium islandicum");
        }

        if is_valid_string_citrus_sinensis_node_name(node_name) {
            return Ok("STRING Citrus sinensis");
        }

        if is_valid_string_synergistes_jonesii_node_name(node_name) {
            return Ok("STRING Synergistes jonesii");
        }

        if is_valid_string_phaeodactylum_tricornutum_node_name(node_name) {
            return Ok("STRING Phaeodactylum tricornutum");
        }

        if is_valid_string_ectocarpus_siliculosus_node_name(node_name) {
            return Ok("STRING Ectocarpus siliculosus");
        }

        if is_valid_string_emiliania_huxleyi_node_name(node_name) {
            return Ok("STRING Emiliania huxleyi");
        }

        if is_valid_string_chlamydomonas_reinhardtii_node_name(node_name) {
            return Ok("STRING Chlamydomonas reinhardtii");
        }

        if is_valid_string_volvox_carteri_node_name(node_name) {
            return Ok("STRING Volvox carteri");
        }

        if is_valid_string_auxenochlorella_protothecoides_node_name(node_name) {
            return Ok("STRING Auxenochlorella protothecoides");
        }

        if is_valid_string_physcomitrella_patens_node_name(node_name) {
            return Ok("STRING Physcomitrella patens");
        }

        if is_valid_string_theobroma_cacao_node_name(node_name) {
            return Ok("STRING Theobroma cacao");
        }

        if is_valid_string_carica_papaya_node_name(node_name) {
            return Ok("STRING Carica papaya");
        }

        if is_valid_string_cucumis_melo_node_name(node_name) {
            return Ok("STRING Cucumis melo");
        }

        if is_valid_string_cucumis_sativus_node_name(node_name) {
            return Ok("STRING Cucumis sativus");
        }

        if is_valid_string_populus_trichocarpa_node_name(node_name) {
            return Ok("STRING Populus trichocarpa");
        }

        if is_valid_string_arabidopsis_thaliana_node_name(node_name) {
            return Ok("STRING Arabidopsis thaliana");
        }

        if is_valid_string_brassica_rapa_node_name(node_name) {
            return Ok("STRING Brassica rapa");
        }

        if is_valid_string_brassica_oleracea_node_name(node_name) {
            return Ok("STRING Brassica oleracea");
        }

        if is_valid_string_malus_domestica_node_name(node_name) {
            return Ok("STRING Malus domestica");
        }

        if is_valid_string_prunus_persica_node_name(node_name) {
            return Ok("STRING Prunus persica");
        }

        if is_valid_string_cicer_arietinum_node_name(node_name) {
            return Ok("STRING Cicer arietinum");
        }

        if is_valid_string_glycine_max_node_name(node_name) {
            return Ok("STRING Glycine max");
        }

        if is_valid_string_medicago_truncatula_node_name(node_name) {
            return Ok("STRING Medicago truncatula");
        }

        if is_valid_string_phaseolus_vulgaris_node_name(node_name) {
            return Ok("STRING Phaseolus vulgaris");
        }

        if is_valid_string_manihot_esculenta_node_name(node_name) {
            return Ok("STRING Manihot esculenta");
        }

        if is_valid_string_ricinus_communis_node_name(node_name) {
            return Ok("STRING Ricinus communis");
        }

        if is_valid_string_linum_usitatissimum_node_name(node_name) {
            return Ok("STRING Linum usitatissimum");
        }

        if is_valid_string_solanum_lycopersicum_node_name(node_name) {
            return Ok("STRING Solanum lycopersicum");
        }

        if is_valid_string_nicotiana_sylvestris_node_name(node_name) {
            return Ok("STRING Nicotiana sylvestris");
        }

        if is_valid_string_nicotiana_tomentosiformis_node_name(node_name) {
            return Ok("STRING Nicotiana tomentosiformis");
        }

        if is_valid_string_solanum_tuberosum_node_name(node_name) {
            return Ok("STRING Solanum tuberosum");
        }

        if is_valid_string_erythranthe_guttata_node_name(node_name) {
            return Ok("STRING Erythranthe guttata");
        }

        if is_valid_string_nelumbo_nucifera_node_name(node_name) {
            return Ok("STRING Nelumbo nucifera");
        }

        if is_valid_string_hordeum_vulgare_node_name(node_name) {
            return Ok("STRING Hordeum vulgare");
        }

        if is_valid_string_oryza_rufipogon_node_name(node_name) {
            return Ok("STRING Oryza rufipogon");
        }

        if is_valid_string_oryza_sativa_node_name(node_name) {
            return Ok("STRING Oryza sativa");
        }

        if is_valid_string_oryza_brachyantha_node_name(node_name) {
            return Ok("STRING Oryza brachyantha");
        }

        if is_valid_string_oryza_nivara_node_name(node_name) {
            return Ok("STRING Oryza nivara");
        }

        if is_valid_string_oryza_punctata_node_name(node_name) {
            return Ok("STRING Oryza punctata");
        }

        if is_valid_string_oryza_glaberrima_node_name(node_name) {
            return Ok("STRING Oryza glaberrima");
        }

        if is_valid_string_setaria_italica_node_name(node_name) {
            return Ok("STRING Setaria italica");
        }

        if is_valid_string_sorghum_bicolor_node_name(node_name) {
            return Ok("STRING Sorghum bicolor");
        }

        if is_valid_string_triticum_aestivum_node_name(node_name) {
            return Ok("STRING Triticum aestivum");
        }

        if is_valid_string_triticum_urartu_node_name(node_name) {
            return Ok("STRING Triticum urartu");
        }

        if is_valid_string_zea_mays_node_name(node_name) {
            return Ok("STRING Zea mays");
        }

        if is_valid_string_musa_acuminata_node_name(node_name) {
            return Ok("STRING Musa acuminata");
        }

        if is_valid_string_phytophthora_infestans_node_name(node_name) {
            return Ok("STRING Phytophthora infestans");
        }

        if is_valid_string_phytophthora_parasitica_node_name(node_name) {
            return Ok("STRING Phytophthora parasitica");
        }

        if is_valid_string_schizosaccharomyces_pombe_node_name(node_name) {
            return Ok("STRING Schizosaccharomyces pombe");
        }

        if is_valid_string_schizosaccharomyces_japonicus_node_name(node_name) {
            return Ok("STRING Schizosaccharomyces japonicus");
        }

        if is_valid_string_schizosaccharomyces_octosporus_node_name(node_name) {
            return Ok("STRING Schizosaccharomyces octosporus");
        }

        if is_valid_string_pichia_kudriavzevii_node_name(node_name) {
            return Ok("STRING Pichia kudriavzevii");
        }

        if is_valid_string_millerozyma_farinosa_node_name(node_name) {
            return Ok("STRING Millerozyma farinosa");
        }

        if is_valid_string_komagataella_pastoris_node_name(node_name) {
            return Ok("STRING Komagataella pastoris");
        }

        if is_valid_string_scheffersomyces_stipitis_node_name(node_name) {
            return Ok("STRING Scheffersomyces stipitis");
        }

        if is_valid_string_meyerozyma_guilliermondii_node_name(node_name) {
            return Ok("STRING Meyerozyma guilliermondii");
        }

        if is_valid_string_saccharomyces_cerevisiae_node_name(node_name) {
            return Ok("STRING Saccharomyces cerevisiae");
        }

        if is_valid_string_torulaspora_delbrueckii_node_name(node_name) {
            return Ok("STRING Torulaspora delbrueckii");
        }

        if is_valid_string_yarrowia_lipolytica_node_name(node_name) {
            return Ok("STRING Yarrowia lipolytica");
        }

        if is_valid_string_zygosaccharomyces_rouxii_node_name(node_name) {
            return Ok("STRING Zygosaccharomyces rouxii");
        }

        if is_valid_string_debaryomyces_hansenii_node_name(node_name) {
            return Ok("STRING Debaryomyces hansenii");
        }

        if is_valid_string_taphrina_deformans_node_name(node_name) {
            return Ok("STRING Taphrina deformans");
        }

        if is_valid_string_bipolaris_maydis_node_name(node_name) {
            return Ok("STRING Bipolaris maydis");
        }

        if is_valid_string_bipolaris_zeicola_node_name(node_name) {
            return Ok("STRING Bipolaris zeicola");
        }

        if is_valid_string_leptosphaeria_maculans_node_name(node_name) {
            return Ok("STRING Leptosphaeria maculans");
        }

        if is_valid_string_histoplasma_capsulatum_node_name(node_name) {
            return Ok("STRING Histoplasma capsulatum");
        }

        if is_valid_string_blastomyces_dermatitidis_node_name(node_name) {
            return Ok("STRING Blastomyces dermatitidis");
        }

        if is_valid_string_aspergillus_clavatus_node_name(node_name) {
            return Ok("STRING Aspergillus clavatus");
        }

        if is_valid_string_aspergillus_flavus_node_name(node_name) {
            return Ok("STRING Aspergillus flavus");
        }

        if is_valid_string_aspergillus_niger_node_name(node_name) {
            return Ok("STRING Aspergillus niger");
        }

        if is_valid_string_aspergillus_oryzae_node_name(node_name) {
            return Ok("STRING Aspergillus oryzae");
        }

        if is_valid_string_claviceps_purpurea_node_name(node_name) {
            return Ok("STRING Claviceps purpurea");
        }

        if is_valid_string_fusarium_fujikuroi_node_name(node_name) {
            return Ok("STRING Fusarium fujikuroi");
        }

        if is_valid_string_neurospora_crassa_node_name(node_name) {
            return Ok("STRING Neurospora crassa");
        }

        if is_valid_string_podospora_anserina_node_name(node_name) {
            return Ok("STRING Podospora anserina");
        }

        if is_valid_string_sordaria_macrospora_node_name(node_name) {
            return Ok("STRING Sordaria macrospora");
        }

        if is_valid_string_sclerotinia_sclerotiorum_node_name(node_name) {
            return Ok("STRING Sclerotinia sclerotiorum");
        }

        if is_valid_string_cryptococcus_neoformans_node_name(node_name) {
            return Ok("STRING Cryptococcus neoformans");
        }

        if is_valid_string_tremella_mesenterica_node_name(node_name) {
            return Ok("STRING Tremella mesenterica");
        }

        if is_valid_string_ustilago_maydis_node_name(node_name) {
            return Ok("STRING Ustilago maydis");
        }

        if is_valid_string_rhodosporidium_toruloides_node_name(node_name) {
            return Ok("STRING Rhodosporidium toruloides");
        }

        if is_valid_string_puccinia_graminis_node_name(node_name) {
            return Ok("STRING Puccinia graminis");
        }

        if is_valid_string_schizophyllum_commune_node_name(node_name) {
            return Ok("STRING Schizophyllum commune");
        }

        if is_valid_string_agaricus_bisporus_node_name(node_name) {
            return Ok("STRING Agaricus bisporus");
        }

        if is_valid_string_coprinopsis_cinerea_node_name(node_name) {
            return Ok("STRING Coprinopsis cinerea");
        }

        if is_valid_string_colletotrichum_orbiculare_node_name(node_name) {
            return Ok("STRING Colletotrichum orbiculare");
        }

        if is_valid_string_candida_albicans_node_name(node_name) {
            return Ok("STRING Candida albicans");
        }

        if is_valid_string_candida_glabrata_node_name(node_name) {
            return Ok("STRING Candida glabrata");
        }

        if is_valid_string_candida_maltosa_node_name(node_name) {
            return Ok("STRING Candida maltosa");
        }

        if is_valid_string_candida_parapsilosis_node_name(node_name) {
            return Ok("STRING Candida parapsilosis");
        }

        if is_valid_string_candida_tropicalis_node_name(node_name) {
            return Ok("STRING Candida tropicalis");
        }

        if is_valid_string_coccidioides_immitis_node_name(node_name) {
            return Ok("STRING Coccidioides immitis");
        }

        if is_valid_string_fusarium_oxysporum_node_name(node_name) {
            return Ok("STRING Fusarium oxysporum");
        }

        if is_valid_string_fusarium_graminearum_node_name(node_name) {
            return Ok("STRING Fusarium graminearum");
        }

        if is_valid_string_trichophyton_rubrum_node_name(node_name) {
            return Ok("STRING Trichophyton rubrum");
        }

        if is_valid_string_leishmania_major_node_name(node_name) {
            return Ok("STRING Leishmania major");
        }

        if is_valid_string_leishmania_infantum_node_name(node_name) {
            return Ok("STRING Leishmania infantum");
        }

        if is_valid_string_leishmania_panamensis_node_name(node_name) {
            return Ok("STRING Leishmania panamensis");
        }

        if is_valid_string_trypanosoma_brucei_node_name(node_name) {
            return Ok("STRING Trypanosoma brucei");
        }

        if is_valid_string_trypanosoma_cruzi_node_name(node_name) {
            return Ok("STRING Trypanosoma cruzi");
        }

        if is_valid_string_trichomonas_vaginalis_node_name(node_name) {
            return Ok("STRING Trichomonas vaginalis");
        }

        if is_valid_string_giardia_intestinalis_node_name(node_name) {
            return Ok("STRING Giardia intestinalis");
        }

        if is_valid_string_entamoeba_histolytica_node_name(node_name) {
            return Ok("STRING Entamoeba histolytica");
        }

        if is_valid_string_naegleria_gruberi_node_name(node_name) {
            return Ok("STRING Naegleria gruberi");
        }

        if is_valid_string_dictyostelium_purpureum_node_name(node_name) {
            return Ok("STRING Dictyostelium purpureum");
        }

        if is_valid_string_cryptosporidium_muris_node_name(node_name) {
            return Ok("STRING Cryptosporidium muris");
        }

        if is_valid_string_toxoplasma_gondii_node_name(node_name) {
            return Ok("STRING Toxoplasma gondii");
        }

        if is_valid_string_plasmodium_berghei_node_name(node_name) {
            return Ok("STRING Plasmodium berghei");
        }

        if is_valid_string_plasmodium_chabaudi_node_name(node_name) {
            return Ok("STRING Plasmodium chabaudi");
        }

        if is_valid_string_plasmodium_cynomolgi_node_name(node_name) {
            return Ok("STRING Plasmodium cynomolgi");
        }

        if is_valid_string_plasmodium_falciparum_node_name(node_name) {
            return Ok("STRING Plasmodium falciparum");
        }

        if is_valid_string_plasmodium_knowlesi_node_name(node_name) {
            return Ok("STRING Plasmodium knowlesi");
        }

        if is_valid_string_plasmodium_vivax_node_name(node_name) {
            return Ok("STRING Plasmodium vivax");
        }

        if is_valid_string_plasmodium_vinckei_node_name(node_name) {
            return Ok("STRING Plasmodium vinckei");
        }

        if is_valid_string_babesia_bovis_node_name(node_name) {
            return Ok("STRING Babesia bovis");
        }

        if is_valid_string_babesia_equi_node_name(node_name) {
            return Ok("STRING Babesia equi");
        }

        if is_valid_string_theileria_annulata_node_name(node_name) {
            return Ok("STRING Theileria annulata");
        }

        if is_valid_string_theileria_parva_node_name(node_name) {
            return Ok("STRING Theileria parva");
        }

        if is_valid_string_paramecium_tetraurelia_node_name(node_name) {
            return Ok("STRING Paramecium tetraurelia");
        }

        if is_valid_string_tetrahymena_thermophila_node_name(node_name) {
            return Ok("STRING Tetrahymena thermophila");
        }

        if is_valid_string_ichthyophthirius_multifiliis_node_name(node_name) {
            return Ok("STRING Ichthyophthirius multifiliis");
        }

        if is_valid_string_exophiala_dermatitidis_node_name(node_name) {
            return Ok("STRING Exophiala dermatitidis");
        }

        if is_valid_string_hydra_vulgaris_node_name(node_name) {
            return Ok("STRING Hydra vulgaris");
        }

        if is_valid_string_schistosoma_mansoni_node_name(node_name) {
            return Ok("STRING Schistosoma mansoni");
        }

        if is_valid_string_echinococcus_multilocularis_node_name(node_name) {
            return Ok("STRING Echinococcus multilocularis");
        }

        if is_valid_string_caenorhabditis_briggsae_node_name(node_name) {
            return Ok("STRING Caenorhabditis briggsae");
        }

        if is_valid_string_caenorhabditis_elegans_node_name(node_name) {
            return Ok("STRING Caenorhabditis elegans");
        }

        if is_valid_string_bursaphelenchus_xylophilus_node_name(node_name) {
            return Ok("STRING Bursaphelenchus xylophilus");
        }

        if is_valid_string_trichinella_spiralis_node_name(node_name) {
            return Ok("STRING Trichinella spiralis");
        }

        if is_valid_string_helobdella_robusta_node_name(node_name) {
            return Ok("STRING Helobdella robusta");
        }

        if is_valid_string_aplysia_californica_node_name(node_name) {
            return Ok("STRING Aplysia californica");
        }

        if is_valid_string_daphnia_pulex_node_name(node_name) {
            return Ok("STRING Daphnia pulex");
        }

        if is_valid_string_acyrthosiphon_pisum_node_name(node_name) {
            return Ok("STRING Acyrthosiphon pisum");
        }

        if is_valid_string_tribolium_castaneum_node_name(node_name) {
            return Ok("STRING Tribolium castaneum");
        }

        if is_valid_string_bombyx_mori_node_name(node_name) {
            return Ok("STRING Bombyx mori");
        }

        if is_valid_string_aedes_aegypti_node_name(node_name) {
            return Ok("STRING Aedes aegypti");
        }

        if is_valid_string_anopheles_gambiae_node_name(node_name) {
            return Ok("STRING Anopheles gambiae");
        }

        if is_valid_string_culex_quinquefasciatus_node_name(node_name) {
            return Ok("STRING Culex quinquefasciatus");
        }

        if is_valid_string_loa_loa_node_name(node_name) {
            return Ok("STRING Loa loa");
        }

        if is_valid_string_ceratitis_capitata_node_name(node_name) {
            return Ok("STRING Ceratitis capitata");
        }

        if is_valid_string_drosophila_ananassae_node_name(node_name) {
            return Ok("STRING Drosophila ananassae");
        }

        if is_valid_string_drosophila_erecta_node_name(node_name) {
            return Ok("STRING Drosophila erecta");
        }

        if is_valid_string_drosophila_grimshawi_node_name(node_name) {
            return Ok("STRING Drosophila grimshawi");
        }

        if is_valid_string_drosophila_melanogaster_node_name(node_name) {
            return Ok("STRING Drosophila melanogaster");
        }

        if is_valid_string_drosophila_mojavensis_node_name(node_name) {
            return Ok("STRING Drosophila mojavensis");
        }

        if is_valid_string_drosophila_persimilis_node_name(node_name) {
            return Ok("STRING Drosophila persimilis");
        }

        if is_valid_string_drosophila_pseudoobscura_node_name(node_name) {
            return Ok("STRING Drosophila pseudoobscura");
        }

        if is_valid_string_drosophila_virilis_node_name(node_name) {
            return Ok("STRING Drosophila virilis");
        }

        if is_valid_string_drosophila_yakuba_node_name(node_name) {
            return Ok("STRING Drosophila yakuba");
        }

        if is_valid_string_drosophila_willistoni_node_name(node_name) {
            return Ok("STRING Drosophila willistoni");
        }

        if is_valid_string_musca_domestica_node_name(node_name) {
            return Ok("STRING Musca domestica");
        }

        if is_valid_string_nasonia_vitripennis_node_name(node_name) {
            return Ok("STRING Nasonia vitripennis");
        }

        if is_valid_string_apis_mellifera_node_name(node_name) {
            return Ok("STRING Apis mellifera");
        }

        if is_valid_string_strongylocentrotus_purpuratus_node_name(node_name) {
            return Ok("STRING Strongylocentrotus purpuratus");
        }

        if is_valid_string_ciona_intestinalis_node_name(node_name) {
            return Ok("STRING Ciona intestinalis");
        }

        if is_valid_string_branchiostoma_floridae_node_name(node_name) {
            return Ok("STRING Branchiostoma floridae");
        }

        if is_valid_string_latimeria_chalumnae_node_name(node_name) {
            return Ok("STRING Latimeria chalumnae");
        }

        if is_valid_string_lepisosteus_oculatus_node_name(node_name) {
            return Ok("STRING Lepisosteus oculatus");
        }

        if is_valid_string_danio_rerio_node_name(node_name) {
            return Ok("STRING Danio rerio");
        }

        if is_valid_string_astyanax_mexicanus_node_name(node_name) {
            return Ok("STRING Astyanax mexicanus");
        }

        if is_valid_string_esox_lucius_node_name(node_name) {
            return Ok("STRING Esox lucius");
        }

        if is_valid_string_gadus_morhua_node_name(node_name) {
            return Ok("STRING Gadus morhua");
        }

        if is_valid_string_poecilia_reticulata_node_name(node_name) {
            return Ok("STRING Poecilia reticulata");
        }

        if is_valid_string_xiphophorus_maculatus_node_name(node_name) {
            return Ok("STRING Xiphophorus maculatus");
        }

        if is_valid_string_oryzias_latipes_node_name(node_name) {
            return Ok("STRING Oryzias latipes");
        }

        if is_valid_string_oreochromis_niloticus_node_name(node_name) {
            return Ok("STRING Oreochromis niloticus");
        }

        if is_valid_string_haplochromis_burtoni_node_name(node_name) {
            return Ok("STRING Haplochromis burtoni");
        }

        if is_valid_string_xenopus_tropicalis_node_name(node_name) {
            return Ok("STRING Xenopus tropicalis");
        }

        if is_valid_string_chelonia_mydas_node_name(node_name) {
            return Ok("STRING Chelonia mydas");
        }

        if is_valid_string_chrysemys_picta_node_name(node_name) {
            return Ok("STRING Chrysemys picta");
        }

        if is_valid_string_alligator_mississippiensis_node_name(node_name) {
            return Ok("STRING Alligator mississippiensis");
        }

        if is_valid_string_columba_livia_node_name(node_name) {
            return Ok("STRING Columba livia");
        }

        if is_valid_string_gallus_gallus_node_name(node_name) {
            return Ok("STRING Gallus gallus");
        }

        if is_valid_string_ornithorhynchus_anatinus_node_name(node_name) {
            return Ok("STRING Ornithorhynchus anatinus");
        }

        if is_valid_string_sarcophilus_harrisii_node_name(node_name) {
            return Ok("STRING Sarcophilus harrisii");
        }

        if is_valid_string_macropus_eugenii_node_name(node_name) {
            return Ok("STRING Macropus eugenii");
        }

        if is_valid_string_dasypus_novemcinctus_node_name(node_name) {
            return Ok("STRING Dasypus novemcinctus");
        }

        if is_valid_string_erinaceus_europaeus_node_name(node_name) {
            return Ok("STRING Erinaceus europaeus");
        }

        if is_valid_string_echinops_telfairi_node_name(node_name) {
            return Ok("STRING Echinops telfairi");
        }

        if is_valid_string_pteropus_alecto_node_name(node_name) {
            return Ok("STRING Pteropus alecto");
        }

        if is_valid_string_tarsius_syrichta_node_name(node_name) {
            return Ok("STRING Tarsius syrichta");
        }

        if is_valid_string_callithrix_jacchus_node_name(node_name) {
            return Ok("STRING Callithrix jacchus");
        }

        if is_valid_string_macaca_fascicularis_node_name(node_name) {
            return Ok("STRING Macaca fascicularis");
        }

        if is_valid_string_macaca_mulatta_node_name(node_name) {
            return Ok("STRING Macaca mulatta");
        }

        if is_valid_string_papio_anubis_node_name(node_name) {
            return Ok("STRING Papio anubis");
        }

        if is_valid_string_gorilla_gorilla_node_name(node_name) {
            return Ok("STRING Gorilla gorilla");
        }

        if is_valid_string_pan_paniscus_node_name(node_name) {
            return Ok("STRING Pan paniscus");
        }

        if is_valid_string_pan_troglodytes_node_name(node_name) {
            return Ok("STRING Pan troglodytes");
        }

        if is_valid_string_pongo_abelii_node_name(node_name) {
            return Ok("STRING Pongo abelii");
        }

        if is_valid_string_homo_sapiens_node_name(node_name) {
            return Ok("STRING Homo sapiens");
        }

        if is_valid_string_canis_lupus_node_name(node_name) {
            return Ok("STRING Canis lupus");
        }

        if is_valid_string_ailuropoda_melanoleuca_node_name(node_name) {
            return Ok("STRING Ailuropoda melanoleuca");
        }

        if is_valid_string_mustela_putorius_node_name(node_name) {
            return Ok("STRING Mustela putorius");
        }

        if is_valid_string_felis_catus_node_name(node_name) {
            return Ok("STRING Felis catus");
        }

        if is_valid_string_panthera_tigris_node_name(node_name) {
            return Ok("STRING Panthera tigris");
        }

        if is_valid_string_odobenus_rosmarus_node_name(node_name) {
            return Ok("STRING Odobenus rosmarus");
        }

        if is_valid_string_leptonychotes_weddellii_node_name(node_name) {
            return Ok("STRING Leptonychotes weddellii");
        }

        if is_valid_string_orcinus_orca_node_name(node_name) {
            return Ok("STRING Orcinus orca");
        }

        if is_valid_string_tursiops_truncatus_node_name(node_name) {
            return Ok("STRING Tursiops truncatus");
        }

        if is_valid_string_balaenoptera_acutorostrata_node_name(node_name) {
            return Ok("STRING Balaenoptera acutorostrata");
        }

        if is_valid_string_trichechus_manatus_node_name(node_name) {
            return Ok("STRING Trichechus manatus");
        }

        if is_valid_string_loxodonta_africana_node_name(node_name) {
            return Ok("STRING Loxodonta africana");
        }

        if is_valid_string_orycteropus_afer_node_name(node_name) {
            return Ok("STRING Orycteropus afer");
        }

        if is_valid_string_sus_scrofa_node_name(node_name) {
            return Ok("STRING Sus scrofa");
        }

        if is_valid_string_bos_taurus_node_name(node_name) {
            return Ok("STRING Bos taurus");
        }

        if is_valid_string_ovis_aries_node_name(node_name) {
            return Ok("STRING Ovis aries");
        }

        if is_valid_string_ochotona_princeps_node_name(node_name) {
            return Ok("STRING Ochotona princeps");
        }

        if is_valid_string_oryctolagus_cuniculus_node_name(node_name) {
            return Ok("STRING Oryctolagus cuniculus");
        }

        if is_valid_string_cricetulus_griseus_node_name(node_name) {
            return Ok("STRING Cricetulus griseus");
        }

        if is_valid_string_mesocricetus_auratus_node_name(node_name) {
            return Ok("STRING Mesocricetus auratus");
        }

        if is_valid_string_peromyscus_maniculatus_node_name(node_name) {
            return Ok("STRING Peromyscus maniculatus");
        }

        if is_valid_string_mus_musculus_node_name(node_name) {
            return Ok("STRING Mus musculus");
        }

        if is_valid_string_rattus_norvegicus_node_name(node_name) {
            return Ok("STRING Rattus norvegicus");
        }

        if is_valid_string_cavia_porcellus_node_name(node_name) {
            return Ok("STRING Cavia porcellus");
        }

        if is_valid_string_octodon_degus_node_name(node_name) {
            return Ok("STRING Octodon degus");
        }

        if is_valid_string_heterocephalus_glaber_node_name(node_name) {
            return Ok("STRING Heterocephalus glaber");
        }

        if is_valid_string_saccoglossus_kowalevskii_node_name(node_name) {
            return Ok("STRING Saccoglossus kowalevskii");
        }

        if is_valid_string_trichoplax_adhaerens_node_name(node_name) {
            return Ok("STRING Trichoplax adhaerens");
        }

        if is_valid_string_atta_cephalotes_node_name(node_name) {
            return Ok("STRING Atta cephalotes");
        }

        if is_valid_string_dactylococcopsis_salina_node_name(node_name) {
            return Ok("STRING Dactylococcopsis salina");
        }

        if is_valid_string_danaus_plexippus_node_name(node_name) {
            return Ok("STRING Danaus plexippus");
        }

        if is_valid_string_rhodnius_prolixus_node_name(node_name) {
            return Ok("STRING Rhodnius prolixus");
        }

        if is_valid_string_amborella_trichopoda_node_name(node_name) {
            return Ok("STRING Amborella trichopoda");
        }

        if is_valid_string_arthrobotrys_oligospora_node_name(node_name) {
            return Ok("STRING Arthrobotrys oligospora");
        }

        if is_valid_string_monodelphis_domestica_node_name(node_name) {
            return Ok("STRING Monodelphis domestica");
        }

        if is_valid_string_parastagonospora_nodorum_node_name(node_name) {
            return Ok("STRING Parastagonospora nodorum");
        }

        if is_valid_string_sphingomonas_paucimobilis_node_name(node_name) {
            return Ok("STRING Sphingomonas paucimobilis");
        }

        if is_valid_string_sphingobium_yanoikuyae_node_name(node_name) {
            return Ok("STRING Sphingobium yanoikuyae");
        }

        if is_valid_string_pelodiscus_sinensis_node_name(node_name) {
            return Ok("STRING Pelodiscus sinensis");
        }

        if is_valid_string_brachypodium_distachyon_node_name(node_name) {
            return Ok("STRING Brachypodium distachyon");
        }

        if is_valid_string_naumovozyma_castellii_node_name(node_name) {
            return Ok("STRING Naumovozyma castellii");
        }

        if is_valid_string_naumovozyma_dairenensis_node_name(node_name) {
            return Ok("STRING Naumovozyma dairenensis");
        }

        if is_valid_string_verticillium_dahliae_node_name(node_name) {
            return Ok("STRING Verticillium dahliae");
        }

        if is_valid_string_saimiri_boliviensis_node_name(node_name) {
            return Ok("STRING Saimiri boliviensis");
        }

        if is_valid_string_mnemiopsis_leidyi_node_name(node_name) {
            return Ok("STRING Mnemiopsis leidyi");
        }

        if is_valid_string_saccharopolyspora_rectivirgula_node_name(node_name) {
            return Ok("STRING Saccharopolyspora rectivirgula");
        }

        if is_valid_string_nostoc_sp_pcc7524_node_name(node_name) {
            return Ok("STRING Nostoc sp. PCC7524");
        }

        if is_valid_string_porphyromonas_macacae_node_name(node_name) {
            return Ok("STRING Porphyromonas macacae");
        }

        if is_valid_string_yersinia_kristensenii_node_name(node_name) {
            return Ok("STRING Yersinia kristensenii");
        }

        if is_valid_string_photodesmus_katoptron_node_name(node_name) {
            return Ok("STRING Photodesmus katoptron");
        }

        if is_valid_string_colwellia_psychrerythraea_node_name(node_name) {
            return Ok("STRING Colwellia psychrerythraea");
        }

        if is_valid_string_cobetia_marina_node_name(node_name) {
            return Ok("STRING Cobetia marina");
        }

        if is_valid_string_anolis_carolinensis_node_name(node_name) {
            return Ok("STRING Anolis carolinensis");
        }

        if is_valid_string_herbidospora_cretacea_node_name(node_name) {
            return Ok("STRING Herbidospora cretacea");
        }

        if is_valid_string_tarenaya_hassleriana_node_name(node_name) {
            return Ok("STRING Tarenaya hassleriana");
        }

        if is_valid_string_talaromyces_stipitatus_node_name(node_name) {
            return Ok("STRING Talaromyces stipitatus");
        }

        if is_valid_string_allomyces_macrogynus_node_name(node_name) {
            return Ok("STRING Allomyces macrogynus");
        }

        if is_valid_string_elephantulus_edwardii_node_name(node_name) {
            return Ok("STRING Elephantulus edwardii");
        }

        if is_valid_string_kluyveromyces_lactis_node_name(node_name) {
            return Ok("STRING Kluyveromyces lactis");
        }

        if is_valid_string_ursus_maritimus_node_name(node_name) {
            return Ok("STRING Ursus maritimus");
        }

        if is_valid_string_eptesicus_fuscus_node_name(node_name) {
            return Ok("STRING Eptesicus fuscus");
        }

        if is_valid_string_neospora_caninum_node_name(node_name) {
            return Ok("STRING Neospora caninum");
        }

        if is_valid_string_streptomyces_griseoluteus_node_name(node_name) {
            return Ok("STRING Streptomyces griseoluteus");
        }

        if is_valid_string_yersinia_ruckeri_node_name(node_name) {
            return Ok("STRING Yersinia ruckeri");
        }

        if is_valid_string_vibrio_navarrensis_node_name(node_name) {
            return Ok("STRING Vibrio navarrensis");
        }

        if is_valid_string_natrialba_asiatica_node_name(node_name) {
            return Ok("STRING Natrialba asiatica");
        }

        if is_valid_string_janthinobacterium_lividum_node_name(node_name) {
            return Ok("STRING Janthinobacterium lividum");
        }

        if is_valid_string_gossypium_raimondii_node_name(node_name) {
            return Ok("STRING Gossypium raimondii");
        }

        if is_valid_string_vitis_vinifera_node_name(node_name) {
            return Ok("STRING Vitis vinifera");
        }

        if is_valid_string_gaeumannomyces_graminis_node_name(node_name) {
            return Ok("STRING Gaeumannomyces graminis");
        }

        if is_valid_string_trichoderma_virens_node_name(node_name) {
            return Ok("STRING Trichoderma virens");
        }

        if is_valid_string_sporothrix_schenckii_node_name(node_name) {
            return Ok("STRING Sporothrix schenckii");
        }

        if is_valid_string_otolemur_garnettii_node_name(node_name) {
            return Ok("STRING Otolemur garnettii");
        }

        if is_valid_string_takifugu_rubripes_node_name(node_name) {
            return Ok("STRING Takifugu rubripes");
        }

        if is_valid_string_caenorhabditis_remanei_node_name(node_name) {
            return Ok("STRING Caenorhabditis remanei");
        }

        if is_valid_string_colletotrichum_graminicola_node_name(node_name) {
            return Ok("STRING Colletotrichum graminicola");
        }

        if is_valid_string_bifidobacterium_pseudolongum_node_name(node_name) {
            return Ok("STRING Bifidobacterium pseudolongum");
        }

        if is_valid_string_clavibacter_michiganensis_sepedonicus_node_name(node_name) {
            return Ok("STRING Clavibacter michiganensis sepedonicus");
        }

        if is_valid_string_campylobacter_sputorum_node_name(node_name) {
            return Ok("STRING Campylobacter sputorum");
        }

        if is_valid_string_pseudomonas_stutzeri_atcc14405_node_name(node_name) {
            return Ok("STRING Pseudomonas stutzeri ATCC14405");
        }

        if is_valid_string_synechococcus_sp_pcc7002_node_name(node_name) {
            return Ok("STRING Synechococcus sp. PCC7002");
        }

        if is_valid_string_synechococcus_sp_wh_7803_node_name(node_name) {
            return Ok("STRING Synechococcus sp. WH 7803");
        }

        if is_valid_string_calothrix_sp_pcc7103_node_name(node_name) {
            return Ok("STRING Calothrix sp. PCC7103");
        }

        if is_valid_string_tetranychus_urticae_node_name(node_name) {
            return Ok("STRING Tetranychus urticae");
        }

        if is_valid_string_neolamprologus_brichardi_node_name(node_name) {
            return Ok("STRING Neolamprologus brichardi");
        }

        if is_valid_string_blautia_producta_node_name(node_name) {
            return Ok("STRING Blautia producta");
        }

        if is_valid_string_eremothecium_gossypii_node_name(node_name) {
            return Ok("STRING Eremothecium gossypii");
        }

        if is_valid_string_aspergillus_terreus_node_name(node_name) {
            return Ok("STRING Aspergillus terreus");
        }

        if is_valid_string_uncinocarpus_reesii_node_name(node_name) {
            return Ok("STRING Uncinocarpus reesii");
        }

        if is_valid_string_catenuloplanes_japonicus_node_name(node_name) {
            return Ok("STRING Catenuloplanes japonicus");
        }

        if is_valid_string_streptomyces_galbus_node_name(node_name) {
            return Ok("STRING Streptomyces galbus");
        }

        if is_valid_string_bifidobacterium_thermophilum_node_name(node_name) {
            return Ok("STRING Bifidobacterium thermophilum");
        }

        if is_valid_string_paracoccus_versutus_node_name(node_name) {
            return Ok("STRING Paracoccus versutus");
        }

        if is_valid_string_mixia_osmundae_node_name(node_name) {
            return Ok("STRING Mixia osmundae");
        }

        if is_valid_string_blumeria_graminis_node_name(node_name) {
            return Ok("STRING Blumeria graminis");
        }

        if is_valid_string_strongyloides_ratti_node_name(node_name) {
            return Ok("STRING Strongyloides ratti");
        }

        if is_valid_string_heliconius_melpomene_node_name(node_name) {
            return Ok("STRING Heliconius melpomene");
        }

        if is_valid_string_chinchilla_lanigera_node_name(node_name) {
            return Ok("STRING Chinchilla lanigera");
        }

        if is_valid_string_thalassiosira_pseudonana_node_name(node_name) {
            return Ok("STRING Thalassiosira pseudonana");
        }

        if is_valid_string_citrobacter_amalonaticus_node_name(node_name) {
            return Ok("STRING Citrobacter amalonaticus");
        }

        if is_valid_string_thielavia_terrestris_node_name(node_name) {
            return Ok("STRING Thielavia terrestris");
        }

        if is_valid_string_macrophomina_phaseolina_node_name(node_name) {
            return Ok("STRING Macrophomina phaseolina");
        }

        if is_valid_string_dactylosporangium_aurantiacum_node_name(node_name) {
            return Ok("STRING Dactylosporangium aurantiacum");
        }

        if is_valid_string_bifidobacterium_choerinum_node_name(node_name) {
            return Ok("STRING Bifidobacterium choerinum");
        }

        if is_valid_string_bacillus_thermoamylovorans_node_name(node_name) {
            return Ok("STRING Bacillus thermoamylovorans");
        }

        if is_valid_string_vanderwaltozyma_polyspora_node_name(node_name) {
            return Ok("STRING Vanderwaltozyma polyspora");
        }

        if is_valid_string_mucor_circinelloides_node_name(node_name) {
            return Ok("STRING Mucor circinelloides");
        }

        if is_valid_string_pythium_irregulare_node_name(node_name) {
            return Ok("STRING Pythium irregulare");
        }

        if is_valid_string_neosartorya_fischeri_node_name(node_name) {
            return Ok("STRING Neosartorya fischeri");
        }

        if is_valid_string_penicillium_digitatum_node_name(node_name) {
            return Ok("STRING Penicillium digitatum");
        }

        if is_valid_string_mycobacterium_abscessus_node_name(node_name) {
            return Ok("STRING Mycobacterium abscessus");
        }

        if is_valid_string_wigglesworthia_glossinidia_sp_gbr_node_name(node_name) {
            return Ok("STRING Wigglesworthia glossinidia sp. Gbr");
        }

        if is_valid_string_porphyromonas_cangingivalis_node_name(node_name) {
            return Ok("STRING Porphyromonas cangingivalis");
        }

        if is_valid_string_porphyromonas_canoris_node_name(node_name) {
            return Ok("STRING Porphyromonas canoris");
        }

        if is_valid_string_lodderomyces_elongisporus_node_name(node_name) {
            return Ok("STRING Lodderomyces elongisporus");
        }

        if is_valid_string_clostridium_algidicarnis_node_name(node_name) {
            return Ok("STRING Clostridium algidicarnis");
        }

        if is_valid_string_aegilops_tauschii_node_name(node_name) {
            return Ok("STRING Aegilops tauschii");
        }

        if is_valid_string_phytoplasma_mali_node_name(node_name) {
            return Ok("STRING Phytoplasma mali");
        }

        if is_valid_string_talaromyces_marneffei_node_name(node_name) {
            return Ok("STRING Talaromyces marneffei");
        }

        if is_valid_string_rhodococcus_opacus_node_name(node_name) {
            return Ok("STRING Rhodococcus opacus");
        }

        if is_valid_string_chaetomium_globosum_node_name(node_name) {
            return Ok("STRING Chaetomium globosum");
        }

        if is_valid_string_alligator_sinensis_node_name(node_name) {
            return Ok("STRING Alligator sinensis");
        }

        if is_valid_string_panicum_virgatum_node_name(node_name) {
            return Ok("STRING Panicum virgatum");
        }

        if is_valid_string_micromonas_pusilla_node_name(node_name) {
            return Ok("STRING Micromonas pusilla");
        }

        if is_valid_string_tuber_melanosporum_node_name(node_name) {
            return Ok("STRING Tuber melanosporum");
        }

        if is_valid_string_streptococcus_equi_node_name(node_name) {
            return Ok("STRING Streptococcus equi");
        }

        if is_valid_string_neurospora_tetrasperma_node_name(node_name) {
            return Ok("STRING Neurospora tetrasperma");
        }

        if is_valid_string_oryza_glumipatula_node_name(node_name) {
            return Ok("STRING Oryza glumipatula");
        }

        if is_valid_string_oryza_meridionalis_node_name(node_name) {
            return Ok("STRING Oryza meridionalis");
        }

        if is_valid_string_acinetobacter_junii_node_name(node_name) {
            return Ok("STRING Acinetobacter junii");
        }

        if is_valid_string_acinetobacter_sp_cipa165_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. CIPA165");
        }

        if is_valid_string_aspergillus_kawachii_node_name(node_name) {
            return Ok("STRING Aspergillus kawachii");
        }

        if is_valid_string_fomitopsis_pinicola_node_name(node_name) {
            return Ok("STRING Fomitopsis pinicola");
        }

        if is_valid_string_stereum_hirsutum_node_name(node_name) {
            return Ok("STRING Stereum hirsutum");
        }

        if is_valid_string_botrytis_cinerea_node_name(node_name) {
            return Ok("STRING Botrytis cinerea");
        }

        if is_valid_string_lentzea_albidocapillata_node_name(node_name) {
            return Ok("STRING Lentzea albidocapillata");
        }

        if is_valid_string_cyanothece_sp_pcc8801_node_name(node_name) {
            return Ok("STRING Cyanothece sp. PCC8801");
        }

        if is_valid_string_bathycoccus_prasinos_node_name(node_name) {
            return Ok("STRING Bathycoccus prasinos");
        }

        if is_valid_string_pneumocystis_jirovecii_node_name(node_name) {
            return Ok("STRING Pneumocystis jirovecii");
        }

        if is_valid_string_pythium_vexans_node_name(node_name) {
            return Ok("STRING Pythium vexans");
        }

        if is_valid_string_sorex_araneus_node_name(node_name) {
            return Ok("STRING Sorex araneus");
        }

        if is_valid_string_rubrobacter_radiotolerans_node_name(node_name) {
            return Ok("STRING Rubrobacter radiotolerans");
        }

        if is_valid_string_phoenix_dactylifera_node_name(node_name) {
            return Ok("STRING Phoenix dactylifera");
        }

        if is_valid_string_candida_dubliniensis_node_name(node_name) {
            return Ok("STRING Candida dubliniensis");
        }

        if is_valid_string_halomonas_salina_node_name(node_name) {
            return Ok("STRING Halomonas salina");
        }

        if is_valid_string_anopheles_darlingi_node_name(node_name) {
            return Ok("STRING Anopheles darlingi");
        }

        if is_valid_string_ictidomys_tridecemlineatus_node_name(node_name) {
            return Ok("STRING Ictidomys tridecemlineatus");
        }

        if is_valid_string_capronia_epimyces_node_name(node_name) {
            return Ok("STRING Capronia epimyces");
        }

        if is_valid_string_capronia_coronata_node_name(node_name) {
            return Ok("STRING Capronia coronata");
        }

        if is_valid_string_prauserella_rugosa_node_name(node_name) {
            return Ok("STRING Prauserella rugosa");
        }

        if is_valid_string_streptomyces_wedmorensis_node_name(node_name) {
            return Ok("STRING Streptomyces wedmorensis");
        }

        if is_valid_string_cyanothece_sp_atcc51142_node_name(node_name) {
            return Ok("STRING Cyanothece sp. ATCC51142");
        }

        if is_valid_string_aureococcus_anophagefferens_node_name(node_name) {
            return Ok("STRING Aureococcus anophagefferens");
        }

        if is_valid_string_streptomyces_megasporus_node_name(node_name) {
            return Ok("STRING Streptomyces megasporus");
        }

        if is_valid_string_paenibacillus_durus_node_name(node_name) {
            return Ok("STRING Paenibacillus durus");
        }

        if is_valid_string_mycobacterium_avium_avium_node_name(node_name) {
            return Ok("STRING Mycobacterium avium avium");
        }

        if is_valid_string_dictyostelium_discoideum_node_name(node_name) {
            return Ok("STRING Dictyostelium discoideum");
        }

        if is_valid_string_bipolaris_sorokiniana_node_name(node_name) {
            return Ok("STRING Bipolaris sorokiniana");
        }

        if is_valid_string_pyrenophora_triticirepentis_node_name(node_name) {
            return Ok("STRING Pyrenophora triticirepentis");
        }

        if is_valid_string_cyanidioschyzon_merolae_node_name(node_name) {
            return Ok("STRING Cyanidioschyzon merolae");
        }

        if is_valid_string_eremothecium_cymbalariae_node_name(node_name) {
            return Ok("STRING Eremothecium cymbalariae");
        }

        if is_valid_string_nematostella_vectensis_node_name(node_name) {
            return Ok("STRING Nematostella vectensis");
        }

        if is_valid_string_mycoplasma_conjunctivae_node_name(node_name) {
            return Ok("STRING Mycoplasma conjunctivae");
        }

        if is_valid_string_candida_tenuis_node_name(node_name) {
            return Ok("STRING Candida tenuis");
        }

        if is_valid_string_anabaena_sp_90_node_name(node_name) {
            return Ok("STRING Anabaena sp. 90");
        }

        if is_valid_string_weissella_hellenica_node_name(node_name) {
            return Ok("STRING Weissella hellenica");
        }

        if is_valid_string_sphingobium_chlorophenolicum_node_name(node_name) {
            return Ok("STRING Sphingobium chlorophenolicum");
        }

        if is_valid_string_entamoeba_dispar_node_name(node_name) {
            return Ok("STRING Entamoeba dispar");
        }

        if is_valid_string_streptomyces_olivaceus_node_name(node_name) {
            return Ok("STRING Streptomyces olivaceus");
        }

        if is_valid_string_streptomyces_lydicus_node_name(node_name) {
            return Ok("STRING Streptomyces lydicus");
        }

        if is_valid_string_mycobacterium_triplex_node_name(node_name) {
            return Ok("STRING Mycobacterium triplex");
        }

        if is_valid_string_poecilia_formosa_node_name(node_name) {
            return Ok("STRING Poecilia formosa");
        }

        if is_valid_string_arabis_alpina_node_name(node_name) {
            return Ok("STRING Arabis alpina");
        }

        if is_valid_string_helicobacter_trogontum_node_name(node_name) {
            return Ok("STRING Helicobacter trogontum");
        }

        if is_valid_string_jaculus_jaculus_node_name(node_name) {
            return Ok("STRING Jaculus jaculus");
        }

        if is_valid_string_trichoderma_reesei_node_name(node_name) {
            return Ok("STRING Trichoderma reesei");
        }

        if is_valid_string_ciona_savignyi_node_name(node_name) {
            return Ok("STRING Ciona savignyi");
        }

        if is_valid_string_sulfitobacter_sp_ee36_node_name(node_name) {
            return Ok("STRING Sulfitobacter sp. EE36");
        }

        if is_valid_string_haliaeetus_leucocephalus_node_name(node_name) {
            return Ok("STRING Haliaeetus leucocephalus");
        }

        if is_valid_string_pyrenophora_teres_node_name(node_name) {
            return Ok("STRING Pyrenophora teres");
        }

        if is_valid_string_pectobacterium_betavasculorum_node_name(node_name) {
            return Ok("STRING Pectobacterium betavasculorum");
        }

        if is_valid_string_guillardia_theta_node_name(node_name) {
            return Ok("STRING Guillardia theta");
        }

        if is_valid_string_vibrio_anguillarum_node_name(node_name) {
            return Ok("STRING Vibrio anguillarum");
        }

        if is_valid_string_streptomyces_toyocaensis_node_name(node_name) {
            return Ok("STRING Streptomyces toyocaensis");
        }

        if is_valid_string_cylindrospermum_stagnale_node_name(node_name) {
            return Ok("STRING Cylindrospermum stagnale");
        }

        if is_valid_string_oscillatoria_acuminata_node_name(node_name) {
            return Ok("STRING Oscillatoria acuminata");
        }

        if is_valid_string_syntrophus_aciditrophicus_node_name(node_name) {
            return Ok("STRING Syntrophus aciditrophicus");
        }

        if is_valid_string_fragaria_vesca_node_name(node_name) {
            return Ok("STRING Fragaria vesca");
        }

        if is_valid_string_spirillospora_albida_node_name(node_name) {
            return Ok("STRING Spirillospora albida");
        }

        if is_valid_string_streptomyces_celluloflavus_node_name(node_name) {
            return Ok("STRING Streptomyces celluloflavus");
        }

        if is_valid_string_rickettsiella_grylli_node_name(node_name) {
            return Ok("STRING Rickettsiella grylli");
        }

        if is_valid_string_fibrobacter_succinogenes_node_name(node_name) {
            return Ok("STRING Fibrobacter succinogenes");
        }

        if is_valid_string_myotis_lucifugus_node_name(node_name) {
            return Ok("STRING Myotis lucifugus");
        }

        if is_valid_string_pantholops_hodgsonii_node_name(node_name) {
            return Ok("STRING Pantholops hodgsonii");
        }

        if is_valid_string_arabidopsis_lyrata_node_name(node_name) {
            return Ok("STRING Arabidopsis lyrata");
        }

        if is_valid_string_phytoplasma_australiense_node_name(node_name) {
            return Ok("STRING Phytoplasma australiense");
        }

        if is_valid_string_ficedula_albicollis_node_name(node_name) {
            return Ok("STRING Ficedula albicollis");
        }

        if is_valid_string_prochlorococcus_marinus_pastoris_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus pastoris");
        }

        if is_valid_string_prochlorococcus_marinus_mit9107_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus MIT9107");
        }

        if is_valid_string_prochlorococcus_marinus_gp2_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus GP2");
        }

        if is_valid_string_prochlorococcus_marinus_sb_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus SB");
        }

        if is_valid_string_synechococcus_sp_wh_7805_node_name(node_name) {
            return Ok("STRING Synechococcus sp. WH 7805");
        }

        if is_valid_string_lactobacillus_paraplantarum_node_name(node_name) {
            return Ok("STRING Lactobacillus paraplantarum");
        }

        if is_valid_string_chlorocebus_sabaeus_node_name(node_name) {
            return Ok("STRING Chlorocebus sabaeus");
        }

        if is_valid_string_ophiostoma_piceae_node_name(node_name) {
            return Ok("STRING Ophiostoma piceae");
        }

        if is_valid_string_coniosporium_apollinis_node_name(node_name) {
            return Ok("STRING Coniosporium apollinis");
        }

        if is_valid_string_rhinopithecus_roxellana_node_name(node_name) {
            return Ok("STRING Rhinopithecus roxellana");
        }

        if is_valid_string_pluralibacter_gergoviae_node_name(node_name) {
            return Ok("STRING Pluralibacter gergoviae");
        }

        if is_valid_string_nomascus_leucogenys_node_name(node_name) {
            return Ok("STRING Nomascus leucogenys");
        }

        if is_valid_string_azoarcus_sp_bh72_node_name(node_name) {
            return Ok("STRING Azoarcus sp. BH72");
        }

        if is_valid_string_acinetobacter_sp_adp1_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. ADP1");
        }

        if is_valid_string_arthroderma_benhamiae_node_name(node_name) {
            return Ok("STRING Arthroderma benhamiae");
        }

        if is_valid_string_microsporum_gypseum_node_name(node_name) {
            return Ok("STRING Microsporum gypseum");
        }

        if is_valid_string_arthroderma_otae_node_name(node_name) {
            return Ok("STRING Arthroderma otae");
        }

        if is_valid_string_trichophyton_equinum_node_name(node_name) {
            return Ok("STRING Trichophyton equinum");
        }

        if is_valid_string_trichoderma_atroviride_node_name(node_name) {
            return Ok("STRING Trichoderma atroviride");
        }

        if is_valid_string_nostoc_punctiforme_node_name(node_name) {
            return Ok("STRING Nostoc punctiforme");
        }

        if is_valid_string_mycosphaerella_pini_node_name(node_name) {
            return Ok("STRING Mycosphaerella pini");
        }

        if is_valid_string_synechococcus_sp_cc9311_node_name(node_name) {
            return Ok("STRING Synechococcus sp. CC9311");
        }

        if is_valid_string_pythium_ultimum_node_name(node_name) {
            return Ok("STRING Pythium ultimum");
        }

        if is_valid_string_halothece_sp_pcc7418_node_name(node_name) {
            return Ok("STRING Halothece sp. PCC7418");
        }

        if is_valid_string_cyanothece_sp_pcc7424_node_name(node_name) {
            return Ok("STRING Cyanothece sp. PCC7424");
        }

        if is_valid_string_oryza_barthii_node_name(node_name) {
            return Ok("STRING Oryza barthii");
        }

        if is_valid_string_actinoalloteichus_cyanogriseus_node_name(node_name) {
            return Ok("STRING Actinoalloteichus cyanogriseus");
        }

        if is_valid_string_piriformospora_indica_node_name(node_name) {
            return Ok("STRING Piriformospora indica");
        }

        if is_valid_string_pantoea_stewartii_node_name(node_name) {
            return Ok("STRING Pantoea stewartii");
        }

        if is_valid_string_streptomyces_niger_node_name(node_name) {
            return Ok("STRING Streptomyces niger");
        }

        if is_valid_string_streptomyces_violens_node_name(node_name) {
            return Ok("STRING Streptomyces violens");
        }

        if is_valid_string_streptomyces_roseoverticillatus_node_name(node_name) {
            return Ok("STRING Streptomyces roseoverticillatus");
        }

        if is_valid_string_bacillus_clausii_node_name(node_name) {
            return Ok("STRING Bacillus clausii");
        }

        if is_valid_string_streptomyces_atroolivaceus_node_name(node_name) {
            return Ok("STRING Streptomyces atroolivaceus");
        }

        if is_valid_string_streptomyces_bicolor_node_name(node_name) {
            return Ok("STRING Streptomyces bicolor");
        }

        if is_valid_string_streptomyces_catenulae_node_name(node_name) {
            return Ok("STRING Streptomyces catenulae");
        }

        if is_valid_string_streptomyces_griseorubens_node_name(node_name) {
            return Ok("STRING Streptomyces griseorubens");
        }

        if is_valid_string_streptomyces_albus_albus_node_name(node_name) {
            return Ok("STRING Streptomyces albus albus");
        }

        if is_valid_string_streptomyces_alboflavus_node_name(node_name) {
            return Ok("STRING Streptomyces alboflavus");
        }

        if is_valid_string_streptomyces_aureocirculatus_node_name(node_name) {
            return Ok("STRING Streptomyces aureocirculatus");
        }

        if is_valid_string_streptomyces_californicus_node_name(node_name) {
            return Ok("STRING Streptomyces californicus");
        }

        if is_valid_string_streptomyces_lavenduligriseus_node_name(node_name) {
            return Ok("STRING Streptomyces lavenduligriseus");
        }

        if is_valid_string_streptomyces_mutabilis_node_name(node_name) {
            return Ok("STRING Streptomyces mutabilis");
        }

        if is_valid_string_streptomyces_purpeofuscus_node_name(node_name) {
            return Ok("STRING Streptomyces purpeofuscus");
        }

        if is_valid_string_streptomyces_resistomycificus_node_name(node_name) {
            return Ok("STRING Streptomyces resistomycificus");
        }

        if is_valid_string_streptomyces_varsoviensis_node_name(node_name) {
            return Ok("STRING Streptomyces varsoviensis");
        }

        if is_valid_string_phytophthora_sojae_node_name(node_name) {
            return Ok("STRING Phytophthora sojae");
        }

        if is_valid_string_lechevalieria_aerocolonigenes_node_name(node_name) {
            return Ok("STRING Lechevalieria aerocolonigenes");
        }

        if is_valid_string_streptomyces_durhamensis_node_name(node_name) {
            return Ok("STRING Streptomyces durhamensis");
        }

        if is_valid_string_streptomyces_flavochromogenes_node_name(node_name) {
            return Ok("STRING Streptomyces flavochromogenes");
        }

        if is_valid_string_streptomyces_iakyrus_node_name(node_name) {
            return Ok("STRING Streptomyces iakyrus");
        }

        if is_valid_string_streptomyces_katrae_node_name(node_name) {
            return Ok("STRING Streptomyces katrae");
        }

        if is_valid_string_streptomyces_pyridomyceticus_node_name(node_name) {
            return Ok("STRING Streptomyces pyridomyceticus");
        }

        if is_valid_string_streptomyces_albulus_node_name(node_name) {
            return Ok("STRING Streptomyces albulus");
        }

        if is_valid_string_theileria_orientalis_node_name(node_name) {
            return Ok("STRING Theileria orientalis");
        }

        if is_valid_string_thermococcus_kodakarensis_node_name(node_name) {
            return Ok("STRING Thermococcus kodakarensis");
        }

        if is_valid_string_synechococcus_sp_wh_5701_node_name(node_name) {
            return Ok("STRING Synechococcus sp. WH 5701");
        }

        if is_valid_string_aquamicrobium_defluvii_node_name(node_name) {
            return Ok("STRING Aquamicrobium defluvii");
        }

        if is_valid_string_gasterosteus_aculeatus_node_name(node_name) {
            return Ok("STRING Gasterosteus aculeatus");
        }

        if is_valid_string_microplitis_demolitor_node_name(node_name) {
            return Ok("STRING Microplitis demolitor");
        }

        if is_valid_string_pseudomonas_sp_vlb120_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. VLB120");
        }

        if is_valid_string_caulobacter_henricii_node_name(node_name) {
            return Ok("STRING Caulobacter henricii");
        }

        if is_valid_string_penicillium_oxalicum_node_name(node_name) {
            return Ok("STRING Penicillium oxalicum");
        }

        if is_valid_string_ostreococcus_tauri_node_name(node_name) {
            return Ok("STRING Ostreococcus tauri");
        }

        if is_valid_string_pyrococcus_horikoshii_node_name(node_name) {
            return Ok("STRING Pyrococcus horikoshii");
        }

        if is_valid_string_eucalyptus_grandis_node_name(node_name) {
            return Ok("STRING Eucalyptus grandis");
        }

        if is_valid_string_haemophilus_influenzae_node_name(node_name) {
            return Ok("STRING Haemophilus influenzae");
        }

        if is_valid_string_bos_mutus_node_name(node_name) {
            return Ok("STRING Bos mutus");
        }

        if is_valid_string_sphaeroforma_arctica_node_name(node_name) {
            return Ok("STRING Sphaeroforma arctica");
        }

        if is_valid_string_ophiocordyceps_sinensis_node_name(node_name) {
            return Ok("STRING Ophiocordyceps sinensis");
        }

        if is_valid_string_sporisorium_reilianum_node_name(node_name) {
            return Ok("STRING Sporisorium reilianum");
        }

        if is_valid_string_boechera_stricta_node_name(node_name) {
            return Ok("STRING Boechera stricta");
        }

        if is_valid_string_eutrema_salsugineum_node_name(node_name) {
            return Ok("STRING Eutrema salsugineum");
        }

        if is_valid_string_streptomyces_seoulensis_node_name(node_name) {
            return Ok("STRING Streptomyces seoulensis");
        }

        if is_valid_string_cordyceps_militaris_node_name(node_name) {
            return Ok("STRING Cordyceps militaris");
        }

        if is_valid_string_prochlorococcus_marinus_mit9302_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus MIT9302");
        }

        if is_valid_string_prochlorococcus_marinus_mit9312_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus MIT9312");
        }

        if is_valid_string_prochlorococcus_marinus_mit9313_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus MIT9313");
        }

        if is_valid_string_thiomonas_intermedia_node_name(node_name) {
            return Ok("STRING Thiomonas intermedia");
        }

        if is_valid_string_aromatoleum_aromaticum_node_name(node_name) {
            return Ok("STRING Aromatoleum aromaticum");
        }

        if is_valid_string_mycetocola_saprophilus_node_name(node_name) {
            return Ok("STRING Mycetocola saprophilus");
        }

        if is_valid_string_pseudomonas_putida_gb1_node_name(node_name) {
            return Ok("STRING Pseudomonas putida GB1");
        }

        if is_valid_string_leersia_perrieri_node_name(node_name) {
            return Ok("STRING Leersia perrieri");
        }

        if is_valid_string_bifidobacterium_subtile_node_name(node_name) {
            return Ok("STRING Bifidobacterium subtile");
        }

        if is_valid_string_xanthobacter_autotrophicus_node_name(node_name) {
            return Ok("STRING Xanthobacter autotrophicus");
        }

        if is_valid_string_bifidobacterium_gallinarum_node_name(node_name) {
            return Ok("STRING Bifidobacterium gallinarum");
        }

        if is_valid_string_bifidobacterium_merycicum_node_name(node_name) {
            return Ok("STRING Bifidobacterium merycicum");
        }

        if is_valid_string_bifidobacterium_ruminantium_node_name(node_name) {
            return Ok("STRING Bifidobacterium ruminantium");
        }

        if is_valid_string_pectobacterium_carotovorum_odoriferum_node_name(node_name) {
            return Ok("STRING Pectobacterium carotovorum odoriferum");
        }

        if is_valid_string_myceliophthora_thermophila_node_name(node_name) {
            return Ok("STRING Myceliophthora thermophila");
        }

        if is_valid_string_mortierella_verticillata_node_name(node_name) {
            return Ok("STRING Mortierella verticillata");
        }

        if is_valid_string_microtus_ochrogaster_node_name(node_name) {
            return Ok("STRING Microtus ochrogaster");
        }

        if is_valid_string_coniophora_puteana_node_name(node_name) {
            return Ok("STRING Coniophora puteana");
        }

        if is_valid_string_colletotrichum_higginsianum_node_name(node_name) {
            return Ok("STRING Colletotrichum higginsianum");
        }

        if is_valid_string_monosiga_brevicollis_node_name(node_name) {
            return Ok("STRING Monosiga brevicollis");
        }

        if is_valid_string_capsella_rubella_node_name(node_name) {
            return Ok("STRING Capsella rubella");
        }

        if is_valid_string_trichosporon_asahii_node_name(node_name) {
            return Ok("STRING Trichosporon asahii");
        }

        if is_valid_string_pseudanabaena_sp_pcc7367_node_name(node_name) {
            return Ok("STRING Pseudanabaena sp. PCC7367");
        }

        if is_valid_string_serratia_grimesii_node_name(node_name) {
            return Ok("STRING Serratia grimesii");
        }

        if is_valid_string_serratia_plymuthica_node_name(node_name) {
            return Ok("STRING Serratia plymuthica");
        }

        if is_valid_string_sulfitobacter_mediterraneus_node_name(node_name) {
            return Ok("STRING Sulfitobacter mediterraneus");
        }

        if is_valid_string_mycobacterium_tuberculosis_h37rv_node_name(node_name) {
            return Ok("STRING Mycobacterium tuberculosis H37Rv");
        }

        if is_valid_string_pseudocercospora_fijiensis_node_name(node_name) {
            return Ok("STRING Pseudocercospora fijiensis");
        }

        if is_valid_string_gammaproteobacteria_bacterium_hdn1_node_name(node_name) {
            return Ok("STRING Gammaproteobacteria bacterium HdN1");
        }

        if is_valid_string_lysobacter_antibioticus_node_name(node_name) {
            return Ok("STRING Lysobacter antibioticus");
        }

        if is_valid_string_synechococcus_sp_wh_8102_node_name(node_name) {
            return Ok("STRING Synechococcus sp. WH 8102");
        }

        if is_valid_string_pseudozyma_flocculosa_node_name(node_name) {
            return Ok("STRING Pseudozyma flocculosa");
        }

        if is_valid_string_thauera_sp_mz1t_node_name(node_name) {
            return Ok("STRING Thauera sp. MZ1T");
        }

        if is_valid_string_citrus_clementina_node_name(node_name) {
            return Ok("STRING Citrus clementina");
        }

        if is_valid_string_sphaerulina_musiva_node_name(node_name) {
            return Ok("STRING Sphaerulina musiva");
        }

        if is_valid_string_helicobacter_pylori_26695_node_name(node_name) {
            return Ok("STRING Helicobacter pylori 26695");
        }

        if is_valid_string_helicobacter_pylori_j99_node_name(node_name) {
            return Ok("STRING Helicobacter pylori J99");
        }

        if is_valid_string_serpula_lacrymans_node_name(node_name) {
            return Ok("STRING Serpula lacrymans");
        }

        if is_valid_string_cladophialophora_carrionii_node_name(node_name) {
            return Ok("STRING Cladophialophora carrionii");
        }

        if is_valid_string_endosymbiont_of_acanthamoeba_node_name(node_name) {
            return Ok("STRING endosymbiont of Acanthamoeba");
        }

        if is_valid_string_clostridium_pasteurianum_bc1_node_name(node_name) {
            return Ok("STRING Clostridium pasteurianum BC1");
        }

        if is_valid_string_pseudoalteromonas_tunicata_node_name(node_name) {
            return Ok("STRING Pseudoalteromonas tunicata");
        }

        if is_valid_string_selaginella_moellendorffii_node_name(node_name) {
            return Ok("STRING Selaginella moellendorffii");
        }

        if is_valid_string_roseovarius_nubinhibens_node_name(node_name) {
            return Ok("STRING Roseovarius nubinhibens");
        }

        if is_valid_string_bubalus_bubalis_node_name(node_name) {
            return Ok("STRING Bubalus bubalis");
        }

        if is_valid_string_salmonella_enterica_typhimurium_node_name(node_name) {
            return Ok("STRING Salmonella enterica Typhimurium");
        }

        if is_valid_string_camelina_sativa_node_name(node_name) {
            return Ok("STRING Camelina sativa");
        }

        if is_valid_string_thiomicrospira_sp_milost1_node_name(node_name) {
            return Ok("STRING Thiomicrospira sp. MilosT1");
        }

        if is_valid_string_thiomicrospira_sp_milost2_node_name(node_name) {
            return Ok("STRING Thiomicrospira sp. MilosT2");
        }

        if is_valid_string_synechococcus_sp_pcc7335_node_name(node_name) {
            return Ok("STRING Synechococcus sp. PCC7335");
        }

        if is_valid_string_paracaedibacter_acanthamoebae_node_name(node_name) {
            return Ok("STRING Paracaedibacter acanthamoebae");
        }

        if is_valid_string_metarhizium_acridum_node_name(node_name) {
            return Ok("STRING Metarhizium acridum");
        }

        if is_valid_string_prochlorococcus_marinus_mit9201_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus MIT9201");
        }

        if is_valid_string_prochlorococcus_marinus_mit9211_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus MIT9211");
        }

        if is_valid_string_prochlorococcus_marinus_mit9215_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus MIT9215");
        }

        if is_valid_string_pandoraea_pnomenusa_node_name(node_name) {
            return Ok("STRING Pandoraea pnomenusa");
        }

        if is_valid_string_setosphaeria_turcica_node_name(node_name) {
            return Ok("STRING Setosphaeria turcica");
        }

        if is_valid_string_shewanella_sp_ana3_node_name(node_name) {
            return Ok("STRING Shewanella sp. ANA3");
        }

        if is_valid_string_bordetella_petrii_node_name(node_name) {
            return Ok("STRING Bordetella petrii");
        }

        if is_valid_string_pseudomonas_sp_m1_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. M1");
        }

        if is_valid_string_desulfococcus_oleovorans_node_name(node_name) {
            return Ok("STRING Desulfococcus oleovorans");
        }

        if is_valid_string_eutypa_lata_node_name(node_name) {
            return Ok("STRING Eutypa lata");
        }

        if is_valid_string_lactobacillus_sp_asf360_node_name(node_name) {
            return Ok("STRING Lactobacillus sp. ASF360");
        }

        if is_valid_string_clostridium_sp_asf356_node_name(node_name) {
            return Ok("STRING Clostridium sp. ASF356");
        }

        if is_valid_string_clostridium_sp_asf502_node_name(node_name) {
            return Ok("STRING Clostridium sp. ASF502");
        }

        if is_valid_string_fischerella_thermalis_node_name(node_name) {
            return Ok("STRING Fischerella thermalis");
        }

        if is_valid_string_hammondia_hammondi_node_name(node_name) {
            return Ok("STRING Hammondia hammondi");
        }

        if is_valid_string_calothrix_sp_pcc7507_node_name(node_name) {
            return Ok("STRING Calothrix sp. PCC7507");
        }

        if is_valid_string_streptomyces_coelicolor_node_name(node_name) {
            return Ok("STRING Streptomyces coelicolor");
        }

        if is_valid_string_wolbachia_sp_ooc_node_name(node_name) {
            return Ok("STRING Wolbachia sp. Ooc");
        }

        if is_valid_string_fusarium_pseudograminearum_node_name(node_name) {
            return Ok("STRING Fusarium pseudograminearum");
        }

        if is_valid_string_bipolaris_oryzae_node_name(node_name) {
            return Ok("STRING Bipolaris oryzae");
        }

        if is_valid_string_rhodococcus_jostii_node_name(node_name) {
            return Ok("STRING Rhodococcus jostii");
        }

        if is_valid_string_glarea_lozoyensis_node_name(node_name) {
            return Ok("STRING Glarea lozoyensis");
        }

        if is_valid_string_prunus_mume_node_name(node_name) {
            return Ok("STRING Prunus mume");
        }

        if is_valid_string_xenococcus_sp_pcc7305_node_name(node_name) {
            return Ok("STRING Xenococcus sp. PCC7305");
        }

        if is_valid_string_leptolyngbya_sp_pcc7375_node_name(node_name) {
            return Ok("STRING Leptolyngbya sp. PCC7375");
        }

        if is_valid_string_gloeocapsa_sp_pcc73106_node_name(node_name) {
            return Ok("STRING Gloeocapsa sp. PCC73106");
        }

        if is_valid_string_acromyrmex_echinatior_node_name(node_name) {
            return Ok("STRING Acromyrmex echinatior");
        }

        if is_valid_string_nostoc_sp_pcc7120_node_name(node_name) {
            return Ok("STRING Nostoc sp. PCC7120");
        }

        if is_valid_string_saccharothrix_syringae_node_name(node_name) {
            return Ok("STRING Saccharothrix syringae");
        }

        if is_valid_string_gloeophyllum_trabeum_node_name(node_name) {
            return Ok("STRING Gloeophyllum trabeum");
        }

        if is_valid_string_serratia_sp_atcc39006_node_name(node_name) {
            return Ok("STRING Serratia sp. ATCC39006");
        }

        if is_valid_string_streptacidiphilus_neutrinimicus_node_name(node_name) {
            return Ok("STRING Streptacidiphilus neutrinimicus");
        }

        if is_valid_string_streptacidiphilus_carbonis_node_name(node_name) {
            return Ok("STRING Streptacidiphilus carbonis");
        }

        if is_valid_string_streptacidiphilus_albus_node_name(node_name) {
            return Ok("STRING Streptacidiphilus albus");
        }

        if is_valid_string_nitrosococcus_watsonii_node_name(node_name) {
            return Ok("STRING Nitrosococcus watsonii");
        }

        if is_valid_string_frankia_sp_cci3_node_name(node_name) {
            return Ok("STRING Frankia sp. CcI3");
        }

        if is_valid_string_maylandia_zebra_node_name(node_name) {
            return Ok("STRING Maylandia zebra");
        }

        if is_valid_string_acinetobacter_bereziniae_node_name(node_name) {
            return Ok("STRING Acinetobacter bereziniae");
        }

        if is_valid_string_methylosinus_sp_lw3_node_name(node_name) {
            return Ok("STRING Methylosinus sp. LW3");
        }

        if is_valid_string_methylosinus_sp_pw1_node_name(node_name) {
            return Ok("STRING Methylosinus sp. PW1");
        }

        if is_valid_string_buchnera_aphidicola_aps_node_name(node_name) {
            return Ok("STRING Buchnera aphidicola APS");
        }

        if is_valid_string_myotis_brandtii_node_name(node_name) {
            return Ok("STRING Myotis brandtii");
        }

        if is_valid_string_spizellomyces_punctatus_node_name(node_name) {
            return Ok("STRING Spizellomyces punctatus");
        }

        if is_valid_string_batrachochytrium_dendrobatidis_node_name(node_name) {
            return Ok("STRING Batrachochytrium dendrobatidis");
        }

        if is_valid_string_nocardioides_sp_cf8_node_name(node_name) {
            return Ok("STRING Nocardioides sp. CF8");
        }

        if is_valid_string_gregarina_niphandrodes_node_name(node_name) {
            return Ok("STRING Gregarina niphandrodes");
        }

        if is_valid_string_synechococcus_sp_cc9605_node_name(node_name) {
            return Ok("STRING Synechococcus sp. CC9605");
        }

        if is_valid_string_synechococcus_sp_cc9616_node_name(node_name) {
            return Ok("STRING Synechococcus sp. CC9616");
        }

        if is_valid_string_porphyromonas_gulae_node_name(node_name) {
            return Ok("STRING Porphyromonas gulae");
        }

        if is_valid_string_stanieria_cyanosphaera_node_name(node_name) {
            return Ok("STRING Stanieria cyanosphaera");
        }

        if is_valid_string_leptolyngbya_sp_pcc7376_node_name(node_name) {
            return Ok("STRING Leptolyngbya sp. PCC7376");
        }

        if is_valid_string_saprolegnia_diclina_node_name(node_name) {
            return Ok("STRING Saprolegnia diclina");
        }

        if is_valid_string_geminocystis_herdmanii_node_name(node_name) {
            return Ok("STRING Geminocystis herdmanii");
        }

        if is_valid_string_bradyrhizobium_sp_tv2a2_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. Tv2a2");
        }

        if is_valid_string_tetrapisispora_phaffii_node_name(node_name) {
            return Ok("STRING Tetrapisispora phaffii");
        }

        if is_valid_string_bradyrhizobium_sp_ors278_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. ORS278");
        }

        if is_valid_string_pythium_iwayamai_node_name(node_name) {
            return Ok("STRING Pythium iwayamai");
        }

        if is_valid_string_chlamydophila_pneumoniae_ar39_node_name(node_name) {
            return Ok("STRING Chlamydophila pneumoniae AR39");
        }

        if is_valid_string_chlamydophila_pneumoniae_cwl029_node_name(node_name) {
            return Ok("STRING Chlamydophila pneumoniae CWL029");
        }

        if is_valid_string_fusarium_verticillioides_node_name(node_name) {
            return Ok("STRING Fusarium verticillioides");
        }

        if is_valid_string_chrysiogenes_arsenatis_node_name(node_name) {
            return Ok("STRING Chrysiogenes arsenatis");
        }

        if is_valid_string_pleurocapsa_sp_pcc7319_node_name(node_name) {
            return Ok("STRING Pleurocapsa sp. PCC7319");
        }

        if is_valid_string_pleurocapsa_sp_pcc7327_node_name(node_name) {
            return Ok("STRING Pleurocapsa sp. PCC7327");
        }

        if is_valid_string_nodosilinea_nodulosa_node_name(node_name) {
            return Ok("STRING Nodosilinea nodulosa");
        }

        if is_valid_string_coleofasciculus_chthonoplastes_node_name(node_name) {
            return Ok("STRING Coleofasciculus chthonoplastes");
        }

        if is_valid_string_pseudanabaena_sp_pcc6802_node_name(node_name) {
            return Ok("STRING Pseudanabaena sp. PCC6802");
        }

        if is_valid_string_lipotes_vexillifer_node_name(node_name) {
            return Ok("STRING Lipotes vexillifer");
        }

        if is_valid_string_ustilago_hordei_node_name(node_name) {
            return Ok("STRING Ustilago hordei");
        }

        if is_valid_string_pediculus_humanus_node_name(node_name) {
            return Ok("STRING Pediculus humanus");
        }

        if is_valid_string_paracoccidioides_brasiliensis_node_name(node_name) {
            return Ok("STRING Paracoccidioides brasiliensis");
        }

        if is_valid_string_persephonella_marina_node_name(node_name) {
            return Ok("STRING Persephonella marina");
        }

        if is_valid_string_bordetella_trematum_node_name(node_name) {
            return Ok("STRING Bordetella trematum");
        }

        if is_valid_string_strigamia_maritima_node_name(node_name) {
            return Ok("STRING Strigamia maritima");
        }

        if is_valid_string_nipponia_nippon_node_name(node_name) {
            return Ok("STRING Nipponia nippon");
        }

        if is_valid_string_galdieria_sulphuraria_node_name(node_name) {
            return Ok("STRING Galdieria sulphuraria");
        }

        if is_valid_string_bombus_impatiens_node_name(node_name) {
            return Ok("STRING Bombus impatiens");
        }

        if is_valid_string_pteropus_vampyrus_node_name(node_name) {
            return Ok("STRING Pteropus vampyrus");
        }

        if is_valid_string_actinoplanes_sp_se50110_node_name(node_name) {
            return Ok("STRING Actinoplanes sp. SE50110");
        }

        if is_valid_string_caenorhabditis_brenneri_node_name(node_name) {
            return Ok("STRING Caenorhabditis brenneri");
        }

        if is_valid_string_zootermopsis_nevadensis_node_name(node_name) {
            return Ok("STRING Zootermopsis nevadensis");
        }

        if is_valid_string_kocuria_polaris_node_name(node_name) {
            return Ok("STRING Kocuria polaris");
        }

        if is_valid_string_methylosinus_sp_lw4_node_name(node_name) {
            return Ok("STRING Methylosinus sp. LW4");
        }

        if is_valid_string_desulfitobacterium_hafniense_node_name(node_name) {
            return Ok("STRING Desulfitobacterium hafniense");
        }

        if is_valid_string_nectria_haematococca_node_name(node_name) {
            return Ok("STRING Nectria haematococca");
        }

        if is_valid_string_lachnobacterium_bovis_node_name(node_name) {
            return Ok("STRING Lachnobacterium bovis");
        }

        if is_valid_string_zobellia_uliginosa_node_name(node_name) {
            return Ok("STRING Zobellia uliginosa");
        }

        if is_valid_string_stegastes_partitus_node_name(node_name) {
            return Ok("STRING Stegastes partitus");
        }

        if is_valid_string_prochlorococcus_marinus_as9601_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus AS9601");
        }

        if is_valid_string_streptomyces_griseofuscus_node_name(node_name) {
            return Ok("STRING Streptomyces griseofuscus");
        }

        if is_valid_string_magnaporthiopsis_poae_node_name(node_name) {
            return Ok("STRING Magnaporthiopsis poae");
        }

        if is_valid_string_lactobacillus_kunkeei_node_name(node_name) {
            return Ok("STRING Lactobacillus kunkeei");
        }

        if is_valid_string_wallemia_sebi_node_name(node_name) {
            return Ok("STRING Wallemia sebi");
        }

        if is_valid_string_kozakia_baliensis_node_name(node_name) {
            return Ok("STRING Kozakia baliensis");
        }

        if is_valid_string_sporocytophaga_myxococcoides_node_name(node_name) {
            return Ok("STRING Sporocytophaga myxococcoides");
        }

        if is_valid_string_nitrosomonas_sp_al212_node_name(node_name) {
            return Ok("STRING Nitrosomonas sp. AL212");
        }

        if is_valid_string_gallibacterium_genomosp_node_name(node_name) {
            return Ok("STRING Gallibacterium genomosp.");
        }

        if is_valid_string_escherichia_coli_o157h7_node_name(node_name) {
            return Ok("STRING Escherichia coli O157H7");
        }

        if is_valid_string_alteromonadales_bacterium_tw7_node_name(node_name) {
            return Ok("STRING Alteromonadales bacterium TW7");
        }

        if is_valid_string_magnetococcus_marinus_node_name(node_name) {
            return Ok("STRING Magnetococcus marinus");
        }

        if is_valid_string_corynebacterium_imitans_node_name(node_name) {
            return Ok("STRING Corynebacterium imitans");
        }

        if is_valid_string_aphanomyces_invadans_node_name(node_name) {
            return Ok("STRING Aphanomyces invadans");
        }

        if is_valid_string_pseudomonas_cremoricolorata_node_name(node_name) {
            return Ok("STRING Pseudomonas cremoricolorata");
        }

        if is_valid_string_sphaerochaeta_globosa_node_name(node_name) {
            return Ok("STRING Sphaerochaeta globosa");
        }

        if is_valid_string_sphaerochaeta_pleomorpha_node_name(node_name) {
            return Ok("STRING Sphaerochaeta pleomorpha");
        }

        if is_valid_string_novosphingobium_resinovorum_node_name(node_name) {
            return Ok("STRING Novosphingobium resinovorum");
        }

        if is_valid_string_bifidobacterium_scardovii_node_name(node_name) {
            return Ok("STRING Bifidobacterium scardovii");
        }

        if is_valid_string_cedecea_neteri_node_name(node_name) {
            return Ok("STRING Cedecea neteri");
        }

        if is_valid_string_dechloromonas_aromatica_node_name(node_name) {
            return Ok("STRING Dechloromonas aromatica");
        }

        if is_valid_string_burkholderia_sacchari_node_name(node_name) {
            return Ok("STRING Burkholderia sacchari");
        }

        if is_valid_string_thalassiosira_oceanica_node_name(node_name) {
            return Ok("STRING Thalassiosira oceanica");
        }

        if is_valid_string_pseudomonas_putida_kt2440_node_name(node_name) {
            return Ok("STRING Pseudomonas putida KT2440");
        }

        if is_valid_string_xylella_fastidiosa_9a5c_node_name(node_name) {
            return Ok("STRING Xylella fastidiosa 9a5c");
        }

        if is_valid_string_paenibacillus_borealis_node_name(node_name) {
            return Ok("STRING Paenibacillus borealis");
        }

        if is_valid_string_auricularia_delicata_node_name(node_name) {
            return Ok("STRING Auricularia delicata");
        }

        if is_valid_string_thermodesulfobacterium_hydrogeniphilum_node_name(node_name) {
            return Ok("STRING Thermodesulfobacterium hydrogeniphilum");
        }

        if is_valid_string_erythrobacter_sp_sd21_node_name(node_name) {
            return Ok("STRING Erythrobacter sp. SD21");
        }

        if is_valid_string_beta_vulgaris_node_name(node_name) {
            return Ok("STRING Beta vulgaris");
        }

        if is_valid_string_aspergillus_nidulans_node_name(node_name) {
            return Ok("STRING Aspergillus nidulans");
        }

        if is_valid_string_wolbachia_sp_dme_node_name(node_name) {
            return Ok("STRING Wolbachia sp. Dme");
        }

        if is_valid_string_anabaena_sp_pcc7108_node_name(node_name) {
            return Ok("STRING Anabaena sp. PCC7108");
        }

        if is_valid_string_phytophthora_ramorum_node_name(node_name) {
            return Ok("STRING Phytophthora ramorum");
        }

        if is_valid_string_mycobacterium_sp_jls_node_name(node_name) {
            return Ok("STRING Mycobacterium sp. JLS");
        }

        if is_valid_string_synechococcus_sp_wh_8109_node_name(node_name) {
            return Ok("STRING Synechococcus sp. WH 8109");
        }

        if is_valid_string_synechococcus_sp_wh_8016_node_name(node_name) {
            return Ok("STRING Synechococcus sp. WH 8016");
        }

        if is_valid_string_prochlorococcus_marinus_ccmp1375_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus CCMP1375");
        }

        if is_valid_string_prochlorococcus_marinus_mit9515_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus MIT9515");
        }

        if is_valid_string_prochlorococcus_marinus_mit9301_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus MIT9301");
        }

        if is_valid_string_prochlorococcus_marinus_mit9314_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus MIT9314");
        }

        if is_valid_string_prochlorococcus_marinus_mit9322_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus MIT9322");
        }

        if is_valid_string_prochlorococcus_marinus_natl1a_node_name(node_name) {
            return Ok("STRING Prochlorococcus marinus NATL1A");
        }

        if is_valid_string_listeria_monocytogenes_egde_node_name(node_name) {
            return Ok("STRING Listeria monocytogenes EGDe");
        }

        if is_valid_string_streptococcus_pneumoniae_tigr4_node_name(node_name) {
            return Ok("STRING Streptococcus pneumoniae TIGR4");
        }

        if is_valid_string_streptococcus_pneumoniae_r6_node_name(node_name) {
            return Ok("STRING Streptococcus pneumoniae R6");
        }

        if is_valid_string_oceanobacillus_picturae_node_name(node_name) {
            return Ok("STRING Oceanobacillus picturae");
        }

        if is_valid_string_elizabethkingia_miricola_node_name(node_name) {
            return Ok("STRING Elizabethkingia miricola");
        }

        if is_valid_string_bradyrhizobium_sp_thb2_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. thb2");
        }

        if is_valid_string_streptococcus_sinensis_node_name(node_name) {
            return Ok("STRING Streptococcus sinensis");
        }

        if is_valid_string_beauveria_bassiana_node_name(node_name) {
            return Ok("STRING Beauveria bassiana");
        }

        if is_valid_string_staphylococcus_epidermidis_rp62a_node_name(node_name) {
            return Ok("STRING Staphylococcus epidermidis RP62A");
        }

        if is_valid_string_staphylococcus_epidermidis_atcc12228_node_name(node_name) {
            return Ok("STRING Staphylococcus epidermidis ATCC12228");
        }

        if is_valid_string_agrobacterium_fabrum_node_name(node_name) {
            return Ok("STRING Agrobacterium fabrum");
        }

        if is_valid_string_python_bivittatus_node_name(node_name) {
            return Ok("STRING Python bivittatus");
        }

        if is_valid_string_francisella_tularensis_schus4_node_name(node_name) {
            return Ok("STRING Francisella tularensis SCHUS4");
        }

        if is_valid_string_desulfobacterium_autotrophicum_node_name(node_name) {
            return Ok("STRING Desulfobacterium autotrophicum");
        }

        if is_valid_string_desulfotalea_psychrophila_node_name(node_name) {
            return Ok("STRING Desulfotalea psychrophila");
        }

        if is_valid_string_pyrobaculum_aerophilum_node_name(node_name) {
            return Ok("STRING Pyrobaculum aerophilum");
        }

        if is_valid_string_acetobacter_malorum_node_name(node_name) {
            return Ok("STRING Acetobacter malorum");
        }

        if is_valid_string_oscillatoria_nigroviridis_node_name(node_name) {
            return Ok("STRING Oscillatoria nigroviridis");
        }

        if is_valid_string_cyanobium_sp_pcc7001_node_name(node_name) {
            return Ok("STRING Cyanobium sp. PCC7001");
        }

        if is_valid_string_robinsoniella_peoriensis_node_name(node_name) {
            return Ok("STRING Robinsoniella peoriensis");
        }

        if is_valid_string_pseudopodoces_humilis_node_name(node_name) {
            return Ok("STRING Pseudopodoces humilis");
        }

        if is_valid_string_chlamydophila_pneumoniae_tw183_node_name(node_name) {
            return Ok("STRING Chlamydophila pneumoniae TW183");
        }

        if is_valid_string_helicobacter_cetorum_mit007128_node_name(node_name) {
            return Ok("STRING Helicobacter cetorum MIT007128");
        }

        if is_valid_string_chrysochloris_asiatica_node_name(node_name) {
            return Ok("STRING Chrysochloris asiatica");
        }

        if is_valid_string_baumannia_cicadellinicola_node_name(node_name) {
            return Ok("STRING Baumannia cicadellinicola");
        }

        if is_valid_string_pyrococcus_furiosus_node_name(node_name) {
            return Ok("STRING Pyrococcus furiosus");
        }

        if is_valid_string_alkalilimnicola_ehrlichii_node_name(node_name) {
            return Ok("STRING Alkalilimnicola ehrlichii");
        }

        if is_valid_string_methylocystis_sp_sc2_node_name(node_name) {
            return Ok("STRING Methylocystis sp. SC2");
        }

        if is_valid_string_dermacoccus_sp_ellin185_node_name(node_name) {
            return Ok("STRING Dermacoccus sp. Ellin185");
        }

        if is_valid_string_paenibacillus_graminis_node_name(node_name) {
            return Ok("STRING Paenibacillus graminis");
        }

        if is_valid_string_paenibacillus_odorifer_node_name(node_name) {
            return Ok("STRING Paenibacillus odorifer");
        }

        if is_valid_string_bradyrhizobium_sp_ec33_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. Ec33");
        }

        if is_valid_string_fusobacterium_nucleatum_atcc25586_node_name(node_name) {
            return Ok("STRING Fusobacterium nucleatum ATCC25586");
        }

        if is_valid_string_xanthomonas_axonopodis_citri_node_name(node_name) {
            return Ok("STRING Xanthomonas axonopodis citri");
        }

        if is_valid_string_caulobacter_crescentus_cb15_node_name(node_name) {
            return Ok("STRING Caulobacter crescentus CB15");
        }

        if is_valid_string_corynebacterium_atypicum_node_name(node_name) {
            return Ok("STRING Corynebacterium atypicum");
        }

        if is_valid_string_campylobacter_jejuni_nctc11168_node_name(node_name) {
            return Ok("STRING Campylobacter jejuni NCTC11168");
        }

        if is_valid_string_capsaspora_owczarzaki_node_name(node_name) {
            return Ok("STRING Capsaspora owczarzaki");
        }

        if is_valid_string_methanosarcina_mazei_node_name(node_name) {
            return Ok("STRING Methanosarcina mazei");
        }

        if is_valid_string_chlorobium_tepidum_node_name(node_name) {
            return Ok("STRING Chlorobium tepidum");
        }

        if is_valid_string_sphingomonas_sp_atcc31555_node_name(node_name) {
            return Ok("STRING Sphingomonas sp. ATCC31555");
        }

        if is_valid_string_clostridium_perfringens_atcc13124_node_name(node_name) {
            return Ok("STRING Clostridium perfringens ATCC13124");
        }

        if is_valid_string_haematobacter_massiliensis_node_name(node_name) {
            return Ok("STRING Haematobacter massiliensis");
        }

        if is_valid_string_synechococcus_sp_pcc7336_node_name(node_name) {
            return Ok("STRING Synechococcus sp. PCC7336");
        }

        if is_valid_string_synechococcus_sp_pcc6312_node_name(node_name) {
            return Ok("STRING Synechococcus sp. PCC6312");
        }

        if is_valid_string_thermococcus_nautili_node_name(node_name) {
            return Ok("STRING Thermococcus nautili");
        }

        if is_valid_string_nocardioides_sp_js614_node_name(node_name) {
            return Ok("STRING Nocardioides sp. JS614");
        }

        if is_valid_string_corynebacterium_efficiens_node_name(node_name) {
            return Ok("STRING Corynebacterium efficiens");
        }

        if is_valid_string_burkholderia_sordidicola_node_name(node_name) {
            return Ok("STRING Burkholderia sordidicola");
        }

        if is_valid_string_bradyrhizobium_sp_ai1a2_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. Ai1a2");
        }

        if is_valid_string_corynebacterium_glutamicum_node_name(node_name) {
            return Ok("STRING Corynebacterium glutamicum");
        }

        if is_valid_string_thermosynechococcus_elongatus_node_name(node_name) {
            return Ok("STRING Thermosynechococcus elongatus");
        }

        if is_valid_string_bacillus_anthracis_ames_node_name(node_name) {
            return Ok("STRING Bacillus anthracis Ames");
        }

        if is_valid_string_shigella_flexneri_node_name(node_name) {
            return Ok("STRING Shigella flexneri");
        }

        if is_valid_string_anoxybacillus_gonensis_node_name(node_name) {
            return Ok("STRING Anoxybacillus gonensis");
        }

        if is_valid_string_dickeya_dadantii_3937_node_name(node_name) {
            return Ok("STRING Dickeya dadantii 3937");
        }

        if is_valid_string_buchnera_aphidicola_sg_node_name(node_name) {
            return Ok("STRING Buchnera aphidicola Sg");
        }

        if is_valid_string_coccidioides_posadasii_node_name(node_name) {
            return Ok("STRING Coccidioides posadasii");
        }

        if is_valid_string_escherichia_coli_cft073_node_name(node_name) {
            return Ok("STRING Escherichia coli CFT073");
        }

        if is_valid_string_punctularia_strigosozonata_node_name(node_name) {
            return Ok("STRING Punctularia strigosozonata");
        }

        if is_valid_string_listeria_ivanovii_londoniensis_node_name(node_name) {
            return Ok("STRING Listeria ivanovii londoniensis");
        }

        if is_valid_string_acinetobacter_gerneri_node_name(node_name) {
            return Ok("STRING Acinetobacter gerneri");
        }

        if is_valid_string_acinetobacter_tandoii_node_name(node_name) {
            return Ok("STRING Acinetobacter tandoii");
        }

        if is_valid_string_acinetobacter_tjernbergiae_node_name(node_name) {
            return Ok("STRING Acinetobacter tjernbergiae");
        }

        if is_valid_string_acinetobacter_towneri_node_name(node_name) {
            return Ok("STRING Acinetobacter towneri");
        }

        if is_valid_string_ruminiclostridium_thermocellum_node_name(node_name) {
            return Ok("STRING Ruminiclostridium thermocellum");
        }

        if is_valid_string_leuconostoc_mesenteroides_node_name(node_name) {
            return Ok("STRING Leuconostoc mesenteroides");
        }

        if is_valid_string_saccharophagus_degradans_node_name(node_name) {
            return Ok("STRING Saccharophagus degradans");
        }

        if is_valid_string_oenococcus_oeni_node_name(node_name) {
            return Ok("STRING Oenococcus oeni");
        }

        if is_valid_string_trichodesmium_erythraeum_node_name(node_name) {
            return Ok("STRING Trichodesmium erythraeum");
        }

        if is_valid_string_tropheryma_whipplei_node_name(node_name) {
            return Ok("STRING Tropheryma whipplei");
        }

        if is_valid_string_tannerella_forsythia_node_name(node_name) {
            return Ok("STRING Tannerella forsythia");
        }

        if is_valid_string_blochmannia_floridanus_node_name(node_name) {
            return Ok("STRING Blochmannia floridanus");
        }

        if is_valid_string_melampsora_laricipopulina_node_name(node_name) {
            return Ok("STRING Melampsora laricipopulina");
        }

        if is_valid_string_sulfurihydrogenibium_azorense_node_name(node_name) {
            return Ok("STRING Sulfurihydrogenibium azorense");
        }

        if is_valid_string_koribacter_versatilis_node_name(node_name) {
            return Ok("STRING Koribacter versatilis");
        }

        if is_valid_string_herminiimonas_arsenicoxydans_node_name(node_name) {
            return Ok("STRING Herminiimonas arsenicoxydans");
        }

        if is_valid_string_haemophilus_somnus_node_name(node_name) {
            return Ok("STRING Haemophilus somnus");
        }

        if is_valid_string_pseudomonas_syringae_b728a_node_name(node_name) {
            return Ok("STRING Pseudomonas syringae B728a");
        }

        if is_valid_string_ehrlichia_chaffeensis_node_name(node_name) {
            return Ok("STRING Ehrlichia chaffeensis");
        }

        if is_valid_string_pseudomonas_fluorescens_pf01_node_name(node_name) {
            return Ok("STRING Pseudomonas fluorescens Pf01");
        }

        if is_valid_string_desulfovibrio_alaskensis_node_name(node_name) {
            return Ok("STRING Desulfovibrio alaskensis");
        }

        if is_valid_string_neptuniibacter_caesariensis_node_name(node_name) {
            return Ok("STRING Neptuniibacter caesariensis");
        }

        if is_valid_string_amycolatopsis_japonica_node_name(node_name) {
            return Ok("STRING Amycolatopsis japonica");
        }

        if is_valid_string_amycolatopsis_vancoresmycina_node_name(node_name) {
            return Ok("STRING Amycolatopsis vancoresmycina");
        }

        if is_valid_string_carnobacterium_sp_174_node_name(node_name) {
            return Ok("STRING Carnobacterium sp. 174");
        }

        if is_valid_string_fomitiporia_mediterranea_node_name(node_name) {
            return Ok("STRING Fomitiporia mediterranea");
        }

        if is_valid_string_chaetomium_thermophilum_node_name(node_name) {
            return Ok("STRING Chaetomium thermophilum");
        }

        if is_valid_string_streptococcus_mutans_ua159_node_name(node_name) {
            return Ok("STRING Streptococcus mutans UA159");
        }

        if is_valid_string_streptococcus_agalactiae_nem316_node_name(node_name) {
            return Ok("STRING Streptococcus agalactiae NEM316");
        }

        if is_valid_string_allokutzneria_albata_node_name(node_name) {
            return Ok("STRING Allokutzneria albata");
        }

        if is_valid_string_chlorogloeopsis_fritschii_node_name(node_name) {
            return Ok("STRING Chlorogloeopsis fritschii");
        }

        if is_valid_string_shewanella_oneidensis_node_name(node_name) {
            return Ok("STRING Shewanella oneidensis");
        }

        if is_valid_string_anaplasma_phagocytophilum_node_name(node_name) {
            return Ok("STRING Anaplasma phagocytophilum");
        }

        if is_valid_string_yersinia_pestis_co92_node_name(node_name) {
            return Ok("STRING Yersinia pestis CO92");
        }

        if is_valid_string_larimichthys_crocea_node_name(node_name) {
            return Ok("STRING Larimichthys crocea");
        }

        if is_valid_string_enhygromyxa_salina_node_name(node_name) {
            return Ok("STRING Enhygromyxa salina");
        }

        if is_valid_string_pseudomonas_rhizosphaerae_node_name(node_name) {
            return Ok("STRING Pseudomonas rhizosphaerae");
        }

        if is_valid_string_croceibacter_atlanticus_node_name(node_name) {
            return Ok("STRING Croceibacter atlanticus");
        }

        if is_valid_string_burkholderia_cenocepacia_node_name(node_name) {
            return Ok("STRING Burkholderia cenocepacia");
        }

        if is_valid_string_mycobacterium_marinum_node_name(node_name) {
            return Ok("STRING Mycobacterium marinum");
        }

        if is_valid_string_pseudomonas_fluorescens_sbw25_node_name(node_name) {
            return Ok("STRING Pseudomonas fluorescens SBW25");
        }

        if is_valid_string_rhizobium_leguminosarum_3841_node_name(node_name) {
            return Ok("STRING Rhizobium leguminosarum 3841");
        }

        if is_valid_string_bifidobacterium_longum_node_name(node_name) {
            return Ok("STRING Bifidobacterium longum");
        }

        if is_valid_string_bifidobacterium_psychraerophilum_node_name(node_name) {
            return Ok("STRING Bifidobacterium psychraerophilum");
        }

        if is_valid_string_bacillus_vietnamensis_node_name(node_name) {
            return Ok("STRING Bacillus vietnamensis");
        }

        if is_valid_string_pectobacterium_atrosepticum_node_name(node_name) {
            return Ok("STRING Pectobacterium atrosepticum");
        }

        if is_valid_string_salmonella_bongori_nctc12419_node_name(node_name) {
            return Ok("STRING Salmonella bongori NCTC12419");
        }

        if is_valid_string_streptococcus_uberis_node_name(node_name) {
            return Ok("STRING Streptococcus uberis");
        }

        if is_valid_string_chlamydophila_abortus_node_name(node_name) {
            return Ok("STRING Chlamydophila abortus");
        }

        if is_valid_string_aquilegia_coerulea_node_name(node_name) {
            return Ok("STRING Aquilegia coerulea");
        }

        if is_valid_string_micromonospora_sp_atcc39149_node_name(node_name) {
            return Ok("STRING Micromonospora sp. ATCC39149");
        }

        if is_valid_string_salmonella_enterica_ct18_node_name(node_name) {
            return Ok("STRING Salmonella enterica CT18");
        }

        if is_valid_string_pseudomonas_protegens_pf5_node_name(node_name) {
            return Ok("STRING Pseudomonas protegens Pf5");
        }

        if is_valid_string_lactobacillus_plantarum_node_name(node_name) {
            return Ok("STRING Lactobacillus plantarum");
        }

        if is_valid_string_treponema_putidum_node_name(node_name) {
            return Ok("STRING Treponema putidum");
        }

        if is_valid_string_moniliophthora_roreri_node_name(node_name) {
            return Ok("STRING Moniliophthora roreri");
        }

        if is_valid_string_oceanobacillus_iheyensis_node_name(node_name) {
            return Ok("STRING Oceanobacillus iheyensis");
        }

        if is_valid_string_mastigocladopsis_repens_node_name(node_name) {
            return Ok("STRING Mastigocladopsis repens");
        }

        if is_valid_string_synechococcus_sp_rs9916_node_name(node_name) {
            return Ok("STRING Synechococcus sp. RS9916");
        }

        if is_valid_string_synechococcus_sp_rs9917_node_name(node_name) {
            return Ok("STRING Synechococcus sp. RS9917");
        }

        if is_valid_string_mannheimia_succiniciproducens_node_name(node_name) {
            return Ok("STRING Mannheimia succiniciproducens");
        }

        if is_valid_string_frankia_sp_bmg512_node_name(node_name) {
            return Ok("STRING Frankia sp. BMG512");
        }

        if is_valid_string_neorickettsia_sennetsu_node_name(node_name) {
            return Ok("STRING Neorickettsia sennetsu");
        }

        if is_valid_string_natrinema_altunense_node_name(node_name) {
            return Ok("STRING Natrinema altunense");
        }

        if is_valid_string_kocuria_marina_node_name(node_name) {
            return Ok("STRING Kocuria marina");
        }

        if is_valid_string_togninia_minima_node_name(node_name) {
            return Ok("STRING Togninia minima");
        }

        if is_valid_string_pseudomonas_syringae_tomato_node_name(node_name) {
            return Ok("STRING Pseudomonas syringae tomato");
        }

        if is_valid_string_vibrio_parahaemolyticus_node_name(node_name) {
            return Ok("STRING Vibrio parahaemolyticus");
        }

        if is_valid_string_bacillus_subtilis_node_name(node_name) {
            return Ok("STRING Bacillus subtilis");
        }

        if is_valid_string_aquifex_aeolicus_node_name(node_name) {
            return Ok("STRING Aquifex aeolicus");
        }

        if is_valid_string_archaeoglobus_fulgidus_node_name(node_name) {
            return Ok("STRING Archaeoglobus fulgidus");
        }

        if is_valid_string_methanobrevibacter_sp_abm4_node_name(node_name) {
            return Ok("STRING Methanobrevibacter sp. AbM4");
        }

        if is_valid_string_bradyrhizobium_diazoefficiens_node_name(node_name) {
            return Ok("STRING Bradyrhizobium diazoefficiens");
        }

        if is_valid_string_brucella_melitensis_node_name(node_name) {
            return Ok("STRING Brucella melitensis");
        }

        if is_valid_string_buchnera_aphidicola_bp_node_name(node_name) {
            return Ok("STRING Buchnera aphidicola Bp");
        }

        if is_valid_string_pyrus_x_bretschneideri_node_name(node_name) {
            return Ok("STRING Pyrus x bretschneideri");
        }

        if is_valid_string_myotis_davidii_node_name(node_name) {
            return Ok("STRING Myotis davidii");
        }

        if is_valid_string_shewanella_piezotolerans_node_name(node_name) {
            return Ok("STRING Shewanella piezotolerans");
        }

        if is_valid_string_marinobacter_adhaerens_node_name(node_name) {
            return Ok("STRING Marinobacter adhaerens");
        }

        if is_valid_string_enterococcus_faecalis_v583_node_name(node_name) {
            return Ok("STRING Enterococcus faecalis V583");
        }

        if is_valid_string_bacteroides_thetaiotaomicron_node_name(node_name) {
            return Ok("STRING Bacteroides thetaiotaomicron");
        }

        if is_valid_string_grosmannia_clavigera_node_name(node_name) {
            return Ok("STRING Grosmannia clavigera");
        }

        if is_valid_string_bigelowiella_natans_node_name(node_name) {
            return Ok("STRING Bigelowiella natans");
        }

        if is_valid_string_coxiella_burnetii_node_name(node_name) {
            return Ok("STRING Coxiella burnetii");
        }

        if is_valid_string_streptomyces_avermitilis_node_name(node_name) {
            return Ok("STRING Streptomyces avermitilis");
        }

        if is_valid_string_chlamydophila_caviae_node_name(node_name) {
            return Ok("STRING Chlamydophila caviae");
        }

        if is_valid_string_actinobacillus_pleuropneumoniae_1_4074_node_name(node_name) {
            return Ok("STRING Actinobacillus pleuropneumoniae 1 4074");
        }

        if is_valid_string_hyphomonas_neptunium_node_name(node_name) {
            return Ok("STRING Hyphomonas neptunium");
        }

        if is_valid_string_nitrosomonas_europaea_node_name(node_name) {
            return Ok("STRING Nitrosomonas europaea");
        }

        if is_valid_string_beijerinckia_mobilis_node_name(node_name) {
            return Ok("STRING Beijerinckia mobilis");
        }

        if is_valid_string_halomonas_alkaliantarctica_node_name(node_name) {
            return Ok("STRING Halomonas alkaliantarctica");
        }

        if is_valid_string_synechococcus_sp_cb0101_node_name(node_name) {
            return Ok("STRING Synechococcus sp. CB0101");
        }

        if is_valid_string_acidovorax_sp_js42_node_name(node_name) {
            return Ok("STRING Acidovorax sp. JS42");
        }

        if is_valid_string_haemophilus_ducreyi_node_name(node_name) {
            return Ok("STRING Haemophilus ducreyi");
        }

        if is_valid_string_solibacter_usitatus_node_name(node_name) {
            return Ok("STRING Solibacter usitatus");
        }

        if is_valid_string_rhodococcus_erythropolis_node_name(node_name) {
            return Ok("STRING Rhodococcus erythropolis");
        }

        if is_valid_string_anaplasma_marginale_maries_node_name(node_name) {
            return Ok("STRING Anaplasma marginale Maries");
        }

        if is_valid_string_pseudoalteromonas_sp_sm9913_node_name(node_name) {
            return Ok("STRING Pseudoalteromonas sp. SM9913");
        }

        if is_valid_string_helicobacter_hepaticus_node_name(node_name) {
            return Ok("STRING Helicobacter hepaticus");
        }

        if is_valid_string_geobacillus_kaustophilus_node_name(node_name) {
            return Ok("STRING Geobacillus kaustophilus");
        }

        if is_valid_string_streptacidiphilus_jiangxiensis_node_name(node_name) {
            return Ok("STRING Streptacidiphilus jiangxiensis");
        }

        if is_valid_string_alcanivorax_sp_dg881_node_name(node_name) {
            return Ok("STRING Alcanivorax sp. DG881");
        }

        if is_valid_string_chryseobacterium_formosense_node_name(node_name) {
            return Ok("STRING Chryseobacterium formosense");
        }

        if is_valid_string_scalindua_brodae_node_name(node_name) {
            return Ok("STRING Scalindua brodae");
        }

        if is_valid_string_pseudomonas_alkylphenolia_node_name(node_name) {
            return Ok("STRING Pseudomonas alkylphenolia");
        }

        if is_valid_string_erythrobacter_sp_nap1_node_name(node_name) {
            return Ok("STRING Erythrobacter sp. NAP1");
        }

        if is_valid_string_acidobacterium_capsulatum_node_name(node_name) {
            return Ok("STRING Acidobacterium capsulatum");
        }

        if is_valid_string_verrucomicrobium_spinosum_node_name(node_name) {
            return Ok("STRING Verrucomicrobium spinosum");
        }

        if is_valid_string_anabaena_variabilis_node_name(node_name) {
            return Ok("STRING Anabaena variabilis");
        }

        if is_valid_string_halobacillus_dabanensis_node_name(node_name) {
            return Ok("STRING Halobacillus dabanensis");
        }

        if is_valid_string_ostreococcus_lucimarinus_node_name(node_name) {
            return Ok("STRING Ostreococcus lucimarinus");
        }

        if is_valid_string_porphyromonas_gingivalis_w83_node_name(node_name) {
            return Ok("STRING Porphyromonas gingivalis W83");
        }

        if is_valid_string_rhodopirellula_baltica_node_name(node_name) {
            return Ok("STRING Rhodopirellula baltica");
        }

        if is_valid_string_acidithiobacillus_ferrooxidans_atcc23270_node_name(node_name) {
            return Ok("STRING Acidithiobacillus ferrooxidans ATCC23270");
        }

        if is_valid_string_burkholderia_mallei_node_name(node_name) {
            return Ok("STRING Burkholderia mallei");
        }

        if is_valid_string_chlamydia_muridarum_node_name(node_name) {
            return Ok("STRING Chlamydia muridarum");
        }

        if is_valid_string_dehalococcoides_mccartyi_195_node_name(node_name) {
            return Ok("STRING Dehalococcoides mccartyi 195");
        }

        if is_valid_string_deinococcus_radiodurans_node_name(node_name) {
            return Ok("STRING Deinococcus radiodurans");
        }

        if is_valid_string_geobacter_sulfurreducens_node_name(node_name) {
            return Ok("STRING Geobacter sulfurreducens");
        }

        if is_valid_string_methanocaldococcus_jannaschii_node_name(node_name) {
            return Ok("STRING Methanocaldococcus jannaschii");
        }

        if is_valid_string_methylococcus_capsulatus_node_name(node_name) {
            return Ok("STRING Methylococcus capsulatus");
        }

        if is_valid_string_photorhabdus_luminescens_node_name(node_name) {
            return Ok("STRING Photorhabdus luminescens");
        }

        if is_valid_string_mycoplasma_arthritidis_node_name(node_name) {
            return Ok("STRING Mycoplasma arthritidis");
        }

        if is_valid_string_mycoplasma_genitalium_node_name(node_name) {
            return Ok("STRING Mycoplasma genitalium");
        }

        if is_valid_string_thermotoga_maritima_node_name(node_name) {
            return Ok("STRING Thermotoga maritima");
        }

        if is_valid_string_treponema_denticola_node_name(node_name) {
            return Ok("STRING Treponema denticola");
        }

        if is_valid_string_treponema_pallidum_node_name(node_name) {
            return Ok("STRING Treponema pallidum");
        }

        if is_valid_string_vibrio_cholerae_o1_node_name(node_name) {
            return Ok("STRING Vibrio cholerae O1");
        }

        if is_valid_string_chromobacterium_violaceum_node_name(node_name) {
            return Ok("STRING Chromobacterium violaceum");
        }

        if is_valid_string_pseudomonas_lutea_node_name(node_name) {
            return Ok("STRING Pseudomonas lutea");
        }

        if is_valid_string_cynoglossus_semilaevis_node_name(node_name) {
            return Ok("STRING Cynoglossus semilaevis");
        }

        if is_valid_string_caedibacter_acanthamoebae_node_name(node_name) {
            return Ok("STRING Caedibacter acanthamoebae");
        }

        if is_valid_string_paracaedibacter_symbiosus_node_name(node_name) {
            return Ok("STRING Paracaedibacter symbiosus");
        }

        if is_valid_string_wallemia_ichthyophaga_node_name(node_name) {
            return Ok("STRING Wallemia ichthyophaga");
        }

        if is_valid_string_carboxydothermus_hydrogenoformans_node_name(node_name) {
            return Ok("STRING Carboxydothermus hydrogenoformans");
        }

        if is_valid_string_dichelobacter_nodosus_node_name(node_name) {
            return Ok("STRING Dichelobacter nodosus");
        }

        if is_valid_string_mycobacterium_smegmatis_mc2155_node_name(node_name) {
            return Ok("STRING Mycobacterium smegmatis MC2155");
        }

        if is_valid_string_myxococcus_xanthus_node_name(node_name) {
            return Ok("STRING Myxococcus xanthus");
        }

        if is_valid_string_ruminococcus_albus_8_node_name(node_name) {
            return Ok("STRING Ruminococcus albus 8");
        }

        if is_valid_string_ruegeria_pomeroyi_node_name(node_name) {
            return Ok("STRING Ruegeria pomeroyi");
        }

        if is_valid_string_streptococcus_mitis_nctc_12261_node_name(node_name) {
            return Ok("STRING Streptococcus mitis NCTC 12261");
        }

        if is_valid_string_tupaia_chinensis_node_name(node_name) {
            return Ok("STRING Tupaia chinensis");
        }

        if is_valid_string_thermococcus_sp_am4_node_name(node_name) {
            return Ok("STRING Thermococcus sp. AM4");
        }

        if is_valid_string_nocardia_farcinica_node_name(node_name) {
            return Ok("STRING Nocardia farcinica");
        }

        if is_valid_string_planctomycete_ksu1_node_name(node_name) {
            return Ok("STRING planctomycete KSU1");
        }

        if is_valid_string_gammaproteobacteria_bacterium_htcc2143_node_name(node_name) {
            return Ok("STRING Gammaproteobacteria bacterium HTCC2143");
        }

        if is_valid_string_gammaproteobacteria_bacterium_htcc2148_node_name(node_name) {
            return Ok("STRING Gammaproteobacteria bacterium HTCC2148");
        }

        if is_valid_string_gammaproteobacteria_bacterium_htcc2080_node_name(node_name) {
            return Ok("STRING Gammaproteobacteria bacterium HTCC2080");
        }

        if is_valid_string_coccomyxa_subellipsoidea_node_name(node_name) {
            return Ok("STRING Coccomyxa subellipsoidea");
        }

        if is_valid_string_gloeobacter_violaceus_node_name(node_name) {
            return Ok("STRING Gloeobacter violaceus");
        }

        if is_valid_string_chroococcidiopsis_thermalis_node_name(node_name) {
            return Ok("STRING Chroococcidiopsis thermalis");
        }

        if is_valid_string_oceanicola_batsensis_node_name(node_name) {
            return Ok("STRING Oceanicola batsensis");
        }

        if is_valid_string_streptomyces_sp_c_node_name(node_name) {
            return Ok("STRING Streptomyces sp. C");
        }

        if is_valid_string_ehrlichia_ruminantium_node_name(node_name) {
            return Ok("STRING Ehrlichia ruminantium");
        }

        if is_valid_string_dehalococcoides_mccartyi_cbdb1_node_name(node_name) {
            return Ok("STRING Dehalococcoides mccartyi CBDB1");
        }

        if is_valid_string_bordetella_bronchiseptica_node_name(node_name) {
            return Ok("STRING Bordetella bronchiseptica");
        }

        if is_valid_string_bordetella_pertussis_tohama_node_name(node_name) {
            return Ok("STRING Bordetella pertussis Tohama");
        }

        if is_valid_string_lactobacillus_johnsonii_node_name(node_name) {
            return Ok("STRING Lactobacillus johnsonii");
        }

        if is_valid_string_rickettsia_typhi_node_name(node_name) {
            return Ok("STRING Rickettsia typhi");
        }

        if is_valid_string_kitasatospora_arboriphila_node_name(node_name) {
            return Ok("STRING Kitasatospora arboriphila");
        }

        if is_valid_string_mycobacterium_cosmeticum_node_name(node_name) {
            return Ok("STRING Mycobacterium cosmeticum");
        }

        if is_valid_string_rhodopseudomonas_palustris_cga009_node_name(node_name) {
            return Ok("STRING Rhodopseudomonas palustris CGA009");
        }

        if is_valid_string_psychrobacter_arcticus_node_name(node_name) {
            return Ok("STRING Psychrobacter arcticus");
        }

        if is_valid_string_bacillus_anthracis_sterne_node_name(node_name) {
            return Ok("STRING Bacillus anthracis Sterne");
        }

        if is_valid_string_nitrosomonas_sp_is79a3_node_name(node_name) {
            return Ok("STRING Nitrosomonas sp. Is79A3");
        }

        if is_valid_string_buchnera_aphidicola_ctu_node_name(node_name) {
            return Ok("STRING Buchnera aphidicola Ctu");
        }

        if is_valid_string_mycobacterium_avium_paratuberculosis_node_name(node_name) {
            return Ok("STRING Mycobacterium avium paratuberculosis");
        }

        if is_valid_string_exiguobacterium_sibiricum_node_name(node_name) {
            return Ok("STRING Exiguobacterium sibiricum");
        }

        if is_valid_string_mycoplasma_hyopneumoniae_j_node_name(node_name) {
            return Ok("STRING Mycoplasma hyopneumoniae J");
        }

        if is_valid_string_mycoplasma_synoviae_node_name(node_name) {
            return Ok("STRING Mycoplasma synoviae");
        }

        if is_valid_string_thermus_thermophilus_hb27_node_name(node_name) {
            return Ok("STRING Thermus thermophilus HB27");
        }

        if is_valid_string_phytoplasma_onion_yellows_node_name(node_name) {
            return Ok("STRING Phytoplasma onion yellows");
        }

        if is_valid_string_verrucosispora_maris_node_name(node_name) {
            return Ok("STRING Verrucosispora maris");
        }

        if is_valid_string_pneumocystis_murina_node_name(node_name) {
            return Ok("STRING Pneumocystis murina");
        }

        if is_valid_string_picrophilus_torridus_node_name(node_name) {
            return Ok("STRING Picrophilus torridus");
        }

        if is_valid_string_ralstonia_eutropha_jmp134_node_name(node_name) {
            return Ok("STRING Ralstonia eutropha JMP134");
        }

        if is_valid_string_streptococcus_thermophilus_lmg18311_node_name(node_name) {
            return Ok("STRING Streptococcus thermophilus LMG18311");
        }

        if is_valid_string_protochlamydia_amoebophila_node_name(node_name) {
            return Ok("STRING Protochlamydia amoebophila");
        }

        if is_valid_string_chlamydophila_felis_node_name(node_name) {
            return Ok("STRING Chlamydophila felis");
        }

        if is_valid_string_zymomonas_mobilis_zm4_node_name(node_name) {
            return Ok("STRING Zymomonas mobilis ZM4");
        }

        if is_valid_string_capsella_grandiflora_node_name(node_name) {
            return Ok("STRING Capsella grandiflora");
        }

        if is_valid_string_bdellovibrio_bacteriovorus_hd100_node_name(node_name) {
            return Ok("STRING Bdellovibrio bacteriovorus HD100");
        }

        if is_valid_string_pseudomonas_syringae_phaseolicola_node_name(node_name) {
            return Ok("STRING Pseudomonas syringae phaseolicola");
        }

        if is_valid_string_prevotella_ruminicola_node_name(node_name) {
            return Ok("STRING Prevotella ruminicola");
        }

        if is_valid_string_moorella_thermoacetica_node_name(node_name) {
            return Ok("STRING Moorella thermoacetica");
        }

        if is_valid_string_byssochlamys_spectabilis_node_name(node_name) {
            return Ok("STRING Byssochlamys spectabilis");
        }

        if is_valid_string_methylobacillus_flagellatus_node_name(node_name) {
            return Ok("STRING Methylobacillus flagellatus");
        }

        if is_valid_string_mesoplasma_florum_node_name(node_name) {
            return Ok("STRING Mesoplasma florum");
        }

        if is_valid_string_bacillus_cibi_node_name(node_name) {
            return Ok("STRING Bacillus cibi");
        }

        if is_valid_string_rubrobacter_xylanophilus_node_name(node_name) {
            return Ok("STRING Rubrobacter xylanophilus");
        }

        if is_valid_string_cupriavidus_metallidurans_node_name(node_name) {
            return Ok("STRING Cupriavidus metallidurans");
        }

        if is_valid_string_burkholderia_xenovorans_node_name(node_name) {
            return Ok("STRING Burkholderia xenovorans");
        }

        if is_valid_string_chryseobacterium_antarcticum_node_name(node_name) {
            return Ok("STRING Chryseobacterium antarcticum");
        }

        if is_valid_string_porphyromonas_gingivicanis_node_name(node_name) {
            return Ok("STRING Porphyromonas gingivicanis");
        }

        if is_valid_string_chelativorans_sp_bnc1_node_name(node_name) {
            return Ok("STRING Chelativorans sp. BNC1");
        }

        if is_valid_string_thalassobacter_stenotrophicus_node_name(node_name) {
            return Ok("STRING Thalassobacter stenotrophicus");
        }

        if is_valid_string_sinorhizobium_meliloti_node_name(node_name) {
            return Ok("STRING Sinorhizobium meliloti");
        }

        if is_valid_string_mesorhizobium_loti_maff303099_node_name(node_name) {
            return Ok("STRING Mesorhizobium loti MAFF303099");
        }

        if is_valid_string_kineococcus_radiotolerans_node_name(node_name) {
            return Ok("STRING Kineococcus radiotolerans");
        }

        if is_valid_string_methanococcus_maripaludis_s2_node_name(node_name) {
            return Ok("STRING Methanococcus maripaludis S2");
        }

        if is_valid_string_ralstonia_solanacearum_gmi1000_node_name(node_name) {
            return Ok("STRING Ralstonia solanacearum GMI1000");
        }

        if is_valid_string_propionibacterium_acnes_kpa171202_node_name(node_name) {
            return Ok("STRING Propionibacterium acnes KPA171202");
        }

        if is_valid_string_mycoplasma_mobile_node_name(node_name) {
            return Ok("STRING Mycoplasma mobile");
        }

        if is_valid_string_paenibacillus_wynnii_node_name(node_name) {
            return Ok("STRING Paenibacillus wynnii");
        }

        if is_valid_string_natronomonas_moolapensis_node_name(node_name) {
            return Ok("STRING Natronomonas moolapensis");
        }

        if is_valid_string_synechococcus_elongatus_pcc6301_node_name(node_name) {
            return Ok("STRING Synechococcus elongatus PCC6301");
        }

        if is_valid_string_burkholderia_vietnamiensis_node_name(node_name) {
            return Ok("STRING Burkholderia vietnamiensis");
        }

        if is_valid_string_ehrlichia_canis_node_name(node_name) {
            return Ok("STRING Ehrlichia canis");
        }

        if is_valid_string_rhodospirillum_rubrum_node_name(node_name) {
            return Ok("STRING Rhodospirillum rubrum");
        }

        if is_valid_string_methanosarcina_barkeri_node_name(node_name) {
            return Ok("STRING Methanosarcina barkeri");
        }

        if is_valid_string_cytophaga_hutchinsonii_node_name(node_name) {
            return Ok("STRING Cytophaga hutchinsonii");
        }

        if is_valid_string_geobacter_metallireducens_node_name(node_name) {
            return Ok("STRING Geobacter metallireducens");
        }

        if is_valid_string_thermobifida_fusca_node_name(node_name) {
            return Ok("STRING Thermobifida fusca");
        }

        if is_valid_string_marinobacter_sp_elb17_node_name(node_name) {
            return Ok("STRING Marinobacter sp. ELB17");
        }

        if is_valid_string_anabaena_cylindrica_node_name(node_name) {
            return Ok("STRING Anabaena cylindrica");
        }

        if is_valid_string_leptolyngbya_boryana_node_name(node_name) {
            return Ok("STRING Leptolyngbya boryana");
        }

        if is_valid_string_aggregatibacter_actinomycetemcomitans_hk1651_node_name(node_name) {
            return Ok("STRING Aggregatibacter actinomycetemcomitans HK1651");
        }

        if is_valid_string_aeropyrum_pernix_node_name(node_name) {
            return Ok("STRING Aeropyrum pernix");
        }

        if is_valid_string_bacillus_halodurans_node_name(node_name) {
            return Ok("STRING Bacillus halodurans");
        }

        if is_valid_string_bacteroides_fragilis_nctc_9343_node_name(node_name) {
            return Ok("STRING Bacteroides fragilis NCTC 9343");
        }

        if is_valid_string_burkholderia_pseudomallei_k96243_node_name(node_name) {
            return Ok("STRING Burkholderia pseudomallei K96243");
        }

        if is_valid_string_clostridium_acetobutylicum_node_name(node_name) {
            return Ok("STRING Clostridium acetobutylicum");
        }

        if is_valid_string_peptoclostridium_difficile_630_node_name(node_name) {
            return Ok("STRING Peptoclostridium difficile 630");
        }

        if is_valid_string_gluconacetobacter_diazotrophicus_node_name(node_name) {
            return Ok("STRING Gluconacetobacter diazotrophicus");
        }

        if is_valid_string_haloarcula_marismortui_node_name(node_name) {
            return Ok("STRING Haloarcula marismortui");
        }

        if is_valid_string_lactobacillus_acidophilus_ncfm_node_name(node_name) {
            return Ok("STRING Lactobacillus acidophilus NCFM");
        }

        if is_valid_string_lactococcus_lactis_il1403_node_name(node_name) {
            return Ok("STRING Lactococcus lactis Il1403");
        }

        if is_valid_string_legionella_pneumophila_philadelphia_node_name(node_name) {
            return Ok("STRING Legionella pneumophila Philadelphia");
        }

        if is_valid_string_listeria_innocua_node_name(node_name) {
            return Ok("STRING Listeria innocua");
        }

        if is_valid_string_methylobacterium_extorquens_am1_node_name(node_name) {
            return Ok("STRING Methylobacterium extorquens AM1");
        }

        if is_valid_string_mycobacterium_leprae_node_name(node_name) {
            return Ok("STRING Mycobacterium leprae");
        }

        if is_valid_string_mycoplasma_mycoides_node_name(node_name) {
            return Ok("STRING Mycoplasma mycoides");
        }

        if is_valid_string_mycoplasma_penetrans_node_name(node_name) {
            return Ok("STRING Mycoplasma penetrans");
        }

        if is_valid_string_mycoplasma_pulmonis_node_name(node_name) {
            return Ok("STRING Mycoplasma pulmonis");
        }

        if is_valid_string_pyrococcus_abyssi_node_name(node_name) {
            return Ok("STRING Pyrococcus abyssi");
        }

        if is_valid_string_rhodobacter_capsulatus_node_name(node_name) {
            return Ok("STRING Rhodobacter capsulatus");
        }

        if is_valid_string_rhodobacter_sphaeroides_241_node_name(node_name) {
            return Ok("STRING Rhodobacter sphaeroides 241");
        }

        if is_valid_string_rickettsia_prowazekii_node_name(node_name) {
            return Ok("STRING Rickettsia prowazekii");
        }

        if is_valid_string_rickettsia_sibirica_node_name(node_name) {
            return Ok("STRING Rickettsia sibirica");
        }

        if is_valid_string_hyaloperonospora_arabidopsidis_node_name(node_name) {
            return Ok("STRING Hyaloperonospora arabidopsidis");
        }

        if is_valid_string_sulfolobus_solfataricus_node_name(node_name) {
            return Ok("STRING Sulfolobus solfataricus");
        }

        if is_valid_string_sulfolobus_tokodaii_node_name(node_name) {
            return Ok("STRING Sulfolobus tokodaii");
        }

        if is_valid_string_caldanaerobacter_subterraneus_node_name(node_name) {
            return Ok("STRING Caldanaerobacter subterraneus");
        }

        if is_valid_string_thermoplasma_acidophilum_node_name(node_name) {
            return Ok("STRING Thermoplasma acidophilum");
        }

        if is_valid_string_thermoplasma_volcanium_node_name(node_name) {
            return Ok("STRING Thermoplasma volcanium");
        }

        if is_valid_string_ureaplasma_parvum_node_name(node_name) {
            return Ok("STRING Ureaplasma parvum");
        }

        if is_valid_string_wolinella_succinogenes_node_name(node_name) {
            return Ok("STRING Wolinella succinogenes");
        }

        if is_valid_string_candida_orthopsilosis_node_name(node_name) {
            return Ok("STRING Candida orthopsilosis");
        }

        if is_valid_string_serratia_marcescens_db11_node_name(node_name) {
            return Ok("STRING Serratia marcescens Db11");
        }

        if is_valid_string_microbacterium_oleivorans_node_name(node_name) {
            return Ok("STRING Microbacterium oleivorans");
        }

        if is_valid_string_pediococcus_pentosaceus_node_name(node_name) {
            return Ok("STRING Pediococcus pentosaceus");
        }

        if is_valid_string_diplosphaera_colitermitum_node_name(node_name) {
            return Ok("STRING Diplosphaera colitermitum");
        }

        if is_valid_string_acidobacteriaceae_bacterium_taa166_node_name(node_name) {
            return Ok("STRING Acidobacteriaceae bacterium TAA166");
        }

        if is_valid_string_bacillus_licheniformis_node_name(node_name) {
            return Ok("STRING Bacillus licheniformis");
        }

        if is_valid_string_novosphingobium_aromaticivorans_node_name(node_name) {
            return Ok("STRING Novosphingobium aromaticivorans");
        }

        if is_valid_string_pseudogulbenkiania_ferrooxidans_node_name(node_name) {
            return Ok("STRING Pseudogulbenkiania ferrooxidans");
        }

        if is_valid_string_staphylococcus_haemolyticus_node_name(node_name) {
            return Ok("STRING Staphylococcus haemolyticus");
        }

        if is_valid_string_leifsonia_xyli_ctcb07_node_name(node_name) {
            return Ok("STRING Leifsonia xyli CTCB07");
        }

        if is_valid_string_caenorhabditis_japonica_node_name(node_name) {
            return Ok("STRING Caenorhabditis japonica");
        }

        if is_valid_string_bartonella_quintana_node_name(node_name) {
            return Ok("STRING Bartonella quintana");
        }

        if is_valid_string_bartonella_henselae_node_name(node_name) {
            return Ok("STRING Bartonella henselae");
        }

        if is_valid_string_pseudoalteromonas_sp_bsw20308_node_name(node_name) {
            return Ok("STRING Pseudoalteromonas sp. Bsw20308");
        }

        if is_valid_string_idiomarina_loihiensis_node_name(node_name) {
            return Ok("STRING Idiomarina loihiensis");
        }

        if is_valid_string_streptomyces_flavovariabilis_node_name(node_name) {
            return Ok("STRING Streptomyces flavovariabilis");
        }

        if is_valid_string_streptomyces_xylophagus_node_name(node_name) {
            return Ok("STRING Streptomyces xylophagus");
        }

        if is_valid_string_streptomyces_fulvoviolaceus_node_name(node_name) {
            return Ok("STRING Streptomyces fulvoviolaceus");
        }

        if is_valid_string_amycolatopsis_rifamycinica_node_name(node_name) {
            return Ok("STRING Amycolatopsis rifamycinica");
        }

        if is_valid_string_bradyrhizobium_sp_btai1_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. BTAi1");
        }

        if is_valid_string_renibacterium_salmoninarum_node_name(node_name) {
            return Ok("STRING Renibacterium salmoninarum");
        }

        if is_valid_string_thermodesulfovibrio_yellowstonii_node_name(node_name) {
            return Ok("STRING Thermodesulfovibrio yellowstonii");
        }

        if is_valid_string_thermodesulfobacterium_commune_node_name(node_name) {
            return Ok("STRING Thermodesulfobacterium commune");
        }

        if is_valid_string_mycoplasma_bovis_node_name(node_name) {
            return Ok("STRING Mycoplasma bovis");
        }

        if is_valid_string_chlorobium_limicola_node_name(node_name) {
            return Ok("STRING Chlorobium limicola");
        }

        if is_valid_string_chlorobium_phaeobacteroides_dsm266_node_name(node_name) {
            return Ok("STRING Chlorobium phaeobacteroides DSM266");
        }

        if is_valid_string_chlorobium_phaeovibrioides_node_name(node_name) {
            return Ok("STRING Chlorobium phaeovibrioides");
        }

        if is_valid_string_arthrobacter_aurescens_node_name(node_name) {
            return Ok("STRING Arthrobacter aurescens");
        }

        if is_valid_string_anaeromyxobacter_dehalogenans_node_name(node_name) {
            return Ok("STRING Anaeromyxobacter dehalogenans");
        }

        if is_valid_string_chromohalobacter_salexigens_node_name(node_name) {
            return Ok("STRING Chromohalobacter salexigens");
        }

        if is_valid_string_arthrobacter_sp_fb24_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. FB24");
        }

        if is_valid_string_jannaschia_sp_ccs1_node_name(node_name) {
            return Ok("STRING Jannaschia sp. CCS1");
        }

        if is_valid_string_clostridium_beijerinckii_node_name(node_name) {
            return Ok("STRING Clostridium beijerinckii");
        }

        if is_valid_string_borrelia_garinii_node_name(node_name) {
            return Ok("STRING Borrelia garinii");
        }

        if is_valid_string_prosthecochloris_aestuarii_node_name(node_name) {
            return Ok("STRING Prosthecochloris aestuarii");
        }

        if is_valid_string_gluconobacter_oxydans_621h_node_name(node_name) {
            return Ok("STRING Gluconobacter oxydans 621H");
        }

        if is_valid_string_photorhabdus_asymbiotica_node_name(node_name) {
            return Ok("STRING Photorhabdus asymbiotica");
        }

        if is_valid_string_blochmannia_pennsylvanicus_node_name(node_name) {
            return Ok("STRING Blochmannia pennsylvanicus");
        }

        if is_valid_string_erythrobacter_vulgaris_node_name(node_name) {
            return Ok("STRING Erythrobacter vulgaris");
        }

        if is_valid_string_ruegeria_sp_tm1040_node_name(node_name) {
            return Ok("STRING Ruegeria sp. TM1040");
        }

        if is_valid_string_thiobacillus_denitrificans_atcc25259_node_name(node_name) {
            return Ok("STRING Thiobacillus denitrificans ATCC25259");
        }

        if is_valid_string_symbiobacterium_thermophilum_node_name(node_name) {
            return Ok("STRING Symbiobacterium thermophilum");
        }

        if is_valid_string_cyanobacterium_stanieri_node_name(node_name) {
            return Ok("STRING Cyanobacterium stanieri");
        }

        if is_valid_string_cyanobium_gracile_node_name(node_name) {
            return Ok("STRING Cyanobium gracile");
        }

        if is_valid_string_wolbachia_sp_bma_node_name(node_name) {
            return Ok("STRING Wolbachia sp. Bma");
        }

        if is_valid_string_cyphellophora_europaea_node_name(node_name) {
            return Ok("STRING Cyphellophora europaea");
        }

        if is_valid_string_rickettsia_canadensis_node_name(node_name) {
            return Ok("STRING Rickettsia canadensis");
        }

        if is_valid_string_rickettsia_akari_node_name(node_name) {
            return Ok("STRING Rickettsia akari");
        }

        if is_valid_string_alkaliphilus_metalliredigens_node_name(node_name) {
            return Ok("STRING Alkaliphilus metalliredigens");
        }

        if is_valid_string_mycoplasma_hyopneumoniae_232_node_name(node_name) {
            return Ok("STRING Mycoplasma hyopneumoniae 232");
        }

        if is_valid_string_micromonas_sp_rcc299_node_name(node_name) {
            return Ok("STRING Micromonas sp. RCC299");
        }

        if is_valid_string_polaromonas_sp_js666_node_name(node_name) {
            return Ok("STRING Polaromonas sp. JS666");
        }

        if is_valid_string_legionella_pneumophila_paris_node_name(node_name) {
            return Ok("STRING Legionella pneumophila Paris");
        }

        if is_valid_string_photobacterium_profundum_node_name(node_name) {
            return Ok("STRING Photobacterium profundum");
        }

        if is_valid_string_frankia_sp_ean1pec_node_name(node_name) {
            return Ok("STRING Frankia sp. EAN1pec");
        }

        if is_valid_string_frankia_sp_eui1c_node_name(node_name) {
            return Ok("STRING Frankia sp. EuI1c");
        }

        if is_valid_string_frankia_sp_cn3_node_name(node_name) {
            return Ok("STRING Frankia sp. CN3");
        }

        if is_valid_string_thermus_thermophilus_hb8_node_name(node_name) {
            return Ok("STRING Thermus thermophilus HB8");
        }

        if is_valid_string_pundamilia_nyererei_node_name(node_name) {
            return Ok("STRING Pundamilia nyererei");
        }

        if is_valid_string_methanocella_paludicola_node_name(node_name) {
            return Ok("STRING Methanocella paludicola");
        }

        if is_valid_string_thauera_sp_27_node_name(node_name) {
            return Ok("STRING Thauera sp. 27");
        }

        if is_valid_string_endozoicomonas_elysicola_node_name(node_name) {
            return Ok("STRING Endozoicomonas elysicola");
        }

        if is_valid_string_campylobacter_lari_node_name(node_name) {
            return Ok("STRING Campylobacter lari");
        }

        if is_valid_string_campylobacter_upsaliensis_node_name(node_name) {
            return Ok("STRING Campylobacter upsaliensis");
        }

        if is_valid_string_fischerella_muscicola_node_name(node_name) {
            return Ok("STRING Fischerella muscicola");
        }

        if is_valid_string_corynebacterium_jeikeium_node_name(node_name) {
            return Ok("STRING Corynebacterium jeikeium");
        }

        if is_valid_string_chryseobacterium_vrystaatense_node_name(node_name) {
            return Ok("STRING Chryseobacterium vrystaatense");
        }

        if is_valid_string_coprothermobacter_proteolyticus_node_name(node_name) {
            return Ok("STRING Coprothermobacter proteolyticus");
        }

        if is_valid_string_dictyoglomus_thermophilum_node_name(node_name) {
            return Ok("STRING Dictyoglomus thermophilum");
        }

        if is_valid_string_haloferax_volcanii_node_name(node_name) {
            return Ok("STRING Haloferax volcanii");
        }

        if is_valid_string_thermomicrobium_roseum_node_name(node_name) {
            return Ok("STRING Thermomicrobium roseum");
        }

        if is_valid_string_thermotoga_neapolitana_node_name(node_name) {
            return Ok("STRING Thermotoga neapolitana");
        }

        if is_valid_string_salinibacter_ruber_node_name(node_name) {
            return Ok("STRING Salinibacter ruber");
        }

        if is_valid_string_neofusicoccum_parvum_node_name(node_name) {
            return Ok("STRING Neofusicoccum parvum");
        }

        if is_valid_string_agrobacterium_vitis_node_name(node_name) {
            return Ok("STRING Agrobacterium vitis");
        }

        if is_valid_string_agrobacterium_radiobacter_node_name(node_name) {
            return Ok("STRING Agrobacterium radiobacter");
        }

        if is_valid_string_dehalococcoides_mccartyi_vs_node_name(node_name) {
            return Ok("STRING Dehalococcoides mccartyi VS");
        }

        if is_valid_string_polynucleobacter_necessarius_asymbioticus_node_name(node_name) {
            return Ok("STRING Polynucleobacter necessarius asymbioticus");
        }

        if is_valid_string_actinobacterium_phsc20c1_node_name(node_name) {
            return Ok("STRING actinobacterium PHSC20C1");
        }

        if is_valid_string_vibrio_fischeri_node_name(node_name) {
            return Ok("STRING Vibrio fischeri");
        }

        if is_valid_string_janibacter_sp_htcc2649_node_name(node_name) {
            return Ok("STRING Janibacter sp. HTCC2649");
        }

        if is_valid_string_dokdonia_sp_med134_node_name(node_name) {
            return Ok("STRING Dokdonia sp. MED134");
        }

        if is_valid_string_polaribacter_irgensii_node_name(node_name) {
            return Ok("STRING Polaribacter irgensii");
        }

        if is_valid_string_psychroflexus_torquis_node_name(node_name) {
            return Ok("STRING Psychroflexus torquis");
        }

        if is_valid_string_robiginitalea_biformata_node_name(node_name) {
            return Ok("STRING Robiginitalea biformata");
        }

        if is_valid_string_polaribacter_sp_med152_node_name(node_name) {
            return Ok("STRING Polaribacter sp. MED152");
        }

        if is_valid_string_maribacter_sp_htcc2170_node_name(node_name) {
            return Ok("STRING Maribacter sp. HTCC2170");
        }

        if is_valid_string_microscilla_marina_node_name(node_name) {
            return Ok("STRING Microscilla marina");
        }

        if is_valid_string_lyngbya_sp_pcc8106_node_name(node_name) {
            return Ok("STRING Lyngbya sp. PCC8106");
        }

        if is_valid_string_nodularia_spumigena_node_name(node_name) {
            return Ok("STRING Nodularia spumigena");
        }

        if is_valid_string_synechococcus_sp_bl107_node_name(node_name) {
            return Ok("STRING Synechococcus sp. BL107");
        }

        if is_valid_string_lentisphaera_araneosa_node_name(node_name) {
            return Ok("STRING Lentisphaera araneosa");
        }

        if is_valid_string_erythrobacter_litoralis_node_name(node_name) {
            return Ok("STRING Erythrobacter litoralis");
        }

        if is_valid_string_blastopirellula_marina_node_name(node_name) {
            return Ok("STRING Blastopirellula marina");
        }

        if is_valid_string_fulvimarina_pelagi_node_name(node_name) {
            return Ok("STRING Fulvimarina pelagi");
        }

        if is_valid_string_loktanella_vestfoldensis_ska53_node_name(node_name) {
            return Ok("STRING Loktanella vestfoldensis SKA53");
        }

        if is_valid_string_oceanicaulis_sp_htcc2633_node_name(node_name) {
            return Ok("STRING Oceanicaulis sp. HTCC2633");
        }

        if is_valid_string_oceanicola_granulosus_node_name(node_name) {
            return Ok("STRING Oceanicola granulosus");
        }

        if is_valid_string_parvularcula_bermudensis_node_name(node_name) {
            return Ok("STRING Parvularcula bermudensis");
        }

        if is_valid_string_roseobacter_sp_med193_node_name(node_name) {
            return Ok("STRING Roseobacter sp. MED193");
        }

        if is_valid_string_roseovarius_sp_217_node_name(node_name) {
            return Ok("STRING Roseovarius sp. 217");
        }

        if is_valid_string_pelagibaca_bermudensis_node_name(node_name) {
            return Ok("STRING Pelagibaca bermudensis");
        }

        if is_valid_string_sphingomonas_sp_ska58_node_name(node_name) {
            return Ok("STRING Sphingomonas sp. SKA58");
        }

        if is_valid_string_rhodobacteraceae_bacterium_htcc2083_node_name(node_name) {
            return Ok("STRING Rhodobacteraceae bacterium HTCC2083");
        }

        if is_valid_string_maritimibacter_alkaliphilus_node_name(node_name) {
            return Ok("STRING Maritimibacter alkaliphilus");
        }

        if is_valid_string_alteromonas_macleodii_deep_node_name(node_name) {
            return Ok("STRING Alteromonas macleodii Deep");
        }

        if is_valid_string_nitrococcus_mobilis_node_name(node_name) {
            return Ok("STRING Nitrococcus mobilis");
        }

        if is_valid_string_psychromonas_sp_cnpt3_node_name(node_name) {
            return Ok("STRING Psychromonas sp. CNPT3");
        }

        if is_valid_string_congregibacter_litoralis_node_name(node_name) {
            return Ok("STRING Congregibacter litoralis");
        }

        if is_valid_string_gammaproteobacteria_bacterium_htcc2207_node_name(node_name) {
            return Ok("STRING Gammaproteobacteria bacterium HTCC2207");
        }

        if is_valid_string_photobacterium_angustum_node_name(node_name) {
            return Ok("STRING Photobacterium angustum");
        }

        if is_valid_string_lactobacillus_sakei_node_name(node_name) {
            return Ok("STRING Lactobacillus sakei");
        }

        if is_valid_string_mariprofundus_ferrooxydans_node_name(node_name) {
            return Ok("STRING Mariprofundus ferrooxydans");
        }

        if is_valid_string_beta_proteobacterium_kb13_node_name(node_name) {
            return Ok("STRING beta proteobacterium KB13");
        }

        if is_valid_string_borrelia_hermsii_node_name(node_name) {
            return Ok("STRING Borrelia hermsii");
        }

        if is_valid_string_borrelia_turicatae_node_name(node_name) {
            return Ok("STRING Borrelia turicatae");
        }

        if is_valid_string_rickettsia_felis_node_name(node_name) {
            return Ok("STRING Rickettsia felis");
        }

        if is_valid_string_bacillus_weihenstephanensis_node_name(node_name) {
            return Ok("STRING Bacillus weihenstephanensis");
        }

        if is_valid_string_bacillus_cytotoxicus_node_name(node_name) {
            return Ok("STRING Bacillus cytotoxicus");
        }

        if is_valid_string_bacillus_pumilus_node_name(node_name) {
            return Ok("STRING Bacillus pumilus");
        }

        if is_valid_string_rhodopseudomonas_palustris_bisa53_node_name(node_name) {
            return Ok("STRING Rhodopseudomonas palustris BisA53");
        }

        if is_valid_string_rhodopseudomonas_palustris_bisb18_node_name(node_name) {
            return Ok("STRING Rhodopseudomonas palustris BisB18");
        }

        if is_valid_string_rhodopseudomonas_palustris_bisb5_node_name(node_name) {
            return Ok("STRING Rhodopseudomonas palustris BisB5");
        }

        if is_valid_string_rhodopseudomonas_palustris_haa2_node_name(node_name) {
            return Ok("STRING Rhodopseudomonas palustris HaA2");
        }

        if is_valid_string_geobacter_daltonii_node_name(node_name) {
            return Ok("STRING Geobacter daltonii");
        }

        if is_valid_string_xanthomonas_campestris_vesicatoria_node_name(node_name) {
            return Ok("STRING Xanthomonas campestris vesicatoria");
        }

        if is_valid_string_herpetosiphon_aurantiacus_node_name(node_name) {
            return Ok("STRING Herpetosiphon aurantiacus");
        }

        if is_valid_string_aliivibrio_salmonicida_node_name(node_name) {
            return Ok("STRING Aliivibrio salmonicida");
        }

        if is_valid_string_synechococcus_sp_rcc307_node_name(node_name) {
            return Ok("STRING Synechococcus sp. RCC307");
        }

        if is_valid_string_synechococcus_sp_cc9902_node_name(node_name) {
            return Ok("STRING Synechococcus sp. CC9902");
        }

        if is_valid_string_escherichia_coli_k12_w3110_node_name(node_name) {
            return Ok("STRING Escherichia coli K12 W3110");
        }

        if is_valid_string_xanthomonas_axonopodis_phaseoli_node_name(node_name) {
            return Ok("STRING Xanthomonas axonopodis phaseoli");
        }

        if is_valid_string_thiomicrospira_crunogena_node_name(node_name) {
            return Ok("STRING Thiomicrospira crunogena");
        }

        if is_valid_string_prochlorothrix_hollandica_node_name(node_name) {
            return Ok("STRING Prochlorothrix hollandica");
        }

        if is_valid_string_sphingopyxis_alaskensis_node_name(node_name) {
            return Ok("STRING Sphingopyxis alaskensis");
        }

        if is_valid_string_nostoc_sp_pcc7107_node_name(node_name) {
            return Ok("STRING Nostoc sp. PCC7107");
        }

        if is_valid_string_shewanella_denitrificans_node_name(node_name) {
            return Ok("STRING Shewanella denitrificans");
        }

        if is_valid_string_shewanella_frigidimarina_node_name(node_name) {
            return Ok("STRING Shewanella frigidimarina");
        }

        if is_valid_string_mycobacterium_rufum_node_name(node_name) {
            return Ok("STRING Mycobacterium rufum");
        }

        if is_valid_string_clostridium_sulfidigenes_node_name(node_name) {
            return Ok("STRING Clostridium sulfidigenes");
        }

        if is_valid_string_paracoccus_denitrificans_node_name(node_name) {
            return Ok("STRING Paracoccus denitrificans");
        }

        if is_valid_string_magnaporthe_oryzae_node_name(node_name) {
            return Ok("STRING Magnaporthe oryzae");
        }

        if is_valid_string_bradyrhizobium_sp_wsm1743_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. WSM1743");
        }

        if is_valid_string_bradyrhizobium_sp_wsm1253_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. WSM1253");
        }

        if is_valid_string_shewanella_putrefaciens_node_name(node_name) {
            return Ok("STRING Shewanella putrefaciens");
        }

        if is_valid_string_chlorobium_luteolum_node_name(node_name) {
            return Ok("STRING Chlorobium luteolum");
        }

        if is_valid_string_nonlabens_sediminis_node_name(node_name) {
            return Ok("STRING Nonlabens sediminis");
        }

        if is_valid_string_deinococcus_geothermalis_node_name(node_name) {
            return Ok("STRING Deinococcus geothermalis");
        }

        if is_valid_string_anaplasma_marginale_florida_node_name(node_name) {
            return Ok("STRING Anaplasma marginale Florida");
        }

        if is_valid_string_synechococcus_sp_ja33ab_node_name(node_name) {
            return Ok("STRING Synechococcus sp. JA33Ab");
        }

        if is_valid_string_synechococcus_sp_ja23ba_node_name(node_name) {
            return Ok("STRING Synechococcus sp. JA23Ba");
        }

        if is_valid_string_pseudomonas_simiae_node_name(node_name) {
            return Ok("STRING Pseudomonas simiae");
        }

        if is_valid_string_brevibacterium_linens_node_name(node_name) {
            return Ok("STRING Brevibacterium linens");
        }

        if is_valid_string_phytoplasma_wheat_blue_node_name(node_name) {
            return Ok("STRING Phytoplasma wheat blue");
        }

        if is_valid_string_phytoplasma_aster_yellows_node_name(node_name) {
            return Ok("STRING Phytoplasma aster yellows");
        }

        if is_valid_string_streptococcus_thermophilus_lmd9_node_name(node_name) {
            return Ok("STRING Streptococcus thermophilus LMD9");
        }

        if is_valid_string_azotobacter_vinelandii_node_name(node_name) {
            return Ok("STRING Azotobacter vinelandii");
        }

        if is_valid_string_nitrobacter_hamburgensis_node_name(node_name) {
            return Ok("STRING Nitrobacter hamburgensis");
        }

        if is_valid_string_nitrobacter_winogradskyi_node_name(node_name) {
            return Ok("STRING Nitrobacter winogradskyi");
        }

        if is_valid_string_methanospirillum_hungatei_node_name(node_name) {
            return Ok("STRING Methanospirillum hungatei");
        }

        if is_valid_string_nitrosococcus_oceani_node_name(node_name) {
            return Ok("STRING Nitrosococcus oceani");
        }

        if is_valid_string_nitrosospira_multiformis_node_name(node_name) {
            return Ok("STRING Nitrosospira multiformis");
        }

        if is_valid_string_shewanella_loihica_node_name(node_name) {
            return Ok("STRING Shewanella loihica");
        }

        if is_valid_string_paenibacillus_sp_jdr2_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. JDR2");
        }

        if is_valid_string_chloroflexus_aurantiacus_node_name(node_name) {
            return Ok("STRING Chloroflexus aurantiacus");
        }

        if is_valid_string_lactobacillus_gasseri_node_name(node_name) {
            return Ok("STRING Lactobacillus gasseri");
        }

        if is_valid_string_pelodictyon_phaeoclathratiforme_node_name(node_name) {
            return Ok("STRING Pelodictyon phaeoclathratiforme");
        }

        if is_valid_string_phytophthora_kernoviae_node_name(node_name) {
            return Ok("STRING Phytophthora kernoviae");
        }

        if is_valid_string_xanthomonas_axonopodis_vasculorum_node_name(node_name) {
            return Ok("STRING Xanthomonas axonopodis vasculorum");
        }

        if is_valid_string_shewanella_amazonensis_node_name(node_name) {
            return Ok("STRING Shewanella amazonensis");
        }

        if is_valid_string_sulfurimonas_denitrificans_node_name(node_name) {
            return Ok("STRING Sulfurimonas denitrificans");
        }

        if is_valid_string_bacillus_amyloliquefaciens_node_name(node_name) {
            return Ok("STRING Bacillus amyloliquefaciens");
        }

        if is_valid_string_frankia_alni_node_name(node_name) {
            return Ok("STRING Frankia alni");
        }

        if is_valid_string_lactobacillus_helveticus_cnrz32_node_name(node_name) {
            return Ok("STRING Lactobacillus helveticus CNRZ32");
        }

        if is_valid_string_bifidobacterium_breve_node_name(node_name) {
            return Ok("STRING Bifidobacterium breve");
        }

        if is_valid_string_chloroflexus_aggregans_node_name(node_name) {
            return Ok("STRING Chloroflexus aggregans");
        }

        if is_valid_string_pseudoalteromonas_haloplanktis_tac125_node_name(node_name) {
            return Ok("STRING Pseudoalteromonas haloplanktis TAC125");
        }

        if is_valid_string_pseudozyma_hubeiensis_node_name(node_name) {
            return Ok("STRING Pseudozyma hubeiensis");
        }

        if is_valid_string_bifidobacterium_crudilactis_node_name(node_name) {
            return Ok("STRING Bifidobacterium crudilactis");
        }

        if is_valid_string_acaryochloris_marina_node_name(node_name) {
            return Ok("STRING Acaryochloris marina");
        }

        if is_valid_string_amycolatopsis_jejuensis_node_name(node_name) {
            return Ok("STRING Amycolatopsis jejuensis");
        }

        if is_valid_string_nitrospira_defluvii_node_name(node_name) {
            return Ok("STRING Nitrospira defluvii");
        }

        if is_valid_string_sulfolobus_acidocaldarius_node_name(node_name) {
            return Ok("STRING Sulfolobus acidocaldarius");
        }

        if is_valid_string_blattabacterium_sp_bge_node_name(node_name) {
            return Ok("STRING Blattabacterium sp. Bge");
        }

        if is_valid_string_simkania_negevensis_node_name(node_name) {
            return Ok("STRING Simkania negevensis");
        }

        if is_valid_string_chlamydophila_pecorum_node_name(node_name) {
            return Ok("STRING Chlamydophila pecorum");
        }

        if is_valid_string_chlamydia_psittaci_6bc_node_name(node_name) {
            return Ok("STRING Chlamydia psittaci 6BC");
        }

        if is_valid_string_chlorobium_phaeobacteroides_bs1_node_name(node_name) {
            return Ok("STRING Chlorobium phaeobacteroides BS1");
        }

        if is_valid_string_alpha_proteobacterium_bal199_node_name(node_name) {
            return Ok("STRING alpha proteobacterium BAL199");
        }

        if is_valid_string_clostridium_drakei_node_name(node_name) {
            return Ok("STRING Clostridium drakei");
        }

        if is_valid_string_bacillus_okhensis_node_name(node_name) {
            return Ok("STRING Bacillus okhensis");
        }

        if is_valid_string_lactobacillus_fermentum_ifo3956_node_name(node_name) {
            return Ok("STRING Lactobacillus fermentum IFO3956");
        }

        if is_valid_string_finegoldia_magna_atcc29328_node_name(node_name) {
            return Ok("STRING Finegoldia magna ATCC29328");
        }

        if is_valid_string_rickettsia_tamurae_node_name(node_name) {
            return Ok("STRING Rickettsia tamurae");
        }

        if is_valid_string_nitrosomonas_eutropha_node_name(node_name) {
            return Ok("STRING Nitrosomonas eutropha");
        }

        if is_valid_string_psychrobacter_cryohalolentis_node_name(node_name) {
            return Ok("STRING Psychrobacter cryohalolentis");
        }

        if is_valid_string_syntrophomonas_wolfei_node_name(node_name) {
            return Ok("STRING Syntrophomonas wolfei");
        }

        if is_valid_string_syntrophobacter_fumaroxidans_node_name(node_name) {
            return Ok("STRING Syntrophobacter fumaroxidans");
        }

        if is_valid_string_bradyrhizobium_sp_s23321_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. S23321");
        }

        if is_valid_string_pelagibacter_ubique_htcc1062_node_name(node_name) {
            return Ok("STRING Pelagibacter ubique HTCC1062");
        }

        if is_valid_string_rickettsia_bellii_rml369c_node_name(node_name) {
            return Ok("STRING Rickettsia bellii RML369C");
        }

        if is_valid_string_pyronema_omphalodes_node_name(node_name) {
            return Ok("STRING Pyronema omphalodes");
        }

        if is_valid_string_gordonia_sp_ktr9_node_name(node_name) {
            return Ok("STRING Gordonia sp. KTR9");
        }

        if is_valid_string_pelobacter_carbinolicus_node_name(node_name) {
            return Ok("STRING Pelobacter carbinolicus");
        }

        if is_valid_string_pelobacter_propionicus_node_name(node_name) {
            return Ok("STRING Pelobacter propionicus");
        }

        if is_valid_string_rhodoferax_ferrireducens_node_name(node_name) {
            return Ok("STRING Rhodoferax ferrireducens");
        }

        if is_valid_string_burkholderia_ambifaria_node_name(node_name) {
            return Ok("STRING Burkholderia ambifaria");
        }

        if is_valid_string_actinobacillus_succinogenes_node_name(node_name) {
            return Ok("STRING Actinobacillus succinogenes");
        }

        if is_valid_string_methanosphaera_stadtmanae_node_name(node_name) {
            return Ok("STRING Methanosphaera stadtmanae");
        }

        if is_valid_string_thermoanaerobacter_pseudethanolicus_node_name(node_name) {
            return Ok("STRING Thermoanaerobacter pseudethanolicus");
        }

        if is_valid_string_spathaspora_passalidarum_node_name(node_name) {
            return Ok("STRING Spathaspora passalidarum");
        }

        if is_valid_string_chlorobium_chlorochromatii_node_name(node_name) {
            return Ok("STRING Chlorobium chlorochromatii");
        }

        if is_valid_string_burkholderia_oklahomensis_node_name(node_name) {
            return Ok("STRING Burkholderia oklahomensis");
        }

        if is_valid_string_staphylococcus_saprophyticus_node_name(node_name) {
            return Ok("STRING Staphylococcus saprophyticus");
        }

        if is_valid_string_pseudoalteromonas_atlantica_node_name(node_name) {
            return Ok("STRING Pseudoalteromonas atlantica");
        }

        if is_valid_string_pyrococcus_sp_na2_node_name(node_name) {
            return Ok("STRING Pyrococcus sp. NA2");
        }

        if is_valid_string_sodalis_glossinidius_node_name(node_name) {
            return Ok("STRING Sodalis glossinidius");
        }

        if is_valid_string_planctomyces_maris_node_name(node_name) {
            return Ok("STRING Planctomyces maris");
        }

        if is_valid_string_vibrio_cholerae_o395_node_name(node_name) {
            return Ok("STRING Vibrio cholerae O395");
        }

        if is_valid_string_bacillus_coagulans_36d1_node_name(node_name) {
            return Ok("STRING Bacillus coagulans 36D1");
        }

        if is_valid_string_kutzneria_sp_744_node_name(node_name) {
            return Ok("STRING Kutzneria sp. 744");
        }

        if is_valid_string_mycoplasma_hominis_node_name(node_name) {
            return Ok("STRING Mycoplasma hominis");
        }

        if is_valid_string_rhizobium_etli_cfn42_node_name(node_name) {
            return Ok("STRING Rhizobium etli CFN42");
        }

        if is_valid_string_natronomonas_pharaonis_node_name(node_name) {
            return Ok("STRING Natronomonas pharaonis");
        }

        if is_valid_string_rhizobium_sp_lpu83_node_name(node_name) {
            return Ok("STRING Rhizobium sp. LPU83");
        }

        if is_valid_string_rhodobacter_sphaeroides_atcc17025_node_name(node_name) {
            return Ok("STRING Rhodobacter sphaeroides ATCC17025");
        }

        if is_valid_string_psychrobacter_sp_prwf1_node_name(node_name) {
            return Ok("STRING Psychrobacter sp. PRwf1");
        }

        if is_valid_string_lactobacillus_reuteri_10023_node_name(node_name) {
            return Ok("STRING Lactobacillus reuteri 10023");
        }

        if is_valid_string_halorhodospira_halophila_node_name(node_name) {
            return Ok("STRING Halorhodospira halophila");
        }

        if is_valid_string_desulfotomaculum_reducens_node_name(node_name) {
            return Ok("STRING Desulfotomaculum reducens");
        }

        if is_valid_string_acidiphilium_cryptum_node_name(node_name) {
            return Ok("STRING Acidiphilium cryptum");
        }

        if is_valid_string_leuconostoc_citreum_node_name(node_name) {
            return Ok("STRING Leuconostoc citreum");
        }

        if is_valid_string_paenibacillus_polymyxa_e681_node_name(node_name) {
            return Ok("STRING Paenibacillus polymyxa E681");
        }

        if is_valid_string_hahella_chejuensis_node_name(node_name) {
            return Ok("STRING Hahella chejuensis");
        }

        if is_valid_string_akkermansia_muciniphila_node_name(node_name) {
            return Ok("STRING Akkermansia muciniphila");
        }

        if is_valid_string_yersinia_intermedia_node_name(node_name) {
            return Ok("STRING Yersinia intermedia");
        }

        if is_valid_string_yersinia_frederiksenii_node_name(node_name) {
            return Ok("STRING Yersinia frederiksenii");
        }

        if is_valid_string_mycobacterium_gilvum_node_name(node_name) {
            return Ok("STRING Mycobacterium gilvum");
        }

        if is_valid_string_mycobacterium_vanbaalenii_node_name(node_name) {
            return Ok("STRING Mycobacterium vanbaalenii");
        }

        if is_valid_string_alkaliphilus_oremlandii_node_name(node_name) {
            return Ok("STRING Alkaliphilus oremlandii");
        }

        if is_valid_string_roseobacter_sp_azwk3b_node_name(node_name) {
            return Ok("STRING Roseobacter sp. AzwK3b");
        }

        if is_valid_string_methanocella_arvoryzae_node_name(node_name) {
            return Ok("STRING Methanocella arvoryzae");
        }

        if is_valid_string_marinobacter_hydrocarbonoclasticus_node_name(node_name) {
            return Ok("STRING Marinobacter hydrocarbonoclasticus");
        }

        if is_valid_string_acidothermus_cellulolyticus_node_name(node_name) {
            return Ok("STRING Acidothermus cellulolyticus");
        }

        if is_valid_string_caldicellulosiruptor_saccharolyticus_node_name(node_name) {
            return Ok("STRING Caldicellulosiruptor saccharolyticus");
        }

        if is_valid_string_pseudomonas_putida_f1_node_name(node_name) {
            return Ok("STRING Pseudomonas putida F1");
        }

        if is_valid_string_pyramidobacter_piscolens_node_name(node_name) {
            return Ok("STRING Pyramidobacter piscolens");
        }

        if is_valid_string_lactobacillus_delbrueckii_2038_node_name(node_name) {
            return Ok("STRING Lactobacillus delbrueckii 2038");
        }

        if is_valid_string_campylobacter_jejuni_81176_node_name(node_name) {
            return Ok("STRING Campylobacter jejuni 81176");
        }

        if is_valid_string_bifidobacterium_tsurumiense_node_name(node_name) {
            return Ok("STRING Bifidobacterium tsurumiense");
        }

        if is_valid_string_micromonospora_chokoriensis_node_name(node_name) {
            return Ok("STRING Micromonospora chokoriensis");
        }

        if is_valid_string_orientia_tsutsugamushi_node_name(node_name) {
            return Ok("STRING Orientia tsutsugamushi");
        }

        if is_valid_string_bacteroides_dorei_node_name(node_name) {
            return Ok("STRING Bacteroides dorei");
        }

        if is_valid_string_psychromonas_ingrahamii_node_name(node_name) {
            return Ok("STRING Psychromonas ingrahamii");
        }

        if is_valid_string_roseiflexus_sp_rs1_node_name(node_name) {
            return Ok("STRING Roseiflexus sp. RS1");
        }

        if is_valid_string_lachnoclostridium_phytofermentans_node_name(node_name) {
            return Ok("STRING Lachnoclostridium phytofermentans");
        }

        if is_valid_string_acidovorax_sp_kks102_node_name(node_name) {
            return Ok("STRING Acidovorax sp. KKS102");
        }

        if is_valid_string_halobiforma_lacisalsi_node_name(node_name) {
            return Ok("STRING Halobiforma lacisalsi");
        }

        if is_valid_string_brevibacillus_brevis_nbrc100599_node_name(node_name) {
            return Ok("STRING Brevibacillus brevis NBRC100599");
        }

        if is_valid_string_streptomyces_olindensis_node_name(node_name) {
            return Ok("STRING Streptomyces olindensis");
        }

        if is_valid_string_xanthomonas_oryzae_pxo99a_node_name(node_name) {
            return Ok("STRING Xanthomonas oryzae PXO99A");
        }

        if is_valid_string_bartonella_bacilliformis_node_name(node_name) {
            return Ok("STRING Bartonella bacilliformis");
        }

        if is_valid_string_campylobacter_concisus_13826_node_name(node_name) {
            return Ok("STRING Campylobacter concisus 13826");
        }

        if is_valid_string_campylobacter_fetus_8240_node_name(node_name) {
            return Ok("STRING Campylobacter fetus 8240");
        }

        if is_valid_string_campylobacter_hominis_node_name(node_name) {
            return Ok("STRING Campylobacter hominis");
        }

        if is_valid_string_bordetella_avium_node_name(node_name) {
            return Ok("STRING Bordetella avium");
        }

        if is_valid_string_exiguobacterium_sp_at1b_node_name(node_name) {
            return Ok("STRING Exiguobacterium sp. AT1b");
        }

        if is_valid_string_mycobacterium_ulcerans_node_name(node_name) {
            return Ok("STRING Mycobacterium ulcerans");
        }

        if is_valid_string_flavobacterium_reichenbachii_node_name(node_name) {
            return Ok("STRING Flavobacterium reichenbachii");
        }

        if is_valid_string_escherichia_coli_536_node_name(node_name) {
            return Ok("STRING Escherichia coli 536");
        }

        if is_valid_string_haloquadratum_walsbyi_dsm16790_node_name(node_name) {
            return Ok("STRING Haloquadratum walsbyi DSM16790");
        }

        if is_valid_string_lawsonia_intracellularis_node_name(node_name) {
            return Ok("STRING Lawsonia intracellularis");
        }

        if is_valid_string_endocarpon_pusillum_node_name(node_name) {
            return Ok("STRING Endocarpon pusillum");
        }

        if is_valid_string_polaromonas_naphthalenivorans_node_name(node_name) {
            return Ok("STRING Polaromonas naphthalenivorans");
        }

        if is_valid_string_ramlibacter_tataouinensis_node_name(node_name) {
            return Ok("STRING Ramlibacter tataouinensis");
        }

        if is_valid_string_frankia_sp_bcu110501_node_name(node_name) {
            return Ok("STRING Frankia sp. BCU110501");
        }

        if is_valid_string_streptococcus_mitis_b6_node_name(node_name) {
            return Ok("STRING Streptococcus mitis B6");
        }

        if is_valid_string_sinorhizobium_medicae_node_name(node_name) {
            return Ok("STRING Sinorhizobium medicae");
        }

        if is_valid_string_caulobacter_sp_k31_node_name(node_name) {
            return Ok("STRING Caulobacter sp. K31");
        }

        if is_valid_string_xanthomonas_fuscans_node_name(node_name) {
            return Ok("STRING Xanthomonas fuscans");
        }

        if is_valid_string_phycicoccus_jejuensis_node_name(node_name) {
            return Ok("STRING Phycicoccus jejuensis");
        }

        if is_valid_string_rhodobacterales_bacterium_htcc2255_node_name(node_name) {
            return Ok("STRING Rhodobacterales bacterium HTCC2255");
        }

        if is_valid_string_arcobacter_butzleri_rm4018_node_name(node_name) {
            return Ok("STRING Arcobacter butzleri RM4018");
        }

        if is_valid_string_methanoculleus_marisnigri_node_name(node_name) {
            return Ok("STRING Methanoculleus marisnigri");
        }

        if is_valid_string_thermofilum_pendens_node_name(node_name) {
            return Ok("STRING Thermofilum pendens");
        }

        if is_valid_string_salinispora_tropica_node_name(node_name) {
            return Ok("STRING Salinispora tropica");
        }

        if is_valid_string_pelotomaculum_thermopropionicum_node_name(node_name) {
            return Ok("STRING Pelotomaculum thermopropionicum");
        }

        if is_valid_string_erwinia_typographi_node_name(node_name) {
            return Ok("STRING Erwinia typographi");
        }

        if is_valid_string_rhodobacter_sp_sw2_node_name(node_name) {
            return Ok("STRING Rhodobacter sp. SW2");
        }

        if is_valid_string_buchnera_aphidicola_bcc_node_name(node_name) {
            return Ok("STRING Buchnera aphidicola BCc");
        }

        if is_valid_string_streptococcus_pneumoniae_d39_node_name(node_name) {
            return Ok("STRING Streptococcus pneumoniae D39");
        }

        if is_valid_string_halothermothrix_orenii_node_name(node_name) {
            return Ok("STRING Halothermothrix orenii");
        }

        if is_valid_string_rivularia_sp_pcc7116_node_name(node_name) {
            return Ok("STRING Rivularia sp. PCC7116");
        }

        if is_valid_string_korarchaeum_cryptofilum_node_name(node_name) {
            return Ok("STRING Korarchaeum cryptofilum");
        }

        if is_valid_string_janthinobacterium_sp_marseille_node_name(node_name) {
            return Ok("STRING Janthinobacterium sp. Marseille");
        }

        if is_valid_string_roseobacter_denitrificans_node_name(node_name) {
            return Ok("STRING Roseobacter denitrificans");
        }

        if is_valid_string_francisella_tularensis_holarctica_node_name(node_name) {
            return Ok("STRING Francisella tularensis holarctica");
        }

        if is_valid_string_flavobacterium_johnsoniae_node_name(node_name) {
            return Ok("STRING Flavobacterium johnsoniae");
        }

        if is_valid_string_paracoccus_halophilus_node_name(node_name) {
            return Ok("STRING Paracoccus halophilus");
        }

        if is_valid_string_teredinibacter_turnerae_t7901_node_name(node_name) {
            return Ok("STRING Teredinibacter turnerae T7901");
        }

        if is_valid_string_kocuria_rhizophila_dc2201_node_name(node_name) {
            return Ok("STRING Kocuria rhizophila DC2201");
        }

        if is_valid_string_stigmatella_aurantiaca_node_name(node_name) {
            return Ok("STRING Stigmatella aurantiaca");
        }

        if is_valid_string_gemmatimonas_aurantiaca_node_name(node_name) {
            return Ok("STRING Gemmatimonas aurantiaca");
        }

        if is_valid_string_pseudomonas_stutzeri_a1501_node_name(node_name) {
            return Ok("STRING Pseudomonas stutzeri A1501");
        }

        if is_valid_string_xanthomonas_albilineans_node_name(node_name) {
            return Ok("STRING Xanthomonas albilineans");
        }

        if is_valid_string_acidithiobacillus_ferrooxidans_atcc53993_node_name(node_name) {
            return Ok("STRING Acidithiobacillus ferrooxidans ATCC53993");
        }

        if is_valid_string_aeromonas_hydrophila_atcc7966_node_name(node_name) {
            return Ok("STRING Aeromonas hydrophila ATCC7966");
        }

        if is_valid_string_hydrogenobaculum_sp_y04aas1_node_name(node_name) {
            return Ok("STRING Hydrogenobaculum sp. Y04AAS1");
        }

        if is_valid_string_lachancea_thermotolerans_node_name(node_name) {
            return Ok("STRING Lachancea thermotolerans");
        }

        if is_valid_string_ralstonia_eutropha_h16_node_name(node_name) {
            return Ok("STRING Ralstonia eutropha H16");
        }

        if is_valid_string_fervidobacterium_nodosum_node_name(node_name) {
            return Ok("STRING Fervidobacterium nodosum");
        }

        if is_valid_string_aeromonas_salmonicida_node_name(node_name) {
            return Ok("STRING Aeromonas salmonicida");
        }

        if is_valid_string_verrucomicrobiae_bacterium_dg1235_node_name(node_name) {
            return Ok("STRING Verrucomicrobiae bacterium DG1235");
        }

        if is_valid_string_helicobacter_acinonychis_node_name(node_name) {
            return Ok("STRING Helicobacter acinonychis");
        }

        if is_valid_string_bartonella_tribocorum_node_name(node_name) {
            return Ok("STRING Bartonella tribocorum");
        }

        if is_valid_string_roseiflexus_castenholzii_node_name(node_name) {
            return Ok("STRING Roseiflexus castenholzii");
        }

        if is_valid_string_erythrobacter_sp_jl475_node_name(node_name) {
            return Ok("STRING Erythrobacter sp. JL475");
        }

        if is_valid_string_xanthomonas_oryzae_oryzicola_node_name(node_name) {
            return Ok("STRING Xanthomonas oryzae oryzicola");
        }

        if is_valid_string_pyrobaculum_islandicum_node_name(node_name) {
            return Ok("STRING Pyrobaculum islandicum");
        }

        if is_valid_string_pseudomonas_entomophila_node_name(node_name) {
            return Ok("STRING Pseudomonas entomophila");
        }

        if is_valid_string_labrenzia_aggregata_node_name(node_name) {
            return Ok("STRING Labrenzia aggregata");
        }

        if is_valid_string_thermophagus_xiamenensis_node_name(node_name) {
            return Ok("STRING Thermophagus xiamenensis");
        }

        if is_valid_string_listeria_welshimeri_node_name(node_name) {
            return Ok("STRING Listeria welshimeri");
        }

        if is_valid_string_clostridium_novyi_nt_node_name(node_name) {
            return Ok("STRING Clostridium novyi NT");
        }

        if is_valid_string_methanobacterium_arcticum_node_name(node_name) {
            return Ok("STRING Methanobacterium arcticum");
        }

        if is_valid_string_nitratiruptor_sp_sb1552_node_name(node_name) {
            return Ok("STRING Nitratiruptor sp. SB1552");
        }

        if is_valid_string_sulfurovum_sp_nbc371_node_name(node_name) {
            return Ok("STRING Sulfurovum sp. NBC371");
        }

        if is_valid_string_lactobacillus_brevis_atcc367_node_name(node_name) {
            return Ok("STRING Lactobacillus brevis ATCC367");
        }

        if is_valid_string_archaeoglobus_sulfaticallidus_node_name(node_name) {
            return Ok("STRING Archaeoglobus sulfaticallidus");
        }

        if is_valid_string_cupriavidus_sp_amp6_node_name(node_name) {
            return Ok("STRING Cupriavidus sp. amp6");
        }

        if is_valid_string_sagittula_stellata_node_name(node_name) {
            return Ok("STRING Sagittula stellata");
        }

        if is_valid_string_rhodobacteraceae_bacterium_htcc2150_node_name(node_name) {
            return Ok("STRING Rhodobacteraceae bacterium HTCC2150");
        }

        if is_valid_string_algoriphagus_machipongonensis_node_name(node_name) {
            return Ok("STRING Algoriphagus machipongonensis");
        }

        if is_valid_string_planktothrix_agardhii_node_name(node_name) {
            return Ok("STRING Planktothrix agardhii");
        }

        if is_valid_string_roseobacter_sp_sk20926_node_name(node_name) {
            return Ok("STRING Roseobacter sp. SK20926");
        }

        if is_valid_string_streptococcus_sanguinis_sk36_node_name(node_name) {
            return Ok("STRING Streptococcus sanguinis SK36");
        }

        if is_valid_string_pseudomonas_putida_w619_node_name(node_name) {
            return Ok("STRING Pseudomonas putida W619");
        }

        if is_valid_string_borrelia_afzelii_node_name(node_name) {
            return Ok("STRING Borrelia afzelii");
        }

        if is_valid_string_lactobacillus_delbrueckii_atcc11842_node_name(node_name) {
            return Ok("STRING Lactobacillus delbrueckii ATCC11842");
        }

        if is_valid_string_thermotoga_petrophila_node_name(node_name) {
            return Ok("STRING Thermotoga petrophila");
        }

        if is_valid_string_actinocatenispora_sera_node_name(node_name) {
            return Ok("STRING Actinocatenispora sera");
        }

        if is_valid_string_stenotrophomonas_maltophilia_r5513_node_name(node_name) {
            return Ok("STRING Stenotrophomonas maltophilia R5513");
        }

        if is_valid_string_thermosipho_melanesiensis_node_name(node_name) {
            return Ok("STRING Thermosipho melanesiensis");
        }

        if is_valid_string_ehrlichia_sp_hf_node_name(node_name) {
            return Ok("STRING Ehrlichia sp. HF");
        }

        if is_valid_string_salinispora_arenicola_node_name(node_name) {
            return Ok("STRING Salinispora arenicola");
        }

        if is_valid_string_burkholderia_phymatum_node_name(node_name) {
            return Ok("STRING Burkholderia phymatum");
        }

        if is_valid_string_granulibacter_bethesdensis_node_name(node_name) {
            return Ok("STRING Granulibacter bethesdensis");
        }

        if is_valid_string_streptococcus_suis_05zyh33_node_name(node_name) {
            return Ok("STRING Streptococcus suis 05ZYH33");
        }

        if is_valid_string_kordia_algicida_node_name(node_name) {
            return Ok("STRING Kordia algicida");
        }

        if is_valid_string_roseobacter_sp_gai101_node_name(node_name) {
            return Ok("STRING Roseobacter sp. GAI101");
        }

        if is_valid_string_roseobacter_sp_ccs2_node_name(node_name) {
            return Ok("STRING Roseobacter sp. CCS2");
        }

        if is_valid_string_roseobacter_litoralis_node_name(node_name) {
            return Ok("STRING Roseobacter litoralis");
        }

        if is_valid_string_pedobacter_sp_bal39_node_name(node_name) {
            return Ok("STRING Pedobacter sp. BAL39");
        }

        if is_valid_string_flavobacteria_bacterium_bal38_node_name(node_name) {
            return Ok("STRING Flavobacteria bacterium BAL38");
        }

        if is_valid_string_brevundimonas_sp_bal3_node_name(node_name) {
            return Ok("STRING Brevundimonas sp. BAL3");
        }

        if is_valid_string_flavobacteriales_bacterium_alc1_node_name(node_name) {
            return Ok("STRING Flavobacteriales bacterium ALC1");
        }

        if is_valid_string_cyanothece_sp_ccy0110_node_name(node_name) {
            return Ok("STRING Cyanothece sp. CCY0110");
        }

        if is_valid_string_roseovarius_sp_tm1035_node_name(node_name) {
            return Ok("STRING Roseovarius sp. TM1035");
        }

        if is_valid_string_gammaproteobacteria_bacterium_htcc5015_node_name(node_name) {
            return Ok("STRING Gammaproteobacteria bacterium HTCC5015");
        }

        if is_valid_string_octadecabacter_arcticus_node_name(node_name) {
            return Ok("STRING Octadecabacter arcticus");
        }

        if is_valid_string_phaeobacter_inhibens_node_name(node_name) {
            return Ok("STRING Phaeobacter inhibens");
        }

        if is_valid_string_thermococcus_barophilus_node_name(node_name) {
            return Ok("STRING Thermococcus barophilus");
        }

        if is_valid_string_oceanibulbus_indolifex_node_name(node_name) {
            return Ok("STRING Oceanibulbus indolifex");
        }

        if is_valid_string_plesiocystis_pacifica_node_name(node_name) {
            return Ok("STRING Plesiocystis pacifica");
        }

        if is_valid_string_octadecabacter_antarcticus_node_name(node_name) {
            return Ok("STRING Octadecabacter antarcticus");
        }

        if is_valid_string_verminephrobacter_eiseniae_node_name(node_name) {
            return Ok("STRING Verminephrobacter eiseniae");
        }

        if is_valid_string_rickettsia_bellii_osu85389_node_name(node_name) {
            return Ok("STRING Rickettsia bellii OSU85389");
        }

        if is_valid_string_nitratireductor_pacificus_node_name(node_name) {
            return Ok("STRING Nitratireductor pacificus");
        }

        if is_valid_string_sphingomonas_wittichii_node_name(node_name) {
            return Ok("STRING Sphingomonas wittichii");
        }

        if is_valid_string_shewanella_woodyi_node_name(node_name) {
            return Ok("STRING Shewanella woodyi");
        }

        if is_valid_string_pestalotiopsis_fici_node_name(node_name) {
            return Ok("STRING Pestalotiopsis fici");
        }

        if is_valid_string_yersinia_enterocolitica_8081_node_name(node_name) {
            return Ok("STRING Yersinia enterocolitica 8081");
        }

        if is_valid_string_fusobacterium_nucleatum_polymorphum_node_name(node_name) {
            return Ok("STRING Fusobacterium nucleatum polymorphum");
        }

        if is_valid_string_alcanivorax_borkumensis_node_name(node_name) {
            return Ok("STRING Alcanivorax borkumensis");
        }

        if is_valid_string_porphyromonas_crevioricanis_node_name(node_name) {
            return Ok("STRING Porphyromonas crevioricanis");
        }

        if is_valid_string_maricaulis_maris_node_name(node_name) {
            return Ok("STRING Maricaulis maris");
        }

        if is_valid_string_clostridium_cellulolyticum_node_name(node_name) {
            return Ok("STRING Clostridium cellulolyticum");
        }

        if is_valid_string_burkholderia_multivorans_node_name(node_name) {
            return Ok("STRING Burkholderia multivorans");
        }

        if is_valid_string_rhizobium_leguminosarum_wsm2304_node_name(node_name) {
            return Ok("STRING Rhizobium leguminosarum WSM2304");
        }

        if is_valid_string_beggiatoa_alba_node_name(node_name) {
            return Ok("STRING Beggiatoa alba");
        }

        if is_valid_string_gallionella_capsiferriformans_node_name(node_name) {
            return Ok("STRING Gallionella capsiferriformans");
        }

        if is_valid_string_leptothrix_cholodnii_node_name(node_name) {
            return Ok("STRING Leptothrix cholodnii");
        }

        if is_valid_string_cyanothece_sp_pcc7425_node_name(node_name) {
            return Ok("STRING Cyanothece sp. PCC7425");
        }

        if is_valid_string_beijerinckia_indica_node_name(node_name) {
            return Ok("STRING Beijerinckia indica");
        }

        if is_valid_string_methylocapsa_acidiphila_node_name(node_name) {
            return Ok("STRING Methylocapsa acidiphila");
        }

        if is_valid_string_methylocella_silvestris_node_name(node_name) {
            return Ok("STRING Methylocella silvestris");
        }

        if is_valid_string_brachybacterium_phenoliresistens_node_name(node_name) {
            return Ok("STRING Brachybacterium phenoliresistens");
        }

        if is_valid_string_staphylococcus_carnosus_node_name(node_name) {
            return Ok("STRING Staphylococcus carnosus");
        }

        if is_valid_string_thioalkalivibrio_sulfidiphilus_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sulfidiphilus");
        }

        if is_valid_string_thioalkalivibrio_sp_k90mix_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. K90mix");
        }

        if is_valid_string_marmoricola_aequoreus_node_name(node_name) {
            return Ok("STRING Marmoricola aequoreus");
        }

        if is_valid_string_lachnospiraceae_bacterium_284_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium 284");
        }

        if is_valid_string_lachnospiraceae_bacterium_31_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium 31");
        }

        if is_valid_string_lachnospiraceae_bacterium_a2_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium A2");
        }

        if is_valid_string_lachnospiraceae_bacterium_a4_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium A4");
        }

        if is_valid_string_acidovorax_citrulli_node_name(node_name) {
            return Ok("STRING Acidovorax citrulli");
        }

        if is_valid_string_caldivirga_maquilingensis_node_name(node_name) {
            return Ok("STRING Caldivirga maquilingensis");
        }

        if is_valid_string_bacillus_pseudofirmus_node_name(node_name) {
            return Ok("STRING Bacillus pseudofirmus");
        }

        if is_valid_string_pseudobacteroides_cellulosolvens_node_name(node_name) {
            return Ok("STRING Pseudobacteroides cellulosolvens");
        }

        if is_valid_string_bifidobacterium_bifidum_ncimb41171_node_name(node_name) {
            return Ok("STRING Bifidobacterium bifidum NCIMB41171");
        }

        if is_valid_string_bradyrhizobium_elkanii_usda76_node_name(node_name) {
            return Ok("STRING Bradyrhizobium elkanii USDA76");
        }

        if is_valid_string_burkholderia_phytofirmans_node_name(node_name) {
            return Ok("STRING Burkholderia phytofirmans");
        }

        if is_valid_string_delftia_acidovorans_node_name(node_name) {
            return Ok("STRING Delftia acidovorans");
        }

        if is_valid_string_shewanella_pealeana_node_name(node_name) {
            return Ok("STRING Shewanella pealeana");
        }

        if is_valid_string_dinoroseobacter_shibae_node_name(node_name) {
            return Ok("STRING Dinoroseobacter shibae");
        }

        if is_valid_string_leeuwenhoekiella_blandensis_node_name(node_name) {
            return Ok("STRING Leeuwenhoekiella blandensis");
        }

        if is_valid_string_geobacter_lovleyi_node_name(node_name) {
            return Ok("STRING Geobacter lovleyi");
        }

        if is_valid_string_metallosphaera_sedula_node_name(node_name) {
            return Ok("STRING Metallosphaera sedula");
        }

        if is_valid_string_staphylothermus_marinus_node_name(node_name) {
            return Ok("STRING Staphylothermus marinus");
        }

        if is_valid_string_pseudomonas_mendocina_ymp_node_name(node_name) {
            return Ok("STRING Pseudomonas mendocina ymp");
        }

        if is_valid_string_serratia_proteamaculans_node_name(node_name) {
            return Ok("STRING Serratia proteamaculans");
        }

        if is_valid_string_enterobacter_sp_638_node_name(node_name) {
            return Ok("STRING Enterobacter sp. 638");
        }

        if is_valid_string_comamonas_testosteroni_kf1_node_name(node_name) {
            return Ok("STRING Comamonas testosteroni KF1");
        }

        if is_valid_string_marinomonas_sp_mwyl1_node_name(node_name) {
            return Ok("STRING Marinomonas sp. MWYL1");
        }

        if is_valid_string_amphimedon_queenslandica_node_name(node_name) {
            return Ok("STRING Amphimedon queenslandica");
        }

        if is_valid_string_terriglobus_saanensis_node_name(node_name) {
            return Ok("STRING Terriglobus saanensis");
        }

        if is_valid_string_bifidobacterium_dentium_node_name(node_name) {
            return Ok("STRING Bifidobacterium dentium");
        }

        if is_valid_string_thermosinus_carboxydivorans_node_name(node_name) {
            return Ok("STRING Thermosinus carboxydivorans");
        }

        if is_valid_string_flavobacterium_psychrophilum_node_name(node_name) {
            return Ok("STRING Flavobacterium psychrophilum");
        }

        if is_valid_string_ralstonia_pickettii_12j_node_name(node_name) {
            return Ok("STRING Ralstonia pickettii 12J");
        }

        if is_valid_string_kamptonema_formosum_node_name(node_name) {
            return Ok("STRING Kamptonema formosum");
        }

        if is_valid_string_methanococcus_maripaludis_c5_node_name(node_name) {
            return Ok("STRING Methanococcus maripaludis C5");
        }

        if is_valid_string_parvibaculum_lavamentivorans_node_name(node_name) {
            return Ok("STRING Parvibaculum lavamentivorans");
        }

        if is_valid_string_petrotoga_mobilis_node_name(node_name) {
            return Ok("STRING Petrotoga mobilis");
        }

        if is_valid_string_geobacter_bemidjiensis_node_name(node_name) {
            return Ok("STRING Geobacter bemidjiensis");
        }

        if is_valid_string_anaeromyxobacter_sp_fw1095_node_name(node_name) {
            return Ok("STRING Anaeromyxobacter sp. Fw1095");
        }

        if is_valid_string_lactobacillus_helveticus_dpc4571_node_name(node_name) {
            return Ok("STRING Lactobacillus helveticus DPC4571");
        }

        if is_valid_string_saccharopolyspora_erythraea_node_name(node_name) {
            return Ok("STRING Saccharopolyspora erythraea");
        }

        if is_valid_string_bacillus_sp_m313_node_name(node_name) {
            return Ok("STRING Bacillus sp. m313");
        }

        if is_valid_string_methanococcus_vannielii_node_name(node_name) {
            return Ok("STRING Methanococcus vannielii");
        }

        if is_valid_string_natrinema_sp_j72_node_name(node_name) {
            return Ok("STRING Natrinema sp. J72");
        }

        if is_valid_string_xenorhabdus_nematophila_node_name(node_name) {
            return Ok("STRING Xenorhabdus nematophila");
        }

        if is_valid_string_xenorhabdus_bovienii_node_name(node_name) {
            return Ok("STRING Xenorhabdus bovienii");
        }

        if is_valid_string_nocardioidaceae_bacterium_broad1_node_name(node_name) {
            return Ok("STRING Nocardioidaceae bacterium Broad1");
        }

        if is_valid_string_methanocorpusculum_labreanum_node_name(node_name) {
            return Ok("STRING Methanocorpusculum labreanum");
        }

        if is_valid_string_pyrobaculum_calidifontis_node_name(node_name) {
            return Ok("STRING Pyrobaculum calidifontis");
        }

        if is_valid_string_gramella_forsetii_node_name(node_name) {
            return Ok("STRING Gramella forsetii");
        }

        if is_valid_string_ruminococcus_obeum_atcc29174_node_name(node_name) {
            return Ok("STRING Ruminococcus obeum ATCC29174");
        }

        if is_valid_string_ruminococcus_torques_atcc27756_node_name(node_name) {
            return Ok("STRING Ruminococcus torques ATCC27756");
        }

        if is_valid_string_dorea_formicigenerans_atcc27755_node_name(node_name) {
            return Ok("STRING Dorea formicigenerans ATCC27755");
        }

        if is_valid_string_dorea_longicatena_dsm13814_node_name(node_name) {
            return Ok("STRING Dorea longicatena DSM13814");
        }

        if is_valid_string_eubacterium_ventriosum_node_name(node_name) {
            return Ok("STRING Eubacterium ventriosum");
        }

        if is_valid_string_desulfovibrio_piger_node_name(node_name) {
            return Ok("STRING Desulfovibrio piger");
        }

        if is_valid_string_parvimonas_micra_node_name(node_name) {
            return Ok("STRING Parvimonas micra");
        }

        if is_valid_string_actinomyces_odontolyticus_node_name(node_name) {
            return Ok("STRING Actinomyces odontolyticus");
        }

        if is_valid_string_pseudoflavonifractor_capillosus_node_name(node_name) {
            return Ok("STRING Pseudoflavonifractor capillosus");
        }

        if is_valid_string_clostridium_scindens_node_name(node_name) {
            return Ok("STRING Clostridium scindens");
        }

        if is_valid_string_eubacterium_hallii_node_name(node_name) {
            return Ok("STRING Eubacterium hallii");
        }

        if is_valid_string_ruminococcus_gnavus_node_name(node_name) {
            return Ok("STRING Ruminococcus gnavus");
        }

        if is_valid_string_subdoligranulum_variabile_node_name(node_name) {
            return Ok("STRING Subdoligranulum variabile");
        }

        if is_valid_string_ruminococcus_callidus_node_name(node_name) {
            return Ok("STRING Ruminococcus callidus");
        }

        if is_valid_string_coprococcus_eutactus_node_name(node_name) {
            return Ok("STRING Coprococcus eutactus");
        }

        if is_valid_string_bacteroides_ovatus_node_name(node_name) {
            return Ok("STRING Bacteroides ovatus");
        }

        if is_valid_string_parabacteroides_merdae_node_name(node_name) {
            return Ok("STRING Parabacteroides merdae");
        }

        if is_valid_string_bacteroides_uniformis_node_name(node_name) {
            return Ok("STRING Bacteroides uniformis");
        }

        if is_valid_string_faecalibacterium_prausnitzii_a2165_node_name(node_name) {
            return Ok("STRING Faecalibacterium prausnitzii A2165");
        }

        if is_valid_string_clostridium_sp_l250_node_name(node_name) {
            return Ok("STRING Clostridium sp. L250");
        }

        if is_valid_string_anaerostipes_caccae_node_name(node_name) {
            return Ok("STRING Anaerostipes caccae");
        }

        if is_valid_string_hoeflea_phototrophica_node_name(node_name) {
            return Ok("STRING Hoeflea phototrophica");
        }

        if is_valid_string_bacteroides_caccae_node_name(node_name) {
            return Ok("STRING Bacteroides caccae");
        }

        if is_valid_string_clostridium_bolteae_node_name(node_name) {
            return Ok("STRING Clostridium bolteae");
        }

        if is_valid_string_borrelia_duttonii_node_name(node_name) {
            return Ok("STRING Borrelia duttonii");
        }

        if is_valid_string_paracoccus_sp_trp_node_name(node_name) {
            return Ok("STRING Paracoccus sp. TRP");
        }

        if is_valid_string_vesicomyosocius_okutanii_node_name(node_name) {
            return Ok("STRING Vesicomyosocius okutanii");
        }

        if is_valid_string_ruthia_magnifica_node_name(node_name) {
            return Ok("STRING Ruthia magnifica");
        }

        if is_valid_string_halorubrum_halophilum_node_name(node_name) {
            return Ok("STRING Halorubrum halophilum");
        }

        if is_valid_string_rhodospirillum_centenum_node_name(node_name) {
            return Ok("STRING Rhodospirillum centenum");
        }

        if is_valid_string_actinopolyspora_erythraea_node_name(node_name) {
            return Ok("STRING Actinopolyspora erythraea");
        }

        if is_valid_string_hyperthermus_butylicus_node_name(node_name) {
            return Ok("STRING Hyperthermus butylicus");
        }

        if is_valid_string_actinobacillus_pleuropneumoniae_5b_l20_node_name(node_name) {
            return Ok("STRING Actinobacillus pleuropneumoniae 5b L20");
        }

        if is_valid_string_halorubrum_lacusprofundi_node_name(node_name) {
            return Ok("STRING Halorubrum lacusprofundi");
        }

        if is_valid_string_thermotoga_lettingae_node_name(node_name) {
            return Ok("STRING Thermotoga lettingae");
        }

        if is_valid_string_lactococcus_lactis_cremoris_node_name(node_name) {
            return Ok("STRING Lactococcus lactis cremoris");
        }

        if is_valid_string_methylobacterium_extorquens_pa1_node_name(node_name) {
            return Ok("STRING Methylobacterium extorquens PA1");
        }

        if is_valid_string_methanococcus_aeolicus_node_name(node_name) {
            return Ok("STRING Methanococcus aeolicus");
        }

        if is_valid_string_mycobacterium_tuberculosis_h37ra_node_name(node_name) {
            return Ok("STRING Mycobacterium tuberculosis H37Ra");
        }

        if is_valid_string_geobacillus_thermodenitrificans_node_name(node_name) {
            return Ok("STRING Geobacillus thermodenitrificans");
        }

        if is_valid_string_methanobrevibacter_smithii_atcc35061_node_name(node_name) {
            return Ok("STRING Methanobrevibacter smithii ATCC35061");
        }

        if is_valid_string_microvirga_lupini_node_name(node_name) {
            return Ok("STRING Microvirga lupini");
        }

        if is_valid_string_methylibium_petroleiphilum_node_name(node_name) {
            return Ok("STRING Methylibium petroleiphilum");
        }

        if is_valid_string_lactococcus_garvieae_lg2_node_name(node_name) {
            return Ok("STRING Lactococcus garvieae Lg2");
        }

        if is_valid_string_acinetobacter_rudis_node_name(node_name) {
            return Ok("STRING Acinetobacter rudis");
        }

        if is_valid_string_epilithonimonas_lactis_node_name(node_name) {
            return Ok("STRING Epilithonimonas lactis");
        }

        if is_valid_string_chryseobacterium_luteum_node_name(node_name) {
            return Ok("STRING Chryseobacterium luteum");
        }

        if is_valid_string_shewanella_sediminis_node_name(node_name) {
            return Ok("STRING Shewanella sediminis");
        }

        if is_valid_string_helicobacter_sp_mit016451_node_name(node_name) {
            return Ok("STRING Helicobacter sp. MIT016451");
        }

        if is_valid_string_thiomonas_arsenitoxydans_node_name(node_name) {
            return Ok("STRING Thiomonas arsenitoxydans");
        }

        if is_valid_string_methylobacterium_sp_446_node_name(node_name) {
            return Ok("STRING Methylobacterium sp. 446");
        }

        if is_valid_string_methylobacterium_radiotolerans_node_name(node_name) {
            return Ok("STRING Methylobacterium radiotolerans");
        }

        if is_valid_string_methanococcus_maripaludis_c7_node_name(node_name) {
            return Ok("STRING Methanococcus maripaludis C7");
        }

        if is_valid_string_nocardia_rhamnosiphila_node_name(node_name) {
            return Ok("STRING Nocardia rhamnosiphila");
        }

        if is_valid_string_clostridium_leptum_node_name(node_name) {
            return Ok("STRING Clostridium leptum");
        }

        if is_valid_string_clostridium_spiroforme_node_name(node_name) {
            return Ok("STRING Clostridium spiroforme");
        }

        if is_valid_string_eubacterium_dolichum_node_name(node_name) {
            return Ok("STRING Eubacterium dolichum");
        }

        if is_valid_string_ammonifex_degensii_node_name(node_name) {
            return Ok("STRING Ammonifex degensii");
        }

        if is_valid_string_dactylellina_haptotyla_node_name(node_name) {
            return Ok("STRING Dactylellina haptotyla");
        }

        if is_valid_string_baudoinia_compniacensis_node_name(node_name) {
            return Ok("STRING Baudoinia compniacensis");
        }

        if is_valid_string_clostridium_kluyveri_node_name(node_name) {
            return Ok("STRING Clostridium kluyveri");
        }

        if is_valid_string_porphyromonas_gingivalis_atcc33277_node_name(node_name) {
            return Ok("STRING Porphyromonas gingivalis ATCC33277");
        }

        if is_valid_string_kazachstania_africana_node_name(node_name) {
            return Ok("STRING Kazachstania africana");
        }

        if is_valid_string_neorickettsia_risticii_node_name(node_name) {
            return Ok("STRING Neorickettsia risticii");
        }

        if is_valid_string_bacteroides_vulgatus_node_name(node_name) {
            return Ok("STRING Bacteroides vulgatus");
        }

        if is_valid_string_parabacteroides_distasonis_node_name(node_name) {
            return Ok("STRING Parabacteroides distasonis");
        }

        if is_valid_string_actinomyces_graevenitzii_node_name(node_name) {
            return Ok("STRING Actinomyces graevenitzii");
        }

        if is_valid_string_neisseria_mucosa_c102_node_name(node_name) {
            return Ok("STRING Neisseria mucosa C102");
        }

        if is_valid_string_staphylococcus_hominis_c80_node_name(node_name) {
            return Ok("STRING Staphylococcus hominis C80");
        }

        if is_valid_string_staphylococcus_capitis_node_name(node_name) {
            return Ok("STRING Staphylococcus capitis");
        }

        if is_valid_string_streptococcus_sp_c150_node_name(node_name) {
            return Ok("STRING Streptococcus sp. C150");
        }

        if is_valid_string_idiomarina_salinarum_node_name(node_name) {
            return Ok("STRING Idiomarina salinarum");
        }

        if is_valid_string_sulfurihydrogenibium_sp_yo3aop1_node_name(node_name) {
            return Ok("STRING Sulfurihydrogenibium sp. YO3AOP1");
        }

        if is_valid_string_streptacidiphilus_jeojiense_node_name(node_name) {
            return Ok("STRING Streptacidiphilus jeojiense");
        }

        if is_valid_string_nitrosopumilus_maritimus_node_name(node_name) {
            return Ok("STRING Nitrosopumilus maritimus");
        }

        if is_valid_string_acinetobacter_oleivorans_node_name(node_name) {
            return Ok("STRING Acinetobacter oleivorans");
        }

        if is_valid_string_azorhizobium_caulinodans_node_name(node_name) {
            return Ok("STRING Azorhizobium caulinodans");
        }

        if is_valid_string_desulfatibacillum_alkenivorans_node_name(node_name) {
            return Ok("STRING Desulfatibacillum alkenivorans");
        }

        if is_valid_string_bacillus_selenitireducens_node_name(node_name) {
            return Ok("STRING Bacillus selenitireducens");
        }

        if is_valid_string_ochrobactrum_anthropi_node_name(node_name) {
            return Ok("STRING Ochrobactrum anthropi");
        }

        if is_valid_string_aciduliprofundum_boonei_node_name(node_name) {
            return Ok("STRING Aciduliprofundum boonei");
        }

        if is_valid_string_pelagibacter_sp_htcc7211_node_name(node_name) {
            return Ok("STRING Pelagibacter sp. HTCC7211");
        }

        if is_valid_string_rhodobacterales_bacterium_y4i_node_name(node_name) {
            return Ok("STRING Rhodobacterales bacterium Y4I");
        }

        if is_valid_string_ruegeria_sp_r11_node_name(node_name) {
            return Ok("STRING Ruegeria sp. R11");
        }

        if is_valid_string_pseudomonas_sp_chol1_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. Chol1");
        }

        if is_valid_string_methylobacterium_populi_node_name(node_name) {
            return Ok("STRING Methylobacterium populi");
        }

        if is_valid_string_acholeplasma_laidlawii_node_name(node_name) {
            return Ok("STRING Acholeplasma laidlawii");
        }

        if is_valid_string_bacillus_coahuilensis_node_name(node_name) {
            return Ok("STRING Bacillus coahuilensis");
        }

        if is_valid_string_geobacter_sp_m18_node_name(node_name) {
            return Ok("STRING Geobacter sp. M18");
        }

        if is_valid_string_geobacter_sp_m21_node_name(node_name) {
            return Ok("STRING Geobacter sp. M21");
        }

        if is_valid_string_marinobacter_algicola_node_name(node_name) {
            return Ok("STRING Marinobacter algicola");
        }

        if is_valid_string_amycolicicoccus_subflavus_node_name(node_name) {
            return Ok("STRING Amycolicicoccus subflavus");
        }

        if is_valid_string_marinitoga_piezophila_node_name(node_name) {
            return Ok("STRING Marinitoga piezophila");
        }

        if is_valid_string_streptomyces_clavuligerus_node_name(node_name) {
            return Ok("STRING Streptomyces clavuligerus");
        }

        if is_valid_string_bradyrhizobium_sp_cp53_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. Cp53");
        }

        if is_valid_string_clavibacter_michiganensis_ncppb382_node_name(node_name) {
            return Ok("STRING Clavibacter michiganensis NCPPB382");
        }

        if is_valid_string_pyrobaculum_neutrophilum_node_name(node_name) {
            return Ok("STRING Pyrobaculum neutrophilum");
        }

        if is_valid_string_methanococcus_maripaludis_c6_node_name(node_name) {
            return Ok("STRING Methanococcus maripaludis C6");
        }

        if is_valid_string_clostridium_botulinum_nctc2916_node_name(node_name) {
            return Ok("STRING Clostridium botulinum NCTC2916");
        }

        if is_valid_string_chryseobacterium_soli_node_name(node_name) {
            return Ok("STRING Chryseobacterium soli");
        }

        if is_valid_string_alistipes_putredinis_node_name(node_name) {
            return Ok("STRING Alistipes putredinis");
        }

        if is_valid_string_anaerofustis_stercorihominis_node_name(node_name) {
            return Ok("STRING Anaerofustis stercorihominis");
        }

        if is_valid_string_anaerotruncus_colihominis_node_name(node_name) {
            return Ok("STRING Anaerotruncus colihominis");
        }

        if is_valid_string_intestinibacter_bartlettii_node_name(node_name) {
            return Ok("STRING Intestinibacter bartlettii");
        }

        if is_valid_string_erysipelatoclostridium_ramosum_node_name(node_name) {
            return Ok("STRING Erysipelatoclostridium ramosum");
        }

        if is_valid_string_collinsella_stercoris_node_name(node_name) {
            return Ok("STRING Collinsella stercoris");
        }

        if is_valid_string_borrelia_valaisiana_vs116_node_name(node_name) {
            return Ok("STRING Borrelia valaisiana VS116");
        }

        if is_valid_string_actinosynnema_mirum_node_name(node_name) {
            return Ok("STRING Actinosynnema mirum");
        }

        if is_valid_string_brachybacterium_faecium_node_name(node_name) {
            return Ok("STRING Brachybacterium faecium");
        }

        if is_valid_string_cellulomonas_flavigena_node_name(node_name) {
            return Ok("STRING Cellulomonas flavigena");
        }

        if is_valid_string_nocardiopsis_dassonvillei_node_name(node_name) {
            return Ok("STRING Nocardiopsis dassonvillei");
        }

        if is_valid_string_sanguibacter_keddieii_node_name(node_name) {
            return Ok("STRING Sanguibacter keddieii");
        }

        if is_valid_string_stackebrandtia_nassauensis_node_name(node_name) {
            return Ok("STRING Stackebrandtia nassauensis");
        }

        if is_valid_string_xylanimonas_cellulosilytica_node_name(node_name) {
            return Ok("STRING Xylanimonas cellulosilytica");
        }

        if is_valid_string_sorangium_cellulosum_so_ce56_node_name(node_name) {
            return Ok("STRING Sorangium cellulosum So ce56");
        }

        if is_valid_string_microcystis_aeruginosa_node_name(node_name) {
            return Ok("STRING Microcystis aeruginosa");
        }

        if is_valid_string_bacteroides_stercoris_node_name(node_name) {
            return Ok("STRING Bacteroides stercoris");
        }

        if is_valid_string_microbacterium_profundi_node_name(node_name) {
            return Ok("STRING Microbacterium profundi");
        }

        if is_valid_string_phenylobacterium_zucineum_node_name(node_name) {
            return Ok("STRING Phenylobacterium zucineum");
        }

        if is_valid_string_amoebophilus_asiaticus_node_name(node_name) {
            return Ok("STRING Amoebophilus asiaticus");
        }

        if is_valid_string_opitutus_terrae_node_name(node_name) {
            return Ok("STRING Opitutus terrae");
        }

        if is_valid_string_polynucleobacter_necessarius_stir1_node_name(node_name) {
            return Ok("STRING Polynucleobacter necessarius STIR1");
        }

        if is_valid_string_kitasatospora_setae_node_name(node_name) {
            return Ok("STRING Kitasatospora setae");
        }

        if is_valid_string_rickettsia_rickettsii_node_name(node_name) {
            return Ok("STRING Rickettsia rickettsii");
        }

        if is_valid_string_sphingobium_japonicum_ut26s_node_name(node_name) {
            return Ok("STRING Sphingobium japonicum UT26S");
        }

        if is_valid_string_arthrobacter_chlorophenolicus_node_name(node_name) {
            return Ok("STRING Arthrobacter chlorophenolicus");
        }

        if is_valid_string_ignicoccus_hospitalis_node_name(node_name) {
            return Ok("STRING Ignicoccus hospitalis");
        }

        if is_valid_string_xanthomonas_arboricola_node_name(node_name) {
            return Ok("STRING Xanthomonas arboricola");
        }

        if is_valid_string_glaciecola_sp_htcc2999_node_name(node_name) {
            return Ok("STRING Glaciecola sp. HTCC2999");
        }

        if is_valid_string_streptomyces_griseus_node_name(node_name) {
            return Ok("STRING Streptomyces griseus");
        }

        if is_valid_string_methanococcus_voltae_node_name(node_name) {
            return Ok("STRING Methanococcus voltae");
        }

        if is_valid_string_methanoregula_boonei_node_name(node_name) {
            return Ok("STRING Methanoregula boonei");
        }

        if is_valid_string_clostridium_sp_7243faa_node_name(node_name) {
            return Ok("STRING Clostridium sp. 7243FAA");
        }

        if is_valid_string_desulfovibrio_sp_31syn3_node_name(node_name) {
            return Ok("STRING Desulfovibrio sp. 31syn3");
        }

        if is_valid_string_fusobacterium_nucleatum_animalis_node_name(node_name) {
            return Ok("STRING Fusobacterium nucleatum animalis");
        }

        if is_valid_string_ruminococcus_sp_5139bfaa_node_name(node_name) {
            return Ok("STRING Ruminococcus sp. 5139BFAA");
        }

        if is_valid_string_synergistes_sp_31syn1_node_name(node_name) {
            return Ok("STRING Synergistes sp. 31syn1");
        }

        if is_valid_string_clostridiales_bacterium_1747faa_node_name(node_name) {
            return Ok("STRING Clostridiales bacterium 1747FAA");
        }

        if is_valid_string_bacteroides_fragilis_3112_node_name(node_name) {
            return Ok("STRING Bacteroides fragilis 3112");
        }

        if is_valid_string_streptomyces_albus_j1074_node_name(node_name) {
            return Ok("STRING Streptomyces albus J1074");
        }

        if is_valid_string_streptomyces_pristinaespiralis_node_name(node_name) {
            return Ok("STRING Streptomyces pristinaespiralis");
        }

        if is_valid_string_natranaerobius_thermophilus_node_name(node_name) {
            return Ok("STRING Natranaerobius thermophilus");
        }

        if is_valid_string_macrococcus_caseolyticus_node_name(node_name) {
            return Ok("STRING Macrococcus caseolyticus");
        }

        if is_valid_string_shewanella_halifaxensis_node_name(node_name) {
            return Ok("STRING Shewanella halifaxensis");
        }

        if is_valid_string_cloacimonas_acidaminovorans_node_name(node_name) {
            return Ok("STRING Cloacimonas acidaminovorans");
        }

        if is_valid_string_arthrospira_platensis_c1_node_name(node_name) {
            return Ok("STRING Arthrospira platensis C1");
        }

        if is_valid_string_methylobacterium_nodulans_node_name(node_name) {
            return Ok("STRING Methylobacterium nodulans");
        }

        if is_valid_string_streptomyces_sviceus_node_name(node_name) {
            return Ok("STRING Streptomyces sviceus");
        }

        if is_valid_string_micrococcus_luteus_nctc_2665_node_name(node_name) {
            return Ok("STRING Micrococcus luteus NCTC 2665");
        }

        if is_valid_string_streptomyces_sp_mg1_node_name(node_name) {
            return Ok("STRING Streptomyces sp. Mg1");
        }

        if is_valid_string_erwinia_tasmaniensis_node_name(node_name) {
            return Ok("STRING Erwinia tasmaniensis");
        }

        if is_valid_string_pelagibacter_ubique_htcc8051_node_name(node_name) {
            return Ok("STRING Pelagibacter ubique HTCC8051");
        }

        if is_valid_string_acinetobacter_sp_ver3_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. Ver3");
        }

        if is_valid_string_streptomyces_griseoflavus_node_name(node_name) {
            return Ok("STRING Streptomyces griseoflavus");
        }

        if is_valid_string_rhodobacteraceae_bacterium_klh11_node_name(node_name) {
            return Ok("STRING Rhodobacteraceae bacterium KLH11");
        }

        if is_valid_string_streptococcus_gordonii_node_name(node_name) {
            return Ok("STRING Streptococcus gordonii");
        }

        if is_valid_string_pedobacter_oryzae_node_name(node_name) {
            return Ok("STRING Pedobacter oryzae");
        }

        if is_valid_string_gordonia_kroppenstedtii_node_name(node_name) {
            return Ok("STRING Gordonia kroppenstedtii");
        }

        if is_valid_string_escherichia_coli_bl21_node_name(node_name) {
            return Ok("STRING Escherichia coli BL21");
        }

        if is_valid_string_thermobispora_bispora_node_name(node_name) {
            return Ok("STRING Thermobispora bispora");
        }

        if is_valid_string_cryptobacterium_curtum_node_name(node_name) {
            return Ok("STRING Cryptobacterium curtum");
        }

        if is_valid_string_dethiosulfovibrio_peptidovorans_node_name(node_name) {
            return Ok("STRING Dethiosulfovibrio peptidovorans");
        }

        if is_valid_string_halogeometricum_borinquense_node_name(node_name) {
            return Ok("STRING Halogeometricum borinquense");
        }

        if is_valid_string_conexibacter_woesei_dsm14684_node_name(node_name) {
            return Ok("STRING Conexibacter woesei DSM14684");
        }

        if is_valid_string_citrobacter_sp_302_node_name(node_name) {
            return Ok("STRING Citrobacter sp. 302");
        }

        if is_valid_string_coprobacillus_sp_291_node_name(node_name) {
            return Ok("STRING Coprobacillus sp. 291");
        }

        if is_valid_string_fusobacterium_nucleatum_vincentii_3136a2_node_name(node_name) {
            return Ok("STRING Fusobacterium nucleatum vincentii 3136A2");
        }

        if is_valid_string_fusobacterium_nucleatum_vincentii_4113_node_name(node_name) {
            return Ok("STRING Fusobacterium nucleatum vincentii 4113");
        }

        if is_valid_string_streptococcus_sp_2136faa_node_name(node_name) {
            return Ok("STRING Streptococcus sp. 2136FAA");
        }

        if is_valid_string_burkholderiales_bacterium_1147_node_name(node_name) {
            return Ok("STRING Burkholderiales bacterium 1147");
        }

        if is_valid_string_fusobacterium_gonidiaformans_node_name(node_name) {
            return Ok("STRING Fusobacterium gonidiaformans");
        }

        if is_valid_string_fusobacterium_mortiferum_node_name(node_name) {
            return Ok("STRING Fusobacterium mortiferum");
        }

        if is_valid_string_fusobacterium_ulcerans_node_name(node_name) {
            return Ok("STRING Fusobacterium ulcerans");
        }

        if is_valid_string_fusobacterium_varium_node_name(node_name) {
            return Ok("STRING Fusobacterium varium");
        }

        if is_valid_string_bacteroides_coprocola_node_name(node_name) {
            return Ok("STRING Bacteroides coprocola");
        }

        if is_valid_string_cladophialophora_yegresii_node_name(node_name) {
            return Ok("STRING Cladophialophora yegresii");
        }

        if is_valid_string_geobacillus_sp_wch70_node_name(node_name) {
            return Ok("STRING Geobacillus sp. WCH70");
        }

        if is_valid_string_thermomonospora_curvata_node_name(node_name) {
            return Ok("STRING Thermomonospora curvata");
        }

        if is_valid_string_beutenbergia_cavernae_node_name(node_name) {
            return Ok("STRING Beutenbergia cavernae");
        }

        if is_valid_string_dyadobacter_fermentans_node_name(node_name) {
            return Ok("STRING Dyadobacter fermentans");
        }

        if is_valid_string_slackia_heliotrinireducens_node_name(node_name) {
            return Ok("STRING Slackia heliotrinireducens");
        }

        if is_valid_string_jonesia_denitrificans_node_name(node_name) {
            return Ok("STRING Jonesia denitrificans");
        }

        if is_valid_string_saccharomonospora_viridis_node_name(node_name) {
            return Ok("STRING Saccharomonospora viridis");
        }

        if is_valid_string_bacteroides_intestinalis_node_name(node_name) {
            return Ok("STRING Bacteroides intestinalis");
        }

        if is_valid_string_providencia_stuartii_node_name(node_name) {
            return Ok("STRING Providencia stuartii");
        }

        if is_valid_string_ruminococcus_lactaris_node_name(node_name) {
            return Ok("STRING Ruminococcus lactaris");
        }

        if is_valid_string_proteus_penneri_node_name(node_name) {
            return Ok("STRING Proteus penneri");
        }

        if is_valid_string_nitratireductor_basaltis_node_name(node_name) {
            return Ok("STRING Nitratireductor basaltis");
        }

        if is_valid_string_nitrosococcus_halophilus_node_name(node_name) {
            return Ok("STRING Nitrosococcus halophilus");
        }

        if is_valid_string_colletotrichum_gloeosporioides_node_name(node_name) {
            return Ok("STRING Colletotrichum gloeosporioides");
        }

        if is_valid_string_blautia_hydrogenotrophica_node_name(node_name) {
            return Ok("STRING Blautia hydrogenotrophica");
        }

        if is_valid_string_achromobacter_arsenitoxydans_node_name(node_name) {
            return Ok("STRING Achromobacter arsenitoxydans");
        }

        if is_valid_string_pseudomonas_stutzeri_ts44_node_name(node_name) {
            return Ok("STRING Pseudomonas stutzeri TS44");
        }

        if is_valid_string_modestobacter_marinus_node_name(node_name) {
            return Ok("STRING Modestobacter marinus");
        }

        if is_valid_string_desulforudis_audaxviator_node_name(node_name) {
            return Ok("STRING Desulforudis audaxviator");
        }

        if is_valid_string_verrucomicrobia_bacterium_lp2a_node_name(node_name) {
            return Ok("STRING Verrucomicrobia bacterium LP2A");
        }

        if is_valid_string_marvinbryantia_formatexigens_node_name(node_name) {
            return Ok("STRING Marvinbryantia formatexigens");
        }

        if is_valid_string_kytococcus_sedentarius_node_name(node_name) {
            return Ok("STRING Kytococcus sedentarius");
        }

        if is_valid_string_nakamurella_multipartita_node_name(node_name) {
            return Ok("STRING Nakamurella multipartita");
        }

        if is_valid_string_streptosporangium_roseum_node_name(node_name) {
            return Ok("STRING Streptosporangium roseum");
        }

        if is_valid_string_catenulispora_acidiphila_node_name(node_name) {
            return Ok("STRING Catenulispora acidiphila");
        }

        if is_valid_string_sphaerobacter_thermophilus_node_name(node_name) {
            return Ok("STRING Sphaerobacter thermophilus");
        }

        if is_valid_string_kribbella_flavida_node_name(node_name) {
            return Ok("STRING Kribbella flavida");
        }

        if is_valid_string_veillonella_parvula_node_name(node_name) {
            return Ok("STRING Veillonella parvula");
        }

        if is_valid_string_eggerthella_lenta_node_name(node_name) {
            return Ok("STRING Eggerthella lenta");
        }

        if is_valid_string_methylacidiphilum_infernorum_node_name(node_name) {
            return Ok("STRING Methylacidiphilum infernorum");
        }

        if is_valid_string_escherichia_coli_atcc8739_node_name(node_name) {
            return Ok("STRING Escherichia coli ATCC8739");
        }

        if is_valid_string_streptococcus_canis_node_name(node_name) {
            return Ok("STRING Streptococcus canis");
        }

        if is_valid_string_galeopterus_variegatus_node_name(node_name) {
            return Ok("STRING Galeopterus variegatus");
        }

        if is_valid_string_bacteroides_finegoldii_node_name(node_name) {
            return Ok("STRING Bacteroides finegoldii");
        }

        if is_valid_string_bacteroides_eggerthii_node_name(node_name) {
            return Ok("STRING Bacteroides eggerthii");
        }

        if is_valid_string_bacteroides_pectinophilus_node_name(node_name) {
            return Ok("STRING Bacteroides pectinophilus");
        }

        if is_valid_string_myxococcus_fulvus_node_name(node_name) {
            return Ok("STRING Myxococcus fulvus");
        }

        if is_valid_string_bacteroides_plebeius_node_name(node_name) {
            return Ok("STRING Bacteroides plebeius");
        }

        if is_valid_string_thermosipho_africanus_node_name(node_name) {
            return Ok("STRING Thermosipho africanus");
        }

        if is_valid_string_francisella_philomiragia_node_name(node_name) {
            return Ok("STRING Francisella philomiragia");
        }

        if is_valid_string_pelosinus_sp_ufo1_node_name(node_name) {
            return Ok("STRING Pelosinus sp. UFO1");
        }

        if is_valid_string_ktedonobacter_racemifer_node_name(node_name) {
            return Ok("STRING Ktedonobacter racemifer");
        }

        if is_valid_string_halomicrobium_mukohataei_node_name(node_name) {
            return Ok("STRING Halomicrobium mukohataei");
        }

        if is_valid_string_desulfohalobium_retbaense_node_name(node_name) {
            return Ok("STRING Desulfohalobium retbaense");
        }

        if is_valid_string_desulfotomaculum_acetoxidans_node_name(node_name) {
            return Ok("STRING Desulfotomaculum acetoxidans");
        }

        if is_valid_string_pedobacter_heparinus_node_name(node_name) {
            return Ok("STRING Pedobacter heparinus");
        }

        if is_valid_string_chitinophaga_pinensis_node_name(node_name) {
            return Ok("STRING Chitinophaga pinensis");
        }

        if is_valid_string_acinetobacter_soli_node_name(node_name) {
            return Ok("STRING Acinetobacter soli");
        }

        if is_valid_string_mycobacterium_intracellulare_node_name(node_name) {
            return Ok("STRING Mycobacterium intracellulare");
        }

        if is_valid_string_flavobacteria_bacterium_ms0242a_node_name(node_name) {
            return Ok("STRING Flavobacteria bacterium MS0242A");
        }

        if is_valid_string_puniceispirillum_marinum_node_name(node_name) {
            return Ok("STRING Puniceispirillum marinum");
        }

        if is_valid_string_neisseria_lactamica_node_name(node_name) {
            return Ok("STRING Neisseria lactamica");
        }

        if is_valid_string_moorea_producens_node_name(node_name) {
            return Ok("STRING Moorea producens");
        }

        if is_valid_string_desulfurococcus_kamchatkensis_node_name(node_name) {
            return Ok("STRING Desulfurococcus kamchatkensis");
        }

        if is_valid_string_chryseobacterium_hispalense_node_name(node_name) {
            return Ok("STRING Chryseobacterium hispalense");
        }

        if is_valid_string_anoxybacillus_flavithermus_wk1_node_name(node_name) {
            return Ok("STRING Anoxybacillus flavithermus WK1");
        }

        if is_valid_string_rhizobium_etli_ciat652_node_name(node_name) {
            return Ok("STRING Rhizobium etli CIAT652");
        }

        if is_valid_string_marinomonas_posidonica_node_name(node_name) {
            return Ok("STRING Marinomonas posidonica");
        }

        if is_valid_string_rhizobium_alamii_node_name(node_name) {
            return Ok("STRING Rhizobium alamii");
        }

        if is_valid_string_glaciecola_arctica_node_name(node_name) {
            return Ok("STRING Glaciecola arctica");
        }

        if is_valid_string_psychrobacter_sp_tb15_node_name(node_name) {
            return Ok("STRING Psychrobacter sp. TB15");
        }

        if is_valid_string_arthrobacter_sp_tb23_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. TB23");
        }

        if is_valid_string_mycoplasma_fermentans_node_name(node_name) {
            return Ok("STRING Mycoplasma fermentans");
        }

        if is_valid_string_thauera_sp_63_node_name(node_name) {
            return Ok("STRING Thauera sp. 63");
        }

        if is_valid_string_chthoniobacter_flavus_node_name(node_name) {
            return Ok("STRING Chthoniobacter flavus");
        }

        if is_valid_string_cyanothece_sp_pcc7822_node_name(node_name) {
            return Ok("STRING Cyanothece sp. PCC7822");
        }

        if is_valid_string_cellvibrio_japonicus_node_name(node_name) {
            return Ok("STRING Cellvibrio japonicus");
        }

        if is_valid_string_borrelia_spielmanii_node_name(node_name) {
            return Ok("STRING Borrelia spielmanii");
        }

        if is_valid_string_heliobacterium_modesticaldum_node_name(node_name) {
            return Ok("STRING Heliobacterium modesticaldum");
        }

        if is_valid_string_thermus_aquaticus_node_name(node_name) {
            return Ok("STRING Thermus aquaticus");
        }

        if is_valid_string_streptomyces_avicenniae_node_name(node_name) {
            return Ok("STRING Streptomyces avicenniae");
        }

        if is_valid_string_tyzzerella_nexilis_node_name(node_name) {
            return Ok("STRING Tyzzerella nexilis");
        }

        if is_valid_string_clostridium_hiranonis_node_name(node_name) {
            return Ok("STRING Clostridium hiranonis");
        }

        if is_valid_string_mitsuokella_multacida_node_name(node_name) {
            return Ok("STRING Mitsuokella multacida");
        }

        if is_valid_string_providencia_rustigianii_node_name(node_name) {
            return Ok("STRING Providencia rustigianii");
        }

        if is_valid_string_citrobacter_youngae_node_name(node_name) {
            return Ok("STRING Citrobacter youngae");
        }

        if is_valid_string_citreicella_sp_se45_node_name(node_name) {
            return Ok("STRING Citreicella sp. SE45");
        }

        if is_valid_string_haliangium_ochraceum_node_name(node_name) {
            return Ok("STRING Haliangium ochraceum");
        }

        if is_valid_string_escherichia_albertii_tw07627_node_name(node_name) {
            return Ok("STRING Escherichia albertii TW07627");
        }

        if is_valid_string_eggerthella_sp_yy7918_node_name(node_name) {
            return Ok("STRING Eggerthella sp. YY7918");
        }

        if is_valid_string_spirosoma_linguale_node_name(node_name) {
            return Ok("STRING Spirosoma linguale");
        }

        if is_valid_string_corynebacterium_urealyticum_node_name(node_name) {
            return Ok("STRING Corynebacterium urealyticum");
        }

        if is_valid_string_jejuia_pallidilutea_node_name(node_name) {
            return Ok("STRING Jejuia pallidilutea");
        }

        if is_valid_string_meiothermus_ruber_node_name(node_name) {
            return Ok("STRING Meiothermus ruber");
        }

        if is_valid_string_oligotropha_carboxidovorans_node_name(node_name) {
            return Ok("STRING Oligotropha carboxidovorans");
        }

        if is_valid_string_rheinheimera_sp_a13l_node_name(node_name) {
            return Ok("STRING Rheinheimera sp. A13L");
        }

        if is_valid_string_caulobacter_segnis_node_name(node_name) {
            return Ok("STRING Caulobacter segnis");
        }

        if is_valid_string_acetivibrio_cellulolyticus_node_name(node_name) {
            return Ok("STRING Acetivibrio cellulolyticus");
        }

        if is_valid_string_pedobacter_sp_v48_node_name(node_name) {
            return Ok("STRING Pedobacter sp. V48");
        }

        if is_valid_string_caldisericum_exile_node_name(node_name) {
            return Ok("STRING Caldisericum exile");
        }

        if is_valid_string_oceanimonas_sp_gk1_node_name(node_name) {
            return Ok("STRING Oceanimonas sp. GK1");
        }

        if is_valid_string_escherichia_coli_k12_mg1655_node_name(node_name) {
            return Ok("STRING Escherichia coli K12 MG1655");
        }

        if is_valid_string_lactobacillus_buchneri_nrrlb30929_node_name(node_name) {
            return Ok("STRING Lactobacillus buchneri NRRLB30929");
        }

        if is_valid_string_butyrivibrio_crossotus_node_name(node_name) {
            return Ok("STRING Butyrivibrio crossotus");
        }

        if is_valid_string_azobacteroides_pseudotrichonymphae_node_name(node_name) {
            return Ok("STRING Azobacteroides pseudotrichonymphae");
        }

        if is_valid_string_mycoplasma_crocodyli_node_name(node_name) {
            return Ok("STRING Mycoplasma crocodyli");
        }

        if is_valid_string_actinoplanes_missouriensis_node_name(node_name) {
            return Ok("STRING Actinoplanes missouriensis");
        }

        if is_valid_string_riesia_pediculicola_node_name(node_name) {
            return Ok("STRING Riesia pediculicola");
        }

        if is_valid_string_eubacterium_eligens_node_name(node_name) {
            return Ok("STRING Eubacterium eligens");
        }

        if is_valid_string_butyrivibrio_proteoclasticus_b316_node_name(node_name) {
            return Ok("STRING Butyrivibrio proteoclasticus B316");
        }

        if is_valid_string_dictyoglomus_turgidum_node_name(node_name) {
            return Ok("STRING Dictyoglomus turgidum");
        }

        if is_valid_string_chlorobaculum_parvum_node_name(node_name) {
            return Ok("STRING Chlorobaculum parvum");
        }

        if is_valid_string_chloroherpeton_thalassium_node_name(node_name) {
            return Ok("STRING Chloroherpeton thalassium");
        }

        if is_valid_string_pantoea_sp_ab_node_name(node_name) {
            return Ok("STRING Pantoea sp. aB");
        }

        if is_valid_string_citromicrobium_sp_jlt1363_node_name(node_name) {
            return Ok("STRING Citromicrobium sp. JLT1363");
        }

        if is_valid_string_bifidobacterium_angulatum_node_name(node_name) {
            return Ok("STRING Bifidobacterium angulatum");
        }

        if is_valid_string_holdemanella_biformis_node_name(node_name) {
            return Ok("STRING Holdemanella biformis");
        }

        if is_valid_string_rhodothermus_marinus_node_name(node_name) {
            return Ok("STRING Rhodothermus marinus");
        }

        if is_valid_string_streptobacillus_moniliformis_node_name(node_name) {
            return Ok("STRING Streptobacillus moniliformis");
        }

        if is_valid_string_halorhabdus_utahensis_node_name(node_name) {
            return Ok("STRING Halorhabdus utahensis");
        }

        if is_valid_string_ectothiorhodospira_sp_phs1_node_name(node_name) {
            return Ok("STRING Ectothiorhodospira sp. PHS1");
        }

        if is_valid_string_acinetobacter_sp_niph973_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. NIPH973");
        }

        if is_valid_string_providencia_alcalifaciens_node_name(node_name) {
            return Ok("STRING Providencia alcalifaciens");
        }

        if is_valid_string_providencia_rettgeri_dsm1131_node_name(node_name) {
            return Ok("STRING Providencia rettgeri DSM1131");
        }

        if is_valid_string_collinsella_intestinalis_node_name(node_name) {
            return Ok("STRING Collinsella intestinalis");
        }

        if is_valid_string_borrelia_bissettii_node_name(node_name) {
            return Ok("STRING Borrelia bissettii");
        }

        if is_valid_string_methanosphaerula_palustris_node_name(node_name) {
            return Ok("STRING Methanosphaerula palustris");
        }

        if is_valid_string_kosmotoga_olearia_node_name(node_name) {
            return Ok("STRING Kosmotoga olearia");
        }

        if is_valid_string_atopobium_parvulum_node_name(node_name) {
            return Ok("STRING Atopobium parvulum");
        }

        if is_valid_string_tsukamurella_paurometabola_node_name(node_name) {
            return Ok("STRING Tsukamurella paurometabola");
        }

        if is_valid_string_capnocytophaga_ochracea_dsm7271_node_name(node_name) {
            return Ok("STRING Capnocytophaga ochracea DSM7271");
        }

        if is_valid_string_alicyclobacillus_acidocaldarius_dsm446_node_name(node_name) {
            return Ok("STRING Alicyclobacillus acidocaldarius DSM446");
        }

        if is_valid_string_actinomyces_timonensis_node_name(node_name) {
            return Ok("STRING Actinomyces timonensis");
        }

        if is_valid_string_caldicellulosiruptor_bescii_node_name(node_name) {
            return Ok("STRING Caldicellulosiruptor bescii");
        }

        if is_valid_string_planctomyces_limnophilus_node_name(node_name) {
            return Ok("STRING Planctomyces limnophilus");
        }

        if is_valid_string_pseudomonas_caeni_node_name(node_name) {
            return Ok("STRING Pseudomonas caeni");
        }

        if is_valid_string_accumulibacter_phosphatis_node_name(node_name) {
            return Ok("STRING Accumulibacter phosphatis");
        }

        if is_valid_string_stenotrophomonas_maltophilia_k279a_node_name(node_name) {
            return Ok("STRING Stenotrophomonas maltophilia K279a");
        }

        if is_valid_string_denitrovibrio_acetiphilus_node_name(node_name) {
            return Ok("STRING Denitrovibrio acetiphilus");
        }

        if is_valid_string_kangiella_koreensis_node_name(node_name) {
            return Ok("STRING Kangiella koreensis");
        }

        if is_valid_string_leptotrichia_buccalis_node_name(node_name) {
            return Ok("STRING Leptotrichia buccalis");
        }

        if is_valid_string_haloferax_mediterranei_node_name(node_name) {
            return Ok("STRING Haloferax mediterranei");
        }

        if is_valid_string_methanothermococcus_thermolithotrophicus_node_name(node_name) {
            return Ok("STRING Methanothermococcus thermolithotrophicus");
        }

        if is_valid_string_thermococcus_onnurineus_node_name(node_name) {
            return Ok("STRING Thermococcus onnurineus");
        }

        if is_valid_string_desulfovibrio_desulfuricans_atcc27774_node_name(node_name) {
            return Ok("STRING Desulfovibrio desulfuricans ATCC27774");
        }

        if is_valid_string_acinetobacter_sp_atcc27244_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. ATCC27244");
        }

        if is_valid_string_actinomyces_coleocanis_node_name(node_name) {
            return Ok("STRING Actinomyces coleocanis");
        }

        if is_valid_string_actinomyces_urogenitalis_dsm15434_node_name(node_name) {
            return Ok("STRING Actinomyces urogenitalis DSM15434");
        }

        if is_valid_string_anaerococcus_lactolyticus_atcc51172_node_name(node_name) {
            return Ok("STRING Anaerococcus lactolyticus ATCC51172");
        }

        if is_valid_string_anaerococcus_tetradius_node_name(node_name) {
            return Ok("STRING Anaerococcus tetradius");
        }

        if is_valid_string_atopobium_vaginae_dsm15829_node_name(node_name) {
            return Ok("STRING Atopobium vaginae DSM15829");
        }

        if is_valid_string_chryseobacterium_gleum_node_name(node_name) {
            return Ok("STRING Chryseobacterium gleum");
        }

        if is_valid_string_corynebacterium_accolens_node_name(node_name) {
            return Ok("STRING Corynebacterium accolens");
        }

        if is_valid_string_corynebacterium_lipophiloflavum_node_name(node_name) {
            return Ok("STRING Corynebacterium lipophiloflavum");
        }

        if is_valid_string_corynebacterium_pseudogenitalium_node_name(node_name) {
            return Ok("STRING Corynebacterium pseudogenitalium");
        }

        if is_valid_string_corynebacterium_striatum_node_name(node_name) {
            return Ok("STRING Corynebacterium striatum");
        }

        if is_valid_string_finegoldia_magna_atcc53516_node_name(node_name) {
            return Ok("STRING Finegoldia magna ATCC53516");
        }

        if is_valid_string_lactobacillus_antri_node_name(node_name) {
            return Ok("STRING Lactobacillus antri");
        }

        if is_valid_string_lactobacillus_buchneri_atcc11577_node_name(node_name) {
            return Ok("STRING Lactobacillus buchneri ATCC11577");
        }

        if is_valid_string_lactobacillus_ultunensis_node_name(node_name) {
            return Ok("STRING Lactobacillus ultunensis");
        }

        if is_valid_string_listeria_grayi_node_name(node_name) {
            return Ok("STRING Listeria grayi");
        }

        if is_valid_string_mycobacterium_parascrofulaceum_node_name(node_name) {
            return Ok("STRING Mycobacterium parascrofulaceum");
        }

        if is_valid_string_sphingobacterium_spiritivorum_node_name(node_name) {
            return Ok("STRING Sphingobacterium spiritivorum");
        }

        if is_valid_string_staphylococcus_caprae_node_name(node_name) {
            return Ok("STRING Staphylococcus caprae");
        }

        if is_valid_string_streptococcus_equinus_node_name(node_name) {
            return Ok("STRING Streptococcus equinus");
        }

        if is_valid_string_desulfomicrobium_baculatum_node_name(node_name) {
            return Ok("STRING Desulfomicrobium baculatum");
        }

        if is_valid_string_sulfurospirillum_deleyianum_node_name(node_name) {
            return Ok("STRING Sulfurospirillum deleyianum");
        }

        if is_valid_string_thermanaerovibrio_acidaminovorans_node_name(node_name) {
            return Ok("STRING Thermanaerovibrio acidaminovorans");
        }

        if is_valid_string_thermobaculum_terrenum_node_name(node_name) {
            return Ok("STRING Thermobaculum terrenum");
        }

        if is_valid_string_acidimicrobium_ferrooxidans_node_name(node_name) {
            return Ok("STRING Acidimicrobium ferrooxidans");
        }

        if is_valid_string_anaerococcus_prevotii_dsm20548_node_name(node_name) {
            return Ok("STRING Anaerococcus prevotii DSM20548");
        }

        if is_valid_string_sebaldella_termitidis_node_name(node_name) {
            return Ok("STRING Sebaldella termitidis");
        }

        if is_valid_string_desulfovibrio_salexigens_node_name(node_name) {
            return Ok("STRING Desulfovibrio salexigens");
        }

        if is_valid_string_brachyspira_murdochii_node_name(node_name) {
            return Ok("STRING Brachyspira murdochii");
        }

        if is_valid_string_geodermatophilus_obscurus_node_name(node_name) {
            return Ok("STRING Geodermatophilus obscurus");
        }

        if is_valid_string_gordonia_bronchialis_node_name(node_name) {
            return Ok("STRING Gordonia bronchialis");
        }

        if is_valid_string_meiothermus_silvanus_node_name(node_name) {
            return Ok("STRING Meiothermus silvanus");
        }

        if is_valid_string_yersinia_aldovae_node_name(node_name) {
            return Ok("STRING Yersinia aldovae");
        }

        if is_valid_string_proteus_mirabilis_node_name(node_name) {
            return Ok("STRING Proteus mirabilis");
        }

        if is_valid_string_pyrococcus_yayanosii_node_name(node_name) {
            return Ok("STRING Pyrococcus yayanosii");
        }

        if is_valid_string_thecamonas_trahens_node_name(node_name) {
            return Ok("STRING Thecamonas trahens");
        }

        if is_valid_string_rhodoluna_lacicola_node_name(node_name) {
            return Ok("STRING Rhodoluna lacicola");
        }

        if is_valid_string_pirellula_staleyi_node_name(node_name) {
            return Ok("STRING Pirellula staleyi");
        }

        if is_valid_string_flavobacteriaceae_bacterium_351910_node_name(node_name) {
            return Ok("STRING Flavobacteriaceae bacterium 351910");
        }

        if is_valid_string_cylindrospermopsis_raciborskii_node_name(node_name) {
            return Ok("STRING Cylindrospermopsis raciborskii");
        }

        if is_valid_string_raphidiopsis_brookii_node_name(node_name) {
            return Ok("STRING Raphidiopsis brookii");
        }

        if is_valid_string_acidovorax_ebreus_node_name(node_name) {
            return Ok("STRING Acidovorax ebreus");
        }

        if is_valid_string_mesorhizobium_opportunistum_node_name(node_name) {
            return Ok("STRING Mesorhizobium opportunistum");
        }

        if is_valid_string_clostridium_carboxidivorans_node_name(node_name) {
            return Ok("STRING Clostridium carboxidivorans");
        }

        if is_valid_string_clostridium_botulinum_a2_kyoto_node_name(node_name) {
            return Ok("STRING Clostridium botulinum A2 Kyoto");
        }

        if is_valid_string_clostridium_botulinum_e1_bont_node_name(node_name) {
            return Ok("STRING Clostridium botulinum E1 BoNT");
        }

        if is_valid_string_blautia_hansenii_node_name(node_name) {
            return Ok("STRING Blautia hansenii");
        }

        if is_valid_string_prevotella_copri_node_name(node_name) {
            return Ok("STRING Prevotella copri");
        }

        if is_valid_string_clostridium_methylpentosum_node_name(node_name) {
            return Ok("STRING Clostridium methylpentosum");
        }

        if is_valid_string_liberibacter_asiaticus_psy62_node_name(node_name) {
            return Ok("STRING Liberibacter asiaticus psy62");
        }

        if is_valid_string_helicobacter_canadensis_node_name(node_name) {
            return Ok("STRING Helicobacter canadensis");
        }

        if is_valid_string_helicobacter_cinaedi_ccug18818_node_name(node_name) {
            return Ok("STRING Helicobacter cinaedi CCUG18818");
        }

        if is_valid_string_helicobacter_pullorum_node_name(node_name) {
            return Ok("STRING Helicobacter pullorum");
        }

        if is_valid_string_haloterrigena_turkmenica_node_name(node_name) {
            return Ok("STRING Haloterrigena turkmenica");
        }

        if is_valid_string_actinoplanes_subtropicus_node_name(node_name) {
            return Ok("STRING Actinoplanes subtropicus");
        }

        if is_valid_string_variovorax_paradoxus_s110_node_name(node_name) {
            return Ok("STRING Variovorax paradoxus S110");
        }

        if is_valid_string_lactobacillus_casei_bl23_node_name(node_name) {
            return Ok("STRING Lactobacillus casei BL23");
        }

        if is_valid_string_beta_proteobacterium_cb_node_name(node_name) {
            return Ok("STRING beta proteobacterium CB");
        }

        if is_valid_string_clostridium_arbusti_node_name(node_name) {
            return Ok("STRING Clostridium arbusti");
        }

        if is_valid_string_thioalkalivibrio_sp_akl11_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. AKL11");
        }

        if is_valid_string_thioalkalivibrio_sp_alj24_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. ALJ24");
        }

        if is_valid_string_bacillus_megaterium_qm_b1551_node_name(node_name) {
            return Ok("STRING Bacillus megaterium QM B1551");
        }

        if is_valid_string_treponema_primitia_zas2_node_name(node_name) {
            return Ok("STRING Treponema primitia ZAS2");
        }

        if is_valid_string_treponema_azotonutricium_node_name(node_name) {
            return Ok("STRING Treponema azotonutricium");
        }

        if is_valid_string_holdemania_filiformis_node_name(node_name) {
            return Ok("STRING Holdemania filiformis");
        }

        if is_valid_string_clostridium_celatum_node_name(node_name) {
            return Ok("STRING Clostridium celatum");
        }

        if is_valid_string_neisseria_cinerea_node_name(node_name) {
            return Ok("STRING Neisseria cinerea");
        }

        if is_valid_string_neisseria_flavescens_nrl30031h210_node_name(node_name) {
            return Ok("STRING Neisseria flavescens NRL30031H210");
        }

        if is_valid_string_neisseria_mucosa_atcc25996_node_name(node_name) {
            return Ok("STRING Neisseria mucosa ATCC25996");
        }

        if is_valid_string_neisseria_polysaccharea_node_name(node_name) {
            return Ok("STRING Neisseria polysaccharea");
        }

        if is_valid_string_neisseria_subflava_node_name(node_name) {
            return Ok("STRING Neisseria subflava");
        }

        if is_valid_string_filifactor_alocis_node_name(node_name) {
            return Ok("STRING Filifactor alocis");
        }

        if is_valid_string_gemella_haemolysans_atcc10379_node_name(node_name) {
            return Ok("STRING Gemella haemolysans ATCC10379");
        }

        if is_valid_string_selenomonas_sputigena_node_name(node_name) {
            return Ok("STRING Selenomonas sputigena");
        }

        if is_valid_string_veillonella_dispar_node_name(node_name) {
            return Ok("STRING Veillonella dispar");
        }

        if is_valid_string_eikenella_corrodens_atcc23834_node_name(node_name) {
            return Ok("STRING Eikenella corrodens ATCC23834");
        }

        if is_valid_string_fusobacterium_periodonticum_node_name(node_name) {
            return Ok("STRING Fusobacterium periodonticum");
        }

        if is_valid_string_deinococcus_deserti_node_name(node_name) {
            return Ok("STRING Deinococcus deserti");
        }

        if is_valid_string_bacteroides_coprophilus_node_name(node_name) {
            return Ok("STRING Bacteroides coprophilus");
        }

        if is_valid_string_bifidobacterium_pseudocatenulatum_node_name(node_name) {
            return Ok("STRING Bifidobacterium pseudocatenulatum");
        }

        if is_valid_string_neisseria_sicca_node_name(node_name) {
            return Ok("STRING Neisseria sicca");
        }

        if is_valid_string_hydrogenobaculum_sp_ho_node_name(node_name) {
            return Ok("STRING Hydrogenobaculum sp. HO");
        }

        if is_valid_string_mycobacterium_vulneris_node_name(node_name) {
            return Ok("STRING Mycobacterium vulneris");
        }

        if is_valid_string_natrialba_magadii_node_name(node_name) {
            return Ok("STRING Natrialba magadii");
        }

        if is_valid_string_corynebacterium_aurimucosum_node_name(node_name) {
            return Ok("STRING Corynebacterium aurimucosum");
        }

        if is_valid_string_corynebacterium_glucuronolyticum_node_name(node_name) {
            return Ok("STRING Corynebacterium glucuronolyticum");
        }

        if is_valid_string_mobiluncus_curtisii_atcc43063_node_name(node_name) {
            return Ok("STRING Mobiluncus curtisii ATCC43063");
        }

        if is_valid_string_ferrimonas_balearica_node_name(node_name) {
            return Ok("STRING Ferrimonas balearica");
        }

        if is_valid_string_nostoc_azollae_node_name(node_name) {
            return Ok("STRING Nostoc azollae");
        }

        if is_valid_string_hirschia_maritima_node_name(node_name) {
            return Ok("STRING Hirschia maritima");
        }

        if is_valid_string_ponticaulis_koreensis_node_name(node_name) {
            return Ok("STRING Ponticaulis koreensis");
        }

        if is_valid_string_erysipelotrichaceae_bacterium_5254faa_node_name(node_name) {
            return Ok("STRING Erysipelotrichaceae bacterium 5254FAA");
        }

        if is_valid_string_ruminococcaceae_bacterium_d16_node_name(node_name) {
            return Ok("STRING Ruminococcaceae bacterium D16");
        }

        if is_valid_string_cryptococcus_gattii_node_name(node_name) {
            return Ok("STRING Cryptococcus gattii");
        }

        if is_valid_string_bifidobacterium_animalis_node_name(node_name) {
            return Ok("STRING Bifidobacterium animalis");
        }

        if is_valid_string_dehalogenimonas_lykanthroporepellens_node_name(node_name) {
            return Ok("STRING Dehalogenimonas lykanthroporepellens");
        }

        if is_valid_string_prevotella_bivia_jcvihmp010_node_name(node_name) {
            return Ok("STRING Prevotella bivia JCVIHMP010");
        }

        if is_valid_string_prevotella_melaninogenica_node_name(node_name) {
            return Ok("STRING Prevotella melaninogenica");
        }

        if is_valid_string_porphyromonas_endodontalis_node_name(node_name) {
            return Ok("STRING Porphyromonas endodontalis");
        }

        if is_valid_string_capnocytophaga_sputigena_node_name(node_name) {
            return Ok("STRING Capnocytophaga sputigena");
        }

        if is_valid_string_capnocytophaga_gingivalis_node_name(node_name) {
            return Ok("STRING Capnocytophaga gingivalis");
        }

        if is_valid_string_atopobium_rimae_node_name(node_name) {
            return Ok("STRING Atopobium rimae");
        }

        if is_valid_string_corynebacterium_amycolatum_node_name(node_name) {
            return Ok("STRING Corynebacterium amycolatum");
        }

        if is_valid_string_corynebacterium_matruchotii_node_name(node_name) {
            return Ok("STRING Corynebacterium matruchotii");
        }

        if is_valid_string_enhydrobacter_aerosaccus_node_name(node_name) {
            return Ok("STRING Enhydrobacter aerosaccus");
        }

        if is_valid_string_campylobacter_rectus_node_name(node_name) {
            return Ok("STRING Campylobacter rectus");
        }

        if is_valid_string_campylobacter_showae_rm3277_node_name(node_name) {
            return Ok("STRING Campylobacter showae RM3277");
        }

        if is_valid_string_campylobacter_gracilis_node_name(node_name) {
            return Ok("STRING Campylobacter gracilis");
        }

        if is_valid_string_cobetia_crustatorum_node_name(node_name) {
            return Ok("STRING Cobetia crustatorum");
        }

        if is_valid_string_clostridium_hylemonae_node_name(node_name) {
            return Ok("STRING Clostridium hylemonae");
        }

        if is_valid_string_chlorella_variabilis_node_name(node_name) {
            return Ok("STRING Chlorella variabilis");
        }

        if is_valid_string_thermosediminibacter_oceani_node_name(node_name) {
            return Ok("STRING Thermosediminibacter oceani");
        }

        if is_valid_string_dethiobacter_alkaliphilus_node_name(node_name) {
            return Ok("STRING Dethiobacter alkaliphilus");
        }

        if is_valid_string_zymomonas_mobilis_atcc10988_node_name(node_name) {
            return Ok("STRING Zymomonas mobilis ATCC10988");
        }

        if is_valid_string_galbibacter_marinus_node_name(node_name) {
            return Ok("STRING Galbibacter marinus");
        }

        if is_valid_string_halothiobacillus_neapolitanus_node_name(node_name) {
            return Ok("STRING Halothiobacillus neapolitanus");
        }

        if is_valid_string_desulfonatronospira_thiodismutans_node_name(node_name) {
            return Ok("STRING Desulfonatronospira thiodismutans");
        }

        if is_valid_string_novosphingobium_sp_rr_217_node_name(node_name) {
            return Ok("STRING Novosphingobium sp. Rr 217");
        }

        if is_valid_string_clostridium_sp_d5_node_name(node_name) {
            return Ok("STRING Clostridium sp. D5");
        }

        if is_valid_string_fusobacterium_necrophorum_d12_node_name(node_name) {
            return Ok("STRING Fusobacterium necrophorum D12");
        }

        if is_valid_string_helicobacter_winghamensis_node_name(node_name) {
            return Ok("STRING Helicobacter winghamensis");
        }

        if is_valid_string_oxalobacter_formigenes_hoxbls_node_name(node_name) {
            return Ok("STRING Oxalobacter formigenes HOxBLS");
        }

        if is_valid_string_oxalobacter_formigenes_oxcc13_node_name(node_name) {
            return Ok("STRING Oxalobacter formigenes OXCC13");
        }

        if is_valid_string_lactobacillus_reuteri_dsm20016_node_name(node_name) {
            return Ok("STRING Lactobacillus reuteri DSM20016");
        }

        if is_valid_string_laribacter_hongkongensis_node_name(node_name) {
            return Ok("STRING Laribacter hongkongensis");
        }

        if is_valid_string_mycobacterium_kansasii_atcc12478_node_name(node_name) {
            return Ok("STRING Mycobacterium kansasii ATCC12478");
        }

        if is_valid_string_chryseobacterium_piperi_node_name(node_name) {
            return Ok("STRING Chryseobacterium piperi");
        }

        if is_valid_string_lentibacillus_jeotgali_node_name(node_name) {
            return Ok("STRING Lentibacillus jeotgali");
        }

        if is_valid_string_corynebacterium_doosanense_node_name(node_name) {
            return Ok("STRING Corynebacterium doosanense");
        }

        if is_valid_string_aeromonas_sp_ae122_node_name(node_name) {
            return Ok("STRING Aeromonas sp. AE122");
        }

        if is_valid_string_actinomadura_flavalba_node_name(node_name) {
            return Ok("STRING Actinomadura flavalba");
        }

        if is_valid_string_anaerococcus_hydrogenalis_node_name(node_name) {
            return Ok("STRING Anaerococcus hydrogenalis");
        }

        if is_valid_string_bifidobacterium_gallicum_node_name(node_name) {
            return Ok("STRING Bifidobacterium gallicum");
        }

        if is_valid_string_dickeya_zeae_node_name(node_name) {
            return Ok("STRING Dickeya zeae");
        }

        if is_valid_string_pectobacterium_carotovorum_pc1_node_name(node_name) {
            return Ok("STRING Pectobacterium carotovorum PC1");
        }

        if is_valid_string_pectobacterium_wasabiae_node_name(node_name) {
            return Ok("STRING Pectobacterium wasabiae");
        }

        if is_valid_string_amphibacillus_jilinensis_node_name(node_name) {
            return Ok("STRING Amphibacillus jilinensis");
        }

        if is_valid_string_kyrpidia_tusciae_node_name(node_name) {
            return Ok("STRING Kyrpidia tusciae");
        }

        if is_valid_string_actinomyces_viscosus_node_name(node_name) {
            return Ok("STRING Actinomyces viscosus");
        }

        if is_valid_string_gemella_haemolysans_m341_node_name(node_name) {
            return Ok("STRING Gemella haemolysans M341");
        }

        if is_valid_string_gemella_morbillorum_node_name(node_name) {
            return Ok("STRING Gemella morbillorum");
        }

        if is_valid_string_gemella_sanguinis_node_name(node_name) {
            return Ok("STRING Gemella sanguinis");
        }

        if is_valid_string_prevotella_oris_c735_node_name(node_name) {
            return Ok("STRING Prevotella oris C735");
        }

        if is_valid_string_prevotella_sp_c561_node_name(node_name) {
            return Ok("STRING Prevotella sp. C561");
        }

        if is_valid_string_streptococcus_sp_m143_node_name(node_name) {
            return Ok("STRING Streptococcus sp. M143");
        }

        if is_valid_string_streptococcus_sp_m334_node_name(node_name) {
            return Ok("STRING Streptococcus sp. M334");
        }

        if is_valid_string_sulfurimonas_autotrophica_node_name(node_name) {
            return Ok("STRING Sulfurimonas autotrophica");
        }

        if is_valid_string_bilophila_wadsworthia_316_node_name(node_name) {
            return Ok("STRING Bilophila wadsworthia 316");
        }

        if is_valid_string_geoglobus_acetivorans_node_name(node_name) {
            return Ok("STRING Geoglobus acetivorans");
        }

        if is_valid_string_brachyspira_hyodysenteriae_node_name(node_name) {
            return Ok("STRING Brachyspira hyodysenteriae");
        }

        if is_valid_string_luminiphilus_syltensis_node_name(node_name) {
            return Ok("STRING Luminiphilus syltensis");
        }

        if is_valid_string_ureaplasma_urealyticum_node_name(node_name) {
            return Ok("STRING Ureaplasma urealyticum");
        }

        if is_valid_string_enterococcus_gallinarum_node_name(node_name) {
            return Ok("STRING Enterococcus gallinarum");
        }

        if is_valid_string_enterococcus_casseliflavus_node_name(node_name) {
            return Ok("STRING Enterococcus casseliflavus");
        }

        if is_valid_string_enterococcus_faecium_c68_node_name(node_name) {
            return Ok("STRING Enterococcus faecium C68");
        }

        if is_valid_string_streptomyces_ghanaensis_node_name(node_name) {
            return Ok("STRING Streptomyces ghanaensis");
        }

        if is_valid_string_gammaproteobacteria_bacterium_nor53_node_name(node_name) {
            return Ok("STRING Gammaproteobacteria bacterium NOR53");
        }

        if is_valid_string_bifidobacterium_catenulatum_node_name(node_name) {
            return Ok("STRING Bifidobacterium catenulatum");
        }

        if is_valid_string_metarhizium_robertsii_node_name(node_name) {
            return Ok("STRING Metarhizium robertsii");
        }

        if is_valid_string_lactobacillus_rhamnosus_node_name(node_name) {
            return Ok("STRING Lactobacillus rhamnosus");
        }

        if is_valid_string_bordetella_pertussis_18323_node_name(node_name) {
            return Ok("STRING Bordetella pertussis 18323");
        }

        if is_valid_string_dickeya_sp_ncppb569_node_name(node_name) {
            return Ok("STRING Dickeya sp. NCPPB569");
        }

        if is_valid_string_acidaminococcus_intestini_node_name(node_name) {
            return Ok("STRING Acidaminococcus intestini");
        }

        if is_valid_string_serratia_symbiotica_node_name(node_name) {
            return Ok("STRING Serratia symbiotica");
        }

        if is_valid_string_nocardiopsis_potens_node_name(node_name) {
            return Ok("STRING Nocardiopsis potens");
        }

        if is_valid_string_wolbachia_sp_culex_node_name(node_name) {
            return Ok("STRING Wolbachia sp. Culex");
        }

        if is_valid_string_fodinicurvata_sediminis_node_name(node_name) {
            return Ok("STRING Fodinicurvata sediminis");
        }

        if is_valid_string_fodinicurvata_fenggangensis_node_name(node_name) {
            return Ok("STRING Fodinicurvata fenggangensis");
        }

        if is_valid_string_sedimentitalea_nanhaiensis_node_name(node_name) {
            return Ok("STRING Sedimentitalea nanhaiensis");
        }

        if is_valid_string_hamiltonella_defensa_node_name(node_name) {
            return Ok("STRING Hamiltonella defensa");
        }

        if is_valid_string_allochromatium_vinosum_node_name(node_name) {
            return Ok("STRING Allochromatium vinosum");
        }

        if is_valid_string_vulcanisaeta_distributa_node_name(node_name) {
            return Ok("STRING Vulcanisaeta distributa");
        }

        if is_valid_string_halanaerobium_praevalens_node_name(node_name) {
            return Ok("STRING Halanaerobium praevalens");
        }

        if is_valid_string_arcobacter_nitrofigilis_node_name(node_name) {
            return Ok("STRING Arcobacter nitrofigilis");
        }

        if is_valid_string_ilyobacter_polytropus_node_name(node_name) {
            return Ok("STRING Ilyobacter polytropus");
        }

        if is_valid_string_archaeoglobus_profundus_node_name(node_name) {
            return Ok("STRING Archaeoglobus profundus");
        }

        if is_valid_string_aminobacterium_colombiense_node_name(node_name) {
            return Ok("STRING Aminobacterium colombiense");
        }

        if is_valid_string_clostridium_cellulovorans_node_name(node_name) {
            return Ok("STRING Clostridium cellulovorans");
        }

        if is_valid_string_methanocaldococcus_infernus_node_name(node_name) {
            return Ok("STRING Methanocaldococcus infernus");
        }

        if is_valid_string_methanocaldococcus_fervens_node_name(node_name) {
            return Ok("STRING Methanocaldococcus fervens");
        }

        if is_valid_string_asticcacaulis_excentricus_node_name(node_name) {
            return Ok("STRING Asticcacaulis excentricus");
        }

        if is_valid_string_desulfovibrio_magneticus_rs1_node_name(node_name) {
            return Ok("STRING Desulfovibrio magneticus RS1");
        }

        if is_valid_string_spirochaeta_smaragdinae_node_name(node_name) {
            return Ok("STRING Spirochaeta smaragdinae");
        }

        if is_valid_string_francisella_sp_tx077308_node_name(node_name) {
            return Ok("STRING Francisella sp. TX077308");
        }

        if is_valid_string_acetohalobium_arabaticum_node_name(node_name) {
            return Ok("STRING Acetohalobium arabaticum");
        }

        if is_valid_string_bacillus_gaemokensis_node_name(node_name) {
            return Ok("STRING Bacillus gaemokensis");
        }

        if is_valid_string_bacillus_manliponensis_node_name(node_name) {
            return Ok("STRING Bacillus manliponensis");
        }

        if is_valid_string_anaplasma_centrale_node_name(node_name) {
            return Ok("STRING Anaplasma centrale");
        }

        if is_valid_string_halomonas_zhanjiangensis_node_name(node_name) {
            return Ok("STRING Halomonas zhanjiangensis");
        }

        if is_valid_string_isosphaera_pallida_node_name(node_name) {
            return Ok("STRING Isosphaera pallida");
        }

        if is_valid_string_acinetobacter_sp_ruh2624_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. RUH2624");
        }

        if is_valid_string_acinetobacter_johnsonii_node_name(node_name) {
            return Ok("STRING Acinetobacter johnsonii");
        }

        if is_valid_string_acinetobacter_lwoffii_sh145_node_name(node_name) {
            return Ok("STRING Acinetobacter lwoffii SH145");
        }

        if is_valid_string_acinetobacter_radioresistens_node_name(node_name) {
            return Ok("STRING Acinetobacter radioresistens");
        }

        if is_valid_string_bacteroidetes_f0058_node_name(node_name) {
            return Ok("STRING Bacteroidetes F0058");
        }

        if is_valid_string_lachnospiraceae_f0167_node_name(node_name) {
            return Ok("STRING Lachnospiraceae F0167");
        }

        if is_valid_string_lactobacillus_coleohominis_node_name(node_name) {
            return Ok("STRING Lactobacillus coleohominis");
        }

        if is_valid_string_lactobacillus_jensenii_1153chn_node_name(node_name) {
            return Ok("STRING Lactobacillus jensenii 1153CHN");
        }

        if is_valid_string_lactobacillus_jensenii_272chn_node_name(node_name) {
            return Ok("STRING Lactobacillus jensenii 272CHN");
        }

        if is_valid_string_peptoniphilus_sp_f0131_node_name(node_name) {
            return Ok("STRING Peptoniphilus sp. F0131");
        }

        if is_valid_string_prevotella_sp_f0108_node_name(node_name) {
            return Ok("STRING Prevotella sp. F0108");
        }

        if is_valid_string_vibrio_tasmaniensis_node_name(node_name) {
            return Ok("STRING Vibrio tasmaniensis");
        }

        if is_valid_string_methanocaldococcus_vulcanius_node_name(node_name) {
            return Ok("STRING Methanocaldococcus vulcanius");
        }

        if is_valid_string_zymomonas_mobilis_pomaceae_node_name(node_name) {
            return Ok("STRING Zymomonas mobilis pomaceae");
        }

        if is_valid_string_dickeya_dadantii_ech703_node_name(node_name) {
            return Ok("STRING Dickeya dadantii Ech703");
        }

        if is_valid_string_thermoanaerobacterium_thermosaccharolyticum_node_name(node_name) {
            return Ok("STRING Thermoanaerobacterium thermosaccharolyticum");
        }

        if is_valid_string_thermoanaerobacter_italicus_node_name(node_name) {
            return Ok("STRING Thermoanaerobacter italicus");
        }

        if is_valid_string_sideroxydans_lithotrophicus_node_name(node_name) {
            return Ok("STRING Sideroxydans lithotrophicus");
        }

        if is_valid_string_thermovirga_lienii_node_name(node_name) {
            return Ok("STRING Thermovirga lienii");
        }

        if is_valid_string_hirschia_baltica_node_name(node_name) {
            return Ok("STRING Hirschia baltica");
        }

        if is_valid_string_rubidibacter_lacunae_node_name(node_name) {
            return Ok("STRING Rubidibacter lacunae");
        }

        if is_valid_string_methylovorus_glucosetrophus_node_name(node_name) {
            return Ok("STRING Methylovorus glucosetrophus");
        }

        if is_valid_string_hyphomicrobium_denitrificans_atcc51888_node_name(node_name) {
            return Ok("STRING Hyphomicrobium denitrificans ATCC51888");
        }

        if is_valid_string_methylotenera_mobilis_jlw8_node_name(node_name) {
            return Ok("STRING Methylotenera mobilis JLW8");
        }

        if is_valid_string_coraliomargarita_akajimensis_node_name(node_name) {
            return Ok("STRING Coraliomargarita akajimensis");
        }

        if is_valid_string_aminomonas_paucivorans_node_name(node_name) {
            return Ok("STRING Aminomonas paucivorans");
        }

        if is_valid_string_mobiluncus_curtisii_atcc35241_node_name(node_name) {
            return Ok("STRING Mobiluncus curtisii ATCC35241");
        }

        if is_valid_string_mobiluncus_mulieris_atcc35243_node_name(node_name) {
            return Ok("STRING Mobiluncus mulieris ATCC35243");
        }

        if is_valid_string_streptococcus_mitis_sk321_node_name(node_name) {
            return Ok("STRING Streptococcus mitis SK321");
        }

        if is_valid_string_streptococcus_mitis_sk564_node_name(node_name) {
            return Ok("STRING Streptococcus mitis SK564");
        }

        if is_valid_string_streptococcus_mitis_sk597_node_name(node_name) {
            return Ok("STRING Streptococcus mitis SK597");
        }

        if is_valid_string_roseburia_hominis_node_name(node_name) {
            return Ok("STRING Roseburia hominis");
        }

        if is_valid_string_synechococcus_sp_kordi49_node_name(node_name) {
            return Ok("STRING Synechococcus sp. KORDI49");
        }

        if is_valid_string_synechococcus_sp_kordi52_node_name(node_name) {
            return Ok("STRING Synechococcus sp. KORDI52");
        }

        if is_valid_string_oribacterium_sinus_node_name(node_name) {
            return Ok("STRING Oribacterium sinus");
        }

        if is_valid_string_prevotella_bergensis_node_name(node_name) {
            return Ok("STRING Prevotella bergensis");
        }

        if is_valid_string_selenomonas_noxia_atcc43541_node_name(node_name) {
            return Ok("STRING Selenomonas noxia ATCC43541");
        }

        if is_valid_string_weissella_paramesenteroides_node_name(node_name) {
            return Ok("STRING Weissella paramesenteroides");
        }

        if is_valid_string_lactobacillus_amylolyticus_node_name(node_name) {
            return Ok("STRING Lactobacillus amylolyticus");
        }

        if is_valid_string_corynebacterium_genitalium_node_name(node_name) {
            return Ok("STRING Corynebacterium genitalium");
        }

        if is_valid_string_brevibacterium_mcbrellneri_node_name(node_name) {
            return Ok("STRING Brevibacterium mcbrellneri");
        }

        if is_valid_string_aeromicrobium_marinum_node_name(node_name) {
            return Ok("STRING Aeromicrobium marinum");
        }

        if is_valid_string_bacteroides_sp_d20_node_name(node_name) {
            return Ok("STRING Bacteroides sp. D20");
        }

        if is_valid_string_oceanobacillus_manasiensis_node_name(node_name) {
            return Ok("STRING Oceanobacillus manasiensis");
        }

        if is_valid_string_terribacillus_aidingensis_node_name(node_name) {
            return Ok("STRING Terribacillus aidingensis");
        }

        if is_valid_string_pseudomonas_chlororaphis_node_name(node_name) {
            return Ok("STRING Pseudomonas chlororaphis");
        }

        if is_valid_string_clostridium_papyrosolvens_node_name(node_name) {
            return Ok("STRING Clostridium papyrosolvens");
        }

        if is_valid_string_rhizophagus_irregularis_node_name(node_name) {
            return Ok("STRING Rhizophagus irregularis");
        }

        if is_valid_string_kazachstania_naganishii_node_name(node_name) {
            return Ok("STRING Kazachstania naganishii");
        }

        if is_valid_string_brevundimonas_naejangsanensis_node_name(node_name) {
            return Ok("STRING Brevundimonas naejangsanensis");
        }

        if is_valid_string_desulfurivibrio_alkaliphilus_node_name(node_name) {
            return Ok("STRING Desulfurivibrio alkaliphilus");
        }

        if is_valid_string_alteromonas_australica_node_name(node_name) {
            return Ok("STRING Alteromonas australica");
        }

        if is_valid_string_ferroglobus_placidus_node_name(node_name) {
            return Ok("STRING Ferroglobus placidus");
        }

        if is_valid_string_dickeya_dadantii_ech586_node_name(node_name) {
            return Ok("STRING Dickeya dadantii Ech586");
        }

        if is_valid_string_cellulomonas_fimi_node_name(node_name) {
            return Ok("STRING Cellulomonas fimi");
        }

        if is_valid_string_acidaminococcus_fermentans_node_name(node_name) {
            return Ok("STRING Acidaminococcus fermentans");
        }

        if is_valid_string_staphylothermus_hellenicus_node_name(node_name) {
            return Ok("STRING Staphylothermus hellenicus");
        }

        if is_valid_string_actinobacillus_minor_202_node_name(node_name) {
            return Ok("STRING Actinobacillus minor 202");
        }

        if is_valid_string_streptomyces_sp_spb78_node_name(node_name) {
            return Ok("STRING Streptomyces sp. SPB78");
        }

        if is_valid_string_streptomyces_sp_aa4_node_name(node_name) {
            return Ok("STRING Streptomyces sp. AA4");
        }

        if is_valid_string_streptomyces_viridochromogenes_node_name(node_name) {
            return Ok("STRING Streptomyces viridochromogenes");
        }

        if is_valid_string_streptomyces_pratensis_node_name(node_name) {
            return Ok("STRING Streptomyces pratensis");
        }

        if is_valid_string_abiotrophia_defectiva_node_name(node_name) {
            return Ok("STRING Abiotrophia defectiva");
        }

        if is_valid_string_anaerobaculum_hydrogeniformans_node_name(node_name) {
            return Ok("STRING Anaerobaculum hydrogeniformans");
        }

        if is_valid_string_catonella_morbi_node_name(node_name) {
            return Ok("STRING Catonella morbi");
        }

        if is_valid_string_clostridium_botulinum_d_1873_node_name(node_name) {
            return Ok("STRING Clostridium botulinum D 1873");
        }

        if is_valid_string_dialister_invisus_node_name(node_name) {
            return Ok("STRING Dialister invisus");
        }

        if is_valid_string_nonlabens_dokdonensis_node_name(node_name) {
            return Ok("STRING Nonlabens dokdonensis");
        }

        if is_valid_string_eubacterium_saphenum_node_name(node_name) {
            return Ok("STRING Eubacterium saphenum");
        }

        if is_valid_string_pantoea_sp_at9b_node_name(node_name) {
            return Ok("STRING Pantoea sp. At9b");
        }

        if is_valid_string_pantoea_sp_sc1_node_name(node_name) {
            return Ok("STRING Pantoea sp. Sc1");
        }

        if is_valid_string_thermococcus_gammatolerans_node_name(node_name) {
            return Ok("STRING Thermococcus gammatolerans");
        }

        if is_valid_string_methanoregula_formicica_node_name(node_name) {
            return Ok("STRING Methanoregula formicica");
        }

        if is_valid_string_cellvibrio_gilvus_node_name(node_name) {
            return Ok("STRING Cellvibrio gilvus");
        }

        if is_valid_string_rhodopirellula_sp_swk7_node_name(node_name) {
            return Ok("STRING Rhodopirellula sp. SWK7");
        }

        if is_valid_string_tolumonas_auensis_node_name(node_name) {
            return Ok("STRING Tolumonas auensis");
        }

        if is_valid_string_methylosinus_trichosporium_node_name(node_name) {
            return Ok("STRING Methylosinus trichosporium");
        }

        if is_valid_string_variovorax_paradoxus_eps_node_name(node_name) {
            return Ok("STRING Variovorax paradoxus EPS");
        }

        if is_valid_string_arthrobacter_sp_a3_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. A3");
        }

        if is_valid_string_desulfovibrio_fructosivorans_node_name(node_name) {
            return Ok("STRING Desulfovibrio fructosivorans");
        }

        if is_valid_string_desulfovibrio_sp_u5l_node_name(node_name) {
            return Ok("STRING Desulfovibrio sp. U5L");
        }

        if is_valid_string_alicycliphilus_denitrificans_bc_node_name(node_name) {
            return Ok("STRING Alicycliphilus denitrificans BC");
        }

        if is_valid_string_alicycliphilus_denitrificans_k601_node_name(node_name) {
            return Ok("STRING Alicycliphilus denitrificans K601");
        }

        if is_valid_string_peptostreptococcus_stomatis_node_name(node_name) {
            return Ok("STRING Peptostreptococcus stomatis");
        }

        if is_valid_string_staphylococcus_warneri_l37603_node_name(node_name) {
            return Ok("STRING Staphylococcus warneri L37603");
        }

        if is_valid_string_neisseria_flavescens_sk114_node_name(node_name) {
            return Ok("STRING Neisseria flavescens SK114");
        }

        if is_valid_string_leptotrichia_goodfellowii_node_name(node_name) {
            return Ok("STRING Leptotrichia goodfellowii");
        }

        if is_valid_string_treponema_vincentii_node_name(node_name) {
            return Ok("STRING Treponema vincentii");
        }

        if is_valid_string_porphyromonas_uenonis_603_node_name(node_name) {
            return Ok("STRING Porphyromonas uenonis 603");
        }

        if is_valid_string_mobiluncus_mulieris_281_node_name(node_name) {
            return Ok("STRING Mobiluncus mulieris 281");
        }

        if is_valid_string_peptostreptococcus_anaerobius_653l_node_name(node_name) {
            return Ok("STRING Peptostreptococcus anaerobius 653L");
        }

        if is_valid_string_peptoniphilus_lacrimalis_node_name(node_name) {
            return Ok("STRING Peptoniphilus lacrimalis");
        }

        if is_valid_string_brenneria_sp_enid312_node_name(node_name) {
            return Ok("STRING Brenneria sp. EniD312");
        }

        if is_valid_string_nautilia_profundicola_node_name(node_name) {
            return Ok("STRING Nautilia profundicola");
        }

        if is_valid_string_blattabacterium_sp_bplan_node_name(node_name) {
            return Ok("STRING Blattabacterium sp. BPLAN");
        }

        if is_valid_string_meiothermus_rufus_node_name(node_name) {
            return Ok("STRING Meiothermus rufus");
        }

        if is_valid_string_thermococcus_sibiricus_node_name(node_name) {
            return Ok("STRING Thermococcus sibiricus");
        }

        if is_valid_string_caldicellulosiruptor_obsidiansis_node_name(node_name) {
            return Ok("STRING Caldicellulosiruptor obsidiansis");
        }

        if is_valid_string_oribacterium_sp_f0262_node_name(node_name) {
            return Ok("STRING Oribacterium sp. F0262");
        }

        if is_valid_string_hydrogenobacter_thermophilus_node_name(node_name) {
            return Ok("STRING Hydrogenobacter thermophilus");
        }

        if is_valid_string_clostridium_saccharolyticum_wm1_node_name(node_name) {
            return Ok("STRING Clostridium saccharolyticum WM1");
        }

        if is_valid_string_helicobacter_bilis_node_name(node_name) {
            return Ok("STRING Helicobacter bilis");
        }

        if is_valid_string_rhodoferax_saidenbachensis_node_name(node_name) {
            return Ok("STRING Rhodoferax saidenbachensis");
        }

        if is_valid_string_arenibacter_algicola_node_name(node_name) {
            return Ok("STRING Arenibacter algicola");
        }

        if is_valid_string_vibrio_splendidus_node_name(node_name) {
            return Ok("STRING Vibrio splendidus");
        }

        if is_valid_string_prevotella_sp_f0295_node_name(node_name) {
            return Ok("STRING Prevotella sp. F0295");
        }

        if is_valid_string_aquimarina_agarilytica_node_name(node_name) {
            return Ok("STRING Aquimarina agarilytica");
        }

        if is_valid_string_paenibacillus_sp_d14_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. D14");
        }

        if is_valid_string_roseburia_inulinivorans_node_name(node_name) {
            return Ok("STRING Roseburia inulinivorans");
        }

        if is_valid_string_methylocystis_sp_atcc49242_node_name(node_name) {
            return Ok("STRING Methylocystis sp. ATCC49242");
        }

        if is_valid_string_granulicatella_elegans_node_name(node_name) {
            return Ok("STRING Granulicatella elegans");
        }

        if is_valid_string_burkholderia_glumae_node_name(node_name) {
            return Ok("STRING Burkholderia glumae");
        }

        if is_valid_string_prevotella_tannerae_node_name(node_name) {
            return Ok("STRING Prevotella tannerae");
        }

        if is_valid_string_shuttleworthia_satelles_node_name(node_name) {
            return Ok("STRING Shuttleworthia satelles");
        }

        if is_valid_string_marinobacter_nanhaiticus_node_name(node_name) {
            return Ok("STRING Marinobacter nanhaiticus");
        }

        if is_valid_string_phascolarctobacterium_succinatutens_node_name(node_name) {
            return Ok("STRING Phascolarctobacterium succinatutens");
        }

        if is_valid_string_sphingobium_sp_syk6_node_name(node_name) {
            return Ok("STRING Sphingobium sp. SYK6");
        }

        if is_valid_string_pseudomonas_syringae_maculicola_node_name(node_name) {
            return Ok("STRING Pseudomonas syringae maculicola");
        }

        if is_valid_string_staphylococcus_hominis_sk119_node_name(node_name) {
            return Ok("STRING Staphylococcus hominis SK119");
        }

        if is_valid_string_sphingomonas_sp_mn802worker_node_name(node_name) {
            return Ok("STRING Sphingomonas sp. Mn802worker");
        }

        if is_valid_string_shimwellia_blattae_node_name(node_name) {
            return Ok("STRING Shimwellia blattae");
        }

        if is_valid_string_thiorhodovibrio_sp_970_node_name(node_name) {
            return Ok("STRING Thiorhodovibrio sp. 970");
        }

        if is_valid_string_lutibaculum_baratangense_node_name(node_name) {
            return Ok("STRING Lutibaculum baratangense");
        }

        if is_valid_string_clostridium_butyricum_node_name(node_name) {
            return Ok("STRING Clostridium butyricum");
        }

        if is_valid_string_caldicellulosiruptor_hydrothermalis_node_name(node_name) {
            return Ok("STRING Caldicellulosiruptor hydrothermalis");
        }

        if is_valid_string_caldicellulosiruptor_kristjanssonii_node_name(node_name) {
            return Ok("STRING Caldicellulosiruptor kristjanssonii");
        }

        if is_valid_string_caldicellulosiruptor_owensensis_node_name(node_name) {
            return Ok("STRING Caldicellulosiruptor owensensis");
        }

        if is_valid_string_thalassobium_sp_r2a62_node_name(node_name) {
            return Ok("STRING Thalassobium sp. R2A62");
        }

        if is_valid_string_olsenella_uli_node_name(node_name) {
            return Ok("STRING Olsenella uli");
        }

        if is_valid_string_thermosphaera_aggregans_node_name(node_name) {
            return Ok("STRING Thermosphaera aggregans");
        }

        if is_valid_string_brevundimonas_subvibrioides_node_name(node_name) {
            return Ok("STRING Brevundimonas subvibrioides");
        }

        if is_valid_string_eubacterium_cellulosolvens_6_node_name(node_name) {
            return Ok("STRING Eubacterium cellulosolvens 6");
        }

        if is_valid_string_aggregatibacter_aphrophilus_node_name(node_name) {
            return Ok("STRING Aggregatibacter aphrophilus");
        }

        if is_valid_string_komagataeibacter_medellinensis_node_name(node_name) {
            return Ok("STRING Komagataeibacter medellinensis");
        }

        if is_valid_string_acetobacter_pasteurianus_ifo328301_node_name(node_name) {
            return Ok("STRING Acetobacter pasteurianus IFO328301");
        }

        if is_valid_string_haloarcula_hispanica_node_name(node_name) {
            return Ok("STRING Haloarcula hispanica");
        }

        if is_valid_string_methanobrevibacter_ruminantium_node_name(node_name) {
            return Ok("STRING Methanobrevibacter ruminantium");
        }

        if is_valid_string_erwinia_pyrifoliae_node_name(node_name) {
            return Ok("STRING Erwinia pyrifoliae");
        }

        if is_valid_string_erwinia_billingiae_node_name(node_name) {
            return Ok("STRING Erwinia billingiae");
        }

        if is_valid_string_bartonella_grahamii_node_name(node_name) {
            return Ok("STRING Bartonella grahamii");
        }

        if is_valid_string_geobacillus_thermoglucosidasius_node_name(node_name) {
            return Ok("STRING Geobacillus thermoglucosidasius");
        }

        if is_valid_string_leptotrichia_hofstadii_node_name(node_name) {
            return Ok("STRING Leptotrichia hofstadii");
        }

        if is_valid_string_thermincola_potens_node_name(node_name) {
            return Ok("STRING Thermincola potens");
        }

        if is_valid_string_acidithiobacillus_caldus_node_name(node_name) {
            return Ok("STRING Acidithiobacillus caldus");
        }

        if is_valid_string_acidithiobacillus_thiooxidans_node_name(node_name) {
            return Ok("STRING Acidithiobacillus thiooxidans");
        }

        if is_valid_string_shewanella_violacea_node_name(node_name) {
            return Ok("STRING Shewanella violacea");
        }

        if is_valid_string_citrobacter_rodentium_node_name(node_name) {
            return Ok("STRING Citrobacter rodentium");
        }

        if is_valid_string_actinobacillus_minor_nm305_node_name(node_name) {
            return Ok("STRING Actinobacillus minor NM305");
        }

        if is_valid_string_granulicatella_adiacens_node_name(node_name) {
            return Ok("STRING Granulicatella adiacens");
        }

        if is_valid_string_selenomonas_flueggei_node_name(node_name) {
            return Ok("STRING Selenomonas flueggei");
        }

        if is_valid_string_thermocrinis_albus_node_name(node_name) {
            return Ok("STRING Thermocrinis albus");
        }

        if is_valid_string_acidobacteria_bacterium_kbs146_node_name(node_name) {
            return Ok("STRING Acidobacteria bacterium KBS146");
        }

        if is_valid_string_deferribacter_desulfuricans_node_name(node_name) {
            return Ok("STRING Deferribacter desulfuricans");
        }

        if is_valid_string_starkeya_novella_node_name(node_name) {
            return Ok("STRING Starkeya novella");
        }

        if is_valid_string_dechlorosoma_suillum_node_name(node_name) {
            return Ok("STRING Dechlorosoma suillum");
        }

        if is_valid_string_segniliparus_rotundus_node_name(node_name) {
            return Ok("STRING Segniliparus rotundus");
        }

        if is_valid_string_burkholderia_sp_ccge1001_node_name(node_name) {
            return Ok("STRING Burkholderia sp. CCGE1001");
        }

        if is_valid_string_burkholderia_sp_ccge1002_node_name(node_name) {
            return Ok("STRING Burkholderia sp. CCGE1002");
        }

        if is_valid_string_burkholderia_sp_ccge1003_node_name(node_name) {
            return Ok("STRING Burkholderia sp. CCGE1003");
        }

        if is_valid_string_enterobacter_asburiae_lf7a_node_name(node_name) {
            return Ok("STRING Enterobacter asburiae LF7a");
        }

        if is_valid_string_clostridium_sp_dlviii_node_name(node_name) {
            return Ok("STRING Clostridium sp. DLVIII");
        }

        if is_valid_string_ruminococcus_flavefaciens_fd1_node_name(node_name) {
            return Ok("STRING Ruminococcus flavefaciens FD1");
        }

        if is_valid_string_capnocytophaga_granulosa_node_name(node_name) {
            return Ok("STRING Capnocytophaga granulosa");
        }

        if is_valid_string_scardovia_inopinata_node_name(node_name) {
            return Ok("STRING Scardovia inopinata");
        }

        if is_valid_string_simonsiella_muelleri_node_name(node_name) {
            return Ok("STRING Simonsiella muelleri");
        }

        if is_valid_string_neisseria_sp_f0314_node_name(node_name) {
            return Ok("STRING Neisseria sp. F0314");
        }

        if is_valid_string_desulfovibrio_desulfuricans_nd132_node_name(node_name) {
            return Ok("STRING Desulfovibrio desulfuricans ND132");
        }

        if is_valid_string_cyclobacterium_qasimii_node_name(node_name) {
            return Ok("STRING Cyclobacterium qasimii");
        }

        if is_valid_string_winogradskyella_psychrotolerans_node_name(node_name) {
            return Ok("STRING Winogradskyella psychrotolerans");
        }

        if is_valid_string_tatumella_morbirosei_node_name(node_name) {
            return Ok("STRING Tatumella morbirosei");
        }

        if is_valid_string_clostridium_lentocellum_node_name(node_name) {
            return Ok("STRING Clostridium lentocellum");
        }

        if is_valid_string_microchaete_sp_pcc7126_node_name(node_name) {
            return Ok("STRING Microchaete sp. PCC7126");
        }

        if is_valid_string_desulfovibrio_aespoeensis_node_name(node_name) {
            return Ok("STRING Desulfovibrio aespoeensis");
        }

        if is_valid_string_syntrophothermus_lipocalidus_node_name(node_name) {
            return Ok("STRING Syntrophothermus lipocalidus");
        }

        if is_valid_string_marivirga_tractuosa_node_name(node_name) {
            return Ok("STRING Marivirga tractuosa");
        }

        if is_valid_string_silicibacter_sp_trichch4b_node_name(node_name) {
            return Ok("STRING Silicibacter sp. TrichCH4B");
        }

        if is_valid_string_silicibacter_lacuscaerulensis_node_name(node_name) {
            return Ok("STRING Silicibacter lacuscaerulensis");
        }

        if is_valid_string_methanocaldococcus_sp_fs40622_node_name(node_name) {
            return Ok("STRING Methanocaldococcus sp. FS40622");
        }

        if is_valid_string_desulfarculus_baarsii_node_name(node_name) {
            return Ok("STRING Desulfarculus baarsii");
        }

        if is_valid_string_micromonospora_aurantiaca_node_name(node_name) {
            return Ok("STRING Micromonospora aurantiaca");
        }

        if is_valid_string_arcanobacterium_haemolyticum_node_name(node_name) {
            return Ok("STRING Arcanobacterium haemolyticum");
        }

        if is_valid_string_gordonia_neofelifaecis_node_name(node_name) {
            return Ok("STRING Gordonia neofelifaecis");
        }

        if is_valid_string_pseudomonas_stutzeri_rch2_node_name(node_name) {
            return Ok("STRING Pseudomonas stutzeri RCH2");
        }

        if is_valid_string_thermaerobacter_marianensis_node_name(node_name) {
            return Ok("STRING Thermaerobacter marianensis");
        }

        if is_valid_string_desulfovibrio_sp_fw1012b_node_name(node_name) {
            return Ok("STRING Desulfovibrio sp. FW1012B");
        }

        if is_valid_string_corynebacterium_kroppenstedtii_node_name(node_name) {
            return Ok("STRING Corynebacterium kroppenstedtii");
        }

        if is_valid_string_streptomyces_sp_e14_node_name(node_name) {
            return Ok("STRING Streptomyces sp. e14");
        }

        if is_valid_string_jonquetella_anthropi_e333e1_node_name(node_name) {
            return Ok("STRING Jonquetella anthropi E333E1");
        }

        if is_valid_string_syntrophobotulus_glycolicus_node_name(node_name) {
            return Ok("STRING Syntrophobotulus glycolicus");
        }

        if is_valid_string_desulfosporosinus_acidiphilus_node_name(node_name) {
            return Ok("STRING Desulfosporosinus acidiphilus");
        }

        if is_valid_string_methanothermococcus_okinawensis_node_name(node_name) {
            return Ok("STRING Methanothermococcus okinawensis");
        }

        if is_valid_string_rhodomicrobium_vannielii_node_name(node_name) {
            return Ok("STRING Rhodomicrobium vannielii");
        }

        if is_valid_string_methylobacterium_sp_mb200_node_name(node_name) {
            return Ok("STRING Methylobacterium sp. MB200");
        }

        if is_valid_string_thermovibrio_ammonificans_node_name(node_name) {
            return Ok("STRING Thermovibrio ammonificans");
        }

        if is_valid_string_leadbetterella_byssophila_node_name(node_name) {
            return Ok("STRING Leadbetterella byssophila");
        }

        if is_valid_string_truepera_radiovictrix_node_name(node_name) {
            return Ok("STRING Truepera radiovictrix");
        }

        if is_valid_string_bacillus_cellulosilyticus_node_name(node_name) {
            return Ok("STRING Bacillus cellulosilyticus");
        }

        if is_valid_string_actinomyces_sp_f0332_node_name(node_name) {
            return Ok("STRING Actinomyces sp. F0332");
        }

        if is_valid_string_aneurinibacillus_aneurinilyticus_node_name(node_name) {
            return Ok("STRING Aneurinibacillus aneurinilyticus");
        }

        if is_valid_string_corynebacterium_ammoniagenes_node_name(node_name) {
            return Ok("STRING Corynebacterium ammoniagenes");
        }

        if is_valid_string_prevotella_veroralis_f0319_node_name(node_name) {
            return Ok("STRING Prevotella veroralis F0319");
        }

        if is_valid_string_slackia_exigua_node_name(node_name) {
            return Ok("STRING Slackia exigua");
        }

        if is_valid_string_actinoplanes_sp_n902109_node_name(node_name) {
            return Ok("STRING Actinoplanes sp. N902109");
        }

        if is_valid_string_erysipelothrix_rhusiopathiae_node_name(node_name) {
            return Ok("STRING Erysipelothrix rhusiopathiae");
        }

        if is_valid_string_rhodopseudomonas_palustris_dx1_node_name(node_name) {
            return Ok("STRING Rhodopseudomonas palustris DX1");
        }

        if is_valid_string_streptomyces_violaceusniger_node_name(node_name) {
            return Ok("STRING Streptomyces violaceusniger");
        }

        if is_valid_string_actinomyces_sp_f0330_node_name(node_name) {
            return Ok("STRING Actinomyces sp. F0330");
        }

        if is_valid_string_desulfurispirillum_indicum_node_name(node_name) {
            return Ok("STRING Desulfurispirillum indicum");
        }

        if is_valid_string_albugo_laibachii_node_name(node_name) {
            return Ok("STRING Albugo laibachii");
        }

        if is_valid_string_anaerococcus_vaginalis_node_name(node_name) {
            return Ok("STRING Anaerococcus vaginalis");
        }

        if is_valid_string_aerococcus_viridans_atcc11563_node_name(node_name) {
            return Ok("STRING Aerococcus viridans ATCC11563");
        }

        if is_valid_string_streptococcus_oralis_atcc35037_node_name(node_name) {
            return Ok("STRING Streptococcus oralis ATCC35037");
        }

        if is_valid_string_zunongwangia_profunda_node_name(node_name) {
            return Ok("STRING Zunongwangia profunda");
        }

        if is_valid_string_pseudogymnoascus_destructans_node_name(node_name) {
            return Ok("STRING Pseudogymnoascus destructans");
        }

        if is_valid_string_frankia_symbiont_node_name(node_name) {
            return Ok("STRING Frankia symbiont");
        }

        if is_valid_string_halanaerobium_hydrogeniformans_node_name(node_name) {
            return Ok("STRING Halanaerobium hydrogeniformans");
        }

        if is_valid_string_bacteroides_xylanisolvens_node_name(node_name) {
            return Ok("STRING Bacteroides xylanisolvens");
        }

        if is_valid_string_faecalibacterium_prausnitzii_sl33_node_name(node_name) {
            return Ok("STRING Faecalibacterium prausnitzii SL33");
        }

        if is_valid_string_lachnospiraceae_bacterium_3157faact1_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium 3157FAACT1");
        }

        if is_valid_string_lachnospiraceae_bacterium_9143bfaa_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium 9143BFAA");
        }

        if is_valid_string_liberibacter_solanacearum_node_name(node_name) {
            return Ok("STRING Liberibacter solanacearum");
        }

        if is_valid_string_legionella_drancourtii_node_name(node_name) {
            return Ok("STRING Legionella drancourtii");
        }

        if is_valid_string_pseudomonas_sp_h2_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. H2");
        }

        if is_valid_string_lachnospiraceae_bacterium_1456faa_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium 1456FAA");
        }

        if is_valid_string_erysipelotrichaceae_bacterium_3153_node_name(node_name) {
            return Ok("STRING Erysipelotrichaceae bacterium 3153");
        }

        if is_valid_string_mesotoga_prima_node_name(node_name) {
            return Ok("STRING Mesotoga prima");
        }

        if is_valid_string_olsenella_sp_f0356_node_name(node_name) {
            return Ok("STRING Olsenella sp. F0356");
        }

        if is_valid_string_legionella_longbeachae_node_name(node_name) {
            return Ok("STRING Legionella longbeachae");
        }

        if is_valid_string_fimbriimonas_ginsengisoli_node_name(node_name) {
            return Ok("STRING Fimbriimonas ginsengisoli");
        }

        if is_valid_string_haloferax_mucosum_node_name(node_name) {
            return Ok("STRING Haloferax mucosum");
        }

        if is_valid_string_corynebacterium_resistens_node_name(node_name) {
            return Ok("STRING Corynebacterium resistens");
        }

        if is_valid_string_ethanoligenens_harbinense_node_name(node_name) {
            return Ok("STRING Ethanoligenens harbinense");
        }

        if is_valid_string_regiella_insecticola_node_name(node_name) {
            return Ok("STRING Regiella insecticola");
        }

        if is_valid_string_methylocapsa_aurea_node_name(node_name) {
            return Ok("STRING Methylocapsa aurea");
        }

        if is_valid_string_acetobacter_aceti_atcc23746_node_name(node_name) {
            return Ok("STRING Acetobacter aceti ATCC23746");
        }

        if is_valid_string_streptococcus_dysgalactiae_atcc27957_node_name(node_name) {
            return Ok("STRING Streptococcus dysgalactiae ATCC27957");
        }

        if is_valid_string_erwinia_amylovora_node_name(node_name) {
            return Ok("STRING Erwinia amylovora");
        }

        if is_valid_string_spirochaeta_thermophila_node_name(node_name) {
            return Ok("STRING Spirochaeta thermophila");
        }

        if is_valid_string_streptomyces_viridosporus_node_name(node_name) {
            return Ok("STRING Streptomyces viridosporus");
        }

        if is_valid_string_desulfovibrio_sp_6146afaa_node_name(node_name) {
            return Ok("STRING Desulfovibrio sp. 6146AFAA");
        }

        if is_valid_string_lachnospiraceae_bacterium_3146faa_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium 3146FAA");
        }

        if is_valid_string_bacillus_smithii_node_name(node_name) {
            return Ok("STRING Bacillus smithii");
        }

        if is_valid_string_subdoligranulum_sp_4354a2faa_node_name(node_name) {
            return Ok("STRING Subdoligranulum sp. 4354A2FAA");
        }

        if is_valid_string_bacillus_sp_2a57ct2_node_name(node_name) {
            return Ok("STRING Bacillus sp. 2A57CT2");
        }

        if is_valid_string_planktomarina_temperata_node_name(node_name) {
            return Ok("STRING Planktomarina temperata");
        }

        if is_valid_string_acidilobus_saccharovorans_node_name(node_name) {
            return Ok("STRING Acidilobus saccharovorans");
        }

        if is_valid_string_methylotenera_versatilis_node_name(node_name) {
            return Ok("STRING Methylotenera versatilis");
        }

        if is_valid_string_afipia_sp_1nls2_node_name(node_name) {
            return Ok("STRING Afipia sp. 1NLS2");
        }

        if is_valid_string_rhodanobacter_denitrificans_node_name(node_name) {
            return Ok("STRING Rhodanobacter denitrificans");
        }

        if is_valid_string_bacillus_sp_1nla3e_node_name(node_name) {
            return Ok("STRING Bacillus sp. 1NLA3E");
        }

        if is_valid_string_thermodesulfatator_indicus_node_name(node_name) {
            return Ok("STRING Thermodesulfatator indicus");
        }

        if is_valid_string_bacteroides_salanitronis_node_name(node_name) {
            return Ok("STRING Bacteroides salanitronis");
        }

        if is_valid_string_edwardsiella_tarda_atcc15947_node_name(node_name) {
            return Ok("STRING Edwardsiella tarda ATCC15947");
        }

        if is_valid_string_burkholderia_sp_jpy347_node_name(node_name) {
            return Ok("STRING Burkholderia sp. JPY347");
        }

        if is_valid_string_mannheimia_haemolytica_bovine_node_name(node_name) {
            return Ok("STRING Mannheimia haemolytica BOVINE");
        }

        if is_valid_string_profftella_armatura_node_name(node_name) {
            return Ok("STRING Profftella armatura");
        }

        if is_valid_string_microvirga_aerilata_node_name(node_name) {
            return Ok("STRING Microvirga aerilata");
        }

        if is_valid_string_hyphomicrobium_denitrificans_1nes1_node_name(node_name) {
            return Ok("STRING Hyphomicrobium denitrificans 1NES1");
        }

        if is_valid_string_oceanithermus_profundus_node_name(node_name) {
            return Ok("STRING Oceanithermus profundus");
        }

        if is_valid_string_metallosphaera_yellowstonensis_node_name(node_name) {
            return Ok("STRING Metallosphaera yellowstonensis");
        }

        if is_valid_string_methylomirabilis_oxyfera_node_name(node_name) {
            return Ok("STRING Methylomirabilis oxyfera");
        }

        if is_valid_string_aciduliprofundum_sp_mar08339_node_name(node_name) {
            return Ok("STRING Aciduliprofundum sp. MAR08339");
        }

        if is_valid_string_babela_massiliensis_node_name(node_name) {
            return Ok("STRING Babela massiliensis");
        }

        if is_valid_string_vibrio_alginolyticus_40b_node_name(node_name) {
            return Ok("STRING Vibrio alginolyticus 40B");
        }

        if is_valid_string_pseudonocardia_dioxanivorans_node_name(node_name) {
            return Ok("STRING Pseudonocardia dioxanivorans");
        }

        if is_valid_string_vibrio_mimicus_node_name(node_name) {
            return Ok("STRING Vibrio mimicus");
        }

        if is_valid_string_grimontia_hollisae_node_name(node_name) {
            return Ok("STRING Grimontia hollisae");
        }

        if is_valid_string_vibrio_metschnikovii_node_name(node_name) {
            return Ok("STRING Vibrio metschnikovii");
        }

        if is_valid_string_vibrio_coralliilyticus_node_name(node_name) {
            return Ok("STRING Vibrio coralliilyticus");
        }

        if is_valid_string_vibrio_sp_rc586_node_name(node_name) {
            return Ok("STRING Vibrio sp. RC586");
        }

        if is_valid_string_vibrio_orientalis_node_name(node_name) {
            return Ok("STRING Vibrio orientalis");
        }

        if is_valid_string_photobacterium_damselae_node_name(node_name) {
            return Ok("STRING Photobacterium damselae");
        }

        if is_valid_string_francisella_tularensis_novicida_node_name(node_name) {
            return Ok("STRING Francisella tularensis novicida");
        }

        if is_valid_string_prevotella_timonensis_cris5cb1_node_name(node_name) {
            return Ok("STRING Prevotella timonensis CRIS5CB1");
        }

        if is_valid_string_prevotella_buccalis_atcc35310_node_name(node_name) {
            return Ok("STRING Prevotella buccalis ATCC35310");
        }

        if is_valid_string_prevotella_amnii_node_name(node_name) {
            return Ok("STRING Prevotella amnii");
        }

        if is_valid_string_bulleidia_extructa_node_name(node_name) {
            return Ok("STRING Bulleidia extructa");
        }

        if is_valid_string_segniliparus_rugosus_node_name(node_name) {
            return Ok("STRING Segniliparus rugosus");
        }

        if is_valid_string_alloprevotella_rava_node_name(node_name) {
            return Ok("STRING Alloprevotella rava");
        }

        if is_valid_string_johnsonella_ignava_node_name(node_name) {
            return Ok("STRING Johnsonella ignava");
        }

        if is_valid_string_selenomonas_infelix_node_name(node_name) {
            return Ok("STRING Selenomonas infelix");
        }

        if is_valid_string_helicobacter_mustelae_node_name(node_name) {
            return Ok("STRING Helicobacter mustelae");
        }

        if is_valid_string_methanoplanus_petrolearius_node_name(node_name) {
            return Ok("STRING Methanoplanus petrolearius");
        }

        if is_valid_string_alistipes_finegoldii_node_name(node_name) {
            return Ok("STRING Alistipes finegoldii");
        }

        if is_valid_string_bacteroides_coprosuis_node_name(node_name) {
            return Ok("STRING Bacteroides coprosuis");
        }

        if is_valid_string_streptomyces_scabiei_node_name(node_name) {
            return Ok("STRING Streptomyces scabiei");
        }

        if is_valid_string_rothia_mucilaginosa_node_name(node_name) {
            return Ok("STRING Rothia mucilaginosa");
        }

        if is_valid_string_granulicella_mallensis_node_name(node_name) {
            return Ok("STRING Granulicella mallensis");
        }

        if is_valid_string_campylobacter_jejuni_414_node_name(node_name) {
            return Ok("STRING Campylobacter jejuni 414");
        }

        if is_valid_string_listeria_seeligeri_12b_node_name(node_name) {
            return Ok("STRING Listeria seeligeri 12b");
        }

        if is_valid_string_alpha_proteobacterium_himb114_node_name(node_name) {
            return Ok("STRING alpha proteobacterium HIMB114");
        }

        if is_valid_string_deinococcus_sp_2009_node_name(node_name) {
            return Ok("STRING Deinococcus sp. 2009");
        }

        if is_valid_string_citromicrobium_bathyomarinum_node_name(node_name) {
            return Ok("STRING Citromicrobium bathyomarinum");
        }

        if is_valid_string_rhodococcus_equi_node_name(node_name) {
            return Ok("STRING Rhodococcus equi");
        }

        if is_valid_string_sphingomonas_sp_pr090111t3t6a_node_name(node_name) {
            return Ok("STRING Sphingomonas sp. PR090111T3T6A");
        }

        if is_valid_string_bartonella_rochalimae_node_name(node_name) {
            return Ok("STRING Bartonella rochalimae");
        }

        if is_valid_string_methylomicrobium_album_node_name(node_name) {
            return Ok("STRING Methylomicrobium album");
        }

        if is_valid_string_pseudomonas_sp_s9_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. S9");
        }

        if is_valid_string_comamonas_testosteroni_cnb2_node_name(node_name) {
            return Ok("STRING Comamonas testosteroni CNB2");
        }

        if is_valid_string_prevotella_multisaccharivorax_node_name(node_name) {
            return Ok("STRING Prevotella multisaccharivorax");
        }

        if is_valid_string_thermotoga_thermarum_node_name(node_name) {
            return Ok("STRING Thermotoga thermarum");
        }

        if is_valid_string_cellulophaga_algicola_node_name(node_name) {
            return Ok("STRING Cellulophaga algicola");
        }

        if is_valid_string_oribacterium_sp_nk2b42_node_name(node_name) {
            return Ok("STRING Oribacterium sp. NK2B42");
        }

        if is_valid_string_rhizobium_vignae_node_name(node_name) {
            return Ok("STRING Rhizobium vignae");
        }

        if is_valid_string_pseudomonas_fluorescens_nz007_node_name(node_name) {
            return Ok("STRING Pseudomonas fluorescens NZ007");
        }

        if is_valid_string_desulfovibrio_africanus_node_name(node_name) {
            return Ok("STRING Desulfovibrio africanus");
        }

        if is_valid_string_fonticula_alba_node_name(node_name) {
            return Ok("STRING Fonticula alba");
        }

        if is_valid_string_enterobacteriaceae_bacterium_strain_node_name(node_name) {
            return Ok("STRING Enterobacteriaceae bacterium strain");
        }

        if is_valid_string_archaeoglobus_veneficus_node_name(node_name) {
            return Ok("STRING Archaeoglobus veneficus");
        }

        if is_valid_string_oscillibacter_valericigenes_node_name(node_name) {
            return Ok("STRING Oscillibacter valericigenes");
        }

        if is_valid_string_deinococcus_proteolyticus_node_name(node_name) {
            return Ok("STRING Deinococcus proteolyticus");
        }

        if is_valid_string_bacteroides_helcogenes_node_name(node_name) {
            return Ok("STRING Bacteroides helcogenes");
        }

        if is_valid_string_methylobacterium_oryzae_node_name(node_name) {
            return Ok("STRING Methylobacterium oryzae");
        }

        if is_valid_string_paludibacter_propionicigenes_node_name(node_name) {
            return Ok("STRING Paludibacter propionicigenes");
        }

        if is_valid_string_pyrolobus_fumarii_node_name(node_name) {
            return Ok("STRING Pyrolobus fumarii");
        }

        if is_valid_string_natronococcus_occultus_node_name(node_name) {
            return Ok("STRING Natronococcus occultus");
        }

        if is_valid_string_desulfurella_acetivorans_node_name(node_name) {
            return Ok("STRING Desulfurella acetivorans");
        }

        if is_valid_string_methanomicrobium_mobile_node_name(node_name) {
            return Ok("STRING Methanomicrobium mobile");
        }

        if is_valid_string_aggregatibacter_actinomycetemcomitans_d7s1_node_name(node_name) {
            return Ok("STRING Aggregatibacter actinomycetemcomitans D7S1");
        }

        if is_valid_string_bartonella_clarridgeiae_node_name(node_name) {
            return Ok("STRING Bartonella clarridgeiae");
        }

        if is_valid_string_desulfotomaculum_ruminis_node_name(node_name) {
            return Ok("STRING Desulfotomaculum ruminis");
        }

        if is_valid_string_desulfotomaculum_nigrificans_node_name(node_name) {
            return Ok("STRING Desulfotomaculum nigrificans");
        }

        if is_valid_string_arthrospira_platensis_nies39_node_name(node_name) {
            return Ok("STRING Arthrospira platensis NIES39");
        }

        if is_valid_string_actinobacillus_suis_node_name(node_name) {
            return Ok("STRING Actinobacillus suis");
        }

        if is_valid_string_mahella_australiensis_node_name(node_name) {
            return Ok("STRING Mahella australiensis");
        }

        if is_valid_string_methylobacter_tundripaludum_node_name(node_name) {
            return Ok("STRING Methylobacter tundripaludum");
        }

        if is_valid_string_paenibacillus_larvae_node_name(node_name) {
            return Ok("STRING Paenibacillus larvae");
        }

        if is_valid_string_thermoanaerobacter_wiegelii_node_name(node_name) {
            return Ok("STRING Thermoanaerobacter wiegelii");
        }

        if is_valid_string_ruminococcus_albus_7_node_name(node_name) {
            return Ok("STRING Ruminococcus albus 7");
        }

        if is_valid_string_marssonina_brunnea_node_name(node_name) {
            return Ok("STRING Marssonina brunnea");
        }

        if is_valid_string_staphylococcus_lugdunensis_hku0901_node_name(node_name) {
            return Ok("STRING Staphylococcus lugdunensis HKU0901");
        }

        if is_valid_string_pyrobaculum_oguniense_node_name(node_name) {
            return Ok("STRING Pyrobaculum oguniense");
        }

        if is_valid_string_amphibacillus_xylanus_node_name(node_name) {
            return Ok("STRING Amphibacillus xylanus");
        }

        if is_valid_string_rhizobium_tropici_node_name(node_name) {
            return Ok("STRING Rhizobium tropici");
        }

        if is_valid_string_virgibacillus_alimentarius_node_name(node_name) {
            return Ok("STRING Virgibacillus alimentarius");
        }

        if is_valid_string_gardnerella_vaginalis_6119v5_node_name(node_name) {
            return Ok("STRING Gardnerella vaginalis 6119V5");
        }

        if is_valid_string_corynebacterium_diphtheriae_node_name(node_name) {
            return Ok("STRING Corynebacterium diphtheriae");
        }

        if is_valid_string_megasphaera_genomosp_node_name(node_name) {
            return Ok("STRING Megasphaera genomosp.");
        }

        if is_valid_string_clostridiales_genomosp_node_name(node_name) {
            return Ok("STRING Clostridiales genomosp.");
        }

        if is_valid_string_streptococcus_ratti_node_name(node_name) {
            return Ok("STRING Streptococcus ratti");
        }

        if is_valid_string_coriobacterium_glomerans_node_name(node_name) {
            return Ok("STRING Coriobacterium glomerans");
        }

        if is_valid_string_mycobacterium_neoaurum_node_name(node_name) {
            return Ok("STRING Mycobacterium neoaurum");
        }

        if is_valid_string_niastella_koreensis_node_name(node_name) {
            return Ok("STRING Niastella koreensis");
        }

        if is_valid_string_vibrio_sp_n418_node_name(node_name) {
            return Ok("STRING Vibrio sp. N418");
        }

        if is_valid_string_enterobacter_lignolyticus_node_name(node_name) {
            return Ok("STRING Enterobacter lignolyticus");
        }

        if is_valid_string_pediococcus_claussenii_node_name(node_name) {
            return Ok("STRING Pediococcus claussenii");
        }

        if is_valid_string_novosphingobium_sp_pp1y_node_name(node_name) {
            return Ok("STRING Novosphingobium sp. PP1Y");
        }

        if is_valid_string_selenomonas_noxia_f0398_node_name(node_name) {
            return Ok("STRING Selenomonas noxia F0398");
        }

        if is_valid_string_prevotella_oulorum_f0390_node_name(node_name) {
            return Ok("STRING Prevotella oulorum F0390");
        }

        if is_valid_string_turicibacter_sanguinis_node_name(node_name) {
            return Ok("STRING Turicibacter sanguinis");
        }

        if is_valid_string_bifidobacterium_bifidum_prl2010_node_name(node_name) {
            return Ok("STRING Bifidobacterium bifidum PRL2010");
        }

        if is_valid_string_pantoea_ananatis_lmg20103_node_name(node_name) {
            return Ok("STRING Pantoea ananatis LMG20103");
        }

        if is_valid_string_solobacterium_moorei_f0204_node_name(node_name) {
            return Ok("STRING Solobacterium moorei F0204");
        }

        if is_valid_string_megasphaera_micronuciformis_node_name(node_name) {
            return Ok("STRING Megasphaera micronuciformis");
        }

        if is_valid_string_capnocytophaga_sp_f0087_node_name(node_name) {
            return Ok("STRING Capnocytophaga sp. F0087");
        }

        if is_valid_string_streptococcus_anginosus_f0211_node_name(node_name) {
            return Ok("STRING Streptococcus anginosus F0211");
        }

        if is_valid_string_actinomyces_sp_f0337_node_name(node_name) {
            return Ok("STRING Actinomyces sp. F0337");
        }

        if is_valid_string_desulfomonile_tiedjei_node_name(node_name) {
            return Ok("STRING Desulfomonile tiedjei");
        }

        if is_valid_string_mycoplasma_gallisepticum_f_node_name(node_name) {
            return Ok("STRING Mycoplasma gallisepticum F");
        }

        if is_valid_string_sulfuricurvum_kujiense_node_name(node_name) {
            return Ok("STRING Sulfuricurvum kujiense");
        }

        if is_valid_string_bradyrhizobiaceae_bacterium_sg6c_node_name(node_name) {
            return Ok("STRING Bradyrhizobiaceae bacterium SG6C");
        }

        if is_valid_string_deinococcus_maricopensis_node_name(node_name) {
            return Ok("STRING Deinococcus maricopensis");
        }

        if is_valid_string_odoribacter_splanchnicus_node_name(node_name) {
            return Ok("STRING Odoribacter splanchnicus");
        }

        if is_valid_string_frankia_sp_qa3_node_name(node_name) {
            return Ok("STRING Frankia sp. QA3");
        }

        if is_valid_string_colletotrichum_fioriniae_node_name(node_name) {
            return Ok("STRING Colletotrichum fioriniae");
        }

        if is_valid_string_helicobacter_suis_node_name(node_name) {
            return Ok("STRING Helicobacter suis");
        }

        if is_valid_string_mycobacterium_chubuense_node_name(node_name) {
            return Ok("STRING Mycobacterium chubuense");
        }

        if is_valid_string_mycobacterium_rhodesiae_nbb3_node_name(node_name) {
            return Ok("STRING Mycobacterium rhodesiae NBB3");
        }

        if is_valid_string_mycobacterium_smegmatis_js623_node_name(node_name) {
            return Ok("STRING Mycobacterium smegmatis JS623");
        }

        if is_valid_string_mycobacterium_tusciae_node_name(node_name) {
            return Ok("STRING Mycobacterium tusciae");
        }

        if is_valid_string_intrasporangium_calvum_node_name(node_name) {
            return Ok("STRING Intrasporangium calvum");
        }

        if is_valid_string_streptomyces_sp_gxt6_node_name(node_name) {
            return Ok("STRING Streptomyces sp. GXT6");
        }

        if is_valid_string_pantoea_vagans_node_name(node_name) {
            return Ok("STRING Pantoea vagans");
        }

        if is_valid_string_thioalkalivibrio_thiocyanodenitrificans_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio thiocyanodenitrificans");
        }

        if is_valid_string_thioalkalivibrio_thiocyanoxidans_arh4_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio thiocyanoxidans ARh4");
        }

        if is_valid_string_lactobacillus_iners_node_name(node_name) {
            return Ok("STRING Lactobacillus iners");
        }

        if is_valid_string_leucobacter_chromiiresistens_node_name(node_name) {
            return Ok("STRING Leucobacter chromiiresistens");
        }

        if is_valid_string_lactobacillus_sanfranciscensis_node_name(node_name) {
            return Ok("STRING Lactobacillus sanfranciscensis");
        }

        if is_valid_string_mucilaginibacter_paludis_node_name(node_name) {
            return Ok("STRING Mucilaginibacter paludis");
        }

        if is_valid_string_lysinibacillus_fusiformis_node_name(node_name) {
            return Ok("STRING Lysinibacillus fusiformis");
        }

        if is_valid_string_asticcacaulis_biprosthecum_node_name(node_name) {
            return Ok("STRING Asticcacaulis biprosthecum");
        }

        if is_valid_string_alteromonas_sp_sn2_node_name(node_name) {
            return Ok("STRING Alteromonas sp. SN2");
        }

        if is_valid_string_enterobacter_cloacae_atcc13047_node_name(node_name) {
            return Ok("STRING Enterobacter cloacae ATCC13047");
        }

        if is_valid_string_waddlia_chondrophila_node_name(node_name) {
            return Ok("STRING Waddlia chondrophila");
        }

        if is_valid_string_ensifer_sojae_node_name(node_name) {
            return Ok("STRING Ensifer sojae");
        }

        if is_valid_string_flexistipes_sinusarabici_node_name(node_name) {
            return Ok("STRING Flexistipes sinusarabici");
        }

        if is_valid_string_thermobacillus_composti_node_name(node_name) {
            return Ok("STRING Thermobacillus composti");
        }

        if is_valid_string_paenibacillus_curdlanolyticus_node_name(node_name) {
            return Ok("STRING Paenibacillus curdlanolyticus");
        }

        if is_valid_string_thioalkalimicrobium_aerophilum_node_name(node_name) {
            return Ok("STRING Thioalkalimicrobium aerophilum");
        }

        if is_valid_string_thioalkalimicrobium_cyclicum_node_name(node_name) {
            return Ok("STRING Thioalkalimicrobium cyclicum");
        }

        if is_valid_string_marinomonas_mediterranea_node_name(node_name) {
            return Ok("STRING Marinomonas mediterranea");
        }

        if is_valid_string_hyphomicrobium_sp_mc1_node_name(node_name) {
            return Ok("STRING Hyphomicrobium sp. MC1");
        }

        if is_valid_string_faecalibacterium_prausnitzii_l26_node_name(node_name) {
            return Ok("STRING Faecalibacterium prausnitzii L26");
        }

        if is_valid_string_clostridium_clariflavum_node_name(node_name) {
            return Ok("STRING Clostridium clariflavum");
        }

        if is_valid_string_bacillus_atrophaeus_node_name(node_name) {
            return Ok("STRING Bacillus atrophaeus");
        }

        if is_valid_string_pseudoalteromonas_haloplanktis_ant505_node_name(node_name) {
            return Ok("STRING Pseudoalteromonas haloplanktis ANT505");
        }

        if is_valid_string_mycoplasma_pneumoniae_node_name(node_name) {
            return Ok("STRING Mycoplasma pneumoniae");
        }

        if is_valid_string_idiomarina_xiamenensis_node_name(node_name) {
            return Ok("STRING Idiomarina xiamenensis");
        }

        if is_valid_string_rahnella_sp_y9602_node_name(node_name) {
            return Ok("STRING Rahnella sp. Y9602");
        }

        if is_valid_string_achromobacter_piechaudii_atcc43553_node_name(node_name) {
            return Ok("STRING Achromobacter piechaudii ATCC43553");
        }

        if is_valid_string_collinsella_sp_4847faa_node_name(node_name) {
            return Ok("STRING Collinsella sp. 4847FAA");
        }

        if is_valid_string_lachnospiraceae_bacterium_2146faa_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium 2146FAA");
        }

        if is_valid_string_alistipes_indistinctus_node_name(node_name) {
            return Ok("STRING Alistipes indistinctus");
        }

        if is_valid_string_barnesiella_intestinihominis_node_name(node_name) {
            return Ok("STRING Barnesiella intestinihominis");
        }

        if is_valid_string_bacteroides_oleiciplenus_node_name(node_name) {
            return Ok("STRING Bacteroides oleiciplenus");
        }

        if is_valid_string_clostridium_citroniae_node_name(node_name) {
            return Ok("STRING Clostridium citroniae");
        }

        if is_valid_string_clostridium_clostridioforme_node_name(node_name) {
            return Ok("STRING Clostridium clostridioforme");
        }

        if is_valid_string_clostridium_orbiscindens_node_name(node_name) {
            return Ok("STRING Clostridium orbiscindens");
        }

        if is_valid_string_clostridium_symbiosum_wal14163_node_name(node_name) {
            return Ok("STRING Clostridium symbiosum WAL14163");
        }

        if is_valid_string_clostridium_symbiosum_wal14673_node_name(node_name) {
            return Ok("STRING Clostridium symbiosum WAL14673");
        }

        if is_valid_string_collinsella_tanakaei_node_name(node_name) {
            return Ok("STRING Collinsella tanakaei");
        }

        if is_valid_string_dialister_succinatiphilus_node_name(node_name) {
            return Ok("STRING Dialister succinatiphilus");
        }

        if is_valid_string_dorea_formicigenerans_4653afaa_node_name(node_name) {
            return Ok("STRING Dorea formicigenerans 4653AFAA");
        }

        if is_valid_string_dysgonomonas_gadei_node_name(node_name) {
            return Ok("STRING Dysgonomonas gadei");
        }

        if is_valid_string_dysgonomonas_mossii_node_name(node_name) {
            return Ok("STRING Dysgonomonas mossii");
        }

        if is_valid_string_odoribacter_laneus_node_name(node_name) {
            return Ok("STRING Odoribacter laneus");
        }

        if is_valid_string_slackia_piriformis_node_name(node_name) {
            return Ok("STRING Slackia piriformis");
        }

        if is_valid_string_sutterella_wadsworthensis_3145b_node_name(node_name) {
            return Ok("STRING Sutterella wadsworthensis 3145B");
        }

        if is_valid_string_sutterella_wadsworthensis_2159bfaa_node_name(node_name) {
            return Ok("STRING Sutterella wadsworthensis 2159BFAA");
        }

        if is_valid_string_acidithiobacillus_ferrivorans_node_name(node_name) {
            return Ok("STRING Acidithiobacillus ferrivorans");
        }

        if is_valid_string_thermus_scotoductus_node_name(node_name) {
            return Ok("STRING Thermus scotoductus");
        }

        if is_valid_string_isoptericola_variabilis_node_name(node_name) {
            return Ok("STRING Isoptericola variabilis");
        }

        if is_valid_string_paenibacillus_lactis_node_name(node_name) {
            return Ok("STRING Paenibacillus lactis");
        }

        if is_valid_string_pseudomonas_fulva_node_name(node_name) {
            return Ok("STRING Pseudomonas fulva");
        }

        if is_valid_string_pseudoxanthomonas_suwonensis_111_node_name(node_name) {
            return Ok("STRING Pseudoxanthomonas suwonensis 111");
        }

        if is_valid_string_sphingobacterium_sp_21_node_name(node_name) {
            return Ok("STRING Sphingobacterium sp. 21");
        }

        if is_valid_string_methylocystis_sp_sb2_node_name(node_name) {
            return Ok("STRING Methylocystis sp. SB2");
        }

        if is_valid_string_mycoplasma_putrefaciens_ks1_node_name(node_name) {
            return Ok("STRING Mycoplasma putrefaciens KS1");
        }

        if is_valid_string_mycoplasma_bovoculi_node_name(node_name) {
            return Ok("STRING Mycoplasma bovoculi");
        }

        if is_valid_string_moraxella_bovoculi_node_name(node_name) {
            return Ok("STRING Moraxella bovoculi");
        }

        if is_valid_string_treponema_caldaria_node_name(node_name) {
            return Ok("STRING Treponema caldaria");
        }

        if is_valid_string_ahrensia_sp_r2a130_node_name(node_name) {
            return Ok("STRING Ahrensia sp. R2A130");
        }

        if is_valid_string_roseibium_sp_trichskd4_node_name(node_name) {
            return Ok("STRING Roseibium sp. TrichSKD4");
        }

        if is_valid_string_alpha_proteobacterium_himb59_node_name(node_name) {
            return Ok("STRING alpha proteobacterium HIMB59");
        }

        if is_valid_string_gammaproteobacteria_bacterium_himb55_node_name(node_name) {
            return Ok("STRING Gammaproteobacteria bacterium HIMB55");
        }

        if is_valid_string_rahnella_aquatilis_cip7865_node_name(node_name) {
            return Ok("STRING Rahnella aquatilis CIP78.65");
        }

        if is_valid_string_sphingomonas_sp_mm1_node_name(node_name) {
            return Ok("STRING Sphingomonas sp. MM1");
        }

        if is_valid_string_gallaecimonas_xiamenensis_node_name(node_name) {
            return Ok("STRING Gallaecimonas xiamenensis");
        }

        if is_valid_string_olleya_sp_vcsm12_node_name(node_name) {
            return Ok("STRING Olleya sp. VCSM12");
        }

        if is_valid_string_deinococcus_gobiensis_node_name(node_name) {
            return Ok("STRING Deinococcus gobiensis");
        }

        if is_valid_string_aspergillus_fumigatus_node_name(node_name) {
            return Ok("STRING Aspergillus fumigatus");
        }

        if is_valid_string_aequorivita_sublithincola_node_name(node_name) {
            return Ok("STRING Aequorivita sublithincola");
        }

        if is_valid_string_thermodesulfobium_narugense_node_name(node_name) {
            return Ok("STRING Thermodesulfobium narugense");
        }

        if is_valid_string_mycoplasma_alligatoris_node_name(node_name) {
            return Ok("STRING Mycoplasma alligatoris");
        }

        if is_valid_string_faecalibacterium_prausnitzii_kle1255_node_name(node_name) {
            return Ok("STRING Faecalibacterium prausnitzii KLE1255");
        }

        if is_valid_string_azoarcus_sp_kh32c_node_name(node_name) {
            return Ok("STRING Azoarcus sp. KH32C");
        }

        if is_valid_string_pseudogulbenkiania_sp_nh8b_node_name(node_name) {
            return Ok("STRING Pseudogulbenkiania sp. NH8B");
        }

        if is_valid_string_halobacteroides_halobius_node_name(node_name) {
            return Ok("STRING Halobacteroides halobius");
        }

        if is_valid_string_thioalkalivibrio_sp_alsr1_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. ALSr1");
        }

        if is_valid_string_lactobacillus_crispatus_node_name(node_name) {
            return Ok("STRING Lactobacillus crispatus");
        }

        if is_valid_string_clostridium_ljungdahlii_node_name(node_name) {
            return Ok("STRING Clostridium ljungdahlii");
        }

        if is_valid_string_nitratifractor_salsuginis_node_name(node_name) {
            return Ok("STRING Nitratifractor salsuginis");
        }

        if is_valid_string_streptomyces_bingchenggensis_node_name(node_name) {
            return Ok("STRING Streptomyces bingchenggensis");
        }

        if is_valid_string_amycolatopsis_mediterranei_node_name(node_name) {
            return Ok("STRING Amycolatopsis mediterranei");
        }

        if is_valid_string_halobacterium_sp_dl1_node_name(node_name) {
            return Ok("STRING Halobacterium sp. DL1");
        }

        if is_valid_string_thermus_oshimai_node_name(node_name) {
            return Ok("STRING Thermus oshimai");
        }

        if is_valid_string_gammaproteobacteria_bacterium_himb30_node_name(node_name) {
            return Ok("STRING Gammaproteobacteria bacterium HIMB30");
        }

        if is_valid_string_treponema_phagedenis_node_name(node_name) {
            return Ok("STRING Treponema phagedenis");
        }

        if is_valid_string_mesorhizobium_australicum_node_name(node_name) {
            return Ok("STRING Mesorhizobium australicum");
        }

        if is_valid_string_propionibacterium_freudenreichii_node_name(node_name) {
            return Ok("STRING Propionibacterium freudenreichii");
        }

        if is_valid_string_escherichia_sp_tw09308_node_name(node_name) {
            return Ok("STRING Escherichia sp. TW09308");
        }

        if is_valid_string_photobacterium_aphoticum_node_name(node_name) {
            return Ok("STRING Photobacterium aphoticum");
        }

        if is_valid_string_methylophaga_nitratireducenticrescens_node_name(node_name) {
            return Ok("STRING Methylophaga nitratireducenticrescens");
        }

        if is_valid_string_methylophaga_frappieri_node_name(node_name) {
            return Ok("STRING Methylophaga frappieri");
        }

        if is_valid_string_cyanobacterium_aponinum_node_name(node_name) {
            return Ok("STRING Cyanobacterium aponinum");
        }

        if is_valid_string_clostridium_sp_bnl1100_node_name(node_name) {
            return Ok("STRING Clostridium sp. BNL1100");
        }

        if is_valid_string_fluviicola_taffensis_node_name(node_name) {
            return Ok("STRING Fluviicola taffensis");
        }

        if is_valid_string_microcoleus_vaginatus_node_name(node_name) {
            return Ok("STRING Microcoleus vaginatus");
        }

        if is_valid_string_planctomyces_brasiliensis_node_name(node_name) {
            return Ok("STRING Planctomyces brasiliensis");
        }

        if is_valid_string_desulfitobacterium_dehalogenans_node_name(node_name) {
            return Ok("STRING Desulfitobacterium dehalogenans");
        }

        if is_valid_string_haloferacales_archaeon_dl31_node_name(node_name) {
            return Ok("STRING Haloferacales archaeon DL31");
        }

        if is_valid_string_herbaspirillum_seropedicae_node_name(node_name) {
            return Ok("STRING Herbaspirillum seropedicae");
        }

        if is_valid_string_ketogulonicigenium_vulgare_node_name(node_name) {
            return Ok("STRING Ketogulonicigenium vulgare");
        }

        if is_valid_string_streptococcus_dysgalactiae_equisimilis_node_name(node_name) {
            return Ok("STRING Streptococcus dysgalactiae equisimilis");
        }

        if is_valid_string_brachyspira_pilosicoli_node_name(node_name) {
            return Ok("STRING Brachyspira pilosicoli");
        }

        if is_valid_string_sphaerochaeta_coccoides_node_name(node_name) {
            return Ok("STRING Sphaerochaeta coccoides");
        }

        if is_valid_string_massilia_consociata_node_name(node_name) {
            return Ok("STRING Massilia consociata");
        }

        if is_valid_string_hippea_maritima_node_name(node_name) {
            return Ok("STRING Hippea maritima");
        }

        if is_valid_string_sulfurospirillum_barnesii_node_name(node_name) {
            return Ok("STRING Sulfurospirillum barnesii");
        }

        if is_valid_string_haliscomenobacter_hydrossis_node_name(node_name) {
            return Ok("STRING Haliscomenobacter hydrossis");
        }

        if is_valid_string_desulfotomaculum_kuznetsovii_node_name(node_name) {
            return Ok("STRING Desulfotomaculum kuznetsovii");
        }

        if is_valid_string_runella_slithyformis_node_name(node_name) {
            return Ok("STRING Runella slithyformis");
        }

        if is_valid_string_leuconostoc_kimchii_node_name(node_name) {
            return Ok("STRING Leuconostoc kimchii");
        }

        if is_valid_string_bifidobacterium_stellenboschense_node_name(node_name) {
            return Ok("STRING Bifidobacterium stellenboschense");
        }

        if is_valid_string_achromobacter_xylosoxidans_a8_node_name(node_name) {
            return Ok("STRING Achromobacter xylosoxidans A8");
        }

        if is_valid_string_leuconostoc_gasicomitatum_node_name(node_name) {
            return Ok("STRING Leuconostoc gasicomitatum");
        }

        if is_valid_string_pedobacter_saltans_node_name(node_name) {
            return Ok("STRING Pedobacter saltans");
        }

        if is_valid_string_rothia_dentocariosa_node_name(node_name) {
            return Ok("STRING Rothia dentocariosa");
        }

        if is_valid_string_actinomyces_sp_f0386_node_name(node_name) {
            return Ok("STRING Actinomyces sp. F0386");
        }

        if is_valid_string_parasutterella_excrementihominis_node_name(node_name) {
            return Ok("STRING Parasutterella excrementihominis");
        }

        if is_valid_string_paraprevotella_clara_node_name(node_name) {
            return Ok("STRING Paraprevotella clara");
        }

        if is_valid_string_paraprevotella_xylaniphila_node_name(node_name) {
            return Ok("STRING Paraprevotella xylaniphila");
        }

        if is_valid_string_succinatimonas_hippei_node_name(node_name) {
            return Ok("STRING Succinatimonas hippei");
        }

        if is_valid_string_bacteroides_clarus_node_name(node_name) {
            return Ok("STRING Bacteroides clarus");
        }

        if is_valid_string_bacteroides_fluxus_node_name(node_name) {
            return Ok("STRING Bacteroides fluxus");
        }

        if is_valid_string_streptococcus_urinalis_228597_node_name(node_name) {
            return Ok("STRING Streptococcus urinalis 228597");
        }

        if is_valid_string_streptococcus_macacae_node_name(node_name) {
            return Ok("STRING Streptococcus macacae");
        }

        if is_valid_string_streptococcus_ictaluri_node_name(node_name) {
            return Ok("STRING Streptococcus ictaluri");
        }

        if is_valid_string_propionibacterium_acnes_hl037pa2_node_name(node_name) {
            return Ok("STRING Propionibacterium acnes HL037PA2");
        }

        if is_valid_string_desulfurococcus_mucosus_node_name(node_name) {
            return Ok("STRING Desulfurococcus mucosus");
        }

        if is_valid_string_oscillochloris_trichoides_node_name(node_name) {
            return Ok("STRING Oscillochloris trichoides");
        }

        if is_valid_string_mesorhizobium_ciceri_biovar_node_name(node_name) {
            return Ok("STRING Mesorhizobium ciceri biovar");
        }

        if is_valid_string_bdellovibrio_bacteriovorus_w_node_name(node_name) {
            return Ok("STRING Bdellovibrio bacteriovorus W");
        }

        if is_valid_string_marichromatium_purpuratum_node_name(node_name) {
            return Ok("STRING Marichromatium purpuratum");
        }

        if is_valid_string_thiocystis_violascens_node_name(node_name) {
            return Ok("STRING Thiocystis violascens");
        }

        if is_valid_string_thioflavicoccus_mobilis_node_name(node_name) {
            return Ok("STRING Thioflavicoccus mobilis");
        }

        if is_valid_string_thiorhodococcus_drewsii_node_name(node_name) {
            return Ok("STRING Thiorhodococcus drewsii");
        }

        if is_valid_string_thiorhodospira_sibirica_node_name(node_name) {
            return Ok("STRING Thiorhodospira sibirica");
        }

        if is_valid_string_parachlamydia_acanthamoebae_node_name(node_name) {
            return Ok("STRING Parachlamydia acanthamoebae");
        }

        if is_valid_string_citreicella_sp_357_node_name(node_name) {
            return Ok("STRING Citreicella sp. 357");
        }

        if is_valid_string_propionibacterium_propionicum_node_name(node_name) {
            return Ok("STRING Propionibacterium propionicum");
        }

        if is_valid_string_prevotella_denticola_node_name(node_name) {
            return Ok("STRING Prevotella denticola");
        }

        if is_valid_string_frateuria_aurantia_node_name(node_name) {
            return Ok("STRING Frateuria aurantia");
        }

        if is_valid_string_desulfotomaculum_gibsoniae_node_name(node_name) {
            return Ok("STRING Desulfotomaculum gibsoniae");
        }

        if is_valid_string_halomonas_elongata_node_name(node_name) {
            return Ok("STRING Halomonas elongata");
        }

        if is_valid_string_enterococcus_hirae_node_name(node_name) {
            return Ok("STRING Enterococcus hirae");
        }

        if is_valid_string_calditerrivibrio_nitroreducens_node_name(node_name) {
            return Ok("STRING Calditerrivibrio nitroreducens");
        }

        if is_valid_string_thiocapsa_marina_node_name(node_name) {
            return Ok("STRING Thiocapsa marina");
        }

        if is_valid_string_desulfurococcus_fermentans_node_name(node_name) {
            return Ok("STRING Desulfurococcus fermentans");
        }

        if is_valid_string_thermoproteus_tenax_node_name(node_name) {
            return Ok("STRING Thermoproteus tenax");
        }

        if is_valid_string_desulfosporosinus_meridiei_node_name(node_name) {
            return Ok("STRING Desulfosporosinus meridiei");
        }

        if is_valid_string_desulfosporosinus_orientis_node_name(node_name) {
            return Ok("STRING Desulfosporosinus orientis");
        }

        if is_valid_string_desulfosporosinus_youngiae_node_name(node_name) {
            return Ok("STRING Desulfosporosinus youngiae");
        }

        if is_valid_string_streptococcus_mitis_2_f0392_node_name(node_name) {
            return Ok("STRING Streptococcus mitis 2 F0392");
        }

        if is_valid_string_fervidobacterium_pennivorans_node_name(node_name) {
            return Ok("STRING Fervidobacterium pennivorans");
        }

        if is_valid_string_sinorhizobium_sp_ccbau05631_node_name(node_name) {
            return Ok("STRING Sinorhizobium sp. CCBAU05631");
        }

        if is_valid_string_opitutaceae_bacterium_tav5_node_name(node_name) {
            return Ok("STRING Opitutaceae bacterium TAV5");
        }

        if is_valid_string_thermodesulfobacterium_geofontis_node_name(node_name) {
            return Ok("STRING Thermodesulfobacterium geofontis");
        }

        if is_valid_string_ralstonia_sp_pba_node_name(node_name) {
            return Ok("STRING Ralstonia sp. PBA");
        }

        if is_valid_string_halalkalicoccus_jeotgali_node_name(node_name) {
            return Ok("STRING Halalkalicoccus jeotgali");
        }

        if is_valid_string_nesterenkonia_sp_f_node_name(node_name) {
            return Ok("STRING Nesterenkonia sp. F");
        }

        if is_valid_string_bacillus_methanolicus_mga3_node_name(node_name) {
            return Ok("STRING Bacillus methanolicus MGA3");
        }

        if is_valid_string_vibrio_caribbeanicus_node_name(node_name) {
            return Ok("STRING Vibrio caribbeanicus");
        }

        if is_valid_string_peptostreptococcaceae_bacterium_cm5_node_name(node_name) {
            return Ok("STRING Peptostreptococcaceae bacterium CM5");
        }

        if is_valid_string_stomatobaculum_longum_node_name(node_name) {
            return Ok("STRING Stomatobaculum longum");
        }

        if is_valid_string_oribacterium_parvum_node_name(node_name) {
            return Ok("STRING Oribacterium parvum");
        }

        if is_valid_string_halosimplex_carlsbadense_node_name(node_name) {
            return Ok("STRING Halosimplex carlsbadense");
        }

        if is_valid_string_haladaptatus_paucihalophilus_node_name(node_name) {
            return Ok("STRING Haladaptatus paucihalophilus");
        }

        if is_valid_string_halopiger_xanaduensis_node_name(node_name) {
            return Ok("STRING Halopiger xanaduensis");
        }

        if is_valid_string_halostagnicola_larsenii_node_name(node_name) {
            return Ok("STRING Halostagnicola larsenii");
        }

        if is_valid_string_halovivax_ruber_node_name(node_name) {
            return Ok("STRING Halovivax ruber");
        }

        if is_valid_string_natrinema_pellirubrum_node_name(node_name) {
            return Ok("STRING Natrinema pellirubrum");
        }

        if is_valid_string_natronobacterium_gregoryi_node_name(node_name) {
            return Ok("STRING Natronobacterium gregoryi");
        }

        if is_valid_string_lactobacillus_parafarraginis_node_name(node_name) {
            return Ok("STRING Lactobacillus parafarraginis");
        }

        if is_valid_string_spiroplasma_mirum_node_name(node_name) {
            return Ok("STRING Spiroplasma mirum");
        }

        if is_valid_string_micavibrio_aeruginosavorus_arl13_node_name(node_name) {
            return Ok("STRING Micavibrio aeruginosavorus ARL13");
        }

        if is_valid_string_methylomonas_methanica_node_name(node_name) {
            return Ok("STRING Methylomonas methanica");
        }

        if is_valid_string_scardovia_wiggsiae_node_name(node_name) {
            return Ok("STRING Scardovia wiggsiae");
        }

        if is_valid_string_caloramator_australicus_node_name(node_name) {
            return Ok("STRING Caloramator australicus");
        }

        if is_valid_string_moraxella_catarrhalis_node_name(node_name) {
            return Ok("STRING Moraxella catarrhalis");
        }

        if is_valid_string_thermoanaerobacterium_xylanolyticum_node_name(node_name) {
            return Ok("STRING Thermoanaerobacterium xylanolyticum");
        }

        if is_valid_string_corynebacterium_variabile_node_name(node_name) {
            return Ok("STRING Corynebacterium variabile");
        }

        if is_valid_string_mycoplasma_haemofelis_node_name(node_name) {
            return Ok("STRING Mycoplasma haemofelis");
        }

        if is_valid_string_alpha_proteobacterium_himb5_node_name(node_name) {
            return Ok("STRING alpha proteobacterium HIMB5");
        }

        if is_valid_string_ralstonia_solanacearum_psi07_node_name(node_name) {
            return Ok("STRING Ralstonia solanacearum PSI07");
        }

        if is_valid_string_capnocytophaga_canimorsus_node_name(node_name) {
            return Ok("STRING Capnocytophaga canimorsus");
        }

        if is_valid_string_agrobacterium_sp_h133_node_name(node_name) {
            return Ok("STRING Agrobacterium sp. H133");
        }

        if is_valid_string_gemmatimonadetes_bacterium_kbs708_node_name(node_name) {
            return Ok("STRING Gemmatimonadetes bacterium KBS708");
        }

        if is_valid_string_arthrobacter_arilaitensis_node_name(node_name) {
            return Ok("STRING Arthrobacter arilaitensis");
        }

        if is_valid_string_anaeroglobus_geminatus_node_name(node_name) {
            return Ok("STRING Anaeroglobus geminatus");
        }

        if is_valid_string_fusobacterium_sp_f0437_node_name(node_name) {
            return Ok("STRING Fusobacterium sp. F0437");
        }

        if is_valid_string_lachnospiraceae_bacterium_f0431_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium F0431");
        }

        if is_valid_string_streptococcus_sp_f0407_node_name(node_name) {
            return Ok("STRING Streptococcus sp. F0407");
        }

        if is_valid_string_staphylococcus_sp_al1_node_name(node_name) {
            return Ok("STRING Staphylococcus sp. AL1");
        }

        if is_valid_string_pediococcus_acidilactici_dsm20284_node_name(node_name) {
            return Ok("STRING Pediococcus acidilactici DSM20284");
        }

        if is_valid_string_prevotella_marshii_node_name(node_name) {
            return Ok("STRING Prevotella marshii");
        }

        if is_valid_string_peptoniphilus_duerdenii_node_name(node_name) {
            return Ok("STRING Peptoniphilus duerdenii");
        }

        if is_valid_string_streptomyces_sp_sirexaae_node_name(node_name) {
            return Ok("STRING Streptomyces sp. SirexAAE");
        }

        if is_valid_string_bacteriovorax_marinus_node_name(node_name) {
            return Ok("STRING Bacteriovorax marinus");
        }

        if is_valid_string_haemophilus_parainfluenzae_node_name(node_name) {
            return Ok("STRING Haemophilus parainfluenzae");
        }

        if is_valid_string_streptococcus_intermedius_node_name(node_name) {
            return Ok("STRING Streptococcus intermedius");
        }

        if is_valid_string_streptococcus_constellatus_pharyngis_c1050_node_name(node_name) {
            return Ok("STRING Streptococcus constellatus pharyngis C1050");
        }

        if is_valid_string_streptococcus_anginosus_c1051_node_name(node_name) {
            return Ok("STRING Streptococcus anginosus C1051");
        }

        if is_valid_string_streptococcus_anginosus_c238_node_name(node_name) {
            return Ok("STRING Streptococcus anginosus C238");
        }

        if is_valid_string_corynebacterium_nuruki_node_name(node_name) {
            return Ok("STRING Corynebacterium nuruki");
        }

        if is_valid_string_xanthomonas_hortorum_node_name(node_name) {
            return Ok("STRING Xanthomonas hortorum");
        }

        if is_valid_string_burkholderiales_bacterium_joshi001_node_name(node_name) {
            return Ok("STRING Burkholderiales bacterium JOSHI001");
        }

        if is_valid_string_microvirga_lotononidis_node_name(node_name) {
            return Ok("STRING Microvirga lotononidis");
        }

        if is_valid_string_herbaspirillum_frisingense_node_name(node_name) {
            return Ok("STRING Herbaspirillum frisingense");
        }

        if is_valid_string_selenomonas_sp_67h29bp_node_name(node_name) {
            return Ok("STRING Selenomonas sp. 67H29BP");
        }

        if is_valid_string_parascardovia_denticolens_dsm10105_node_name(node_name) {
            return Ok("STRING Parascardovia denticolens DSM10105");
        }

        if is_valid_string_eubacterium_yurii_node_name(node_name) {
            return Ok("STRING Eubacterium yurii");
        }

        if is_valid_string_streptococcus_mitis_atcc6249_node_name(node_name) {
            return Ok("STRING Streptococcus mitis ATCC6249");
        }

        if is_valid_string_streptococcus_sp_73h25ap_node_name(node_name) {
            return Ok("STRING Streptococcus sp. 73H25AP");
        }

        if is_valid_string_oscillatoriales_cyanobacterium_node_name(node_name) {
            return Ok("STRING Oscillatoriales cyanobacterium");
        }

        if is_valid_string_clostridium_ultunense_node_name(node_name) {
            return Ok("STRING Clostridium ultunense");
        }

        if is_valid_string_gillisia_limnaea_node_name(node_name) {
            return Ok("STRING Gillisia limnaea");
        }

        if is_valid_string_weeksella_virosa_node_name(node_name) {
            return Ok("STRING Weeksella virosa");
        }

        if is_valid_string_belliella_baltica_node_name(node_name) {
            return Ok("STRING Belliella baltica");
        }

        if is_valid_string_schizosaccharomyces_cryophilus_node_name(node_name) {
            return Ok("STRING Schizosaccharomyces cryophilus");
        }

        if is_valid_string_prevotella_disiens_fb03509an_node_name(node_name) {
            return Ok("STRING Prevotella disiens FB03509AN");
        }

        if is_valid_string_atopobium_vaginae_pb189t14_node_name(node_name) {
            return Ok("STRING Atopobium vaginae PB189T14");
        }

        if is_valid_string_aerococcus_urinae_node_name(node_name) {
            return Ok("STRING Aerococcus urinae");
        }

        if is_valid_string_veillonella_atypica_acs049vsch6_node_name(node_name) {
            return Ok("STRING Veillonella atypica ACS049VSch6");
        }

        if is_valid_string_halobacillus_halophilus_node_name(node_name) {
            return Ok("STRING Halobacillus halophilus");
        }

        if is_valid_string_chloroflexus_sp_y3961_node_name(node_name) {
            return Ok("STRING Chloroflexus sp. Y3961");
        }

        if is_valid_string_cellulophaga_lytica_node_name(node_name) {
            return Ok("STRING Cellulophaga lytica");
        }

        if is_valid_string_ornithobacterium_rhinotracheale_node_name(node_name) {
            return Ok("STRING Ornithobacterium rhinotracheale");
        }

        if is_valid_string_thermaerobacter_subterraneus_node_name(node_name) {
            return Ok("STRING Thermaerobacter subterraneus");
        }

        if is_valid_string_methanobacterium_paludis_node_name(node_name) {
            return Ok("STRING Methanobacterium paludis");
        }

        if is_valid_string_desulfotomaculum_carboxydivorans_node_name(node_name) {
            return Ok("STRING Desulfotomaculum carboxydivorans");
        }

        if is_valid_string_desulfurobacterium_thermolithotrophum_node_name(node_name) {
            return Ok("STRING Desulfurobacterium thermolithotrophum");
        }

        if is_valid_string_treponema_succinifaciens_node_name(node_name) {
            return Ok("STRING Treponema succinifaciens");
        }

        if is_valid_string_marinithermus_hydrothermalis_node_name(node_name) {
            return Ok("STRING Marinithermus hydrothermalis");
        }

        if is_valid_string_cytophaga_fermentans_node_name(node_name) {
            return Ok("STRING Cytophaga fermentans");
        }

        if is_valid_string_thiothrix_nivea_node_name(node_name) {
            return Ok("STRING Thiothrix nivea");
        }

        if is_valid_string_vibrio_scophthalmi_node_name(node_name) {
            return Ok("STRING Vibrio scophthalmi");
        }

        if is_valid_string_acinetobacter_calcoaceticus_node_name(node_name) {
            return Ok("STRING Acinetobacter calcoaceticus");
        }

        if is_valid_string_desulfitobacterium_dichloroeliminans_node_name(node_name) {
            return Ok("STRING Desulfitobacterium dichloroeliminans");
        }

        if is_valid_string_desulfitobacterium_metallireducens_node_name(node_name) {
            return Ok("STRING Desulfitobacterium metallireducens");
        }

        if is_valid_string_streptococcus_parauberis_ncfd2020_node_name(node_name) {
            return Ok("STRING Streptococcus parauberis NCFD2020");
        }

        if is_valid_string_streptococcus_porcinus_node_name(node_name) {
            return Ok("STRING Streptococcus porcinus");
        }

        if is_valid_string_streptococcus_criceti_node_name(node_name) {
            return Ok("STRING Streptococcus criceti");
        }

        if is_valid_string_prevotella_buccae_atcc33574_node_name(node_name) {
            return Ok("STRING Prevotella buccae ATCC33574");
        }

        if is_valid_string_capnocytophaga_ochracea_f0287_node_name(node_name) {
            return Ok("STRING Capnocytophaga ochracea F0287");
        }

        if is_valid_string_prevotella_oralis_atcc33269_node_name(node_name) {
            return Ok("STRING Prevotella oralis ATCC33269");
        }

        if is_valid_string_mycobacterium_sp_jdm601_node_name(node_name) {
            return Ok("STRING Mycobacterium sp. JDM601");
        }

        if is_valid_string_peptoniphilus_rhinitidis_node_name(node_name) {
            return Ok("STRING Peptoniphilus rhinitidis");
        }

        if is_valid_string_gammaproteobacteria_bacterium_imcc3088_node_name(node_name) {
            return Ok("STRING Gammaproteobacteria bacterium IMCC3088");
        }

        if is_valid_string_methyloferula_stellata_node_name(node_name) {
            return Ok("STRING Methyloferula stellata");
        }

        if is_valid_string_ruminococcus_sp_nk3a76_node_name(node_name) {
            return Ok("STRING Ruminococcus sp. NK3A76");
        }

        if is_valid_string_clostridiales_bacterium_nk3b98_node_name(node_name) {
            return Ok("STRING Clostridiales bacterium NK3B98");
        }

        if is_valid_string_erysipelotrichaceae_bacterium_nk3d112_node_name(node_name) {
            return Ok("STRING Erysipelotrichaceae bacterium NK3D112");
        }

        if is_valid_string_treponema_bryantii_node_name(node_name) {
            return Ok("STRING Treponema bryantii");
        }

        if is_valid_string_lachnospiraceae_bacterium_nk4a136_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium NK4A136");
        }

        if is_valid_string_lachnospiraceae_bacterium_nk4a144_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium NK4A144");
        }

        if is_valid_string_lachnospiraceae_bacterium_nk4a179_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium NK4A179");
        }

        if is_valid_string_methanobacterium_lacus_node_name(node_name) {
            return Ok("STRING Methanobacterium lacus");
        }

        if is_valid_string_desulfobacter_postgatei_node_name(node_name) {
            return Ok("STRING Desulfobacter postgatei");
        }

        if is_valid_string_porphyromonas_asaccharolytica_node_name(node_name) {
            return Ok("STRING Porphyromonas asaccharolytica");
        }

        if is_valid_string_anaerococcus_prevotii_acs065vcol13_node_name(node_name) {
            return Ok("STRING Anaerococcus prevotii ACS065VCol13");
        }

        if is_valid_string_peptoniphilus_sp_f0436_node_name(node_name) {
            return Ok("STRING Peptoniphilus sp. F0436");
        }

        if is_valid_string_veillonella_sp_f0412_node_name(node_name) {
            return Ok("STRING Veillonella sp. F0412");
        }

        if is_valid_string_selenomonas_sp_f0430_node_name(node_name) {
            return Ok("STRING Selenomonas sp. F0430");
        }

        if is_valid_string_cyclobacterium_marinum_node_name(node_name) {
            return Ok("STRING Cyclobacterium marinum");
        }

        if is_valid_string_flexibacter_litoralis_node_name(node_name) {
            return Ok("STRING Flexibacter litoralis");
        }

        if is_valid_string_desulfobacca_acetoxidans_node_name(node_name) {
            return Ok("STRING Desulfobacca acetoxidans");
        }

        if is_valid_string_caldithrix_abyssi_node_name(node_name) {
            return Ok("STRING Caldithrix abyssi");
        }

        if is_valid_string_barnesiella_viscericola_node_name(node_name) {
            return Ok("STRING Barnesiella viscericola");
        }

        if is_valid_string_mycoplasma_leachii_node_name(node_name) {
            return Ok("STRING Mycoplasma leachii");
        }

        if is_valid_string_rikenella_microfusus_node_name(node_name) {
            return Ok("STRING Rikenella microfusus");
        }

        if is_valid_string_listeria_ivanovii_pam55_node_name(node_name) {
            return Ok("STRING Listeria ivanovii PAM55");
        }

        if is_valid_string_saccharomonospora_cyanea_node_name(node_name) {
            return Ok("STRING Saccharomonospora cyanea");
        }

        if is_valid_string_saccharomonospora_marina_node_name(node_name) {
            return Ok("STRING Saccharomonospora marina");
        }

        if is_valid_string_saccharomonospora_xinjiangensis_node_name(node_name) {
            return Ok("STRING Saccharomonospora xinjiangensis");
        }

        if is_valid_string_burkholderia_rhizoxinica_node_name(node_name) {
            return Ok("STRING Burkholderia rhizoxinica");
        }

        if is_valid_string_actinobaculum_massiliae_node_name(node_name) {
            return Ok("STRING Actinobaculum massiliae");
        }

        if is_valid_string_actinobaculum_schaalii_node_name(node_name) {
            return Ok("STRING Actinobaculum schaalii");
        }

        if is_valid_string_actinomyces_europaeus_node_name(node_name) {
            return Ok("STRING Actinomyces europaeus");
        }

        if is_valid_string_actinomyces_turicensis_node_name(node_name) {
            return Ok("STRING Actinomyces turicensis");
        }

        if is_valid_string_afipia_broomeae_node_name(node_name) {
            return Ok("STRING Afipia broomeae");
        }

        if is_valid_string_afipia_felis_node_name(node_name) {
            return Ok("STRING Afipia felis");
        }

        if is_valid_string_alloiococcus_otitis_node_name(node_name) {
            return Ok("STRING Alloiococcus otitis");
        }

        if is_valid_string_bergeyella_zoohelcum_node_name(node_name) {
            return Ok("STRING Bergeyella zoohelcum");
        }

        if is_valid_string_dolosigranulum_pigrum_node_name(node_name) {
            return Ok("STRING Dolosigranulum pigrum");
        }

        if is_valid_string_eubacterium_infirmum_node_name(node_name) {
            return Ok("STRING Eubacterium infirmum");
        }

        if is_valid_string_facklamia_hominis_node_name(node_name) {
            return Ok("STRING Facklamia hominis");
        }

        if is_valid_string_facklamia_ignava_node_name(node_name) {
            return Ok("STRING Facklamia ignava");
        }

        if is_valid_string_facklamia_languida_node_name(node_name) {
            return Ok("STRING Facklamia languida");
        }

        if is_valid_string_helcococcus_kunzii_node_name(node_name) {
            return Ok("STRING Helcococcus kunzii");
        }

        if is_valid_string_massilia_timonae_node_name(node_name) {
            return Ok("STRING Massilia timonae");
        }

        if is_valid_string_veillonella_ratti_node_name(node_name) {
            return Ok("STRING Veillonella ratti");
        }

        if is_valid_string_prevotella_micans_node_name(node_name) {
            return Ok("STRING Prevotella micans");
        }

        if is_valid_string_streptococcus_urinalis_fb127cna2_node_name(node_name) {
            return Ok("STRING Streptococcus urinalis FB127CNA2");
        }

        if is_valid_string_turicella_otitidis_node_name(node_name) {
            return Ok("STRING Turicella otitidis");
        }

        if is_valid_string_jonquetella_anthropi_dsm22815_node_name(node_name) {
            return Ok("STRING Jonquetella anthropi DSM22815");
        }

        if is_valid_string_fukomys_damarensis_node_name(node_name) {
            return Ok("STRING Fukomys damarensis");
        }

        if is_valid_string_singulisphaera_acidiphila_node_name(node_name) {
            return Ok("STRING Singulisphaera acidiphila");
        }

        if is_valid_string_muricauda_ruestringensis_node_name(node_name) {
            return Ok("STRING Muricauda ruestringensis");
        }

        if is_valid_string_anaerophaga_thermohalophila_node_name(node_name) {
            return Ok("STRING Anaerophaga thermohalophila");
        }

        if is_valid_string_paenibacillus_polymyxa_sc2_node_name(node_name) {
            return Ok("STRING Paenibacillus polymyxa SC2");
        }

        if is_valid_string_hylemonella_gracilis_atcc19624_node_name(node_name) {
            return Ok("STRING Hylemonella gracilis ATCC19624");
        }

        if is_valid_string_lachnoanaerobaculum_saburreum_dsm3986_node_name(node_name) {
            return Ok("STRING Lachnoanaerobaculum saburreum DSM3986");
        }

        if is_valid_string_kingella_kingae_node_name(node_name) {
            return Ok("STRING Kingella kingae");
        }

        if is_valid_string_lautropia_mirabilis_node_name(node_name) {
            return Ok("STRING Lautropia mirabilis");
        }

        if is_valid_string_pseudoramibacter_alactolyticus_node_name(node_name) {
            return Ok("STRING Pseudoramibacter alactolyticus");
        }

        if is_valid_string_streptococcus_parasanguinis_atcc903_node_name(node_name) {
            return Ok("STRING Streptococcus parasanguinis ATCC903");
        }

        if is_valid_string_streptococcus_oralis_atcc49296_node_name(node_name) {
            return Ok("STRING Streptococcus oralis ATCC49296");
        }

        if is_valid_string_actinomyces_cardiffensis_node_name(node_name) {
            return Ok("STRING Actinomyces cardiffensis");
        }

        if is_valid_string_actinomyces_sp_f0310_node_name(node_name) {
            return Ok("STRING Actinomyces sp. F0310");
        }

        if is_valid_string_leptotrichia_wadei_node_name(node_name) {
            return Ok("STRING Leptotrichia wadei");
        }

        if is_valid_string_actinomyces_sp_f0400_node_name(node_name) {
            return Ok("STRING Actinomyces sp. F0400");
        }

        if is_valid_string_capnocytophaga_sp_f0234_node_name(node_name) {
            return Ok("STRING Capnocytophaga sp. F0234");
        }

        if is_valid_string_centipeda_periodontii_node_name(node_name) {
            return Ok("STRING Centipeda periodontii");
        }

        if is_valid_string_dialister_micraerophilus_node_name(node_name) {
            return Ok("STRING Dialister micraerophilus");
        }

        if is_valid_string_enterococcus_italicus_node_name(node_name) {
            return Ok("STRING Enterococcus italicus");
        }

        if is_valid_string_actinomyces_neuii_bvs029a5_node_name(node_name) {
            return Ok("STRING Actinomyces neuii BVS029A5");
        }

        if is_valid_string_eubacterium_sulci_node_name(node_name) {
            return Ok("STRING Eubacterium sulci");
        }

        if is_valid_string_prevotella_multiformis_node_name(node_name) {
            return Ok("STRING Prevotella multiformis");
        }

        if is_valid_string_streptococcus_peroris_node_name(node_name) {
            return Ok("STRING Streptococcus peroris");
        }

        if is_valid_string_streptococcus_sanguinis_sk49_node_name(node_name) {
            return Ok("STRING Streptococcus sanguinis SK49");
        }

        if is_valid_string_streptococcus_sanguinis_sk355_node_name(node_name) {
            return Ok("STRING Streptococcus sanguinis SK355");
        }

        if is_valid_string_streptococcus_sanguinis_sk1057_node_name(node_name) {
            return Ok("STRING Streptococcus sanguinis SK1057");
        }

        if is_valid_string_prevotella_salivae_node_name(node_name) {
            return Ok("STRING Prevotella salivae");
        }

        if is_valid_string_streptococcus_australis_node_name(node_name) {
            return Ok("STRING Streptococcus australis");
        }

        if is_valid_string_streptococcus_cristatus_node_name(node_name) {
            return Ok("STRING Streptococcus cristatus");
        }

        if is_valid_string_streptococcus_infantis_atcc700779_node_name(node_name) {
            return Ok("STRING Streptococcus infantis ATCC700779");
        }

        if is_valid_string_spirochaeta_africana_node_name(node_name) {
            return Ok("STRING Spirochaeta africana");
        }

        if is_valid_string_lactobacillus_acidophilus_30sc_node_name(node_name) {
            return Ok("STRING Lactobacillus acidophilus 30SC");
        }

        if is_valid_string_anaerobaculum_mobile_node_name(node_name) {
            return Ok("STRING Anaerobaculum mobile");
        }

        if is_valid_string_plautia_stali_node_name(node_name) {
            return Ok("STRING Plautia stali");
        }

        if is_valid_string_moranella_endobia_node_name(node_name) {
            return Ok("STRING Moranella endobia");
        }

        if is_valid_string_eubacterium_limosum_node_name(node_name) {
            return Ok("STRING Eubacterium limosum");
        }

        if is_valid_string_holophaga_foetida_node_name(node_name) {
            return Ok("STRING Holophaga foetida");
        }

        if is_valid_string_gardnerella_vaginalis_101_node_name(node_name) {
            return Ok("STRING Gardnerella vaginalis 101");
        }

        if is_valid_string_streptococcus_downei_node_name(node_name) {
            return Ok("STRING Streptococcus downei");
        }

        if is_valid_string_streptococcus_sp_f0418_node_name(node_name) {
            return Ok("STRING Streptococcus sp. F0418");
        }

        if is_valid_string_oribacterium_sp_f0425_node_name(node_name) {
            return Ok("STRING Oribacterium sp. F0425");
        }

        if is_valid_string_streptococcus_vestibularis_node_name(node_name) {
            return Ok("STRING Streptococcus vestibularis");
        }

        if is_valid_string_staphylococcus_pettenkoferi_node_name(node_name) {
            return Ok("STRING Staphylococcus pettenkoferi");
        }

        if is_valid_string_nonlabens_ulvanivorans_node_name(node_name) {
            return Ok("STRING Nonlabens ulvanivorans");
        }

        if is_valid_string_treponema_brennaborense_node_name(node_name) {
            return Ok("STRING Treponema brennaborense");
        }

        if is_valid_string_helicobacter_pylori_southafrica7_node_name(node_name) {
            return Ok("STRING Helicobacter pylori SouthAfrica7");
        }

        if is_valid_string_treponema_saccharophilum_node_name(node_name) {
            return Ok("STRING Treponema saccharophilum");
        }

        if is_valid_string_leuconostoc_fallax_node_name(node_name) {
            return Ok("STRING Leuconostoc fallax");
        }

        if is_valid_string_eremococcus_coleocola_acs139vcol8_node_name(node_name) {
            return Ok("STRING Eremococcus coleocola ACS139VCol8");
        }

        if is_valid_string_peptoniphilus_harei_node_name(node_name) {
            return Ok("STRING Peptoniphilus harei");
        }

        if is_valid_string_lactobacillus_oris_pb013t23_node_name(node_name) {
            return Ok("STRING Lactobacillus oris PB013T23");
        }

        if is_valid_string_clostridium_sp_hgf2_node_name(node_name) {
            return Ok("STRING Clostridium sp. HGF2");
        }

        if is_valid_string_alistipes_sp_hgb5_node_name(node_name) {
            return Ok("STRING Alistipes sp. HGB5");
        }

        if is_valid_string_prevotella_dentalis_node_name(node_name) {
            return Ok("STRING Prevotella dentalis");
        }

        if is_valid_string_actinokineospora_sp_eg49_node_name(node_name) {
            return Ok("STRING Actinokineospora sp. EG49");
        }

        if is_valid_string_syntrophorhabdus_aromaticivorans_node_name(node_name) {
            return Ok("STRING Syntrophorhabdus aromaticivorans");
        }

        if is_valid_string_alphaproteobacterium_sar116_node_name(node_name) {
            return Ok("STRING Alphaproteobacterium SAR116");
        }

        if is_valid_string_streptococcus_pseudoporcinus_node_name(node_name) {
            return Ok("STRING Streptococcus pseudoporcinus");
        }

        if is_valid_string_dialister_microaerophilus_node_name(node_name) {
            return Ok("STRING Dialister microaerophilus");
        }

        if is_valid_string_ewingella_americana_node_name(node_name) {
            return Ok("STRING Ewingella americana");
        }

        if is_valid_string_leclercia_adecarboxylata_node_name(node_name) {
            return Ok("STRING Leclercia adecarboxylata");
        }

        if is_valid_string_pseudovibrio_sp_fobeg1_node_name(node_name) {
            return Ok("STRING Pseudovibrio sp. FOBEG1");
        }

        if is_valid_string_weissella_cibaria_node_name(node_name) {
            return Ok("STRING Weissella cibaria");
        }

        if is_valid_string_pseudomonas_sp_cf149_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. CF149");
        }

        if is_valid_string_lysobacter_arseniciresistens_node_name(node_name) {
            return Ok("STRING Lysobacter arseniciresistens");
        }

        if is_valid_string_lactobacillus_coryniformis_kctc3167_node_name(node_name) {
            return Ok("STRING Lactobacillus coryniformis KCTC3167");
        }

        if is_valid_string_desulfosporosinus_sp_ot_node_name(node_name) {
            return Ok("STRING Desulfosporosinus sp. OT");
        }

        if is_valid_string_sediminibacterium_sp_or53_node_name(node_name) {
            return Ok("STRING Sediminibacterium sp. OR53");
        }

        if is_valid_string_xanthomonas_vesicatoria_node_name(node_name) {
            return Ok("STRING Xanthomonas vesicatoria");
        }

        if is_valid_string_adhaeribacter_aquaticus_node_name(node_name) {
            return Ok("STRING Adhaeribacter aquaticus");
        }

        if is_valid_string_caldilinea_aerophila_node_name(node_name) {
            return Ok("STRING Caldilinea aerophila");
        }

        if is_valid_string_capnocytophaga_cynodegmi_node_name(node_name) {
            return Ok("STRING Capnocytophaga cynodegmi");
        }

        if is_valid_string_deinococcus_pimensis_node_name(node_name) {
            return Ok("STRING Deinococcus pimensis");
        }

        if is_valid_string_echinicola_vietnamensis_node_name(node_name) {
            return Ok("STRING Echinicola vietnamensis");
        }

        if is_valid_string_joostella_marina_node_name(node_name) {
            return Ok("STRING Joostella marina");
        }

        if is_valid_string_meiothermus_chliarophilus_node_name(node_name) {
            return Ok("STRING Meiothermus chliarophilus");
        }

        if is_valid_string_orenia_marismortui_node_name(node_name) {
            return Ok("STRING Orenia marismortui");
        }

        if is_valid_string_owenweeksia_hongkongensis_node_name(node_name) {
            return Ok("STRING Owenweeksia hongkongensis");
        }

        if is_valid_string_promicromonospora_kroppenstedtii_node_name(node_name) {
            return Ok("STRING Promicromonospora kroppenstedtii");
        }

        if is_valid_string_terriglobus_roseus_node_name(node_name) {
            return Ok("STRING Terriglobus roseus");
        }

        if is_valid_string_thermanaerovibrio_velox_node_name(node_name) {
            return Ok("STRING Thermanaerovibrio velox");
        }

        if is_valid_string_anaerolinea_thermophila_node_name(node_name) {
            return Ok("STRING Anaerolinea thermophila");
        }

        if is_valid_string_haloplanus_natans_node_name(node_name) {
            return Ok("STRING Haloplanus natans");
        }

        if is_valid_string_halonatronum_saccharophilum_node_name(node_name) {
            return Ok("STRING Halonatronum saccharophilum");
        }

        if is_valid_string_alkaliflexus_imshenetskii_node_name(node_name) {
            return Ok("STRING Alkaliflexus imshenetskii");
        }

        if is_valid_string_streptococcus_oralis_uo5_node_name(node_name) {
            return Ok("STRING Streptococcus oralis Uo5");
        }

        if is_valid_string_synechocystis_sp_pcc7509_node_name(node_name) {
            return Ok("STRING Synechocystis sp. PCC7509");
        }

        if is_valid_string_leuconostoc_gelidum_kctc3527_node_name(node_name) {
            return Ok("STRING Leuconostoc gelidum KCTC3527");
        }

        if is_valid_string_selenomonas_ruminantium_lactilytica_node_name(node_name) {
            return Ok("STRING Selenomonas ruminantium lactilytica");
        }

        if is_valid_string_bartonella_sp_r4_node_name(node_name) {
            return Ok("STRING Bartonella sp. R4");
        }

        if is_valid_string_saccharomonospora_glauca_node_name(node_name) {
            return Ok("STRING Saccharomonospora glauca");
        }

        if is_valid_string_clostridium_botulinum_bkt015925_node_name(node_name) {
            return Ok("STRING Clostridium botulinum BKT015925");
        }

        if is_valid_string_solitalea_canadensis_node_name(node_name) {
            return Ok("STRING Solitalea canadensis");
        }

        if is_valid_string_sulfurimonas_gotlandica_node_name(node_name) {
            return Ok("STRING Sulfurimonas gotlandica");
        }

        if is_valid_string_emticicia_oligotrophica_node_name(node_name) {
            return Ok("STRING Emticicia oligotrophica");
        }

        if is_valid_string_flectobacillus_major_node_name(node_name) {
            return Ok("STRING Flectobacillus major");
        }

        if is_valid_string_myroides_odoratus_node_name(node_name) {
            return Ok("STRING Myroides odoratus");
        }

        if is_valid_string_patulibacter_minatonensis_node_name(node_name) {
            return Ok("STRING Patulibacter minatonensis");
        }

        if is_valid_string_niabella_soli_node_name(node_name) {
            return Ok("STRING Niabella soli");
        }

        if is_valid_string_bartonella_senegalensis_node_name(node_name) {
            return Ok("STRING Bartonella senegalensis");
        }

        if is_valid_string_pseudomonas_brassicacearum_node_name(node_name) {
            return Ok("STRING Pseudomonas brassicacearum");
        }

        if is_valid_string_alcanivorax_dieselolei_node_name(node_name) {
            return Ok("STRING Alcanivorax dieselolei");
        }

        if is_valid_string_arthrobacter_phenanthrenivorans_node_name(node_name) {
            return Ok("STRING Arthrobacter phenanthrenivorans");
        }

        if is_valid_string_sulfolobus_islandicus_node_name(node_name) {
            return Ok("STRING Sulfolobus islandicus");
        }

        if is_valid_string_fructobacillus_fructosus_node_name(node_name) {
            return Ok("STRING Fructobacillus fructosus");
        }

        if is_valid_string_clostridium_saccharoperbutylacetonicum_node_name(node_name) {
            return Ok("STRING Clostridium saccharoperbutylacetonicum");
        }

        if is_valid_string_halococcus_morrhuae_node_name(node_name) {
            return Ok("STRING Halococcus morrhuae");
        }

        if is_valid_string_acetobacterium_woodii_node_name(node_name) {
            return Ok("STRING Acetobacterium woodii");
        }

        if is_valid_string_mycobacterium_rhodesiae_js60_node_name(node_name) {
            return Ok("STRING Mycobacterium rhodesiae JS60");
        }

        if is_valid_string_serratia_sp_m24t3_node_name(node_name) {
            return Ok("STRING Serratia sp. M24T3");
        }

        if is_valid_string_pantoea_ananatis_aj13355_node_name(node_name) {
            return Ok("STRING Pantoea ananatis AJ13355");
        }

        if is_valid_string_thermocrinis_ruber_node_name(node_name) {
            return Ok("STRING Thermocrinis ruber");
        }

        if is_valid_string_planococcus_donghaensis_node_name(node_name) {
            return Ok("STRING Planococcus donghaensis");
        }

        if is_valid_string_desulfosarcina_sp_bus5_node_name(node_name) {
            return Ok("STRING Desulfosarcina sp. BuS5");
        }

        if is_valid_string_acidianus_hospitalis_node_name(node_name) {
            return Ok("STRING Acidianus hospitalis");
        }

        if is_valid_string_aminobacter_sp_j41_node_name(node_name) {
            return Ok("STRING Aminobacter sp. J41");
        }

        if is_valid_string_mesorhizobium_loti_r88b_node_name(node_name) {
            return Ok("STRING Mesorhizobium loti R88b");
        }

        if is_valid_string_sinorhizobium_arboris_node_name(node_name) {
            return Ok("STRING Sinorhizobium arboris");
        }

        if is_valid_string_paracoccus_pantotrophus_node_name(node_name) {
            return Ok("STRING Paracoccus pantotrophus");
        }

        if is_valid_string_pseudoxanthomonas_suwonensis_j43_node_name(node_name) {
            return Ok("STRING Pseudoxanthomonas suwonensis J43");
        }

        if is_valid_string_bacillus_sp_j33_node_name(node_name) {
            return Ok("STRING Bacillus sp. J33");
        }

        if is_valid_string_bacillus_sp_j37_node_name(node_name) {
            return Ok("STRING Bacillus sp. J37");
        }

        if is_valid_string_cellulosimicrobium_cellulans_node_name(node_name) {
            return Ok("STRING Cellulosimicrobium cellulans");
        }

        if is_valid_string_chelativorans_sp_j32_node_name(node_name) {
            return Ok("STRING Chelativorans sp. J32");
        }

        if is_valid_string_paenibacillus_sp_j14_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. J14");
        }

        if is_valid_string_paracoccus_sp_j39_node_name(node_name) {
            return Ok("STRING Paracoccus sp. J39");
        }

        if is_valid_string_luteimonas_sp_j29_node_name(node_name) {
            return Ok("STRING Luteimonas sp. J29");
        }

        if is_valid_string_nocardioides_sp_j54_node_name(node_name) {
            return Ok("STRING Nocardioides sp. J54");
        }

        if is_valid_string_caldanaerobius_polysaccharolyticus_node_name(node_name) {
            return Ok("STRING Caldanaerobius polysaccharolyticus");
        }

        if is_valid_string_rhizopus_delemar_node_name(node_name) {
            return Ok("STRING Rhizopus delemar");
        }

        if is_valid_string_rhizobium_leguminosarum_248_node_name(node_name) {
            return Ok("STRING Rhizobium leguminosarum 248");
        }

        if is_valid_string_lactobacillus_farciminis_node_name(node_name) {
            return Ok("STRING Lactobacillus farciminis");
        }

        if is_valid_string_streptococcus_parauberis_kctc11537_node_name(node_name) {
            return Ok("STRING Streptococcus parauberis KCTC11537");
        }

        if is_valid_string_helicobacter_felis_node_name(node_name) {
            return Ok("STRING Helicobacter felis");
        }

        if is_valid_string_mogibacterium_sp_cm50_node_name(node_name) {
            return Ok("STRING Mogibacterium sp. CM50");
        }

        if is_valid_string_bradyrhizobium_genosp_node_name(node_name) {
            return Ok("STRING Bradyrhizobium genosp.");
        }

        if is_valid_string_actinomyces_sp_icm47_node_name(node_name) {
            return Ok("STRING Actinomyces sp. ICM47");
        }

        if is_valid_string_atopobium_sp_bs2_node_name(node_name) {
            return Ok("STRING Atopobium sp. BS2");
        }

        if is_valid_string_selenomonas_sp_fobrc6_node_name(node_name) {
            return Ok("STRING Selenomonas sp. FOBRC6");
        }

        if is_valid_string_selenomonas_sp_fobrc9_node_name(node_name) {
            return Ok("STRING Selenomonas sp. FOBRC9");
        }

        if is_valid_string_shuttleworthia_sp_msx8b_node_name(node_name) {
            return Ok("STRING Shuttleworthia sp. MSX8B");
        }

        if is_valid_string_streptococcus_sp_cm6_node_name(node_name) {
            return Ok("STRING Streptococcus sp. CM6");
        }

        if is_valid_string_veillonella_sp_as16_node_name(node_name) {
            return Ok("STRING Veillonella sp. AS16");
        }

        if is_valid_string_lachnoanaerobaculum_sp_msx33_node_name(node_name) {
            return Ok("STRING Lachnoanaerobaculum sp. MSX33");
        }

        if is_valid_string_taylorella_equigenitalis_node_name(node_name) {
            return Ok("STRING Taylorella equigenitalis");
        }

        if is_valid_string_deinococcus_peraridilitoris_node_name(node_name) {
            return Ok("STRING Deinococcus peraridilitoris");
        }

        if is_valid_string_clostridiales_bacterium_9400853_node_name(node_name) {
            return Ok("STRING Clostridiales bacterium 9400853");
        }

        if is_valid_string_clostridiales_bacterium_9401234_node_name(node_name) {
            return Ok("STRING Clostridiales bacterium 9401234");
        }

        if is_valid_string_clostridiales_bacterium_9403326_node_name(node_name) {
            return Ok("STRING Clostridiales bacterium 9403326");
        }

        if is_valid_string_anaerococcus_sp_9402080_node_name(node_name) {
            return Ok("STRING Anaerococcus sp. 9402080");
        }

        if is_valid_string_bacteroidetes_bacterium_scgc_aaa027n21_node_name(node_name) {
            return Ok("STRING Bacteroidetes bacterium SCGC AAA027N21");
        }

        if is_valid_string_acetobacter_pasteurianus_3p3_node_name(node_name) {
            return Ok("STRING Acetobacter pasteurianus 3P3");
        }

        if is_valid_string_desulfovibrio_sp_x2_node_name(node_name) {
            return Ok("STRING Desulfovibrio sp. X2");
        }

        if is_valid_string_bacillus_coagulans_26_node_name(node_name) {
            return Ok("STRING Bacillus coagulans 26");
        }

        if is_valid_string_lactobacillus_fructivorans_node_name(node_name) {
            return Ok("STRING Lactobacillus fructivorans");
        }

        if is_valid_string_thermobrachium_celere_node_name(node_name) {
            return Ok("STRING Thermobrachium celere");
        }

        if is_valid_string_burkholderia_sp_wsm2230_node_name(node_name) {
            return Ok("STRING Burkholderia sp. WSM2230");
        }

        if is_valid_string_hippea_jasoniae_node_name(node_name) {
            return Ok("STRING Hippea jasoniae");
        }

        if is_valid_string_hippea_alviniae_node_name(node_name) {
            return Ok("STRING Hippea alviniae");
        }

        if is_valid_string_hippea_sp_km1_node_name(node_name) {
            return Ok("STRING Hippea sp. KM1");
        }

        if is_valid_string_arcobacter_butzleri_ed1_node_name(node_name) {
            return Ok("STRING Arcobacter butzleri ED1");
        }

        if is_valid_string_arcobacter_sp_l_node_name(node_name) {
            return Ok("STRING Arcobacter sp. L");
        }

        if is_valid_string_actinomyces_sp_f0384_node_name(node_name) {
            return Ok("STRING Actinomyces sp. F0384");
        }

        if is_valid_string_lactobacillus_oris_f0423_node_name(node_name) {
            return Ok("STRING Lactobacillus oris F0423");
        }

        if is_valid_string_veillonella_sp_f0422_node_name(node_name) {
            return Ok("STRING Veillonella sp. F0422");
        }

        if is_valid_string_parvimonas_sp_f0440_node_name(node_name) {
            return Ok("STRING Parvimonas sp. F0440");
        }

        if is_valid_string_tetragenococcus_halophilus_node_name(node_name) {
            return Ok("STRING Tetragenococcus halophilus");
        }

        if is_valid_string_vibrio_brasiliensis_node_name(node_name) {
            return Ok("STRING Vibrio brasiliensis");
        }

        if is_valid_string_vibrio_sinaloensis_node_name(node_name) {
            return Ok("STRING Vibrio sinaloensis");
        }

        if is_valid_string_corynebacterium_ulcerans_node_name(node_name) {
            return Ok("STRING Corynebacterium ulcerans");
        }

        if is_valid_string_ignavibacterium_album_node_name(node_name) {
            return Ok("STRING Ignavibacterium album");
        }

        if is_valid_string_imtechella_halotolerans_node_name(node_name) {
            return Ok("STRING Imtechella halotolerans");
        }

        if is_valid_string_oceanobacillus_massiliensis_node_name(node_name) {
            return Ok("STRING Oceanobacillus massiliensis");
        }

        if is_valid_string_salpingoeca_rosetta_node_name(node_name) {
            return Ok("STRING Salpingoeca rosetta");
        }

        if is_valid_string_symbiobacter_mobilis_node_name(node_name) {
            return Ok("STRING Symbiobacter mobilis");
        }

        if is_valid_string_burkholderia_dilworthii_node_name(node_name) {
            return Ok("STRING Burkholderia dilworthii");
        }

        if is_valid_string_avibacterium_paragallinarum_72_node_name(node_name) {
            return Ok("STRING Avibacterium paragallinarum 72");
        }

        if is_valid_string_streptomyces_venezuelae_node_name(node_name) {
            return Ok("STRING Streptomyces venezuelae");
        }

        if is_valid_string_cupriavidus_taiwanensis_node_name(node_name) {
            return Ok("STRING Cupriavidus taiwanensis");
        }

        if is_valid_string_microbacterium_testaceum_node_name(node_name) {
            return Ok("STRING Microbacterium testaceum");
        }

        if is_valid_string_flavobacteriaceae_bacterium_hqm9_node_name(node_name) {
            return Ok("STRING Flavobacteriaceae bacterium HQM9");
        }

        if is_valid_string_morus_notabilis_node_name(node_name) {
            return Ok("STRING Morus notabilis");
        }

        if is_valid_string_acinetobacter_sp_nctc_7422_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. NCTC 7422");
        }

        if is_valid_string_acinetobacter_lwoffii_nctc_5866_node_name(node_name) {
            return Ok("STRING Acinetobacter lwoffii NCTC 5866");
        }

        if is_valid_string_acinetobacter_ursingii_node_name(node_name) {
            return Ok("STRING Acinetobacter ursingii");
        }

        if is_valid_string_streptacidiphilus_rugosus_node_name(node_name) {
            return Ok("STRING Streptacidiphilus rugosus");
        }

        if is_valid_string_ornithinibacillus_scapharcae_node_name(node_name) {
            return Ok("STRING Ornithinibacillus scapharcae");
        }

        if is_valid_string_ruegeria_conchae_node_name(node_name) {
            return Ok("STRING Ruegeria conchae");
        }

        if is_valid_string_campylobacter_fetus_venerealis_node_name(node_name) {
            return Ok("STRING Campylobacter fetus venerealis");
        }

        if is_valid_string_lacinutrix_sp_5h374_node_name(node_name) {
            return Ok("STRING Lacinutrix sp. 5H374");
        }

        if is_valid_string_glaciecola_sp_4h37ye5_node_name(node_name) {
            return Ok("STRING Glaciecola sp. 4H37YE5");
        }

        if is_valid_string_dokdonia_sp_4h375_node_name(node_name) {
            return Ok("STRING Dokdonia sp. 4H375");
        }

        if is_valid_string_rubrivivax_gelatinosus_node_name(node_name) {
            return Ok("STRING Rubrivivax gelatinosus");
        }

        if is_valid_string_novosphingobium_nitrogenifigens_node_name(node_name) {
            return Ok("STRING Novosphingobium nitrogenifigens");
        }

        if is_valid_string_saprospira_grandis_node_name(node_name) {
            return Ok("STRING Saprospira grandis");
        }

        if is_valid_string_staphylococcus_pseudintermedius_node_name(node_name) {
            return Ok("STRING Staphylococcus pseudintermedius");
        }

        if is_valid_string_heterobasidion_irregulare_node_name(node_name) {
            return Ok("STRING Heterobasidion irregulare");
        }

        if is_valid_string_vulcanisaeta_moutnovskia_node_name(node_name) {
            return Ok("STRING Vulcanisaeta moutnovskia");
        }

        if is_valid_string_ruegeria_halocynthiae_node_name(node_name) {
            return Ok("STRING Ruegeria halocynthiae");
        }

        if is_valid_string_gillisia_sp_cal575_node_name(node_name) {
            return Ok("STRING Gillisia sp. CAL575");
        }

        if is_valid_string_paenibacillus_terrae_node_name(node_name) {
            return Ok("STRING Paenibacillus terrae");
        }

        if is_valid_string_staphylococcus_agnetis_node_name(node_name) {
            return Ok("STRING Staphylococcus agnetis");
        }

        if is_valid_string_odyssella_thessalonicensis_node_name(node_name) {
            return Ok("STRING Odyssella thessalonicensis");
        }

        if is_valid_string_caldalkalibacillus_thermarum_node_name(node_name) {
            return Ok("STRING Caldalkalibacillus thermarum");
        }

        if is_valid_string_rubrivivax_benzoatilyticus_node_name(node_name) {
            return Ok("STRING Rubrivivax benzoatilyticus");
        }

        if is_valid_string_lebetimonas_sp_js170_node_name(node_name) {
            return Ok("STRING Lebetimonas sp. JS170");
        }

        if is_valid_string_rhizobium_grahamii_node_name(node_name) {
            return Ok("STRING Rhizobium grahamii");
        }

        if is_valid_string_polymorphum_gilvum_node_name(node_name) {
            return Ok("STRING Polymorphum gilvum");
        }

        if is_valid_string_riemerella_anatipestifer_node_name(node_name) {
            return Ok("STRING Riemerella anatipestifer");
        }

        if is_valid_string_saccharopolyspora_spinosa_node_name(node_name) {
            return Ok("STRING Saccharopolyspora spinosa");
        }

        if is_valid_string_youngiibacter_fragilis_node_name(node_name) {
            return Ok("STRING Youngiibacter fragilis");
        }

        if is_valid_string_streptococcus_suis_r61_node_name(node_name) {
            return Ok("STRING Streptococcus suis R61");
        }

        if is_valid_string_streptomyces_griseoaurantiacus_node_name(node_name) {
            return Ok("STRING Streptomyces griseoaurantiacus");
        }

        if is_valid_string_bacillus_methanolicus_pb1_node_name(node_name) {
            return Ok("STRING Bacillus methanolicus PB1");
        }

        if is_valid_string_desmospora_sp_8437_node_name(node_name) {
            return Ok("STRING Desmospora sp. 8437");
        }

        if is_valid_string_peptoniphilus_indolicus_node_name(node_name) {
            return Ok("STRING Peptoniphilus indolicus");
        }

        if is_valid_string_prevotella_nigrescens_node_name(node_name) {
            return Ok("STRING Prevotella nigrescens");
        }

        if is_valid_string_prevotella_pallens_node_name(node_name) {
            return Ok("STRING Prevotella pallens");
        }

        if is_valid_string_porphyromonas_sp_kle1280_node_name(node_name) {
            return Ok("STRING Porphyromonas sp. KLE1280");
        }

        if is_valid_string_streptococcus_infantis_x_node_name(node_name) {
            return Ok("STRING Streptococcus infantis X");
        }

        if is_valid_string_bacteroides_nordii_node_name(node_name) {
            return Ok("STRING Bacteroides nordii");
        }

        if is_valid_string_aeromonas_veronii_node_name(node_name) {
            return Ok("STRING Aeromonas veronii");
        }

        if is_valid_string_leucothrix_mucor_node_name(node_name) {
            return Ok("STRING Leucothrix mucor");
        }

        if is_valid_string_halomonas_sp_td01_node_name(node_name) {
            return Ok("STRING Halomonas sp. TD01");
        }

        if is_valid_string_clostridium_colicanis_node_name(node_name) {
            return Ok("STRING Clostridium colicanis");
        }

        if is_valid_string_clostridium_innocuum_node_name(node_name) {
            return Ok("STRING Clostridium innocuum");
        }

        if is_valid_string_eggerthia_catenaformis_node_name(node_name) {
            return Ok("STRING Eggerthia catenaformis");
        }

        if is_valid_string_parabacteroides_johnsonii_node_name(node_name) {
            return Ok("STRING Parabacteroides johnsonii");
        }

        if is_valid_string_selenomonas_sp_f0473_node_name(node_name) {
            return Ok("STRING Selenomonas sp. F0473");
        }

        if is_valid_string_streptococcus_sp_f0442_node_name(node_name) {
            return Ok("STRING Streptococcus sp. F0442");
        }

        if is_valid_string_burkholderia_gladioli_node_name(node_name) {
            return Ok("STRING Burkholderia gladioli");
        }

        if is_valid_string_leisingera_daeponensis_node_name(node_name) {
            return Ok("STRING Leisingera daeponensis");
        }

        if is_valid_string_leisingera_caerulea_node_name(node_name) {
            return Ok("STRING Leisingera caerulea");
        }

        if is_valid_string_pseudophaeobacter_arcticus_node_name(node_name) {
            return Ok("STRING Pseudophaeobacter arcticus");
        }

        if is_valid_string_leisingera_aquimarina_node_name(node_name) {
            return Ok("STRING Leisingera aquimarina");
        }

        if is_valid_string_thermoproteus_uzoniensis_node_name(node_name) {
            return Ok("STRING Thermoproteus uzoniensis");
        }

        if is_valid_string_methyloversatilis_universalis_fam5_node_name(node_name) {
            return Ok("STRING Methyloversatilis universalis FAM5");
        }

        if is_valid_string_megasphaera_sp_upii135e_node_name(node_name) {
            return Ok("STRING Megasphaera sp. UPII135E");
        }

        if is_valid_string_streptococcus_anginosus_sk52_node_name(node_name) {
            return Ok("STRING Streptococcus anginosus SK52");
        }

        if is_valid_string_streptococcus_mitis_2_sk95_node_name(node_name) {
            return Ok("STRING Streptococcus mitis 2 SK95");
        }

        if is_valid_string_cryobacterium_roopkundense_node_name(node_name) {
            return Ok("STRING Cryobacterium roopkundense");
        }

        if is_valid_string_photobacterium_leiognathi_mandapamensis_node_name(node_name) {
            return Ok("STRING Photobacterium leiognathi mandapamensis");
        }

        if is_valid_string_pseudomonas_mendocina_nk01_node_name(node_name) {
            return Ok("STRING Pseudomonas mendocina NK01");
        }

        if is_valid_string_psychrobacter_sp_1501_node_name(node_name) {
            return Ok("STRING Psychrobacter sp. 1501");
        }

        if is_valid_string_leisingera_sp_ang1_node_name(node_name) {
            return Ok("STRING Leisingera sp. ANG1");
        }

        if is_valid_string_prevotella_stercorea_node_name(node_name) {
            return Ok("STRING Prevotella stercorea");
        }

        if is_valid_string_pelagibacter_sp_imcc9063_node_name(node_name) {
            return Ok("STRING Pelagibacter sp. IMCC9063");
        }

        if is_valid_string_helicobacter_bizzozeronii_node_name(node_name) {
            return Ok("STRING Helicobacter bizzozeronii");
        }

        if is_valid_string_solibacillus_silvestris_node_name(node_name) {
            return Ok("STRING Solibacillus silvestris");
        }

        if is_valid_string_streptomyces_cattleya_node_name(node_name) {
            return Ok("STRING Streptomyces cattleya");
        }

        if is_valid_string_achromobacter_insuavis_node_name(node_name) {
            return Ok("STRING Achromobacter insuavis");
        }

        if is_valid_string_mesoflavibacter_zeaxanth_s86_node_name(node_name) {
            return Ok("STRING Mesoflavibacter zeaxanth. S86");
        }

        if is_valid_string_alteromonas_macleodii_black_node_name(node_name) {
            return Ok("STRING Alteromonas macleodii Black");
        }

        if is_valid_string_collimonas_fungivorans_node_name(node_name) {
            return Ok("STRING Collimonas fungivorans");
        }

        if is_valid_string_buchnera_aphidicola_ua_node_name(node_name) {
            return Ok("STRING Buchnera aphidicola Ua");
        }

        if is_valid_string_gallibacterium_anatis_node_name(node_name) {
            return Ok("STRING Gallibacterium anatis");
        }

        if is_valid_string_buchnera_aphidicola_ak_node_name(node_name) {
            return Ok("STRING Buchnera aphidicola Ak");
        }

        if is_valid_string_pseudomonas_putida_csv86_node_name(node_name) {
            return Ok("STRING Pseudomonas putida CSV86");
        }

        if is_valid_string_streptococcus_oralis_sk255_node_name(node_name) {
            return Ok("STRING Streptococcus oralis SK255");
        }

        if is_valid_string_streptococcus_infantis_sk1076_node_name(node_name) {
            return Ok("STRING Streptococcus infantis SK1076");
        }

        if is_valid_string_ogataea_parapolymorpha_node_name(node_name) {
            return Ok("STRING Ogataea parapolymorpha");
        }

        if is_valid_string_trabulsiella_guamensis_node_name(node_name) {
            return Ok("STRING Trabulsiella guamensis");
        }

        if is_valid_string_tatumella_ptyseos_node_name(node_name) {
            return Ok("STRING Tatumella ptyseos");
        }

        if is_valid_string_leminorella_grimontii_node_name(node_name) {
            return Ok("STRING Leminorella grimontii");
        }

        if is_valid_string_kluyvera_ascorbata_node_name(node_name) {
            return Ok("STRING Kluyvera ascorbata");
        }

        if is_valid_string_buttiauxella_agrestis_node_name(node_name) {
            return Ok("STRING Buttiauxella agrestis");
        }

        if is_valid_string_metallosphaera_cuprina_node_name(node_name) {
            return Ok("STRING Metallosphaera cuprina");
        }

        if is_valid_string_mycoplasma_gallisepticum_s6_node_name(node_name) {
            return Ok("STRING Mycoplasma gallisepticum S6");
        }

        if is_valid_string_oscillibacter_ruminantium_node_name(node_name) {
            return Ok("STRING Oscillibacter ruminantium");
        }

        if is_valid_string_paenibacillus_elgii_node_name(node_name) {
            return Ok("STRING Paenibacillus elgii");
        }

        if is_valid_string_sphingomonas_sp_s17_node_name(node_name) {
            return Ok("STRING Sphingomonas sp. S17");
        }

        if is_valid_string_pusillimonas_sp_t77_node_name(node_name) {
            return Ok("STRING Pusillimonas sp. T77");
        }

        if is_valid_string_streptococcus_mitis_sk1080_node_name(node_name) {
            return Ok("STRING Streptococcus mitis SK1080");
        }

        if is_valid_string_myroides_injenensis_node_name(node_name) {
            return Ok("STRING Myroides injenensis");
        }

        if is_valid_string_taylorella_asinigenitalis_node_name(node_name) {
            return Ok("STRING Taylorella asinigenitalis");
        }

        if is_valid_string_acetonema_longum_node_name(node_name) {
            return Ok("STRING Acetonema longum");
        }

        if is_valid_string_buchnera_aphidicola_g002_node_name(node_name) {
            return Ok("STRING Buchnera aphidicola G002");
        }

        if is_valid_string_methylophaga_aminisulfidivorans_node_name(node_name) {
            return Ok("STRING Methylophaga aminisulfidivorans");
        }

        if is_valid_string_nannospalax_galili_node_name(node_name) {
            return Ok("STRING Nannospalax galili");
        }

        if is_valid_string_endozoicomonas_montiporae_node_name(node_name) {
            return Ok("STRING Endozoicomonas montiporae");
        }

        if is_valid_string_sporosarcina_newyorkensis_node_name(node_name) {
            return Ok("STRING Sporosarcina newyorkensis");
        }

        if is_valid_string_gordonia_alkanivorans_node_name(node_name) {
            return Ok("STRING Gordonia alkanivorans");
        }

        if is_valid_string_listeria_monocytogenes_scott_node_name(node_name) {
            return Ok("STRING Listeria monocytogenes Scott");
        }

        if is_valid_string_enterobacter_aerogenes_node_name(node_name) {
            return Ok("STRING Enterobacter aerogenes");
        }

        if is_valid_string_neorhizobium_galegae_orientalis_node_name(node_name) {
            return Ok("STRING Neorhizobium galegae orientalis");
        }

        if is_valid_string_neorhizobium_galegae_officinalis_node_name(node_name) {
            return Ok("STRING Neorhizobium galegae officinalis");
        }

        if is_valid_string_haemophilus_haemolyticus_m19501_node_name(node_name) {
            return Ok("STRING Haemophilus haemolyticus M19501");
        }

        if is_valid_string_haemophilus_haemolyticus_m21621_node_name(node_name) {
            return Ok("STRING Haemophilus haemolyticus M21621");
        }

        if is_valid_string_haemophilus_haemolyticus_m21639_node_name(node_name) {
            return Ok("STRING Haemophilus haemolyticus M21639");
        }

        if is_valid_string_arthromitus_sp_sfbmousejapan_node_name(node_name) {
            return Ok("STRING Arthromitus sp. SFBmouseJapan");
        }

        if is_valid_string_acinetobacter_sp_p838_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. P838");
        }

        if is_valid_string_kocuria_rhizophila_p74_node_name(node_name) {
            return Ok("STRING Kocuria rhizophila P74");
        }

        if is_valid_string_sphingomonas_sp_kc8_node_name(node_name) {
            return Ok("STRING Sphingomonas sp. KC8");
        }

        if is_valid_string_caloramator_sp_ald01_node_name(node_name) {
            return Ok("STRING Caloramator sp. ALD01");
        }

        if is_valid_string_ralstonia_solanacearum_po82_node_name(node_name) {
            return Ok("STRING Ralstonia solanacearum Po82");
        }

        if is_valid_string_microlunatus_phosphovorus_node_name(node_name) {
            return Ok("STRING Microlunatus phosphovorus");
        }

        if is_valid_string_aeromicrobium_massiliense_node_name(node_name) {
            return Ok("STRING Aeromicrobium massiliense");
        }

        if is_valid_string_alistipes_senegalensis_node_name(node_name) {
            return Ok("STRING Alistipes senegalensis");
        }

        if is_valid_string_anaerococcus_senegalensis_node_name(node_name) {
            return Ok("STRING Anaerococcus senegalensis");
        }

        if is_valid_string_bacillus_timonensis_node_name(node_name) {
            return Ok("STRING Bacillus timonensis");
        }

        if is_valid_string_brevibacterium_senegalense_node_name(node_name) {
            return Ok("STRING Brevibacterium senegalense");
        }

        if is_valid_string_clostridium_senegalense_node_name(node_name) {
            return Ok("STRING Clostridium senegalense");
        }

        if is_valid_string_kurthia_sp_dielmo_node_name(node_name) {
            return Ok("STRING Kurthia sp. Dielmo");
        }

        if is_valid_string_kurthia_massiliensis_node_name(node_name) {
            return Ok("STRING Kurthia massiliensis");
        }

        if is_valid_string_kurthia_sp_jc8e_node_name(node_name) {
            return Ok("STRING Kurthia sp. JC8E");
        }

        if is_valid_string_paenibacillus_senegalensis_node_name(node_name) {
            return Ok("STRING Paenibacillus senegalensis");
        }

        if is_valid_string_peptoniphilus_senegalensis_node_name(node_name) {
            return Ok("STRING Peptoniphilus senegalensis");
        }

        if is_valid_string_salinisphaera_shabanensis_node_name(node_name) {
            return Ok("STRING Salinisphaera shabanensis");
        }

        if is_valid_string_halorhabdus_tiamatea_node_name(node_name) {
            return Ok("STRING Halorhabdus tiamatea");
        }

        if is_valid_string_haloplasma_contractile_node_name(node_name) {
            return Ok("STRING Haloplasma contractile");
        }

        if is_valid_string_lactobacillus_kefiranofaciens_node_name(node_name) {
            return Ok("STRING Lactobacillus kefiranofaciens");
        }

        if is_valid_string_rhizobium_leguminosarum_cb782_node_name(node_name) {
            return Ok("STRING Rhizobium leguminosarum CB782");
        }

        if is_valid_string_senegalimassilia_anaerobia_node_name(node_name) {
            return Ok("STRING Senegalimassilia anaerobia");
        }

        if is_valid_string_bacillus_massiliosenegalensis_node_name(node_name) {
            return Ok("STRING Bacillus massiliosenegalensis");
        }

        if is_valid_string_paenibacillus_sp_hw567_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. HW567");
        }

        if is_valid_string_flavobacterium_branchiophilum_node_name(node_name) {
            return Ok("STRING Flavobacterium branchiophilum");
        }

        if is_valid_string_mycoplasma_anatis_node_name(node_name) {
            return Ok("STRING Mycoplasma anatis");
        }

        if is_valid_string_staphylococcus_lugdunensis_n920143_node_name(node_name) {
            return Ok("STRING Staphylococcus lugdunensis N920143");
        }

        if is_valid_string_legionella_massiliensis_node_name(node_name) {
            return Ok("STRING Legionella massiliensis");
        }

        if is_valid_string_streptococcus_constellatus_pharyngis_sk1060_node_name(node_name) {
            return Ok("STRING Streptococcus constellatus pharyngis SK1060");
        }

        if is_valid_string_streptococcus_mitis_sk569_node_name(node_name) {
            return Ok("STRING Streptococcus mitis SK569");
        }

        if is_valid_string_streptococcus_infantis_sk970_node_name(node_name) {
            return Ok("STRING Streptococcus infantis SK970");
        }

        if is_valid_string_brevundimonas_diminuta_4704_node_name(node_name) {
            return Ok("STRING Brevundimonas diminuta 4704");
        }

        if is_valid_string_capnocytophaga_sp_f0382_node_name(node_name) {
            return Ok("STRING Capnocytophaga sp. F0382");
        }

        if is_valid_string_corynebacterium_durum_node_name(node_name) {
            return Ok("STRING Corynebacterium durum");
        }

        if is_valid_string_peptostreptococcus_anaerobius_vpi4330_node_name(node_name) {
            return Ok("STRING Peptostreptococcus anaerobius VPI4330");
        }

        if is_valid_string_prevotella_sp_f0040_node_name(node_name) {
            return Ok("STRING Prevotella sp. F0040");
        }

        if is_valid_string_dehalobacter_sp_fth1_node_name(node_name) {
            return Ok("STRING Dehalobacter sp. FTH1");
        }

        if is_valid_string_haemophilus_sputorum_node_name(node_name) {
            return Ok("STRING Haemophilus sputorum");
        }

        if is_valid_string_idiomarina_sp_a28l_node_name(node_name) {
            return Ok("STRING Idiomarina sp. A28L");
        }

        if is_valid_string_bradyrhizobium_japonicum_usda6_node_name(node_name) {
            return Ok("STRING Bradyrhizobium japonicum USDA6");
        }

        if is_valid_string_mycoplasma_columbinum_node_name(node_name) {
            return Ok("STRING Mycoplasma columbinum");
        }

        if is_valid_string_azorhizobium_doebereinerae_node_name(node_name) {
            return Ok("STRING Azorhizobium doebereinerae");
        }

        if is_valid_string_bradyrhizobium_elkanii_wsm1741_node_name(node_name) {
            return Ok("STRING Bradyrhizobium elkanii WSM1741");
        }

        if is_valid_string_bradyrhizobium_elkanii_wsm2783_node_name(node_name) {
            return Ok("STRING Bradyrhizobium elkanii WSM2783");
        }

        if is_valid_string_bradyrhizobium_japonicum_usda124_node_name(node_name) {
            return Ok("STRING Bradyrhizobium japonicum USDA124");
        }

        if is_valid_string_bradyrhizobium_sp_wsm2793_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. WSM2793");
        }

        if is_valid_string_bradyrhizobium_sp_wsm3983_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. WSM3983");
        }

        if is_valid_string_burkholderia_mimosarum_node_name(node_name) {
            return Ok("STRING Burkholderia mimosarum");
        }

        if is_valid_string_pseudomonas_fluorescens_q287_node_name(node_name) {
            return Ok("STRING Pseudomonas fluorescens Q287");
        }

        if is_valid_string_mesorhizobium_loti_cj3sym_node_name(node_name) {
            return Ok("STRING Mesorhizobium loti CJ3sym");
        }

        if is_valid_string_mesorhizobium_loti_usda3471_node_name(node_name) {
            return Ok("STRING Mesorhizobium loti USDA3471");
        }

        if is_valid_string_mesorhizobium_sp_wsm3224_node_name(node_name) {
            return Ok("STRING Mesorhizobium sp. WSM3224");
        }

        if is_valid_string_mesorhizobium_sp_wsm3626_node_name(node_name) {
            return Ok("STRING Mesorhizobium sp. WSM3626");
        }

        if is_valid_string_bradyrhizobium_sp_arr65_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. ARR65");
        }

        if is_valid_string_rhizobium_gallicum_node_name(node_name) {
            return Ok("STRING Rhizobium gallicum");
        }

        if is_valid_string_rhizobium_giardinii_node_name(node_name) {
            return Ok("STRING Rhizobium giardinii");
        }

        if is_valid_string_rhizobium_leguminosarum_gb30_node_name(node_name) {
            return Ok("STRING Rhizobium leguminosarum GB30");
        }

        if is_valid_string_rhizobium_sullae_node_name(node_name) {
            return Ok("STRING Rhizobium sullae");
        }

        if is_valid_string_rhizobium_leucaenae_node_name(node_name) {
            return Ok("STRING Rhizobium leucaenae");
        }

        if is_valid_string_ensifer_sp_wsm1721_node_name(node_name) {
            return Ok("STRING Ensifer sp. WSM1721");
        }

        if is_valid_string_arthromitus_sp_sfbratyit_node_name(node_name) {
            return Ok("STRING Arthromitus sp. SFBratYit");
        }

        if is_valid_string_mycobacterium_colombiense_node_name(node_name) {
            return Ok("STRING Mycobacterium colombiense");
        }

        if is_valid_string_wickerhamomyces_ciferrii_node_name(node_name) {
            return Ok("STRING Wickerhamomyces ciferrii");
        }

        if is_valid_string_flavobacterium_columnare_node_name(node_name) {
            return Ok("STRING Flavobacterium columnare");
        }

        if is_valid_string_methanocella_conradii_node_name(node_name) {
            return Ok("STRING Methanocella conradii");
        }

        if is_valid_string_clostridium_sp_sy8519_node_name(node_name) {
            return Ok("STRING Clostridium sp. SY8519");
        }

        if is_valid_string_brevibacillus_laterosporus_lmg15441_node_name(node_name) {
            return Ok("STRING Brevibacillus laterosporus LMG15441");
        }

        if is_valid_string_pseudomonas_fluorescens_hk44_node_name(node_name) {
            return Ok("STRING Pseudomonas fluorescens HK44");
        }

        if is_valid_string_rhizobium_sp_ibun_node_name(node_name) {
            return Ok("STRING Rhizobium sp. IBUN");
        }

        if is_valid_string_marinobacterium_stanieri_node_name(node_name) {
            return Ok("STRING Marinobacterium stanieri");
        }

        if is_valid_string_flavobacteriaceae_bacterium_s85_node_name(node_name) {
            return Ok("STRING Flavobacteriaceae bacterium S85");
        }

        if is_valid_string_microbulbifer_agarilyticus_node_name(node_name) {
            return Ok("STRING Microbulbifer agarilyticus");
        }

        if is_valid_string_pseudomonas_putida_s16_node_name(node_name) {
            return Ok("STRING Pseudomonas putida S16");
        }

        if is_valid_string_thermococcus_sp_4557_node_name(node_name) {
            return Ok("STRING Thermococcus sp. 4557");
        }

        if is_valid_string_serinicoccus_profundi_node_name(node_name) {
            return Ok("STRING Serinicoccus profundi");
        }

        if is_valid_string_lysinimicrobium_mangrovi_node_name(node_name) {
            return Ok("STRING Lysinimicrobium mangrovi");
        }

        if is_valid_string_oenococcus_kitaharae_node_name(node_name) {
            return Ok("STRING Oenococcus kitaharae");
        }

        if is_valid_string_citricoccus_sp_ch26a_node_name(node_name) {
            return Ok("STRING Citricoccus sp. CH26A");
        }

        if is_valid_string_weissella_koreensis_node_name(node_name) {
            return Ok("STRING Weissella koreensis");
        }

        if is_valid_string_pseudoxanthomonas_spadix_node_name(node_name) {
            return Ok("STRING Pseudoxanthomonas spadix");
        }

        if is_valid_string_enterobacter_cloacae_ecwsu1_node_name(node_name) {
            return Ok("STRING Enterobacter cloacae EcWSU1");
        }

        if is_valid_string_brachyspira_intermedia_node_name(node_name) {
            return Ok("STRING Brachyspira intermedia");
        }

        if is_valid_string_acinetobacter_lwoffii_wj10621_node_name(node_name) {
            return Ok("STRING Acinetobacter lwoffii WJ10621");
        }

        if is_valid_string_bizionia_argentinensis_node_name(node_name) {
            return Ok("STRING Bizionia argentinensis");
        }

        if is_valid_string_streptococcus_salivarius_57i_node_name(node_name) {
            return Ok("STRING Streptococcus salivarius 57I");
        }

        if is_valid_string_gayadomonas_joobiniege_node_name(node_name) {
            return Ok("STRING Gayadomonas joobiniege");
        }

        if is_valid_string_marinobacter_lipolyticus_bf04cf4_node_name(node_name) {
            return Ok("STRING Marinobacter lipolyticus BF04CF4");
        }

        if is_valid_string_aminicenantes_bacterium_aaa252a02_node_name(node_name) {
            return Ok("STRING Aminicenantes bacterium AAA252A02");
        }

        if is_valid_string_zymoseptoria_tritici_node_name(node_name) {
            return Ok("STRING Zymoseptoria tritici");
        }

        if is_valid_string_sporichthya_polymorpha_node_name(node_name) {
            return Ok("STRING Sporichthya polymorpha");
        }

        if is_valid_string_paracoccidioides_sp_lutzii_node_name(node_name) {
            return Ok("STRING Paracoccidioides sp. lutzii");
        }

        if is_valid_string_mycoplasma_iowae_node_name(node_name) {
            return Ok("STRING Mycoplasma iowae");
        }

        if is_valid_string_alicyclobacillus_acidocaldarius_tc41_node_name(node_name) {
            return Ok("STRING Alicyclobacillus acidocaldarius Tc41");
        }

        if is_valid_string_anditalea_andensis_node_name(node_name) {
            return Ok("STRING Anditalea andensis");
        }

        if is_valid_string_endosymbiont_of_tevnia_node_name(node_name) {
            return Ok("STRING endosymbiont of Tevnia");
        }

        if is_valid_string_allobaculum_stercoricanis_node_name(node_name) {
            return Ok("STRING Allobaculum stercoricanis");
        }

        if is_valid_string_actinopolyspora_mortivallis_node_name(node_name) {
            return Ok("STRING Actinopolyspora mortivallis");
        }

        if is_valid_string_propionibacterium_acnes_sk182bjcvi_node_name(node_name) {
            return Ok("STRING Propionibacterium acnes SK182BJCVI");
        }

        if is_valid_string_bacillus_mojavensis_node_name(node_name) {
            return Ok("STRING Bacillus mojavensis");
        }

        if is_valid_string_verticillium_alfalfae_node_name(node_name) {
            return Ok("STRING Verticillium alfalfae");
        }

        if is_valid_string_sulfobacillus_acidophilus_node_name(node_name) {
            return Ok("STRING Sulfobacillus acidophilus");
        }

        if is_valid_string_vibrio_tubiashii_node_name(node_name) {
            return Ok("STRING Vibrio tubiashii");
        }

        if is_valid_string_neisseria_weaveri_node_name(node_name) {
            return Ok("STRING Neisseria weaveri");
        }

        if is_valid_string_paenibacillus_polymyxa_m1_node_name(node_name) {
            return Ok("STRING Paenibacillus polymyxa M1");
        }

        if is_valid_string_acetobacteraceae_bacterium_at5844_node_name(node_name) {
            return Ok("STRING Acetobacteraceae bacterium AT5844");
        }

        if is_valid_string_thermoplasmatales_archaeon_node_name(node_name) {
            return Ok("STRING Thermoplasmatales archaeon");
        }

        if is_valid_string_streptococcus_pseudopneumoniae_node_name(node_name) {
            return Ok("STRING Streptococcus pseudopneumoniae");
        }

        if is_valid_string_streptomyces_purpureus_node_name(node_name) {
            return Ok("STRING Streptomyces purpureus");
        }

        if is_valid_string_psychrobacter_sp_tb67_node_name(node_name) {
            return Ok("STRING Psychrobacter sp. TB67");
        }

        if is_valid_string_caldisphaera_lagunensis_node_name(node_name) {
            return Ok("STRING Caldisphaera lagunensis");
        }

        if is_valid_string_grimontia_indica_node_name(node_name) {
            return Ok("STRING Grimontia indica");
        }

        if is_valid_string_nocardia_sp_bmg51109_node_name(node_name) {
            return Ok("STRING Nocardia sp. BMG51109");
        }

        if is_valid_string_teredinibacter_turnerae_t7902_node_name(node_name) {
            return Ok("STRING Teredinibacter turnerae T7902");
        }

        if is_valid_string_ensifer_sp_br816_node_name(node_name) {
            return Ok("STRING Ensifer sp. BR816");
        }

        if is_valid_string_megasphaera_elsdenii_node_name(node_name) {
            return Ok("STRING Megasphaera elsdenii");
        }

        if is_valid_string_brachybacterium_paraconglomeratum_node_name(node_name) {
            return Ok("STRING Brachybacterium paraconglomeratum");
        }

        if is_valid_string_amycolatopsis_methanolica_node_name(node_name) {
            return Ok("STRING Amycolatopsis methanolica");
        }

        if is_valid_string_amycolatopsis_nigrescens_node_name(node_name) {
            return Ok("STRING Amycolatopsis nigrescens");
        }

        if is_valid_string_succinispira_mobilis_node_name(node_name) {
            return Ok("STRING Succinispira mobilis");
        }

        if is_valid_string_streptococcus_infantarius_node_name(node_name) {
            return Ok("STRING Streptococcus infantarius");
        }

        if is_valid_string_lactobacillus_ruminis_atcc27782_node_name(node_name) {
            return Ok("STRING Lactobacillus ruminis ATCC27782");
        }

        if is_valid_string_glomeribacter_gigasporarum_node_name(node_name) {
            return Ok("STRING Glomeribacter gigasporarum");
        }

        if is_valid_string_halonotius_sp_j07hn4_node_name(node_name) {
            return Ok("STRING Halonotius sp. J07HN4");
        }

        if is_valid_string_bacillus_sp_nsp222_node_name(node_name) {
            return Ok("STRING Bacillus sp. NSP222");
        }

        if is_valid_string_haloarchaeon_3a1dgr_node_name(node_name) {
            return Ok("STRING haloarchaeon 3A1DGR");
        }

        if is_valid_string_tetrapisispora_blattae_node_name(node_name) {
            return Ok("STRING Tetrapisispora blattae");
        }

        if is_valid_string_lactobacillus_buchneri_cd034_node_name(node_name) {
            return Ok("STRING Lactobacillus buchneri CD034");
        }

        if is_valid_string_burkholderia_grimmiae_node_name(node_name) {
            return Ok("STRING Burkholderia grimmiae");
        }

        if is_valid_string_basilea_psittacipulmonis_node_name(node_name) {
            return Ok("STRING Basilea psittacipulmonis");
        }

        if is_valid_string_cronobacter_condimenti_node_name(node_name) {
            return Ok("STRING Cronobacter condimenti");
        }

        if is_valid_string_lactobacillus_curvatus_node_name(node_name) {
            return Ok("STRING Lactobacillus curvatus");
        }

        if is_valid_string_brachybacterium_squillarum_node_name(node_name) {
            return Ok("STRING Brachybacterium squillarum");
        }

        if is_valid_string_blattabacterium_sp_madar_node_name(node_name) {
            return Ok("STRING Blattabacterium sp. MADAR");
        }

        if is_valid_string_gordonia_amarae_node_name(node_name) {
            return Ok("STRING Gordonia amarae");
        }

        if is_valid_string_blattabacterium_sp_cpu_node_name(node_name) {
            return Ok("STRING Blattabacterium sp. Cpu");
        }

        if is_valid_string_pantoea_rwandensis_node_name(node_name) {
            return Ok("STRING Pantoea rwandensis");
        }

        if is_valid_string_dietzia_alimentaria_node_name(node_name) {
            return Ok("STRING Dietzia alimentaria");
        }

        if is_valid_string_bacteroides_faecis_node_name(node_name) {
            return Ok("STRING Bacteroides faecis");
        }

        if is_valid_string_arthrobacter_globiformis_node_name(node_name) {
            return Ok("STRING Arthrobacter globiformis");
        }

        if is_valid_string_gordonia_effusa_node_name(node_name) {
            return Ok("STRING Gordonia effusa");
        }

        if is_valid_string_mycobacterium_thermoresistibile_node_name(node_name) {
            return Ok("STRING Mycobacterium thermoresistibile");
        }

        if is_valid_string_staphylococcus_sp_hgb0015_node_name(node_name) {
            return Ok("STRING Staphylococcus sp. HGB0015");
        }

        if is_valid_string_paenisporosarcina_sp_hgh0030_node_name(node_name) {
            return Ok("STRING Paenisporosarcina sp. HGH0030");
        }

        if is_valid_string_rhizobium_mongolense_node_name(node_name) {
            return Ok("STRING Rhizobium mongolense");
        }

        if is_valid_string_streptomyces_chartreusis_node_name(node_name) {
            return Ok("STRING Streptomyces chartreusis");
        }

        if is_valid_string_citrobacter_sp_s77_node_name(node_name) {
            return Ok("STRING Citrobacter sp. S77");
        }

        if is_valid_string_sphingomonas_elodea_node_name(node_name) {
            return Ok("STRING Sphingomonas elodea");
        }

        if is_valid_string_aquiluna_sp_imcc13023_node_name(node_name) {
            return Ok("STRING Aquiluna sp. IMCC13023");
        }

        if is_valid_string_lonsdalea_quercina_node_name(node_name) {
            return Ok("STRING Lonsdalea quercina");
        }

        if is_valid_string_pelagibacterium_halotolerans_node_name(node_name) {
            return Ok("STRING Pelagibacterium halotolerans");
        }

        if is_valid_string_agrobacterium_tumefaciens_ccnwgs0286_node_name(node_name) {
            return Ok("STRING Agrobacterium tumefaciens CCNWGS0286");
        }

        if is_valid_string_mesorhizobium_amorphae_node_name(node_name) {
            return Ok("STRING Mesorhizobium amorphae");
        }

        if is_valid_string_glaciecola_nitratireducens_node_name(node_name) {
            return Ok("STRING Glaciecola nitratireducens");
        }

        if is_valid_string_flavobacterium_frigoris_node_name(node_name) {
            return Ok("STRING Flavobacterium frigoris");
        }

        if is_valid_string_exiguobacterium_antarcticum_node_name(node_name) {
            return Ok("STRING Exiguobacterium antarcticum");
        }

        if is_valid_string_paenibacillus_peoriae_node_name(node_name) {
            return Ok("STRING Paenibacillus peoriae");
        }

        if is_valid_string_novosphingobium_pentaromativorans_node_name(node_name) {
            return Ok("STRING Novosphingobium pentaromativorans");
        }

        if is_valid_string_commensalibacter_intestini_node_name(node_name) {
            return Ok("STRING Commensalibacter intestini");
        }

        if is_valid_string_gluconobacter_morbifer_node_name(node_name) {
            return Ok("STRING Gluconobacter morbifer");
        }

        if is_valid_string_fangia_hongkongensis_node_name(node_name) {
            return Ok("STRING Fangia hongkongensis");
        }

        if is_valid_string_aggregatibacter_actinomycetemcomitans_rhaa1_node_name(node_name) {
            return Ok("STRING Aggregatibacter actinomycetemcomitans RhAA1");
        }

        if is_valid_string_gordonia_sputi_node_name(node_name) {
            return Ok("STRING Gordonia sputi");
        }

        if is_valid_string_mobilicoccus_pelagius_node_name(node_name) {
            return Ok("STRING Mobilicoccus pelagius");
        }

        if is_valid_string_amycolatopsis_benzoatilytica_node_name(node_name) {
            return Ok("STRING Amycolatopsis benzoatilytica");
        }

        if is_valid_string_amycolatopsis_balhimycina_node_name(node_name) {
            return Ok("STRING Amycolatopsis balhimycina");
        }

        if is_valid_string_actinopolyspora_halophila_node_name(node_name) {
            return Ok("STRING Actinopolyspora halophila");
        }

        if is_valid_string_rudanella_lutea_node_name(node_name) {
            return Ok("STRING Rudanella lutea");
        }

        if is_valid_string_thermicanus_aegyptius_node_name(node_name) {
            return Ok("STRING Thermicanus aegyptius");
        }

        if is_valid_string_haloglycomyces_albus_node_name(node_name) {
            return Ok("STRING Haloglycomyces albus");
        }

        if is_valid_string_salisaeta_longa_node_name(node_name) {
            return Ok("STRING Salisaeta longa");
        }

        if is_valid_string_geminicoccus_roseus_node_name(node_name) {
            return Ok("STRING Geminicoccus roseus");
        }

        if is_valid_string_rhodovibrio_salinarum_node_name(node_name) {
            return Ok("STRING Rhodovibrio salinarum");
        }

        if is_valid_string_thermacetogenium_phaeum_node_name(node_name) {
            return Ok("STRING Thermacetogenium phaeum");
        }

        if is_valid_string_sphingomonas_phyllosphaerae_52_node_name(node_name) {
            return Ok("STRING Sphingomonas phyllosphaerae 52");
        }

        if is_valid_string_sphingomonas_phyllosphaerae_fa2_node_name(node_name) {
            return Ok("STRING Sphingomonas phyllosphaerae FA2");
        }

        if is_valid_string_sphingomonas_melonis_node_name(node_name) {
            return Ok("STRING Sphingomonas melonis");
        }

        if is_valid_string_xanthomonas_campestris_musacearum_node_name(node_name) {
            return Ok("STRING Xanthomonas campestris musacearum");
        }

        if is_valid_string_flavobacterium_indicum_node_name(node_name) {
            return Ok("STRING Flavobacterium indicum");
        }

        if is_valid_string_bartonella_australis_node_name(node_name) {
            return Ok("STRING Bartonella australis");
        }

        if is_valid_string_bartonella_bovis_node_name(node_name) {
            return Ok("STRING Bartonella bovis");
        }

        if is_valid_string_bartonella_schoenbuchensis_node_name(node_name) {
            return Ok("STRING Bartonella schoenbuchensis");
        }

        if is_valid_string_bartonella_vinsonii_berkhoffii_node_name(node_name) {
            return Ok("STRING Bartonella vinsonii berkhoffii");
        }

        if is_valid_string_thermoanaerobacterium_saccharolyticum_node_name(node_name) {
            return Ok("STRING Thermoanaerobacterium saccharolyticum");
        }

        if is_valid_string_bartonella_alsatica_node_name(node_name) {
            return Ok("STRING Bartonella alsatica");
        }

        if is_valid_string_bartonella_doshiae_node_name(node_name) {
            return Ok("STRING Bartonella doshiae");
        }

        if is_valid_string_bartonella_rattimassiliensis_node_name(node_name) {
            return Ok("STRING Bartonella rattimassiliensis");
        }

        if is_valid_string_bartonella_melophagi_node_name(node_name) {
            return Ok("STRING Bartonella melophagi");
        }

        if is_valid_string_bartonella_tamiae_node_name(node_name) {
            return Ok("STRING Bartonella tamiae");
        }

        if is_valid_string_bartonella_taylorii_node_name(node_name) {
            return Ok("STRING Bartonella taylorii");
        }

        if is_valid_string_bartonella_vinsonii_arupensis_node_name(node_name) {
            return Ok("STRING Bartonella vinsonii arupensis");
        }

        if is_valid_string_bartonella_washoensis_node_name(node_name) {
            return Ok("STRING Bartonella washoensis");
        }

        if is_valid_string_fluoribacter_dumoffii_node_name(node_name) {
            return Ok("STRING Fluoribacter dumoffii");
        }

        if is_valid_string_bartonella_sp_db56_node_name(node_name) {
            return Ok("STRING Bartonella sp. DB56");
        }

        if is_valid_string_methanolobus_psychrophilus_node_name(node_name) {
            return Ok("STRING Methanolobus psychrophilus");
        }

        if is_valid_string_streptococcus_sp_sk140_node_name(node_name) {
            return Ok("STRING Streptococcus sp. SK140");
        }

        if is_valid_string_streptococcus_sp_sk643_node_name(node_name) {
            return Ok("STRING Streptococcus sp. SK643");
        }

        if is_valid_string_streptococcus_mitis_sk579_node_name(node_name) {
            return Ok("STRING Streptococcus mitis SK579");
        }

        if is_valid_string_streptococcus_oralis_sk1074_node_name(node_name) {
            return Ok("STRING Streptococcus oralis SK1074");
        }

        if is_valid_string_haemophilus_paraphrohaemolyticus_node_name(node_name) {
            return Ok("STRING Haemophilus paraphrohaemolyticus");
        }

        if is_valid_string_fusobacterium_necrophorum_funduliforme_node_name(node_name) {
            return Ok("STRING Fusobacterium necrophorum funduliforme");
        }

        if is_valid_string_pasteurella_bettyae_node_name(node_name) {
            return Ok("STRING Pasteurella bettyae");
        }

        if is_valid_string_lachnoanaerobaculum_saburreum_f0468_node_name(node_name) {
            return Ok("STRING Lachnoanaerobaculum saburreum F0468");
        }

        if is_valid_string_prevotella_sp_f0472_node_name(node_name) {
            return Ok("STRING Prevotella sp. F0472");
        }

        if is_valid_string_cellulomonas_massiliensis_node_name(node_name) {
            return Ok("STRING Cellulomonas massiliensis");
        }

        if is_valid_string_herbaspirillum_massiliense_node_name(node_name) {
            return Ok("STRING Herbaspirillum massiliense");
        }

        if is_valid_string_peptoniphilus_timonensis_node_name(node_name) {
            return Ok("STRING Peptoniphilus timonensis");
        }

        if is_valid_string_timonella_senegalensis_node_name(node_name) {
            return Ok("STRING Timonella senegalensis");
        }

        if is_valid_string_methylobacterium_sp_gxf4_node_name(node_name) {
            return Ok("STRING Methylobacterium sp. GXF4");
        }

        if is_valid_string_arthrobacter_sp_pao19_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. PAO19");
        }

        if is_valid_string_pelagibacter_ubique_himb083_node_name(node_name) {
            return Ok("STRING Pelagibacter ubique HIMB083");
        }

        if is_valid_string_novosphingobium_lindaniclasticum_node_name(node_name) {
            return Ok("STRING Novosphingobium lindaniclasticum");
        }

        if is_valid_string_burkholderia_sp_yi23_node_name(node_name) {
            return Ok("STRING Burkholderia sp. YI23");
        }

        if is_valid_string_limnohabitans_sp_rim28_node_name(node_name) {
            return Ok("STRING Limnohabitans sp. Rim28");
        }

        if is_valid_string_limnohabitans_sp_rim47_node_name(node_name) {
            return Ok("STRING Limnohabitans sp. Rim47");
        }

        if is_valid_string_arthrobacter_sp_man2_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. MAN2");
        }

        if is_valid_string_paracoccus_sp_n5_node_name(node_name) {
            return Ok("STRING Paracoccus sp. N5");
        }

        if is_valid_string_methylopila_sp_m107_node_name(node_name) {
            return Ok("STRING Methylopila sp. M107");
        }

        if is_valid_string_methylobacterium_sp_10_node_name(node_name) {
            return Ok("STRING Methylobacterium sp. 10");
        }

        if is_valid_string_methylobacterium_sp_77_node_name(node_name) {
            return Ok("STRING Methylobacterium sp. 77");
        }

        if is_valid_string_methylophilaceae_bacterium_11_node_name(node_name) {
            return Ok("STRING Methylophilaceae bacterium 11");
        }

        if is_valid_string_pyrobaculum_sp_1860_node_name(node_name) {
            return Ok("STRING Pyrobaculum sp. 1860");
        }

        if is_valid_string_enterococcus_faecium_nrrlb2354_node_name(node_name) {
            return Ok("STRING Enterococcus faecium NRRLB2354");
        }

        if is_valid_string_actinomyces_sp_icm39_node_name(node_name) {
            return Ok("STRING Actinomyces sp. ICM39");
        }

        if is_valid_string_clostridium_sp_mste9_node_name(node_name) {
            return Ok("STRING Clostridium sp. MSTE9");
        }

        if is_valid_string_rickettsia_australis_node_name(node_name) {
            return Ok("STRING Rickettsia australis");
        }

        if is_valid_string_paenirhodobacter_enshiensis_node_name(node_name) {
            return Ok("STRING Paenirhodobacter enshiensis");
        }

        if is_valid_string_flavobacterium_enshiense_node_name(node_name) {
            return Ok("STRING Flavobacterium enshiense");
        }

        if is_valid_string_gordonia_rhizosphera_node_name(node_name) {
            return Ok("STRING Gordonia rhizosphera");
        }

        if is_valid_string_penicillium_rubens_node_name(node_name) {
            return Ok("STRING Penicillium rubens");
        }

        if is_valid_string_pseudomonas_stutzeri_sdmlac_node_name(node_name) {
            return Ok("STRING Pseudomonas stutzeri SDMLAC");
        }

        if is_valid_string_tistrella_mobilis_node_name(node_name) {
            return Ok("STRING Tistrella mobilis");
        }

        if is_valid_string_nocardia_asteroides_node_name(node_name) {
            return Ok("STRING Nocardia asteroides");
        }

        if is_valid_string_thermus_sp_ccbus3uf1_node_name(node_name) {
            return Ok("STRING Thermus sp. CCBUS3UF1");
        }

        if is_valid_string_atopobium_sp_bv3ac4_node_name(node_name) {
            return Ok("STRING Atopobium sp. BV3Ac4");
        }

        if is_valid_string_propionimicrobium_sp_bv2f7_node_name(node_name) {
            return Ok("STRING Propionimicrobium sp. BV2F7");
        }

        if is_valid_string_peptoniphilus_sp_bv3c26_node_name(node_name) {
            return Ok("STRING Peptoniphilus sp. BV3C26");
        }

        if is_valid_string_coriobacteriaceae_bacterium_bv3ac1_node_name(node_name) {
            return Ok("STRING Coriobacteriaceae bacterium BV3Ac1");
        }

        if is_valid_string_megasphaera_sp_bv3c161_node_name(node_name) {
            return Ok("STRING Megasphaera sp. BV3C161");
        }

        if is_valid_string_alicyclobacillus_pomorum_node_name(node_name) {
            return Ok("STRING Alicyclobacillus pomorum");
        }

        if is_valid_string_budvicia_aquatica_node_name(node_name) {
            return Ok("STRING Budvicia aquatica");
        }

        if is_valid_string_corynebacterium_sputi_node_name(node_name) {
            return Ok("STRING Corynebacterium sputi");
        }

        if is_valid_string_flavobacterium_antarcticum_node_name(node_name) {
            return Ok("STRING Flavobacterium antarcticum");
        }

        if is_valid_string_ignatzschineria_larvae_node_name(node_name) {
            return Ok("STRING Ignatzschineria larvae");
        }

        if is_valid_string_gordonia_polyisoprenivorans_node_name(node_name) {
            return Ok("STRING Gordonia polyisoprenivorans");
        }

        if is_valid_string_psychrobacter_sp_pamc21119_node_name(node_name) {
            return Ok("STRING Psychrobacter sp. PAMC21119");
        }

        if is_valid_string_sphingomonas_echinoides_node_name(node_name) {
            return Ok("STRING Sphingomonas echinoides");
        }

        if is_valid_string_sphingomonas_sp_pamc26605_node_name(node_name) {
            return Ok("STRING Sphingomonas sp. PAMC26605");
        }

        if is_valid_string_sphingomonas_sp_pamc26617_node_name(node_name) {
            return Ok("STRING Sphingomonas sp. PAMC26617");
        }

        if is_valid_string_pseudomonas_psychrotolerans_node_name(node_name) {
            return Ok("STRING Pseudomonas psychrotolerans");
        }

        if is_valid_string_methylophilus_sp_5_node_name(node_name) {
            return Ok("STRING Methylophilus sp. 5");
        }

        if is_valid_string_natronorubrum_tibetense_node_name(node_name) {
            return Ok("STRING Natronorubrum tibetense");
        }

        if is_valid_string_citrobacter_farmeri_node_name(node_name) {
            return Ok("STRING Citrobacter farmeri");
        }

        if is_valid_string_saccharomonospora_azurea_node_name(node_name) {
            return Ok("STRING Saccharomonospora azurea");
        }

        if is_valid_string_sphingobium_baderi_node_name(node_name) {
            return Ok("STRING Sphingobium baderi");
        }

        if is_valid_string_streptococcus_parasanguinis_fw213_node_name(node_name) {
            return Ok("STRING Streptococcus parasanguinis FW213");
        }

        if is_valid_string_pseudomonas_fluorescens_f113_node_name(node_name) {
            return Ok("STRING Pseudomonas fluorescens F113");
        }

        if is_valid_string_lactobacillus_rossiae_node_name(node_name) {
            return Ok("STRING Lactobacillus rossiae");
        }

        if is_valid_string_escherichia_hermannii_node_name(node_name) {
            return Ok("STRING Escherichia hermannii");
        }

        if is_valid_string_escherichia_vulneris_node_name(node_name) {
            return Ok("STRING Escherichia vulneris");
        }

        if is_valid_string_arthrobacter_sp_31y_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. 31Y");
        }

        if is_valid_string_actinomyces_naeslundii_node_name(node_name) {
            return Ok("STRING Actinomyces naeslundii");
        }

        if is_valid_string_streptococcus_macedonicus_node_name(node_name) {
            return Ok("STRING Streptococcus macedonicus");
        }

        if is_valid_string_streptomyces_acidiscabies_node_name(node_name) {
            return Ok("STRING Streptomyces acidiscabies");
        }

        if is_valid_string_hoeflea_sp_108_node_name(node_name) {
            return Ok("STRING Hoeflea sp. 108");
        }

        if is_valid_string_vibrio_sp_ejy3_node_name(node_name) {
            return Ok("STRING Vibrio sp. EJY3");
        }

        if is_valid_string_methyloglobulus_morosus_node_name(node_name) {
            return Ok("STRING Methyloglobulus morosus");
        }

        if is_valid_string_paenibacillus_alvei_node_name(node_name) {
            return Ok("STRING Paenibacillus alvei");
        }

        if is_valid_string_pseudoalteromonas_citrea_node_name(node_name) {
            return Ok("STRING Pseudoalteromonas citrea");
        }

        if is_valid_string_pseudoalteromonas_haloplanktis_atcc14393_node_name(node_name) {
            return Ok("STRING Pseudoalteromonas haloplanktis ATCC14393");
        }

        if is_valid_string_pseudoalteromonas_rubra_node_name(node_name) {
            return Ok("STRING Pseudoalteromonas rubra");
        }

        if is_valid_string_pseudoalteromonas_spongiae_node_name(node_name) {
            return Ok("STRING Pseudoalteromonas spongiae");
        }

        if is_valid_string_bacillus_bataviensis_node_name(node_name) {
            return Ok("STRING Bacillus bataviensis");
        }

        if is_valid_string_mycoplasma_canis_node_name(node_name) {
            return Ok("STRING Mycoplasma canis");
        }

        if is_valid_string_simiduia_agarivorans_node_name(node_name) {
            return Ok("STRING Simiduia agarivorans");
        }

        if is_valid_string_sinorhizobium_fredii_hh103_node_name(node_name) {
            return Ok("STRING Sinorhizobium fredii HH103");
        }

        if is_valid_string_pseudomonas_extremaustralis_node_name(node_name) {
            return Ok("STRING Pseudomonas extremaustralis");
        }

        if is_valid_string_brevibacillus_massiliensis_node_name(node_name) {
            return Ok("STRING Brevibacillus massiliensis");
        }

        if is_valid_string_anaerococcus_sp_ph9_node_name(node_name) {
            return Ok("STRING Anaerococcus sp. PH9");
        }

        if is_valid_string_peptoniphilus_grossensis_node_name(node_name) {
            return Ok("STRING Peptoniphilus grossensis");
        }

        if is_valid_string_actinomyces_sp_ph3_node_name(node_name) {
            return Ok("STRING Actinomyces sp. ph3");
        }

        if is_valid_string_kallipyga_massiliensis_node_name(node_name) {
            return Ok("STRING Kallipyga massiliensis");
        }

        if is_valid_string_enorma_massiliensis_node_name(node_name) {
            return Ok("STRING Enorma massiliensis");
        }

        if is_valid_string_halomonas_sp_gfaj1_node_name(node_name) {
            return Ok("STRING Halomonas sp. GFAJ1");
        }

        if is_valid_string_stenotrophomonas_maltophilia_pml168_node_name(node_name) {
            return Ok("STRING Stenotrophomonas maltophilia PML168");
        }

        if is_valid_string_mycoplasma_hyorhinis_sk76_node_name(node_name) {
            return Ok("STRING Mycoplasma hyorhinis SK76");
        }

        if is_valid_string_sphingopyxis_sp_lc363_node_name(node_name) {
            return Ok("STRING Sphingopyxis sp. LC363");
        }

        if is_valid_string_bacterium_sp_ms4_node_name(node_name) {
            return Ok("STRING Bacterium sp. MS4");
        }

        if is_valid_string_methylopila_sp_73b_node_name(node_name) {
            return Ok("STRING Methylopila sp. 73B");
        }

        if is_valid_string_mycobacterium_sp_141_node_name(node_name) {
            return Ok("STRING Mycobacterium sp. 141");
        }

        if is_valid_string_acaricomes_phytoseiuli_node_name(node_name) {
            return Ok("STRING Acaricomes phytoseiuli");
        }

        if is_valid_string_acetobacter_nitrogenifigens_node_name(node_name) {
            return Ok("STRING Acetobacter nitrogenifigens");
        }

        if is_valid_string_acinetobacter_bouvetii_node_name(node_name) {
            return Ok("STRING Acinetobacter bouvetii");
        }

        if is_valid_string_actinobacillus_capsulatus_node_name(node_name) {
            return Ok("STRING Actinobacillus capsulatus");
        }

        if is_valid_string_actinobaculum_urinale_node_name(node_name) {
            return Ok("STRING Actinobaculum urinale");
        }

        if is_valid_string_actinokineospora_enzanensis_node_name(node_name) {
            return Ok("STRING Actinokineospora enzanensis");
        }

        if is_valid_string_actinomadura_atramentaria_node_name(node_name) {
            return Ok("STRING Actinomadura atramentaria");
        }

        if is_valid_string_actinomyces_dentalis_node_name(node_name) {
            return Ok("STRING Actinomyces dentalis");
        }

        if is_valid_string_actinomyces_georgiae_node_name(node_name) {
            return Ok("STRING Actinomyces georgiae");
        }

        if is_valid_string_actinomyces_israelii_node_name(node_name) {
            return Ok("STRING Actinomyces israelii");
        }

        if is_valid_string_actinomyces_neuii_dsm8576_node_name(node_name) {
            return Ok("STRING Actinomyces neuii DSM8576");
        }

        if is_valid_string_actinomyces_suimastitidis_node_name(node_name) {
            return Ok("STRING Actinomyces suimastitidis");
        }

        if is_valid_string_actinomyces_vaccimaxillae_node_name(node_name) {
            return Ok("STRING Actinomyces vaccimaxillae");
        }

        if is_valid_string_actinomycetospora_chiangmaiensis_node_name(node_name) {
            return Ok("STRING Actinomycetospora chiangmaiensis");
        }

        if is_valid_string_actinoplanes_globisporus_node_name(node_name) {
            return Ok("STRING Actinoplanes globisporus");
        }

        if is_valid_string_actinopolymorpha_alba_node_name(node_name) {
            return Ok("STRING Actinopolymorpha alba");
        }

        if is_valid_string_aequorivita_capsosiphonis_node_name(node_name) {
            return Ok("STRING Aequorivita capsosiphonis");
        }

        if is_valid_string_aestuariibacter_salexigens_node_name(node_name) {
            return Ok("STRING Aestuariibacter salexigens");
        }

        if is_valid_string_aestuariimicrobium_kwangyangense_node_name(node_name) {
            return Ok("STRING Aestuariimicrobium kwangyangense");
        }

        if is_valid_string_afifella_pfennigii_node_name(node_name) {
            return Ok("STRING Afifella pfennigii");
        }

        if is_valid_string_agrococcus_lahaulensis_node_name(node_name) {
            return Ok("STRING Agrococcus lahaulensis");
        }

        if is_valid_string_agromyces_italicus_node_name(node_name) {
            return Ok("STRING Agromyces italicus");
        }

        if is_valid_string_agromyces_subbeticus_node_name(node_name) {
            return Ok("STRING Agromyces subbeticus");
        }

        if is_valid_string_algicola_sagamiensis_node_name(node_name) {
            return Ok("STRING Algicola sagamiensis");
        }

        if is_valid_string_algoriphagus_mannitolivorans_node_name(node_name) {
            return Ok("STRING Algoriphagus mannitolivorans");
        }

        if is_valid_string_algoriphagus_marincola_dsm16067_node_name(node_name) {
            return Ok("STRING Algoriphagus marincola DSM16067");
        }

        if is_valid_string_algoriphagus_vanfongensis_node_name(node_name) {
            return Ok("STRING Algoriphagus vanfongensis");
        }

        if is_valid_string_aliagarivorans_taiwanensis_node_name(node_name) {
            return Ok("STRING Aliagarivorans taiwanensis");
        }

        if is_valid_string_alicyclobacillus_contaminans_node_name(node_name) {
            return Ok("STRING Alicyclobacillus contaminans");
        }

        if is_valid_string_alicyclobacillus_herbarius_node_name(node_name) {
            return Ok("STRING Alicyclobacillus herbarius");
        }

        if is_valid_string_alicyclobacillus_pohliae_node_name(node_name) {
            return Ok("STRING Alicyclobacillus pohliae");
        }

        if is_valid_string_alkanindiges_illinoisensis_node_name(node_name) {
            return Ok("STRING Alkanindiges illinoisensis");
        }

        if is_valid_string_allofustis_seminis_node_name(node_name) {
            return Ok("STRING Allofustis seminis");
        }

        if is_valid_string_alloscardovia_omnicolens_node_name(node_name) {
            return Ok("STRING Alloscardovia omnicolens");
        }

        if is_valid_string_alysiella_crassa_node_name(node_name) {
            return Ok("STRING Alysiella crassa");
        }

        if is_valid_string_amorphus_coralli_node_name(node_name) {
            return Ok("STRING Amorphus coralli");
        }

        if is_valid_string_anaeroarcus_burkinensis_node_name(node_name) {
            return Ok("STRING Anaeroarcus burkinensis");
        }

        if is_valid_string_anaerobiospirillum_succiniciproducens_node_name(node_name) {
            return Ok("STRING Anaerobiospirillum succiniciproducens");
        }

        if is_valid_string_anaerovorax_odorimutans_node_name(node_name) {
            return Ok("STRING Anaerovorax odorimutans");
        }

        if is_valid_string_andreprevotia_chitinilytica_node_name(node_name) {
            return Ok("STRING Andreprevotia chitinilytica");
        }

        if is_valid_string_aquaspirillum_serpens_node_name(node_name) {
            return Ok("STRING Aquaspirillum serpens");
        }

        if is_valid_string_aquimarina_muelleri_node_name(node_name) {
            return Ok("STRING Aquimarina muelleri");
        }

        if is_valid_string_arenibacter_certesii_node_name(node_name) {
            return Ok("STRING Arenibacter certesii");
        }

        if is_valid_string_arenibacter_latericius_node_name(node_name) {
            return Ok("STRING Arenibacter latericius");
        }

        if is_valid_string_arenimonas_composti_node_name(node_name) {
            return Ok("STRING Arenimonas composti");
        }

        if is_valid_string_arenimonas_oryziterrae_node_name(node_name) {
            return Ok("STRING Arenimonas oryziterrae");
        }

        if is_valid_string_arsenicicoccus_bolidensis_node_name(node_name) {
            return Ok("STRING Arsenicicoccus bolidensis");
        }

        if is_valid_string_arthrobacter_castelli_node_name(node_name) {
            return Ok("STRING Arthrobacter castelli");
        }

        if is_valid_string_arthrobacter_sanguinis_node_name(node_name) {
            return Ok("STRING Arthrobacter sanguinis");
        }

        if is_valid_string_asticcacaulis_benevestitus_node_name(node_name) {
            return Ok("STRING Asticcacaulis benevestitus");
        }

        if is_valid_string_atopococcus_tabaci_node_name(node_name) {
            return Ok("STRING Atopococcus tabaci");
        }

        if is_valid_string_aureimonas_ureilytica_node_name(node_name) {
            return Ok("STRING Aureimonas ureilytica");
        }

        if is_valid_string_azospirillum_halopraeferens_node_name(node_name) {
            return Ok("STRING Azospirillum halopraeferens");
        }

        if is_valid_string_azovibrio_restrictus_node_name(node_name) {
            return Ok("STRING Azovibrio restrictus");
        }

        if is_valid_string_bacillus_aidingensis_node_name(node_name) {
            return Ok("STRING Bacillus aidingensis");
        }

        if is_valid_string_bacillus_chagannorensis_node_name(node_name) {
            return Ok("STRING Bacillus chagannorensis");
        }

        if is_valid_string_bacillus_fordii_node_name(node_name) {
            return Ok("STRING Bacillus fordii");
        }

        if is_valid_string_bacillus_gelatini_node_name(node_name) {
            return Ok("STRING Bacillus gelatini");
        }

        if is_valid_string_bacteroides_barnesiae_node_name(node_name) {
            return Ok("STRING Bacteroides barnesiae");
        }

        if is_valid_string_bacteroides_graminisolvens_node_name(node_name) {
            return Ok("STRING Bacteroides graminisolvens");
        }

        if is_valid_string_bacteroides_massiliensis_b84634_node_name(node_name) {
            return Ok("STRING Bacteroides massiliensis B84634");
        }

        if is_valid_string_bacteroides_pyogenes_dsm20611_node_name(node_name) {
            return Ok("STRING Bacteroides pyogenes DSM20611");
        }

        if is_valid_string_bacteroides_salyersiae_node_name(node_name) {
            return Ok("STRING Bacteroides salyersiae");
        }

        if is_valid_string_balneola_vulgaris_node_name(node_name) {
            return Ok("STRING Balneola vulgaris");
        }

        if is_valid_string_bavariicoccus_seileri_node_name(node_name) {
            return Ok("STRING Bavariicoccus seileri");
        }

        if is_valid_string_belnapia_moabensis_node_name(node_name) {
            return Ok("STRING Belnapia moabensis");
        }

        if is_valid_string_blautia_wexlerae_node_name(node_name) {
            return Ok("STRING Blautia wexlerae");
        }

        if is_valid_string_brachymonas_chironomi_node_name(node_name) {
            return Ok("STRING Brachymonas chironomi");
        }

        if is_valid_string_brevibacillus_laterosporus_dsm25_node_name(node_name) {
            return Ok("STRING Brevibacillus laterosporus DSM25");
        }

        if is_valid_string_brevundimonas_aveniformis_node_name(node_name) {
            return Ok("STRING Brevundimonas aveniformis");
        }

        if is_valid_string_brevundimonas_bacteroides_node_name(node_name) {
            return Ok("STRING Brevundimonas bacteroides");
        }

        if is_valid_string_burkholderia_nodosa_node_name(node_name) {
            return Ok("STRING Burkholderia nodosa");
        }

        if is_valid_string_butyricimonas_synergistica_node_name(node_name) {
            return Ok("STRING Butyricimonas synergistica");
        }

        if is_valid_string_campylobacter_cuniculorum_node_name(node_name) {
            return Ok("STRING Campylobacter cuniculorum");
        }

        if is_valid_string_gemmobacter_nectariphilus_node_name(node_name) {
            return Ok("STRING Gemmobacter nectariphilus");
        }

        if is_valid_string_catelliglobosispora_koreensis_node_name(node_name) {
            return Ok("STRING Catelliglobosispora koreensis");
        }

        if is_valid_string_chryseobacterium_caeni_node_name(node_name) {
            return Ok("STRING Chryseobacterium caeni");
        }

        if is_valid_string_chryseobacterium_daeguense_node_name(node_name) {
            return Ok("STRING Chryseobacterium daeguense");
        }

        if is_valid_string_chryseobacterium_gregarium_node_name(node_name) {
            return Ok("STRING Chryseobacterium gregarium");
        }

        if is_valid_string_chryseobacterium_palustre_node_name(node_name) {
            return Ok("STRING Chryseobacterium palustre");
        }

        if is_valid_string_clostridiisalibacter_paucivorans_node_name(node_name) {
            return Ok("STRING Clostridiisalibacter paucivorans");
        }

        if is_valid_string_clostridium_aminophilum_node_name(node_name) {
            return Ok("STRING Clostridium aminophilum");
        }

        if is_valid_string_clostridium_litorale_node_name(node_name) {
            return Ok("STRING Clostridium litorale");
        }

        if is_valid_string_clostridium_saccharogumia_node_name(node_name) {
            return Ok("STRING Clostridium saccharogumia");
        }

        if is_valid_string_clostridium_sporosphaeroides_node_name(node_name) {
            return Ok("STRING Clostridium sporosphaeroides");
        }

        if is_valid_string_clostridium_stercorarium_node_name(node_name) {
            return Ok("STRING Clostridium stercorarium");
        }

        if is_valid_string_clostridium_tyrobutyricum_node_name(node_name) {
            return Ok("STRING Clostridium tyrobutyricum");
        }

        if is_valid_string_clostridium_viride_node_name(node_name) {
            return Ok("STRING Clostridium viride");
        }

        if is_valid_string_cohnella_laeviribosi_node_name(node_name) {
            return Ok("STRING Cohnella laeviribosi");
        }

        if is_valid_string_conchiformibius_kuhniae_node_name(node_name) {
            return Ok("STRING Conchiformibius kuhniae");
        }

        if is_valid_string_conchiformibius_steedae_node_name(node_name) {
            return Ok("STRING Conchiformibius steedae");
        }

        if is_valid_string_corynebacterium_callunae_node_name(node_name) {
            return Ok("STRING Corynebacterium callunae");
        }

        if is_valid_string_corynebacterium_capitovis_node_name(node_name) {
            return Ok("STRING Corynebacterium capitovis");
        }

        if is_valid_string_corynebacterium_caspium_node_name(node_name) {
            return Ok("STRING Corynebacterium caspium");
        }

        if is_valid_string_corynebacterium_ciconiae_node_name(node_name) {
            return Ok("STRING Corynebacterium ciconiae");
        }

        if is_valid_string_corynebacterium_freiburgense_node_name(node_name) {
            return Ok("STRING Corynebacterium freiburgense");
        }

        if is_valid_string_corynebacterium_halotolerans_node_name(node_name) {
            return Ok("STRING Corynebacterium halotolerans");
        }

        if is_valid_string_corynebacterium_lubricantis_node_name(node_name) {
            return Ok("STRING Corynebacterium lubricantis");
        }

        if is_valid_string_corynebacterium_massiliense_node_name(node_name) {
            return Ok("STRING Corynebacterium massiliense");
        }

        if is_valid_string_corynebacterium_mastitidis_node_name(node_name) {
            return Ok("STRING Corynebacterium mastitidis");
        }

        if is_valid_string_corynebacterium_pilosum_node_name(node_name) {
            return Ok("STRING Corynebacterium pilosum");
        }

        if is_valid_string_corynebacterium_propinquum_node_name(node_name) {
            return Ok("STRING Corynebacterium propinquum");
        }

        if is_valid_string_corynebacterium_ulceribovis_node_name(node_name) {
            return Ok("STRING Corynebacterium ulceribovis");
        }

        if is_valid_string_gryllotalpicola_ginsengisoli_node_name(node_name) {
            return Ok("STRING Gryllotalpicola ginsengisoli");
        }

        if is_valid_string_cytophaga_aurantiaca_node_name(node_name) {
            return Ok("STRING Cytophaga aurantiaca");
        }

        if is_valid_string_dasania_marina_node_name(node_name) {
            return Ok("STRING Dasania marina");
        }

        if is_valid_string_deinococcus_apachensis_node_name(node_name) {
            return Ok("STRING Deinococcus apachensis");
        }

        if is_valid_string_deinococcus_aquatilis_node_name(node_name) {
            return Ok("STRING Deinococcus aquatilis");
        }

        if is_valid_string_deinococcus_frigens_node_name(node_name) {
            return Ok("STRING Deinococcus frigens");
        }

        if is_valid_string_deinococcus_marmoris_node_name(node_name) {
            return Ok("STRING Deinococcus marmoris");
        }

        if is_valid_string_deinococcus_misasensis_node_name(node_name) {
            return Ok("STRING Deinococcus misasensis");
        }

        if is_valid_string_demetria_terragena_node_name(node_name) {
            return Ok("STRING Demetria terragena");
        }

        if is_valid_string_desulfobacter_curvatus_node_name(node_name) {
            return Ok("STRING Desulfobacter curvatus");
        }

        if is_valid_string_desulfobulbus_japonicus_node_name(node_name) {
            return Ok("STRING Desulfobulbus japonicus");
        }

        if is_valid_string_desulfococcus_multivorans_node_name(node_name) {
            return Ok("STRING Desulfococcus multivorans");
        }

        if is_valid_string_desulfocurvus_vexinensis_node_name(node_name) {
            return Ok("STRING Desulfocurvus vexinensis");
        }

        if is_valid_string_desulfonatronovibrio_hydrogenovorans_node_name(node_name) {
            return Ok("STRING Desulfonatronovibrio hydrogenovorans");
        }

        if is_valid_string_desulfotomaculum_alcoholivorax_node_name(node_name) {
            return Ok("STRING Desulfotomaculum alcoholivorax");
        }

        if is_valid_string_desulfotomaculum_alkaliphilum_node_name(node_name) {
            return Ok("STRING Desulfotomaculum alkaliphilum");
        }

        if is_valid_string_desulfotomaculum_hydrothermale_node_name(node_name) {
            return Ok("STRING Desulfotomaculum hydrothermale");
        }

        if is_valid_string_desulfotomaculum_thermocisternum_node_name(node_name) {
            return Ok("STRING Desulfotomaculum thermocisternum");
        }

        if is_valid_string_desulfovibrio_acrylicus_node_name(node_name) {
            return Ok("STRING Desulfovibrio acrylicus");
        }

        if is_valid_string_desulfovibrio_alcoholivorans_node_name(node_name) {
            return Ok("STRING Desulfovibrio alcoholivorans");
        }

        if is_valid_string_desulfovibrio_alkalitolerans_node_name(node_name) {
            return Ok("STRING Desulfovibrio alkalitolerans");
        }

        if is_valid_string_desulfovibrio_aminophilus_node_name(node_name) {
            return Ok("STRING Desulfovibrio aminophilus");
        }

        if is_valid_string_desulfovibrio_bastinii_node_name(node_name) {
            return Ok("STRING Desulfovibrio bastinii");
        }

        if is_valid_string_desulfovibrio_desulfuricans_dsm642_node_name(node_name) {
            return Ok("STRING Desulfovibrio desulfuricans DSM642");
        }

        if is_valid_string_desulfovibrio_frigidus_node_name(node_name) {
            return Ok("STRING Desulfovibrio frigidus");
        }

        if is_valid_string_desulfovibrio_gigas_node_name(node_name) {
            return Ok("STRING Desulfovibrio gigas");
        }

        if is_valid_string_desulfovibrio_hydrothermalis_node_name(node_name) {
            return Ok("STRING Desulfovibrio hydrothermalis");
        }

        if is_valid_string_desulfovibrio_longus_node_name(node_name) {
            return Ok("STRING Desulfovibrio longus");
        }

        if is_valid_string_desulfovibrio_oxyclinae_node_name(node_name) {
            return Ok("STRING Desulfovibrio oxyclinae");
        }

        if is_valid_string_desulfovirgula_thermocuniculi_node_name(node_name) {
            return Ok("STRING Desulfovirgula thermocuniculi");
        }

        if is_valid_string_desulfurispora_thermophila_node_name(node_name) {
            return Ok("STRING Desulfurispora thermophila");
        }

        if is_valid_string_donghicola_xiamenensis_node_name(node_name) {
            return Ok("STRING Donghicola xiamenensis");
        }

        if is_valid_string_dyadobacter_alkalitolerans_node_name(node_name) {
            return Ok("STRING Dyadobacter alkalitolerans");
        }

        if is_valid_string_echinicola_pacifica_node_name(node_name) {
            return Ok("STRING Echinicola pacifica");
        }

        if is_valid_string_elioraea_tepidiphila_node_name(node_name) {
            return Ok("STRING Elioraea tepidiphila");
        }

        if is_valid_string_enterococcus_cecorum_node_name(node_name) {
            return Ok("STRING Enterococcus cecorum");
        }

        if is_valid_string_enterococcus_columbae_node_name(node_name) {
            return Ok("STRING Enterococcus columbae");
        }

        if is_valid_string_enterorhabdus_mucosicola_node_name(node_name) {
            return Ok("STRING Enterorhabdus mucosicola");
        }

        if is_valid_string_epilithonimonas_tenax_node_name(node_name) {
            return Ok("STRING Epilithonimonas tenax");
        }

        if is_valid_string_eremococcus_coleocola_dsm15696_node_name(node_name) {
            return Ok("STRING Eremococcus coleocola DSM15696");
        }

        if is_valid_string_erysipelothrix_tonsillarum_node_name(node_name) {
            return Ok("STRING Erysipelothrix tonsillarum");
        }

        if is_valid_string_eudoraea_adriatica_node_name(node_name) {
            return Ok("STRING Eudoraea adriatica");
        }

        if is_valid_string_ferrimicrobium_acidiphilum_node_name(node_name) {
            return Ok("STRING Ferrimicrobium acidiphilum");
        }

        if is_valid_string_ferrimonas_futtsuensis_node_name(node_name) {
            return Ok("STRING Ferrimonas futtsuensis");
        }

        if is_valid_string_flavobacterium_daejeonense_node_name(node_name) {
            return Ok("STRING Flavobacterium daejeonense");
        }

        if is_valid_string_flavobacterium_filum_node_name(node_name) {
            return Ok("STRING Flavobacterium filum");
        }

        if is_valid_string_flavobacterium_frigidarium_node_name(node_name) {
            return Ok("STRING Flavobacterium frigidarium");
        }

        if is_valid_string_flavobacterium_rivuli_node_name(node_name) {
            return Ok("STRING Flavobacterium rivuli");
        }

        if is_valid_string_flavobacterium_sasangense_node_name(node_name) {
            return Ok("STRING Flavobacterium sasangense");
        }

        if is_valid_string_flavobacterium_soli_node_name(node_name) {
            return Ok("STRING Flavobacterium soli");
        }

        if is_valid_string_flavobacterium_subsaxonicum_node_name(node_name) {
            return Ok("STRING Flavobacterium subsaxonicum");
        }

        if is_valid_string_flavobacterium_suncheonense_node_name(node_name) {
            return Ok("STRING Flavobacterium suncheonense");
        }

        if is_valid_string_flexithrix_dorotheae_node_name(node_name) {
            return Ok("STRING Flexithrix dorotheae");
        }

        if is_valid_string_gelidibacter_mesophilus_node_name(node_name) {
            return Ok("STRING Gelidibacter mesophilus");
        }

        if is_valid_string_gemella_cuniculi_node_name(node_name) {
            return Ok("STRING Gemella cuniculi");
        }

        if is_valid_string_geopsychrobacter_electrodiphilus_node_name(node_name) {
            return Ok("STRING Geopsychrobacter electrodiphilus");
        }

        if is_valid_string_geothrix_fermentans_node_name(node_name) {
            return Ok("STRING Geothrix fermentans");
        }

        if is_valid_string_gilvimarinus_chinensis_node_name(node_name) {
            return Ok("STRING Gilvimarinus chinensis");
        }

        if is_valid_string_glaciecola_punicea_node_name(node_name) {
            return Ok("STRING Glaciecola punicea");
        }

        if is_valid_string_glaciibacter_superstes_node_name(node_name) {
            return Ok("STRING Glaciibacter superstes");
        }

        if is_valid_string_glycomyces_arizonensis_node_name(node_name) {
            return Ok("STRING Glycomyces arizonensis");
        }

        if is_valid_string_gordonia_hirsuta_node_name(node_name) {
            return Ok("STRING Gordonia hirsuta");
        }

        if is_valid_string_gordonia_shandongensis_node_name(node_name) {
            return Ok("STRING Gordonia shandongensis");
        }

        if is_valid_string_gracilibacillus_lacisalsi_node_name(node_name) {
            return Ok("STRING Gracilibacillus lacisalsi");
        }

        if is_valid_string_gracilimonas_tropica_node_name(node_name) {
            return Ok("STRING Gracilimonas tropica");
        }

        if is_valid_string_gramella_echinicola_node_name(node_name) {
            return Ok("STRING Gramella echinicola");
        }

        if is_valid_string_granulicoccus_phenolivorans_node_name(node_name) {
            return Ok("STRING Granulicoccus phenolivorans");
        }

        if is_valid_string_gulosibacter_molinativorax_node_name(node_name) {
            return Ok("STRING Gulosibacter molinativorax");
        }

        if is_valid_string_hahella_ganghwensis_node_name(node_name) {
            return Ok("STRING Hahella ganghwensis");
        }

        if is_valid_string_halalkalibacillus_halophilus_node_name(node_name) {
            return Ok("STRING Halalkalibacillus halophilus");
        }

        if is_valid_string_haliea_salexigens_node_name(node_name) {
            return Ok("STRING Haliea salexigens");
        }

        if is_valid_string_halobacillus_kuroshimensis_node_name(node_name) {
            return Ok("STRING Halobacillus kuroshimensis");
        }

        if is_valid_string_halomonas_anticariensis_node_name(node_name) {
            return Ok("STRING Halomonas anticariensis");
        }

        if is_valid_string_halomonas_halocynthiae_node_name(node_name) {
            return Ok("STRING Halomonas halocynthiae");
        }

        if is_valid_string_halomonas_lutea_node_name(node_name) {
            return Ok("STRING Halomonas lutea");
        }

        if is_valid_string_halorubrum_ezzemoulense_node_name(node_name) {
            return Ok("STRING Halorubrum ezzemoulense");
        }

        if is_valid_string_hamadaea_tsunoensis_node_name(node_name) {
            return Ok("STRING Hamadaea tsunoensis");
        }

        if is_valid_string_helcococcus_sueciensis_node_name(node_name) {
            return Ok("STRING Helcococcus sueciensis");
        }

        if is_valid_string_hellea_balneolensis_node_name(node_name) {
            return Ok("STRING Hellea balneolensis");
        }

        if is_valid_string_henriciella_marina_node_name(node_name) {
            return Ok("STRING Henriciella marina");
        }

        if is_valid_string_humibacter_albus_node_name(node_name) {
            return Ok("STRING Humibacter albus");
        }

        if is_valid_string_hymenobacter_norwichensis_node_name(node_name) {
            return Ok("STRING Hymenobacter norwichensis");
        }

        if is_valid_string_jeotgalicoccus_marinus_node_name(node_name) {
            return Ok("STRING Jeotgalicoccus marinus");
        }

        if is_valid_string_jeotgalicoccus_psychrophilus_node_name(node_name) {
            return Ok("STRING Jeotgalicoccus psychrophilus");
        }

        if is_valid_string_jonesia_quinghaiensis_node_name(node_name) {
            return Ok("STRING Jonesia quinghaiensis");
        }

        if is_valid_string_kaistia_granuli_node_name(node_name) {
            return Ok("STRING Kaistia granuli");
        }

        if is_valid_string_kangiella_aquimarina_node_name(node_name) {
            return Ok("STRING Kangiella aquimarina");
        }

        if is_valid_string_kiloniella_laminariae_node_name(node_name) {
            return Ok("STRING Kiloniella laminariae");
        }

        if is_valid_string_kordiimonas_gwangyangensis_node_name(node_name) {
            return Ok("STRING Kordiimonas gwangyangensis");
        }

        if is_valid_string_kribbella_catacumbae_node_name(node_name) {
            return Ok("STRING Kribbella catacumbae");
        }

        if is_valid_string_kushneria_aurantia_node_name(node_name) {
            return Ok("STRING Kushneria aurantia");
        }

        if is_valid_string_lacticigenium_naphtae_node_name(node_name) {
            return Ok("STRING Lacticigenium naphtae");
        }

        if is_valid_string_lactobacillus_ceti_node_name(node_name) {
            return Ok("STRING Lactobacillus ceti");
        }

        if is_valid_string_lactobacillus_harbinensis_node_name(node_name) {
            return Ok("STRING Lactobacillus harbinensis");
        }

        if is_valid_string_lactobacillus_malefermentans_node_name(node_name) {
            return Ok("STRING Lactobacillus malefermentans");
        }

        if is_valid_string_lactobacillus_psittaci_node_name(node_name) {
            return Ok("STRING Lactobacillus psittaci");
        }

        if is_valid_string_legionella_lansingensis_node_name(node_name) {
            return Ok("STRING Legionella lansingensis");
        }

        if is_valid_string_legionella_moravica_node_name(node_name) {
            return Ok("STRING Legionella moravica");
        }

        if is_valid_string_legionella_shakespearei_node_name(node_name) {
            return Ok("STRING Legionella shakespearei");
        }

        if is_valid_string_leptotrichia_shahii_node_name(node_name) {
            return Ok("STRING Leptotrichia shahii");
        }

        if is_valid_string_leptotrichia_trevisanii_node_name(node_name) {
            return Ok("STRING Leptotrichia trevisanii");
        }

        if is_valid_string_leucobacter_chironomi_node_name(node_name) {
            return Ok("STRING Leucobacter chironomi");
        }

        if is_valid_string_lewinella_cohaerens_node_name(node_name) {
            return Ok("STRING Lewinella cohaerens");
        }

        if is_valid_string_lewinella_persica_node_name(node_name) {
            return Ok("STRING Lewinella persica");
        }

        if is_valid_string_loktanella_hongkongensis_node_name(node_name) {
            return Ok("STRING Loktanella hongkongensis");
        }

        if is_valid_string_longispora_albida_node_name(node_name) {
            return Ok("STRING Longispora albida");
        }

        if is_valid_string_lysobacter_concretionis_node_name(node_name) {
            return Ok("STRING Lysobacter concretionis");
        }

        if is_valid_string_marinimicrobium_agarilyticum_node_name(node_name) {
            return Ok("STRING Marinimicrobium agarilyticum");
        }

        if is_valid_string_marinobacter_daepoensis_node_name(node_name) {
            return Ok("STRING Marinobacter daepoensis");
        }

        if is_valid_string_marinobacterium_litorale_node_name(node_name) {
            return Ok("STRING Marinobacterium litorale");
        }

        if is_valid_string_marinomonas_ushuaiensis_node_name(node_name) {
            return Ok("STRING Marinomonas ushuaiensis");
        }

        if is_valid_string_marinospirillum_insulare_node_name(node_name) {
            return Ok("STRING Marinospirillum insulare");
        }

        if is_valid_string_marinospirillum_minutulum_node_name(node_name) {
            return Ok("STRING Marinospirillum minutulum");
        }

        if is_valid_string_martelella_mediterranea_node_name(node_name) {
            return Ok("STRING Martelella mediterranea");
        }

        if is_valid_string_megamonas_hypermegale_node_name(node_name) {
            return Ok("STRING Megamonas hypermegale");
        }

        if is_valid_string_megamonas_rupellensis_node_name(node_name) {
            return Ok("STRING Megamonas rupellensis");
        }

        if is_valid_string_meganema_perideroedes_node_name(node_name) {
            return Ok("STRING Meganema perideroedes");
        }

        if is_valid_string_meiothermus_cerbereus_node_name(node_name) {
            return Ok("STRING Meiothermus cerbereus");
        }

        if is_valid_string_meiothermus_taiwanensis_node_name(node_name) {
            return Ok("STRING Meiothermus taiwanensis");
        }

        if is_valid_string_meiothermus_timidus_node_name(node_name) {
            return Ok("STRING Meiothermus timidus");
        }

        if is_valid_string_mesoflavibacter_zeaxanth_dsm18436_node_name(node_name) {
            return Ok("STRING Mesoflavibacter zeaxanth. DSM18436");
        }

        if is_valid_string_mesonia_mobilis_node_name(node_name) {
            return Ok("STRING Mesonia mobilis");
        }

        if is_valid_string_metascardovia_criceti_node_name(node_name) {
            return Ok("STRING Metascardovia criceti");
        }

        if is_valid_string_methylophilus_methylotrophus_node_name(node_name) {
            return Ok("STRING Methylophilus methylotrophus");
        }

        if is_valid_string_microbacterium_gubbeenense_node_name(node_name) {
            return Ok("STRING Microbacterium gubbeenense");
        }

        if is_valid_string_microbacterium_indicum_node_name(node_name) {
            return Ok("STRING Microbacterium indicum");
        }

        if is_valid_string_microbacterium_luticocti_node_name(node_name) {
            return Ok("STRING Microbacterium luticocti");
        }

        if is_valid_string_moraxella_boevrei_node_name(node_name) {
            return Ok("STRING Moraxella boevrei");
        }

        if is_valid_string_moraxella_caprae_node_name(node_name) {
            return Ok("STRING Moraxella caprae");
        }

        if is_valid_string_mycobacterium_hassiacum_node_name(node_name) {
            return Ok("STRING Mycobacterium hassiacum");
        }

        if is_valid_string_neptunomonas_japonica_node_name(node_name) {
            return Ok("STRING Neptunomonas japonica");
        }

        if is_valid_string_nesterenkonia_alba_node_name(node_name) {
            return Ok("STRING Nesterenkonia alba");
        }

        if is_valid_string_nevskia_ramosa_node_name(node_name) {
            return Ok("STRING Nevskia ramosa");
        }

        if is_valid_string_nevskia_soli_node_name(node_name) {
            return Ok("STRING Nevskia soli");
        }

        if is_valid_string_niabella_aurantiaca_node_name(node_name) {
            return Ok("STRING Niabella aurantiaca");
        }

        if is_valid_string_nocardioides_halotolerans_node_name(node_name) {
            return Ok("STRING Nocardioides halotolerans");
        }

        if is_valid_string_nonomuraea_coxensis_node_name(node_name) {
            return Ok("STRING Nonomuraea coxensis");
        }

        if is_valid_string_novosphingobium_acidiphilum_node_name(node_name) {
            return Ok("STRING Novosphingobium acidiphilum");
        }

        if is_valid_string_oceanicaulis_alexandrii_node_name(node_name) {
            return Ok("STRING Oceanicaulis alexandrii");
        }

        if is_valid_string_oceanicola_nanhaiensis_node_name(node_name) {
            return Ok("STRING Oceanicola nanhaiensis");
        }

        if is_valid_string_oligella_ureolytica_node_name(node_name) {
            return Ok("STRING Oligella ureolytica");
        }

        if is_valid_string_olivibacter_sitiensis_node_name(node_name) {
            return Ok("STRING Olivibacter sitiensis");
        }

        if is_valid_string_ornithinimicrobium_pekingense_node_name(node_name) {
            return Ok("STRING Ornithinimicrobium pekingense");
        }

        if is_valid_string_paenibacillus_alginolyticus_node_name(node_name) {
            return Ok("STRING Paenibacillus alginolyticus");
        }

        if is_valid_string_paenibacillus_daejeonensis_node_name(node_name) {
            return Ok("STRING Paenibacillus daejeonensis");
        }

        if is_valid_string_paenibacillus_fonticola_node_name(node_name) {
            return Ok("STRING Paenibacillus fonticola");
        }

        if is_valid_string_paenibacillus_ginsengihumi_node_name(node_name) {
            return Ok("STRING Paenibacillus ginsengihumi");
        }

        if is_valid_string_paenibacillus_massiliensis_node_name(node_name) {
            return Ok("STRING Paenibacillus massiliensis");
        }

        if is_valid_string_paenibacillus_sanguinis_node_name(node_name) {
            return Ok("STRING Paenibacillus sanguinis");
        }

        if is_valid_string_paenibacillus_terrigena_node_name(node_name) {
            return Ok("STRING Paenibacillus terrigena");
        }

        if is_valid_string_pannonibacter_phragmitetus_node_name(node_name) {
            return Ok("STRING Pannonibacter phragmitetus");
        }

        if is_valid_string_parabacteroides_gordonii_node_name(node_name) {
            return Ok("STRING Parabacteroides gordonii");
        }

        if is_valid_string_paraoerskovia_marina_node_name(node_name) {
            return Ok("STRING Paraoerskovia marina");
        }

        if is_valid_string_patulibacter_americanus_node_name(node_name) {
            return Ok("STRING Patulibacter americanus");
        }

        if is_valid_string_pelosinus_fermentans_node_name(node_name) {
            return Ok("STRING Pelosinus fermentans");
        }

        if is_valid_string_perlucidibaca_piscinae_node_name(node_name) {
            return Ok("STRING Perlucidibaca piscinae");
        }

        if is_valid_string_pleomorphomonas_koreensis_node_name(node_name) {
            return Ok("STRING Pleomorphomonas koreensis");
        }

        if is_valid_string_pleomorphomonas_oryzae_node_name(node_name) {
            return Ok("STRING Pleomorphomonas oryzae");
        }

        if is_valid_string_porphyrobacter_cryptus_node_name(node_name) {
            return Ok("STRING Porphyrobacter cryptus");
        }

        if is_valid_string_porphyromonas_bennonis_node_name(node_name) {
            return Ok("STRING Porphyromonas bennonis");
        }

        if is_valid_string_porphyromonas_levii_node_name(node_name) {
            return Ok("STRING Porphyromonas levii");
        }

        if is_valid_string_porphyromonas_somerae_node_name(node_name) {
            return Ok("STRING Porphyromonas somerae");
        }

        if is_valid_string_prevotella_albensis_node_name(node_name) {
            return Ok("STRING Prevotella albensis");
        }

        if is_valid_string_prevotella_corporis_node_name(node_name) {
            return Ok("STRING Prevotella corporis");
        }

        if is_valid_string_prevotella_falsenii_node_name(node_name) {
            return Ok("STRING Prevotella falsenii");
        }

        if is_valid_string_prevotella_loescheii_node_name(node_name) {
            return Ok("STRING Prevotella loescheii");
        }

        if is_valid_string_prevotella_maculosa_node_name(node_name) {
            return Ok("STRING Prevotella maculosa");
        }

        if is_valid_string_prevotella_oris_dsm18711_node_name(node_name) {
            return Ok("STRING Prevotella oris DSM18711");
        }

        if is_valid_string_prevotella_paludivivens_node_name(node_name) {
            return Ok("STRING Prevotella paludivivens");
        }

        if is_valid_string_prevotella_shahii_node_name(node_name) {
            return Ok("STRING Prevotella shahii");
        }

        if is_valid_string_prevotella_timonensis_4401737_node_name(node_name) {
            return Ok("STRING Prevotella timonensis 4401737");
        }

        if is_valid_string_prevotella_veroralis_dsm19559_node_name(node_name) {
            return Ok("STRING Prevotella veroralis DSM19559");
        }

        if is_valid_string_propionibacterium_acidifaciens_node_name(node_name) {
            return Ok("STRING Propionibacterium acidifaciens");
        }

        if is_valid_string_propionibacterium_jensenii_node_name(node_name) {
            return Ok("STRING Propionibacterium jensenii");
        }

        if is_valid_string_propionibacterium_thoenii_node_name(node_name) {
            return Ok("STRING Propionibacterium thoenii");
        }

        if is_valid_string_proteiniphilum_acetatigenes_node_name(node_name) {
            return Ok("STRING Proteiniphilum acetatigenes");
        }

        if is_valid_string_proteocatella_sphenisci_node_name(node_name) {
            return Ok("STRING Proteocatella sphenisci");
        }

        if is_valid_string_pseudoclavibacter_soli_node_name(node_name) {
            return Ok("STRING Pseudoclavibacter soli");
        }

        if is_valid_string_pseudomonas_resinovorans_dsm21078_node_name(node_name) {
            return Ok("STRING Pseudomonas resinovorans DSM21078");
        }

        if is_valid_string_pseudonocardia_acaciae_node_name(node_name) {
            return Ok("STRING Pseudonocardia acaciae");
        }

        if is_valid_string_pseudonocardia_asaccharolytica_node_name(node_name) {
            return Ok("STRING Pseudonocardia asaccharolytica");
        }

        if is_valid_string_psychrobacter_lutiphocae_node_name(node_name) {
            return Ok("STRING Psychrobacter lutiphocae");
        }

        if is_valid_string_psychrobacter_phenylpyruvicus_node_name(node_name) {
            return Ok("STRING Psychrobacter phenylpyruvicus");
        }

        if is_valid_string_psychroflexus_tropicus_node_name(node_name) {
            return Ok("STRING Psychroflexus tropicus");
        }

        if is_valid_string_psychroserpens_burtonensis_node_name(node_name) {
            return Ok("STRING Psychroserpens burtonensis");
        }

        if is_valid_string_rathayibacter_toxicus_node_name(node_name) {
            return Ok("STRING Rathayibacter toxicus");
        }

        if is_valid_string_rheinheimera_baltica_node_name(node_name) {
            return Ok("STRING Rheinheimera baltica");
        }

        if is_valid_string_rheinheimera_perlucida_node_name(node_name) {
            return Ok("STRING Rheinheimera perlucida");
        }

        if is_valid_string_rhodonellum_psychrophilum_node_name(node_name) {
            return Ok("STRING Rhodonellum psychrophilum");
        }

        if is_valid_string_riemerella_columbina_node_name(node_name) {
            return Ok("STRING Riemerella columbina");
        }

        if is_valid_string_robiginitomaculum_antarcticum_node_name(node_name) {
            return Ok("STRING Robiginitomaculum antarcticum");
        }

        if is_valid_string_roseomonas_aerilata_node_name(node_name) {
            return Ok("STRING Roseomonas aerilata");
        }

        if is_valid_string_ruania_albidiflava_node_name(node_name) {
            return Ok("STRING Ruania albidiflava");
        }

        if is_valid_string_rubritalea_marina_node_name(node_name) {
            return Ok("STRING Rubritalea marina");
        }

        if is_valid_string_rubritepida_flocculans_node_name(node_name) {
            return Ok("STRING Rubritepida flocculans");
        }

        if is_valid_string_rudaea_cellulosilytica_node_name(node_name) {
            return Ok("STRING Rudaea cellulosilytica");
        }

        if is_valid_string_ruminococcus_gauvreauii_node_name(node_name) {
            return Ok("STRING Ruminococcus gauvreauii");
        }

        if is_valid_string_saccharibacillus_kuerlensis_node_name(node_name) {
            return Ok("STRING Saccharibacillus kuerlensis");
        }

        if is_valid_string_saccharibacter_floricola_node_name(node_name) {
            return Ok("STRING Saccharibacter floricola");
        }

        if is_valid_string_saccharospirillum_impatiens_node_name(node_name) {
            return Ok("STRING Saccharospirillum impatiens");
        }

        if is_valid_string_salinarimonas_rosea_node_name(node_name) {
            return Ok("STRING Salinarimonas rosea");
        }

        if is_valid_string_salinicoccus_albus_node_name(node_name) {
            return Ok("STRING Salinicoccus albus");
        }

        if is_valid_string_salinimicrobium_terrae_node_name(node_name) {
            return Ok("STRING Salinimicrobium terrae");
        }

        if is_valid_string_salinimonas_chungwhensis_node_name(node_name) {
            return Ok("STRING Salinimonas chungwhensis");
        }

        if is_valid_string_salipiger_mucosus_node_name(node_name) {
            return Ok("STRING Salipiger mucosus");
        }

        if is_valid_string_salsuginibacillus_kocurii_node_name(node_name) {
            return Ok("STRING Salsuginibacillus kocurii");
        }

        if is_valid_string_sandarakinorhabdus_limnophila_node_name(node_name) {
            return Ok("STRING Sandarakinorhabdus limnophila");
        }

        if is_valid_string_schlesneria_paludicola_node_name(node_name) {
            return Ok("STRING Schlesneria paludicola");
        }

        if is_valid_string_sediminimonas_qiaohouensis_node_name(node_name) {
            return Ok("STRING Sediminimonas qiaohouensis");
        }

        if is_valid_string_segetibacter_koreensis_node_name(node_name) {
            return Ok("STRING Segetibacter koreensis");
        }

        if is_valid_string_selenomonas_bovis_node_name(node_name) {
            return Ok("STRING Selenomonas bovis");
        }

        if is_valid_string_serinicoccus_marinus_node_name(node_name) {
            return Ok("STRING Serinicoccus marinus");
        }

        if is_valid_string_shimazuella_kribbensis_node_name(node_name) {
            return Ok("STRING Shimazuella kribbensis");
        }

        if is_valid_string_silanimonas_lenta_node_name(node_name) {
            return Ok("STRING Silanimonas lenta");
        }

        if is_valid_string_simplicispira_psychrophila_node_name(node_name) {
            return Ok("STRING Simplicispira psychrophila");
        }

        if is_valid_string_solimonas_variicoloris_node_name(node_name) {
            return Ok("STRING Solimonas variicoloris");
        }

        if is_valid_string_solimonas_flava_node_name(node_name) {
            return Ok("STRING Solimonas flava");
        }

        if is_valid_string_smaragdicoccus_niigatensis_node_name(node_name) {
            return Ok("STRING Smaragdicoccus niigatensis");
        }

        if is_valid_string_solimonas_soli_node_name(node_name) {
            return Ok("STRING Solimonas soli");
        }

        if is_valid_string_solobacterium_moorei_dsm22971_node_name(node_name) {
            return Ok("STRING Solobacterium moorei DSM22971");
        }

        if is_valid_string_sphingomonas_astaxanthinifaciens_node_name(node_name) {
            return Ok("STRING Sphingomonas astaxanthinifaciens");
        }

        if is_valid_string_sphingomonas_sanxanigenens_node_name(node_name) {
            return Ok("STRING Sphingomonas sanxanigenens");
        }

        if is_valid_string_sphingopyxis_baekryungensis_node_name(node_name) {
            return Ok("STRING Sphingopyxis baekryungensis");
        }

        if is_valid_string_spirochaeta_bajacaliforniensis_node_name(node_name) {
            return Ok("STRING Spirochaeta bajacaliforniensis");
        }

        if is_valid_string_spirosoma_luteum_node_name(node_name) {
            return Ok("STRING Spirosoma luteum");
        }

        if is_valid_string_spirosoma_panaciterrae_node_name(node_name) {
            return Ok("STRING Spirosoma panaciterrae");
        }

        if is_valid_string_spirosoma_spitsbergense_node_name(node_name) {
            return Ok("STRING Spirosoma spitsbergense");
        }

        if is_valid_string_spongiibacter_tropicus_node_name(node_name) {
            return Ok("STRING Spongiibacter tropicus");
        }

        if is_valid_string_sporolactobacillus_vineae_node_name(node_name) {
            return Ok("STRING Sporolactobacillus vineae");
        }

        if is_valid_string_sporomusa_ovata_node_name(node_name) {
            return Ok("STRING Sporomusa ovata");
        }

        if is_valid_string_sporosarcina_ureae_node_name(node_name) {
            return Ok("STRING Sporosarcina ureae");
        }

        if is_valid_string_stenoxybacter_acetivorans_node_name(node_name) {
            return Ok("STRING Stenoxybacter acetivorans");
        }

        if is_valid_string_streptococcus_caballi_node_name(node_name) {
            return Ok("STRING Streptococcus caballi");
        }

        if is_valid_string_streptococcus_castoreus_node_name(node_name) {
            return Ok("STRING Streptococcus castoreus");
        }

        if is_valid_string_streptococcus_devriesei_node_name(node_name) {
            return Ok("STRING Streptococcus devriesei");
        }

        if is_valid_string_streptococcus_didelphis_node_name(node_name) {
            return Ok("STRING Streptococcus didelphis");
        }

        if is_valid_string_streptococcus_entericus_node_name(node_name) {
            return Ok("STRING Streptococcus entericus");
        }

        if is_valid_string_streptococcus_ferus_node_name(node_name) {
            return Ok("STRING Streptococcus ferus");
        }

        if is_valid_string_streptococcus_henryi_node_name(node_name) {
            return Ok("STRING Streptococcus henryi");
        }

        if is_valid_string_streptococcus_marimammalium_node_name(node_name) {
            return Ok("STRING Streptococcus marimammalium");
        }

        if is_valid_string_streptococcus_massiliensis_node_name(node_name) {
            return Ok("STRING Streptococcus massiliensis");
        }

        if is_valid_string_streptococcus_merionis_node_name(node_name) {
            return Ok("STRING Streptococcus merionis");
        }

        if is_valid_string_streptococcus_minor_node_name(node_name) {
            return Ok("STRING Streptococcus minor");
        }

        if is_valid_string_streptococcus_orisratti_node_name(node_name) {
            return Ok("STRING Streptococcus orisratti");
        }

        if is_valid_string_streptococcus_ovis_node_name(node_name) {
            return Ok("STRING Streptococcus ovis");
        }

        if is_valid_string_faecalicoccus_pleomorphus_node_name(node_name) {
            return Ok("STRING Faecalicoccus pleomorphus");
        }

        if is_valid_string_streptococcus_plurextorum_node_name(node_name) {
            return Ok("STRING Streptococcus plurextorum");
        }

        if is_valid_string_streptococcus_porci_node_name(node_name) {
            return Ok("STRING Streptococcus porci");
        }

        if is_valid_string_streptococcus_thoraltensis_node_name(node_name) {
            return Ok("STRING Streptococcus thoraltensis");
        }

        if is_valid_string_streptomyces_flavidovirens_node_name(node_name) {
            return Ok("STRING Streptomyces flavidovirens");
        }

        if is_valid_string_streptomyces_scabrisporus_node_name(node_name) {
            return Ok("STRING Streptomyces scabrisporus");
        }

        if is_valid_string_streptomyces_sulphureus_node_name(node_name) {
            return Ok("STRING Streptomyces sulphureus");
        }

        if is_valid_string_streptomyces_vitaminophilus_node_name(node_name) {
            return Ok("STRING Streptomyces vitaminophilus");
        }

        if is_valid_string_sulfurihydrogenibium_subterraneum_node_name(node_name) {
            return Ok("STRING Sulfurihydrogenibium subterraneum");
        }

        if is_valid_string_sulfurospirillum_arcachonense_node_name(node_name) {
            return Ok("STRING Sulfurospirillum arcachonense");
        }

        if is_valid_string_tepidiphilus_margaritifer_node_name(node_name) {
            return Ok("STRING Tepidiphilus margaritifer");
        }

        if is_valid_string_terasakiella_pusilla_node_name(node_name) {
            return Ok("STRING Terasakiella pusilla");
        }

        if is_valid_string_tetragenococcus_muriaticus_node_name(node_name) {
            return Ok("STRING Tetragenococcus muriaticus");
        }

        if is_valid_string_thalassobacter_arenae_node_name(node_name) {
            return Ok("STRING Thalassobacter arenae");
        }

        if is_valid_string_thalassospira_xiamenensis_node_name(node_name) {
            return Ok("STRING Thalassospira xiamenensis");
        }

        if is_valid_string_thauera_linaloolentis_node_name(node_name) {
            return Ok("STRING Thauera linaloolentis");
        }

        if is_valid_string_thermithiobacillus_tepidarius_node_name(node_name) {
            return Ok("STRING Thermithiobacillus tepidarius");
        }

        if is_valid_string_thermodesulfatator_atlanticus_node_name(node_name) {
            return Ok("STRING Thermodesulfatator atlanticus");
        }

        if is_valid_string_thermodesulfobacterium_hveragerdense_node_name(node_name) {
            return Ok("STRING Thermodesulfobacterium hveragerdense");
        }

        if is_valid_string_thermodesulfobacterium_thermophilum_node_name(node_name) {
            return Ok("STRING Thermodesulfobacterium thermophilum");
        }

        if is_valid_string_thermodesulfovibrio_thiophilus_node_name(node_name) {
            return Ok("STRING Thermodesulfovibrio thiophilus");
        }

        if is_valid_string_thermomonas_fusca_node_name(node_name) {
            return Ok("STRING Thermomonas fusca");
        }

        if is_valid_string_thermus_antranikianii_node_name(node_name) {
            return Ok("STRING Thermus antranikianii");
        }

        if is_valid_string_thermus_igniterrae_node_name(node_name) {
            return Ok("STRING Thermus igniterrae");
        }

        if is_valid_string_thermus_islandicus_node_name(node_name) {
            return Ok("STRING Thermus islandicus");
        }

        if is_valid_string_thiobacillus_denitrificans_dsm12475_node_name(node_name) {
            return Ok("STRING Thiobacillus denitrificans DSM12475");
        }

        if is_valid_string_thiobacillus_thioparus_node_name(node_name) {
            return Ok("STRING Thiobacillus thioparus");
        }

        if is_valid_string_thiothrix_disciformis_node_name(node_name) {
            return Ok("STRING Thiothrix disciformis");
        }

        if is_valid_string_thiothrix_flexilis_node_name(node_name) {
            return Ok("STRING Thiothrix flexilis");
        }

        if is_valid_string_thiothrix_lacustris_node_name(node_name) {
            return Ok("STRING Thiothrix lacustris");
        }

        if is_valid_string_tuberibacillus_calidus_node_name(node_name) {
            return Ok("STRING Tuberibacillus calidus");
        }

        if is_valid_string_uliginosibacterium_gangwonense_node_name(node_name) {
            return Ok("STRING Uliginosibacterium gangwonense");
        }

        if is_valid_string_varibaculum_cambriense_dsm15806_node_name(node_name) {
            return Ok("STRING Varibaculum cambriense DSM15806");
        }

        if is_valid_string_veillonella_magna_node_name(node_name) {
            return Ok("STRING Veillonella magna");
        }

        if is_valid_string_vitreoscilla_stercoraria_node_name(node_name) {
            return Ok("STRING Vitreoscilla stercoraria");
        }

        if is_valid_string_weissella_halotolerans_node_name(node_name) {
            return Ok("STRING Weissella halotolerans");
        }

        if is_valid_string_wenxinia_marina_node_name(node_name) {
            return Ok("STRING Wenxinia marina");
        }

        if is_valid_string_wohlfahrtiimonas_chitiniclastica_node_name(node_name) {
            return Ok("STRING Wohlfahrtiimonas chitiniclastica");
        }

        if is_valid_string_woodsholea_maritima_node_name(node_name) {
            return Ok("STRING Woodsholea maritima");
        }

        if is_valid_string_xenophilus_azovorans_node_name(node_name) {
            return Ok("STRING Xenophilus azovorans");
        }

        if is_valid_string_yaniella_halotolerans_node_name(node_name) {
            return Ok("STRING Yaniella halotolerans");
        }

        if is_valid_string_zavarzinella_formosa_node_name(node_name) {
            return Ok("STRING Zavarzinella formosa");
        }

        if is_valid_string_zymophilus_raffinosivorans_node_name(node_name) {
            return Ok("STRING Zymophilus raffinosivorans");
        }

        if is_valid_string_thiomicrospira_arctica_node_name(node_name) {
            return Ok("STRING Thiomicrospira arctica");
        }

        if is_valid_string_thiomicrospira_pelophila_node_name(node_name) {
            return Ok("STRING Thiomicrospira pelophila");
        }

        if is_valid_string_thiomicrospira_sp_kp2_node_name(node_name) {
            return Ok("STRING Thiomicrospira sp. Kp2");
        }

        if is_valid_string_pseudomonas_stutzeri_dsm10701_node_name(node_name) {
            return Ok("STRING Pseudomonas stutzeri DSM10701");
        }

        if is_valid_string_nafulsella_turpanensis_node_name(node_name) {
            return Ok("STRING Nafulsella turpanensis");
        }

        if is_valid_string_treponema_sp_jc4_node_name(node_name) {
            return Ok("STRING Treponema sp. JC4");
        }

        if is_valid_string_pseudomonas_protegens_cha0_node_name(node_name) {
            return Ok("STRING Pseudomonas protegens CHA0");
        }

        if is_valid_string_morganella_morganii_node_name(node_name) {
            return Ok("STRING Morganella morganii");
        }

        if is_valid_string_treponema_maltophilum_node_name(node_name) {
            return Ok("STRING Treponema maltophilum");
        }

        if is_valid_string_treponema_medium_node_name(node_name) {
            return Ok("STRING Treponema medium");
        }

        if is_valid_string_treponema_socranskii_paredis_node_name(node_name) {
            return Ok("STRING Treponema socranskii paredis");
        }

        if is_valid_string_olsenella_profusa_node_name(node_name) {
            return Ok("STRING Olsenella profusa");
        }

        if is_valid_string_actinomyces_massiliensis_node_name(node_name) {
            return Ok("STRING Actinomyces massiliensis");
        }

        if is_valid_string_treponema_socranskii_vpidr56br1116_node_name(node_name) {
            return Ok("STRING Treponema socranskii VPIDR56BR1116");
        }

        if is_valid_string_corynebacterium_pyruviciproducens_node_name(node_name) {
            return Ok("STRING Corynebacterium pyruviciproducens");
        }

        if is_valid_string_deferrisoma_camini_node_name(node_name) {
            return Ok("STRING Deferrisoma camini");
        }

        if is_valid_string_amycolatopsis_orientalis_node_name(node_name) {
            return Ok("STRING Amycolatopsis orientalis");
        }

        if is_valid_string_bosea_sp_117_node_name(node_name) {
            return Ok("STRING Bosea sp. 117");
        }

        if is_valid_string_bradyrhizobium_sp_doa9_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. DOA9");
        }

        if is_valid_string_weissella_confusa_node_name(node_name) {
            return Ok("STRING Weissella confusa");
        }

        if is_valid_string_nocardia_cyriacigeorgica_node_name(node_name) {
            return Ok("STRING Nocardia cyriacigeorgica");
        }

        if is_valid_string_glaciecola_lipolytica_node_name(node_name) {
            return Ok("STRING Glaciecola lipolytica");
        }

        if is_valid_string_capnocytophaga_sp_f0381_node_name(node_name) {
            return Ok("STRING Capnocytophaga sp. F0381");
        }

        if is_valid_string_selenomonas_sp_f0429_node_name(node_name) {
            return Ok("STRING Selenomonas sp. F0429");
        }

        if is_valid_string_porphyromonas_catoniae_node_name(node_name) {
            return Ok("STRING Porphyromonas catoniae");
        }

        if is_valid_string_veillonella_atypica_kon_node_name(node_name) {
            return Ok("STRING Veillonella atypica KON");
        }

        if is_valid_string_clostridium_acidurici_node_name(node_name) {
            return Ok("STRING Clostridium acidurici");
        }

        if is_valid_string_bacterium_sp_jkg1_node_name(node_name) {
            return Ok("STRING Bacterium sp. JKG1");
        }

        if is_valid_string_filamentous_cyanobacterium_node_name(node_name) {
            return Ok("STRING filamentous cyanobacterium");
        }

        if is_valid_string_glaciecola_mesophila_node_name(node_name) {
            return Ok("STRING Glaciecola mesophila");
        }

        if is_valid_string_spiroplasma_melliferum_node_name(node_name) {
            return Ok("STRING Spiroplasma melliferum");
        }

        if is_valid_string_mycoplasma_hyorhinis_gdl1_node_name(node_name) {
            return Ok("STRING Mycoplasma hyorhinis GDL1");
        }

        if is_valid_string_alishewanella_jeotgali_node_name(node_name) {
            return Ok("STRING Alishewanella jeotgali");
        }

        if is_valid_string_glaciecola_psychrophila_node_name(node_name) {
            return Ok("STRING Glaciecola psychrophila");
        }

        if is_valid_string_thaumarchaeota_sp_scgc_ab629i23_node_name(node_name) {
            return Ok("STRING Thaumarchaeota sp. SCGC AB629I23");
        }

        if is_valid_string_nitrospina_sp_ab629b18_node_name(node_name) {
            return Ok("STRING Nitrospina sp. AB629B18");
        }

        if is_valid_string_xanthomonas_fragariae_node_name(node_name) {
            return Ok("STRING Xanthomonas fragariae");
        }

        if is_valid_string_dehalobacter_sp_cf_node_name(node_name) {
            return Ok("STRING Dehalobacter sp. CF");
        }

        if is_valid_string_nitrosomonas_cryotolerans_node_name(node_name) {
            return Ok("STRING Nitrosomonas cryotolerans");
        }

        if is_valid_string_bacillus_vireti_node_name(node_name) {
            return Ok("STRING Bacillus vireti");
        }

        if is_valid_string_flavobacterium_sp_83_node_name(node_name) {
            return Ok("STRING Flavobacterium sp. 83");
        }

        if is_valid_string_methylobacterium_sp_88a_node_name(node_name) {
            return Ok("STRING Methylobacterium sp. 88A");
        }

        if is_valid_string_xanthobacter_sp_126_node_name(node_name) {
            return Ok("STRING Xanthobacter sp. 126");
        }

        if is_valid_string_arthrobacter_sp_35w_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. 35W");
        }

        if is_valid_string_bacillus_sp_37ma_node_name(node_name) {
            return Ok("STRING Bacillus sp. 37MA");
        }

        if is_valid_string_halococcus_hamelinensis_node_name(node_name) {
            return Ok("STRING Halococcus hamelinensis");
        }

        if is_valid_string_rhizobium_sp_ccge510_node_name(node_name) {
            return Ok("STRING Rhizobium sp. CCGE510");
        }

        if is_valid_string_methylotenera_mobilis_13_node_name(node_name) {
            return Ok("STRING Methylotenera mobilis 13");
        }

        if is_valid_string_lactobacillus_vini_node_name(node_name) {
            return Ok("STRING Lactobacillus vini");
        }

        if is_valid_string_nocardia_brasiliensis_node_name(node_name) {
            return Ok("STRING Nocardia brasiliensis");
        }

        if is_valid_string_streptomyces_hygroscopicus_node_name(node_name) {
            return Ok("STRING Streptomyces hygroscopicus");
        }

        if is_valid_string_bacillus_sp_l1_node_name(node_name) {
            return Ok("STRING Bacillus sp. L1");
        }

        if is_valid_string_streptomyces_somaliensis_node_name(node_name) {
            return Ok("STRING Streptomyces somaliensis");
        }

        if is_valid_string_cellvibrio_sp_br_node_name(node_name) {
            return Ok("STRING Cellvibrio sp. BR");
        }

        if is_valid_string_bartonella_koehlerae_node_name(node_name) {
            return Ok("STRING Bartonella koehlerae");
        }

        if is_valid_string_methylocystis_parvus_node_name(node_name) {
            return Ok("STRING Methylocystis parvus");
        }

        if is_valid_string_pseudomonas_fragi_node_name(node_name) {
            return Ok("STRING Pseudomonas fragi");
        }

        if is_valid_string_vibrio_cyclitrophicus_node_name(node_name) {
            return Ok("STRING Vibrio cyclitrophicus");
        }

        if is_valid_string_lactobacillus_pentosus_node_name(node_name) {
            return Ok("STRING Lactobacillus pentosus");
        }

        if is_valid_string_salinispora_pacifica_node_name(node_name) {
            return Ok("STRING Salinispora pacifica");
        }

        if is_valid_string_nocardiopsis_sp_cnt312_node_name(node_name) {
            return Ok("STRING Nocardiopsis sp. CNT312");
        }

        if is_valid_string_streptomyces_sp_cnh099_node_name(node_name) {
            return Ok("STRING Streptomyces sp. CNH099");
        }

        if is_valid_string_saccharomonospora_sp_cnq490_node_name(node_name) {
            return Ok("STRING Saccharomonospora sp. CNQ490");
        }

        if is_valid_string_formosa_sp_ak20_node_name(node_name) {
            return Ok("STRING Formosa sp. AK20");
        }

        if is_valid_string_endozoicomonas_numazuensis_node_name(node_name) {
            return Ok("STRING Endozoicomonas numazuensis");
        }

        if is_valid_string_lactobacillus_curieae_node_name(node_name) {
            return Ok("STRING Lactobacillus curieae");
        }

        if is_valid_string_enterococcus_dispar_node_name(node_name) {
            return Ok("STRING Enterococcus dispar");
        }

        if is_valid_string_enterococcus_saccharolyticus_node_name(node_name) {
            return Ok("STRING Enterococcus saccharolyticus");
        }

        if is_valid_string_enterococcus_durans_node_name(node_name) {
            return Ok("STRING Enterococcus durans");
        }

        if is_valid_string_enterococcus_avium_node_name(node_name) {
            return Ok("STRING Enterococcus avium");
        }

        if is_valid_string_enterococcus_sulfureus_node_name(node_name) {
            return Ok("STRING Enterococcus sulfureus");
        }

        if is_valid_string_staphylococcus_intermedius_node_name(node_name) {
            return Ok("STRING Staphylococcus intermedius");
        }

        if is_valid_string_providencia_burhodogranariea_node_name(node_name) {
            return Ok("STRING Providencia burhodogranariea");
        }

        if is_valid_string_providencia_rettgeri_dmel1_node_name(node_name) {
            return Ok("STRING Providencia rettgeri Dmel1");
        }

        if is_valid_string_phycisphaera_mikurensis_node_name(node_name) {
            return Ok("STRING Phycisphaera mikurensis");
        }

        if is_valid_string_wigglesworthia_glossinidia_sp_gmo_node_name(node_name) {
            return Ok("STRING Wigglesworthia glossinidia sp. Gmo");
        }

        if is_valid_string_chlamydia_gallinacea_node_name(node_name) {
            return Ok("STRING Chlamydia gallinacea");
        }

        if is_valid_string_corallococcus_coralloides_node_name(node_name) {
            return Ok("STRING Corallococcus coralloides");
        }

        if is_valid_string_novosphingobium_sp_ap12_node_name(node_name) {
            return Ok("STRING Novosphingobium sp. AP12");
        }

        if is_valid_string_sphingobium_sp_ap49_node_name(node_name) {
            return Ok("STRING Sphingobium sp. AP49");
        }

        if is_valid_string_rhizobium_sp_cf080_node_name(node_name) {
            return Ok("STRING Rhizobium sp. CF080");
        }

        if is_valid_string_rhizobium_sp_cf122_node_name(node_name) {
            return Ok("STRING Rhizobium sp. CF122");
        }

        if is_valid_string_flavobacterium_sp_cf136_node_name(node_name) {
            return Ok("STRING Flavobacterium sp. CF136");
        }

        if is_valid_string_herbaspirillum_sp_cf444_node_name(node_name) {
            return Ok("STRING Herbaspirillum sp. CF444");
        }

        if is_valid_string_pseudomonas_sp_gm21_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. GM21");
        }

        if is_valid_string_herbaspirillum_sp_yr522_node_name(node_name) {
            return Ok("STRING Herbaspirillum sp. YR522");
        }

        if is_valid_string_phyllobacterium_sp_yr531_node_name(node_name) {
            return Ok("STRING Phyllobacterium sp. YR531");
        }

        if is_valid_string_acinetobacter_sp_cip102129_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. CIP102129");
        }

        if is_valid_string_acinetobacter_sp_cip562_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. CIP56.2");
        }

        if is_valid_string_rickettsia_helvetica_node_name(node_name) {
            return Ok("STRING Rickettsia helvetica");
        }

        if is_valid_string_pelagibacter_ubique_himb058_node_name(node_name) {
            return Ok("STRING Pelagibacter ubique HIMB058");
        }

        if is_valid_string_lysinibacillus_varians_node_name(node_name) {
            return Ok("STRING Lysinibacillus varians");
        }

        if is_valid_string_blastococcus_saxobsidens_node_name(node_name) {
            return Ok("STRING Blastococcus saxobsidens");
        }

        if is_valid_string_bifidobacterium_asteroides_node_name(node_name) {
            return Ok("STRING Bifidobacterium asteroides");
        }

        if is_valid_string_pseudomonas_pseudoalcaligenes_kf707_node_name(node_name) {
            return Ok("STRING Pseudomonas pseudoalcaligenes KF707");
        }

        if is_valid_string_arthrobacter_sp_9mfcol31_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. 9MFCol31");
        }

        if is_valid_string_leifsonia_sp_109_node_name(node_name) {
            return Ok("STRING Leifsonia sp. 109");
        }

        if is_valid_string_rhodospirillum_photometricum_node_name(node_name) {
            return Ok("STRING Rhodospirillum photometricum");
        }

        if is_valid_string_mesoaciditoga_lauensis_node_name(node_name) {
            return Ok("STRING Mesoaciditoga lauensis");
        }

        if is_valid_string_mycobacterium_phlei_node_name(node_name) {
            return Ok("STRING Mycobacterium phlei");
        }

        if is_valid_string_arcticibacter_svalbardensis_node_name(node_name) {
            return Ok("STRING Arcticibacter svalbardensis");
        }

        if is_valid_string_sulfurospirillum_multivorans_node_name(node_name) {
            return Ok("STRING Sulfurospirillum multivorans");
        }

        if is_valid_string_phaeospirillum_molischianum_node_name(node_name) {
            return Ok("STRING Phaeospirillum molischianum");
        }

        if is_valid_string_micromonospora_lupini_node_name(node_name) {
            return Ok("STRING Micromonospora lupini");
        }

        if is_valid_string_tsukamurella_sp_1534_node_name(node_name) {
            return Ok("STRING Tsukamurella sp. 1534");
        }

        if is_valid_string_rahnella_aquatilis_hx2_node_name(node_name) {
            return Ok("STRING Rahnella aquatilis HX2");
        }

        if is_valid_string_thermococcus_zilligii_node_name(node_name) {
            return Ok("STRING Thermococcus zilligii");
        }

        if is_valid_string_arthrobacter_sp_161mfsha21_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. 161MFSha21");
        }

        if is_valid_string_arthrobacter_sp_162mfsha11_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. 162MFSha11");
        }

        if is_valid_string_microbacterium_sp_292mf_node_name(node_name) {
            return Ok("STRING Microbacterium sp. 292MF");
        }

        if is_valid_string_microbacterium_paraoxydans_77mftsu32_node_name(node_name) {
            return Ok("STRING Microbacterium paraoxydans 77MFTsu32");
        }

        if is_valid_string_pseudomonas_mandelii_36mfcvi11_node_name(node_name) {
            return Ok("STRING Pseudomonas mandelii 36MFCvi11");
        }

        if is_valid_string_peptoclostridium_difficile_cd160_node_name(node_name) {
            return Ok("STRING Peptoclostridium difficile CD160");
        }

        if is_valid_string_leuconostoc_pseudomesenteroides_node_name(node_name) {
            return Ok("STRING Leuconostoc pseudomesenteroides");
        }

        if is_valid_string_streptococcus_agalactiae_lmg14747_node_name(node_name) {
            return Ok("STRING Streptococcus agalactiae LMG14747");
        }

        if is_valid_string_streptomyces_sp_lappah108_node_name(node_name) {
            return Ok("STRING Streptomyces sp. LaPpAH108");
        }

        if is_valid_string_streptomyces_sp_mspmpm5_node_name(node_name) {
            return Ok("STRING Streptomyces sp. MspMPM5");
        }

        if is_valid_string_streptomyces_sp_hmica12_node_name(node_name) {
            return Ok("STRING Streptomyces sp. HmicA12");
        }

        if is_valid_string_achromobacter_piechaudii_hle_node_name(node_name) {
            return Ok("STRING Achromobacter piechaudii HLE");
        }

        if is_valid_string_agrobacterium_albertimagni_node_name(node_name) {
            return Ok("STRING Agrobacterium albertimagni");
        }

        if is_valid_string_methylacidiphilum_fumariolicum_node_name(node_name) {
            return Ok("STRING Methylacidiphilum fumariolicum");
        }

        if is_valid_string_tumebacillus_flagellatus_node_name(node_name) {
            return Ok("STRING Tumebacillus flagellatus");
        }

        if is_valid_string_streptomyces_sp_lappah95_node_name(node_name) {
            return Ok("STRING Streptomyces sp. LaPpAH95");
        }

        if is_valid_string_streptomyces_sp_amel2xe9_node_name(node_name) {
            return Ok("STRING Streptomyces sp. Amel2xE9");
        }

        if is_valid_string_streptomyces_sp_atexabd23_node_name(node_name) {
            return Ok("STRING Streptomyces sp. ATexABD23");
        }

        if is_valid_string_streptomyces_sp_bolea5_node_name(node_name) {
            return Ok("STRING Streptomyces sp. BoleA5");
        }

        if is_valid_string_streptomyces_sp_pstaah124_node_name(node_name) {
            return Ok("STRING Streptomyces sp. PsTaAH124");
        }

        if is_valid_string_streptomyces_sp_fxanac1_node_name(node_name) {
            return Ok("STRING Streptomyces sp. FxanaC1");
        }

        if is_valid_string_variovorax_paradoxus_110b_node_name(node_name) {
            return Ok("STRING Variovorax paradoxus 110B");
        }

        if is_valid_string_mycobacterium_sp_155_node_name(node_name) {
            return Ok("STRING Mycobacterium sp. 155");
        }

        if is_valid_string_arthrobacter_sp_135mfcol51_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. 135MFCol51");
        }

        if is_valid_string_thioalkalivibrio_sp_aljt_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. ALJT");
        }

        if is_valid_string_thioalkalivibrio_sp_ald1_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. ALD1");
        }

        if is_valid_string_thioalkalivibrio_sp_almg11_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. ALMg11");
        }

        if is_valid_string_thioalkalivibrio_sp_ale31_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. ALE31");
        }

        if is_valid_string_thiomonas_sp_fbcd_node_name(node_name) {
            return Ok("STRING Thiomonas sp. FBCd");
        }

        if is_valid_string_prevotella_sp_10_node_name(node_name) {
            return Ok("STRING Prevotella sp. 10");
        }

        if is_valid_string_desulfurobacterium_sp_tc51_node_name(node_name) {
            return Ok("STRING Desulfurobacterium sp. TC51");
        }

        if is_valid_string_persephonella_sp_if05l8_node_name(node_name) {
            return Ok("STRING Persephonella sp. IF05L8");
        }

        if is_valid_string_persephonella_sp_km09lau8_node_name(node_name) {
            return Ok("STRING Persephonella sp. KM09Lau8");
        }

        if is_valid_string_enterococcus_malodoratus_node_name(node_name) {
            return Ok("STRING Enterococcus malodoratus");
        }

        if is_valid_string_enterococcus_raffinosus_node_name(node_name) {
            return Ok("STRING Enterococcus raffinosus");
        }

        if is_valid_string_enterococcus_villorum_node_name(node_name) {
            return Ok("STRING Enterococcus villorum");
        }

        if is_valid_string_enterococcus_asini_node_name(node_name) {
            return Ok("STRING Enterococcus asini");
        }

        if is_valid_string_enterococcus_pallens_node_name(node_name) {
            return Ok("STRING Enterococcus pallens");
        }

        if is_valid_string_enterococcus_haemoperoxidus_node_name(node_name) {
            return Ok("STRING Enterococcus haemoperoxidus");
        }

        if is_valid_string_enterococcus_moraviensis_node_name(node_name) {
            return Ok("STRING Enterococcus moraviensis");
        }

        if is_valid_string_enterococcus_phoeniculicola_node_name(node_name) {
            return Ok("STRING Enterococcus phoeniculicola");
        }

        if is_valid_string_enterococcus_caccae_node_name(node_name) {
            return Ok("STRING Enterococcus caccae");
        }

        if is_valid_string_enterococcus_gilvus_node_name(node_name) {
            return Ok("STRING Enterococcus gilvus");
        }

        if is_valid_string_thioalkalivibrio_sp_alj7_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. ALJ7");
        }

        if is_valid_string_thioalkalivibrio_sp_alj11_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. ALJ11");
        }

        if is_valid_string_thioalkalivibrio_sp_alj16_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. ALJ16");
        }

        if is_valid_string_staphylococcus_equorum_node_name(node_name) {
            return Ok("STRING Staphylococcus equorum");
        }

        if is_valid_string_bordetella_sp_fb8_node_name(node_name) {
            return Ok("STRING Bordetella sp. FB8");
        }

        if is_valid_string_nocardia_sp_bmg111209_node_name(node_name) {
            return Ok("STRING Nocardia sp. BMG111209");
        }

        if is_valid_string_ureibacillus_thermosphaericus_node_name(node_name) {
            return Ok("STRING Ureibacillus thermosphaericus");
        }

        if is_valid_string_streptomyces_auratus_node_name(node_name) {
            return Ok("STRING Streptomyces auratus");
        }

        if is_valid_string_ruminococcus_bicirculans_node_name(node_name) {
            return Ok("STRING Ruminococcus bicirculans");
        }

        if is_valid_string_maricaulis_sp_jl2009_node_name(node_name) {
            return Ok("STRING Maricaulis sp. JL2009");
        }

        if is_valid_string_streptococcus_sp_acc21_node_name(node_name) {
            return Ok("STRING Streptococcus sp. ACC21");
        }

        if is_valid_string_eubacterium_nodatum_node_name(node_name) {
            return Ok("STRING Eubacterium nodatum");
        }

        if is_valid_string_leptospirillum_ferrooxidans_node_name(node_name) {
            return Ok("STRING Leptospirillum ferrooxidans");
        }

        if is_valid_string_phytoplasma_peanut_witchesbroom_node_name(node_name) {
            return Ok("STRING Phytoplasma peanut witchesbroom");
        }

        if is_valid_string_francisella_noatunensis_node_name(node_name) {
            return Ok("STRING Francisella noatunensis");
        }

        if is_valid_string_pseudomonas_sp_hys_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. HYS");
        }

        if is_valid_string_rhodanobacter_spathiphylli_node_name(node_name) {
            return Ok("STRING Rhodanobacter spathiphylli");
        }

        if is_valid_string_rhodanobacter_fulvus_node_name(node_name) {
            return Ok("STRING Rhodanobacter fulvus");
        }

        if is_valid_string_rhodanobacter_thiooxydans_node_name(node_name) {
            return Ok("STRING Rhodanobacter thiooxydans");
        }

        if is_valid_string_sulfuricella_denitrificans_node_name(node_name) {
            return Ok("STRING Sulfuricella denitrificans");
        }

        if is_valid_string_clostridium_sp_12_node_name(node_name) {
            return Ok("STRING Clostridium sp. 12");
        }

        if is_valid_string_fervidicoccus_fontis_node_name(node_name) {
            return Ok("STRING Fervidicoccus fontis");
        }

        if is_valid_string_helicobacter_cetorum_mit995656_node_name(node_name) {
            return Ok("STRING Helicobacter cetorum MIT995656");
        }

        if is_valid_string_richelia_intracellularis_node_name(node_name) {
            return Ok("STRING Richelia intracellularis");
        }

        if is_valid_string_methylotenera_sp_73s_node_name(node_name) {
            return Ok("STRING Methylotenera sp. 73s");
        }

        if is_valid_string_sulfurovum_sp_ar_node_name(node_name) {
            return Ok("STRING Sulfurovum sp. AR");
        }

        if is_valid_string_pectobacterium_sp_scc3193_node_name(node_name) {
            return Ok("STRING Pectobacterium sp. SCC3193");
        }

        if is_valid_string_fibrella_aestuarina_node_name(node_name) {
            return Ok("STRING Fibrella aestuarina");
        }

        if is_valid_string_enterobacter_sp_r4368_node_name(node_name) {
            return Ok("STRING Enterobacter sp. R4368");
        }

        if is_valid_string_halomonas_xinjiangensis_node_name(node_name) {
            return Ok("STRING Halomonas xinjiangensis");
        }

        if is_valid_string_desulfocapsa_sulfexigens_node_name(node_name) {
            return Ok("STRING Desulfocapsa sulfexigens");
        }

        if is_valid_string_staphylococcus_vitulinus_node_name(node_name) {
            return Ok("STRING Staphylococcus vitulinus");
        }

        if is_valid_string_draconibacterium_orientale_node_name(node_name) {
            return Ok("STRING Draconibacterium orientale");
        }

        if is_valid_string_xanthobacteraceae_bacterium_501b_node_name(node_name) {
            return Ok("STRING Xanthobacteraceae bacterium 501b");
        }

        if is_valid_string_gammaproteobacteria_bacterium_bdw918_node_name(node_name) {
            return Ok("STRING Gammaproteobacteria bacterium BDW918");
        }

        if is_valid_string_thiomicrospira_kuenenii_node_name(node_name) {
            return Ok("STRING Thiomicrospira kuenenii");
        }

        if is_valid_string_marinilabilia_salmonicolor_node_name(node_name) {
            return Ok("STRING Marinilabilia salmonicolor");
        }

        if is_valid_string_burkholderia_bryophila_node_name(node_name) {
            return Ok("STRING Burkholderia bryophila");
        }

        if is_valid_string_bacillus_sp_123mfchir2_node_name(node_name) {
            return Ok("STRING Bacillus sp. 123MFChir2");
        }

        if is_valid_string_nocardia_sp_cny236_node_name(node_name) {
            return Ok("STRING Nocardia sp. CNY236");
        }

        if is_valid_string_streptomyces_sp_cnt372_node_name(node_name) {
            return Ok("STRING Streptomyces sp. CNT372");
        }

        if is_valid_string_streptomyces_sp_cny243_node_name(node_name) {
            return Ok("STRING Streptomyces sp. CNY243");
        }

        if is_valid_string_propionibacterium_avidum_node_name(node_name) {
            return Ok("STRING Propionibacterium avidum");
        }

        if is_valid_string_calothrix_sp_pcc6303_node_name(node_name) {
            return Ok("STRING Calothrix sp. PCC6303");
        }

        if is_valid_string_propionibacterium_acidipropionici_node_name(node_name) {
            return Ok("STRING Propionibacterium acidipropionici");
        }

        if is_valid_string_streptomyces_sp_142mfcol31_node_name(node_name) {
            return Ok("STRING Streptomyces sp. 142MFCol31");
        }

        if is_valid_string_streptomyces_sp_351mftsu51_node_name(node_name) {
            return Ok("STRING Streptomyces sp. 351MFTsu51");
        }

        if is_valid_string_streptomyces_sp_303mfcol52_node_name(node_name) {
            return Ok("STRING Streptomyces sp. 303MFCol52");
        }

        if is_valid_string_nocardia_sp_348mftsu51_node_name(node_name) {
            return Ok("STRING Nocardia sp. 348MFTsu51");
        }

        if is_valid_string_mycobacterium_sp_360mftsu51_node_name(node_name) {
            return Ok("STRING Mycobacterium sp. 360MFTsu51");
        }

        if is_valid_string_terracoccus_sp_273mftsu31_node_name(node_name) {
            return Ok("STRING Terracoccus sp. 273MFTsu31");
        }

        if is_valid_string_sulfurimonas_sp_ast10_node_name(node_name) {
            return Ok("STRING Sulfurimonas sp. AST10");
        }

        if is_valid_string_helicobacter_cinaedi_pagu611_node_name(node_name) {
            return Ok("STRING Helicobacter cinaedi PAGU611");
        }

        if is_valid_string_chamaesiphon_minutus_node_name(node_name) {
            return Ok("STRING Chamaesiphon minutus");
        }

        if is_valid_string_cyanobacterium_sp_pcc7702_node_name(node_name) {
            return Ok("STRING Cyanobacterium sp. PCC7702");
        }

        if is_valid_string_crinalium_epipsammum_node_name(node_name) {
            return Ok("STRING Crinalium epipsammum");
        }

        if is_valid_string_fischerella_sp_pcc9431_node_name(node_name) {
            return Ok("STRING Fischerella sp. PCC9431");
        }

        if is_valid_string_fischerella_sp_pcc9605_node_name(node_name) {
            return Ok("STRING Fischerella sp. PCC9605");
        }

        if is_valid_string_geitlerinema_sp_pcc7407_node_name(node_name) {
            return Ok("STRING Geitlerinema sp. PCC7407");
        }

        if is_valid_string_gloeocapsa_sp_pcc7428_node_name(node_name) {
            return Ok("STRING Gloeocapsa sp. PCC7428");
        }

        if is_valid_string_microcoleus_sp_pcc7113_node_name(node_name) {
            return Ok("STRING Microcoleus sp. PCC7113");
        }

        if is_valid_string_oscillatoria_sp_pcc10802_node_name(node_name) {
            return Ok("STRING Oscillatoria sp. PCC10802");
        }

        if is_valid_string_spirulina_subsalsa_node_name(node_name) {
            return Ok("STRING Spirulina subsalsa");
        }

        if is_valid_string_synechococcus_sp_pcc7502_node_name(node_name) {
            return Ok("STRING Synechococcus sp. PCC7502");
        }

        if is_valid_string_leptolyngbya_sp_pcc6406_node_name(node_name) {
            return Ok("STRING Leptolyngbya sp. PCC6406");
        }

        if is_valid_string_colletotrichum_sublineola_node_name(node_name) {
            return Ok("STRING Colletotrichum sublineola");
        }

        if is_valid_string_bacillus_psychrosaccharolyticus_node_name(node_name) {
            return Ok("STRING Bacillus psychrosaccharolyticus");
        }

        if is_valid_string_fischerella_sp_pcc9339_node_name(node_name) {
            return Ok("STRING Fischerella sp. PCC9339");
        }

        if is_valid_string_liberibacter_asiaticus_gxpsy_node_name(node_name) {
            return Ok("STRING Liberibacter asiaticus gxpsy");
        }

        if is_valid_string_sphingopyxis_sp_mc1_node_name(node_name) {
            return Ok("STRING Sphingopyxis sp. MC1");
        }

        if is_valid_string_herbaspirillum_sp_gw103_node_name(node_name) {
            return Ok("STRING Herbaspirillum sp. GW103");
        }

        if is_valid_string_aerococcus_viridans_ll1_node_name(node_name) {
            return Ok("STRING Aerococcus viridans LL1");
        }

        if is_valid_string_brevibacterium_massiliense_node_name(node_name) {
            return Ok("STRING Brevibacterium massiliense");
        }

        if is_valid_string_alcanivorax_sp_19m6_node_name(node_name) {
            return Ok("STRING Alcanivorax sp. 19m6");
        }

        if is_valid_string_alcanivorax_hongdengensis_node_name(node_name) {
            return Ok("STRING Alcanivorax hongdengensis");
        }

        if is_valid_string_alcanivorax_jadensis_node_name(node_name) {
            return Ok("STRING Alcanivorax jadensis");
        }

        if is_valid_string_microbacterium_sp_c448_node_name(node_name) {
            return Ok("STRING Microbacterium sp. C448");
        }

        if is_valid_string_thalassospira_profundimaris_node_name(node_name) {
            return Ok("STRING Thalassospira profundimaris");
        }

        if is_valid_string_halomonas_sp_bjgmmb45_node_name(node_name) {
            return Ok("STRING Halomonas sp. BJGMMB45");
        }

        if is_valid_string_bacillus_xiamenensis_node_name(node_name) {
            return Ok("STRING Bacillus xiamenensis");
        }

        if is_valid_string_bacillus_sp_dw54_node_name(node_name) {
            return Ok("STRING Bacillus sp. DW54");
        }

        if is_valid_string_arenitalea_lutea_node_name(node_name) {
            return Ok("STRING Arenitalea lutea");
        }

        if is_valid_string_photodesmus_blepharus_node_name(node_name) {
            return Ok("STRING Photodesmus blepharus");
        }

        if is_valid_string_staphylococcus_lentus_node_name(node_name) {
            return Ok("STRING Staphylococcus lentus");
        }

        if is_valid_string_saccharothrix_espanaensis_node_name(node_name) {
            return Ok("STRING Saccharothrix espanaensis");
        }

        if is_valid_string_mycoplasma_sp_g5847_node_name(node_name) {
            return Ok("STRING Mycoplasma sp. G5847");
        }

        if is_valid_string_pseudomonas_sp_m47t1_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. M47T1");
        }

        if is_valid_string_cladophialophora_psammophila_node_name(node_name) {
            return Ok("STRING Cladophialophora psammophila");
        }

        if is_valid_string_pseudomonas_pseudoalcaligenes_cect5344_node_name(node_name) {
            return Ok("STRING Pseudomonas pseudoalcaligenes CECT5344");
        }

        if is_valid_string_pyrococcus_sp_st04_node_name(node_name) {
            return Ok("STRING Pyrococcus sp. ST04");
        }

        if is_valid_string_gloeobacter_kilaueensis_node_name(node_name) {
            return Ok("STRING Gloeobacter kilaueensis");
        }

        if is_valid_string_thermogladius_cellulolyticus_node_name(node_name) {
            return Ok("STRING Thermogladius cellulolyticus");
        }

        if is_valid_string_bdellovibrio_exovorus_node_name(node_name) {
            return Ok("STRING Bdellovibrio exovorus");
        }

        if is_valid_string_austwickia_chelonae_node_name(node_name) {
            return Ok("STRING Austwickia chelonae");
        }

        if is_valid_string_kineosphaera_limosa_node_name(node_name) {
            return Ok("STRING Kineosphaera limosa");
        }

        if is_valid_string_sinorhizobium_fredii_usda257_node_name(node_name) {
            return Ok("STRING Sinorhizobium fredii USDA257");
        }

        if is_valid_string_planococcus_antarcticus_node_name(node_name) {
            return Ok("STRING Planococcus antarcticus");
        }

        if is_valid_string_thioclava_dalianensis_node_name(node_name) {
            return Ok("STRING Thioclava dalianensis");
        }

        if is_valid_string_fibrisoma_limi_node_name(node_name) {
            return Ok("STRING Fibrisoma limi");
        }

        if is_valid_string_blattabacterium_sp_bgi_node_name(node_name) {
            return Ok("STRING Blattabacterium sp. Bgi");
        }

        if is_valid_string_vibrio_genomosp_node_name(node_name) {
            return Ok("STRING Vibrio genomosp.");
        }

        if is_valid_string_rhodovulum_sp_ph10_node_name(node_name) {
            return Ok("STRING Rhodovulum sp. PH10");
        }

        if is_valid_string_mycoplasma_auris_node_name(node_name) {
            return Ok("STRING Mycoplasma auris");
        }

        if is_valid_string_mycoplasma_alkalescens_node_name(node_name) {
            return Ok("STRING Mycoplasma alkalescens");
        }

        if is_valid_string_mycoplasma_bovigenitalium_node_name(node_name) {
            return Ok("STRING Mycoplasma bovigenitalium");
        }

        if is_valid_string_mycoplasma_arginini_node_name(node_name) {
            return Ok("STRING Mycoplasma arginini");
        }

        if is_valid_string_mycoplasma_ovipneumoniae_node_name(node_name) {
            return Ok("STRING Mycoplasma ovipneumoniae");
        }

        if is_valid_string_mycoplasma_yeatsii_node_name(node_name) {
            return Ok("STRING Mycoplasma yeatsii");
        }

        if is_valid_string_ureaplasma_diversum_node_name(node_name) {
            return Ok("STRING Ureaplasma diversum");
        }

        if is_valid_string_vibrio_rumoiensis_node_name(node_name) {
            return Ok("STRING Vibrio rumoiensis");
        }

        if is_valid_string_rhodovulum_sulfidophilum_node_name(node_name) {
            return Ok("STRING Rhodovulum sulfidophilum");
        }

        if is_valid_string_indibacter_alkaliphilus_node_name(node_name) {
            return Ok("STRING Indibacter alkaliphilus");
        }

        if is_valid_string_psychroflexus_gondwanensis_node_name(node_name) {
            return Ok("STRING Psychroflexus gondwanensis");
        }

        if is_valid_string_flavobacterium_sp_acam123_node_name(node_name) {
            return Ok("STRING Flavobacterium sp. ACAM123");
        }

        if is_valid_string_enterovibrio_norvegicus_node_name(node_name) {
            return Ok("STRING Enterovibrio norvegicus");
        }

        if is_valid_string_enterovibrio_calviensis_node_name(node_name) {
            return Ok("STRING Enterovibrio calviensis");
        }

        if is_valid_string_vibrio_kanaloae_node_name(node_name) {
            return Ok("STRING Vibrio kanaloae");
        }

        if is_valid_string_acinetobacter_venetianus_node_name(node_name) {
            return Ok("STRING Acinetobacter venetianus");
        }

        if is_valid_string_melioribacter_roseus_node_name(node_name) {
            return Ok("STRING Melioribacter roseus");
        }

        if is_valid_string_chondromyces_apiculatus_node_name(node_name) {
            return Ok("STRING Chondromyces apiculatus");
        }

        if is_valid_string_burkholderia_sp_lig30_node_name(node_name) {
            return Ok("STRING Burkholderia sp. lig30");
        }

        if is_valid_string_sphingobium_xenophagum_node_name(node_name) {
            return Ok("STRING Sphingobium xenophagum");
        }

        if is_valid_string_pseudaminobacter_salicylatoxidans_node_name(node_name) {
            return Ok("STRING Pseudaminobacter salicylatoxidans");
        }

        if is_valid_string_parascardovia_denticolens_ipla20019_node_name(node_name) {
            return Ok("STRING Parascardovia denticolens IPLA20019");
        }

        if is_valid_string_tetrasphaera_elongata_node_name(node_name) {
            return Ok("STRING Tetrasphaera elongata");
        }

        if is_valid_string_endolissoclinum_faulkneri_l2_node_name(node_name) {
            return Ok("STRING Endolissoclinum faulkneri L2");
        }

        if is_valid_string_microbacterium_yannicii_node_name(node_name) {
            return Ok("STRING Microbacterium yannicii");
        }

        if is_valid_string_staphylococcus_warneri_sg1_node_name(node_name) {
            return Ok("STRING Staphylococcus warneri SG1");
        }

        if is_valid_string_mycobacterium_vaccae_node_name(node_name) {
            return Ok("STRING Mycobacterium vaccae");
        }

        if is_valid_string_clostridium_termitidis_node_name(node_name) {
            return Ok("STRING Clostridium termitidis");
        }

        if is_valid_string_alishewanella_agri_node_name(node_name) {
            return Ok("STRING Alishewanella agri");
        }

        if is_valid_string_virgibacillus_halodenitrificans_node_name(node_name) {
            return Ok("STRING Virgibacillus halodenitrificans");
        }

        if is_valid_string_bacillus_endophyticus_node_name(node_name) {
            return Ok("STRING Bacillus endophyticus");
        }

        if is_valid_string_bacillus_oceanisediminis_node_name(node_name) {
            return Ok("STRING Bacillus oceanisediminis");
        }

        if is_valid_string_snodgrassella_alvi_node_name(node_name) {
            return Ok("STRING Snodgrassella alvi");
        }

        if is_valid_string_gilliamella_apicola_node_name(node_name) {
            return Ok("STRING Gilliamella apicola");
        }

        if is_valid_string_clostridium_sp_maddingley_node_name(node_name) {
            return Ok("STRING Clostridium sp. Maddingley");
        }

        if is_valid_string_paenibacillus_sp_osyse_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. OSYSE");
        }

        if is_valid_string_bacillus_macauensis_node_name(node_name) {
            return Ok("STRING Bacillus macauensis");
        }

        if is_valid_string_pseudomonas_stutzeri_ccug_29243_node_name(node_name) {
            return Ok("STRING Pseudomonas stutzeri CCUG 29243");
        }

        if is_valid_string_halococcus_sp_197a_node_name(node_name) {
            return Ok("STRING Halococcus sp. 197A");
        }

        if is_valid_string_mangrovimonas_yunxiaonensis_node_name(node_name) {
            return Ok("STRING Mangrovimonas yunxiaonensis");
        }

        if is_valid_string_arthrobacter_sp_m2012083_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. M2012083");
        }

        if is_valid_string_salmonella_bongori_n26808_node_name(node_name) {
            return Ok("STRING Salmonella bongori N26808");
        }

        if is_valid_string_afipia_birgiae_node_name(node_name) {
            return Ok("STRING Afipia birgiae");
        }

        if is_valid_string_granulicella_tundricola_node_name(node_name) {
            return Ok("STRING Granulicella tundricola");
        }

        if is_valid_string_cycloclasticus_zancles_node_name(node_name) {
            return Ok("STRING Cycloclasticus zancles");
        }

        if is_valid_string_aeropyrum_camini_node_name(node_name) {
            return Ok("STRING Aeropyrum camini");
        }

        if is_valid_string_janthinobacterium_sp_hh01_node_name(node_name) {
            return Ok("STRING Janthinobacterium sp. HH01");
        }

        if is_valid_string_streptococcus_mutans_gs5_node_name(node_name) {
            return Ok("STRING Streptococcus mutans GS5");
        }

        if is_valid_string_corynebacterium_terpenotabidum_node_name(node_name) {
            return Ok("STRING Corynebacterium terpenotabidum");
        }

        if is_valid_string_anaerovibrio_sp_rm50_node_name(node_name) {
            return Ok("STRING Anaerovibrio sp. RM50");
        }

        if is_valid_string_ruminobacter_sp_rm87_node_name(node_name) {
            return Ok("STRING Ruminobacter sp. RM87");
        }

        if is_valid_string_brevibacillus_brevis_x23_node_name(node_name) {
            return Ok("STRING Brevibacillus brevis X23");
        }

        if is_valid_string_bartonella_birtlesii_node_name(node_name) {
            return Ok("STRING Bartonella birtlesii");
        }

        if is_valid_string_bacteriovorax_sp_seq25v_node_name(node_name) {
            return Ok("STRING Bacteriovorax sp. Seq25V");
        }

        if is_valid_string_bacteriovorax_sp_bal6x_node_name(node_name) {
            return Ok("STRING Bacteriovorax sp. BAL6X");
        }

        if is_valid_string_enterococcus_faecalis_atcc29212_node_name(node_name) {
            return Ok("STRING Enterococcus faecalis ATCC29212");
        }

        if is_valid_string_moritella_dasanensis_node_name(node_name) {
            return Ok("STRING Moritella dasanensis");
        }

        if is_valid_string_flavobacterium_sp_f52_node_name(node_name) {
            return Ok("STRING Flavobacterium sp. F52");
        }

        if is_valid_string_halopiger_salifodinae_node_name(node_name) {
            return Ok("STRING Halopiger salifodinae");
        }

        if is_valid_string_moritella_marina_node_name(node_name) {
            return Ok("STRING Moritella marina");
        }

        if is_valid_string_lactobacillus_ingluviei_node_name(node_name) {
            return Ok("STRING Lactobacillus ingluviei");
        }

        if is_valid_string_corynebacterium_timonense_node_name(node_name) {
            return Ok("STRING Corynebacterium timonense");
        }

        if is_valid_string_prevotella_oralis_hga0225_node_name(node_name) {
            return Ok("STRING Prevotella oralis HGA0225");
        }

        if is_valid_string_sutterella_wadsworthensis_hga0223_node_name(node_name) {
            return Ok("STRING Sutterella wadsworthensis HGA0223");
        }

        if is_valid_string_actinomyces_sp_hpa0247_node_name(node_name) {
            return Ok("STRING Actinomyces sp. HPA0247");
        }

        if is_valid_string_corynebacterium_sp_kpl1859_node_name(node_name) {
            return Ok("STRING Corynebacterium sp. KPL1859");
        }

        if is_valid_string_corynebacterium_sp_kpl1860_node_name(node_name) {
            return Ok("STRING Corynebacterium sp. KPL1860");
        }

        if is_valid_string_dermabacter_sp_hfh0086_node_name(node_name) {
            return Ok("STRING Dermabacter sp. HFH0086");
        }

        if is_valid_string_streptococcus_sp_hph0090_node_name(node_name) {
            return Ok("STRING Streptococcus sp. HPH0090");
        }

        if is_valid_string_atopobium_sp_f0494_node_name(node_name) {
            return Ok("STRING Atopobium sp. F0494");
        }

        if is_valid_string_propionibacterium_sp_f0372_node_name(node_name) {
            return Ok("STRING Propionibacterium sp. F0372");
        }

        if is_valid_string_butyricicoccus_pullicaecorum_node_name(node_name) {
            return Ok("STRING Butyricicoccus pullicaecorum");
        }

        if is_valid_string_alistipes_onderdonkii_node_name(node_name) {
            return Ok("STRING Alistipes onderdonkii");
        }

        if is_valid_string_corynebacterium_sp_kpl1989_node_name(node_name) {
            return Ok("STRING Corynebacterium sp. KPL1989");
        }

        if is_valid_string_corynebacterium_sp_kpl2004_node_name(node_name) {
            return Ok("STRING Corynebacterium sp. KPL2004");
        }

        if is_valid_string_reyranella_massiliensis_node_name(node_name) {
            return Ok("STRING Reyranella massiliensis");
        }

        if is_valid_string_bartonella_rattaustraliani_node_name(node_name) {
            return Ok("STRING Bartonella rattaustraliani");
        }

        if is_valid_string_yersinia_massiliensis_node_name(node_name) {
            return Ok("STRING Yersinia massiliensis");
        }

        if is_valid_string_xanthomonas_translucens_node_name(node_name) {
            return Ok("STRING Xanthomonas translucens");
        }

        if is_valid_string_vibrio_breoganii_node_name(node_name) {
            return Ok("STRING Vibrio breoganii");
        }

        if is_valid_string_nocardiopsis_alba_node_name(node_name) {
            return Ok("STRING Nocardiopsis alba");
        }

        if is_valid_string_streptomyces_sp_cnr698_node_name(node_name) {
            return Ok("STRING Streptomyces sp. CNR698");
        }

        if is_valid_string_nocardia_aobensis_node_name(node_name) {
            return Ok("STRING Nocardia aobensis");
        }

        if is_valid_string_nocardia_brevicatena_node_name(node_name) {
            return Ok("STRING Nocardia brevicatena");
        }

        if is_valid_string_nocardia_carnea_node_name(node_name) {
            return Ok("STRING Nocardia carnea");
        }

        if is_valid_string_nocardia_exalbida_node_name(node_name) {
            return Ok("STRING Nocardia exalbida");
        }

        if is_valid_string_nocardia_higoensis_node_name(node_name) {
            return Ok("STRING Nocardia higoensis");
        }

        if is_valid_string_nocardia_jiangxiensis_node_name(node_name) {
            return Ok("STRING Nocardia jiangxiensis");
        }

        if is_valid_string_nocardia_otitidiscaviarum_node_name(node_name) {
            return Ok("STRING Nocardia otitidiscaviarum");
        }

        if is_valid_string_nocardia_niigatensis_node_name(node_name) {
            return Ok("STRING Nocardia niigatensis");
        }

        if is_valid_string_nocardia_takedensis_node_name(node_name) {
            return Ok("STRING Nocardia takedensis");
        }

        if is_valid_string_nocardia_pneumoniae_node_name(node_name) {
            return Ok("STRING Nocardia pneumoniae");
        }

        if is_valid_string_nocardia_testacea_node_name(node_name) {
            return Ok("STRING Nocardia testacea");
        }

        if is_valid_string_nocardia_concava_node_name(node_name) {
            return Ok("STRING Nocardia concava");
        }

        if is_valid_string_nocardia_veterana_node_name(node_name) {
            return Ok("STRING Nocardia veterana");
        }

        if is_valid_string_nocardia_transvalensis_node_name(node_name) {
            return Ok("STRING Nocardia transvalensis");
        }

        if is_valid_string_pseudomonas_sp_lz4w_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. Lz4W");
        }

        if is_valid_string_sphingobium_sp_c100_node_name(node_name) {
            return Ok("STRING Sphingobium sp. C100");
        }

        if is_valid_string_hyphomonas_sp_l53140_node_name(node_name) {
            return Ok("STRING Hyphomonas sp. L53140");
        }

        if is_valid_string_oceanibaculum_indicum_node_name(node_name) {
            return Ok("STRING Oceanibaculum indicum");
        }

        if is_valid_string_pseudomonas_sp_uw4_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. UW4");
        }

        if is_valid_string_pseudomonas_luteola_node_name(node_name) {
            return Ok("STRING Pseudomonas luteola");
        }

        if is_valid_string_marinomonas_sp_d104_node_name(node_name) {
            return Ok("STRING Marinomonas sp. D104");
        }

        if is_valid_string_celeribacter_baekdonensis_node_name(node_name) {
            return Ok("STRING Celeribacter baekdonensis");
        }

        if is_valid_string_commensalibacter_sp_mx01_node_name(node_name) {
            return Ok("STRING Commensalibacter sp. MX01");
        }

        if is_valid_string_kinetoplastibacterium_crithidii_node_name(node_name) {
            return Ok("STRING Kinetoplastibacterium crithidii");
        }

        if is_valid_string_kinetoplastibacterium_desouzaii_node_name(node_name) {
            return Ok("STRING Kinetoplastibacterium desouzaii");
        }

        if is_valid_string_kinetoplastibacterium_oncopeltii_node_name(node_name) {
            return Ok("STRING Kinetoplastibacterium oncopeltii");
        }

        if is_valid_string_kinetoplastibacterium_galatii_node_name(node_name) {
            return Ok("STRING Kinetoplastibacterium galatii");
        }

        if is_valid_string_kinetoplastibacterium_blastocrithidii_node_name(node_name) {
            return Ok("STRING Kinetoplastibacterium blastocrithidii");
        }

        if is_valid_string_cellvibrio_mixtus_node_name(node_name) {
            return Ok("STRING Cellvibrio mixtus");
        }

        if is_valid_string_mycobacterium_mageritense_node_name(node_name) {
            return Ok("STRING Mycobacterium mageritense");
        }

        if is_valid_string_tepidanaerobacter_acetatoxydans_node_name(node_name) {
            return Ok("STRING Tepidanaerobacter acetatoxydans");
        }

        if is_valid_string_streptomyces_sp_aa0539_node_name(node_name) {
            return Ok("STRING Streptomyces sp. AA0539");
        }

        if is_valid_string_janibacter_hoylei_node_name(node_name) {
            return Ok("STRING Janibacter hoylei");
        }

        if is_valid_string_gemmata_sp_iil30_node_name(node_name) {
            return Ok("STRING Gemmata sp. IIL30");
        }

        if is_valid_string_halogranum_salarium_node_name(node_name) {
            return Ok("STRING Halogranum salarium");
        }

        if is_valid_string_lysinibacillus_massiliensis_node_name(node_name) {
            return Ok("STRING Lysinibacillus massiliensis");
        }

        if is_valid_string_pseudomonas_psychrophila_node_name(node_name) {
            return Ok("STRING Pseudomonas psychrophila");
        }

        if is_valid_string_pseudoxanthomonas_sp_gw2_node_name(node_name) {
            return Ok("STRING Pseudoxanthomonas sp. GW2");
        }

        if is_valid_string_chelatococcus_sp_gw1_node_name(node_name) {
            return Ok("STRING Chelatococcus sp. GW1");
        }

        if is_valid_string_pseudomonas_putida_nbrc14164_node_name(node_name) {
            return Ok("STRING Pseudomonas putida NBRC14164");
        }

        if is_valid_string_rhizobium_mesoamericanum_node_name(node_name) {
            return Ok("STRING Rhizobium mesoamericanum");
        }

        if is_valid_string_alistipes_marseilloanorexicus_node_name(node_name) {
            return Ok("STRING Alistipes marseilloanorexicus");
        }

        if is_valid_string_bacillus_massilioanorexius_node_name(node_name) {
            return Ok("STRING Bacillus massilioanorexius");
        }

        if is_valid_string_blastococcus_massiliensis_node_name(node_name) {
            return Ok("STRING Blastococcus massiliensis");
        }

        if is_valid_string_clostridium_anorexicamassiliense_node_name(node_name) {
            return Ok("STRING Clostridium anorexicamassiliense");
        }

        if is_valid_string_holdemania_massiliensis_node_name(node_name) {
            return Ok("STRING Holdemania massiliensis");
        }

        if is_valid_string_stoquefichus_massiliensis_node_name(node_name) {
            return Ok("STRING Stoquefichus massiliensis");
        }

        if is_valid_string_pseudomonas_stutzeri_nf13_node_name(node_name) {
            return Ok("STRING Pseudomonas stutzeri NF13");
        }

        if is_valid_string_enterobacteriaceae_bacterium_b14_node_name(node_name) {
            return Ok("STRING Enterobacteriaceae bacterium B14");
        }

        if is_valid_string_streptomyces_davawensis_node_name(node_name) {
            return Ok("STRING Streptomyces davawensis");
        }

        if is_valid_string_streptococcus_suis_865192_node_name(node_name) {
            return Ok("STRING Streptococcus suis 865192");
        }

        if is_valid_string_streptococcus_suis_22083_node_name(node_name) {
            return Ok("STRING Streptococcus suis 22083");
        }

        if is_valid_string_streptococcus_suis_ys72_node_name(node_name) {
            return Ok("STRING Streptococcus suis YS72");
        }

        if is_valid_string_streptomyces_collinus_node_name(node_name) {
            return Ok("STRING Streptomyces collinus");
        }

        if is_valid_string_pseudomonas_alcaligenes_nbrc14159_node_name(node_name) {
            return Ok("STRING Pseudomonas alcaligenes NBRC14159");
        }

        if is_valid_string_pseudomonas_parafulva_node_name(node_name) {
            return Ok("STRING Pseudomonas parafulva");
        }

        if is_valid_string_liberibacter_crescens_node_name(node_name) {
            return Ok("STRING Liberibacter crescens");
        }

        if is_valid_string_lactococcus_raffinolactis_node_name(node_name) {
            return Ok("STRING Lactococcus raffinolactis");
        }

        if is_valid_string_pseudoalteromonas_ruthenica_node_name(node_name) {
            return Ok("STRING Pseudoalteromonas ruthenica");
        }

        if is_valid_string_fusobacterium_hwasookii_node_name(node_name) {
            return Ok("STRING Fusobacterium hwasookii");
        }

        if is_valid_string_clostridium_sp_m240_node_name(node_name) {
            return Ok("STRING Clostridium sp. M240");
        }

        if is_valid_string_edwardsiella_hoshinae_node_name(node_name) {
            return Ok("STRING Edwardsiella hoshinae");
        }

        if is_valid_string_elizabethkingia_meningoseptica_node_name(node_name) {
            return Ok("STRING Elizabethkingia meningoseptica");
        }

        if is_valid_string_achromobacter_xylosoxidans_nbrc15126_node_name(node_name) {
            return Ok("STRING Achromobacter xylosoxidans NBRC15126");
        }

        if is_valid_string_acinetobacter_beijerinckii_node_name(node_name) {
            return Ok("STRING Acinetobacter beijerinckii");
        }

        if is_valid_string_acinetobacter_brisouii_node_name(node_name) {
            return Ok("STRING Acinetobacter brisouii");
        }

        if is_valid_string_acinetobacter_guillouiae_node_name(node_name) {
            return Ok("STRING Acinetobacter guillouiae");
        }

        if is_valid_string_acinetobacter_gyllenbergii_node_name(node_name) {
            return Ok("STRING Acinetobacter gyllenbergii");
        }

        if is_valid_string_acinetobacter_sp_anc4105_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. ANC4105");
        }

        if is_valid_string_acinetobacter_sp_anc3862_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. ANC3862");
        }

        if is_valid_string_acinetobacter_sp_niph2100_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. NIPH2100");
        }

        if is_valid_string_acinetobacter_sp_niph899_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. NIPH899");
        }

        if is_valid_string_acinetobacter_sp_niph758_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. NIPH758");
        }

        if is_valid_string_acinetobacter_sp_niph809_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. NIPH809");
        }

        if is_valid_string_acinetobacter_sp_anc3789_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. ANC3789");
        }

        if is_valid_string_acinetobacter_bohemicus_node_name(node_name) {
            return Ok("STRING Acinetobacter bohemicus");
        }

        if is_valid_string_cupriavidus_sp_bis7_node_name(node_name) {
            return Ok("STRING Cupriavidus sp. BIS7");
        }

        if is_valid_string_roseomonas_sp_b5_node_name(node_name) {
            return Ok("STRING Roseomonas sp. B5");
        }

        if is_valid_string_burkholderia_acidipaludis_node_name(node_name) {
            return Ok("STRING Burkholderia acidipaludis");
        }

        if is_valid_string_burkholderia_bannensis_node_name(node_name) {
            return Ok("STRING Burkholderia bannensis");
        }

        if is_valid_string_burkholderia_ferrariae_node_name(node_name) {
            return Ok("STRING Burkholderia ferrariae");
        }

        if is_valid_string_burkholderia_terrae_node_name(node_name) {
            return Ok("STRING Burkholderia terrae");
        }

        if is_valid_string_citrobacter_sedlakii_node_name(node_name) {
            return Ok("STRING Citrobacter sedlakii");
        }

        if is_valid_string_chryseobacterium_indologenes_node_name(node_name) {
            return Ok("STRING Chryseobacterium indologenes");
        }

        if is_valid_string_empedobacter_brevis_node_name(node_name) {
            return Ok("STRING Empedobacter brevis");
        }

        if is_valid_string_bacillus_alcalophilus_node_name(node_name) {
            return Ok("STRING Bacillus alcalophilus");
        }

        if is_valid_string_pseudomonas_stutzeri_kos6_node_name(node_name) {
            return Ok("STRING Pseudomonas stutzeri KOS6");
        }

        if is_valid_string_comamonas_aquatica_node_name(node_name) {
            return Ok("STRING Comamonas aquatica");
        }

        if is_valid_string_novosphingobium_tardaugens_node_name(node_name) {
            return Ok("STRING Novosphingobium tardaugens");
        }

        if is_valid_string_sphingobium_herbicidovorans_node_name(node_name) {
            return Ok("STRING Sphingobium herbicidovorans");
        }

        if is_valid_string_sphingomonas_parapaucimobilis_node_name(node_name) {
            return Ok("STRING Sphingomonas parapaucimobilis");
        }

        if is_valid_string_vibrio_proteolyticus_node_name(node_name) {
            return Ok("STRING Vibrio proteolyticus");
        }

        if is_valid_string_vibrio_halioticoli_node_name(node_name) {
            return Ok("STRING Vibrio halioticoli");
        }

        if is_valid_string_vibrio_alginolyticus_nbrc15630_node_name(node_name) {
            return Ok("STRING Vibrio alginolyticus NBRC15630");
        }

        if is_valid_string_vibrio_azureus_node_name(node_name) {
            return Ok("STRING Vibrio azureus");
        }

        if is_valid_string_vibrio_ezurae_node_name(node_name) {
            return Ok("STRING Vibrio ezurae");
        }

        if is_valid_string_thermotoga_hypogea_node_name(node_name) {
            return Ok("STRING Thermotoga hypogea");
        }

        if is_valid_string_xanthomonas_cassavae_node_name(node_name) {
            return Ok("STRING Xanthomonas cassavae");
        }

        if is_valid_string_actinomyces_sp_s4c9_node_name(node_name) {
            return Ok("STRING Actinomyces sp. S4C9");
        }

        if is_valid_string_arcanobacterium_sp_s3pf19_node_name(node_name) {
            return Ok("STRING Arcanobacterium sp. S3PF19");
        }

        if is_valid_string_peptostreptococcus_sp_mv1_node_name(node_name) {
            return Ok("STRING Peptostreptococcus sp. MV1");
        }

        if is_valid_string_methanobacterium_sp_maddingley_node_name(node_name) {
            return Ok("STRING Methanobacterium sp. Maddingley");
        }

        if is_valid_string_alpha_proteobacterium_imcc14465_node_name(node_name) {
            return Ok("STRING alpha proteobacterium IMCC14465");
        }

        if is_valid_string_staphylococcus_chromogenes_node_name(node_name) {
            return Ok("STRING Staphylococcus chromogenes");
        }

        if is_valid_string_rhizobium_rubi_node_name(node_name) {
            return Ok("STRING Rhizobium rubi");
        }

        if is_valid_string_gordonia_aichiensis_node_name(node_name) {
            return Ok("STRING Gordonia aichiensis");
        }

        if is_valid_string_lysinibacillus_odysseyi_node_name(node_name) {
            return Ok("STRING Lysinibacillus odysseyi");
        }

        if is_valid_string_pseudomonas_fluorescens_ncimb11764_node_name(node_name) {
            return Ok("STRING Pseudomonas fluorescens NCIMB11764");
        }

        if is_valid_string_lactobacillus_florum_node_name(node_name) {
            return Ok("STRING Lactobacillus florum");
        }

        if is_valid_string_altibacter_lentus_node_name(node_name) {
            return Ok("STRING Altibacter lentus");
        }

        if is_valid_string_comamonas_granuli_node_name(node_name) {
            return Ok("STRING Comamonas granuli");
        }

        if is_valid_string_streptomyces_mobaraensis_node_name(node_name) {
            return Ok("STRING Streptomyces mobaraensis");
        }

        if is_valid_string_gordonia_malaquae_node_name(node_name) {
            return Ok("STRING Gordonia malaquae");
        }

        if is_valid_string_gordonia_paraffinivorans_node_name(node_name) {
            return Ok("STRING Gordonia paraffinivorans");
        }

        if is_valid_string_gordonia_sihwensis_node_name(node_name) {
            return Ok("STRING Gordonia sihwensis");
        }

        if is_valid_string_gordonia_soli_node_name(node_name) {
            return Ok("STRING Gordonia soli");
        }

        if is_valid_string_enterobacteriaceae_bacterium_lsjc7_node_name(node_name) {
            return Ok("STRING Enterobacteriaceae bacterium LSJC7");
        }

        if is_valid_string_corynebacterium_maris_node_name(node_name) {
            return Ok("STRING Corynebacterium maris");
        }

        if is_valid_string_corynebacterium_vitaeruminis_node_name(node_name) {
            return Ok("STRING Corynebacterium vitaeruminis");
        }

        if is_valid_string_mangrovibacter_sp_mfb070_node_name(node_name) {
            return Ok("STRING Mangrovibacter sp. MFB070");
        }

        if is_valid_string_gluconobacter_oxydans_h24_node_name(node_name) {
            return Ok("STRING Gluconobacter oxydans H24");
        }

        if is_valid_string_pantoea_sp_a4_node_name(node_name) {
            return Ok("STRING Pantoea sp. A4");
        }

        if is_valid_string_dickeya_sp_dw_0440_node_name(node_name) {
            return Ok("STRING Dickeya sp. DW 0440");
        }

        if is_valid_string_oscillibacter_sp_kle1728_node_name(node_name) {
            return Ok("STRING Oscillibacter sp. KLE1728");
        }

        if is_valid_string_clostridium_sp_kle1755_node_name(node_name) {
            return Ok("STRING Clostridium sp. KLE1755");
        }

        if is_valid_string_pseudomonas_nitroreducens_tx1_node_name(node_name) {
            return Ok("STRING Pseudomonas nitroreducens TX1");
        }

        if is_valid_string_actinobaculum_sp_f0552_node_name(node_name) {
            return Ok("STRING Actinobaculum sp. F0552");
        }

        if is_valid_string_capnocytophaga_sp_f0517_node_name(node_name) {
            return Ok("STRING Capnocytophaga sp. F0517");
        }

        if is_valid_string_leptotrichia_sp_f0557_node_name(node_name) {
            return Ok("STRING Leptotrichia sp. F0557");
        }

        if is_valid_string_porphyromonas_sp_w7784_node_name(node_name) {
            return Ok("STRING Porphyromonas sp. W7784");
        }

        if is_valid_string_prevotella_sp_f0091_node_name(node_name) {
            return Ok("STRING Prevotella sp. F0091");
        }

        if is_valid_string_paenibacillus_sp_fslh7689_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. FSLH7689");
        }

        if is_valid_string_paenibacillus_sp_fslr7277_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. FSLR7277");
        }

        if is_valid_string_viridibacillus_arenosi_node_name(node_name) {
            return Ok("STRING Viridibacillus arenosi");
        }

        if is_valid_string_haloarcula_japonica_node_name(node_name) {
            return Ok("STRING Haloarcula japonica");
        }

        if is_valid_string_halobiforma_nitratireducens_node_name(node_name) {
            return Ok("STRING Halobiforma nitratireducens");
        }

        if is_valid_string_halococcus_thailandensis_node_name(node_name) {
            return Ok("STRING Halococcus thailandensis");
        }

        if is_valid_string_halorubrum_saccharovorum_node_name(node_name) {
            return Ok("STRING Halorubrum saccharovorum");
        }

        if is_valid_string_halosarcina_pallida_node_name(node_name) {
            return Ok("STRING Halosarcina pallida");
        }

        if is_valid_string_haloterrigena_salina_node_name(node_name) {
            return Ok("STRING Haloterrigena salina");
        }

        if is_valid_string_natrinema_pallidum_node_name(node_name) {
            return Ok("STRING Natrinema pallidum");
        }

        if is_valid_string_natronococcus_amylolyticus_node_name(node_name) {
            return Ok("STRING Natronococcus amylolyticus");
        }

        if is_valid_string_natronolimnobius_innermongolicus_node_name(node_name) {
            return Ok("STRING Natronolimnobius innermongolicus");
        }

        if is_valid_string_natronorubrum_bangense_node_name(node_name) {
            return Ok("STRING Natronorubrum bangense");
        }

        if is_valid_string_hymenobacter_swuensis_node_name(node_name) {
            return Ok("STRING Hymenobacter swuensis");
        }

        if is_valid_string_leptolyngbya_sp_kiost1_node_name(node_name) {
            return Ok("STRING Leptolyngbya sp. KIOST1");
        }

        if is_valid_string_actinobacterium_llx17_node_name(node_name) {
            return Ok("STRING actinobacterium LLX17");
        }

        if is_valid_string_alpha_proteobacterium_l41a_node_name(node_name) {
            return Ok("STRING alpha proteobacterium L41A");
        }

        if is_valid_string_burkholderia_phenoliruptrix_node_name(node_name) {
            return Ok("STRING Burkholderia phenoliruptrix");
        }

        if is_valid_string_sphingobacterium_sp_accc05744_node_name(node_name) {
            return Ok("STRING Sphingobacterium sp. ACCC05744");
        }

        if is_valid_string_gammaproteobacteria_bacterium_wg36_node_name(node_name) {
            return Ok("STRING Gammaproteobacteria bacterium WG36");
        }

        if is_valid_string_flavobacterium_sp_wg21_node_name(node_name) {
            return Ok("STRING Flavobacterium sp. WG21");
        }

        if is_valid_string_blattabacterium_sp_bpaa_node_name(node_name) {
            return Ok("STRING Blattabacterium sp. BPAA");
        }

        if is_valid_string_lactococcus_garvieae_i113_node_name(node_name) {
            return Ok("STRING Lactococcus garvieae I113");
        }

        if is_valid_string_alkalibacterium_sp_ak22_node_name(node_name) {
            return Ok("STRING Alkalibacterium sp. AK22");
        }

        if is_valid_string_leuconostoc_gelidum_jb7_node_name(node_name) {
            return Ok("STRING Leuconostoc gelidum JB7");
        }

        if is_valid_string_leuconostoc_carnosum_node_name(node_name) {
            return Ok("STRING Leuconostoc carnosum");
        }

        if is_valid_string_microthrix_parvicella_node_name(node_name) {
            return Ok("STRING Microthrix parvicella");
        }

        if is_valid_string_brevibacterium_casei_node_name(node_name) {
            return Ok("STRING Brevibacterium casei");
        }

        if is_valid_string_staphylococcus_massiliensis_node_name(node_name) {
            return Ok("STRING Staphylococcus massiliensis");
        }

        if is_valid_string_chlamydia_avium_node_name(node_name) {
            return Ok("STRING Chlamydia avium");
        }

        if is_valid_string_nitrosopumilus_sp_ar2_node_name(node_name) {
            return Ok("STRING Nitrosopumilus sp. AR2");
        }

        if is_valid_string_moraxella_macacae_node_name(node_name) {
            return Ok("STRING Moraxella macacae");
        }

        if is_valid_string_salimicrobium_sp_mj3_node_name(node_name) {
            return Ok("STRING Salimicrobium sp. MJ3");
        }

        if is_valid_string_clostridium_tetanomorphum_node_name(node_name) {
            return Ok("STRING Clostridium tetanomorphum");
        }

        if is_valid_string_legionella_anisa_node_name(node_name) {
            return Ok("STRING Legionella anisa");
        }

        if is_valid_string_haloterrigena_limicola_node_name(node_name) {
            return Ok("STRING Haloterrigena limicola");
        }

        if is_valid_string_natronorubrum_sulfidifaciens_node_name(node_name) {
            return Ok("STRING Natronorubrum sulfidifaciens");
        }

        if is_valid_string_bradyrhizobium_sp_dfci1_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. DFCI1");
        }

        if is_valid_string_paenisporosarcina_sp_tg14_node_name(node_name) {
            return Ok("STRING Paenisporosarcina sp. TG14");
        }

        if is_valid_string_nitratireductor_aquibiodomus_node_name(node_name) {
            return Ok("STRING Nitratireductor aquibiodomus");
        }

        if is_valid_string_nitratireductor_indicus_node_name(node_name) {
            return Ok("STRING Nitratireductor indicus");
        }

        if is_valid_string_thermotoga_sp_mc24_node_name(node_name) {
            return Ok("STRING Thermotoga sp. Mc24");
        }

        if is_valid_string_lactobacillus_shenzhenensis_node_name(node_name) {
            return Ok("STRING Lactobacillus shenzhenensis");
        }

        if is_valid_string_lactococcus_garvieae_dcc43_node_name(node_name) {
            return Ok("STRING Lactococcus garvieae DCC43");
        }

        if is_valid_string_pusillimonas_noertemannii_node_name(node_name) {
            return Ok("STRING Pusillimonas noertemannii");
        }

        if is_valid_string_oceaniovalibus_guishaninsula_node_name(node_name) {
            return Ok("STRING Oceaniovalibus guishaninsula");
        }

        if is_valid_string_cardinium_endosymbiont_ceper1_node_name(node_name) {
            return Ok("STRING Cardinium endosymbiont cEper1");
        }

        if is_valid_string_desulfuromonas_sp_tf_node_name(node_name) {
            return Ok("STRING Desulfuromonas sp. TF");
        }

        if is_valid_string_corynebacterium_sp_gd7_node_name(node_name) {
            return Ok("STRING Corynebacterium sp. GD7");
        }

        if is_valid_string_megasphaera_sp_np3_node_name(node_name) {
            return Ok("STRING Megasphaera sp. NP3");
        }

        if is_valid_string_nesterenkonia_sp_np1_node_name(node_name) {
            return Ok("STRING Nesterenkonia sp. NP1");
        }

        if is_valid_string_nosocomiicoccus_sp_np2_node_name(node_name) {
            return Ok("STRING Nosocomiicoccus sp. NP2");
        }

        if is_valid_string_coriobacteriaceae_bacterium_gd5_node_name(node_name) {
            return Ok("STRING Coriobacteriaceae bacterium GD5");
        }

        if is_valid_string_desulfobacula_sp_ts_node_name(node_name) {
            return Ok("STRING Desulfobacula sp. TS");
        }

        if is_valid_string_clostridiales_bacterium_ve20213_node_name(node_name) {
            return Ok("STRING Clostridiales bacterium VE20213");
        }

        if is_valid_string_clostridiales_bacterium_ve20218_node_name(node_name) {
            return Ok("STRING Clostridiales bacterium VE20218");
        }

        if is_valid_string_clostridiales_bacterium_ve20209_node_name(node_name) {
            return Ok("STRING Clostridiales bacterium VE20209");
        }

        if is_valid_string_clostridiales_bacterium_ve20208_node_name(node_name) {
            return Ok("STRING Clostridiales bacterium VE20208");
        }

        if is_valid_string_clostridiales_bacterium_ve20214_node_name(node_name) {
            return Ok("STRING Clostridiales bacterium VE20214");
        }

        if is_valid_string_clostridiales_bacterium_ve20221_node_name(node_name) {
            return Ok("STRING Clostridiales bacterium VE20221");
        }

        if is_valid_string_staphylococcus_sciuri_node_name(node_name) {
            return Ok("STRING Staphylococcus sciuri");
        }

        if is_valid_string_marinobacterium_sp_ak27_node_name(node_name) {
            return Ok("STRING Marinobacterium sp. AK27");
        }

        if is_valid_string_chryseobacterium_sp_jm1_node_name(node_name) {
            return Ok("STRING Chryseobacterium sp. JM1");
        }

        if is_valid_string_epilithonimonas_sp_fh1_node_name(node_name) {
            return Ok("STRING Epilithonimonas sp. FH1");
        }

        if is_valid_string_dyella_ginsengisoli_node_name(node_name) {
            return Ok("STRING Dyella ginsengisoli");
        }

        if is_valid_string_catellicoccus_marimammalium_node_name(node_name) {
            return Ok("STRING Catellicoccus marimammalium");
        }

        if is_valid_string_staphylococcus_sp_e463_node_name(node_name) {
            return Ok("STRING Staphylococcus sp. E463");
        }

        if is_valid_string_alpha_proteobacterium_jlt2015_node_name(node_name) {
            return Ok("STRING alpha proteobacterium JLT2015");
        }

        if is_valid_string_geobacillus_caldoxylosilyticus_node_name(node_name) {
            return Ok("STRING Geobacillus caldoxylosilyticus");
        }

        if is_valid_string_carnobacterium_maltaromaticum_node_name(node_name) {
            return Ok("STRING Carnobacterium maltaromaticum");
        }

        if is_valid_string_bhargavaea_cecembensis_node_name(node_name) {
            return Ok("STRING Bhargavaea cecembensis");
        }

        if is_valid_string_ralstonia_sp_au1208_node_name(node_name) {
            return Ok("STRING Ralstonia sp. AU1208");
        }

        if is_valid_string_salinicoccus_carnicancri_node_name(node_name) {
            return Ok("STRING Salinicoccus carnicancri");
        }

        if is_valid_string_bacteroides_massiliensis_dnlkv3_node_name(node_name) {
            return Ok("STRING Bacteroides massiliensis dnLKV3");
        }

        if is_valid_string_eubacterium_sp_142_node_name(node_name) {
            return Ok("STRING Eubacterium sp. 142");
        }

        if is_valid_string_lachnospiraceae_bacterium_m181_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium M181");
        }

        if is_valid_string_lachnospiraceae_bacterium_coe1_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium COE1");
        }

        if is_valid_string_enterorhabdus_caecimuris_node_name(node_name) {
            return Ok("STRING Enterorhabdus caecimuris");
        }

        if is_valid_string_firmicutes_bacterium_m102_node_name(node_name) {
            return Ok("STRING Firmicutes bacterium M102");
        }

        if is_valid_string_oscillibacter_sp_13_node_name(node_name) {
            return Ok("STRING Oscillibacter sp. 13");
        }

        if is_valid_string_dorea_sp_52_node_name(node_name) {
            return Ok("STRING Dorea sp. 52");
        }

        if is_valid_string_lachnospiraceae_bacterium_32_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium 32");
        }

        if is_valid_string_lachnospiraceae_bacterium_101_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium 101");
        }

        if is_valid_string_lactobacillus_murinus_node_name(node_name) {
            return Ok("STRING Lactobacillus murinus");
        }

        if is_valid_string_eubacterium_plexicaudatum_node_name(node_name) {
            return Ok("STRING Eubacterium plexicaudatum");
        }

        if is_valid_string_parabacteroides_sp_asf519_node_name(node_name) {
            return Ok("STRING Parabacteroides sp. ASF519");
        }

        if is_valid_string_prevotella_disiens_jcm6334_node_name(node_name) {
            return Ok("STRING Prevotella disiens JCM6334");
        }

        if is_valid_string_bacteroides_pyogenes_jcm10003_node_name(node_name) {
            return Ok("STRING Bacteroides pyogenes JCM10003");
        }

        if is_valid_string_prevotella_enoeca_node_name(node_name) {
            return Ok("STRING Prevotella enoeca");
        }

        if is_valid_string_anaerotruncus_sp_g3_node_name(node_name) {
            return Ok("STRING Anaerotruncus sp. G3");
        }

        if is_valid_string_halyomorpha_halys_node_name(node_name) {
            return Ok("STRING Halyomorpha halys");
        }

        if is_valid_string_prevotella_pleuritidis_node_name(node_name) {
            return Ok("STRING Prevotella pleuritidis");
        }

        if is_valid_string_prevotella_oulorum_jcm14966_node_name(node_name) {
            return Ok("STRING Prevotella oulorum JCM14966");
        }

        if is_valid_string_acetobacter_okinawensis_node_name(node_name) {
            return Ok("STRING Acetobacter okinawensis");
        }

        if is_valid_string_prevotella_histicola_node_name(node_name) {
            return Ok("STRING Prevotella histicola");
        }

        if is_valid_string_prevotella_aurantiaca_node_name(node_name) {
            return Ok("STRING Prevotella aurantiaca");
        }

        if is_valid_string_bacteroides_stercorirosoris_node_name(node_name) {
            return Ok("STRING Bacteroides stercorirosoris");
        }

        if is_valid_string_prevotella_fusca_node_name(node_name) {
            return Ok("STRING Prevotella fusca");
        }

        if is_valid_string_prevotella_scopos_node_name(node_name) {
            return Ok("STRING Prevotella scopos");
        }

        if is_valid_string_shewanella_haliotis_node_name(node_name) {
            return Ok("STRING Shewanella haliotis");
        }

        if is_valid_string_shewanella_marina_node_name(node_name) {
            return Ok("STRING Shewanella marina");
        }

        if is_valid_string_methanomethylophilus_alvus_node_name(node_name) {
            return Ok("STRING Methanomethylophilus alvus");
        }

        if is_valid_string_nocardiopsis_baichengensis_node_name(node_name) {
            return Ok("STRING Nocardiopsis baichengensis");
        }

        if is_valid_string_wolbachia_sp_dsi_node_name(node_name) {
            return Ok("STRING Wolbachia sp. Dsi");
        }

        if is_valid_string_methylobacillus_glycogenes_node_name(node_name) {
            return Ok("STRING Methylobacillus glycogenes");
        }

        if is_valid_string_bacillus_akibai_node_name(node_name) {
            return Ok("STRING Bacillus akibai");
        }

        if is_valid_string_paenibacillus_pini_node_name(node_name) {
            return Ok("STRING Paenibacillus pini");
        }

        if is_valid_string_fulvivirga_imtechensis_node_name(node_name) {
            return Ok("STRING Fulvivirga imtechensis");
        }

        if is_valid_string_nocardiopsis_ganjiahuensis_node_name(node_name) {
            return Ok("STRING Nocardiopsis ganjiahuensis");
        }

        if is_valid_string_caenispirillum_salinarum_node_name(node_name) {
            return Ok("STRING Caenispirillum salinarum");
        }

        if is_valid_string_oceanobacillus_kimchii_node_name(node_name) {
            return Ok("STRING Oceanobacillus kimchii");
        }

        if is_valid_string_leucobacter_salsicius_node_name(node_name) {
            return Ok("STRING Leucobacter salsicius");
        }

        if is_valid_string_halomonas_jeotgali_node_name(node_name) {
            return Ok("STRING Halomonas jeotgali");
        }

        if is_valid_string_chlamydia_psittaci_10139811_node_name(node_name) {
            return Ok("STRING Chlamydia psittaci 10139811");
        }

        if is_valid_string_haloquadratum_walsbyi_j07hqw2_node_name(node_name) {
            return Ok("STRING Haloquadratum walsbyi J07HQW2");
        }

        if is_valid_string_vibrio_nigripulchritudo_node_name(node_name) {
            return Ok("STRING Vibrio nigripulchritudo");
        }

        if is_valid_string_dokdonia_sp_pro95_node_name(node_name) {
            return Ok("STRING Dokdonia sp. PRO95");
        }

        if is_valid_string_mariniradius_saccharolyticus_node_name(node_name) {
            return Ok("STRING Mariniradius saccharolyticus");
        }

        if is_valid_string_rhodococcus_ruber_node_name(node_name) {
            return Ok("STRING Rhodococcus ruber");
        }

        if is_valid_string_pseudomonas_putida_mtcc5279_node_name(node_name) {
            return Ok("STRING Pseudomonas putida MTCC5279");
        }

        if is_valid_string_cystobacter_fuscus_node_name(node_name) {
            return Ok("STRING Cystobacter fuscus");
        }

        if is_valid_string_campylobacter_concisus_atcc51562_node_name(node_name) {
            return Ok("STRING Campylobacter concisus ATCC51562");
        }

        if is_valid_string_bacillus_sp_g2_node_name(node_name) {
            return Ok("STRING Bacillus sp. G2");
        }

        if is_valid_string_campylobacter_showae_csunswcd_node_name(node_name) {
            return Ok("STRING Campylobacter showae CSUNSWCD");
        }

        if is_valid_string_campylobacter_fetus_testudinum_node_name(node_name) {
            return Ok("STRING Campylobacter fetus testudinum");
        }

        if is_valid_string_campylobacter_sp_1485e_node_name(node_name) {
            return Ok("STRING Campylobacter sp. 1485E");
        }

        if is_valid_string_magnetospirillum_sp_so1_node_name(node_name) {
            return Ok("STRING Magnetospirillum sp. SO1");
        }

        if is_valid_string_bradyrhizobium_oligotrophicum_node_name(node_name) {
            return Ok("STRING Bradyrhizobium oligotrophicum");
        }

        if is_valid_string_pseudomonas_resinovorans_nbrc106553_node_name(node_name) {
            return Ok("STRING Pseudomonas resinovorans NBRC106553");
        }

        if is_valid_string_nocardiopsis_prasina_node_name(node_name) {
            return Ok("STRING Nocardiopsis prasina");
        }

        if is_valid_string_nocardiopsis_kunsanensis_node_name(node_name) {
            return Ok("STRING Nocardiopsis kunsanensis");
        }

        if is_valid_string_nocardiopsis_valliformis_node_name(node_name) {
            return Ok("STRING Nocardiopsis valliformis");
        }

        if is_valid_string_rhizobium_sp_2mfcol31_node_name(node_name) {
            return Ok("STRING Rhizobium sp. 2MFCol31");
        }

        if is_valid_string_nocardiopsis_xinjiangensis_node_name(node_name) {
            return Ok("STRING Nocardiopsis xinjiangensis");
        }

        if is_valid_string_halobacillus_sp_bab2008_node_name(node_name) {
            return Ok("STRING Halobacillus sp. BAB2008");
        }

        if is_valid_string_bacillus_lehensis_node_name(node_name) {
            return Ok("STRING Bacillus lehensis");
        }

        if is_valid_string_mycoplasma_cynos_node_name(node_name) {
            return Ok("STRING Mycoplasma cynos");
        }

        if is_valid_string_actinoplanes_friuliensis_node_name(node_name) {
            return Ok("STRING Actinoplanes friuliensis");
        }

        if is_valid_string_arsenophonus_endosymbiont_node_name(node_name) {
            return Ok("STRING Arsenophonus endosymbiont");
        }

        if is_valid_string_bordetella_holmesii_node_name(node_name) {
            return Ok("STRING Bordetella holmesii");
        }

        if is_valid_string_advenella_mimigardefordensis_node_name(node_name) {
            return Ok("STRING Advenella mimigardefordensis");
        }

        if is_valid_string_parvularcula_oceani_node_name(node_name) {
            return Ok("STRING Parvularcula oceani");
        }

        if is_valid_string_photobacterium_leiognathi_lrivu41_node_name(node_name) {
            return Ok("STRING Photobacterium leiognathi lrivu41");
        }

        if is_valid_string_blastomonas_sp_aap53_node_name(node_name) {
            return Ok("STRING Blastomonas sp. AAP53");
        }

        if is_valid_string_sandarakinorhabdus_sp_aap62_node_name(node_name) {
            return Ok("STRING Sandarakinorhabdus sp. AAP62");
        }

        if is_valid_string_porphyrobacter_sp_aap82_node_name(node_name) {
            return Ok("STRING Porphyrobacter sp. AAP82");
        }

        if is_valid_string_sulfuricurvum_sp_rifrc1_node_name(node_name) {
            return Ok("STRING Sulfuricurvum sp. RIFRC1");
        }

        if is_valid_string_thiorhodococcus_sp_ak35_node_name(node_name) {
            return Ok("STRING Thiorhodococcus sp. AK35");
        }

        if is_valid_string_serratia_marcescens_fgi94_node_name(node_name) {
            return Ok("STRING Serratia marcescens FGI94");
        }

        if is_valid_string_gillisia_sp_heli29_node_name(node_name) {
            return Ok("STRING Gillisia sp. HelI29");
        }

        if is_valid_string_maribacter_sp_heli7_node_name(node_name) {
            return Ok("STRING Maribacter sp. HelI7");
        }

        if is_valid_string_polaribacter_sp_hel185_node_name(node_name) {
            return Ok("STRING Polaribacter sp. Hel185");
        }

        if is_valid_string_polaribacter_sp_heli88_node_name(node_name) {
            return Ok("STRING Polaribacter sp. HelI88");
        }

        if is_valid_string_muricauda_sp_mar201075_node_name(node_name) {
            return Ok("STRING Muricauda sp. MAR201075");
        }

        if is_valid_string_salegentibacter_sp_heli6_node_name(node_name) {
            return Ok("STRING Salegentibacter sp. HelI6");
        }

        if is_valid_string_sorangium_cellulosum_so01572_node_name(node_name) {
            return Ok("STRING Sorangium cellulosum So01572");
        }

        if is_valid_string_thioalkalivibrio_nitratireducens_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio nitratireducens");
        }

        if is_valid_string_eubacterium_ramulus_node_name(node_name) {
            return Ok("STRING Eubacterium ramulus");
        }

        if is_valid_string_coprothermobacter_platensis_node_name(node_name) {
            return Ok("STRING Coprothermobacter platensis");
        }

        if is_valid_string_spiribacter_salinus_node_name(node_name) {
            return Ok("STRING Spiribacter salinus");
        }

        if is_valid_string_enterococcus_faecalis_13sdw01_node_name(node_name) {
            return Ok("STRING Enterococcus faecalis 13SDW01");
        }

        if is_valid_string_liberibacter_americanus_node_name(node_name) {
            return Ok("STRING Liberibacter americanus");
        }

        if is_valid_string_halarchaeum_acidiphilum_node_name(node_name) {
            return Ok("STRING Halarchaeum acidiphilum");
        }

        if is_valid_string_clostridium_pasteurianum_dsm525_node_name(node_name) {
            return Ok("STRING Clostridium pasteurianum DSM525");
        }

        if is_valid_string_phascolarctobacterium_sp_cag207_node_name(node_name) {
            return Ok("STRING Phascolarctobacterium sp. CAG207");
        }

        if is_valid_string_phascolarctobacterium_sp_cag266_node_name(node_name) {
            return Ok("STRING Phascolarctobacterium sp. CAG266");
        }

        if is_valid_string_bibersteinia_trehalosi_node_name(node_name) {
            return Ok("STRING Bibersteinia trehalosi");
        }

        if is_valid_string_mycobacterium_asiaticum_node_name(node_name) {
            return Ok("STRING Mycobacterium asiaticum");
        }

        if is_valid_string_haliea_rubra_node_name(node_name) {
            return Ok("STRING Haliea rubra");
        }

        if is_valid_string_pseudomonas_sp_urmo17wk12i8_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. URMO17WK12I8");
        }

        if is_valid_string_caldimonas_manganoxidans_node_name(node_name) {
            return Ok("STRING Caldimonas manganoxidans");
        }

        if is_valid_string_colwellia_piezophila_node_name(node_name) {
            return Ok("STRING Colwellia piezophila");
        }

        if is_valid_string_desulfospira_joergensenii_node_name(node_name) {
            return Ok("STRING Desulfospira joergensenii");
        }

        if is_valid_string_succinimonas_amylolytica_node_name(node_name) {
            return Ok("STRING Succinimonas amylolytica");
        }

        if is_valid_string_pelagibacter_ubique_htcc9022_node_name(node_name) {
            return Ok("STRING Pelagibacter ubique HTCC9022");
        }

        if is_valid_string_listeria_weihenstephanensis_node_name(node_name) {
            return Ok("STRING Listeria weihenstephanensis");
        }

        if is_valid_string_carnobacterium_sp_wn1359_node_name(node_name) {
            return Ok("STRING Carnobacterium sp. WN1359");
        }

        if is_valid_string_thioalkalivibrio_sp_ale6_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. ALE6");
        }

        if is_valid_string_thioalkalivibrio_sp_ale19_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. ALE19");
        }

        if is_valid_string_thioalkalivibrio_sp_akl19_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. AKL19");
        }

        if is_valid_string_nitrosospira_briensis_node_name(node_name) {
            return Ok("STRING Nitrosospira briensis");
        }

        if is_valid_string_paracoccus_zeaxanthinifaciens_node_name(node_name) {
            return Ok("STRING Paracoccus zeaxanthinifaciens");
        }

        if is_valid_string_lactobacillus_parabrevis_node_name(node_name) {
            return Ok("STRING Lactobacillus parabrevis");
        }

        if is_valid_string_hyphomicrobium_zavarzinii_node_name(node_name) {
            return Ok("STRING Hyphomicrobium zavarzinii");
        }

        if is_valid_string_sediminibacterium_sp_c3_node_name(node_name) {
            return Ok("STRING Sediminibacterium sp. C3");
        }

        if is_valid_string_acidobacteriaceae_bacterium_kbs83_node_name(node_name) {
            return Ok("STRING Acidobacteriaceae bacterium KBS83");
        }

        if is_valid_string_acidobacteriaceae_bacterium_kbs89_node_name(node_name) {
            return Ok("STRING Acidobacteriaceae bacterium KBS89");
        }

        if is_valid_string_acidobacteriaceae_bacterium_kbs96_node_name(node_name) {
            return Ok("STRING Acidobacteriaceae bacterium KBS96");
        }

        if is_valid_string_anoxybacillus_flavithermus_tno09006_node_name(node_name) {
            return Ok("STRING Anoxybacillus flavithermus TNO09006");
        }

        if is_valid_string_pantoea_sp_imh_node_name(node_name) {
            return Ok("STRING Pantoea sp. IMH");
        }

        if is_valid_string_pseudomonas_sp_g5_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. G5");
        }

        if is_valid_string_paenibacillus_sabinae_node_name(node_name) {
            return Ok("STRING Paenibacillus sabinae");
        }

        if is_valid_string_aeromonas_diversa_node_name(node_name) {
            return Ok("STRING Aeromonas diversa");
        }

        if is_valid_string_pseudoalteromonas_luteoviolacea_node_name(node_name) {
            return Ok("STRING Pseudoalteromonas luteoviolacea");
        }

        if is_valid_string_bacteroides_cellulosilyticus_node_name(node_name) {
            return Ok("STRING Bacteroides cellulosilyticus");
        }

        if is_valid_string_rhodococcus_sp_aw25m09_node_name(node_name) {
            return Ok("STRING Rhodococcus sp. AW25M09");
        }

        if is_valid_string_acidovorax_sp_mrs7_node_name(node_name) {
            return Ok("STRING Acidovorax sp. MRS7");
        }

        if is_valid_string_legionella_oakridgensis_node_name(node_name) {
            return Ok("STRING Legionella oakridgensis");
        }

        if is_valid_string_thioalkalivibrio_sp_alr1721_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. ALR1721");
        }

        if is_valid_string_flavobacterium_sp_kjj_node_name(node_name) {
            return Ok("STRING Flavobacterium sp. KJJ");
        }

        if is_valid_string_pedobacter_sp_r2019_node_name(node_name) {
            return Ok("STRING Pedobacter sp. R2019");
        }

        if is_valid_string_megasphaera_sp_nm10_node_name(node_name) {
            return Ok("STRING Megasphaera sp. NM10");
        }

        if is_valid_string_rhodococcus_rhodnii_node_name(node_name) {
            return Ok("STRING Rhodococcus rhodnii");
        }

        if is_valid_string_planomicrobium_glaciei_node_name(node_name) {
            return Ok("STRING Planomicrobium glaciei");
        }

        if is_valid_string_paenibacillus_sp_gd11_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. GD11");
        }

        if is_valid_string_hepatobacter_penaei_node_name(node_name) {
            return Ok("STRING Hepatobacter penaei");
        }

        if is_valid_string_bacillus_sonorensis_node_name(node_name) {
            return Ok("STRING Bacillus sonorensis");
        }

        if is_valid_string_spiroplasma_taiwanense_node_name(node_name) {
            return Ok("STRING Spiroplasma taiwanense");
        }

        if is_valid_string_spiroplasma_diminutum_node_name(node_name) {
            return Ok("STRING Spiroplasma diminutum");
        }

        if is_valid_string_spiroplasma_chrysopicola_node_name(node_name) {
            return Ok("STRING Spiroplasma chrysopicola");
        }

        if is_valid_string_spiroplasma_syrphidicola_node_name(node_name) {
            return Ok("STRING Spiroplasma syrphidicola");
        }

        if is_valid_string_spiroplasma_culicicola_node_name(node_name) {
            return Ok("STRING Spiroplasma culicicola");
        }

        if is_valid_string_spiroplasma_sabaudiense_node_name(node_name) {
            return Ok("STRING Spiroplasma sabaudiense");
        }

        if is_valid_string_spiroplasma_apis_node_name(node_name) {
            return Ok("STRING Spiroplasma apis");
        }

        if is_valid_string_acidovorax_sp_jhl9_node_name(node_name) {
            return Ok("STRING Acidovorax sp. JHL9");
        }

        if is_valid_string_arthrobacter_gangotriensis_node_name(node_name) {
            return Ok("STRING Arthrobacter gangotriensis");
        }

        if is_valid_string_myxococcus_stipitatus_node_name(node_name) {
            return Ok("STRING Myxococcus stipitatus");
        }

        if is_valid_string_rhodococcus_triatomae_node_name(node_name) {
            return Ok("STRING Rhodococcus triatomae");
        }

        if is_valid_string_acholeplasma_granularum_node_name(node_name) {
            return Ok("STRING Acholeplasma granularum");
        }

        if is_valid_string_fusobacterium_russii_node_name(node_name) {
            return Ok("STRING Fusobacterium russii");
        }

        if is_valid_string_psychromonas_ossibalaenae_node_name(node_name) {
            return Ok("STRING Psychromonas ossibalaenae");
        }

        if is_valid_string_zimmermannella_faecalis_node_name(node_name) {
            return Ok("STRING Zimmermannella faecalis");
        }

        if is_valid_string_amphritea_japonica_node_name(node_name) {
            return Ok("STRING Amphritea japonica");
        }

        if is_valid_string_acholeplasma_axanthum_node_name(node_name) {
            return Ok("STRING Acholeplasma axanthum");
        }

        if is_valid_string_avibacterium_paragallinarum_221_node_name(node_name) {
            return Ok("STRING Avibacterium paragallinarum 221");
        }

        if is_valid_string_cesiribacter_andamanensis_node_name(node_name) {
            return Ok("STRING Cesiribacter andamanensis");
        }

        if is_valid_string_oceanimonas_smirnovii_node_name(node_name) {
            return Ok("STRING Oceanimonas smirnovii");
        }

        if is_valid_string_microbulbifer_variabilis_node_name(node_name) {
            return Ok("STRING Microbulbifer variabilis");
        }

        if is_valid_string_thioalkalivibrio_thiocyanoxidans_arh2_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio thiocyanoxidans ARh2");
        }

        if is_valid_string_novispirillum_itersonii_node_name(node_name) {
            return Ok("STRING Novispirillum itersonii");
        }

        if is_valid_string_vibrio_jasicida_node_name(node_name) {
            return Ok("STRING Vibrio jasicida");
        }

        if is_valid_string_synechococcus_sp_kordi100_node_name(node_name) {
            return Ok("STRING Synechococcus sp. KORDI100");
        }

        if is_valid_string_paenibacillaceae_bacterium_g5_node_name(node_name) {
            return Ok("STRING Paenibacillaceae bacterium G5");
        }

        if is_valid_string_butyrivibrio_sp_ae2015_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. AE2015");
        }

        if is_valid_string_butyrivibrio_sp_vcd2006_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. VCD2006");
        }

        if is_valid_string_butyrivibrio_sp_ae3009_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. AE3009");
        }

        if is_valid_string_butyrivibrio_sp_xpd2006_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. XPD2006");
        }

        if is_valid_string_butyrivibrio_sp_fc2001_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. FC2001");
        }

        if is_valid_string_butyrivibrio_sp_ae3006_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. AE3006");
        }

        if is_valid_string_prevotella_sp_agr2160_node_name(node_name) {
            return Ok("STRING Prevotella sp. AGR2160");
        }

        if is_valid_string_butyrivibrio_sp_wcd3002_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. WCD3002");
        }

        if is_valid_string_butyrivibrio_sp_vcb2006_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. VCB2006");
        }

        if is_valid_string_butyrivibrio_sp_lc3010_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. LC3010");
        }

        if is_valid_string_butyrivibrio_sp_wcd2001_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. WCD2001");
        }

        if is_valid_string_butyrivibrio_sp_xbb1001_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. XBB1001");
        }

        if is_valid_string_butyrivibrio_sp_nc3005_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. NC3005");
        }

        if is_valid_string_butyrivibrio_sp_mc2013_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. MC2013");
        }

        if is_valid_string_pseudobutyrivibrio_ruminis_cf1b_node_name(node_name) {
            return Ok("STRING Pseudobutyrivibrio ruminis CF1b");
        }

        if is_valid_string_clostridium_paraputrificum_node_name(node_name) {
            return Ok("STRING Clostridium paraputrificum");
        }

        if is_valid_string_clostridium_cadaveris_node_name(node_name) {
            return Ok("STRING Clostridium cadaveris");
        }

        if is_valid_string_pseudobutyrivibrio_ruminis_ad2017_node_name(node_name) {
            return Ok("STRING Pseudobutyrivibrio ruminis AD2017");
        }

        if is_valid_string_butyrivibrio_fibrisolvens_nd3005_node_name(node_name) {
            return Ok("STRING Butyrivibrio fibrisolvens ND3005");
        }

        if is_valid_string_dorea_longicatena_agr2136_node_name(node_name) {
            return Ok("STRING Dorea longicatena AGR2136");
        }

        if is_valid_string_selenomonas_ruminantium_atcc12561_node_name(node_name) {
            return Ok("STRING Selenomonas ruminantium ATCC12561");
        }

        if is_valid_string_hyphomonas_sp_t16b2_node_name(node_name) {
            return Ok("STRING Hyphomonas sp. T16B2");
        }

        if is_valid_string_hyphomonas_sp_cy54118_node_name(node_name) {
            return Ok("STRING Hyphomonas sp. CY54118");
        }

        if is_valid_string_hyphomonas_sp_25b141_node_name(node_name) {
            return Ok("STRING Hyphomonas sp. 25B141");
        }

        if is_valid_string_hyphomonas_sp_bhbn044_node_name(node_name) {
            return Ok("STRING Hyphomonas sp. BHBN044");
        }

        if is_valid_string_hyphomonas_sp_22ii122f38_node_name(node_name) {
            return Ok("STRING Hyphomonas sp. 22II122F38");
        }

        if is_valid_string_hyphomonas_adhaerens_node_name(node_name) {
            return Ok("STRING Hyphomonas adhaerens");
        }

        if is_valid_string_hyphomonas_johnsonii_node_name(node_name) {
            return Ok("STRING Hyphomonas johnsonii");
        }

        if is_valid_string_hyphomonas_jannaschiana_node_name(node_name) {
            return Ok("STRING Hyphomonas jannaschiana");
        }

        if is_valid_string_hyphomonas_oceanitis_node_name(node_name) {
            return Ok("STRING Hyphomonas oceanitis");
        }

        if is_valid_string_hyphomonas_polymorpha_node_name(node_name) {
            return Ok("STRING Hyphomonas polymorpha");
        }

        if is_valid_string_agrobacterium_tumefaciens_cherry_node_name(node_name) {
            return Ok("STRING Agrobacterium tumefaciens Cherry");
        }

        if is_valid_string_pseudomonas_poae_node_name(node_name) {
            return Ok("STRING Pseudomonas poae");
        }

        if is_valid_string_asticcacaulis_sp_ac460_node_name(node_name) {
            return Ok("STRING Asticcacaulis sp. AC460");
        }

        if is_valid_string_asticcacaulis_sp_ac402_node_name(node_name) {
            return Ok("STRING Asticcacaulis sp. AC402");
        }

        if is_valid_string_asticcacaulis_sp_ac466_node_name(node_name) {
            return Ok("STRING Asticcacaulis sp. AC466");
        }

        if is_valid_string_streptococcus_tigurinus_az3a_node_name(node_name) {
            return Ok("STRING Streptococcus tigurinus AZ3a");
        }

        if is_valid_string_streptococcus_tigurinus_1366_node_name(node_name) {
            return Ok("STRING Streptococcus tigurinus 1366");
        }

        if is_valid_string_alpha_proteobacterium_mf_105b01_node_name(node_name) {
            return Ok("STRING alpha proteobacterium Mf 105b01");
        }

        if is_valid_string_lachnospira_multipara_atcc19207_node_name(node_name) {
            return Ok("STRING Lachnospira multipara ATCC19207");
        }

        if is_valid_string_frankia_sp_iso899_node_name(node_name) {
            return Ok("STRING Frankia sp. Iso899");
        }

        if is_valid_string_tolumonas_sp_brl61_node_name(node_name) {
            return Ok("STRING Tolumonas sp. BRL61");
        }

        if is_valid_string_nocardioides_sp_iso805n_node_name(node_name) {
            return Ok("STRING Nocardioides sp. Iso805N");
        }

        if is_valid_string_conexibacter_woesei_iso977n_node_name(node_name) {
            return Ok("STRING Conexibacter woesei Iso977N");
        }

        if is_valid_string_methylohalobius_crimeensis_node_name(node_name) {
            return Ok("STRING Methylohalobius crimeensis");
        }

        if is_valid_string_paenibacillus_sp_a9_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. A9");
        }

        if is_valid_string_actinomyces_urogenitalis_s6c4_node_name(node_name) {
            return Ok("STRING Actinomyces urogenitalis S6C4");
        }

        if is_valid_string_actinomyces_sp_s6spd3_node_name(node_name) {
            return Ok("STRING Actinomyces sp. S6Spd3");
        }

        if is_valid_string_anaerococcus_lactolyticus_s7113_node_name(node_name) {
            return Ok("STRING Anaerococcus lactolyticus S7113");
        }

        if is_valid_string_clostridiales_bacterium_s714_node_name(node_name) {
            return Ok("STRING Clostridiales bacterium S714");
        }

        if is_valid_string_prevotella_sp_s718_node_name(node_name) {
            return Ok("STRING Prevotella sp. S718");
        }

        if is_valid_string_corynebacterium_casei_node_name(node_name) {
            return Ok("STRING Corynebacterium casei");
        }

        if is_valid_string_lysinibacillus_sphaericus_ot4b31_node_name(node_name) {
            return Ok("STRING Lysinibacillus sphaericus OT4b31");
        }

        if is_valid_string_pandoraea_sp_sd62_node_name(node_name) {
            return Ok("STRING Pandoraea sp. SD62");
        }

        if is_valid_string_methylophaga_lonarensis_node_name(node_name) {
            return Ok("STRING Methylophaga lonarensis");
        }

        if is_valid_string_raoultella_ornithinolytica_node_name(node_name) {
            return Ok("STRING Raoultella ornithinolytica");
        }

        if is_valid_string_eubacterium_acidaminophilum_node_name(node_name) {
            return Ok("STRING Eubacterium acidaminophilum");
        }

        if is_valid_string_sphaerotilus_natans_node_name(node_name) {
            return Ok("STRING Sphaerotilus natans");
        }

        if is_valid_string_zhouia_amylolytica_node_name(node_name) {
            return Ok("STRING Zhouia amylolytica");
        }

        if is_valid_string_mesorhizobium_sp_l2c084a000_node_name(node_name) {
            return Ok("STRING Mesorhizobium sp. L2C084A000");
        }

        if is_valid_string_mesorhizobium_sp_lnjc398b00_node_name(node_name) {
            return Ok("STRING Mesorhizobium sp. LNJC398B00");
        }

        if is_valid_string_corynebacterium_freneyi_node_name(node_name) {
            return Ok("STRING Corynebacterium freneyi");
        }

        if is_valid_string_prevotella_bivia_dnf00188_node_name(node_name) {
            return Ok("STRING Prevotella bivia DNF00188");
        }

        if is_valid_string_prevotella_sp_s7_ms_2_node_name(node_name) {
            return Ok("STRING Prevotella sp. S7 MS 2");
        }

        if is_valid_string_streptomyces_sp_cnt318_node_name(node_name) {
            return Ok("STRING Streptomyces sp. CNT318");
        }

        if is_valid_string_streptomyces_sp_taa040_node_name(node_name) {
            return Ok("STRING Streptomyces sp. TAA040");
        }

        if is_valid_string_roseovarius_mucosus_node_name(node_name) {
            return Ok("STRING Roseovarius mucosus");
        }

        if is_valid_string_deinococcus_wulumuqiensis_node_name(node_name) {
            return Ok("STRING Deinococcus wulumuqiensis");
        }

        if is_valid_string_nitrosospira_sp_apg3_node_name(node_name) {
            return Ok("STRING Nitrosospira sp. APG3");
        }

        if is_valid_string_marinobacter_santoriniensis_node_name(node_name) {
            return Ok("STRING Marinobacter santoriniensis");
        }

        if is_valid_string_cyclobacteriaceae_bacterium_ak24_node_name(node_name) {
            return Ok("STRING Cyclobacteriaceae bacterium AK24");
        }

        if is_valid_string_brachyspira_hampsonii_node_name(node_name) {
            return Ok("STRING Brachyspira hampsonii");
        }

        if is_valid_string_streptomyces_sp_taa204_node_name(node_name) {
            return Ok("STRING Streptomyces sp. TAA204");
        }

        if is_valid_string_clostridium_josui_node_name(node_name) {
            return Ok("STRING Clostridium josui");
        }

        if is_valid_string_lactobacillus_oryzae_node_name(node_name) {
            return Ok("STRING Lactobacillus oryzae");
        }

        if is_valid_string_dietzia_sp_ucdthp_node_name(node_name) {
            return Ok("STRING Dietzia sp. UCDTHP");
        }

        if is_valid_string_mycoplasma_putrefaciens_mput9231_node_name(node_name) {
            return Ok("STRING Mycoplasma putrefaciens Mput9231");
        }

        if is_valid_string_caulobacter_crescentus_or37_node_name(node_name) {
            return Ok("STRING Caulobacter crescentus OR37");
        }

        if is_valid_string_clostridium_sordellii_node_name(node_name) {
            return Ok("STRING Clostridium sordellii");
        }

        if is_valid_string_propionibacterium_granulosum_node_name(node_name) {
            return Ok("STRING Propionibacterium granulosum");
        }

        if is_valid_string_halopiger_sp_iih2_node_name(node_name) {
            return Ok("STRING Halopiger sp. IIH2");
        }

        if is_valid_string_halopiger_sp_iih3_node_name(node_name) {
            return Ok("STRING Halopiger sp. IIH3");
        }

        if is_valid_string_halanaerobium_saccharolyticum_node_name(node_name) {
            return Ok("STRING Halanaerobium saccharolyticum");
        }

        if is_valid_string_lactobacillus_equicursoris_node_name(node_name) {
            return Ok("STRING Lactobacillus equicursoris");
        }

        if is_valid_string_clostridium_intestinale_node_name(node_name) {
            return Ok("STRING Clostridium intestinale");
        }

        if is_valid_string_pseudomonas_denitrificans_node_name(node_name) {
            return Ok("STRING Pseudomonas denitrificans");
        }

        if is_valid_string_bacillus_boroniphilus_node_name(node_name) {
            return Ok("STRING Bacillus boroniphilus");
        }

        if is_valid_string_roseibacterium_elongatum_node_name(node_name) {
            return Ok("STRING Roseibacterium elongatum");
        }

        if is_valid_string_geobacillus_stearothermophilus_node_name(node_name) {
            return Ok("STRING Geobacillus stearothermophilus");
        }

        if is_valid_string_aquimarina_sp_sw150_node_name(node_name) {
            return Ok("STRING Aquimarina sp. SW150");
        }

        if is_valid_string_aquimarina_megaterium_node_name(node_name) {
            return Ok("STRING Aquimarina megaterium");
        }

        if is_valid_string_komagataeibacter_xylinus_node_name(node_name) {
            return Ok("STRING Komagataeibacter xylinus");
        }

        if is_valid_string_streptococcus_dentisani_7746_node_name(node_name) {
            return Ok("STRING Streptococcus dentisani 7746");
        }

        if is_valid_string_mesorhizobium_metallidurans_node_name(node_name) {
            return Ok("STRING Mesorhizobium metallidurans");
        }

        if is_valid_string_mesorhizobium_sp_stm4661_node_name(node_name) {
            return Ok("STRING Mesorhizobium sp. STM4661");
        }

        if is_valid_string_anoxybacillus_flavithermus_ak1_node_name(node_name) {
            return Ok("STRING Anoxybacillus flavithermus AK1");
        }

        if is_valid_string_intestinimonas_butyriciproducens_node_name(node_name) {
            return Ok("STRING Intestinimonas butyriciproducens");
        }

        if is_valid_string_myxococcus_sp_node_name(node_name) {
            return Ok("STRING Myxococcus sp.");
        }

        if is_valid_string_streptococcus_dentisani_7747_node_name(node_name) {
            return Ok("STRING Streptococcus dentisani 7747");
        }

        if is_valid_string_afipia_sp_ohsuic4_node_name(node_name) {
            return Ok("STRING Afipia sp. OHSUIC4");
        }

        if is_valid_string_bradyrhizobium_sp_ohsuiii_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. OHSUIII");
        }

        if is_valid_string_thalassolituus_oleivorans_node_name(node_name) {
            return Ok("STRING Thalassolituus oleivorans");
        }

        if is_valid_string_gracilibacillus_boraciitolerans_node_name(node_name) {
            return Ok("STRING Gracilibacillus boraciitolerans");
        }

        if is_valid_string_psychrobacter_sp_jcm18900_node_name(node_name) {
            return Ok("STRING Psychrobacter sp. JCM18900");
        }

        if is_valid_string_mesorhizobium_sp_urha0056_node_name(node_name) {
            return Ok("STRING Mesorhizobium sp. URHA0056");
        }

        if is_valid_string_microbacterium_sp_urha0036_node_name(node_name) {
            return Ok("STRING Microbacterium sp. URHA0036");
        }

        if is_valid_string_marmoricola_sp_urhb0036_node_name(node_name) {
            return Ok("STRING Marmoricola sp. URHB0036");
        }

        if is_valid_string_mycobacterium_sp_urhd0025_node_name(node_name) {
            return Ok("STRING Mycobacterium sp. URHD0025");
        }

        if is_valid_string_alteromonas_sp_alt199_node_name(node_name) {
            return Ok("STRING Alteromonas sp. ALT199");
        }

        if is_valid_string_bradyrhizobium_sp_urha0002_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. URHA0002");
        }

        if is_valid_string_streptomyces_sp_taa486_node_name(node_name) {
            return Ok("STRING Streptomyces sp. TAA486");
        }

        if is_valid_string_desulfotomaculum_guttoideum_node_name(node_name) {
            return Ok("STRING Desulfotomaculum guttoideum");
        }

        if is_valid_string_mycobacterium_kansasii_732_node_name(node_name) {
            return Ok("STRING Mycobacterium kansasii 732");
        }

        if is_valid_string_chryseobacterium_oranimense_node_name(node_name) {
            return Ok("STRING Chryseobacterium oranimense");
        }

        if is_valid_string_enterococcus_mundtii_node_name(node_name) {
            return Ok("STRING Enterococcus mundtii");
        }

        if is_valid_string_lysobacter_dokdonensis_node_name(node_name) {
            return Ok("STRING Lysobacter dokdonensis");
        }

        if is_valid_string_sulfitobacter_donghicola_node_name(node_name) {
            return Ok("STRING Sulfitobacter donghicola");
        }

        if is_valid_string_pseudomonas_knackmussii_node_name(node_name) {
            return Ok("STRING Pseudomonas knackmussii");
        }

        if is_valid_string_clostridium_sp_01_node_name(node_name) {
            return Ok("STRING Clostridium sp. 01");
        }

        if is_valid_string_lactobacillus_namurensis_node_name(node_name) {
            return Ok("STRING Lactobacillus namurensis");
        }

        if is_valid_string_borrelia_miyamotoi_node_name(node_name) {
            return Ok("STRING Borrelia miyamotoi");
        }

        if is_valid_string_streptococcus_oligofermentans_node_name(node_name) {
            return Ok("STRING Streptococcus oligofermentans");
        }

        if is_valid_string_chthonomonas_calidirosea_node_name(node_name) {
            return Ok("STRING Chthonomonas calidirosea");
        }

        if is_valid_string_streptomyces_fulvissimus_node_name(node_name) {
            return Ok("STRING Streptomyces fulvissimus");
        }

        if is_valid_string_salinisphaera_hydrothermalis_node_name(node_name) {
            return Ok("STRING Salinisphaera hydrothermalis");
        }

        if is_valid_string_clostridiaceae_bacterium_l21thd2_node_name(node_name) {
            return Ok("STRING Clostridiaceae bacterium L21THD2");
        }

        if is_valid_string_cellulomonas_sp_krmcy2_node_name(node_name) {
            return Ok("STRING Cellulomonas sp. KRMCY2");
        }

        if is_valid_string_clostridium_sp_asbs410_node_name(node_name) {
            return Ok("STRING Clostridium sp. ASBs410");
        }

        if is_valid_string_desulfovibrio_magneticus_ifrc170_node_name(node_name) {
            return Ok("STRING Desulfovibrio magneticus IFRC170");
        }

        if is_valid_string_aminiphilus_circumscriptus_node_name(node_name) {
            return Ok("STRING Aminiphilus circumscriptus");
        }

        if is_valid_string_aminobacterium_mobile_node_name(node_name) {
            return Ok("STRING Aminobacterium mobile");
        }

        if is_valid_string_arthrobacter_nicotinovorans_node_name(node_name) {
            return Ok("STRING Arthrobacter nicotinovorans");
        }

        if is_valid_string_bradyrhizobium_japonicum_22_node_name(node_name) {
            return Ok("STRING Bradyrhizobium japonicum 22");
        }

        if is_valid_string_bradyrhizobium_japonicum_in8p8_node_name(node_name) {
            return Ok("STRING Bradyrhizobium japonicum in8p8");
        }

        if is_valid_string_caldicoprobacter_oshimai_node_name(node_name) {
            return Ok("STRING Caldicoprobacter oshimai");
        }

        if is_valid_string_dechloromonas_agitata_node_name(node_name) {
            return Ok("STRING Dechloromonas agitata");
        }

        if is_valid_string_desulforegula_conservatrix_node_name(node_name) {
            return Ok("STRING Desulforegula conservatrix");
        }

        if is_valid_string_geovibrio_sp_l21acebes_node_name(node_name) {
            return Ok("STRING Geovibrio sp. L21AceBES");
        }

        if is_valid_string_microbacterium_sp_krocy2_node_name(node_name) {
            return Ok("STRING Microbacterium sp. KROCY2");
        }

        if is_valid_string_oceanicola_sp_hl35_node_name(node_name) {
            return Ok("STRING Oceanicola sp. HL35");
        }

        if is_valid_string_algoriphagus_marincola_hl49_node_name(node_name) {
            return Ok("STRING Algoriphagus marincola HL49");
        }

        if is_valid_string_sporosarcina_sp_eur3222_node_name(node_name) {
            return Ok("STRING Sporosarcina sp. EUR3222");
        }

        if is_valid_string_kineosporia_aurantiaca_node_name(node_name) {
            return Ok("STRING Kineosporia aurantiaca");
        }

        if is_valid_string_streptomyces_thermolilacinus_node_name(node_name) {
            return Ok("STRING Streptomyces thermolilacinus");
        }

        if is_valid_string_candidate_division_tm6_node_name(node_name) {
            return Ok("STRING candidate division TM6");
        }

        if is_valid_string_streptomyces_sp_r1ns10_node_name(node_name) {
            return Ok("STRING Streptomyces sp. R1NS10");
        }

        if is_valid_string_bacillus_firmus_node_name(node_name) {
            return Ok("STRING Bacillus firmus");
        }

        if is_valid_string_pseudoalteromonas_agarivorans_node_name(node_name) {
            return Ok("STRING Pseudoalteromonas agarivorans");
        }

        if is_valid_string_desulfovibrio_sp_l21syrab_node_name(node_name) {
            return Ok("STRING Desulfovibrio sp. L21SyrAB");
        }

        if is_valid_string_spirochaeta_sp_l21rpuld2_node_name(node_name) {
            return Ok("STRING Spirochaeta sp. L21RPulD2");
        }

        if is_valid_string_gluconobacter_frateurii_node_name(node_name) {
            return Ok("STRING Gluconobacter frateurii");
        }

        if is_valid_string_gracilibacillus_halophilus_node_name(node_name) {
            return Ok("STRING Gracilibacillus halophilus");
        }

        if is_valid_string_arthrobacter_sp_br18_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. Br18");
        }

        if is_valid_string_arthrobacter_sp_h14_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. H14");
        }

        if is_valid_string_ilumatobacter_coccineus_node_name(node_name) {
            return Ok("STRING Ilumatobacter coccineus");
        }

        if is_valid_string_thermocrinis_sp_gbs_node_name(node_name) {
            return Ok("STRING Thermocrinis sp. GBS");
        }

        if is_valid_string_borrelia_coriaceae_node_name(node_name) {
            return Ok("STRING Borrelia coriaceae");
        }

        if is_valid_string_borrelia_anserina_node_name(node_name) {
            return Ok("STRING Borrelia anserina");
        }

        if is_valid_string_borrelia_parkeri_node_name(node_name) {
            return Ok("STRING Borrelia parkeri");
        }

        if is_valid_string_thermonema_rossianum_node_name(node_name) {
            return Ok("STRING Thermonema rossianum");
        }

        if is_valid_string_chitinivibrio_alkaliphilus_node_name(node_name) {
            return Ok("STRING Chitinivibrio alkaliphilus");
        }

        if is_valid_string_aureispira_sp_ccbqb1_node_name(node_name) {
            return Ok("STRING Aureispira sp. CCBQB1");
        }

        if is_valid_string_streptococcus_sp_hsism1_node_name(node_name) {
            return Ok("STRING Streptococcus sp. HSISM1");
        }

        if is_valid_string_blattabacterium_sp_nci_node_name(node_name) {
            return Ok("STRING Blattabacterium sp. Nci");
        }

        if is_valid_string_pseudomonas_corrugata_node_name(node_name) {
            return Ok("STRING Pseudomonas corrugata");
        }

        if is_valid_string_mannheimia_haemolytica_m42548_node_name(node_name) {
            return Ok("STRING Mannheimia haemolytica M42548");
        }

        if is_valid_string_phaeospirillum_fulvum_node_name(node_name) {
            return Ok("STRING Phaeospirillum fulvum");
        }

        if is_valid_string_roseivivax_sp_22iis10s_node_name(node_name) {
            return Ok("STRING Roseivivax sp. 22IIs10s");
        }

        if is_valid_string_aquimarina_sp_22iis11z7_node_name(node_name) {
            return Ok("STRING Aquimarina sp. 22IIS11z7");
        }

        if is_valid_string_thioclava_sp_13d2w2_node_name(node_name) {
            return Ok("STRING Thioclava sp. 13D2W2");
        }

        if is_valid_string_mycoplasma_girerdii_node_name(node_name) {
            return Ok("STRING Mycoplasma girerdii");
        }

        if is_valid_string_marinobacter_lipolyticus_sm19_node_name(node_name) {
            return Ok("STRING Marinobacter lipolyticus SM19");
        }

        if is_valid_string_cetobacterium_somerae_node_name(node_name) {
            return Ok("STRING Cetobacterium somerae");
        }

        if is_valid_string_mesorhizobium_sp_nbimcp2c3_node_name(node_name) {
            return Ok("STRING Mesorhizobium sp. NBIMCP2C3");
        }

        if is_valid_string_streptococcus_suis_ea183292_node_name(node_name) {
            return Ok("STRING Streptococcus suis EA183292");
        }

        if is_valid_string_atopobium_sp_f0209_node_name(node_name) {
            return Ok("STRING Atopobium sp. F0209");
        }

        if is_valid_string_leptotrichia_sp_f0581_node_name(node_name) {
            return Ok("STRING Leptotrichia sp. F0581");
        }

        if is_valid_string_actinomyces_sp_f0311_node_name(node_name) {
            return Ok("STRING Actinomyces sp. F0311");
        }

        if is_valid_string_clostridiales_bacterium_f0540_node_name(node_name) {
            return Ok("STRING Clostridiales bacterium F0540");
        }

        if is_valid_string_leptotrichia_sp_w9775_node_name(node_name) {
            return Ok("STRING Leptotrichia sp. W9775");
        }

        if is_valid_string_mitsuokella_sp_w9106_node_name(node_name) {
            return Ok("STRING Mitsuokella sp. W9106");
        }

        if is_valid_string_oribacterium_sp_f0263_node_name(node_name) {
            return Ok("STRING Oribacterium sp. F0263");
        }

        if is_valid_string_peptostreptococcaceae_bacterium_w5053_node_name(node_name) {
            return Ok("STRING Peptostreptococcaceae bacterium W5053");
        }

        if is_valid_string_selenomonas_sp_f0426_node_name(node_name) {
            return Ok("STRING Selenomonas sp. F0426");
        }

        if is_valid_string_eubacterium_brachy_node_name(node_name) {
            return Ok("STRING Eubacterium brachy");
        }

        if is_valid_string_treponema_lecithinolyticum_node_name(node_name) {
            return Ok("STRING Treponema lecithinolyticum");
        }

        if is_valid_string_gemella_bergeriae_node_name(node_name) {
            return Ok("STRING Gemella bergeriae");
        }

        if is_valid_string_desulfovibrio_piezophilus_node_name(node_name) {
            return Ok("STRING Desulfovibrio piezophilus");
        }

        if is_valid_string_rhodococcus_defluvii_node_name(node_name) {
            return Ok("STRING Rhodococcus defluvii");
        }

        if is_valid_string_pseudomonas_pelagia_node_name(node_name) {
            return Ok("STRING Pseudomonas pelagia");
        }

        if is_valid_string_halobonum_tyrrellensis_node_name(node_name) {
            return Ok("STRING Halobonum tyrrellensis");
        }

        if is_valid_string_helicobacter_fennelliae_node_name(node_name) {
            return Ok("STRING Helicobacter fennelliae");
        }

        if is_valid_string_catenovulum_agarivorans_ds2_node_name(node_name) {
            return Ok("STRING Catenovulum agarivorans DS2");
        }

        if is_valid_string_weissella_oryzae_node_name(node_name) {
            return Ok("STRING Weissella oryzae");
        }

        if is_valid_string_thermoactinomyces_daqus_node_name(node_name) {
            return Ok("STRING Thermoactinomyces daqus");
        }

        if is_valid_string_thermus_caliditerrae_node_name(node_name) {
            return Ok("STRING Thermus caliditerrae");
        }

        if is_valid_string_sphingobium_lactosutens_node_name(node_name) {
            return Ok("STRING Sphingobium lactosutens");
        }

        if is_valid_string_acinetobacter_haemolyticus_node_name(node_name) {
            return Ok("STRING Acinetobacter haemolyticus");
        }

        if is_valid_string_serratia_fonticola_auap2c_node_name(node_name) {
            return Ok("STRING Serratia fonticola AUAP2C");
        }

        if is_valid_string_pseudoalteromonas_haloplanktis_tb64_node_name(node_name) {
            return Ok("STRING Pseudoalteromonas haloplanktis TB64");
        }

        if is_valid_string_salinarchaeum_sp_harchtbsk1_node_name(node_name) {
            return Ok("STRING Salinarchaeum sp. HarchtBsk1");
        }

        if is_valid_string_pseudomonas_stutzeri_mf28_node_name(node_name) {
            return Ok("STRING Pseudomonas stutzeri MF28");
        }

        if is_valid_string_alpha_proteobacterium_ma2_node_name(node_name) {
            return Ok("STRING alpha proteobacterium MA2");
        }

        if is_valid_string_kurthia_huakuii_node_name(node_name) {
            return Ok("STRING Kurthia huakuii");
        }

        if is_valid_string_spiribacter_sp_uahsp71_node_name(node_name) {
            return Ok("STRING Spiribacter sp. UAHSP71");
        }

        if is_valid_string_sphingobium_sp_yl23_node_name(node_name) {
            return Ok("STRING Sphingobium sp. YL23");
        }

        if is_valid_string_roseomonas_gilardii_node_name(node_name) {
            return Ok("STRING Roseomonas gilardii");
        }

        if is_valid_string_shewanella_waksmanii_node_name(node_name) {
            return Ok("STRING Shewanella waksmanii");
        }

        if is_valid_string_atopobacter_phocae_node_name(node_name) {
            return Ok("STRING Atopobacter phocae");
        }

        if is_valid_string_rhizobium_selenitireducens_node_name(node_name) {
            return Ok("STRING Rhizobium selenitireducens");
        }

        if is_valid_string_pseudomonas_flectens_node_name(node_name) {
            return Ok("STRING Pseudomonas flectens");
        }

        if is_valid_string_eubacterium_xylanophilum_node_name(node_name) {
            return Ok("STRING Eubacterium xylanophilum");
        }

        if is_valid_string_microvirga_flocculans_node_name(node_name) {
            return Ok("STRING Microvirga flocculans");
        }

        if is_valid_string_carnimonas_nigrificans_node_name(node_name) {
            return Ok("STRING Carnimonas nigrificans");
        }

        if is_valid_string_rhizobium_larrymoorei_node_name(node_name) {
            return Ok("STRING Rhizobium larrymoorei");
        }

        if is_valid_string_polaribacter_sp_hel13349_node_name(node_name) {
            return Ok("STRING Polaribacter sp. Hel13349");
        }

        if is_valid_string_loktanella_cinnabarina_node_name(node_name) {
            return Ok("STRING Loktanella cinnabarina");
        }

        if is_valid_string_calothrix_sp_3363_node_name(node_name) {
            return Ok("STRING Calothrix sp. 3363");
        }

        if is_valid_string_elizabethkingia_anophelis_node_name(node_name) {
            return Ok("STRING Elizabethkingia anophelis");
        }

        if is_valid_string_bacillus_sp_unc438cl73tsus30_node_name(node_name) {
            return Ok("STRING Bacillus sp. UNC438CL73TsuS30");
        }

        if is_valid_string_bryobacter_aggregatus_node_name(node_name) {
            return Ok("STRING Bryobacter aggregatus");
        }

        if is_valid_string_laceyella_sacchari_node_name(node_name) {
            return Ok("STRING Laceyella sacchari");
        }

        if is_valid_string_flavobacterium_saliperosum_node_name(node_name) {
            return Ok("STRING Flavobacterium saliperosum");
        }

        if is_valid_string_ruminococcus_flavefaciens_007c_node_name(node_name) {
            return Ok("STRING Ruminococcus flavefaciens 007c");
        }

        if is_valid_string_flavobacterium_limnosediminis_node_name(node_name) {
            return Ok("STRING Flavobacterium limnosediminis");
        }

        if is_valid_string_mycobacterium_septicum_node_name(node_name) {
            return Ok("STRING Mycobacterium septicum");
        }

        if is_valid_string_acinetobacter_indicus_node_name(node_name) {
            return Ok("STRING Acinetobacter indicus");
        }

        if is_valid_string_bifidobacterium_bombi_node_name(node_name) {
            return Ok("STRING Bifidobacterium bombi");
        }

        if is_valid_string_sulfitobacter_sp_mm124_node_name(node_name) {
            return Ok("STRING Sulfitobacter sp. MM124");
        }

        if is_valid_string_sulfitobacter_sp_nb77_node_name(node_name) {
            return Ok("STRING Sulfitobacter sp. NB77");
        }

        if is_valid_string_sulfitobacter_sp_nb68_node_name(node_name) {
            return Ok("STRING Sulfitobacter sp. NB68");
        }

        if is_valid_string_saccharibacter_sp_am169_node_name(node_name) {
            return Ok("STRING Saccharibacter sp. AM169");
        }

        if is_valid_string_palaeococcus_pacificus_node_name(node_name) {
            return Ok("STRING Palaeococcus pacificus");
        }

        if is_valid_string_streptomyces_rapamycinicus_node_name(node_name) {
            return Ok("STRING Streptomyces rapamycinicus");
        }

        if is_valid_string_tatumella_sp_nml063099_node_name(node_name) {
            return Ok("STRING Tatumella sp. NML063099");
        }

        if is_valid_string_exiguobacterium_pavilionensis_node_name(node_name) {
            return Ok("STRING Exiguobacterium pavilionensis");
        }

        if is_valid_string_helicobacter_pylori_sa213a_node_name(node_name) {
            return Ok("STRING Helicobacter pylori SA213A");
        }

        if is_valid_string_clostridium_saccharobutylicum_node_name(node_name) {
            return Ok("STRING Clostridium saccharobutylicum");
        }

        if is_valid_string_geobacillus_sp_jf8_node_name(node_name) {
            return Ok("STRING Geobacillus sp. JF8");
        }

        if is_valid_string_sphingobacterium_paucimobilis_node_name(node_name) {
            return Ok("STRING Sphingobacterium paucimobilis");
        }

        if is_valid_string_sphingobium_ummariense_node_name(node_name) {
            return Ok("STRING Sphingobium ummariense");
        }

        if is_valid_string_bacillus_sp_eb01_node_name(node_name) {
            return Ok("STRING Bacillus sp. EB01");
        }

        if is_valid_string_paucisalibacillus_sp_eb02_node_name(node_name) {
            return Ok("STRING Paucisalibacillus sp. EB02");
        }

        if is_valid_string_formosa_agariphila_node_name(node_name) {
            return Ok("STRING Formosa agariphila");
        }

        if is_valid_string_bacillus_sp_ff3_node_name(node_name) {
            return Ok("STRING Bacillus sp. FF3");
        }

        if is_valid_string_bacillus_sp_ff4_node_name(node_name) {
            return Ok("STRING Bacillus sp. FF4");
        }

        if is_valid_string_anaerosalibacter_sp_nd1_node_name(node_name) {
            return Ok("STRING Anaerosalibacter sp. ND1");
        }

        if is_valid_string_bacteroidaceae_bacterium_ms4_node_name(node_name) {
            return Ok("STRING Bacteroidaceae bacterium MS4");
        }

        if is_valid_string_pseudoalteromonas_sp_ocn003_node_name(node_name) {
            return Ok("STRING Pseudoalteromonas sp. OCN003");
        }

        if is_valid_string_leifsonia_rubra_node_name(node_name) {
            return Ok("STRING Leifsonia rubra");
        }

        if is_valid_string_cellulophaga_baltica_node_name(node_name) {
            return Ok("STRING Cellulophaga baltica");
        }

        if is_valid_string_vibrio_diazotrophicus_node_name(node_name) {
            return Ok("STRING Vibrio diazotrophicus");
        }

        if is_valid_string_thauera_terpenica_node_name(node_name) {
            return Ok("STRING Thauera terpenica");
        }

        if is_valid_string_corynebacterium_argentoratense_node_name(node_name) {
            return Ok("STRING Corynebacterium argentoratense");
        }

        if is_valid_string_kitasatospora_cheerisanensis_node_name(node_name) {
            return Ok("STRING Kitasatospora cheerisanensis");
        }

        if is_valid_string_bacillus_megaterium_msp201_node_name(node_name) {
            return Ok("STRING Bacillus megaterium MSP201");
        }

        if is_valid_string_janthinobacterium_agaricidamnosum_node_name(node_name) {
            return Ok("STRING Janthinobacterium agaricidamnosum");
        }

        if is_valid_string_tenacibaculum_maritimum_node_name(node_name) {
            return Ok("STRING Tenacibaculum maritimum");
        }

        if is_valid_string_arthrobacter_sp_akyn10_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. AKYN10");
        }

        if is_valid_string_coprobacter_fastidiosus_node_name(node_name) {
            return Ok("STRING Coprobacter fastidiosus");
        }

        if is_valid_string_streptomyces_niveus_node_name(node_name) {
            return Ok("STRING Streptomyces niveus");
        }

        if is_valid_string_olleya_marilimosa_node_name(node_name) {
            return Ok("STRING Olleya marilimosa");
        }

        if is_valid_string_thioclava_sp_dt234_node_name(node_name) {
            return Ok("STRING Thioclava sp. DT234");
        }

        if is_valid_string_bacteriovorax_sp_bsw11iv_node_name(node_name) {
            return Ok("STRING Bacteriovorax sp. BSW11IV");
        }

        if is_valid_string_ensifer_sp_tw10_node_name(node_name) {
            return Ok("STRING Ensifer sp. TW10");
        }

        if is_valid_string_thioclava_pacifica_node_name(node_name) {
            return Ok("STRING Thioclava pacifica");
        }

        if is_valid_string_peptoniphilus_sp_chdc_node_name(node_name) {
            return Ok("STRING Peptoniphilus sp. ChDC");
        }

        if is_valid_string_psychrobacter_aquaticus_node_name(node_name) {
            return Ok("STRING Psychrobacter aquaticus");
        }

        if is_valid_string_cardinium_endosymbiont_cbtq1_node_name(node_name) {
            return Ok("STRING Cardinium endosymbiont cBtQ1");
        }

        if is_valid_string_roseovarius_sp_mctg156_node_name(node_name) {
            return Ok("STRING Roseovarius sp. MCTG156");
        }

        if is_valid_string_arcobacter_sp_af1028_node_name(node_name) {
            return Ok("STRING Arcobacter sp. AF1028");
        }

        if is_valid_string_arcobacter_sp_af1440_node_name(node_name) {
            return Ok("STRING Arcobacter sp. AF1440");
        }

        if is_valid_string_hymenobacter_sp_apr13_node_name(node_name) {
            return Ok("STRING Hymenobacter sp. APR13");
        }

        if is_valid_string_alicyclobacillus_acidoterrestris_node_name(node_name) {
            return Ok("STRING Alicyclobacillus acidoterrestris");
        }

        if is_valid_string_pseudomonas_syringae_cc1417_node_name(node_name) {
            return Ok("STRING Pseudomonas syringae CC1417");
        }

        if is_valid_string_pseudomonas_syringae_cc1513_node_name(node_name) {
            return Ok("STRING Pseudomonas syringae CC1513");
        }

        if is_valid_string_pseudomonas_syringae_cc1557_node_name(node_name) {
            return Ok("STRING Pseudomonas syringae CC1557");
        }

        if is_valid_string_helicobacter_canis_node_name(node_name) {
            return Ok("STRING Helicobacter canis");
        }

        if is_valid_string_helicobacter_macacae_node_name(node_name) {
            return Ok("STRING Helicobacter macacae");
        }

        if is_valid_string_sphingobacterium_antarcticus_node_name(node_name) {
            return Ok("STRING Sphingobacterium antarcticus");
        }

        if is_valid_string_thermofilum_sp_1910b_node_name(node_name) {
            return Ok("STRING Thermofilum sp. 1910b");
        }

        if is_valid_string_rhodobacteraceae_bacterium_himb11_node_name(node_name) {
            return Ok("STRING Rhodobacteraceae bacterium HIMB11");
        }

        if is_valid_string_ralstonia_pickettii_dtp0602_node_name(node_name) {
            return Ok("STRING Ralstonia pickettii DTP0602");
        }

        if is_valid_string_campylobacter_coli_76339_node_name(node_name) {
            return Ok("STRING Campylobacter coli 76339");
        }

        if is_valid_string_paracoccus_aminophilus_node_name(node_name) {
            return Ok("STRING Paracoccus aminophilus");
        }

        if is_valid_string_mycobacterium_sp_umwgj_node_name(node_name) {
            return Ok("STRING Mycobacterium sp. UMWGJ");
        }

        if is_valid_string_mycobacterium_sp_umwwy_node_name(node_name) {
            return Ok("STRING Mycobacterium sp. UMWWY");
        }

        if is_valid_string_rhizobium_undicola_node_name(node_name) {
            return Ok("STRING Rhizobium undicola");
        }

        if is_valid_string_mycobacterium_iranicum_node_name(node_name) {
            return Ok("STRING Mycobacterium iranicum");
        }

        if is_valid_string_firmicutes_bacterium_asf500_node_name(node_name) {
            return Ok("STRING Firmicutes bacterium ASF500");
        }

        if is_valid_string_gemmatimonas_sp_ap64_node_name(node_name) {
            return Ok("STRING Gemmatimonas sp. AP64");
        }

        if is_valid_string_desulfonauticus_sp_a7a_node_name(node_name) {
            return Ok("STRING Desulfonauticus sp. A7A");
        }

        if is_valid_string_candidate_division_zixi_node_name(node_name) {
            return Ok("STRING candidate division ZIXI");
        }

        if is_valid_string_sphingomonas_sp_fukuswis1_node_name(node_name) {
            return Ok("STRING Sphingomonas sp. FUKUSWIS1");
        }

        if is_valid_string_mucispirillum_schaedleri_node_name(node_name) {
            return Ok("STRING Mucispirillum schaedleri");
        }

        if is_valid_string_streptomyces_sp_urha0041_node_name(node_name) {
            return Ok("STRING Streptomyces sp. URHA0041");
        }

        if is_valid_string_geodermatophilaceae_bacterium_urhb0048_node_name(node_name) {
            return Ok("STRING Geodermatophilaceae bacterium URHB0048");
        }

        if is_valid_string_mesorhizobium_sp_urhc0008_node_name(node_name) {
            return Ok("STRING Mesorhizobium sp. URHC0008");
        }

        if is_valid_string_cellulomonas_sp_urhe0023_node_name(node_name) {
            return Ok("STRING Cellulomonas sp. URHE0023");
        }

        if is_valid_string_bradyrhizobium_sp_urhd0069_node_name(node_name) {
            return Ok("STRING Bradyrhizobium sp. URHD0069");
        }

        if is_valid_string_blastococcus_sp_urhd0036_node_name(node_name) {
            return Ok("STRING Blastococcus sp. URHD0036");
        }

        if is_valid_string_halomonas_sp_23gom1509m_node_name(node_name) {
            return Ok("STRING Halomonas sp. 23GOM1509m");
        }

        if is_valid_string_sulfitobacter_sp_20gpm1509m_node_name(node_name) {
            return Ok("STRING Sulfitobacter sp. 20GPM1509m");
        }

        if is_valid_string_intrasporangiaceae_bacterium_urhb0013_node_name(node_name) {
            return Ok("STRING Intrasporangiaceae bacterium URHB0013");
        }

        if is_valid_string_ahrensia_sp_13gom1096m_node_name(node_name) {
            return Ok("STRING Ahrensia sp. 13GOM1096m");
        }

        if is_valid_string_tenacibaculum_sp_47agom205m_node_name(node_name) {
            return Ok("STRING Tenacibaculum sp. 47AGOM205m");
        }

        if is_valid_string_mycobacterium_sp_urhb0044_node_name(node_name) {
            return Ok("STRING Mycobacterium sp. URHB0044");
        }

        if is_valid_string_alcanivorax_sp_43bgom46m_node_name(node_name) {
            return Ok("STRING Alcanivorax sp. 43BGOM46m");
        }

        if is_valid_string_solirubrobacterales_bacterium_urhd0059_node_name(node_name) {
            return Ok("STRING Solirubrobacterales bacterium URHD0059");
        }

        if is_valid_string_dongia_sp_urhe0060_node_name(node_name) {
            return Ok("STRING Dongia sp. URHE0060");
        }

        if is_valid_string_geodermatophilaceae_bacterium_urha0031_node_name(node_name) {
            return Ok("STRING Geodermatophilaceae bacterium URHA0031");
        }

        if is_valid_string_rhodospirillales_bacterium_urhd0088_node_name(node_name) {
            return Ok("STRING Rhodospirillales bacterium URHD0088");
        }

        if is_valid_string_anoxybacillus_flavithermus_yunnanensis_node_name(node_name) {
            return Ok("STRING Anoxybacillus flavithermus yunnanensis");
        }

        if is_valid_string_flaviramulus_ichthyoenteri_node_name(node_name) {
            return Ok("STRING Flaviramulus ichthyoenteri");
        }

        if is_valid_string_paenibacillus_darwinianus_node_name(node_name) {
            return Ok("STRING Paenibacillus darwinianus");
        }

        if is_valid_string_aliihoeflea_sp_2ww_node_name(node_name) {
            return Ok("STRING Aliihoeflea sp. 2WW");
        }

        if is_valid_string_brevibacterium_sp_vcm10_node_name(node_name) {
            return Ok("STRING Brevibacterium sp. VCM10");
        }

        if is_valid_string_asaia_platycodi_sf21_node_name(node_name) {
            return Ok("STRING Asaia platycodi SF21");
        }

        if is_valid_string_caulobacteraceae_bacterium_pmmr1_node_name(node_name) {
            return Ok("STRING Caulobacteraceae bacterium PMMR1");
        }

        if is_valid_string_alicyclobacillus_macrosporangiidus_node_name(node_name) {
            return Ok("STRING Alicyclobacillus macrosporangiidus");
        }

        if is_valid_string_sporosarcina_sp_d27_node_name(node_name) {
            return Ok("STRING Sporosarcina sp. D27");
        }

        if is_valid_string_thermogemmatispora_sp_pm5_node_name(node_name) {
            return Ok("STRING Thermogemmatispora sp. PM5");
        }

        if is_valid_string_geobacillus_vulcani_node_name(node_name) {
            return Ok("STRING Geobacillus vulcani");
        }

        if is_valid_string_thermomicrobiales_sp_ki4_node_name(node_name) {
            return Ok("STRING Thermomicrobiales sp. KI4");
        }

        if is_valid_string_anoxybacillus_tepidamans_node_name(node_name) {
            return Ok("STRING Anoxybacillus tepidamans");
        }

        if is_valid_string_acidobacterium_sp_pmmr2_node_name(node_name) {
            return Ok("STRING Acidobacterium sp. PMMR2");
        }

        if is_valid_string_lysinibacillus_manganicus_node_name(node_name) {
            return Ok("STRING Lysinibacillus manganicus");
        }

        if is_valid_string_arenimonas_malthae_node_name(node_name) {
            return Ok("STRING Arenimonas malthae");
        }

        if is_valid_string_arenimonas_metalli_node_name(node_name) {
            return Ok("STRING Arenimonas metalli");
        }

        if is_valid_string_lysinibacillus_sinduriensis_node_name(node_name) {
            return Ok("STRING Lysinibacillus sinduriensis");
        }

        if is_valid_string_ruminococcus_albus_ad2013_node_name(node_name) {
            return Ok("STRING Ruminococcus albus AD2013");
        }

        if is_valid_string_ruminococcus_flavefaciens_ae3010_node_name(node_name) {
            return Ok("STRING Ruminococcus flavefaciens AE3010");
        }

        if is_valid_string_adlercreutzia_equolifaciens_node_name(node_name) {
            return Ok("STRING Adlercreutzia equolifaciens");
        }

        if is_valid_string_francisella_sp_w121067_node_name(node_name) {
            return Ok("STRING Francisella sp. W121067");
        }

        if is_valid_string_pontibacillus_halophilus_node_name(node_name) {
            return Ok("STRING Pontibacillus halophilus");
        }

        if is_valid_string_pontibacillus_marinus_node_name(node_name) {
            return Ok("STRING Pontibacillus marinus");
        }

        if is_valid_string_pontibacillus_litoralis_node_name(node_name) {
            return Ok("STRING Pontibacillus litoralis");
        }

        if is_valid_string_pontibacillus_chungwhensis_node_name(node_name) {
            return Ok("STRING Pontibacillus chungwhensis");
        }

        if is_valid_string_pontibacillus_yanchengensis_node_name(node_name) {
            return Ok("STRING Pontibacillus yanchengensis");
        }

        if is_valid_string_lysobacter_defluvii_node_name(node_name) {
            return Ok("STRING Lysobacter defluvii");
        }

        if is_valid_string_lysobacter_daejeonensis_node_name(node_name) {
            return Ok("STRING Lysobacter daejeonensis");
        }

        if is_valid_string_knoellia_flava_node_name(node_name) {
            return Ok("STRING Knoellia flava");
        }

        if is_valid_string_knoellia_aerolata_node_name(node_name) {
            return Ok("STRING Knoellia aerolata");
        }

        if is_valid_string_knoellia_sinensis_node_name(node_name) {
            return Ok("STRING Knoellia sinensis");
        }

        if is_valid_string_knoellia_subterranea_node_name(node_name) {
            return Ok("STRING Knoellia subterranea");
        }

        if is_valid_string_leptolyngbya_sp_heron_node_name(node_name) {
            return Ok("STRING Leptolyngbya sp. Heron");
        }

        if is_valid_string_intrasporangium_oryzae_node_name(node_name) {
            return Ok("STRING Intrasporangium oryzae");
        }

        if is_valid_string_gordonia_amicalis_node_name(node_name) {
            return Ok("STRING Gordonia amicalis");
        }

        if is_valid_string_schmidhempelia_bombi_node_name(node_name) {
            return Ok("STRING Schmidhempelia bombi");
        }

        if is_valid_string_methylophilus_sp_oh31_node_name(node_name) {
            return Ok("STRING Methylophilus sp. OH31");
        }

        if is_valid_string_pseudomonas_mosselii_node_name(node_name) {
            return Ok("STRING Pseudomonas mosselii");
        }

        if is_valid_string_leifsonia_xyli_cynodontis_node_name(node_name) {
            return Ok("STRING Leifsonia xyli cynodontis");
        }

        if is_valid_string_pseudomonas_mendocina_egdaq5_node_name(node_name) {
            return Ok("STRING Pseudomonas mendocina EGDAQ5");
        }

        if is_valid_string_clostridium_bifermentans_node_name(node_name) {
            return Ok("STRING Clostridium bifermentans");
        }

        if is_valid_string_clostridium_sp_adety_node_name(node_name) {
            return Ok("STRING Clostridium sp. AdeTY");
        }

        if is_valid_string_pseudozyma_brasiliensis_node_name(node_name) {
            return Ok("STRING Pseudozyma brasiliensis");
        }

        if is_valid_string_prevotella_sp_hun102_node_name(node_name) {
            return Ok("STRING Prevotella sp. HUN102");
        }

        if is_valid_string_eubacterium_sp_ab3007_node_name(node_name) {
            return Ok("STRING Eubacterium sp. AB3007");
        }

        if is_valid_string_leeuwenhoekiella_sp_heli48_node_name(node_name) {
            return Ok("STRING Leeuwenhoekiella sp. HelI48");
        }

        if is_valid_string_leeuwenhoekiella_sp_mar2009132_node_name(node_name) {
            return Ok("STRING Leeuwenhoekiella sp. MAR2009132");
        }

        if is_valid_string_sediminibacter_sp_heli10_node_name(node_name) {
            return Ok("STRING Sediminibacter sp. HelI10");
        }

        if is_valid_string_ruminococcaceae_bacterium_ae2021_node_name(node_name) {
            return Ok("STRING Ruminococcaceae bacterium AE2021");
        }

        if is_valid_string_lachnospiraceae_bacterium_ac2031_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium AC2031");
        }

        if is_valid_string_maribacter_forsetii_node_name(node_name) {
            return Ok("STRING Maribacter forsetii");
        }

        if is_valid_string_selenomonas_ruminantium_ac2024_node_name(node_name) {
            return Ok("STRING Selenomonas ruminantium AC2024");
        }

        if is_valid_string_selenomonas_ruminantium_ab3002_node_name(node_name) {
            return Ok("STRING Selenomonas ruminantium AB3002");
        }

        if is_valid_string_acinetobacter_nectaris_node_name(node_name) {
            return Ok("STRING Acinetobacter nectaris");
        }

        if is_valid_string_bordetella_hinzii_node_name(node_name) {
            return Ok("STRING Bordetella hinzii");
        }

        if is_valid_string_bifidobacterium_sp_7101_node_name(node_name) {
            return Ok("STRING Bifidobacterium sp. 7101");
        }

        if is_valid_string_bifidobacterium_sp_a11_node_name(node_name) {
            return Ok("STRING Bifidobacterium sp. A11");
        }

        if is_valid_string_actinomadura_madurae_node_name(node_name) {
            return Ok("STRING Actinomadura madurae");
        }

        if is_valid_string_sporolactobacillus_laevolacticus_node_name(node_name) {
            return Ok("STRING Sporolactobacillus laevolacticus");
        }

        if is_valid_string_pseudomonas_moraviensis_node_name(node_name) {
            return Ok("STRING Pseudomonas moraviensis");
        }

        if is_valid_string_pseudomonas_taeanensis_node_name(node_name) {
            return Ok("STRING Pseudomonas taeanensis");
        }

        if is_valid_string_paenibacillus_sp_maepy2_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. MAEPY2");
        }

        if is_valid_string_haloferula_sp_bvorr071_node_name(node_name) {
            return Ok("STRING Haloferula sp. BvORR071");
        }

        if is_valid_string_verrucomicrobium_sp_bvorr034_node_name(node_name) {
            return Ok("STRING Verrucomicrobium sp. BvORR034");
        }

        if is_valid_string_marinobacter_sp_es1_node_name(node_name) {
            return Ok("STRING Marinobacter sp. ES1");
        }

        if is_valid_string_leucobacter_sp_ph1c_node_name(node_name) {
            return Ok("STRING Leucobacter sp. PH1c");
        }

        if is_valid_string_serratia_sp_h1n_node_name(node_name) {
            return Ok("STRING Serratia sp. H1n");
        }

        if is_valid_string_alcanivorax_sp_p2s70_node_name(node_name) {
            return Ok("STRING Alcanivorax sp. P2S70");
        }

        if is_valid_string_halomonas_sp_pbn3_node_name(node_name) {
            return Ok("STRING Halomonas sp. PBN3");
        }

        if is_valid_string_alpha_proteobacterium_rs24_node_name(node_name) {
            return Ok("STRING alpha proteobacterium RS24");
        }

        if is_valid_string_exiguobacterium_undae_node_name(node_name) {
            return Ok("STRING Exiguobacterium undae");
        }

        if is_valid_string_exiguobacterium_marinum_node_name(node_name) {
            return Ok("STRING Exiguobacterium marinum");
        }

        if is_valid_string_exiguobacterium_oxidotolerans_node_name(node_name) {
            return Ok("STRING Exiguobacterium oxidotolerans");
        }

        if is_valid_string_borrelia_valaisiana_tom4006_node_name(node_name) {
            return Ok("STRING Borrelia valaisiana Tom4006");
        }

        if is_valid_string_exiguobacterium_sp_mh3_node_name(node_name) {
            return Ok("STRING Exiguobacterium sp. MH3");
        }

        if is_valid_string_holospora_obtusa_node_name(node_name) {
            return Ok("STRING Holospora obtusa");
        }

        if is_valid_string_enterobacter_cloacae_jd6301_node_name(node_name) {
            return Ok("STRING Enterobacter cloacae JD6301");
        }

        if is_valid_string_lactobacillus_fabifermentans_node_name(node_name) {
            return Ok("STRING Lactobacillus fabifermentans");
        }

        if is_valid_string_pelagibacter_ubique_htcc7214_node_name(node_name) {
            return Ok("STRING Pelagibacter ubique HTCC7214");
        }

        if is_valid_string_pelagibacter_ubique_htcc7217_node_name(node_name) {
            return Ok("STRING Pelagibacter ubique HTCC7217");
        }

        if is_valid_string_corynebacterium_tuscaniense_node_name(node_name) {
            return Ok("STRING Corynebacterium tuscaniense");
        }

        if is_valid_string_oligella_urethralis_node_name(node_name) {
            return Ok("STRING Oligella urethralis");
        }

        if is_valid_string_veillonella_montpellierensis_node_name(node_name) {
            return Ok("STRING Veillonella montpellierensis");
        }

        if is_valid_string_prevotella_buccalis_dnf00985_node_name(node_name) {
            return Ok("STRING Prevotella buccalis DNF00985");
        }

        if is_valid_string_endolissoclinum_faulkneri_l5_node_name(node_name) {
            return Ok("STRING Endolissoclinum faulkneri L5");
        }

        if is_valid_string_sulfitobacter_sp_h3_node_name(node_name) {
            return Ok("STRING Sulfitobacter sp. H3");
        }

        if is_valid_string_bacillus_simplex_node_name(node_name) {
            return Ok("STRING Bacillus simplex");
        }

        if is_valid_string_verrucomicrobium_sp_bvorr106_node_name(node_name) {
            return Ok("STRING Verrucomicrobium sp. BvORR106");
        }

        if is_valid_string_streptococcus_anginosus_dora7_node_name(node_name) {
            return Ok("STRING Streptococcus anginosus DORA7");
        }

        if is_valid_string_varibaculum_cambriense_dora20_node_name(node_name) {
            return Ok("STRING Varibaculum cambriense DORA20");
        }

        if is_valid_string_corynebacterium_glycinophilum_node_name(node_name) {
            return Ok("STRING Corynebacterium glycinophilum");
        }

        if is_valid_string_chlamydia_suis_node_name(node_name) {
            return Ok("STRING Chlamydia suis");
        }

        if is_valid_string_staphylococcus_simulans_node_name(node_name) {
            return Ok("STRING Staphylococcus simulans");
        }

        if is_valid_string_flavobacterium_beibuense_node_name(node_name) {
            return Ok("STRING Flavobacterium beibuense");
        }

        if is_valid_string_synechococcus_sp_nkbg15041c_node_name(node_name) {
            return Ok("STRING Synechococcus sp. NKBG15041c");
        }

        if is_valid_string_betaproteobacteria_bacterium_mola814_node_name(node_name) {
            return Ok("STRING Betaproteobacteria bacterium MOLA814");
        }

        if is_valid_string_sinorhizobium_americanum_node_name(node_name) {
            return Ok("STRING Sinorhizobium americanum");
        }

        if is_valid_string_vagococcus_lutrae_node_name(node_name) {
            return Ok("STRING Vagococcus lutrae");
        }

        if is_valid_string_brevibacillus_panacihumi_node_name(node_name) {
            return Ok("STRING Brevibacillus panacihumi");
        }

        if is_valid_string_fusobacterium_nucleatum_w1481_node_name(node_name) {
            return Ok("STRING Fusobacterium nucleatum W1481");
        }

        if is_valid_string_bacillus_sp_mb2021_node_name(node_name) {
            return Ok("STRING Bacillus sp. MB2021");
        }

        if is_valid_string_butyrivibrio_sp_fcs014_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. FCS014");
        }

        if is_valid_string_butyrivibrio_sp_mc2021_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. MC2021");
        }

        if is_valid_string_prevotella_sp_ma2016_node_name(node_name) {
            return Ok("STRING Prevotella sp. MA2016");
        }

        if is_valid_string_oribacterium_sp_fc2011_node_name(node_name) {
            return Ok("STRING Oribacterium sp. FC2011");
        }

        if is_valid_string_pseudobutyrivibrio_sp_lb2011_node_name(node_name) {
            return Ok("STRING Pseudobutyrivibrio sp. LB2011");
        }

        if is_valid_string_lachnospiraceae_bacterium_ac2028_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium AC2028");
        }

        if is_valid_string_lachnospiraceae_bacterium_ac3007_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium AC3007");
        }

        if is_valid_string_lachnospiraceae_bacterium_ma2020_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium MA2020");
        }

        if is_valid_string_lachnospiraceae_bacterium_mc2017_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium MC2017");
        }

        if is_valid_string_acholeplasma_equifetale_node_name(node_name) {
            return Ok("STRING Acholeplasma equifetale");
        }

        if is_valid_string_acholeplasma_hippikon_node_name(node_name) {
            return Ok("STRING Acholeplasma hippikon");
        }

        if is_valid_string_acholeplasma_modicum_node_name(node_name) {
            return Ok("STRING Acholeplasma modicum");
        }

        if is_valid_string_acidiphilium_angustum_node_name(node_name) {
            return Ok("STRING Acidiphilium angustum");
        }

        if is_valid_string_acidocella_facilis_node_name(node_name) {
            return Ok("STRING Acidocella facilis");
        }

        if is_valid_string_alkaliphilus_transvaalensis_node_name(node_name) {
            return Ok("STRING Alkaliphilus transvaalensis");
        }

        if is_valid_string_anaerovibrio_lipolyticus_node_name(node_name) {
            return Ok("STRING Anaerovibrio lipolyticus");
        }

        if is_valid_string_bacillus_bogoriensis_node_name(node_name) {
            return Ok("STRING Bacillus bogoriensis");
        }

        if is_valid_string_bartonella_elizabethae_node_name(node_name) {
            return Ok("STRING Bartonella elizabethae");
        }

        if is_valid_string_bilophila_wadsworthia_atcc49260_node_name(node_name) {
            return Ok("STRING Bilophila wadsworthia ATCC49260");
        }

        if is_valid_string_crocinitomix_catalasitica_node_name(node_name) {
            return Ok("STRING Crocinitomix catalasitica");
        }

        if is_valid_string_eubacterium_cellulosolvens_ld2006_node_name(node_name) {
            return Ok("STRING Eubacterium cellulosolvens LD2006");
        }

        if is_valid_string_eubacterium_desmolans_node_name(node_name) {
            return Ok("STRING Eubacterium desmolans");
        }

        if is_valid_string_facklamia_sourekii_node_name(node_name) {
            return Ok("STRING Facklamia sourekii");
        }

        if is_valid_string_fusobacterium_perfoetens_node_name(node_name) {
            return Ok("STRING Fusobacterium perfoetens");
        }

        if is_valid_string_helicobacter_pametensis_node_name(node_name) {
            return Ok("STRING Helicobacter pametensis");
        }

        if is_valid_string_legionella_fairfieldensis_node_name(node_name) {
            return Ok("STRING Legionella fairfieldensis");
        }

        if is_valid_string_legionella_sainthelensi_node_name(node_name) {
            return Ok("STRING Legionella sainthelensi");
        }

        if is_valid_string_mycobacterium_genavense_node_name(node_name) {
            return Ok("STRING Mycobacterium genavense");
        }

        if is_valid_string_prolixibacter_bellariivorans_node_name(node_name) {
            return Ok("STRING Prolixibacter bellariivorans");
        }

        if is_valid_string_sphingobacterium_sp_h1ai_node_name(node_name) {
            return Ok("STRING Sphingobacterium sp. H1ai");
        }

        if is_valid_string_clostridium_mangenotii_node_name(node_name) {
            return Ok("STRING Clostridium mangenotii");
        }

        if is_valid_string_bacteroides_sp_ga6a2_node_name(node_name) {
            return Ok("STRING Bacteroides sp. Ga6A2");
        }

        if is_valid_string_treponema_sp_c6a8_node_name(node_name) {
            return Ok("STRING Treponema sp. C6A8");
        }

        if is_valid_string_oribacterium_sp_p6a1_node_name(node_name) {
            return Ok("STRING Oribacterium sp. P6A1");
        }

        if is_valid_string_prevotella_sp_p6b1_node_name(node_name) {
            return Ok("STRING Prevotella sp. P6B1");
        }

        if is_valid_string_pseudobutyrivibrio_sp_md2005_node_name(node_name) {
            return Ok("STRING Pseudobutyrivibrio sp. MD2005");
        }

        if is_valid_string_ruminococcus_sp_fc2018_node_name(node_name) {
            return Ok("STRING Ruminococcus sp. FC2018");
        }

        if is_valid_string_selenomonas_sp_nd2010_node_name(node_name) {
            return Ok("STRING Selenomonas sp. ND2010");
        }

        if is_valid_string_serratia_sp_dd3_node_name(node_name) {
            return Ok("STRING Serratia sp. DD3");
        }

        if is_valid_string_shinella_sp_dd12_node_name(node_name) {
            return Ok("STRING Shinella sp. DD12");
        }

        if is_valid_string_lachnospiraceae_bacterium_c6a11_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium C6A11");
        }

        if is_valid_string_lachnospiraceae_bacterium_fe2018_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium FE2018");
        }

        if is_valid_string_lachnospiraceae_bacterium_md2004_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium MD2004");
        }

        if is_valid_string_lachnospiraceae_bacterium_nc2004_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium NC2004");
        }

        if is_valid_string_lachnospiraceae_bacterium_nd2006_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium ND2006");
        }

        if is_valid_string_lachnospiraceae_bacterium_p6a3_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium P6A3");
        }

        if is_valid_string_lachnospiraceae_bacterium_p6b14_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium P6B14");
        }

        if is_valid_string_lachnospiraceae_bacterium_v9d3004_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium V9D3004");
        }

        if is_valid_string_lachnospiraceae_bacterium_ysb2008_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium YSB2008");
        }

        if is_valid_string_propionibacteriaceae_bacterium_p6a17_node_name(node_name) {
            return Ok("STRING Propionibacteriaceae bacterium P6A17");
        }

        if is_valid_string_ruminococcaceae_bacterium_ab4001_node_name(node_name) {
            return Ok("STRING Ruminococcaceae bacterium AB4001");
        }

        if is_valid_string_butyrivibrio_proteoclasticus_p6b7_node_name(node_name) {
            return Ok("STRING Butyrivibrio proteoclasticus P6B7");
        }

        if is_valid_string_clostridium_lundense_node_name(node_name) {
            return Ok("STRING Clostridium lundense");
        }

        if is_valid_string_kandleria_vitulina_node_name(node_name) {
            return Ok("STRING Kandleria vitulina");
        }

        if is_valid_string_lachnospira_multipara_lb2003_node_name(node_name) {
            return Ok("STRING Lachnospira multipara LB2003");
        }

        if is_valid_string_mitsuokella_jalaludinii_node_name(node_name) {
            return Ok("STRING Mitsuokella jalaludinii");
        }

        if is_valid_string_prevotella_brevis_node_name(node_name) {
            return Ok("STRING Prevotella brevis");
        }

        if is_valid_string_proteiniclasticum_ruminis_node_name(node_name) {
            return Ok("STRING Proteiniclasticum ruminis");
        }

        if is_valid_string_ruminococcus_flavefaciens_ma2007_node_name(node_name) {
            return Ok("STRING Ruminococcus flavefaciens MA2007");
        }

        if is_valid_string_sharpea_azabuensis_node_name(node_name) {
            return Ok("STRING Sharpea azabuensis");
        }

        if is_valid_string_succinivibrio_dextrinosolvens_node_name(node_name) {
            return Ok("STRING Succinivibrio dextrinosolvens");
        }

        if is_valid_string_rhizobiales_bacterium_yim77505_node_name(node_name) {
            return Ok("STRING Rhizobiales bacterium YIM77505");
        }

        if is_valid_string_gammaproteobacteria_bacterium_mola455_node_name(node_name) {
            return Ok("STRING Gammaproteobacteria bacterium MOLA455");
        }

        if is_valid_string_corynebacterium_sp_jcb_node_name(node_name) {
            return Ok("STRING Corynebacterium sp. JCB");
        }

        if is_valid_string_clostridium_sp_jcc_node_name(node_name) {
            return Ok("STRING Clostridium sp. JCC");
        }

        if is_valid_string_nocardia_nova_node_name(node_name) {
            return Ok("STRING Nocardia nova");
        }

        if is_valid_string_pseudomonas_sp_tkp_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. TKP");
        }

        if is_valid_string_marinobacter_sp_mctg268_node_name(node_name) {
            return Ok("STRING Marinobacter sp. MCTG268");
        }

        if is_valid_string_halomonas_sp_tg39a_node_name(node_name) {
            return Ok("STRING Halomonas sp. TG39a");
        }

        if is_valid_string_oceanicola_sp_mctg156_node_name(node_name) {
            return Ok("STRING Oceanicola sp. MCTG156");
        }

        if is_valid_string_clostridium_botulinum_202f_node_name(node_name) {
            return Ok("STRING Clostridium botulinum 202F");
        }

        if is_valid_string_clostridium_baratii_node_name(node_name) {
            return Ok("STRING Clostridium baratii");
        }

        if is_valid_string_porticoccus_hydrocarbonoclasticus_node_name(node_name) {
            return Ok("STRING Porticoccus hydrocarbonoclasticus");
        }

        if is_valid_string_polycyclovorans_algicola_node_name(node_name) {
            return Ok("STRING Polycyclovorans algicola");
        }

        if is_valid_string_algiphilus_aromaticivorans_node_name(node_name) {
            return Ok("STRING Algiphilus aromaticivorans");
        }

        if is_valid_string_microbacterium_paraoxydans_dh1b_node_name(node_name) {
            return Ok("STRING Microbacterium paraoxydans DH1b");
        }

        if is_valid_string_leifsonia_aquatica_node_name(node_name) {
            return Ok("STRING Leifsonia aquatica");
        }

        if is_valid_string_myroides_odoratimimus_node_name(node_name) {
            return Ok("STRING Myroides odoratimimus");
        }

        if is_valid_string_borrelia_persica_node_name(node_name) {
            return Ok("STRING Borrelia persica");
        }

        if is_valid_string_defluviimonas_sp_20v17_node_name(node_name) {
            return Ok("STRING Defluviimonas sp. 20V17");
        }

        if is_valid_string_pseudomonas_mandelii_pd30_node_name(node_name) {
            return Ok("STRING Pseudomonas mandelii PD30");
        }

        if is_valid_string_streptococcus_sp_vt_162_node_name(node_name) {
            return Ok("STRING Streptococcus sp. VT 162");
        }

        if is_valid_string_phaeobacter_gallaeciensis_node_name(node_name) {
            return Ok("STRING Phaeobacter gallaeciensis");
        }

        if is_valid_string_bacillus_sp_sjs_node_name(node_name) {
            return Ok("STRING Bacillus sp. SJS");
        }

        if is_valid_string_lactobacillus_apodemi_node_name(node_name) {
            return Ok("STRING Lactobacillus apodemi");
        }

        if is_valid_string_lactobacillus_casei_dsm20011_node_name(node_name) {
            return Ok("STRING Lactobacillus casei DSM20011");
        }

        if is_valid_string_lactobacillus_composti_node_name(node_name) {
            return Ok("STRING Lactobacillus composti");
        }

        if is_valid_string_lactobacillus_farraginis_node_name(node_name) {
            return Ok("STRING Lactobacillus farraginis");
        }

        if is_valid_string_lactobacillus_fuchuensis_node_name(node_name) {
            return Ok("STRING Lactobacillus fuchuensis");
        }

        if is_valid_string_lactobacillus_gallinarum_node_name(node_name) {
            return Ok("STRING Lactobacillus gallinarum");
        }

        if is_valid_string_lactobacillus_hamsteri_node_name(node_name) {
            return Ok("STRING Lactobacillus hamsteri");
        }

        if is_valid_string_lactobacillus_hayakitensis_node_name(node_name) {
            return Ok("STRING Lactobacillus hayakitensis");
        }

        if is_valid_string_lactobacillus_hominis_node_name(node_name) {
            return Ok("STRING Lactobacillus hominis");
        }

        if is_valid_string_lactobacillus_kitasatonis_node_name(node_name) {
            return Ok("STRING Lactobacillus kitasatonis");
        }

        if is_valid_string_lactobacillus_nodensis_node_name(node_name) {
            return Ok("STRING Lactobacillus nodensis");
        }

        if is_valid_string_lactobacillus_otakiensis_node_name(node_name) {
            return Ok("STRING Lactobacillus otakiensis");
        }

        if is_valid_string_lactobacillus_pasteurii_node_name(node_name) {
            return Ok("STRING Lactobacillus pasteurii");
        }

        if is_valid_string_lactobacillus_sucicola_node_name(node_name) {
            return Ok("STRING Lactobacillus sucicola");
        }

        if is_valid_string_lactobacillus_suebicus_node_name(node_name) {
            return Ok("STRING Lactobacillus suebicus");
        }

        if is_valid_string_lactobacillus_vaginalis_node_name(node_name) {
            return Ok("STRING Lactobacillus vaginalis");
        }

        if is_valid_string_lactobacillus_versmoldensis_node_name(node_name) {
            return Ok("STRING Lactobacillus versmoldensis");
        }

        if is_valid_string_lactobacillus_zeae_node_name(node_name) {
            return Ok("STRING Lactobacillus zeae");
        }

        if is_valid_string_advenella_kashmirensis_w13003_node_name(node_name) {
            return Ok("STRING Advenella kashmirensis W13003");
        }

        if is_valid_string_hepatoplasma_crinochetorum_node_name(node_name) {
            return Ok("STRING Hepatoplasma crinochetorum");
        }

        if is_valid_string_rhodococcus_rhodochrous_node_name(node_name) {
            return Ok("STRING Rhodococcus rhodochrous");
        }

        if is_valid_string_stenotrophomonas_maltophilia_5bai2_node_name(node_name) {
            return Ok("STRING Stenotrophomonas maltophilia 5BAI2");
        }

        if is_valid_string_afipia_sp_p5210_node_name(node_name) {
            return Ok("STRING Afipia sp. P5210");
        }

        if is_valid_string_geobacillus_sp_g1w1_node_name(node_name) {
            return Ok("STRING Geobacillus sp. G1w1");
        }

        if is_valid_string_magnetospirillum_gryphiswaldense_node_name(node_name) {
            return Ok("STRING Magnetospirillum gryphiswaldense");
        }

        if is_valid_string_rhizobium_etli_mimosae_node_name(node_name) {
            return Ok("STRING Rhizobium etli mimosae");
        }

        if is_valid_string_komagataeibacter_rhaeticus_node_name(node_name) {
            return Ok("STRING Komagataeibacter rhaeticus");
        }

        if is_valid_string_mannheimia_varigena_usmarc1261_node_name(node_name) {
            return Ok("STRING Mannheimia varigena USMARC1261");
        }

        if is_valid_string_rikenellaceae_bacterium_m3_node_name(node_name) {
            return Ok("STRING Rikenellaceae bacterium M3");
        }

        if is_valid_string_mannheimia_varigena_usmarc1296_node_name(node_name) {
            return Ok("STRING Mannheimia varigena USMARC1296");
        }

        if is_valid_string_dyadobacter_tibetensis_node_name(node_name) {
            return Ok("STRING Dyadobacter tibetensis");
        }

        if is_valid_string_burkholderia_pseudomallei_abcpw111_node_name(node_name) {
            return Ok("STRING Burkholderia pseudomallei ABCPW111");
        }

        if is_valid_string_bifidobacterium_moukalabense_node_name(node_name) {
            return Ok("STRING Bifidobacterium moukalabense");
        }

        if is_valid_string_rhodococcus_pyridinivorans_node_name(node_name) {
            return Ok("STRING Rhodococcus pyridinivorans");
        }

        if is_valid_string_criblamydia_sequanensis_node_name(node_name) {
            return Ok("STRING Criblamydia sequanensis");
        }

        if is_valid_string_ochrobactrum_rhizosphaerae_node_name(node_name) {
            return Ok("STRING Ochrobactrum rhizosphaerae");
        }

        if is_valid_string_bifidobacterium_pullorum_node_name(node_name) {
            return Ok("STRING Bifidobacterium pullorum");
        }

        if is_valid_string_bifidobacterium_mongoliense_node_name(node_name) {
            return Ok("STRING Bifidobacterium mongoliense");
        }

        if is_valid_string_bifidobacterium_actinocoloniiforme_node_name(node_name) {
            return Ok("STRING Bifidobacterium actinocoloniiforme");
        }

        if is_valid_string_bifidobacterium_bohemicum_node_name(node_name) {
            return Ok("STRING Bifidobacterium bohemicum");
        }

        if is_valid_string_bifidobacterium_biavatii_node_name(node_name) {
            return Ok("STRING Bifidobacterium biavatii");
        }

        if is_valid_string_bifidobacterium_callitrichos_node_name(node_name) {
            return Ok("STRING Bifidobacterium callitrichos");
        }

        if is_valid_string_bifidobacterium_reuteri_node_name(node_name) {
            return Ok("STRING Bifidobacterium reuteri");
        }

        if is_valid_string_castellaniella_defragrans_node_name(node_name) {
            return Ok("STRING Castellaniella defragrans");
        }

        if is_valid_string_pseudomonas_nitroreducens_hbp1_node_name(node_name) {
            return Ok("STRING Pseudomonas nitroreducens HBP1");
        }

        if is_valid_string_pseudomonas_sp_bay1663_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. BAY1663");
        }

        if is_valid_string_escherichia_albertii_kf1_node_name(node_name) {
            return Ok("STRING Escherichia albertii KF1");
        }

        if is_valid_string_streptomyces_scopuliridis_node_name(node_name) {
            return Ok("STRING Streptomyces scopuliridis");
        }

        if is_valid_string_mycobacterium_aromaticivorans_node_name(node_name) {
            return Ok("STRING Mycobacterium aromaticivorans");
        }

        if is_valid_string_pseudomonas_cichorii_node_name(node_name) {
            return Ok("STRING Pseudomonas cichorii");
        }

        if is_valid_string_serratia_fonticola_rb25_node_name(node_name) {
            return Ok("STRING Serratia fonticola RB25");
        }

        if is_valid_string_arcobacter_cibarius_node_name(node_name) {
            return Ok("STRING Arcobacter cibarius");
        }

        if is_valid_string_luteimonas_huabeiensis_node_name(node_name) {
            return Ok("STRING Luteimonas huabeiensis");
        }

        if is_valid_string_sulfitobacter_guttiformis_node_name(node_name) {
            return Ok("STRING Sulfitobacter guttiformis");
        }

        if is_valid_string_yersinia_enterocolitica_lc20_node_name(node_name) {
            return Ok("STRING Yersinia enterocolitica LC20");
        }

        if is_valid_string_clostridium_novyi_b_nctc9691_node_name(node_name) {
            return Ok("STRING Clostridium novyi B NCTC9691");
        }

        if is_valid_string_clostridium_botulinum_cd_bkt12695_node_name(node_name) {
            return Ok("STRING Clostridium botulinum CD BKT12695");
        }

        if is_valid_string_aquimarina_macrocephali_node_name(node_name) {
            return Ok("STRING Aquimarina macrocephali");
        }

        if is_valid_string_sporolactobacillus_terrae_node_name(node_name) {
            return Ok("STRING Sporolactobacillus terrae");
        }

        if is_valid_string_brevibacillus_borstelensis_node_name(node_name) {
            return Ok("STRING Brevibacillus borstelensis");
        }

        if is_valid_string_bacillus_flexus_node_name(node_name) {
            return Ok("STRING Bacillus flexus");
        }

        if is_valid_string_chlamydia_sp_diamant_node_name(node_name) {
            return Ok("STRING Chlamydia sp. Diamant");
        }

        if is_valid_string_chlamydia_sp_rubis_node_name(node_name) {
            return Ok("STRING Chlamydia sp. Rubis");
        }

        if is_valid_string_xylella_fastidiosa_pls229_node_name(node_name) {
            return Ok("STRING Xylella fastidiosa PLS229");
        }

        if is_valid_string_sciscionella_sp_se31_node_name(node_name) {
            return Ok("STRING Sciscionella sp. SE31");
        }

        if is_valid_string_paracoccus_yeei_node_name(node_name) {
            return Ok("STRING Paracoccus yeei");
        }

        if is_valid_string_aeromonas_hydrophila_yl17_node_name(node_name) {
            return Ok("STRING Aeromonas hydrophila YL17");
        }

        if is_valid_string_tomitella_biformata_node_name(node_name) {
            return Ok("STRING Tomitella biformata");
        }

        if is_valid_string_campylobacter_sp_cit045_node_name(node_name) {
            return Ok("STRING Campylobacter sp. CIT045");
        }

        if is_valid_string_halobellus_rufus_node_name(node_name) {
            return Ok("STRING Halobellus rufus");
        }

        if is_valid_string_arthrobacter_sp_unc362mftsu51_node_name(node_name) {
            return Ok("STRING Arthrobacter sp. UNC362MFTsu51");
        }

        if is_valid_string_mycobacterium_sp_unc280mftsu51_node_name(node_name) {
            return Ok("STRING Mycobacterium sp. UNC280MFTsu51");
        }

        if is_valid_string_caulobacter_sp_unc358mftsu51_node_name(node_name) {
            return Ok("STRING Caulobacter sp. UNC358MFTsu51");
        }

        if is_valid_string_clostridium_sp_knhs205_node_name(node_name) {
            return Ok("STRING Clostridium sp. KNHs205");
        }

        if is_valid_string_microbacterium_sp_unccl10_node_name(node_name) {
            return Ok("STRING Microbacterium sp. UNCCL10");
        }

        if is_valid_string_paenibacillus_sp_unc451mf_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. UNC451MF");
        }

        if is_valid_string_phyllobacterium_sp_unc302mfcol52_node_name(node_name) {
            return Ok("STRING Phyllobacterium sp. UNC302MFCol52");
        }

        if is_valid_string_rhodococcus_sp_unc23mfcrub11_node_name(node_name) {
            return Ok("STRING Rhodococcus sp. UNC23MFCrub11");
        }

        if is_valid_string_rhodococcus_sp_unc363mftsu51_node_name(node_name) {
            return Ok("STRING Rhodococcus sp. UNC363MFTsu51");
        }

        if is_valid_string_sphingomonas_sp_unc305mfcol52_node_name(node_name) {
            return Ok("STRING Sphingomonas sp. UNC305MFCol52");
        }

        if is_valid_string_thermus_sp_yim77409_node_name(node_name) {
            return Ok("STRING Thermus sp. YIM77409");
        }

        if is_valid_string_clostridiales_bacterium_dri13_node_name(node_name) {
            return Ok("STRING Clostridiales bacterium DRI13");
        }

        if is_valid_string_carnobacterium_alterfunditum_node_name(node_name) {
            return Ok("STRING Carnobacterium alterfunditum");
        }

        if is_valid_string_carnobacterium_divergens_node_name(node_name) {
            return Ok("STRING Carnobacterium divergens");
        }

        if is_valid_string_carnobacterium_funditum_node_name(node_name) {
            return Ok("STRING Carnobacterium funditum");
        }

        if is_valid_string_carnobacterium_gallinarum_node_name(node_name) {
            return Ok("STRING Carnobacterium gallinarum");
        }

        if is_valid_string_carnobacterium_mobile_node_name(node_name) {
            return Ok("STRING Carnobacterium mobile");
        }

        if is_valid_string_carnobacterium_pleistocenium_node_name(node_name) {
            return Ok("STRING Carnobacterium pleistocenium");
        }

        if is_valid_string_helicobacter_rodentium_node_name(node_name) {
            return Ok("STRING Helicobacter rodentium");
        }

        if is_valid_string_kitasatospora_azatica_node_name(node_name) {
            return Ok("STRING Kitasatospora azatica");
        }

        if is_valid_string_kitasatospora_mediocidica_node_name(node_name) {
            return Ok("STRING Kitasatospora mediocidica");
        }

        if is_valid_string_roseivivax_halodurans_node_name(node_name) {
            return Ok("STRING Roseivivax halodurans");
        }

        if is_valid_string_roseivivax_isoporae_node_name(node_name) {
            return Ok("STRING Roseivivax isoporae");
        }

        if is_valid_string_streptacidiphilus_oryzae_node_name(node_name) {
            return Ok("STRING Streptacidiphilus oryzae");
        }

        if is_valid_string_streptomyces_yeochonensis_node_name(node_name) {
            return Ok("STRING Streptomyces yeochonensis");
        }

        if is_valid_string_thermus_tengchongensis_node_name(node_name) {
            return Ok("STRING Thermus tengchongensis");
        }

        if is_valid_string_kutzneria_albida_node_name(node_name) {
            return Ok("STRING Kutzneria albida");
        }

        if is_valid_string_flavobacterium_succinicans_node_name(node_name) {
            return Ok("STRING Flavobacterium succinicans");
        }

        if is_valid_string_bacillus_sp_ts2_node_name(node_name) {
            return Ok("STRING Bacillus sp. TS2");
        }

        if is_valid_string_corynebacterium_falsenii_node_name(node_name) {
            return Ok("STRING Corynebacterium falsenii");
        }

        if is_valid_string_microbacterium_sp_mrs1_node_name(node_name) {
            return Ok("STRING Microbacterium sp. MRS1");
        }

        if is_valid_string_microbacterium_sp_crk32_node_name(node_name) {
            return Ok("STRING Microbacterium sp. CrK32");
        }

        if is_valid_string_microbacterium_sp_crk20_node_name(node_name) {
            return Ok("STRING Microbacterium sp. CrK20");
        }

        if is_valid_string_pseudomonas_sp_rl_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. RL");
        }

        if is_valid_string_hafnia_alvei_node_name(node_name) {
            return Ok("STRING Hafnia alvei");
        }

        if is_valid_string_flavobacterium_aquatile_node_name(node_name) {
            return Ok("STRING Flavobacterium aquatile");
        }

        if is_valid_string_schleiferia_thermophila_node_name(node_name) {
            return Ok("STRING Schleiferia thermophila");
        }

        if is_valid_string_microbulbifer_sp_hz11_node_name(node_name) {
            return Ok("STRING Microbulbifer sp. HZ11");
        }

        if is_valid_string_pseudomonas_pseudoalcaligenes_ad6_node_name(node_name) {
            return Ok("STRING Pseudomonas pseudoalcaligenes AD6");
        }

        if is_valid_string_flavobacterium_chungangense_node_name(node_name) {
            return Ok("STRING Flavobacterium chungangense");
        }

        if is_valid_string_accumulibacter_sp_ba93_node_name(node_name) {
            return Ok("STRING Accumulibacter sp. BA93");
        }

        if is_valid_string_pedobacter_borealis_node_name(node_name) {
            return Ok("STRING Pedobacter borealis");
        }

        if is_valid_string_cellulomonas_sp_hzm_node_name(node_name) {
            return Ok("STRING Cellulomonas sp. HZM");
        }

        if is_valid_string_photobacterium_phosphoreum_node_name(node_name) {
            return Ok("STRING Photobacterium phosphoreum");
        }

        if is_valid_string_haladaptatus_cibarius_node_name(node_name) {
            return Ok("STRING Haladaptatus cibarius");
        }

        if is_valid_string_halapricum_salinum_node_name(node_name) {
            return Ok("STRING Halapricum salinum");
        }

        if is_valid_string_acetobacter_aceti_1023_node_name(node_name) {
            return Ok("STRING Acetobacter aceti 1023");
        }

        if is_valid_string_hylemonella_gracilis_niagara_node_name(node_name) {
            return Ok("STRING Hylemonella gracilis Niagara");
        }

        if is_valid_string_burkholderia_sp_mp1_node_name(node_name) {
            return Ok("STRING Burkholderia sp. MP1");
        }

        if is_valid_string_comamonadaceae_bacterium_h1_node_name(node_name) {
            return Ok("STRING Comamonadaceae bacterium H1");
        }

        if is_valid_string_lachnospiraceae_bacterium_ac2029_node_name(node_name) {
            return Ok("STRING Lachnospiraceae bacterium AC2029");
        }

        if is_valid_string_nitrososphaera_evergladensis_node_name(node_name) {
            return Ok("STRING Nitrososphaera evergladensis");
        }

        if is_valid_string_geomicrobium_sp_jcm19037_node_name(node_name) {
            return Ok("STRING Geomicrobium sp. JCM19037");
        }

        if is_valid_string_geomicrobium_sp_jcm19038_node_name(node_name) {
            return Ok("STRING Geomicrobium sp. JCM19038");
        }

        if is_valid_string_bacillus_sp_jcm19046_node_name(node_name) {
            return Ok("STRING Bacillus sp. JCM19046");
        }

        if is_valid_string_weeksella_sp_ff8_node_name(node_name) {
            return Ok("STRING Weeksella sp. FF8");
        }

        if is_valid_string_haemophilus_sp_ff7_node_name(node_name) {
            return Ok("STRING Haemophilus sp. FF7");
        }

        if is_valid_string_bacillus_sp_jce_node_name(node_name) {
            return Ok("STRING Bacillus sp. JCE");
        }

        if is_valid_string_jeotgalicoccus_sp_13mg44air_node_name(node_name) {
            return Ok("STRING Jeotgalicoccus sp. 13MG44air");
        }

        if is_valid_string_actibacterium_atlanticum_node_name(node_name) {
            return Ok("STRING Actibacterium atlanticum");
        }

        if is_valid_string_oceanicola_sp_22iis11g_node_name(node_name) {
            return Ok("STRING Oceanicola sp. 22IIS11g");
        }

        if is_valid_string_virgibacillus_sp_vm5_node_name(node_name) {
            return Ok("STRING Virgibacillus sp. Vm5");
        }

        if is_valid_string_oceanobacillus_sp_s5_node_name(node_name) {
            return Ok("STRING Oceanobacillus sp. S5");
        }

        if is_valid_string_actinosporangium_sp_nrrlb3428_node_name(node_name) {
            return Ok("STRING Actinosporangium sp. NRRLB3428");
        }

        if is_valid_string_glycomyces_sp_nrrlb16210_node_name(node_name) {
            return Ok("STRING Glycomyces sp. NRRLB16210");
        }

        if is_valid_string_saccharothrix_sp_nrrlb16314_node_name(node_name) {
            return Ok("STRING Saccharothrix sp. NRRLB16314");
        }

        if is_valid_string_streptomyces_sp_nrrlf2580_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLF2580");
        }

        if is_valid_string_streptomyces_sp_nrrlf2890_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLF2890");
        }

        if is_valid_string_streptomyces_sp_nrrlf5008_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLF5008");
        }

        if is_valid_string_streptomyces_sp_nrrlf5053_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLF5053");
        }

        if is_valid_string_streptomyces_sp_nrrlf5065_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLF5065");
        }

        if is_valid_string_streptomyces_sp_nrrlf5123_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLF5123");
        }

        if is_valid_string_streptomyces_sp_nrrlf5126_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLF5126");
        }

        if is_valid_string_streptomyces_sp_nrrlf5135_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLF5135");
        }

        if is_valid_string_streptomyces_sp_nrrlf525_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLF525");
        }

        if is_valid_string_streptomyces_sp_nrrlf5630_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLF5630");
        }

        if is_valid_string_streptomyces_sp_nrrlf6677_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLF6677");
        }

        if is_valid_string_streptomyces_sp_nrrls118_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLS118");
        }

        if is_valid_string_streptomyces_sp_nrrls149_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLS149");
        }

        if is_valid_string_streptomyces_sp_nrrls1777_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLS1777");
        }

        if is_valid_string_streptomyces_sp_nrrls237_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLS237");
        }

        if is_valid_string_streptomyces_sp_nrrls337_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLS337");
        }

        if is_valid_string_streptomyces_sp_nrrls340_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLS340");
        }

        if is_valid_string_streptomyces_sp_nrrls37_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLS37");
        }

        if is_valid_string_streptomyces_sp_nrrls474_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLS474");
        }

        if is_valid_string_streptomyces_sp_nrrls646_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLS646");
        }

        if is_valid_string_streptomyces_sp_nrrls87_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLS87");
        }

        if is_valid_string_streptomyces_sp_nrrls920_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLS920");
        }

        if is_valid_string_streptomyces_sp_nrrlwc3626_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLWC3626");
        }

        if is_valid_string_streptomyces_sp_nrrlwc3742_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLWC3742");
        }

        if is_valid_string_streptomyces_sp_nrrlwc3773_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLWC3773");
        }

        if is_valid_string_micromonospora_parva_node_name(node_name) {
            return Ok("STRING Micromonospora parva");
        }

        if is_valid_string_thioalkalivibrio_sp_hk1_node_name(node_name) {
            return Ok("STRING Thioalkalivibrio sp. HK1");
        }

        if is_valid_string_lacinutrix_sp_pamc27137_node_name(node_name) {
            return Ok("STRING Lacinutrix sp. PAMC27137");
        }

        if is_valid_string_scytonema_hofmanni_node_name(node_name) {
            return Ok("STRING Scytonema hofmanni");
        }

        if is_valid_string_rhodovulum_sp_ni22_node_name(node_name) {
            return Ok("STRING Rhodovulum sp. NI22");
        }

        if is_valid_string_clostridium_sp_knhs209_node_name(node_name) {
            return Ok("STRING Clostridium sp. KNHs209");
        }

        if is_valid_string_sphingomonas_sp_rit328_node_name(node_name) {
            return Ok("STRING Sphingomonas sp. RIT328");
        }

        if is_valid_string_pseudomonas_sp_rit357_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. RIT357");
        }

        if is_valid_string_prochlorococcus_sp_scb241528o2_node_name(node_name) {
            return Ok("STRING Prochlorococcus sp. scB241528O2");
        }

        if is_valid_string_prochlorococcus_sp_scb245a520k10_node_name(node_name) {
            return Ok("STRING Prochlorococcus sp. scB245a520K10");
        }

        if is_valid_string_falsirhodobacter_sp_alg1_node_name(node_name) {
            return Ok("STRING Falsirhodobacter sp. alg1");
        }

        if is_valid_string_burkholderia_sp_k24_node_name(node_name) {
            return Ok("STRING Burkholderia sp. K24");
        }

        if is_valid_string_lysinibacillus_sp_bf4_node_name(node_name) {
            return Ok("STRING Lysinibacillus sp. BF4");
        }

        if is_valid_string_deinococcus_phoenicis_node_name(node_name) {
            return Ok("STRING Deinococcus phoenicis");
        }

        if is_valid_string_streptomyces_sp_nrrlb24720_node_name(node_name) {
            return Ok("STRING Streptomyces sp. NRRLB24720");
        }

        if is_valid_string_peptostreptococcaceae_bacterium_va2_node_name(node_name) {
            return Ok("STRING Peptostreptococcaceae bacterium VA2");
        }

        if is_valid_string_halomonas_sp_hl48_node_name(node_name) {
            return Ok("STRING Halomonas sp. HL48");
        }

        if is_valid_string_marinobacter_sp_hl58_node_name(node_name) {
            return Ok("STRING Marinobacter sp. HL58");
        }

        if is_valid_string_oceanicaulis_sp_hl87_node_name(node_name) {
            return Ok("STRING Oceanicaulis sp. HL87");
        }

        if is_valid_string_porphyrobacter_sp_hl46_node_name(node_name) {
            return Ok("STRING Porphyrobacter sp. HL46");
        }

        if is_valid_string_curtobacterium_sp_s6_node_name(node_name) {
            return Ok("STRING Curtobacterium sp. S6");
        }

        if is_valid_string_spirochaeta_sp_jc230_node_name(node_name) {
            return Ok("STRING Spirochaeta sp. JC230");
        }

        if is_valid_string_pantoea_sp_psnih2_node_name(node_name) {
            return Ok("STRING Pantoea sp. PSNIH2");
        }

        if is_valid_string_pantoea_sp_psnih1_node_name(node_name) {
            return Ok("STRING Pantoea sp. PSNIH1");
        }

        if is_valid_string_psychroserpens_sp_pamc27130_node_name(node_name) {
            return Ok("STRING Psychroserpens sp. PAMC27130");
        }

        if is_valid_string_exiguobacterium_sp_ab2_node_name(node_name) {
            return Ok("STRING Exiguobacterium sp. AB2");
        }

        if is_valid_string_selenomonas_sp_ae3005_node_name(node_name) {
            return Ok("STRING Selenomonas sp. AE3005");
        }

        if is_valid_string_ferriphaselus_sp_r1_node_name(node_name) {
            return Ok("STRING Ferriphaselus sp. R1");
        }

        if is_valid_string_zetaproteobacteria_bacterium_tag1_node_name(node_name) {
            return Ok("STRING Zetaproteobacteria bacterium TAG1");
        }

        if is_valid_string_clostridium_sp_hmp27_node_name(node_name) {
            return Ok("STRING Clostridium sp. HMP27");
        }

        if is_valid_string_desulfosporosinus_sp_hmp52_node_name(node_name) {
            return Ok("STRING Desulfosporosinus sp. HMP52");
        }

        if is_valid_string_leptolyngbya_sp_jsc1_node_name(node_name) {
            return Ok("STRING Leptolyngbya sp. JSC1");
        }

        if is_valid_string_corynebacterium_sp_atcc6931_node_name(node_name) {
            return Ok("STRING Corynebacterium sp. ATCC6931");
        }

        if is_valid_string_pseudomonas_sp_ant303_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. Ant303");
        }

        if is_valid_string_deinococcus_sp_rl_node_name(node_name) {
            return Ok("STRING Deinococcus sp. RL");
        }

        if is_valid_string_flavobacterium_sp_em1308_node_name(node_name) {
            return Ok("STRING Flavobacterium sp. EM1308");
        }

        if is_valid_string_flavobacterium_sp_em1321_node_name(node_name) {
            return Ok("STRING Flavobacterium sp. EM1321");
        }

        if is_valid_string_gammaproteobacteria_bacterium_mfb021_node_name(node_name) {
            return Ok("STRING Gammaproteobacteria bacterium MFB021");
        }

        if is_valid_string_cyanobium_sp_caciam14_node_name(node_name) {
            return Ok("STRING Cyanobium sp. CACIAM14");
        }

        if is_valid_string_listeriaceae_bacterium_fsla50209_node_name(node_name) {
            return Ok("STRING Listeriaceae bacterium FSLA50209");
        }

        if is_valid_string_prochlorococcus_sp_mit0601_node_name(node_name) {
            return Ok("STRING Prochlorococcus sp. MIT0601");
        }

        if is_valid_string_prochlorococcus_sp_mit0602_node_name(node_name) {
            return Ok("STRING Prochlorococcus sp. MIT0602");
        }

        if is_valid_string_prochlorococcus_sp_mit0701_node_name(node_name) {
            return Ok("STRING Prochlorococcus sp. MIT0701");
        }

        if is_valid_string_bacillus_sp_mt2_node_name(node_name) {
            return Ok("STRING Bacillus sp. MT2");
        }

        if is_valid_string_clostridium_sp_cl6_node_name(node_name) {
            return Ok("STRING Clostridium sp. CL6");
        }

        if is_valid_string_clostridium_sp_cl2_node_name(node_name) {
            return Ok("STRING Clostridium sp. CL2");
        }

        if is_valid_string_bacillus_sp_kw12_node_name(node_name) {
            return Ok("STRING Bacillus sp. KW12");
        }

        if is_valid_string_pseudomonas_sp_20bn_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. 20BN");
        }

        if is_valid_string_clostridium_sp_lf2_node_name(node_name) {
            return Ok("STRING Clostridium sp. LF2");
        }

        if is_valid_string_bacterium_uasb270_node_name(node_name) {
            return Ok("STRING Bacterium UASB270");
        }

        if is_valid_string_paenibacillus_sp_tca20_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. TCA20");
        }

        if is_valid_string_rhizobium_sp_yr295_node_name(node_name) {
            return Ok("STRING Rhizobium sp. YR295");
        }

        if is_valid_string_rhizobium_sp_yr519_node_name(node_name) {
            return Ok("STRING Rhizobium sp. YR519");
        }

        if is_valid_string_chryseobacterium_sp_cf284_node_name(node_name) {
            return Ok("STRING Chryseobacterium sp. CF284");
        }

        if is_valid_string_rhizobium_sp_cf097_node_name(node_name) {
            return Ok("STRING Rhizobium sp. CF097");
        }

        if is_valid_string_rhizobium_sp_cf394_node_name(node_name) {
            return Ok("STRING Rhizobium sp. CF394");
        }

        if is_valid_string_rhizobium_sp_ok494_node_name(node_name) {
            return Ok("STRING Rhizobium sp. OK494");
        }

        if is_valid_string_luteibacter_sp_9143_node_name(node_name) {
            return Ok("STRING Luteibacter sp. 9143");
        }

        if is_valid_string_luteibacter_sp_9135_node_name(node_name) {
            return Ok("STRING Luteibacter sp. 9135");
        }

        if is_valid_string_massilia_sp_9096_node_name(node_name) {
            return Ok("STRING Massilia sp. 9096");
        }

        if is_valid_string_burkholderia_sp_9120_node_name(node_name) {
            return Ok("STRING Burkholderia sp. 9120");
        }

        if is_valid_string_paenibacillus_sp_mst1_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. MSt1");
        }

        if is_valid_string_prochlorococcus_sp_mit0604_node_name(node_name) {
            return Ok("STRING Prochlorococcus sp. MIT0604");
        }

        if is_valid_string_prochlorococcus_sp_mit0801_node_name(node_name) {
            return Ok("STRING Prochlorococcus sp. MIT0801");
        }

        if is_valid_string_alistipes_sp_627_node_name(node_name) {
            return Ok("STRING Alistipes sp. 627");
        }

        if is_valid_string_devosia_sp_lc5_node_name(node_name) {
            return Ok("STRING Devosia sp. LC5");
        }

        if is_valid_string_methylotenera_sp_l2l1_node_name(node_name) {
            return Ok("STRING Methylotenera sp. L2L1");
        }

        if is_valid_string_sphingopyxis_sp_lc81_node_name(node_name) {
            return Ok("STRING Sphingopyxis sp. LC81");
        }

        if is_valid_string_bosea_sp_lc85_node_name(node_name) {
            return Ok("STRING Bosea sp. LC85");
        }

        if is_valid_string_massilia_sp_lc238_node_name(node_name) {
            return Ok("STRING Massilia sp. LC238");
        }

        if is_valid_string_actinobacterium_acamd5_node_name(node_name) {
            return Ok("STRING actinobacterium acAMD5");
        }

        if is_valid_string_polaromonas_sp_cg912_node_name(node_name) {
            return Ok("STRING Polaromonas sp. CG912");
        }

        if is_valid_string_bacterium_sp_ol1_node_name(node_name) {
            return Ok("STRING Bacterium sp. OL1");
        }

        if is_valid_string_bacterium_sp_lf3_node_name(node_name) {
            return Ok("STRING Bacterium sp. LF3");
        }

        if is_valid_string_halomonas_sp_ko116_node_name(node_name) {
            return Ok("STRING Halomonas sp. KO116");
        }

        if is_valid_string_flavobacterium_sp_fl_node_name(node_name) {
            return Ok("STRING Flavobacterium sp. Fl");
        }

        if is_valid_string_butyrivibrio_sp_ae3004_node_name(node_name) {
            return Ok("STRING Butyrivibrio sp. AE3004");
        }

        if is_valid_string_arthromitus_sp_sfbmousenl_node_name(node_name) {
            return Ok("STRING Arthromitus sp. SFBmouseNL");
        }

        if is_valid_string_acinetobacter_sp_hr7_node_name(node_name) {
            return Ok("STRING Acinetobacter sp. HR7");
        }

        if is_valid_string_rhizobium_sp_r1200b2_node_name(node_name) {
            return Ok("STRING Rhizobium sp. R1200B2");
        }

        if is_valid_string_bosea_sp_unc402clcol_node_name(node_name) {
            return Ok("STRING Bosea sp. UNC402CLCol");
        }

        if is_valid_string_ruminococcus_sp_hun007_node_name(node_name) {
            return Ok("STRING Ruminococcus sp. HUN007");
        }

        if is_valid_string_porphyromonas_sp_cot239oh1446_node_name(node_name) {
            return Ok("STRING Porphyromonas sp. COT239OH1446");
        }

        if is_valid_string_porphyromonas_sp_cot290oh860_node_name(node_name) {
            return Ok("STRING Porphyromonas sp. COT290OH860");
        }

        if is_valid_string_shewanella_sp_yqh10_node_name(node_name) {
            return Ok("STRING Shewanella sp. YQH10");
        }

        if is_valid_string_idiomarina_sp_mccc1a10513_node_name(node_name) {
            return Ok("STRING Idiomarina sp. MCCC1A10513");
        }

        if is_valid_string_vibrio_sp_er1a_node_name(node_name) {
            return Ok("STRING Vibrio sp. ER1A");
        }

        if is_valid_string_porphyromonadaceae_cot184oh4590_node_name(node_name) {
            return Ok("STRING Porphyromonadaceae COT184OH4590");
        }

        if is_valid_string_oscillibacter_sp_er4_node_name(node_name) {
            return Ok("STRING Oscillibacter sp. ER4");
        }

        if is_valid_string_chlorobium_sp_gbchlb_node_name(node_name) {
            return Ok("STRING Chlorobium sp. GBChlB");
        }

        if is_valid_string_chloroflexus_sp_msg_node_name(node_name) {
            return Ok("STRING Chloroflexus sp. MSG");
        }

        if is_valid_string_sphingobium_sp_ba1_node_name(node_name) {
            return Ok("STRING Sphingobium sp. ba1");
        }

        if is_valid_string_pseudomonas_sp_ml96_node_name(node_name) {
            return Ok("STRING Pseudomonas sp. ML96");
        }

        if is_valid_string_serratia_sp_ag1_node_name(node_name) {
            return Ok("STRING Serratia sp. Ag1");
        }

        if is_valid_string_paracoccus_sp_4681_node_name(node_name) {
            return Ok("STRING Paracoccus sp. 4681");
        }

        if is_valid_string_planococcus_sp_pamc21323_node_name(node_name) {
            return Ok("STRING Planococcus sp. PAMC21323");
        }

        if is_valid_string_atelocyanobacterium_thalassa_sio64986_node_name(node_name) {
            return Ok("STRING Atelocyanobacterium thalassa SIO64986");
        }

        if is_valid_string_rickettsiales_bacterium_ac37b_node_name(node_name) {
            return Ok("STRING Rickettsiales bacterium Ac37b");
        }

        if is_valid_string_thalassospira_australica_node_name(node_name) {
            return Ok("STRING Thalassospira australica");
        }

        if is_valid_string_maribius_sp_mola401_node_name(node_name) {
            return Ok("STRING Maribius sp. MOLA401");
        }

        if is_valid_string_achromobacter_sp_rta_node_name(node_name) {
            return Ok("STRING Achromobacter sp. RTa");
        }

        if is_valid_string_rhizobium_sp_ys1r_node_name(node_name) {
            return Ok("STRING Rhizobium sp. YS1r");
        }

        if is_valid_string_devosia_sp_172e8_node_name(node_name) {
            return Ok("STRING Devosia sp. 172E8");
        }

        if is_valid_string_thalassotalea_sp_nd16a_node_name(node_name) {
            return Ok("STRING Thalassotalea sp. ND16A");
        }

        if is_valid_string_paenibacillus_sp_fslp40081_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. FSLP40081");
        }

        if is_valid_string_paenibacillus_sp_fslr50345_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. FSLR50345");
        }

        if is_valid_string_paenibacillus_sp_fslr70273_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. FSLR70273");
        }

        if is_valid_string_paenibacillus_sp_fslr70331_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. FSLR70331");
        }

        if is_valid_string_paenibacillus_sp_fslh70357_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. FSLH70357");
        }

        if is_valid_string_paenibacillus_sp_fslh70737_node_name(node_name) {
            return Ok("STRING Paenibacillus sp. FSLH70737");
        }

        if is_valid_string_sphingopyxis_sp_mwb1_node_name(node_name) {
            return Ok("STRING Sphingopyxis sp. MWB1");
        }

        if is_valid_string_sulfurospirillum_sp_scadc_node_name(node_name) {
            return Ok("STRING Sulfurospirillum sp. SCADC");
        }

        if is_valid_string_sulfuricurvum_sp_mlsb_node_name(node_name) {
            return Ok("STRING Sulfuricurvum sp. MLSB");
        }

        if is_valid_string_alteromonas_sp_lor_node_name(node_name) {
            return Ok("STRING Alteromonas sp. LOR");
        }

        if is_valid_string_aquabacterium_sp_nj1_node_name(node_name) {
            return Ok("STRING Aquabacterium sp. NJ1");
        }

        if is_valid_string_sphingobacterium_sp_ml3w_node_name(node_name) {
            return Ok("STRING Sphingobacterium sp. ML3W");
        }

        if is_valid_string_treponema_sp_omz838_node_name(node_name) {
            return Ok("STRING Treponema sp. OMZ838");
        }

        if is_valid_string_deinococcus_sp_yim77859_node_name(node_name) {
            return Ok("STRING Deinococcus sp. YIM77859");
        }

        if is_valid_string_clostridium_sp_knhs214_node_name(node_name) {
            return Ok("STRING Clostridium sp. KNHs214");
        }

        if is_valid_string_myxosarcina_sp_gi1_node_name(node_name) {
            return Ok("STRING Myxosarcina sp. GI1");
        }

        if is_valid_string_mollicutes_bacterium_hr1_node_name(node_name) {
            return Ok("STRING Mollicutes bacterium HR1");
        }

        if is_valid_string_mollicutes_bacterium_hr2_node_name(node_name) {
            return Ok("STRING Mollicutes bacterium HR2");
        }

        if is_valid_string_lactobacillus_sp_wkb10_node_name(node_name) {
            return Ok("STRING Lactobacillus sp. wkB10");
        }

        if is_valid_string_lactobacillus_sp_wkb8_node_name(node_name) {
            return Ok("STRING Lactobacillus sp. wkB8");
        }

        if is_valid_string_sphingomonas_sp_3524zxx_node_name(node_name) {
            return Ok("STRING Sphingomonas sp. 3524ZXX");
        }

        if is_valid_string_hoeflea_sp_bal378_node_name(node_name) {
            return Ok("STRING Hoeflea sp. BAL378");
        }

        if is_valid_string_francisella_sp_fsc1006_node_name(node_name) {
            return Ok("STRING Francisella sp. FSC1006");
        }

        if is_valid_string_helicobacter_sp_mit115569_node_name(node_name) {
            return Ok("STRING Helicobacter sp. MIT115569");
        }

        if is_valid_string_campylobacter_sp_mit975078_node_name(node_name) {
            return Ok("STRING Campylobacter sp. MIT975078");
        }

        if is_valid_string_sphingomonas_taxi_node_name(node_name) {
            return Ok("STRING Sphingomonas taxi");
        }

        if is_valid_string_sphingomonas_sp_37zxx_node_name(node_name) {
            return Ok("STRING Sphingomonas sp. 37zxx");
        }

        if is_valid_string_sphingobacteriaceae_bacterium_dw12_node_name(node_name) {
            return Ok("STRING Sphingobacteriaceae bacterium DW12");
        }

        if is_valid_string_listeriaceae_bacterium_fsla50281_node_name(node_name) {
            return Ok("STRING Listeriaceae bacterium FSLA50281");
        }

        if is_valid_string_xanthomonas_sp_nyagatare_node_name(node_name) {
            return Ok("STRING Xanthomonas sp. Nyagatare");
        }

        if is_valid_string_caenorhabditis_tropicalis_node_name(node_name) {
            return Ok("STRING Caenorhabditis tropicalis");
        }

        if is_valid_string_burkholderiaceae_sp_mortierella_node_name(node_name) {
            return Ok("STRING Burkholderiaceae sp. Mortierella");
        }

        if is_valid_string_shewanella_sp_ecsmb14101_node_name(node_name) {
            return Ok("STRING Shewanella sp. ECSMB14101");
        }

        if is_valid_string_sulfurospirillum_sp_mes_node_name(node_name) {
            return Ok("STRING Sulfurospirillum sp. MES");
        }

        if is_valid_string_paracoccus_sp_pamc22219_node_name(node_name) {
            return Ok("STRING Paracoccus sp. PAMC22219");
        }

        if is_valid_string_dickeya_sp_2b12_node_name(node_name) {
            return Ok("STRING Dickeya sp. 2B12");
        }
    }

    Err(format!(
        concat!(
            "There is no known repository with a pattern for the provided node name {:?}.\n",
            "If you believe there should be one, please do open a pull request to ",
            "add it to the library!"
        ),
        node_name
    ))
}

/// Returns name of the graph repository from the given node type name.
///
/// # Implementative details
/// Currently we have support for retrieving the repositories for:
/// * [BioLink](https://biolink.github.io/biolink-model/)
/// * Angular links
///
/// # Arguments
/// * `node_type_name`: &str - Node type name to query for.
///
/// # Raises
/// * If there is no known repository source for the given node type name.
///
pub fn get_node_type_repository_from_node_type_name(node_type_name: &str) -> Result<&str> {
    if is_valid_biolink_from_object(node_type_name) {
        return Ok("BioLink");
    }

    Err(format!(
        concat!(
            "There is no known repository with a pattern for the provided node type name {:?}.\n",
            "If you believe there should be one, please do open a pull request to ",
            "add it to the library!"
        ),
        node_type_name
    ))
}

/// Returns url describing the given node type if a pattern is known.
///
/// # Implementative details
/// Currently we have support for building the URLs for:
/// * [BioLink](https://biolink.github.io/biolink-model/)
///
/// # Arguments
/// * `node_type_name`: &str - Node name to query for.
///
/// # Raises
/// * If there is no known url source for the given node type.
pub fn get_node_type_source_url_from_node_type_name(node_type_name: &str) -> Result<String> {
    if is_valid_biolink_from_object(node_type_name) {
        return Ok(unsafe { format_biolink_from_object(node_type_name) });
    }
    if is_valid_angular_link(node_type_name) {
        return Ok(format_angular_link_url_from_object(node_type_name));
    }
    Err(format!(
        concat!(
            "There is no known url with a pattern for the provided node type {:?}.\n",
            "If you believe there should be one, please do open a pull request to ",
            "add it to the library!"
        ),
        node_type_name
    ))
}

/// Returns name of the graph repository from the given edge type name.
///
/// # Implementative details
/// Currently we have support for retrieving the repositories for:
/// * [BioLink](https://biolink.github.io/biolink-model/)
/// * Angular links
///
/// # Arguments
/// * `edge_type_name`: &str - edge type name to query for.
///
/// # Raises
/// * If there is no known repository source for the given edge type name.
///
pub fn get_edge_type_repository_from_edge_type_name(edge_type_name: &str) -> Result<&str> {
    if is_valid_biolink_from_object(edge_type_name) {
        return Ok("BioLink");
    }

    Err(format!(
        concat!(
            "There is no known repository with a pattern for the provided edge type name {:?}.\n",
            "If you believe there should be one, please do open a pull request to ",
            "add it to the library!"
        ),
        edge_type_name
    ))
}

/// Returns url describing the given edge type if a pattern is known.
///
/// # Implementative details
/// Currently we have support for building the URLs for:
/// * [BioLink](https://biolink.github.io/biolink-model/)
///
/// # Arguments
/// * `edge_type_name`: &str - edge name to query for.
///
/// # Raises
/// * If there is no known url source for the given edge type.
pub fn get_edge_type_source_url_from_edge_type_name(edge_type_name: &str) -> Result<String> {
    if is_valid_biolink_from_object(edge_type_name) {
        return Ok(unsafe { format_biolink_from_object(edge_type_name) });
    }
    if is_valid_angular_link(edge_type_name) {
        return Ok(format_angular_link_url_from_object(edge_type_name));
    }
    Err(format!(
        concat!(
            "There is no known url with a pattern for the provided edge type {:?}.\n",
            "If you believe there should be one, please do open a pull request to ",
            "add it to the library!"
        ),
        edge_type_name
    ))
}

/// Returns html-formatted source of given node name if known.
///
/// # Implementative details
/// If an URL is detected from the provided node name then a standard
/// html URL formatting is returned, otherwise the node name is returned.
/// Refer to the `get_node_source_url_from_node_name` method documentation
/// to see which node databases are supported currently.
///
/// # Arguments
/// * `node_name`: &str - Node name to query for.
pub fn get_node_source_html_url_from_node_name(node_name: &str) -> String {
    match get_node_source_url_from_node_name(node_name) {
        Ok(url) => get_url_formatted(
            url.as_str(),
            node_name,
            get_node_repository_from_node_name(node_name).ok(),
        ),
        Err(_) => escape(node_name).to_string(),
    }
}

/// Returns html-formatted source of given node type name if known.
///
/// # Implementative details
/// If an URL is detected from the provided node type name then a standard
/// html URL formatting is returned, otherwise the node type name is returned.
/// Refer to the `get_node_source_url_from_node_type_name` method documentation
/// to see which node databases are supported currently.
///
/// # Arguments
/// * `node_type_name`: &str - Node name to query for.
pub fn get_node_type_source_html_url_from_node_type_name(node_type_name: &str) -> String {
    match get_node_type_source_url_from_node_type_name(node_type_name) {
        Ok(url) => get_url_formatted(url.as_str(), node_type_name, get_node_type_repository_from_node_type_name(node_type_name).ok()),
        Err(_) => escape(node_type_name).to_string(),
    }
}

/// Returns html-formatted source of given edge type name if known.
///
/// # Implementative details
/// If an URL is detected from the provided edge type name then a standard
/// html URL formatting is returned, otherwise the edge type name is returned.
/// Refer to the `get_edge_source_url_from_edge_type_name` method documentation
/// to see which edge databases are supported currently.
///
/// # Arguments
/// * `edge_type_name`: &str - edge name to query for.
pub fn get_edge_type_source_html_url_from_edge_type_name(edge_type_name: &str) -> String {
    match get_edge_type_source_url_from_edge_type_name(edge_type_name) {
        Ok(url) => get_url_formatted(
            url.as_str(),
            edge_type_name,
            get_edge_type_repository_from_edge_type_name(edge_type_name).ok(),
        ),
        Err(_) => escape(edge_type_name).to_string(),
    }
}
